#! /bin/bash

# This script is used to extract information about all MarkTypes
# used in ODEMx. It uses 'grep' to automatically extract all
# MarkType definitions from the .cpp files stored in subfolders of odemx.

# The extracted lines are stored in a temporary file from which they
# are fed to the function 'processMarkLine()' which uses 'sed' to
# extract the MarkType details like event, name, scope and description.
# This of course requires that the MarkType definitions have a certain
# format:

# const MarkType Condq::markCreate( "create", baseMarkId+1, typeid(Condq) ); ///< Call of Condq constructor

# MarkType details for each line are temporarily stored in global variables
# and then written to the HTML file $DOXYGENFILE, thereby creating a table
# which stores all necessary information about the MarkTypes for all
# ODEMx classes supporting the Trace mechanism.

# --------------------- GLOBALS ------------------------------------------------

# specify the HTML file which is included by doxygen in odemxMarkTypes.dox
DOXYGENFILE="odemxMarkTypes.html"

# temporary variables for storing the extracted
# information describing each mark type
MARKTYPE=""
MARKTEXT=""
MARKOFFSET=""
MARKSCOPE=""
MARKDESCRIPTION=""
MARKCOUNTER=0
LASTSCOPE=""

# --------------------- CHECKS --------------------------------------------------

#~ # check if html file already exists and prompt user for overwrite
#~ if [ -f $DOXYGENFILE ]
#~ then
	#~ echo -n "$DOXYGENFILE already exists. Overwrite? (Y/n):"
	#~ read -n 1 CONTINUE
	#~ echo
#~
	#~ # finish only with 'n', continue on any other key
	#~ if [ "$CONTINUE" = "n" ]
	#~ then
		#~ exit
	#~ fi
#~
	#~ # remove the old file, will be recreated with new HTML table
	#~ rm -f $DOXYGENFILE
#~ fi

# ---------------------- FUNCTIONS ----------------------------------------------


# mark type pattern:
# const MarkType Condq::markCreate( "create", baseMarkId+1, typeid(Condq) ); ///< Call of Condq constructor

# this function can be used to test if the extraction of MarkTypes and
# the corresponding details works properly
testMarkLine() {
	LINE="$@"
	echo $MARKCOUNTER : $LINE

	# remove 'const MarkType Condq::' -- ensures there is no '(' in the greedy part so it won't delete too much
	SHORTLINE=$( echo $LINE | sed s!const.*MarkType[^\(]*:!! )
	echo $SHORTLINE

	# name: remove everything after first '(...'
	MARKTYPE=$( echo $SHORTLINE | sed s![\(].*\$!! )
	echo -n "$MARKTYPE "

	# trace string: remove '^..."' and '"...$'
	MARKTEXT=$( echo $SHORTLINE | sed -e s!^.*[\(][\ ]*\"!! -e s!\".*\$!! )
	echo -n "\"$MARKTEXT\" "

	# trace string: remove '^..."' and '"...$'
	MARKOFFSET=$( echo $SHORTLINE | sed -e s!^.*baseMarkId[\ ]*\+!! -e s!,.*\$!! )
	echo -n "$MARKOFFSET "

	# scope: remove '^...typeid(' and ')...$' to get typeid value
	MARKSCOPE=$( echo $SHORTLINE | sed -e s!^.*typeid[\(]!! -e s![\)].*\$!! )
	echo "$MARKSCOPE"

	# description: remove '^... ///<' to get commentary
	MARKDESCRIPTION=$( echo $SHORTLINE | sed -e s!^.*/[\<]!! )
	echo -- $MARKDESCRIPTION

	let MARKCOUNTER=MARKCOUNTER+1
}

# this function processes the mark lines from the temporary file
# it fills the global variables with the details about each MarkType
processMarkLine() {
	LINE="$@"

	# set the variables needed for
	SHORTLINE=$( echo $LINE | sed s!const.*MarkType[^\(]*:!! )
	MARKTYPE=$( echo $SHORTLINE | sed s![\(].*\$!! )
	MARKTEXT=$( echo $SHORTLINE | sed -e s!^.*[\(][\ ]*\"!! -e s!\".*\$!! )
	MARKOFFSET=$( echo $SHORTLINE | sed -e s!^.*baseMarkId[\ ]*\+!! -e s!,.*\$!! )
	MARKSCOPE=$( echo $SHORTLINE | sed -e s!^.*typeid[\(]!! -e s![\)].*\$!! )
	MARKDESCRIPTION=$( echo $SHORTLINE | sed -e s!^.*/[\<]!! )
}


# ------------------- MAIN SCRIPT ----------------------------------------------


# create HTML file, write HEADER with inline stylesheet
echo "<html>" > $DOXYGENFILE
echo "<head>" >> $DOXYGENFILE
echo "<style>" >> $DOXYGENFILE
cat odemxMarkTypes.css >> $DOXYGENFILE
echo "</style>" >> $DOXYGENFILE
echo "</head>" >> $DOXYGENFILE
echo >> $DOXYGENFILE

# write HTML BODY with scripted MarkType table
echo "<body>" >> $DOXYGENFILE
echo "<table cellspacing='0' cols='3' width='99%'>" >> $DOXYGENFILE
echo "<col width='25%'>" >> $DOXYGENFILE

# color switcher for even and odd lines
EVEN="false"

# find all mark types in .cpp files and store them in tmp file
grep -r -h --include=*.cpp 'const.*MarkType.*typeid' .. | grep -v 'testing' | grep -v 'getNewBaseMarkId' > unsorted

# sort the lines in unsorted file and store them in tmp file
sort unsorted > tmp
rm -f unsorted

# assign file tmp to file descriptor 3 for read access
exec 3< tmp

# read lines from tmp file
while read <&3 LINE
do

# uncomment the next line for testing only
#	testMarkLine $LINE

	# extract MarkType details from mark line
	processMarkLine $LINE

	# check if a new scope is to be written to the HTML file
	if [ ! "$LASTSCOPE" = "$MARKSCOPE" ]
	then

		echo "<tr>" >> $DOXYGENFILE
		echo "	<td colspan='3' class='emph'><b>$MARKSCOPE</b></td>" >> $DOXYGENFILE
		echo "</tr>" >> $DOXYGENFILE
		echo "<tr class='colhead'>" >> $DOXYGENFILE
		echo "	<td>Trace Event</td>" >> $DOXYGENFILE
		echo "	<td>Mark Type</td>" >> $DOXYGENFILE
		echo "	<td class='right'>BaseMarkId Offset</td>" >> $DOXYGENFILE
		echo "</tr>" >> $DOXYGENFILE

		# remember the new MarkType scope
		LASTSCOPE=$MARKSCOPE
	fi

	# check which color to draw in this row
	if [ "$EVEN" = "true" ]
	then echo "<tr class='even'>" >> $DOXYGENFILE
	else echo "<tr class='odd'>" >> $DOXYGENFILE
	fi

	# insert MarkType details into table
	echo "	<td rowspan='2'>&quot;$MARKTEXT&quot;</td>" >> $DOXYGENFILE
	echo "	<td class='italic'>$MARKTYPE</td>" >> $DOXYGENFILE
	echo "	<td class='right'>$MARKOFFSET</td>" >> $DOXYGENFILE
	echo "</tr>" >> $DOXYGENFILE

	if [ "$EVEN" = "true" ]
	then echo "<tr class='even'>" >> $DOXYGENFILE
	else echo "<tr class='odd'>" >> $DOXYGENFILE
	fi

	# write MarkType description to table
	echo "	<td colspan='2' class='bold'>$MARKDESCRIPTION</td>" >> $DOXYGENFILE
	echo "</tr>" >> $DOXYGENFILE

	# set color switcher variable
	if [ "$EVEN" = "true" ];
	then EVEN="false"
	else EVEN="true"
	fi
done

# close the HTML table
echo "</table>" >> $DOXYGENFILE

# write closing tags to file
echo "</body></html>" >> $DOXYGENFILE

# remove tmp file
rm -f tmp

#~ # prompt user whether to run doxygen on the new data
#~ echo -n "Run Doxygen? (Y/n):"
#~ read -n 1 DOXYEXEC
#~ echo
#~
#~ # finish only with 'n', continue on any other key
#~ if [ "$DOXYEXEC" = "n" ]
#~ then
	#~ exit
#~ fi
#~
#~ # run doxygen
#~ doxygen Doxyfile
