//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Event.cpp

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Implementation of classes in Event.h

	\sa Event.h

	<!-- [detailed description] -->

	\since 2.0
*/

#include <odemx/base/Event.h>
#include <odemx/util/ErrorHandling.h>
#include <string>
#include <cassert>

using namespace odemx;

Event::Event ( Label l, EventObserver* eo ):
	Sched( EVENT ),
	DefLabeledObject( getDefaultSimulation(), l ),
	Observable<EventObserver>( eo ),
	eventTime( static_cast< SimTime >(0) ),
	env( getDefaultSimulation() ),
	priority( 0.0 ) // same default priority like Process
	{
		// trace
		getTrace() -> mark( this, markCreate );

		// observer
		_obsForEach(EventObserver, Create(this));
	}

Event::Event ( Simulation* sim, Label l, EventObserver* eo ):
	Sched( EVENT ),
	DefLabeledObject( sim, l ),
	Observable<EventObserver>( eo ),
	eventTime( static_cast< SimTime >(0) ),
	env( sim ),
	priority( 0.0 ) // same default priority like Process
	{
		// trace
		getTrace() -> mark( this, markCreate );

		// observer
		_obsForEach(EventObserver, Create(this));
	}

Event::~Event ()
	{
		// trace
		getTrace() -> mark( this, markDestroy );

		// observer
		_obsForEach(EventObserver, Destroy(this));
	}

SimTime Event::getExecutionTime() const {

	return eventTime;
}


SimTime Event::setExecutionTime( SimTime newTime ) {

	SimTime oldTime = eventTime;
	eventTime = newTime;

	// trace
	getTrace()->mark(this, markChangeExecutionTime, newTime, oldTime);

	// observer
	_obsAForEach(EventObserver, ExecutionTime, oldTime, eventTime);

	return oldTime;
}


Priority Event::getPriority() const {

	return priority;
}


SimTime Event::getCurrentTime() {

	return env -> getTime();
}


Simulation* Event::getSimulation() {

	return env;
}


void Event::schedule() {

	//trace
	getTrace() -> mark( this, markSchedule, getPartner() );

	// observer
	_obsForEach( EventObserver, Schedule( this ) );

	// scheduling
	setExecutionTime( getCurrentTime() );
	env -> insertSched( this );
}


void Event::scheduleIn( SimTime delay ) {
	assert( ! (delay < 0) );

	// trace
	getTrace() -> beginMark( this, markScheduleIn );
	getTrace() -> addTag( tagCurrent, getPartner() );
	getTrace() -> addTag( tagRelSimTime, delay );
	getTrace() -> endMark();

	// observer
	_obsForEach( EventObserver, ScheduleIn( this, delay ) );

	// scheduling
	setExecutionTime( getCurrentTime() + delay );
	env -> insertSched( this );
}


void Event::scheduleAt( SimTime absoluteTime ) {
	assert( ! (absoluteTime < getCurrentTime()) );

	// trace
	getTrace() -> beginMark( this, markScheduleAt );
	getTrace() -> addTag( tagCurrent, getPartner() );
	getTrace() -> addTag( tagAbsSimTime, absoluteTime );
	getTrace() -> endMark();

	// observer
	_obsForEach( EventObserver, ScheduleAt( this,  absoluteTime ) );

	// scheduling
	setExecutionTime( absoluteTime );
	env -> insertSched( this );
}


void Event::scheduleAppend() {

	//trace
	getTrace() -> mark( this, markScheduleAppend, getPartner());

	// observer
	_obsForEach( EventObserver, ScheduleAppend( this ) );

	// scheduling
	setExecutionTime( getCurrentTime() );
	env -> addSched( this );
}


void Event::scheduleAppendIn( SimTime delay ) {
	assert( ! (delay < 0) );

	// trace
	getTrace() -> beginMark( this, markScheduleAppendIn );
	getTrace() -> addTag( tagCurrent, getPartner() );
	getTrace() -> addTag( tagRelSimTime, delay );
	getTrace() -> endMark();

	// observer
	_obsForEach( EventObserver, ScheduleAppendIn( this, delay ) );

	// scheduling
	setExecutionTime( getCurrentTime() + delay );
	env -> addSched( this );
}


void Event::scheduleAppendAt( SimTime absoluteTime ) {
	assert( ! (absoluteTime < getCurrentTime()) );

	// trace
	getTrace() -> beginMark( this, markScheduleAt );
	getTrace() -> addTag( tagCurrent, getPartner() );
	getTrace() -> addTag( tagAbsSimTime, absoluteTime );
	getTrace() -> endMark();

	// observer
	_obsForEach( EventObserver, ScheduleAppendAt( this,  absoluteTime ) );

	// scheduling
	setExecutionTime( absoluteTime );
	env -> addSched( this );
}


void Event::removeFromSchedule() {

	if ( Sched::scheduled ) {

		env -> removeSched( this );

		// trace
		getTrace() -> mark( this, markRemoveFromSchedule, getPartner() );

		// observer
		_obsForEach( EventObserver, RemoveFromSchedule( this ) );
	}
	else {

		error( this,  "Event::removeFromSchedule(): Event not scheduled" );
	}
}


void Event::execute() {

	if ( ! Sched::scheduled ) {
		// error: Event has to be scheduled to be executed
		fatalError( this, "Event::executeEvent(); event is not in schedule", -1 );
	}

	// remove Event from ExecutionList
	env -> removeSched( this );

	// trace
	getTrace() -> mark( this, markExecute );

	// observer
	_obsForEach( EventObserver, ExecuteEvent( this ) );

	// current must be set here, the old value is needed for the trace above
	env -> setCurrentSched( this );

	// execute the action represented by this Event
	eventAction();
}


Priority Event::setPriority( Priority newPriority ) {

	Priority oldPriority = priority;
	priority = newPriority;

	// trace
	getTrace() -> mark( this, markChangePriority, newPriority, oldPriority );

	// observer
	_obsAForEach( EventObserver, Priority, oldPriority, newPriority );

	if ( Sched::scheduled ) {
		// setPriority() influences ExecutionList
		env -> inSort( this );
	}

	return oldPriority;
}


TraceProducer* Event::getPartner() {

	if ( env -> getCurrentSched() != 0 ) {

		return env -> getCurrentSched();
	}
	else {

		return env;
	}
}


const MarkTypeId Event::baseMarkId = MarkType::getNewBaseMarkId( typeid(Event) );

const MarkType Event::markCreate("create", baseMarkId+1, typeid(Event)); ///< Call of Event constructor
const MarkType Event::markDestroy("destroy", baseMarkId+2, typeid(Event)); ///< Call of Event destructor

const MarkType Event::markSchedule("schedule", baseMarkId+3, typeid(Event)); ///< Event scheduled at SimTime now with LIFO strategy (inserted in front)
const MarkType Event::markScheduleAt("schedule at", baseMarkId+4, typeid(Event)); ///< Event scheduled at absolute SimTime t with LIFO strategy (inserted in front)
const MarkType Event::markScheduleIn("schedule in", baseMarkId+5, typeid(Event)); ///< Event scheduled at relative SimTime now+t with LIFO strategy (inserted in front)

const MarkType Event::markScheduleAppend("schedule append", baseMarkId+6, typeid(Event)); ///< Event scheduled at SimTime now with FIFO strategy (inserted at the end)
const MarkType Event::markScheduleAppendAt("schedule append at", baseMarkId+7, typeid(Event)); ///< Event scheduled at absolute SimTime t with FIFO strategy (inserted at the end)
const MarkType Event::markScheduleAppendIn("schedule append in", baseMarkId+8, typeid(Event)); ///< Event scheduled at relative SimTime now+t with FIFO strategy (inserted at the end)

const MarkType Event::markRemoveFromSchedule("remove from schedule", baseMarkId+9, typeid(Event)); ///< Event was removed from schedule (ordered execution list)
const MarkType Event::markExecute("execute", baseMarkId+10, typeid(Event)); ///< Event executed by the Simulation, call of virtual function eventAction()

const MarkType Event::markChangePriority("change priority", baseMarkId+11, typeid(Event)); ///< Change of Event priority, may influence order of execution list
const MarkType Event::markChangeExecutionTime("change execution time", baseMarkId+12, typeid(Event)); ///< Change of Event execution time, may influence order of execution list

const TagId Event::baseTagId = Event::baseMarkId;

const Tag Event::tagCurrent( "current", baseTagId+1 ); ///< Displays the currently active Sched object
const Tag Event::tagAbsSimTime( "absolute time", baseTagId+2 ); ///< Displays details about scheduling at absolute time
const Tag Event::tagRelSimTime( "relative time", baseTagId+3 ); ///< Displays details about scheduling at relative time
