//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ExecutionList.cpp

	\author Ralf Gerstenberger
	<!-- [\author <author>]* -->

	\date created at 2002/01/25

	\brief Implementation of ExecutionList

	\sa ExecutionList.h

	<!-- [detailed description] -->

	\since 1.0
*/

#include <odemx/base/Event.h>
#include <odemx/base/Process.h>
#include <odemx/base/ExecutionList.h>
#include <odemx/base/InSort.h>
#include <odemx/util/ErrorHandling.h>

#include <cassert>
#include <iostream>
#include <algorithm>

using namespace std;
using namespace odemx;

ExecutionList::ExecutionList(ExecutionListObserver* o):
	Observable<ExecutionListObserver>(o)
	{
		// observer
		_obsForEach(ExecutionListObserver, Create(this));
	}

ExecutionList::~ExecutionList() {

	// observer
	_obsForEach(ExecutionListObserver, Destroy(this));
}

// attach Sched s at its ExecutionTime
// to ExecutionList (FIFO), considering priority
void ExecutionList::addSched( Sched* s ) {
	assert( s );

	// adjust execution time
	if ( s -> getExecutionTime() < getTime() )
		s -> setExecutionTime( getTime() );

	inSort( s );

	// observer
	_obsForEach(ExecutionListObserver, AddSched(this, s));
}

// insert Sched s at ExecutionTime in
// ExecutionList (LIFO), considering priority
void ExecutionList::insertSched( Sched* s ) {
	assert( s );

	// adjust execution time
	if ( s -> getExecutionTime() < getTime() )
		s -> setExecutionTime( getTime() );

	inSort( s, false );

	// observer
	_obsForEach(ExecutionListObserver, InsertSched(this, s));
}

// insert Sched s after 'previous', changing priority if necessary
void ExecutionList::insertSchedAfter( Sched* s, Sched* previous ) {
	assert( s );
	assert( previous );

	if ( s == previous )
		return;

	if ( ! previous -> scheduled ) {
		// error: 'previous' is not listed
		error( "ExecutionList::insertSchedAfter(); 'previous' not scheduled" );
		return;
	}

	// remove Sched s from ExecutionList if scheduled
	if ( s -> scheduled ) {
		// assert( find(l.begin(), l.end(), s) != l.end() );
		l.erase( s -> execListIter );
	}

	Priority prevPrio = previous -> getPriority();

	// adjust priority so the list order remains correct for s and previous
	if ( s -> getPriority() > prevPrio )
		s -> setPriority( prevPrio );

	// set s to execution time of previous
	s -> setExecutionTime( previous -> getExecutionTime() );

	// insert Sched p after 'pre' (before 'next' after 'pre')
	std::list<Sched*>::iterator pre = previous -> execListIter;
	++pre;
	s -> execListIter = l.insert( pre, s );
	s -> scheduled = true;

	_obsForEach(ExecutionListObserver, InsertSchedAfter(this, s, previous));
}

void ExecutionList::insertSchedBefore( Sched* s, Sched* next ) {
	// insert Sched s before 'next', changing priority if necessary
	assert( s );
	assert( next );

	if ( s == next )
		return;

	if ( ! next -> scheduled ) {
		// error: 'next' is not listed
		error( "ExecutionList::insertSchedBefore(); 'next' not scheduled" );
		return;
	}

	// remove Sched s from ExecutionList if scheduled
	if ( s -> scheduled ) {
		// assert( find(l.begin(), l.end(), s) != l.end() );
		l.erase( s -> execListIter );
	}

	Priority nextPrio = next -> getPriority();

	// adjust priority
	if ( s -> getPriority() < nextPrio )
		s -> setPriority( nextPrio );

	// set s to execution time of next
	s -> setExecutionTime( next -> getExecutionTime() );

	// insert Sched s before 'next'
	s -> execListIter = l.insert( next -> execListIter, s );
	s -> scheduled = true;

	_obsForEach(ExecutionListObserver, InsertSchedBefore(this, s, next));
}

void ExecutionList::removeSched( Sched* s ) {

	if ( ! s -> scheduled ) {
		error( "ExecutionList::removeSched(): attempt to remove unscheduled object: "
				+ s -> getLabel() );
		return;
	}

	l.erase( s -> execListIter );
	s -> scheduled = false;

	_obsForEach(ExecutionListObserver, RemoveSched(this, s));
}

Sched* ExecutionList::getNextSched() {

	return l.empty()  ?  0 : l.front();
}


void ExecutionList::inSort( Sched* s, bool fifo/* = true*/ ) {
	assert( s != 0 );

	// remove the object from the list if it is already scheduled
	if ( s -> scheduled ) {
		// assert( std::find(l.begin(), l.end(), s) != l.end() );
		l.erase( s -> execListIter );
	}

	// insert s into ordered list and store the location as iterator
	s -> execListIter = ListInSort< Sched*, DefaultCmp >( l, s, defCmp, fifo );
	s -> scheduled = true;
}


void ExecutionList::printExecutionList() {

	std::list< Sched* >::iterator iter;

	cout << "--------------" << endl;
	cout << "ExecutionList: " << l.size() << " objects" << endl;

	for ( iter = l.begin(); iter != l.end(); ++iter ) {

		if ( (*iter) -> getSchedType() == Sched::PROCESS ) {
			cout << "  Process \"" << (*iter) -> getLabel()
				 << "\" scheduled at " << (*iter) -> getExecutionTime();
		}

		if ( (*iter) -> getSchedType() == Sched::EVENT ) {
			cout << "  Event \"" << (*iter) -> getLabel()
				 << "\" scheduled at " << (*iter) -> getExecutionTime();
		}
		cout << endl;
	}
	cout << "--------------" << endl;
}
