//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file InSort.h

	\author Ronald Kluth
	<!-- [\author <author>]* -->

	\date created at 2007/11/08

	\brief InSort template for ordered std::list<> container

	<!-- [\sa <pair-file>] -->

	<!-- [detailed description] -->

	

	\since 2.1
*/

#ifndef ODEMX_INSORT_INCLUDED
#define ODEMX_INSORT_INCLUDED

#include <odemx/base/Process.h>

#include <list>
#include <algorithm>

namespace odemx {

	/** \struct DefaultCmp

		\author Ronald Kluth

		\brief DefaultCmp is used for ordering objects in std::list<> 
		
		\sa ListInSort, PriorityCmp
	*/
	extern struct DefaultCmp {
		
		bool operator ()( const Sched* f, const Sched* s ) const {
			return *f < *s;
		}
		
	} defCmp;

	/** \struct QPriorityCmp

		\author Ronald Kluth

		\brief QPriorityCmp is used for ordering objects in std::list<> 
		
		\sa ListInSort, DefaultCmp
	*/
	extern struct QPriorityCmp {
		
		bool operator ()( const Process* f, const Process* s ) const {
			return f -> getQueuePriority() > s -> getQueuePriority();
		}
		
	} qPrioCmp;
	
	/**
		\brief sorted insertion of elements into std::list<> of various types

		\param list
			reference to std::list<> container

		\param newElement
			the element to be sorted into the given list 

		\param cmp
			the chosen comparison functor by which the list is ordered

		\param fifo
			use FIFO or LIFO strategy

		ListInSort is used for insertion of Sched objects into a 
		std::list< Sched* >	container by the ExecutionList. It is also
		employed to sort a ProcessQueue of type std::list< Process* >
		using a user-defined predicate as sorting criterion.
		
		\sa DefaultCmp, QPriorityCmp
	*/
	template < typename ListElementType, typename ComparisonFunctorType >
	typename std::list< ListElementType >::iterator 
	inline ListInSort( std::list< ListElementType >& l, 
					   ListElementType newElement, 
					   ComparisonFunctorType& cmp, 
					   bool fifo = true ) 
	{
		// an iterator to store the position 
		// where the new element should be inserted
		typename std::list< ListElementType >::iterator insertLocation;
		
		// according to fifo or lifo, the new element is inserted
		// either at the end or at the beginning of the possible range
		// the insert location is determined by standard algorithms
		if( fifo )
			insertLocation = std::upper_bound( l.begin(), l.end(), newElement, cmp );
		else // lifo
			insertLocation = std::lower_bound( l.begin(), l.end(), newElement, cmp );

		// insert returns an iterator to the location 
		// where the new element was placed in the list
		return l.insert( insertLocation, newElement );
	}
}

#endif

