//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Process.cpp

	\author Ralf Gerstenberger
	<!-- [\author <author>]* -->

	\date created at 2002/01/30

	\brief Implementation of classes in Process.h

	\sa Process.h

	<!-- [detailed description] -->

	\since 1.0
*/

#include <odemx/base/Process.h>
#include <odemx/base/Event.h>
#include <odemx/base/Simulation.h>
#include <odemx/util/ErrorHandling.h>

#include <string>

using namespace odemx;
using namespace std;

Process::Process( Label l, ProcessObserver* o ):
	Sched( PROCESS ),
	Coroutine(l, getDefaultSimulation(), o),
	Observable<ProcessObserver>(o),
	processState(CREATED),
	p(0.0),
	t(static_cast< SimTime >(0)),
	env(getDefaultSimulation()),
	validReturn(false),
	returnValue(0),
	q(0),
	q_in(static_cast< SimTime >(0)),
	q_out(static_cast< SimTime >(0)),
	qPriority( 0.0 ),
	interrupted(false),
	interrupter(0),
	alerted(false),
	alerter(0),
	internalMemVec( false )
	{
		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEach(ProcessObserver, Create(this));

		env -> setProcessAsCreated(this);
	}


Process::Process(Simulation* s, Label l, ProcessObserver* o) :
	Sched(PROCESS),
	Coroutine(l, s, o),
	Observable<ProcessObserver>(o),
	processState(CREATED),
	p(0.0),
	t(static_cast< SimTime >(0)),
	env(s),
	validReturn(false),
	returnValue(0),
	q(0),
	q_in(static_cast< SimTime >(0)),
	q_out(static_cast< SimTime >(0)),
	qPriority( 0.0 ),
	interrupted(false),
	interrupter(0),
	alerted(false),
	alerter(0),
	internalMemVec( false )
	{
		if (s==0) {
			// error: Process was created without reference to simulation.
			fatalError( this, "Process(); invalid argument: Simulation* s", -1);
		}

		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEach(ProcessObserver, Create(this));

		env->setProcessAsCreated(this);
	}

Process::~Process() {

	// trace
	getTrace()->mark(this, markDestroy);

	// observer
	_obsForEach(ProcessObserver, Destroy(this));

	if (!env->isFinished()) {
		// warning: Process should not be destroyed while simulation is running.
		warning( this, "~Process(); simulation not finished");
	}

	// the simulation must not have a pointer to a destroyed process
	env->removeProcessFromList(this);
}


void Process::activate() {
	// trace
	getTrace()->mark(this, markActivate, getPartner());

	// observer
	_obsForEach(ProcessObserver, Activate(this));

	// state management
	if (!setProcessState(RUNNABLE))
		return;

	// reset interrupt state
	resetInterrupt();

	// reset Memory::alert state
	resetAlert();

	// scheduling
	setExecutionTime(getCurrentTime());
	env->insertSched(this);

	// continue simulation
	env->compSimulation();
}

void Process::activateIn(SimTime t) {
	// trace
	traceActivateIn(t);

	// observer
	_obsForEach(ProcessObserver, ActivateIn(this, t));

	// state management
	if (!setProcessState(RUNNABLE))
		return;

	// reset interrupt state
	resetInterrupt();

	// reset Memory::alert state
	resetAlert();

	// scheduling
	setExecutionTime(getCurrentTime() + t);
	env->insertSched(this);

	// continue simulation
	env->compSimulation();
}

void Process::activateAt(SimTime t) {
	// trace
	traceActivateAt(t);

	// observer
	_obsForEach(ProcessObserver, ActivateAt(this, t));

	// state management
	if (!setProcessState(RUNNABLE))
		return;

	// reset interrupt state
	resetInterrupt();

	// reset Memory::alert state
	resetAlert();

	// scheduling
	setExecutionTime(t);
	env->insertSched(this);

	// continue simulation
	env->compSimulation();
}

void Process::activateBefore(Sched* p) {
	// trace
	traceActivateBefore(p);

	// observer
	_obsForEach(ProcessObserver, ActivateBefore(this, p));

	// state management
	if (!setProcessState(RUNNABLE))
		return;

	// reset interrupt state
	resetInterrupt();

	// reset Memory::alert state
	resetAlert();

	// scheduling
	env->insertSchedBefore(this, p);

	// continue simulation
	env->compSimulation();
}

void Process::activateAfter(Sched* p) {
	// trace
	traceActivateAfter(p);

	// observer
	_obsForEach(ProcessObserver, ActivateAfter(this, p));

	// state management
	if (!setProcessState(RUNNABLE))
		return;

	// reset interrupt state
	resetInterrupt();

	// reset Memory::alert state
	resetAlert();

	// scheduling
	env->insertSchedAfter(this, p);

	// continue simulation ?
	// this function should never require a context change
	// because scheduling after another Sched object requires
	// for that to be executed first;

	// if compSimulation() is called here, it leads to odd
	// stepping results because the caller of this function
	// has to continue its life cycle anyway

	// env->compSimulation();
}

void Process::hold() {
	// trace
	getTrace()->mark(this, markHold, getPartner());

	// observer
	_obsForEach(ProcessObserver, Hold(this));

	// state management
	if (!setProcessState(RUNNABLE))
		return;

	// reset interrupt state
	resetInterrupt();

	// reset Memory::alert state
	resetAlert();

	// scheduling
	setExecutionTime(getCurrentTime());
	env->addSched(this);

	// continue simulation
	env->compSimulation();
}

void Process::holdFor(SimTime t) {
	// trace
	traceHoldFor(t);

	// observer
	_obsForEach(ProcessObserver, HoldFor(this, t));

	// state management
	if (!setProcessState(RUNNABLE))
		return;

	// reset interrupt state
	resetInterrupt();

	// reset Memory::alert state
	resetAlert();

	// scheduling
	setExecutionTime(getCurrentTime() + t);
	env->addSched(this);

	// continue simulation
	env->compSimulation();
}


void Process::holdUntil(SimTime t) {
	// trace
	traceHoldUntil(t);

	// observer
	_obsForEach(ProcessObserver, HoldUntil(this, t));

	// state management
	if (!setProcessState(RUNNABLE))
		return;

	// reset interrupt state
	resetInterrupt();

	// reset Memory::alert state
	resetAlert();

	// scheduling
	setExecutionTime(t);
	env->addSched(this);

	// continue simulation
	env->compSimulation();
}


// m0 always has to be given, else no matching call
Memory* Process::wait( Memory* m0, Memory* m1, Memory* m2,
					   Memory* m3, Memory* m4, Memory* m5 ) {

	MemoryVector* memvec = new MemoryVector;
	internalMemVec = true;

	memvec -> reserve( 6 );
	memvec -> push_back( m0 );

	// add more parameters to vector, if they are given
	if ( m1 ) memvec->push_back( m1 );
	if ( m2 ) memvec->push_back( m2 );
	if ( m3 ) memvec->push_back( m3 );
	if ( m4 ) memvec->push_back( m4 );
	if ( m5 ) memvec->push_back( m5 );

	return wait( memvec );
}


Memory* Process::wait( MemoryVector* memvec ) {

	if ( ( !memvec ) || ( memvec -> empty() ) ) {
		// warning no alerter given
		warning( this,  "Process::wait(): No alerter given" );
		return 0;
	}

	// trace
	traceWait( memvec );

	// observer
	_obsForEach( ProcessObserver, Wait( this, memvec ) );

	// before suspending, check availability of resources
	MemoryVector::iterator iter;
	for ( iter = memvec -> begin(); iter != memvec -> end(); ++iter ) {

		// check Memory object availability
		if ( (*iter) -> available() ) {

			Memory* theAvailable = (*iter);

			// trace
			getTrace() -> mark( this, markAvailable, theAvailable -> getLabel() );

			// merely warn if Timer is not set
			if ( theAvailable -> getMemoryType() == Memory::TIMER )
			{
				warning( this,  "Process::wait(): Timer available due to not being set: "
						 + theAvailable -> getLabel() );
			}
			else // other Memo objects will be returned
			{
				// clean up if vector was created by wait()
				if ( internalMemVec ) {
					delete memvec;
					internalMemVec = false;
				}

				return theAvailable;
			}
		}
	}

	// since none are available, register this process with the given Memory objects
	for ( iter = memvec -> begin(); iter != memvec -> end(); ++iter ) {
		(*iter) -> remember( this );
	}

	// the alert mechanism is normally only used with wait()
	// reset alerter and alerted anyway, if set somewhere else
	resetAlert();

	// suspend process
	sleep();

	// ************* beyond this point, the process was awakened ************

	// check for Memory alert
	if ( isAlerted() ){

		Memory* currentAlerter = getAlerter();

		for ( iter = memvec -> begin(); iter != memvec -> end(); ++iter ) {
			if ( (*iter) != currentAlerter ) // alerter has already removed this process
				(*iter) -> forget(this);
		}

		// reset so outdated alerted/alerter cannot be used elsewhere by accident
		resetAlert();

		// clean up internal vector created by wait()
		if ( internalMemVec ) {
			delete memvec;
			internalMemVec = false;
		}

		return currentAlerter;
	}
	// the sleep() period can be interrupted by other processes
	else if ( isInterrupted() ) {

		// remove this from all Memos
		for ( iter = memvec->begin(); iter != memvec->end(); ++iter ) {

			(*iter) -> forget(this);
		}

		warning( this,  "Process::wait(): waiting period ended by interrupt" );

		if ( internalMemVec ) {
			delete memvec;
			internalMemVec = false;
		}
		// when wait returns 0, a Process should check for interruption
		return 0;
	}
	else
		error( this,  "Process::wait(): Process awakened without alert or interrupt" );

	return 0;
}


void Process::alertProcess( Memory* theAlerter ) {

	// trace
	getTrace() -> mark( this, markAlert );

	// observer
	_obsForEach( ProcessObserver, Alert( this, theAlerter ) );

	alerted = true;
	alerter = theAlerter;

	// state management
	if (!setProcessState(RUNNABLE))
		return;

	// scheduling
	setExecutionTime(getCurrentTime());
	env->insertSched(this);

	// env->addSched(this) might cause problems with scheduling ??
}


bool Process::isAlerted() const {
	return alerted;
}


Memory* Process::getAlerter() const {
	return alerter;
}


void Process::resetAlert() {
	alerted = false;
	alerter = 0;
}


void Process::interrupt() {
	if (this == env -> getCurrentSched()) {
		// ignore self interrupt
		return;
	}

	if (isInterrupted()) {
		// Warning: Process already interrupted
		warning( this, "interrupt(); Process already interrupted");
	}

	// trace
	getTrace()->mark(this, markInterrupt, getPartner());

	// observer
	_obsForEach(ProcessObserver, Interrupt(this));

	// Set interrupt data
	interrupted = true;
	interrupter = env -> getCurrentSched();

	// state management
	if (!setProcessState(RUNNABLE))
		return;

	// scheduling
	setExecutionTime(getCurrentTime());
	env->insertSched(this);

	// continue simulation
	env->compSimulation();
}


void Process::sleep() {
	// trace
	getTrace()->mark(this, markSleep, getPartner());

	// observer
	_obsForEach(ProcessObserver, Sleep(this));

	// reset interrupt state
	resetInterrupt();

	// state management
	if (!setProcessState(IDLE))
		return;

	// scheduling
	env->removeSched(this);

	// continue simulation
	env->compSimulation();
}

void Process::cancel() {
	// trace
	getTrace()->mark(this, markCancel, getPartner());

	// observer
	_obsForEach(ProcessObserver, Cancel(this));

	if (setProcessState(TERMINATED)) {
		// scheduling
		if( isScheduled() )
			env->removeSched(this);

		// continue simulation
		env->compSimulation();
	}
}

Sched* Process::getCurrentSched() {
	return env->getCurrentSched();
}

Process* Process::getCurrentProcess() {
	return env->getCurrentProcess();
}

SimTime Process::getCurrentTime() {
	return env->getTime();
}

Simulation* Process::getSimulation() const {
	return env;
}

int Process::getReturnValue() const {
	if (!validReturn) {
		// Warning: return value is not valid
		warning( this, "getReturnValue(); return value is not valid");
	}

	return returnValue;
}


void Process::execute() {

	if (processState!=RUNNABLE && processState!=CURRENT) {
		// error: Process has to be RUNNABLE or CURRENT to be executed.
		fatalError( this, "execute(); process is not RUNNABLE or CURRENT", -1);
	}

	// trace
	getTrace()->mark(this, markExecute);

	// observer
	_obsForEach(ProcessObserver, Execute(this));

	// old currentSched value is needed in getPartner()
	setProcessState(CURRENT);

	// switch to Coroutine
	switchTo();
}


Process::ProcessState Process::getProcessState() const {
	return processState;
}


bool Process::setProcessState(ProcessState newState) {
	if (processState==TERMINATED) {
		// error: Process already terminated
		error( this, "setProcessState(); changing state of terminated process is impossible");
		return false;
	}

	if (newState==CREATED) {
		// error: Process already created
		error( this, "setProcessState(); Process already created");
		return false;
	}

	ProcessState oldState = processState;
	
	switch (newState) {
	case CREATED:
		env->setProcessAsCreated(this);
		break;
	case RUNNABLE:
		env->setProcessAsRunnable(this);
		break;
	case CURRENT:
		env->setProcessAsCurrent(this);
		break;
	case IDLE:
		env->setProcessAsIdle(this);
		break;
	case TERMINATED:
		env->setProcessAsTerminated(this);
		break;
	}

	// must set state here because the old state is still needed
	// by env -> setProcessAsXXX
	processState = newState;

	// trace
	getTrace()->mark(this, markChangeProcessState, newState, oldState);

	// observer
	_obsAForEach(ProcessObserver, ProcessState, oldState, newState);

	return true;
}

void Process::start() {
	returnValue = main();

	// Process finished [returned from main()]
	//if (setProcessState(TERMINATED))
	//	env->removeSched(this);

	env->removeSched(this);

	// The return value is valid now
	validReturn=true;

	// observer
	_obsForEach(ProcessObserver, Return(this));

	setProcessState(TERMINATED);

	// in case the context is switched, this coroutine will not be reactivated
	// so it must be cleared here
	Coroutine::clear();

	// continue simulation
	env->compSimulation();
}

Priority Process::getPriority() const {
	return p;
}

Priority Process::setPriority(Priority newPriority) {

	Priority oldPriority = p;
	p = newPriority;

	// trace
	getTrace()->mark(this, markChangePriority, newPriority, oldPriority);

	// observer
	_obsAForEach(ProcessObserver, Priority, oldPriority, newPriority);

	// changes in priority may influence ExecutionList
	if( isScheduled() || getProcessState() == CURRENT )
		env->inSort(this);

	// precise check for context change because running insertSchedAfter()
	// creates unexpected stepping if its call to setPriority() causes
	// compSimulation() to be called without need
	if( env -> getCurrentSched() &&
		env -> getNextSched() != env -> getCurrentSched() )
	{
		env->compSimulation();
	}

	return oldPriority;
}

SimTime Process::getExecutionTime() const {
	if (processState!=RUNNABLE &&
		processState!=CURRENT)
		return static_cast< SimTime >(0);

	return t;
}

SimTime Process::setExecutionTime( SimTime time ) {
	SimTime oldTime = t;
	t = time;

	// trace
	getTrace()->mark(this, markChangeExTime, time, oldTime);

	// observer
	_obsAForEach(ProcessObserver, ExecutionTime, oldTime, t);

	return oldTime;
}

void Process::enqueue(ProcessQueue* nq) {
	if (q!=0) {
		// error: Process is already listed in ProcessQueue.
		fatalError( this, "Process::enqueue(); Process already listed in a ProcessQueue", -1);
	}

	q=nq;
	q_in=getCurrentTime();
}

void Process::dequeue(ProcessQueue* nq) {
	if (q!=nq) {
		// error: Process dequeued() by wrong ProcessQueue
		error( this, "dequeue(); Process dequeued() by wrong ProcessQueue");
		return;
	}

	q=0;
	q_out=getCurrentTime();
}

Priority Process::getQueuePriority() const {
	return qPriority;
}

Priority Process::setQueuePriority( Priority newQPriority, bool reactivate ) {

	Priority oldQPriority = qPriority;
	qPriority = newQPriority;

	// trace
	getTrace() -> mark(this, markChangeQueuePriority, newQPriority, oldQPriority);

	// observer
	_obsAForEach(ProcessObserver, QueuePriority, oldQPriority, newQPriority);

	// check if this Process is in a queue, sort queue again
	if( q != 0 ) {
		q -> inSort( this );
	}

	// this Process might have become the first in a waiting queue,
	// it will still be blocked in queue if condition not met
	if( reactivate ) {
		activate();
		env -> compSimulation();
	}

	return oldQPriority;
}

TraceProducer* Process::getPartner() {

	if ( env -> getCurrentSched() != 0 ) {
		return env -> getCurrentSched();
	}
	else {
		return env;
	}
}



inline void Process::traceActivateIn(SimTime t) {
	getTrace()->beginMark(this, markActivateIn);
	getTrace()->addTag(tagCurrent, getPartner());
	getTrace()->addTag(tagRelSimTime, t);
	getTrace()->endMark();
}

inline void Process::traceActivateAt(SimTime t) {
	getTrace()->beginMark(this, markActivateAt);
	getTrace()->addTag(tagCurrent, getPartner());
	getTrace()->addTag(tagAbsSimTime, t);
	getTrace()->endMark();
}

void Process::traceActivateBefore(Sched* p) {
	getTrace()->beginMark(this, markActivateBefore);
	getTrace()->addTag(tagCurrent, getPartner());
	getTrace()->addTag(tagPartner, p);
	getTrace()->endMark();
}

void Process::traceActivateAfter(Sched* p) {
	getTrace()->beginMark(this, markActivateAfter);
	getTrace()->addTag(tagCurrent, getPartner());
	getTrace()->addTag(tagPartner, p);
	getTrace()->endMark();
}

void Process::traceHoldFor(SimTime t) {
	getTrace()->beginMark(this, markHoldFor);
	getTrace()->addTag(tagCurrent, getPartner());
	getTrace()->addTag(tagRelSimTime, t);
	getTrace()->endMark();
}

void Process::traceHoldUntil(SimTime t) {
	getTrace()->beginMark(this, markHoldUntil);
	getTrace()->addTag(tagCurrent, getPartner());
	getTrace()->addTag(tagAbsSimTime, t);
	getTrace()->endMark();
}

void Process::traceWait( MemoryVector* memvec ) {
	getTrace() -> beginMark( this, markWait );
	MemoryVector::iterator iter;
	for ( iter = memvec -> begin(); iter != memvec -> end(); ++iter ) {
		getTrace() -> addTag( tagPartner, *iter );
	}
	getTrace() -> endMark();
}


const MarkTypeId Process::baseMarkId = MarkType::getNewBaseMarkId( typeid(Process) );

const MarkType Process::markCreate("create", baseMarkId+1, typeid(Process)); ///< Call of Process constructor
const MarkType Process::markDestroy("destroy", baseMarkId+2, typeid(Process)); ///< Call of Process destructor

const MarkType Process::markActivate("activate", baseMarkId+3, typeid(Process)); ///< Process scheduled for activation at SimTime now with LIFO strategy (inserted in front)
const MarkType Process::markActivateAt("activate at", baseMarkId+4, typeid(Process)); ///< Process scheduled for activation at absolute SimTime t with LIFO strategy (inserted in front)
const MarkType Process::markActivateIn("activate in", baseMarkId+5, typeid(Process)); ///< Process scheduled for activation at relative SimTime now+t with LIFO strategy (inserted in front)

const MarkType Process::markActivateBefore("activate before", baseMarkId+6, typeid(Process)); ///< Process scheduled for activation before a given partner process
const MarkType Process::markActivateAfter("activate after", baseMarkId+7, typeid(Process)); ///< Process scheduled for activation after a given partner process
const MarkType Process::markHold("hold", baseMarkId+8, typeid(Process)); ///< Process scheduled for activation at SimTime now with FIFO strategy (inserted at the end)
const MarkType Process::markHoldUntil("hold until", baseMarkId+9, typeid(Process)); ///< Process scheduled for activation at absolute SimTime t with FIFO strategy (inserted at the end)
const MarkType Process::markHoldFor("hold for", baseMarkId+10, typeid(Process)); ///< Process scheduled for activation at relative SimTime now+t with FIFO strategy (inserted at the end)

const MarkType Process::markSleep("sleep", baseMarkId+11, typeid(Process)); ///< Suspension of a Process, removal from schedule (execution list)
const MarkType Process::markInterrupt("interrupt", baseMarkId+12, typeid(Process)); ///< Interruption of suspended Process, activation at SimTime now with LIFO strategy (inserted in front)
const MarkType Process::markCancel("cancel", baseMarkId+13, typeid(Process)); ///< Termination of a Process, removal from schedule

const MarkType Process::markExecute("execute", baseMarkId+14, typeid(Process)); ///< Process executed by the Simulation, describes start or continuation of virtual function main()

const MarkType Process::markWait("wait", baseMarkId+15, typeid(Process)); ///< Process object is suspended, waiting for Memory object(s) to become available
const MarkType Process::markAvailable("found available", baseMarkId+16, typeid(Process)); ///< Process has found an available Memory object
const MarkType Process::markAlert("alerted", baseMarkId+17, typeid(Process)); ///< Process was alerted by a Memory object, activation at SimTime now with LIFO strategy (inserted in front)

const MarkType Process::markChangeProcessState("change state", baseMarkId+18, typeid(Process)); ///< Process has changed its state, displays old and new value
const MarkType Process::markChangePriority("change priority", baseMarkId+19, typeid(Process)); ///< Process priority changed, displays old and new value
const MarkType Process::markChangeQueuePriority("change queue priority", baseMarkId+20, typeid(Process)); ///< Process queue priority changed, displays old and new value
const MarkType Process::markChangeExTime("change execution time", baseMarkId+21, typeid(Process)); ///< Process execution time changed, displays old and new value

const TagId Process::baseTagId = Process::baseMarkId;

const Tag Process::tagCurrent( "current", baseTagId+1 ); ///< Displays the currently active Sched object
const Tag Process::tagAbsSimTime( "absolute time", baseTagId+2 ); ///< Displays details about scheduling at absolute time
const Tag Process::tagRelSimTime( "relative time", baseTagId+3 ); ///< Displays details about scheduling at relative time
const Tag Process::tagPartner( "partner", baseTagId+4 ); ///< Displays details about a partner process
