//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ProcessQueue.cpp

	\author Ralf Gerstenberger
	<!-- [\author <author>]* -->

	\date created at 2002/03/25

	\brief Implementation of ProcessQueue, ProcessOrder, 
	DefaultOrder and PriorityOrder

	\sa ProcessQueue.h

	<!-- [detailed description] -->

	\since 1.0
*/

#include <odemx/base/InSort.h>
#include <odemx/base/Process.h>
#include <odemx/base/ProcessQueue.h>

#include <algorithm>

using namespace odemx;
using namespace std;

ProcessQueue::ProcessQueue(): 
	size( 0 )
	{}

ProcessQueue::~ProcessQueue()
	{}

Process* ProcessQueue::getTop() const {
	if (isEmpty())
		return 0;

	return *l.begin();
}

const list<Process*>& ProcessQueue::getList() const {
	return l;
}

unsigned int ProcessQueue::getLength() const {
	return size; // a lot faster than calling list::size()
}

const unsigned int ProcessQueue::getPosition( Process* p ) const {
	int pos = 1;
	
	list< Process* >::const_iterator i;
	
	// count positions until p is found
	for( i = getList().begin(); i != getList().end() && (*i) != p; ++i )
		++pos;

	// p found
	if( i != getList().end() && (*i) == p ) {
		return pos;
	}
	
	return 0;
}

bool ProcessQueue::isEmpty() const {
	return l.empty();
}

// wraps remove() of first element
void ProcessQueue::popQueue() {
	remove( *l.begin() );
}

void ProcessQueue::remove( Process* p ) {
	assert( p != 0 );

	// Process p forgets this queue
	p -> dequeue( this );

	l.erase( p -> qListIter );
	--size;
}


void ProcessQueue::inSort( Process* p,  bool fifo/* = true*/ ) {
	assert( p != 0 );
	
	// if already in this queue, remove previous entry of Process* p
	if ( p -> q == this ) {
		l.erase( p -> qListIter );
		p -> q = 0; // q != 0 would cause error in Process::enqueue() 
	}
	else { // not yet in queue, increment size
		++size;
	}
	
	// Process p has to remember this queue
	p -> enqueue( this );

	// insert s into ordered list and store the location as iterator
	p -> qListIter = ListInSort< Process*, QPriorityCmp >( l, p, qPrioCmp, fifo );
}

void odemx::awakeAll( ProcessQueue* q ) {
	if( q -> isEmpty() )
		return;

	Sched* current = q -> getTop() -> getCurrentSched(); // may return 0 !
	
	list< Process* >::const_iterator i;
	for( i = q -> getList().begin(); i != q -> getList().end(); ++i ) {
		assert( (Sched*)*i != current );

		// cannot activate after current event, must use activate
		if( ! current || current -> getSchedType() == Sched::EVENT ) {
			(*i) -> activate();
		}
		else {
			(*i) -> activateAfter( current );
		}
	}
}

void odemx::awakeFirst( ProcessQueue* q ) {
	if( q -> isEmpty() )
		return;

	Sched* current = q -> getTop() -> getCurrentSched(); // may return 0 !

	// cannot activate after current event, must use activate
	if( ! current || current -> getSchedType() == Sched::EVENT ) {
		q -> getTop() -> activate();
	}
	else {
		if( q -> getTop() != current )
			q -> getTop() -> activateAfter( current );
	}
}

void odemx::awakeNext( ProcessQueue* q, Process* p ) {
	if( q -> isEmpty() )
		return;

	Sched* current = p -> getCurrentSched(); // may return 0 !
	
	list< Process* >::const_iterator i;
	for( i = q -> getList().begin();
		 i != q -> getList().end() && (*i) != p;
		 ++i );

	if( (*i) == p ) {
		++i;

		if( i != q -> getList().end() ) {
			
			// cannot activate after current event, must use activate
			if( ! current || current -> getSchedType() == Sched::EVENT ) {
				(*i) -> activate();
			}
			else {
				if( (*i) != current )
					(*i) -> activateAfter( current );
			}
		}
	}
}
