//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ProcessQueue.h

	\author Ralf Gerstenberger
	<!-- [\author <author>]* -->

	\date created at 2002/03/25

	\brief Declaration of ProcessQueue, ProcessOrder,
	DefaultOrder and PriorityOrder

	\sa ProcessQueue.cpp

	<!-- [detailed description] -->

	\since 1.0
*/

#ifndef ODEMX_PROCESSQUEUE_INCLUDED
#define ODEMX_PROCESSQUEUE_INCLUDED

#include <list>

namespace odemx {
	class Process;

	/** \class ProcessQueue

		\ingroup base

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief ProcessQueue is used for process synchronization

		<!-- [\note {notes}]* -->

		<!-- [\sa {references to other classes}]* -->

		ProcessQueue is a general tool for storing Process objects. It is used
		by synchronization objects. Every Process can stay in no more than one
		ProcessQueue at a time (provisional). If the queue priority of a 
		Process in a ProcessQueue is changed, the ProcessQueue is updated.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class ProcessQueue {
		// Interface
	public:
		ProcessQueue();
		virtual ~ProcessQueue();

		/* \name Queue access and information methods
		 * @{
		 */
		
		/// get pointer to the first process in the queue
		Process* getTop() const;
		/// get a reference to the list object representing the queue
		const std::list<Process*>& getList() const;
		/// check if the queue is empty
		bool isEmpty() const;
		/// get the current queue length, i.e. number of stored processes
		unsigned int getLength() const;
		/**
		 * \brief
		 * 		get the queue position of a specific process
		 * \param p
		 * 		Process pointer to be found in queue
		 * \return
		 *		Position of \p p, or 0 when not found 
		 * \note
		 * 		The first queue position is \c 1. \c 0 is reserved for errors.
		 */
		const unsigned int getPosition( Process* p ) const;
		//@}
		
		/* \name Queue manipulation methods
		 * @{
		 */
		/// remove the first Process from the queue
		virtual void popQueue();
		/// remove a specific Process \p p from the queue
		virtual void remove(Process* p);
		/// add Process \p p to the queue, considering order and priority
		virtual void inSort(Process* p,  bool fifo = true);
		//@}
		
	protected:
		/// list of Process pointers representing the queue
		std::list<Process*> l;
		/// size counter for efficiency to avoid calling list::size()
		unsigned int size;
	};

	/**
		\brief awakes all Process objects in queue \p q
	*/
	void awakeAll(ProcessQueue* q);

	/**
	 	\brief awake first Process object in queue \p q
	*/
	void awakeFirst(ProcessQueue* q);

	/**
		\brief awake next Process to \p p in queue \p q
	*/
	void awakeNext(ProcessQueue* q, Process* p);
}

#endif

