//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Simulation.cpp

	\author Ralf Gerstenberger
	<!-- [\author <author>]* -->

	\date created at 2002/02/22

	\brief Implementation of classes in Simulation.h

	\sa Simulation.h

	<!-- [detailed description] -->



	\since 1.0
*/

#include <odemx/base/Event.h>
#include <odemx/base/Process.h>
#include <odemx/base/Simulation.h>
#include <odemx/util/ErrorHandling.h>

using namespace odemx;

Simulation::Simulation(Label l, SimulationObserver* o)
:	ExecutionList(o),
	CoroutineContext(l, o),
	Observable<SimulationObserver>(o),
	isInitialized(false),
	isStopped(false),
	executingEvent(false),
	simMode(UNLIMITED),
	startTime(0),
	endTime(0),
	now(0),
	current(0),
	currentSched(0)
	{
		_obsForEach(SimulationObserver, Create(this));
	}

Simulation::Simulation(Label l, SimTime startAt, SimulationObserver* o)
:	ExecutionList(o),
	CoroutineContext(l, o),
	Observable<SimulationObserver>(o),
	isInitialized(false),
	isStopped(false),
	executingEvent(false),
	simMode(UNLIMITED),
	startTime(startAt),
	endTime(startAt),
	now(startAt),
	current(0),
	currentSched(0)
	{
		_obsForEach(SimulationObserver, Create(this));
	}

Simulation::~Simulation()
	{
		_obsForEach(SimulationObserver, Destroy(this));
	}

void Simulation::run() {
	// trace
	getTrace()->mark(this, markRun);

	// observer
	_obsForEach(SimulationObserver, Run(this));

	init();
	simMode = UNLIMITED;
	compSimulation(false);
}

void Simulation::step() {
	// trace
	getTrace()->mark(this, markStep);

	// observer
	_obsForEach(SimulationObserver, Step(this));

	init();
	simMode = STEPPING;
	compSimulation(false);
}

void Simulation::step( unsigned int stepCount )
{
	for ( unsigned int i = 0; i < stepCount; ++i )
	{
		step();
	}
}

void Simulation::runUntil(SimTime t) {
	// trace
	getTrace()->mark(this, markRunUntil);

	// observer
	_obsForEach(SimulationObserver, RunUntil(this, t));

	init();
	simMode = UNTILTIME;
	endTime = t;
	compSimulation(false);
}

void Simulation::exitSimulation() {
	// trace
	getTrace()->mark(this, markExitCall);

	// observer
	_obsForEach(SimulationObserver, ExitSimulation(this));

	// setting isInitialized may not be formally correct, but it will
	// effect that isFinished() returns true, which makes sense
	// after calling exit on a simulation
	isInitialized = true;
	isStopped = true;
	switchTo();
}

bool Simulation::isFinished() {
	return isInitialized && (isStopped || ExecutionList::isEmpty());
}


void Simulation::init() {
	if (!isInitialized) {
		// trace
		getTrace()->mark(this, markInit);

		// observer
		_obsForEach(SimulationObserver, Initialization(this));

		initSimulation();
		isInitialized=true;
	}
}


// execute next Sched object
void Simulation::exec() {

	Sched* next = ExecutionList::getNextSched();

	// check time consistency
	if ( next -> getExecutionTime() < now ) {
		// Error: ExecutionList corrupted
		error( this,  "exec(); ExecutionList corrupted: invalid execution time" );

		printExecutionList();

		fatalError( this,  "exec(); ExecutionList corrupted: ending simulation", -1 );
	}

	// trace
	getTrace()->mark( this, markTime, next -> getExecutionTime(), now );

	// observer
	_obsAForEach( SimulationObserver, Time, now, next -> getExecutionTime() );

	// update current SimTime
	now = next -> getExecutionTime();

	// execute according to type of scheduled object
	switch ( next -> getSchedType() ) {

		case Sched::PROCESS:

			// trace
			getTrace() -> mark( this, markExecProc, next );

			// observer
			_obsForEach( SimulationObserver, ExecuteProcess( this, static_cast< Process* >(next) ) );

			next -> execute();
			break;

		case Sched::EVENT:

			// trace
			getTrace() -> mark( this, markExecEvent, next );

			// observer
			_obsForEach( SimulationObserver, ExecuteEvent( this, static_cast< Event* >(next) ) );

			// ensure complete execution, even when scheduling processes
			executingEvent = true;
			next -> execute();
			executingEvent = false;

			break;

		default:
			error( this,  "Simulation::exec(): cannot determine SchedType" );
	}
}


Process* Simulation::getCurrentProcess() {
	return current;
}

Sched* Simulation::getCurrentSched() {
	return currentSched;
}

std::list<Process*>& Simulation::getCreatedProcesses() {
	return created;
}

std::list<Process*>& Simulation::getRunnableProcesses() {
	return runnable;
}

std::list<Process*>& Simulation::getIdleProcesses() {
	return idle;
}

std::list<Process*>& Simulation::getTerminatedProcesses() {
	return terminated;
}

void Simulation::setCurrentSched( Sched* s ) {
	currentSched = s;
}


void Simulation::setProcessAsCurrent(Process* p) {

	// trace
	getTrace()->mark(this, markCurrProc, p);

	// observer
	_obsAForEach(SimulationObserver, CurrentProcess, current, p);

	// erase process from previous state list
	removeProcessFromList(p);

	current = p;
	currentSched = p;
}

void Simulation::setProcessAsCreated(Process* p) {

	_obsForEach(SimulationObserver, ChangeProcessList(this, p, CREATED));

	// insert p
	p -> simListIter = created.insert( created.end(), p );
}

void Simulation::setProcessAsRunnable(Process* p) {

	// erase process from previous state list
	removeProcessFromList(p);

	_obsForEach(SimulationObserver, ChangeProcessList(this, p, RUNNABLE));

	// insert p
	p -> simListIter = runnable.insert( runnable.end(), p );
}

void Simulation::setProcessAsIdle(Process* p) {

	// erase process from previous state list
	removeProcessFromList(p);

	_obsForEach(SimulationObserver, ChangeProcessList(this, p, IDLE));

	// insert p
	p -> simListIter = idle.insert( idle.end(), p );
}

void Simulation::setProcessAsTerminated(Process* p) {

	// erase process from previous state list
	removeProcessFromList(p);

	_obsForEach(SimulationObserver, ChangeProcessList(this, p, TERMINATED));

	// insert p
	p -> simListIter = terminated.insert( terminated.end(), p );
}

void Simulation::removeProcessFromList(Process* p) {

	switch ( p -> processState ) {
	case Process::CREATED:
		created.erase( p -> simListIter );
		break;
	case Process::RUNNABLE:
		runnable.erase( p -> simListIter );
		break;
	case Process::IDLE:
		idle.erase( p -> simListIter );
		break;
	// should never happen, no reactivation after termination
	case Process::TERMINATED:
		terminated.erase( p -> simListIter );
		break;
	// Process::CURRENT, no list needed
	default:
		break;
	}
}


// compute simulation according to simulation mode
void Simulation::compSimulation(bool inSimulation/* = true*/) {

	if ( executingEvent )
		return;

	// execute all events scheduled before the next process object
	while( isInitialized &&
		   !ExecutionList::isEmpty() &&
		   getNextSched() -> getSchedType() == Sched::EVENT ) {

		switch( simMode ) {

			case UNLIMITED:

				if ( !isStopped )
					exec();

				break;

			case STEPPING:

				if ( !isStopped ) {

					if ( inSimulation ) {
						current = 0;
						currentSched = 0;
						switchTo();	// called from process scheduling fct, jump to main()
									// remember this position for the current process
					}
					else
						exec();		// coming from main(), execute Event
					return;			// do nothing else after returning
				}
				break;

			case UNTILTIME:
				if ( !isStopped && getTime() <= endTime ) {

					// check next Sched to avoid Event firing
					if ( getNextSched() -> getExecutionTime() > endTime ) {
						if ( inSimulation ) {
							current = 0;
							currentSched = 0;
							switchTo();	// called from Process scheduling
						}
						else
							return;		// return to main()
					}
					else {
						exec(); // execTime <= endTime
					}
				}
			}
		} // done with events

	// execute process object if there is one scheduled
	if (!inSimulation) {
		// if called from main program [run(), step(), runUntil()]:
		// start next Process object
		switch(simMode) {
		case UNLIMITED:
		case STEPPING:
			if (!ExecutionList::isEmpty() && !isStopped)
				exec();
			break;
		case UNTILTIME:
			if (!ExecutionList::isEmpty() && getTime() <= endTime && !isStopped) {

				// check next Sched to avoid Event firing
				if ( getNextSched() -> getExecutionTime() > endTime ) {
					current = 0;
					currentSched = 0;
					break; // return to main
				}
				else {
					exec();
				}
			}
		}
	} else {
		if (!isInitialized)
			return;

		// if called from Process: handle different simulation modes
		switch(simMode) {
		case UNLIMITED:
			// continue with next Process in Execution list
			if (!ExecutionList::isEmpty() && !isStopped)
				exec();
			else {
				current = 0;
				currentSched = 0;
				switchTo(); // return to main program
			}
			break;
		case STEPPING:
			// return to main program
			current = 0;
			currentSched = 0;
			switchTo();
			break;
		case UNTILTIME:
			// continue if possible
			if (!ExecutionList::isEmpty() && getTime() <= endTime && !isStopped) {

				// check next Sched to avoid Event firing
				if ( getNextSched() -> getExecutionTime() > endTime ) {
					current = 0;
					currentSched = 0;
					switchTo(); // return to main
				}
				else {
					exec();
				}
			}
			else {
				current = 0;
				currentSched = 0;
				switchTo(); // return to main program
			}
			break;
		}
	}
}


// ============== DEFAULT SIMULATION / MARK INITIALIZATION ====================


/**
	\return	pointer to the DefaultSimulation
	\relates DefaultSimulation

	default Simulation provided for convenience
*/
DefaultSimulation* odemx::getDefaultSimulation()
{
	static DefaultSimulation defaultSim;
	return &defaultSim;
}

void odemx::simulate()
{
	getDefaultSimulation() -> run();
}

void odemx::simulateStep()
{
	getDefaultSimulation() -> step();
}

void odemx::simulateUntil( SimTime t )
{
	getDefaultSimulation() -> runUntil( t );
}


const MarkTypeId Simulation::baseMarkId = MarkType::getNewBaseMarkId( typeid(Simulation) );

const MarkType Simulation::markInit( "init", baseMarkId+1, typeid(Simulation)); ///< Simulation is initialized, has executed user-defined initSimulation()
const MarkType Simulation::markExitCall( "exit call", baseMarkId+2, typeid(Simulation)); ///< Simulation was stopped by call of exitSimulation()
const MarkType Simulation::markRun( "run", baseMarkId+3, typeid(Simulation)); ///< Simulation was started using run(), continues until execution list is empty
const MarkType Simulation::markStep( "step", baseMarkId+4, typeid(Simulation)); ///< Simulation was started using step(), only executes one entry in the execution list and returns to main
const MarkType Simulation::markRunUntil( "run until", baseMarkId+5, typeid(Simulation)); ///< Simulation was started using runUntil(), continues through execution list until stop time is reached
const MarkType Simulation::markExecProc( "execute process", baseMarkId+6, typeid(Simulation)); ///< Simulation has started the execution or continuation of a Process
const MarkType Simulation::markExecEvent( "execute event", baseMarkId+7, typeid(Simulation)); ///< Simulation has started an Event object
const MarkType Simulation::markCurrProc( "current process", baseMarkId+8, typeid(Simulation)); ///< Displays the currently active Process object
const MarkType Simulation::markTime( "time", baseMarkId+9, typeid(Simulation)); ///< Displays the current simulation time when scheduled objects are executed
