//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file HtmlGenerator.cpp

	\author Ronald Kluth

	\date created at 2007/11/01

	\brief Implementation of functions in HtmlGenerator.h

	\sa HtmlGenerator.h

	<!-- [detailed description] -->

	\since 2.0
*/

#include <odemx/util/ErrorHandling.h>
#include <odemx/display/HtmlGenerator.h>
#include <odemx/util/Version.h>

#if defined(__GNUC__)
	/* GNU Compiler, means make generated ODEMxPath.h */
	#include <odemx/util/ODEMxPath.h>
#endif

using namespace odemx;
using namespace std;

HtmlGenerator::HtmlGenerator():
	env( getDefaultSimulation() )
	{}

HtmlGenerator::HtmlGenerator( Simulation* sim ):
	env( sim )
	{}

void HtmlGenerator::generateHtml( const string& fileName, const string& filter ) {

	// build file name
	ostringstream file;
	file << fileName << ".html";

	htmlStream = new ofstream( file.str().c_str() );

	if ( htmlStream==0 || !(*htmlStream) ) {
		// Error: cannot open file
		error( "HtmlGenerator::generateHtml(): cannot open file, sending output to stdout" );
		htmlStream = &cout;
	}

	// build the file
	htmlHead( fileName );
	htmlBodyCaption( filter );
	htmlBodyNavigation();
	htmlBodyFoot();

	// delete old file pointer
	reinterpret_cast<ofstream*>( htmlStream ) -> close();
	delete htmlStream;
}



void HtmlGenerator::htmlHead( const string& xmlFileName ) {

	*htmlStream
	<< "<!DOCTYPE html" << endl
	<< "PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"" << endl
	<< "\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">" << endl
	<< "<html xmlns='http://www.w3.org/1999/xhtml'>" << endl
	<< "<head>" << endl
	<< "<meta http-equiv='Content-Script-Type' content='text/javascript' />" << endl
	<< "<link rel='stylesheet' type='text/css' href='";

#if defined(_MSC_VER)

	// not using 'make' means we have to get the library location
	// using other means...

	// get full path of this file, includes the location of display dir
	string fullPath(__FILE__);
	// cut off the substring containing this file's name
	int end = fullPath.find( "htmlgenerator.cpp" );
	string odemxPath = fullPath.substr(0, end);

	// output the directory
	*htmlStream
	<< "file://localhost/"
	<< odemxPath + "stylesheet\\XmlTrace.css' />" << endl;
#else

	*htmlStream << odemxPath + "/display/stylesheet/XmlTrace.css' />" << endl;

#endif

	*htmlStream
	<< "<script type='text/javascript'>" << endl
	<< "<!--" << endl
	<< "var xmlListFile = '"
	<< xmlFileName << "_files.xml';" << endl
	<< "var markInfoFile = '"
	<< xmlFileName << "_markinfo.xml';" << endl
	<< "var xmlTraceIncluded = false;" << endl
	<< "//-->" << endl
	<< "</script>" << endl
	<< "<script type='text/javascript' src='";

#if defined(_MSC_VER)

	*htmlStream
	<< "file://localhost/"
	<< odemxPath + "javascript\\XmlTrace.js'></script>" << endl;

#else

	*htmlStream
	<< odemxPath + "/display/javascript/XmlTrace.js'></script>" << endl;

#endif

	*htmlStream
	<< "<script type='text/javascript'>" << endl
	<< "<!--" << endl
	<< "if ( !xmlTraceIncluded ) { " << endl
	<< "	alert( 'JavaScript file XmlTrace.js not found!' );" << endl
	<< "}" << endl
	<< "//-->" << endl
	<< "</script>" << endl
	<< "<title>ODEMx XML Trace: " << env -> getLabel() << "</title>" << endl
	<< "</head>" << endl;
}

void HtmlGenerator::htmlBodyCaption( const string& filter ) {

	*htmlStream
	<< "<body onload='init();'>" << endl << endl
	<< "<div id='caption'>" << endl << endl
	<< "<h2>ODEMx XML Trace</h2>" << endl
	<< "<div>" << env -> getLabel() << "</div>" << endl
	<< "<div>ODEMx Version: " << Version::getString() << "</div>" << endl
	<< "<div>" << filter << "</div>" << endl
	<< "</div>" << endl;
}

void HtmlGenerator::htmlBodyNavigation() {

	*htmlStream
	<< "<div id='nav'>" << endl
	<< "	<div id='filterNav'>" << endl
	<< "	Filter &gt;&gt;" << endl
	<< "		<div id='markNav' class='subNav'>" << endl
	<< "			<div id='mtn' class='filterToggle'>" << endl
	<< "			Mark Type Name" << endl
	<< "				<div id='mtnDisplay' class='filterChoice'></div>" << endl
	<< "			</div>" << endl
	<< "			<div id='mti' class='filterToggle'>" << endl
	<< "			Mark Type Id" << endl
	<< "				<div id='mtiDisplay' class='filterChoice'></div>" << endl
	<< "			</div>" << endl
	<< "			<div id='mts' class='filterToggle'>" << endl
	<< "			Mark Type Scope" << endl
	<< "				<div id='mtsDisplay' class='filterChoice'></div>" << endl
	<< "			</div>" << endl
	<< "			<div id='sl' class='filterToggle'>" << endl
	<< "			Sender Label" << endl
	<< "				<div id='slDisplay' class='filterChoice'></div>" << endl
	<< "			</div>" << endl
	<< "			<div id='st' class='filterToggle'>" << endl
	<< "			Sender Type" << endl
	<< "				<div id='stDisplay' class='filterChoice'></div>" << endl
	<< "			</div>" << endl
	<< "			<div id='filterReset'>Reset</div>" << endl
	<< "		</div>" << endl
	<< "	</div>" << endl
	<< "	<div id='logToggle'>Toggle JS Log</div>" << endl
	<< "</div>" << endl;
}

void HtmlGenerator::htmlBodyFoot() {

	*htmlStream
	<< "<div id='timeNav'> Simulation Period <div id='periodSelect'></div> </div>" << endl
	<< "<div id='log'></div>" << endl
	<< "<div id='info'></div>" << endl
	<< "<div id='content'></div>" << endl
	<< "</body>" << endl
	<< "</html>" << endl;
}

