//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file TimeFormat.cpp

	\author Ronald Kluth

	\date created at 2008/03/14

	\brief Definition of odemx::Time and output Formats

	\sa TimeFormat.h

	<!-- [detailed description] -->

	\since 2.1
*/

#if defined(_MSC_VER)
#pragma warning(disable : 4244) // conversion of SimTime to double possible data loss
#endif

#include <odemx/display/TimeFormat.h>
#include <odemx/util/ErrorHandling.h>
#include <cmath>
#include <cassert>

using namespace odemx;
using namespace std;

// extern definition of default time string format
odemx::Time::DefaultFormat odemx::Time::defaultFormat;

// convert SimTime to string using a given time format
std::string odemx::simTimeToString( SimTime t, Time::Format& format ) {
	return format.timeToString( t );
}

// compute number of day of a month, considering leap years
unsigned short Time::daysOfMonth( unsigned short month, unsigned short y ) {

	bool leapYear = false;

	if( y % 4 == 0 ) leapYear = true;
	if( y % 100 == 0 ) leapYear = false;
	if( y % 400 == 0 ) leapYear = true;

    switch( month ) {
        case 4:
        case 6:
        case 9:
        case 11:
            return 30;

        case 2:
			if( leapYear )
				return 29;
			else
				return 28;

        default:
            return 31;
    }
}

void Time::Format::resetTimeValues() {
	simTime = static_cast< SimTime >(0);
	year = 0;
	month = 0;
	day = 0;
	hour = 0;
	min = 0;
	sec = 0.0;
}

void Time::Format::computeTimeValues( SimTime time ) {

	resetTimeValues();

	if( getTimeBase().unit == Time::no_unit )
	{
		error( "Time::Format::computeTimeValues(): cannot compute time, no base unit given" );
		return;
	}

	// store the SimTime in minutes for easy computation
	double simMinutes = 0.0;
	// store the daytime offset
	double dayTimeOffset = 0.0;

	// convert SimTime to minutes for convenience
	// by evaluating the given Time::unit of the Format definition
	switch( getTimeBase().unit )
	{
	case Time::milliseconds:
		simMinutes = time / double(60000);
		dayTimeOffset = getTimeBase().offset / double(60000);
		break;

	case Time::seconds:
		simMinutes = time / double(60);
		dayTimeOffset = getTimeBase().offset / double(60);
		break;

	case Time::minutes:
		simMinutes = time;
		dayTimeOffset = getTimeBase().offset;
		break;

	case Time::hours:
		simMinutes = time * 60;
		dayTimeOffset = getTimeBase().offset * 60;
		break;

	case Time::no_unit:
		assert(false);
	}

	// compute number of days from minutes
	day = (unsigned short) floor( simMinutes / 1440.0 ); // 24h * 60' = 1440'
	// subtract days from minutes to get the day time
	simMinutes = simMinutes - ( day * 1440.0 );

	// add time base data representing the starting point of the simulation
	year += getTimeBase().year;
	month += getTimeBase().month;
	day += getTimeBase().day;
	simMinutes += dayTimeOffset;

	// increase number of days if offset addition put minutes over 1440
	while( simMinutes > 1440.0 ) {
		++day;
		simMinutes -= 1440.0;
	}

	// determine month and year
	while( day > Time::daysOfMonth( month, year ) ) {

		day -= Time::daysOfMonth(month, year);
		++month;

		if( month > 12 ) {
			month = 1;
			++year;
		}
	}

	// compute the day time in hours, minutes, seconds and milliseconds
	hour = (unsigned short) floor( simMinutes/60.0 );
	simMinutes -= hour * 60.0;

	// get minutes integer value
	min = (unsigned short) floor( simMinutes );

	// get seconds with milliseconds as double value
	sec = ( simMinutes - min ) * 60.0;
}

const string Time::German::makeString() const {
	std::ostringstream out;

	out.fill('0');

	if( (year - 100) > 0 )
		out << setw(4) << year;
	else
		out << setw(2) << year;

	out << '-' << setw(2) << month << '-';
	out << setw(2) << day;

	out << " / ";
	out << setw(2) << hour << ':';
	out << setw(2) << min;

	if( base.unit == Time::milliseconds )
		out << ':' << setw(6) << fixed << setprecision(3) << sec;
	else
		out << ':' << setw(2) << fixed << setprecision(0) << sec;

	out.fill(' ');

	return out.str();
}

const string Time::ISO8601::makeString() const {
	std::ostringstream out;

	out.fill('0');

	out << setw(4) << year << '-';
	out << setw(2) << month << '-';
	out << setw(2) << day;
	out << 'T';
	out << setw(2) << hour << ':';
	out << setw(2) << min;

	// forcing fixed value output rounds up or down
	if( precision == Time::precision_sec ) {
		out << ':' << setw(2) << fixed << setprecision(0) << sec;
	}

	// print seconds with milliseconds
	if( precision == Time::precision_milli ) {
		out << ':' << setw(6) << fixed << setprecision(3) << sec;
	}

	// print UTC time zone
	if( zone.hour || zone.minute ) {

		out << ( zone.hour > 0 ? '+' : '-' );
		out << setw(2) << std::abs( static_cast< float >( zone.hour ) ) << ':';
		out << setw(2) << zone.minute;
	}
	else {
		out << 'Z'; // represents UTC-0
	}

	return out.str();
}
