//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file TimeFormat.h

	\author Ronald Kluth

	\date created at 2008/03/14

	\brief Declaration of odemx::Time, Unit, Precision, and various output Formats

	\sa TimeFormat.cpp

	<!-- [detailed description] -->

	\since 2.1
*/

#ifndef ODEMX_TIMEFORMAT_INCLUDED
#define ODEMX_TIMEFORMAT_INCLUDED

#include <iostream>
#include <iomanip>
#include <sstream>
#include <odemx/util/SimTime.h>
#include <odemx/util/ErrorHandling.h>

namespace odemx {

	/// grouping namespace for time unit, time base, and format definitions
	namespace Time {

		/// SimTime always represents one of these time units
		enum Unit {
			no_unit,
			milliseconds,
			seconds,
			minutes,
			hours
		};

		/// Precision of the printed time output format
		enum Precision {
			precision_min = 0, ///< date, hours and minutes
			precision_sec = 1, ///< date, hours, minutes and seconds
			precision_milli = 2 ///< date, hours, minutes, seconds with milliseconds
		};

		/** \class Base

			\ingroup display

			\author Ronald Kluth

			\brief describes the time base in day, month, year, day time offset and unit

			\since 2.1
		*/
		class Base {
		public:

			/// Construction
			Base( unsigned short y, unsigned short m, unsigned short d,
				  SimTime dayTimeOffset, Time::Unit u ):
				year( y ),
				month( m ),
				day( d ),
				offset( dayTimeOffset ),
				unit( u )
				{}

			unsigned short year; ///< time base year
			unsigned short month; ///< time base month
			unsigned short day; ///< time base day
			SimTime offset; ///< time offset on the given day, must respect unit
			Time::Unit unit; ///< time base unit
		};

		/** \class Zone

			\ingroup display

			\author Ronald Kluth

			\brief describes the UTC time zone offset in hours and minutes,
			to be used in time format output

			\since 2.1
		*/
		class Zone {
		public:

			/// Construction
			Zone( short h, unsigned short m )
				{
					if ( h || m ) {

						if( h < -12 || h > 14 )
							error( "Time::Zone(): hour value range is -12 through 14" );

						if( ! (m == 0 || m == 15 || m == 30 || m == 45) )
							error( "Time::Zone(): minute requires values 0, 15, 30, or 45" );
					}
					hour = h;
					minute = m;
				}

			short hour; ///< time zone hour offset from UTC
			unsigned short minute; ///< time zone minute offset from UTC
		};

		/** \class Format

			\ingroup display

			\author Ronald Kluth

			\brief abstract base class for formatted simulation time output

			\since 2.1
		*/
		class Format {
		public:

			/// default constructor for uninitialized TimeFormat objects
			Format():
				base( 0, 0, 0, static_cast< SimTime >(0), Time::no_unit )
				{
					resetTimeValues();
				}

			/// constructor for given time base and unit
			Format( Time::Base tb ):
				base( tb )
				{
					resetTimeValues();
				}

			/// Destruction
			virtual ~Format() {}

			/// set the time base (date, time of day, unit) for the formatted output
			void setTimeBase( unsigned short y, unsigned short m, unsigned short d,
							  SimTime dayTimeOffset, Time::Unit u ) {
				base.year = y;
				base.month = m;
				base.day = d;
				base.offset = dayTimeOffset;
				base.unit = u;
			}

			/// overload
			void setTimeBase( Time::Base& tb ) { base = tb; }

			/// get a reference to the format's time base
			const Time::Base& getTimeBase() const { return base; }

			/** \brief transform SimTime into a formatted string
			 *
			 *  This function is used by all time formats to transform
			 *  SimTime double values into human-readable time formats.
			 *
			 *  It uses the virtual function compute time values to
			 *  determine the date and time according to the time unit.
			 *
			 * 	Finally, it calls the pure virtual function makeString(),
			 *  which must be defined by all subclasses to produce the
			 *  corresponding formatted string.
			 */
			const std::string timeToString( SimTime t ) {
				computeTimeValues( t );
				return makeString();
			}

		protected:

			/// computes a real date and clock time from SimTime, stores values in member data
			virtual void computeTimeValues( SimTime t );

			/// produces the formatted time string using the previously computed values
			virtual const std::string makeString() const = 0;

			/// time base, the starting point of the simulation time
			Time::Base base;

			// values to be set by computeTimeValues()
			SimTime simTime; ///< the SimTime to be converted considering base
			unsigned short year; ///< computed year
			unsigned short month; ///< computed month
			unsigned short day; ///< computed day
			unsigned short hour; ///< computed hour
			unsigned short min; ///< computed minutes
			double sec; ///< computed seconds and milliseconds

		private:
			/// reset member data, used for initialization and reset
			void resetTimeValues();
		};

		// function to get the days of a month, considering the year
		unsigned short daysOfMonth( unsigned short m, unsigned short y );

		/** \class DefaultFormat

			\ingroup display

			\author Ronald Kluth

			\brief simple default implementation for Time::Form0.0at

			\since 2.1
		*/
		class DefaultFormat: public Format {
		public:
			/// Constructor
			DefaultFormat()
			:	Format( Time::Base( 0, 0, 0, static_cast< SimTime >(0), Time::minutes ) )
			{}

			/// redefines computation function to avoid overhead
			virtual void computeTimeValues( SimTime t ) { simTime = t; }

			/// print the normal SimTime without computing anything
			virtual const std::string makeString() const {

				std::ostringstream stream;
				stream << simTime;

				return stream.str();
			}
		};

		/// default time format is the string representation of SimTime
		extern DefaultFormat defaultFormat;

		/** \class German

			\ingroup display

			\author Ronald Kluth

			\brief implementation of German time format that uses the
			predefined computation function from Format

			\since 2.1
		*/
		class German: public Format {
		public:

			/// Default constructor
			German(): Format() {}

			/// Construction with given time base
			German( Time::Base b ): Format( b ) {}

			// uses predefined computation function from Format()

			/// defines the output format of the time string
		    virtual const std::string makeString() const;
		};

		/** \class ISO8601

			\ingroup display

			\author Ronald Kluth

			\brief implementation of ISO-8601 time format

			This implementation features a subset, as proposed
			by the W3C, of the whole ISO-8601 standard for
			displaying international dates and time.

			It enables display of different time precisions
			ranging from minutes over seconds up to milliseconds.

			\since 2.1
		*/
		class ISO8601: public Format {
		public:

			/// Construction with time base, precision and time zone
			ISO8601( Time::Base b, Time::Precision p = precision_min,
					 Time::Zone z = Zone( 0, 0 ) ):
				Format( b ),
				precision( p ),
				zone( z )
				{}

			// uses predefined computation function from Format()

			/// defines the output format of the time string
		    virtual const std::string makeString() const;

		    /// change precision value
		    void setPrecision( Time::Precision tp ) { precision = tp; }

		    /// change time zone
		    void setZone( short hour, short minute ) {
		    	zone = Zone( hour, minute );
		    }

		    /// get the set precision
		    const Time::Precision getPrecision() const { return precision; }
		    /// get the set time zone
		    const Time::Zone getZone() const { return zone; }

		private:
			Precision precision; ///< precision to be displayed
			Zone zone; ///< time zone to be displayed
		};

	} // end of namespace Time


	/// global function to convert any SimTime value to a string of the given format
	std::string simTimeToString( SimTime t, Time::Format& format );
}

#endif /*ODEMX_TIMEFORMAT_INCLUDED*/
