//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2003, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file TraceFilter.cpp

	\author Ralf Gerstenberger

	\date created at 2003/06/10

	\brief {brief description}

	\sa TraceFilter.h

	<!-- [detailed description] -->

	\since 1.0
*/

#if defined(_MSC_VER)
#pragma warning(disable : 4996) // sscanf unsafe warning
#endif

#include <odemx/display/TraceFilter.h>
#include <odemx/util/ErrorHandling.h>

#include <cstdio>
#include <cassert>
#include <sstream>

using namespace odemx;
using namespace std;

TraceFilter::TraceFilter() :
	filterOut( false ),
	isSet( false ),
	patt( "" )
	{
		filter.def = TraceFilter::PASSALL;
	}

TraceFilter::~TraceFilter()
	{}


void TraceFilter::setFilter( const string& f ) {
	if ( f.empty() )
		return;
	else
		patt = f;

	// set default and reset
	filter.def = TraceFilter::PASSALL;
	filter.MarkTypeIds.clear();
	filter.MarkTypeNames.clear();
	filter.MarkTypeScopes.clear();
	filter.SenderLabels.clear();
	filter.SenderTypes.clear();
	filterOut = false;

	// remember that the filter is set
	isSet = true;

	using namespace std;

	string token; // temporary storage
	string::const_iterator i;
	int j;
	enum MODE {ID, MTI, MTN, MTS, SL, ST, DEFALL, DEFNONE, END} parseMode = ID;
	static const string idTab[END]={"", "mti:", "mtn:", "mts:", "sl:", "st:", "all", "none"};

	// scan the given filter pattern and act according to
	// parseMode and the current token
	for ( i = patt.begin(); ; ++i ) {

		// while pattern contains chars, continue tokenizing
		if ( i != patt.end() ) {

			switch( *i ) {
				case ' ':	// white space
				case '\n':
				case '\t':
					if ( parseMode != ID ) // ID mode to determine next parse mode
						token += *i;       // store this char
					continue;
				case ';':
				case ',':
					break; // end of token found
				case ':':  // found a parseMode token
					token += *i;
					if (parseMode==ID) // end of parseMode token found
						break;
					continue;
				default:
					token+=*i; // no special char, append current char to token
					continue;
			}
		}

		// remove whitespace at the beginning and at the end
		if ( !token.empty() && parseMode!=ID ) {
			string::size_type bg = token.find_first_not_of(" \n\t");
			string::size_type eg = token.find_last_not_of(" \n\t");
			token = token.substr(bg, eg+1);
		}

		// handle token according to parseMode
		switch( parseMode ) {
		case ID: // find new parse mode
			// compare token with idTab filter keywords to determine mode
			for( j = MTI; j < END && token != idTab[j]; ++j );

			if ( j == DEFNONE ) {  // found 'none' key

				filter.def = TraceFilter::PASSNONE;
				parseMode = ID;
			}
			else if ( j == DEFALL ) {  // found 'all' key

				filter.def = TraceFilter::PASSALL;
				parseMode = ID;
			}
			else if ( j != END ) {  // found different key

				parseMode = MODE( j ); // set new parse mode
			}
			else { // could not determine mode

				error( "TraceFilter::setFilter(): error in filter pattern, mode expected" );
				return;
			}
			break;
		case MTI: // get MarkTypeId
			{
				unsigned int l;

				// store string-to-unsigned int conversion in l
				if ( sscanf( token.c_str(), "%u", &l ) == 1 ) {

					filter.MarkTypeIds.insert(l);
				}
				else { // error: number expected

					error("TraceFilter::setFilter(): error in filter pattern, number expected");
					return;
				}
			}
			break;
		case MTN: // insert string tokens into filter sets

			filter.MarkTypeNames.insert(token);
			break;
		case MTS:

			filter.MarkTypeScopes.insert(token);
			break;
		case SL:

			filter.SenderLabels.insert(token);
			break;
		case ST:

			filter.SenderTypes.insert(token);
			break;
		default:

			/* MPi unhandled (by Ralf): DEFALL, DEFNONE, END */
			assert(false);
		}

		// done with parsing
		if ( i == patt.end())
			return;

		// done with parsing in one mode, try to determine next mode
		if (*i==';')
			parseMode=ID;

		token.erase();

	} // end of pattern-scanning for-loop
}

void TraceFilter::resetFilter()
{
	filterOut = false ;
	isSet = false;
	patt.clear();

	filter.def = TraceFilter::PASSALL;
	filter.MarkTypeIds.clear();
	filter.MarkTypeNames.clear();
	filter.MarkTypeScopes.clear();
	filter.SenderLabels.clear();
	filter.SenderTypes.clear();
}

bool TraceFilter::foundMatch() {

	filterOut = ( filter.def == TraceFilter::PASSALL );

	// true means: pass-all except matches: if match found, return false
	// false means: pass none, except matches: if match found, return true

	return !filterOut;
}

// helper for the pass() function, checks occurrence of element in set
template< typename SetType, typename ElemType >
bool foundIn( SetType theSet, ElemType element )
{
	return theSet.find( element ) != theSet.end();
}

bool TraceFilter::pass( const MarkType* m, const TraceProducer* sender )
{
	if( ! isSet )
	{
		filterOut = false;
		return true;
	}

	// look for the given values in the filter sets
	if( foundIn( filter.MarkTypeIds, m -> getId() )
		|| foundIn( filter.MarkTypeNames, m -> getName() )
		|| foundIn( filter.MarkTypeScopes, typeToString( m -> getScope() ) )
		|| foundIn( filter.SenderLabels, sender -> getLabel() )
		|| foundIn( filter.SenderTypes, typeToString( sender -> getType() ) ) )
	{
		return foundMatch();
	}

	/* -------- NO MATCH found when arriving here ------- */

	if ( filter.def == TraceFilter::PASSALL )
	{
		// must let this mark pass
		filterOut = false;
	}
	else if ( filter.def == TraceFilter::PASSNONE )
	{
		// must filter out this mark
		filterOut = true;
	}
	else // error
		error( "TraceFilter::pass(): cannot determine filter mode" );

	return !filterOut;
}

void TraceFilter::passNone() {
	isSet = true;
	filter.def = TraceFilter::PASSNONE;
}

void TraceFilter::passAll() {
	isSet = true;
	filter.def = TraceFilter::PASSALL;
}

// helper for the getFilter() print function
template< typename SetType >
void addToStream( ostringstream& stream, SetType& theSet, const std::string& caption )
{
	if ( ! theSet.empty() )
	{
		stream << " | " << caption << ": ";
		bool firstLoop = true;
		for ( typename SetType::const_iterator iter = theSet.begin();
			  iter != theSet.end(); ++iter )
		{
			if( ! firstLoop )
			{
				stream << ", ";
			}
			stream << *iter;
			firstLoop = false;
		}
	}
}

const string TraceFilter::getFilter() const
{
	ostringstream filterStream;

	// print mode
	filterStream << "FilterMode: ";
	if ( filter.def == TraceFilter::PASSALL )
		filterStream << "PASS ALL";
	else
		filterStream << "PASS NONE";

	// add filter information to a stream object
	addToStream( filterStream, filter.MarkTypeIds, "MarkTypeIds" );
	addToStream( filterStream, filter.MarkTypeNames, "MarkTypeNames" );
	addToStream( filterStream, filter.MarkTypeScopes, "MarkTypeScopes" );
	addToStream( filterStream, filter.SenderLabels, "SenderLabels" );
	addToStream( filterStream, filter.SenderTypes, "SenderTypes" );

	// return the filter information as string
	return filterStream.str();
}

std::set<MarkTypeId>& odemx::operator <<( std::set<MarkTypeId>& filterSet,
										  const MarkTypeId& id ) {
	// insert new filter value
	filterSet.insert( id );
	return filterSet;
}

std::set<std::string>& odemx::operator <<( std::set<std::string>& filterSet,
										   const std::string& str ) {
	// insert new filter value
	filterSet.insert( str );
	return filterSet;
}

std::set<std::string>& odemx::operator <<( std::set<std::string>& filterSet,
										   const std::type_info& type ) {
	// insert new filter value
	filterSet.insert( typeToString( type ) );
	return filterSet;
}

