//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2003, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file TraceFilter.h

	\author Ralf Gerstenberger

	\date created at 2003/06/10

	\brief Filter for trace marks

	\sa TraceFilter.cpp

	<!-- [detailed description] -->

	\since 1.0
*/

#ifndef ODEMX_TRACEFILTER_INCLUDED
#define ODEMX_TRACEFILTER_INCLUDED

#include <odemx/util/Trace.h>

#include <string>
#include <set>

namespace odemx {

	/** \class TraceFilter

		\ingroup display

		\author Ralf Gerstenberger

		\brief Filter for trace marks

		<!-- [\note {notes}]* -->

		<!-- [\sa {references to other classes}]* -->

		TraceFilter provides a crude filter for trace marks. It
		can be used by TraceConsumer implementations to filter
		trace marks.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class TraceFilter {
	public:
		/// Construction
		TraceFilter();
		/// Destruction
		virtual ~TraceFilter();

		/**
			\name output filter

			TraceFilter provides a crude filter to reduce
			the generated output. The filter can be set
			before the trace starts or in pause mode through
			setFilter().

			@{
		*/
		/**
			TraceFilter provides a filter to remove marks from the trace.
			This filter is set with the string \p f. The string is read
			from left to right. If the keywords \c none or \c all are
			found the filter policy is set. The last occurrence of one of
			these keywords is used. If none of these keywords is found
			\c all is used as pre-set. The \c none policy blocks all marks
			except those specified in the string \p f. The \c
			all policy passes all marks except those specified in the
			string \p f.
			A mark can be specified by its MarkType or its sender. The
			MarkType consists of the MarkTypeId, the MarkTypeName and the
			scope of the mark. Each of these fields can be used to specify
			Marks. The keyword \c mti: is used to specify MarkTypeIds:
			\code
		mti:1
		mti: 345; mti:1334;
		mti: 15, 2354, 2;
			\endcode
			The keyword \c mtn: can be used to specify MarktTypeNames:
			\code
		mtn:begin
		mtn:end , activate; mtn: destroy;
			\endcode
			The keyword \c mts: is used to specify a scope:
			\code
		mts:odemx::Process
		mts:odemx::Res, odemx::Bin; mts: odemx::WaitQ;
			\endcode
			The sender can be specified by its label or its type. The
			keyword \c sl: is used to specify the label:
			\code
		sl:MyWaitq, Car
		sl: Client 2, Client 3; sl: Servant;
			\endcode
			The keyword \c st: specifies the sender type:
			\code
		st:ServantType, Bin
		st:MyTestProcess1; st:Storage
			\endcode
			A mark is specified if at least one of the MarkType fields or its
			sender label or its sender type is specified.

			\note The filter can only be set in pause mode or
			before trace starts.
		*/

		enum FilterMode {PASSALL, PASSNONE};

		/**
		 * @brief set filter by providing a command string
		 * @note The string representation used for MarkTypeScope and
		 * SenderType should be obtained using \c odemx::typeToString().
		 */
		virtual void setFilter(const std::string& f);

		/// Get filter
		const std::string getFilter() const;

		/// reset filter to default values
		void resetFilter();

		/// Only filter out the trace marks specified in the filter
		void passAll();

		/// Filter out all trace marks except the ones specified
		void passNone();

		/// Filter by MarkType ID
		std::set<MarkTypeId>& filterMarkById() {
			// remember that the filter is set
			isSet = true;
			return filter.MarkTypeIds;
		}

		/// Filter by MarkType name
		std::set<std::string>& filterMarkByName() {
			// remember that the filter is set
			isSet = true;
			return filter.MarkTypeNames;
		}

		/// Filter by MarkType scope
		std::set<std::string>& filterMarkByScope() {
			// remember that the filter is set
			isSet = true;
			return filter.MarkTypeScopes;
		}

		/// Filter by sender label
		std::set<std::string>& filterSenderByLabel() {
			// remember that the filter is set
			isSet = true;
			return filter.SenderLabels;
		}

		/// Filter by sender class type
		std::set<std::string>& filterSenderByType() {
			// remember that the filter is set
			isSet = true;
			return filter.SenderTypes;
		}
		//@}

	protected:
		/// Filter mark
		bool pass(const MarkType* m, const TraceProducer* sender);

		/// Last filter result
		bool hasPassed() const {return !filterOut;}

		/// sets filterOut according to policy and returns its negation
		bool foundMatch();

	private:
		bool filterOut; ///< \c true if last mark did not pass
		bool isSet;		///< \c true if object is set to filter input
		std::string patt; ///< filter configuration string

		/// Filter
		struct Filter {
			FilterMode def;
			std::set<MarkTypeId> MarkTypeIds;
			std::set<std::string> MarkTypeNames;
			std::set<std::string> MarkTypeScopes;
			std::set<std::string> SenderLabels;
			std::set<std::string> SenderTypes;
		} filter;

	};

	std::set<MarkTypeId>& operator <<( std::set<MarkTypeId>& filterSet, const MarkTypeId& id );
	std::set<std::string>& operator <<( std::set<std::string>& filterSet, const std::string& str );
	std::set<std::string>& operator <<( std::set<std::string>& filterSet, const std::type_info& type );
}

#endif
