//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file XmlReport.cpp

	\author Ronald Kluth

	\date created at 2008/03/09

	\brief Implementation of odemx::XmlReport

	\sa XmlReport.h

	<!-- [detailed description] -->

	\since 2.1
*/

#include <odemx/display/XmlReport.h>
#include <odemx/util/Translate.h>
#include <odemx/util/Version.h>
#include <odemx/util/ErrorHandling.h>

#if defined(__GNUC__)
	/* GNU Compiler, means make generated ODEMxPath.h */
	#include <odemx/util/ODEMxPath.h>
#endif

#include <vector>
#include <fstream>
#include <cassert>

using namespace std;
using namespace odemx;

XmlReport::XmlReport( const string& fileName, Time::Format& f ):
	sim( getDefaultSimulation() ),
	manageOut( true ),
	out( &cout ),
	format( f )
	{
		if ( fileName.empty() )
			manageOut = false;
		else {
			out = new ofstream( fileName.c_str() );
			if (out==0 || !(*out)) {
				// Error: cannot open file
				error("XmlReport cannot open file; sending output to stdout.");
				manageOut = false;
				out = &cout;
			}
		}
	}

XmlReport::XmlReport( Simulation* s, const string& fileName, Time::Format& f ):
	sim(s), manageOut(true), out(&cout), format( f )
	{
		assert(sim!=0);

		if ( fileName.empty() )
			manageOut = false;
		else {
			out = new ofstream( fileName.c_str() );
			if (out==0 || !(*out)) {
				// Error: cannot open file
				error("XmlReport cannot open file; sending output to stdout.");
				manageOut = false;
				out = &cout;
			}
		}
	}

XmlReport::XmlReport( Simulation* s, std::ostream& os, Time::Format& f ) :
	sim(s), manageOut(false), out(&os), format( f )
	{
		assert(sim!=0);

		if (!(*out)) {
			// Error: invalid ostream
			error("XmlReport invalid ostream; sending output to stdout.");
			out = &cout;
		}
	}

XmlReport::XmlReport( std::ostream& os, Time::Format& f ):
	sim( getDefaultSimulation() ),
	manageOut( false ),
	out( &os ),
	format( f )
	{
		assert(sim!=0);

		if (!(*out)) {
			// Error: invalid ostream
			error("XmlReport invalid ostream; sending output to stdout.");
			out = &cout;
		}
	}

XmlReport::~XmlReport() {
	if (manageOut) {
		static_cast<ofstream*>(out)->close();
		delete out;
	}
}

void XmlReport::processTables() {

	Table* tab = 0;
	vector< Table* >::iterator it;

	// write DOCTYPE and start XML tree
	*out
	<< "<?xml version=\"1.0\" encoding=\"utf-8\" ?>" << endl
	<< "<?xml-stylesheet type=\"text/xsl\" href=\"";

#if defined(_MSC_VER)

	// not using 'make' means we have to get the library location
	// using other means...

	// get full path of this file, includes the location of display dir
	string fullPath(__FILE__);
	// cut off the substring containing this file's name
	int end = fullPath.find( "xmlreport.cpp" );
	string odemxPath = fullPath.substr(0, end);

	// output the directory
	*out
	<< "file://localhost/"
	<< odemxPath + "xsl\\XmlReport.xsl\"?>" << endl << endl;

#else

	*out
	<< "file://"
	<< odemxPath + "/display/xsl/XmlReport.xsl\"?>" << endl << endl;

#endif

	*out
	<< "<xmlreport simulation=\"" << translate( sim -> getLabel() )
	<< "\" simtime=\"" << format.timeToString( sim -> getTime() )
	<< "\" odemxversion=\"" << translate( Version::getString() )
	<< "\">" << endl;

	// write tables
	for( it = ts.begin(); it != ts.end(); ++it ) {
		tab = *it;

		if( tab == 0 )
			continue;

		// start table
		*out
		<< "<reporttable label=\"" << translate( tab -> getLabel() )
		<< "\" columns=\"" << tab -> getNumberOfColumns()
		<< "\" lines=\"" << tab -> getNumberOfLines() << "\">" << endl;

		// write column descriptions with label, number, and type
		for( unsigned int col = 0; col < tab -> getNumberOfColumns(); ++col ) {

			*out
			<< "<column number=\"" << col + 1
			<< "\" label=\"" << translate( tab -> getLabelOfColumn( col ) )
			<< "\" type=\"" << Column::typeToString( tab -> getTypeOfColumn( col ) )
			<< "\" />" << endl;
		}

		// write cells to XML
		for ( unsigned int line = 0; line < tab -> getNumberOfLines(); ++line ) {

			// start line
			*out << "<line number=\"" << line + 1 << "\">" << endl;

			for( unsigned int col = 0; col < tab -> getNumberOfColumns(); ++col ) {

				// start cell
				*out << "<cell col=\"" << col + 1 << "\">";

				// write table cell content
				*out << translate( tab -> getSTRING( col, line ) ) ;

				// close cell
				*out << "</cell>" << endl;
			}
			// close line
			*out << "</line>" << endl;
		}
		*out << "</reporttable>" << endl;
	}

	// close XML tree
	*out << "</xmlreport>" << endl;
}
