//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file XmlReport.h

	\author Ronald Kluth

	\date created at 2008/03/09

	\brief Declaration of odemx::XmlReport

	\sa XmlReport.cpp

	<!-- [detailed description] -->

	\since 2.1
*/

#ifndef ODEMX_XMLREPORT_INCLUDED
#define ODEMX_XMLREPORT_INCLUDED

#include <odemx/base/Simulation.h>
#include <odemx/display/TimeFormat.h>
#include <odemx/util/Report.h>

#include <iostream>
#include <string>


namespace odemx {

	/** \class XmlReport

		\ingroup display

		\author Ronald Kluth

		\brief Report class with XML output

		%XmlReport gathers information from report-generating model
		components and generates XML output. This output is written
		to a file or an output stream.

		\note XmlReport requires the stylesheet XmlReport.xsl
		located in odemx/display/xsl/
	*/
	class XmlReport : public Report {
	public:
		/**
			\name Construction
			@{
		*/
		/**
			\param s
				pointer to the Simulation object
			\param filename
				filename of output file

			%XmlReport opens the file \p filename and writes into
			this file. When %XmlReport is deleted the file is closed.
			If %XmlReport fails to open the file the output is sent
			to \c stdout.
		*/
		XmlReport( Simulation* s, const std::string& filename, Time::Format& f = Time::defaultFormat );

		/**
			\brief Overloaded constructor for DefaultSimulation
			\param filename
				filename of output file
		*/
		XmlReport( const std::string& filename, Time::Format& f = Time::defaultFormat );

		/**
			\param s
				pointer to the Simulation object
			\param os
				ostream for output

			%XmlReport writes to the output stream \p os.
		*/
		XmlReport( Simulation* s, std::ostream& os, Time::Format& f = Time::defaultFormat );

		/**
			\brief Overloaded constructor for DefaultSimulation
			\param os
				ostream for output

			%XmlReport writes to the output stream \p os.
		*/
		XmlReport( std::ostream& os, Time::Format& f = Time::defaultFormat );

		//@}

		/**
			\name Destruction
			@{
		*/
		/**
			In case %XmlReport has opened a file it is closed by the
			destructor.
		*/
		virtual ~XmlReport();
		//@}

	protected:
		/**
			\brief processTables implementation

			%XmlReport is derived from the Report class. Report implements
			management of report producer and data transfer. The data
			are provided by registered report producer and transferred as
			tables. %XmlReport implements the processTables() function to
			generate html output from these tables.
		*/
		virtual void processTables();

		// Implementation
	private:
		Simulation* sim; ///< Simulation
		bool manageOut; ///< manage out
		std::ostream* out; ///< outstream
		Time::Format& format; ///< time display format
	};

} // end namespace odemx


#endif /* ODEMX_XMLREPORT_INCLUDED */
