//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file XmlTrace.cpp

	\author Ronald Kluth

	\date created at 2007/07/15

	\brief Implementation of XmlTrace

	\sa XmlTrace.h

	<!-- [detailed description] -->

	<!-- [\since {since when in project}] -->
*/

#include <odemx/display/XmlTrace.h>
#include <odemx/util/ErrorHandling.h>
#include <odemx/util/Translate.h>

#include <fstream>
#include <cassert>
#include <cstring>

using namespace std;
using namespace odemx;

static const Tag tagPartner( "partner" );
static const Tag tagPartner1( "first partner" );
static const Tag tagPartner2( "second partner" );

XmlTrace::XmlTrace( const string& fileName/* = "" */,
					Time::Format& tf /* = odemx::Time::defaultFormat */,
					unsigned int limit/* = 2500 */ ):
	HtmlGenerator(),
	sim( getDefaultSimulation() ),
	manageOut( true ),
	pause( false ),
	active( false ),
	filterOut( false ),
	out( &cout ),
	fileNumber( 0 ),
	timeFormat( tf ),
	oldTime( 0 ),
	timeString( "???" ),
	xmlMarkCount( 0 ),
	xmlMarkLimit( limit ),
	limitReachTime( -1 )
	{
		if ( fileName.empty() ) {
			// output to stdout
			manageOut = false;
		}
		else {
			xmlFileName = fileName;
		}
	}


XmlTrace::XmlTrace( Simulation* s,
					const string& fileName /* = "" */,
					Time::Format& tf /* = odemx::Time::defaultFormat */,
					unsigned int limit/* = 2500 */ ):
	HtmlGenerator( s ),
	sim(s),
	manageOut(true),
	pause(false),
	active(false),
	filterOut(false),
	out(&cout),
	fileNumber( 0 ),
	timeFormat( tf ),
	oldTime( 0 ),
	timeString( "???" ),
	xmlMarkCount( 0 ),
	xmlMarkLimit( limit ),
	limitReachTime( -1 )
	{
		assert(sim!=0);

		if ( fileName.empty() ) {
			// output to stdout
			manageOut = false;
		}
		else {
			xmlFileName = fileName;
		}
	}

XmlTrace::XmlTrace( std::ostream& os,
					Time::Format& tf /* = odemx::Time::defaultFormat */ ):
	HtmlGenerator(),
	sim( getDefaultSimulation() ),
	manageOut( false ),
	pause( false ),
	active( false ),
	filterOut( false ),
	out( &os ),
	fileNumber( 0 ),
	timeFormat( tf ),
	oldTime( 0 ),
	timeString( "???" ),
	xmlMarkCount( 0 ),
	xmlMarkLimit( 0 ),
	limitReachTime( -1 )
	{
		if (!(*out)) {
			// Error: invalid ostream
			error( "XmlTrace(): invalid ostream; sending output to stdout" );
			out = &cout;
		}
	}


XmlTrace::XmlTrace( Simulation* s,
					std::ostream& os,
					Time::Format& tf /* = odemx::Time::defaultFormat */ ):
	HtmlGenerator( s ),
	sim(s),
	manageOut(false),
	pause(false),
	active(false),
	filterOut(false),
	out(&os),
	fileNumber( 0 ),
	timeFormat( tf ),
	oldTime( 0 ),
	timeString( "???" ),
	xmlMarkCount( 0 ),
	xmlMarkLimit( 0 ),
	limitReachTime( -1 )
	{
		assert(sim!=0);

		if (!(*out)) {
			// Error: invalid ostream
			error( "XmlTrace(): invalid ostream; sending output to stdout" );
			out = &cout;
		}
	}

XmlTrace::~XmlTrace()
	{
	}

//
// Simple trace marks are handled by the following methods.
//
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	beginMarkTree(sender, m);
	endMarkTree(comment);
}

// Marks that include information about producer associations:
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, const TraceProducer* partner, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	beginMarkTree(sender, m);
	writeValue(partner, tagPartner);
	endMarkTree(comment);
}

void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, const TraceProducer* firstPartner, const TraceProducer* secondPartner, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	beginMarkTree(sender, m);
	writeValue(firstPartner, tagPartner1);
	writeValue(secondPartner, tagPartner2);
	endMarkTree(comment);
}

// Marks that include information about producer attributes:
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, bool newValue, bool oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, char newValue, char oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, short newValue, short oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, int newValue, int oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, long newValue, long oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, long long newValue, long long oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, float newValue, float oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, double newValue, double oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, const char* newValue, const char* oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, unsigned char newValue, unsigned char oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, unsigned short newValue, unsigned short oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, unsigned int newValue, unsigned int oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}
void XmlTrace::mark(const TraceProducer* sender, const MarkType* m, unsigned long newValue, unsigned long oldValue, const std::string& comment /*=""*/) {
	// filter
	if (!pass(m, sender))
		return;

	writeNvOv(sender, m, newValue, oldValue, comment);
}

//
// Complex trace marks are transmitted with the following methods.
// A complex trace mark can be interpreted as a tree of tag-value pairs.
//
// beginMark indicates the construction of a new mark
void XmlTrace::beginMark(const TraceProducer* sender, const MarkType* m) {
	// filter
	if (!pass(m, sender))
		return;

	// open mark tree
	beginMarkTree(sender, m);
}

// endMarks signals the end of construction
void XmlTrace::endMark( const std::string& comment /*=""*/ ) {
	// close mark tree
	endMarkTree(comment);
}

// This Methodes are called to describe the construction of complex marks.
void XmlTrace::beginTag(Tag t) {
	openTag(t);
}

void XmlTrace::endTag(Tag) {
	closeTag();
}

void XmlTrace::addTag(Tag t, bool value) {
	writeValue(value, t);
}

void XmlTrace::addTag(Tag t, char value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, short value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, int value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, long value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, long long value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, float value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, double value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, const char* value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, const TraceProducer* value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, unsigned char value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, unsigned short value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, unsigned int value) {
		writeValue(value, t);
}

void XmlTrace::addTag(Tag t, unsigned long value) {
		writeValue(value, t);
}

//
// trace control
//
// XmlTrace starts a new trace
void XmlTrace::startTrace() {
	active=true;

	if( manageOut )
	{
		openXmlFile();
	}

	beginTraceTree();
}

// XmlTrace stops the trace
void XmlTrace::stopTrace() {
	active=false;

	endTraceTree();

	if ( manageOut ) {
		closeXmlFile();
		outputXmlFileList();
		outputXmlMarkInfo();
		generateHtml( xmlFileName, translate(getFilter()) );
	}
}

// This call signals a break in trace: A TraceConsumer should not produce output after this call is send,
// but await further marks for upkeeping internal data if required.
void XmlTrace::pauseTrace()
{
	if( ! active )
	{
		error( "XmlTrace::pauseTrace(): trace is not active, call startTrace()" );
	}
	pause = true;

	*out << "<mark time=\"" << getTime() << "\" id=\"0000\"";
	*out << " name=\"pause\" scope=\"odemx::Simulation\">" << endl;

	*out << "<sender type=\"" << translate( typeToString(sim -> getType()) ) << "\">";
	*out << translate(sim ->getLabel()) << "</sender>" << endl;

	// write details tag
	*out << "<tag name=\"start time\">";
	// write time
	*out << getTime() << "</tag>" << endl;

}

// This call signals the end of a break in trace: A TraceConsumer should update its output if required,
// and report following marks.
void XmlTrace::continueTrace()
{
	if( ! active )
	{
		error( "XmlTrace::continueTrace(): trace is not active, call startTrace()" );
	}

	if( ! pause )
	{
		error( "XmlTrace::continueTrace(): trace not paused, call pauseTrace()" );
	}

	pause = false;

	// write details tag
	*out << "<tag name=\"end time\">";
	// write time
	*out << getTime() << "</tag>" << endl;
	*out << "</mark>" << endl;
}


// help functions
void XmlTrace::openXmlFile() {

	// delete last file data
	resetCurrentFile();

	// build file name
	ostringstream file;
	file << xmlFileName << "_" << fileNumber++ << ".xml";

	// open new file
	out = new ofstream( file.str().c_str() );

	if ( out==0 || !(*out) ) {
		// Error: cannot open file
		error( "XmlTrace::openXmlFile(): cannot open file, sending output to stdout" );
		manageOut = false;
		out = &cout;
	}
	else {
		currentFile.name = file.str();
	}
}


void XmlTrace::closeXmlFile() {

	// delete old file pointer
	reinterpret_cast<ofstream*>(out) -> close();
	delete out; out = 0;
}

void XmlTrace::outputXmlFileList() {

	// build file name
	ostringstream file;
	file << xmlFileName << "_files.xml";

	// open new file
	ostream* xmlStream = new ofstream( file.str().c_str() );

	if ( xmlStream==0 || !(*xmlStream) ) {
		// Error: cannot open file
		error( "XmlTrace::outputXmlFileList(): cannot open file, sending output to stdout" );
		xmlStream = &cout;
	}

	*xmlStream << "<?xml version=\"1.0\" encoding=\"utf-8\" ?>" << endl;
	*xmlStream << "<xmlfiles simulation=\"" << translate(sim ->getLabel());
	*xmlStream << "\">" << endl;

	FileList::iterator iter;
	for( iter = xmlFiles.begin(); iter != xmlFiles.end(); ++iter ) {

		*xmlStream << "<file name=\"" << translate((*iter).name) << "\" ";
		*xmlStream << "starttime=\"" << (*iter).startTime << "\" ";
		*xmlStream << "endtime=\"" << (*iter).endTime << "\" />" << endl;
	}

	*xmlStream << "</xmlfiles>" << endl;

	reinterpret_cast<ofstream*>(xmlStream) -> close();
	delete xmlStream;
}


void XmlTrace::outputXmlMarkInfo() {

	// build file name
	ostringstream file;
	file << xmlFileName << "_markinfo.xml";

	// open new file
	ostream* xmlStream = new ofstream( file.str().c_str() );

	if ( xmlStream==0 || !(*xmlStream) ) {
		// Error: cannot open file
		error( "XmlTrace::outputXmlTraceMarkFile(): cannot open file, sending output to stdout" );
		xmlStream = &cout;
	}

	// build header
	*xmlStream << "<?xml version=\"1.0\" encoding=\"utf-8\" ?>" << endl;
	*xmlStream << "<xmlmarkinfo simulation=\"" << translate(sim ->getLabel());
	*xmlStream << "\">" << endl;

	// list basemark scope with basemarkid
	MarkScopeIdMap::const_iterator map_i;
	for ( map_i = MarkType::getScopeIdMap().begin();
		  map_i != MarkType::getScopeIdMap().end(); ++map_i )
	{
		*xmlStream
		<< "<basemark scope=\""
		<< translate( (*map_i).first )
		<< "\" id=\""
		<< (*map_i).second
		<< "\" />" << endl;
	}

	// list trace mark types
	MarkTypeList::const_iterator list_i;
	for ( list_i = MarkType::getList().begin(); list_i != MarkType::getList().end(); ++list_i )
	{
		*xmlStream
		<< "<marktype scope=\""
		<< translate( typeToString( (*list_i)->getScope() ) )
		<< "\" id=\""
		<< (*list_i) -> getId()
		<< "\" name=\""
		<< translate( (*list_i) -> getName() )
		<< "\" />" << endl;
	}

	// end mark tree
	*xmlStream << "</xmlmarkinfo>" << endl;

	reinterpret_cast<ofstream*>(xmlStream) -> close();
	delete xmlStream;
}


void XmlTrace::resetCurrentFile() {
	currentFile.name = "";
	currentFile.startTime = "";
	currentFile.endTime = "";
}


void XmlTrace::beginTraceTree() {

	// write XML-Header
	*out << "<?xml version=\"1.0\" encoding=\"utf-8\" ?>" << endl;
	// open XmlTrace-Tree
	*out << "<xmltrace simulation=\"" << translate(sim ->getLabel()) << "\" time=\"" << getTime() << "\">" << endl;
	// remember start time
	currentFile.startTime = getTime();
}



void XmlTrace::endTraceTree() {

	// close XmlTrace-Tree
	*out << "</xmltrace>" << endl;

	currentFile.endTime = getTime();
	xmlFiles.push_back( currentFile );
}



void XmlTrace::checkMarkCount() { // only called if manageOut is true

	// not too many marks continue with current file
	if ( xmlMarkCount <= xmlMarkLimit ){
		return;
	}
	// beyond the mark limit
	else if ( limitReachTime != -1 ) {

		 // same time mark must still be added to current file
		if ( limitReachTime == sim -> getTime() ) {
			return;
		}
		// time change, close file and open new one
		else {

			// finish old file
			endTraceTree();
			closeXmlFile();

			// start next file
			openXmlFile();
			beginTraceTree();

			// reset
			limitReachTime = -1;
			xmlMarkCount = 0;
		}
		return;
	}
	// limit reached, but time not set yet, remember the current time
	else {
		limitReachTime = sim -> getTime();
	}
}


//inline void XmlTrace::openXmlTag(const string& n) {
//	if (skipOutput())
//		return;
//
//	// open Xml tag
//	*out << "<" << n << ">";
//}
//
//inline void XmlTrace::closeXmlTag(const string& n) {
//	if (skipOutput())
//		return;
//
//	// close Xml tag
//	*out << "</" << n << ">" << endl;
//}

void XmlTrace::openTag(Tag t) {
	if (skipOutput())
		return;

	// open tag
	*out << "<tag ";
	*out << "name=\"" << translate(t.getName()) << "\"";

	if (t.getId()!=Tag::NOID) {
		*out << " id=\"" << t.getId() << "\"";
	}

	*out << ">";
}

void XmlTrace::closeTag() {
	if (skipOutput())
		return;

	// close tag
	*out << "</tag>" << endl;
}

inline void XmlTrace::writeValue(bool v, const string& n) {
	if (skipOutput())
		return;

	openXmlTag(n);
	// write value
	*out << (v ? "true" : "false");
	closeXmlTag(n);
}

inline void XmlTrace::writeValue(const TraceProducer* tp, const string& n) {
	if (skipOutput())
		return;

	*out << "<" << n;
	*out << " type=\"" << translate( typeToString(tp->getType()) ) << "\">";
	*out << translate(tp->getLabel());
	*out << "</" << n << ">" << endl;
}

inline void XmlTrace::writeValue(bool v, Tag t) {
	if (skipOutput())
		return;

	openTag(t);
	// write value
	*out << (v ? "true" : "false");
	closeTag();
}

inline void XmlTrace::writeValue(const TraceProducer* tp, Tag t) {
	if (skipOutput())
		return;

	*out << "<tag ";
	*out << "name=\"" << translate(t.getName()) << "\"";

	if (t.getId() != Tag::NOID) {
		*out << " id=\"" << t.getId() << "\"";
	}

	*out << " type=\"" << translate( typeToString(tp->getType()) ) << "\">";
	*out << translate(tp->getLabel());

	*out << "</tag>" << endl;

}


template <class T>
void XmlTrace::writeNvOv( const TraceProducer* sender, const MarkType* m,
				T newValue, T oldValue, const std::string& comment /*=""*/ )
{
	if (skipOutput())
		return;

	beginMarkTree(sender, m);

	*out << "<tag name=\"new value\">";
	if ( m -> getId() == Process::markChangeProcessState.getId() )
		*out << printState( (long)newValue );
	else
		*out << newValue;

	*out << "</tag>" << std::endl;

	*out << "<tag name=\"old value\">";
	if ( m -> getId() == Process::markChangeProcessState.getId() )
		*out << printState( (long)oldValue );
	else
		*out << oldValue;

	*out << "</tag>" << std::endl;

	endMarkTree(comment);
}


//void XmlTrace::writeNvOv( const TraceProducer* sender, const MarkType* m,
//				SimTime newValue, SimTime oldValue, const std::string& comment /*=""*/ )
//{
//	if (skipOutput())
//		return;
//
//	beginMarkTree(sender, m);
//
//	*out << "<tag name=\"new value\">";
//	*out << simTimeToString( newValue, timeFormat );
//	*out << "</tag>" << std::endl;
//
//	*out << "<tag name=\"old value\">";
//	*out << simTimeToString( oldValue, timeFormat );
//	*out << "</tag>" << std::endl;
//
//	endMarkTree(comment);
//}


const string XmlTrace::printState( int val ) {
	switch( val ) {
		case( 0 ): return "CREATED";
		case( 1 ): return "CURRENT";
		case( 2 ): return "RUNNABLE";
		case( 3 ): return "IDLE";
		case( 4 ): return "TERMINATED";
	}

	// never reached
	error( "XmlTrace::printState(): unknown process state" );
	return "Trace ERROR";
}


inline void XmlTrace::beginMarkTree(const TraceProducer* sender, const MarkType* m) {
	if (skipOutput())
		return;

	// if dumping to file, check size and open new one if needed
	if (manageOut)
		checkMarkCount();

	// count marks for file size limit
	++xmlMarkCount;

	// open mark tree
	*out << "<mark time=\"" << getTime() << "\" id=\"";
	*out << m->getId();
	*out << "\" name=\"";
	*out << translate(m->getName());

	*out << "\" scope=\"" << translate( typeToString(m->getScope()) ) << "\"";
	*out << ">" << endl;

	// write sender
	writeValue( sender, string("sender") );
}

inline void XmlTrace::endMarkTree( const std::string& comment /*=""*/ ) {
	if (skipOutput())
		return;

	if ( !comment.empty() ) {
		// writeValue(comment, "comment");

		*out << "<tag name=\"comment\">";
		*out << translate( comment );
		//*out << comment;
		*out << "</tag>" << std::endl;
	}

	// close mark tree
	*out << "</mark>" << endl;
}


const string XmlTrace::getTime() {

	SimTime currentTime = sim -> getTime();

	// only compute string when SimTime changed or is still 0
	if( currentTime != oldTime || currentTime == 0 )
	{
		oldTime = currentTime;
		timeString = simTimeToString( currentTime, timeFormat );
	}

	return translate( timeString );
}
