//----------------------------------------------------------------------------
//	Copyright (C) 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file XmlTrace.h

	\author Ronald Kluth

	\date created at 2007/07/15

	\brief Declaration of XmlTrace

	\sa XmlTrace.cpp

	<!-- [detailed description] -->

	\since 2.0
*/

#ifdef _MSC_VER
#pragma warning(disable : 4786)
#endif

#ifndef ODEMX_XMLTRACE_INCLUDED
#define ODEMX_XMLTRACE_INCLUDED

#include <odemx/base/Simulation.h>
#include <odemx/base/Process.h>
#include <odemx/util/Trace.h>
#include <odemx/util/FileInfo.h>
#include <odemx/display/HtmlGenerator.h>
#include <odemx/display/TraceFilter.h>
#include <odemx/display/TimeFormat.h>

#include <iostream>
#include <string>

namespace odemx {

	/** \class XmlTrace

		\ingroup display

		\author Ronald Kluth

		\sa Trace, TraceProducer, TraceConsumer, HtmlGenerator

		\brief XML trace

		%XmlTrace logs simulation events of all model components to one or
		sevaral xml files or an open output stream. If the user provides a
		base filename, %XmlTrace creates the necessary files automatically.

		\note XmlTrace requires the CSS stylesheet XmlTrace.css
		located in odemx/display/stylesheet/

		\note XmlTrace requires the JavaScript XmlTrace.js
		located in odemx/display/javascript/

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.0
	*/
	class XmlTrace : public virtual TraceConsumer,
					 public TraceFilter,
					 public HtmlGenerator
	{
	public:
		/// Constructor for use with DefaultSimulation with file output
		XmlTrace( const std::string& fileName = "",
				  Time::Format& tf = odemx::Time::defaultFormat,
				  unsigned int limit = 2500 );

		/// Constructor for user-defined Simulation with file output
		XmlTrace( Simulation* s, const std::string& fileName = "",
				  Time::Format& tf = odemx::Time::defaultFormat,
				  unsigned int limit = 2500 );

		/// Constructor for use with DefaultSimulation with ostream output
		XmlTrace( std::ostream& os,
				  Time::Format& tf = odemx::Time::defaultFormat );

		/// Constructor for user-defined Simulation with ostream output
		XmlTrace( Simulation* s, std::ostream& os,
				  Time::Format& tf = odemx::Time::defaultFormat );

		/// destruction
		virtual ~XmlTrace();

		// Implementation
	private:

		/**
			\name Simple trace marks

			Simple trace marks are handled by the following methods.

			@{

		*/
		virtual void mark(const TraceProducer* sender, const MarkType* m, const std::string& comment = "");

		// Marks that include information about producer associations:
		virtual void mark(const TraceProducer* sender, const MarkType* m, const TraceProducer* partner, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, const TraceProducer* firstPartner, const TraceProducer* secondPartner, const std::string& comment = "");

		// Marks that include information about producer attributes:
		virtual void mark(const TraceProducer* sender, const MarkType* m, bool newValue, bool oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, char newValue, char oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, short newValue, short oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, int newValue, int oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, long newValue, long oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, long long newValue, long long oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, float newValue, float oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, double newValue, double oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, const char* newValue, const char* oldValue, const std::string& comment = "");

		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned char newValue, unsigned char oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned short newValue, unsigned short oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned int newValue, unsigned int oldValue, const std::string& comment = "");
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned long newValue, unsigned long oldValue, const std::string& comment = "");
		//@}

		/**
			\name Complex trace marks

			Complex trace marks are transmitted with the folowing methodes.
			A complex trace mark can be interpreted as a tree of tag-value pairs.

			@{
		*/
		/// beginMark indicates the construction of a new mark
		virtual void beginMark(const TraceProducer* sender, const MarkType* m);
		/// endMarks signals the end of construction
		virtual void endMark(const std::string& comment = "");

		// These methods are called to construct complex xml marks.
		virtual void beginTag(Tag t);
		virtual void endTag(Tag t);

		virtual void addTag(Tag t, bool value);
		virtual void addTag(Tag t, char value);
		virtual void addTag(Tag t, short value);
		virtual void addTag(Tag t, int value);
		virtual void addTag(Tag t, long value);
		virtual void addTag(Tag t, long long value);
		virtual void addTag(Tag t, float value);
		virtual void addTag(Tag t, double value);
		virtual void addTag(Tag t, const char* value);
		virtual void addTag(Tag t, const TraceProducer* value);

		virtual void addTag(Tag t, unsigned char value);
		virtual void addTag(Tag t, unsigned short value);
		virtual void addTag(Tag t, unsigned int value);
		virtual void addTag(Tag t, unsigned long value);
		//@}

		/**
			\name Trace Control
			@{
		*/
		virtual void startTrace(); ///< XmlTrace initializes a new trace
		virtual void stopTrace(); ///< XmlTrace cleans up
		virtual void pauseTrace(); ///< XmlTrace starts a break
		virtual void continueTrace(); ///< XmlTrace ends a break
		//@}

	private:
		Simulation* sim; ///< pointer to simulation
		bool manageOut; ///< manage the outstream
		bool pause; ///< in pause mode
		bool active; ///< is active
		bool filterOut; ///< is filtering out
		std::ostream* out; ///< outstream for XML text
		std::string xmlFileName; ///< file name base for XML files
		FileInfo currentFile; ///< store name, start and end time
		FileList xmlFiles; ///< stores all xml file names
		unsigned int fileNumber; ///< file counter to be added to each file
		Time::Format& timeFormat; ///< describes the SimTime output format
		SimTime oldTime; ///< used for getTime() format conversion
		std::string timeString; ///< the last computed SimTime as string

		unsigned int xmlMarkCount; ///< count marks to estimate file size
		unsigned int xmlMarkLimit; ///< maximum mark count per xml file
		SimTime limitReachTime; ///< store the SimTime when mark limit was reached

		/**
		 \name Help Procedures
		 @{
		*/

		void openXmlFile();
		void closeXmlFile();
		void resetCurrentFile();
		void outputXmlFileList();
		void outputXmlMarkInfo();

		void checkMarkCount();
		bool limitReached();

		void beginTraceTree();
		void endTraceTree();
		/// begin complex mark
		void beginMarkTree(const TraceProducer* sender, const MarkType* m);
		/// end complex mark
		void endMarkTree(const std::string& comment = "");
		void openXmlTag(const std::string& n); ///< open XML tag
		void closeXmlTag(const std::string& n); ///< close XML tag
		void openTag(Tag t); ///< open a trace tag
		void closeTag(); ///< close a trace tag

		/// write value \p v inside xml tag \p n
		template <class T>
		void writeValue(T v, const std::string& n) {
			if (skipOutput())
				return;

			openXmlTag(n);
			// write value
			*out << v;
			closeXmlTag(n);
		}

		/// write value \p v inside trace tag \p t
		template <class T>
		void writeValue(T v, Tag t) {
			if (skipOutput())
				return;

			openTag(t);
			// write value
			*out << v;
			closeTag();
		}

		/// write bool value \p v inside xml tag \p n
		void writeValue(bool v, const std::string& n);
		/// write TraceProducer \p tp inside xml tag \p n
		void writeValue(const TraceProducer* tp, const std::string& n);

		/// write bool value \p v inside trace tag \p t
		void writeValue(bool v, Tag t);
		/// write TraceProducer \p tp inside trace tag \p t
		void writeValue(const TraceProducer* tp, Tag t);
		/// print process state as string
		static const std::string printState( int val );

		/// write new value old value pair
		template< typename T >
		void writeNvOv( const TraceProducer* sender, const MarkType* m,
						T newValue, T oldValue, const std::string& comment = "" );

//		/// write new value old value for SimTime
//		void writeNvOv( const TraceProducer* sender, const MarkType* m,
//						SimTime newValue, SimTime oldValue, const std::string& comment /*=""*/ );

		/// is current mark filtered out
		bool skipOutput() {return pause || !hasPassed();}
		/// get current SimTime in form of a formatted string
		const std::string getTime();
		//@}
	};
}

#endif /* ODEMX_XMLTRACE_INCLUDED */
