//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file XmlTrace.js

	\author Ronald Kluth
	<!-- [\author <author>]* -->

	\date created at 2007/04/04

	\brief A javascript which handles parsing and display of XmlTraces

	\sa XmlTrace, HtmlGenerator

	<!-- [detailed description] -->

	<!-- [\todo {todos for this file}]* -->

	\since 2.0
*/



// =================================== GLOBALS =================================

// 'var xmlListFile' is defined before this script is included
// 'var markInfoFile' is defined before this script is included

// provide a means to check for the inclusion of this script,
// the value of this variable is checked in the HTML document
// after this script was supposed to be included
xmlTraceIncluded = true;

// the filter object which stores several hashes of entries
// that can be filtered in the trace display
var filter = new TraceFilter();

// the trace handler is responsible for handling the trace files
// and which one of them is currently displayed
var traceHandler = new TraceFileHandler();


// =============================== INITIALIZATION FUNCTIONS ====================


// start trace display after parsing the first trace file
// and initialization of the necessary menus
function init() {
log('init()');

	parseXmlTrace( xmlListFile );
	initMenus();
	showFilteredData();
}


// use trace handler info to fill menu boxes for filtering the trace display
var menusInitialized = false;
function initMenus() {
log('initMenus()');

	// period selection and log only need to be initialized once
	if ( ! menusInitialized ) {

		// fill an HTML select menu with the file values
		var html = '<select ' +
				   'onchange="' +
				   'changeFile( this.options[this.selectedIndex].value )">';

		// iterate over the whole xml snapshot
		for ( var j = 0; j < traceHandler.file.length; ++j ) {
			// create a new select option with value = traceHandler.file index
			html += '<option value="'+ j +'">';
			html += traceHandler.file[j].start + "-";
			html += traceHandler.file[j].end
			html += '</option>';
		}

		// add an otion for all files
		//html += '<option value="all">all</option>';
		html += '<option value="all">all</option>';

		html += '</select>';

		// populate the select menu
		$('periodSelect').innerHTML = html;

		// initialize the log window
		$('logToggle').onclick = function () { toggle( 'log' ); };

		$('filterReset').onclick = function () { filter.reset(); };
	}

	// ----------------------------------------------------------------
	// this part will be executed again when the displayed file changes
	// new filter marks might have been added when parsing a new file

	// populate the filter panels
	$('mtnDisplay').innerHTML = hashToHTML( filter.markNames, 'mtnEntry', "markName" );
	$('mtiDisplay').innerHTML = hashToHTML( filter.markIds, 'mtiEntry', "markId" );
	$('mtsDisplay').innerHTML = hashToHTML( filter.markScopes, 'mtsEntry', "markScope" );
	$('slDisplay').innerHTML = hashToHTML( filter.senderLabels, 'slEntry', "senderLabel" );
	$('stDisplay').innerHTML = hashToHTML( filter.senderTypes, 'stEntry', "senderType" );

	// add onclick events to all mark filters in xxxDisplays
	// which call toggleFilter() for the selection
	addClickHandler( "mtnEntry", "markName" );
	addClickHandler( "mtiEntry", "markId" );
	addClickHandler( "mtsEntry", "markScope" );
	addClickHandler( "slEntry", "senderLabel" );
	addClickHandler( "stEntry", "senderType" );

	menusInitialized = true;
}



// =========================== PARSING =========================================


// parsing requires the xml list file which contains information about all
// files generated during the simulation: file name, start time, end time
function parseXmlTrace( listFile ) {
log('parseXmlTrace()');

	showInfo( "Retrieving simulation data..." );

	// read the xml list file, returns XMLDocument object
	var xmlFiles = loadXml( listFile );

	// get all 'file' tags from XML list file, returns XPathResult object
	var files = xmlFiles.evaluate( "//file",
								   xmlFiles,
								   null,
								   XPathResult.ORDERED_NODE_SNAPSHOT_TYPE,
								   null );

	var name, start, end;

	// fill the traceHandler with the file information:
	// file name, start time, end time
	for ( var i = 0; i < files.snapshotLength; ++i ) {

		name = files.snapshotItem(i).getAttribute('name');
		start = files.snapshotItem(i).getAttribute('starttime');
		end = files.snapshotItem(i).getAttribute('endtime');

		// construct trace file object and add it to the traceHandler
		// the mark tree is left out here because it takes too long
		// to parse all files at once -> see: changeFile()
		traceHandler.file.push( new File ( name, start, end ) );
	}

	// set the handler to display the first trace file by default
	if ( traceHandler.file[0].name )
		traceHandler.currentFileIndex = 0;
	else
		alert( "Error: XML start file not initialized" )

	// --------------------------------------------------
	// only the first file is parsed 'onload' of the page

	// get the mark tree from the current file
	loadTraceMarks( 0 );
}


// fill the trace handler mark tree from file stored at the given index
function loadTraceMarks( fileIndex ) {
log('loadTraceMarks( ' + fileIndex + ' )')

	// if the file has not been parsed yet, do it now
	if ( traceHandler.file[ fileIndex ].allMarks == 0 ) {

		// get the mark tree from the current file
		var xmlTree = loadXml( traceHandler.file[ fileIndex ].name );

		// get all mark nodes from the XML tree
		var tree = xmlTree.evaluate( "//mark",
									  xmlTree,
									  null,
									  XPathResult.ORDERED_NODE_SNAPSHOT_TYPE,
									  null );

		// set trace handler mark tree
		traceHandler.file[ fileIndex ].allMarks = buildMarkArrayAndFilter( tree, 0 );

		// the filter displays must be rechecked, possibly new handlers added
		initMenus();
	}
}


// =========================== HTML PRODUCTION =================================



// this function takes an XPathResult xml snapshot,
// which in this case is the mark tree of a file,
// and builds the MarkArray and the possible filters
function buildMarkArrayAndFilter( xpathMarkTree ) {
log('buildMarkArrayAndFilter()');

	// result empty, display message
	if ( xpathMarkTree.snapshotLength == 0 ) {
		showInfo( "Nothing to display." );
		return 0;
	}

	// the return Array which gets filled with Mark()s
	var markArray = new Array();

	// variables needed to describe a mark without tags
	var markName, markScope, markId, markTime;
	var senderType, senderLabel;

	// iterate over the whole xml mark snapshot
	for ( var i = 0; i < xpathMarkTree.snapshotLength; ++i )
	{
		// get complete mark info and add fields to filter
		markName = xpathMarkTree.snapshotItem(i).getAttribute('name');
		filter.add( "markName", markName );
		markScope = xpathMarkTree.snapshotItem(i).getAttribute('scope').replace( /</g, '&lt;' ).replace( />/g, '&gt;' );
		filter.add( "markScope", markScope );
		markId = xpathMarkTree.snapshotItem(i).getAttribute('id');
		filter.add( "markId", markId );
		senderLabel = xpathMarkTree.snapshotItem(i).getElementsByTagName('sender')[0].textContent;
		filter.add( "senderLabel", senderLabel );
		senderType = xpathMarkTree.snapshotItem(i).getElementsByTagName('sender')[0].getAttribute('type').replace( /</g, '&lt;' ).replace( />/g, '&gt;' );
		filter.add( "senderType", senderType );
		markTime = xpathMarkTree.snapshotItem(i).getAttribute('time');

		// store the retrieved information in a mark
		var currentMark = new Mark( markTime, markName, markId,
									markScope, senderLabel, senderType );

		// print tags belonging to this event
		var tags = xpathMarkTree.snapshotItem(i).getElementsByTagName('tag');

		// temporary local vars
		var tagName;
		var tagValue;

		// add tags to the xpathMarkTree tag list
		for ( var tag_i = 0; tag_i < tags.length; ++tag_i )
		{
			// get name and value
			tagName = tags[tag_i].getAttribute('name');
			tagValue = tags[tag_i].textContent;

			currentMark.tags.push( new Tag( tagName, tagValue ) );
		}

		markArray.push( currentMark );
	} // end of mark loop

	return markArray;
}


// this function builds HTML from the MarkArray
// of the currently active trace file
function markArrayToHTML( fileIndex ) {
log('markArrayToHTML()');

	var newHTML = ""; 		// the output stream
	var lastSimTime = null;	// remember the marks' simulation time
	var lastSender = "";	// suppress printing the same sender
	var firstMarkAtTime = false;

	// variables needed to describe a mark without tags
	var markName, markScope, markId, markTime;
	var senderType, senderLabel;

	// global traceHandler.currentFileIndex is set
	// whenever the time selector is accessed
	var marks = traceHandler.file[ fileIndex ].allMarks;


	// iterate over the whole mark array
	for ( var i = 0; i < marks.length; ++i )
	{
		// get complete mark info and immediately check the filter
		// if the mark does not pass all checks, the loop continues immediately
		markName = marks[i].name;
		if ( !filter.pass( "markName", markName ) ) continue;
		markScope = marks[i].scope;
		if ( !filter.pass( "markScope", markScope ) ) continue;
		markId = marks[i].id;
		if ( !filter.pass( "markId", markId ) ) continue;
		senderLabel = marks[i].senderLabel;
		if ( !filter.pass( "senderLabel", senderLabel ) ) continue;
		senderType = marks[i].senderType;
		if ( !filter.pass( "senderType", senderType ) ) continue;
		markTime = marks[i].time;

		// to print sender at start of time box, even if repeated
		firstMarkAtTime = false;

		// check for change in the simulation time of a mark
		if ( markTime != lastSimTime ) {

			// new simulation time: close time box
			if ( lastSimTime != null )
			{
				newHTML += "</table>" + // timeTable
						   "</div>";    // timeBox
			}
			// remember time of current timeBox;
			lastSimTime = markTime;

			// build new time box containing a table
			newHTML += "<div class='timeBox' title='Simulation Time:" + markTime + "'>" +
					   "<h3>Simulation Time: " + markTime + "</h3>" +
					   "<table class='timeTable'>" +
					   "<tr><th class='th_sender'>Sender Label</th>" +
					   "<th class='th_event'>Mark Type</th>" +
					   "<th class='th_tag'>Details</th></tr>";

			// remember the start of a new time box, so a sender will be shown
			firstMarkAtTime = true;
		}

		// build new mark row
		newHTML += "<tr>";

		// check if sender label needs to be displayed or if it's still the same
		if ( senderLabel != lastSender || firstMarkAtTime ) {
			// get sender label
			newHTML += "<td class='sender' title='" + senderType + "'>" +
					   senderLabel +
					   "</td>";

			lastSender = senderLabel;
		}
		else {
			// leave senderlabel empty if still the same
			newHTML += "<td class='empty'>&nbsp;</td>";
		}
		// get event name
		newHTML += "<td class='event' " +
				"title='" + markScope + " " + markId +"'>" +
				markName +
				"</td>";

		// print tags belonging to this event
		var tags = marks[i].tags
		if ( tags.length > 0 )
		{
			// print tags in the third column
			for ( var tag_i = 0; tag_i < tags.length; ++tag_i )
			{
				if ( tag_i > 0 ) {
					// start row, leave first two columns empty
					newHTML += "</tr><tr><td class='empty' colspan='2'>&nbsp;</td>";
				}
				newHTML += "<td class='tag'>";

				// get name and value
				tagName = tags[tag_i].name;
				tagValue = tags[tag_i].value;

				// print tags and treat comments
				if ( tagName != "comment" ) {
					newHTML += "<div class='tagName'>" +
							   tagName +
							   "</div>" +
							   "<div class='tagValue'>" +
							   tagValue +
							   "</div>";
				}
				else
				{
					newHTML += "<div class='comment'>" +
							   tagValue +
							   "</div>";
				}

				// close tag
				newHTML += "</td>";
			}
		}
		else { // no tags
			newHTML += "<td>&nbsp;</td>";
		}

		// close event box and draw spacer line to next event
		newHTML += "</tr>";
		newHTML += "<tr><td class='spacer' colspan='3'></td></tr>";

	} // end of mark loop

	// close last time box
	newHTML += "</table>";
	newHTML += "</div>";

	return newHTML;
}


// ============================== DISPLAY FUNCTIONS ============================


// display the mark array of the currently selected file,
// called on init and on click of a filter choice
function showFilteredData() {
log('showFilteredData()');

	// in some browsers the info works, in others it doesn't
	// still, make an attempt to have this shown
	showInfo( "Retrieving simulation data..." );

	// this is where the parsed info gets displayed
	$('content').innerHTML = "";

	// the user may choose to display all info at once, or only one file
	if ( traceHandler.currentFileIndex != "all" ) {

		// display the chosen file's content
		$('content').innerHTML = markArrayToHTML( traceHandler.currentFileIndex );
	}
	else {

		// display all files' content by appendinng to 'content'
		for ( var file_i = 0; file_i < traceHandler.file.length; ++file_i ) {
			$('content').innerHTML += markArrayToHTML( file_i );
		}
	}

	// in case there's nothinig to display, tell the user
	if ( $('content').innerHTML.length > 0 )
		showContent();
	else
		showInfo( "Nothing to display" );

	// redraw the filter, i.e. correct the filtered entries' style
	// sometimes when loading a new file, the 'line-through' is gone
	filter.redraw();
}


// a file is parsed into a mark array if it has not been loaded before,
// called when the period selector is accessed
function changeFile( newFileIndex ) {
log('changeFile( ' + newFileIndex + ' )');

	// if all files are to be shown, all trees have to be built in memory
	if ( newFileIndex == "all" ) {

		// confirm potentially long load time
		if ( confirm( "Loading data from all files may take a long time.\n\n" +
					  "Do you want to continue?" ) )
		{
			// build all mark trees
			for ( var i = 0; i < traceHandler.file.length; ++i ) {

				loadTraceMarks( i );
			}
		}
		else { // if the action was cancelled above
			return;
		}
	}
	else { // only one file to be loaded, do it
		loadTraceMarks( newFileIndex );
	}

	// remember which file is currently active
	traceHandler.currentFileIndex = newFileIndex;

	// display the current file
	showFilteredData();
}


// =============================== HELP FUNCTIONS ==============================


// show an info box to tell the user what's being done
// this, however is not shown consistently on all browsers
function showInfo( infoText ) {
log('showInfo( ' + infoText + ' )');

	$('content').style.display = 'none';
	$('info').innerHTML = "<p>" + infoText + "</p>";
	$('info').style.display = 'block';
}

// display the content area, called after 'content' was re-filled
function showContent() {
log('showContent()');

	$('info').style.display = 'none';
	$('content').style.display = 'block';
}

// provides easy access to all id'ed elements of the document
function $( elementId ) {
	return document.getElementById( elementId );
}

// numeric ordering function
function ascendingOrder(a, b){ return (a-b); }

// fill an HTML table with the contents of the given filter hash
function hashToHTML( outputHash, tdClass, which ) {
log('hashToHTML( ' + tdClass + ' )');

	// use a temporary array to sort the output
	var tmp = new Array();

	// copy hash keys ( filter entries ) into array
	for ( var element in outputHash ) {
		tmp.push( element );
	}

	// sort the filter entries to be displayed
	// check for numeric values, which need a different sorting function
	if ( !isNaN( tmp[0] ) ){
		tmp.sort( ascendingOrder );
	}
	else {
		tmp.sort();
	}

	// out put is stored in a 3-column-table
	var html = "<table>";
	var entries = 0;

	// get the corresponding HTML ID hash
	var ids = eval( "filter." + which + "ElementHtmlIDs;" );

	// add buttons to the table for switching the selection (all, none, invert)
	html += "<tr>" +
			"<td nowrap id='" + tdClass + "All' class='selSwitcher'>" +
			"Filter All" +
			"</td>" +
			"<td nowrap id='" + tdClass + "None' class='selSwitcher'>" +
			"Filter None" +
			"</td>" +
			"<td nowrap id='" + tdClass + "Invert' class='selSwitcher'>" +
			"Invert Selection" +
			"</td>" +
			"</tr>";

	// fill the table
	for ( var i = 0; i < tmp.length; ++i ) {

		// start row
		if ( (entries%3) == 0 )
			html += "<tr>";

		// remember IDs given to corresponding HTML elements
		ids[ tmp[i] ] = tdClass + i;

		// one entry, nowrap to avoid line wrapping in some browsers
		html += "<td nowrap id='" + tdClass + i + "' class='" + tdClass + "'>";
		html += tmp[i] + "</td>";

		// end row
		if ( (entries%3) == 2 )
			html += "</tr>";

		// counter to end row after every third entry
		++entries;
	}

	// add proper amount of empty cells
	if ( (entries%3) == 1 )
		html += "<td></td><td></td></tr>";

	if ( (entries%3) == 2 )
		html += "<td></td></tr>"

	// end table
	html += "</table>";

	return html;
}

// useful function to get all elements belonging to one class
function getElementsByClassName( className, parentElement ) {
log('getElementsByClassName( ' + className + ' )');

	// xpath expression to find objects by className
  	xpath = ".//*[contains(concat(' ', @class, ' '), ' " + className + " ')]";

    // apply xpath to the document node
    var query = document.evaluate(xpath, $(parentElement) || document,
      			null, XPathResult.ORDERED_NODE_SNAPSHOT_TYPE, null);

    // fill result list and return it
    var results = [];
    for (var i = 0, length = query.snapshotLength; i < length; ++i) {
      results.push(query.snapshotItem(i));
    }
    return results;
}

// all filter entries need to have an onclick event attached,
// which will toggle the filtering of the corresponding mark type
function addClickHandler( className, which ) {
log("addClickHandler( " + className + ", " + which + " )");

	// add onclick handler to selection switches (all, none, invert)
	eval( "$('" + className + "All').onclick = " +
		  "function () { filter.selectAll('" + which + "'); };" );

	eval( "$('" + className + "None').onclick = " +
		  "function () { filter.selectNone('" + which + "'); };" );

	eval( "$('" + className + "Invert').onclick = " +
		  "function () { filter.invert('" + which + "'); };" );

	// get list of all HTML elements with className
	try { var list = getElementsByClassName( className ); }
	catch(e) { alert( 'Error initializing dynamic mark type choice' ); }

	// init mark type entries with onclick event listener
	for ( var i = 0; i < list.length; ++i ) {
		$( list[i].id ).onclick =
		function () {
			filter.toggle( which, this.id );
		};
	}
}

// add the given string to the JS log window
function log( string ) {
	// add to the top row of the log-<div>
	var tmp = string + "<br />";
	tmp += $('log').innerHTML;
	$('log').innerHTML = tmp;
}

// visibility toggle for given element
function toggle( elementId ) {
log( 'toggle( ' + elementId + ' )' );

	$( elementId ).style.display =
		( $( elementId ).style.display == 'block' ) ? 'none' : 'block';
}


// ================================ XML LOADER =================================

// get XMLHttp object
function createXmlHttp() {
log('createXmlHttp()');

	// get the object the standard way, or the Microsoft way
	try {
		return( new XMLHttpRequest() );
	}
	catch(e) {
		return( new ActiveXObject('Microsoft.XMLHTTP') );
	}
}

// load a given XML file by name
function loadXml( file ) {
log('loadXml( ' + file + ' )');

	// get the XMLHttp object
	var xmlhttp = createXmlHttp();

	// if object initialized, make request for XML file contents
	if ( xmlhttp != null )
	{
		xmlhttp.open( "GET", file, false );
		xmlhttp.send( null );
	}
	else // error
	{
		alert( "Cannot load XML file, this browser does not support XMLHTTP." )
	}

	// read the xml document (implicit cast)
	return xmlhttp.responseXML;
}

// ============================== TYPE DEFINITIONS =============================


// object to store the complete mark type information
function Mark( mtime, mname, mid, mscope, slabel, stype ) {
	this.time = mtime;
	this.name = mname;
	this.id = mid;
	this.scope = mscope;
	this.senderLabel = slabel;
	this.senderType = stype;
	this.tags = [] // Array containing Tag()s
}

// object to store mark tag information
function Tag( tname, tvalue ) {
	this.name = tname;
	this.value = tvalue;
}

// the filter object in JSON notation:
// uses hash arrays { keyStringToFilter : bool } to filter out marks
// choices offered: MTS, MTN, MTI, SL, ST -- analogous to TraceFilter
function TraceFilter() {
	return {

		// filter entries
		markScopes		: {},
		markNames		: {},
		markIds			: {},
		senderLabels	: {},
		senderTypes		: {},

		// corresponding filter entry IDs in the HTML document
		markScopeElementHtmlIDs		: {},
		markNameElementHtmlIDs		: {},
		markIdElementHtmlIDs		: {},
		senderLabelElementHtmlIDs	: {},
		senderTypeElementHtmlIDs	: {},

		// add an Element to the filter
		add:
		function( which, newElement ) {
			// if the element exists and is being filtered, don't
			if ( ! eval( "this." + which + "s[ newElement ] == true;" ) )
				eval( "this." + which + "s[ newElement ] = false;" );
		},

		// checks whether element value is true, meaning it must be filtered
		pass:
		function( which, testElement ) {
			return !( eval( "this." + which + "s[ testElement ] == true;" ) );
		},

		// sets the value of the element to block as true
		block:
		function( which, blockedElement ) {
			eval( "this." + which + "s[ blockedElement ] = true");
		},

		// sets the value of blocked element to false, so it will pass
		unblock:
		function( which, blockedElement ) {
			eval( "this." + which + "s[ blockedElement ] = false");
		},

		// switches the filtering of an entry between true and false and
		toggle:
		function( which, entryId ) {

			// replacement of '<' and '>' needed for C++ template type names
			var entry = $(entryId).textContent.replace( /</g, '&lt;' ).replace( />/g, '&gt;' );

			log('filter.toggle( ' + which + ', ' + entry + ' )');

			// if this mark name is being filtered, unblock it
			if ( ! this.pass( which, entry ) ) {
				this.unblock( which, entry );
			}
			// this mark name is currently not being filtered
			else {
				this.block( which, entry );
			}

			// display the filtered content
			showFilteredData();
		},

		// set all filter entries to true
		selectAll:
		function ( which ) {
		log('filter.selectAll( ' + which + ' )');

			// get filter hash and corresponding entry ID hash
			var hash = eval( "this." + which + "s;" );

			// set all entries in filter hash to true
			for ( var entry in hash ) {

				hash[entry] = true;
			}

			// display newly filtered trace
			showFilteredData();
		},

		// set all filter entries to false
		selectNone:
		function ( which ) {
		log('filter.selectNone( ' + which + ' )');

			// get filter hash and corresponding entry ID hash
			var hash = eval( "this." + which + "s;" );

			// set all entries in filter hash to false
			for ( var entry in hash ) {

				hash[entry] = false;
			}

			// display newly filtered trace
			showFilteredData();
		},

		// set all filter entries to their negation
		invert:
		function ( which ) {
		log('filter.invert( ' + which + ' )');

			// get filter hash and corresponding entry ID hash
			var hash = eval( "this." + which + "s;" );

			// negate all entries in the given filter hash
			for ( var entry in hash ) {

				hash[entry] = !hash[entry];
			}

			// display newly filtered trace
			showFilteredData();
		},

		// redraw the style of all filter entries
		redraw:
		function() {
		log('filter.redraw()');

			// all allowed eval selections
			which = [ "markScope", "markName", "markId", "senderLabel", "senderType" ];

			// iterate over all filter hashes
			for ( var i = 0; i < which.length; ++i ) {

				// get current filter hash and ID hash
				var hash = eval( "this." + which[i] + "s;" );
				var ids = eval( "this." + which[i] + "ElementHtmlIDs;" );

				// iterate over all elements in the current hash
				for ( var entry in hash ) {

					// get the entry's ID in the document
					var entryId = ids[ entry ];

					// set style according to filtering status
					if ( hash[entry] == false ) {
						$( entryId ).style.textDecoration = "none";
						//$( entryId ).style.color = "white";
						$( entryId ).style.backgroundColor = "firebrick";
						hash[entry] = false;
					}
					else {
						$( entryId ).style.textDecoration = "line-through";
						//$( entryId ).style.color = "red";
						$( entryId ).style.backgroundColor = "black";
						hash[entry] = true;
					}
				}
			}
		},

		// reset all filter entries to false
		reset:
		function () {
		log('filter.reset()');

			// all allowed eval selections
			which = [ "markScope", "markName", "markId", "senderLabel", "senderType" ];

			// iterate over all filter hashes
			for ( var i = 0; i < which.length; ++i ) {

				// get current filter hash and ID hash
				var hash = eval( "this." + which[i] + "s;" );

				// iterate over all elements in the current hash
				for ( var entry in hash ) {

					hash[entry] = false;
				}
			}

			// display newly filtered trace
			showFilteredData();
		}
	}; // end tmp filter object
}

// the handler object constructor for all trace file data
function TraceFileHandler() {
	this.currentFileIndex = undefined;	// the currently active trace file
	this.file = new Array(); 			// of File() objects
}

// the file object constructor
function File( name, start, end ) {

	this.name = name;	// actual file name
	this.start = start;	// simulation time of first trace mark
	this.end = end;		// simulation time of last trace mark
	this.allMarks = 0;
}
