//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2003, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file SimpleProtocol.cpp

	\author Ronald Kluth

	\date created at 2007/06/10

	\brief Example for ODEMx protocol simulation techniques

	\sa SimpleProtocolMessage.h SimpleProtocolEntities.h

	This example shows essential techniques for
	implementing a protocol simulation with ODEMx.
	Protocol behavior is specified by employing the OSI
	layer model.

	\since 2.0
*/

/**
	\example SimpleProtocol.cpp

	With this example, the essential protocol simulation techniques of
	ODEMx are introduced.

	Protocol models for ODEMx are based on the OSI layer model approach,
	where the protocol stack is separated into several layers, each with
	a distinct function.

	A protocol simulation starts out with a ProtocolSimulation object,
	which is basically a normal Simulation object, extended by functions
	to create and access the network topology, the transmission medium,
	and the protocol layers.

	The network topology is described by ProtocolStack objects and their
	connections, which are maintained by the class NetTopology. Derived
	from this class is the TransmissionMedium, which also handles
	message transmissions between the network's nodes.

	A node itself has an input and an output port, which can be used
	by the system environment to pass data to a node, and to register
	successful message transmissions at a receiving node. The behavior
	of a node is modeled in its active components, so-called entities.

	A ProtocolEntity is a process representing a protocol layer function.
	Each node maintains exactly one entity per layer, which are
	automatically created upon ProtocolStack construction. Entities can receive
	input from the layer above or below and a behavior must be specified
	according to the data passed between these layers.

	In this example, a simple protocol simulation is defined, which shows
	the essential techniques to model and run a protocol simulation with
	ODEMx. Here, two protocol layers are modeled, one for the hardware
	part of a node, and one for the software part. The network contains
	simple nodes, whose behavior is modeled by the definition of a
	HardwareEntity and a SoftwareEntity. The network's topology is
	definied in the simulation objects's initialization function. Lastly,
	messages between nodes are triggered by external events, which are also
	scheduled during the initialization phase.

	\sa SimpleProtocolMessage.h SimpleProtocolEntities.h
*/

#include <odemx/odemx.h>
#include "SimpleProtocolMessage.h"
#include "SimpleProtocolEntities.h"

#include <iostream>
#include <string>
#include <vector>

using namespace odemx;
using namespace std;
//
// In this example, a simple protocol simulation is defined, which shows
// the essential techniques to model and run a protocol simulation with
// ODEMx. There are two protocol layers, one for the hardware
// part of a node, and one for the software part. The network contains
// simple nodes, whose behavior is modeled by defining a
// HardwareEntity and a SoftwareEntity. The network's topology is
// definied in the simulation objects's initialization function. Lastly,
// messages between nodes are triggered by external events, which are also
// scheduled during the initialization phase.
//
// NetNodes are basic components of protocol simulations. In this case,
// a very simple derivation is used.
//
class SimpleNetNode: public ProtocolStack {
public:
	SimpleNetNode( ProtocolSimulation* sim, Label l ):
		ProtocolStack( sim, l ) {}

protected:
	//
	// Since ProtocolStack is an abstract class, a definition
	// of the pure virtual function receivedData() must be
	// provided to define a node's behavior upon message reception.
	//
	virtual void receivedData( PortData* msg ) {
		UserData* msgData = (UserData*)msg;

		cout <<  "Message received: \"" << msgData -> content
			 << "\" by " << getLabel()
			 << " at SimTime " << getSimulation() -> getTime() << endl;
	}
};

//
// Triggering environment data input and message transmissions
// can be done using events.
//
class MessageEvent: public Event {
public:
	MessageEvent( ProtocolSimulation* sim,
				  ProtocolStack* src, ProtocolStack* dest, string msg ):
		Event( sim, "Message Event" ),
		source( src ),
		destination( dest ),
		message( msg )
		{}
protected:
	//
	// The event action is executed when the simulation reaches
	// its given excution time. At that point, a simple message
	// is handed to the source node, which will initiate a transmission
	// of the message to the destination node.
	//
	virtual void eventAction() {

		cout << "Message event: \"" << message
			 << "\" from " << source -> getLabel()
			 << " to " << destination -> getLabel()
			 << " at SimTime " << getCurrentTime() << endl;

		UserData* udat = new UserData( source, destination, message );
		source -> inputData( udat );
	}
private:
	ProtocolStack* source;
	ProtocolStack* destination;
	string message;
};


//
// The simulation object initalizes all protocol components
// and schedules message events.
//
class SimpleProtocolSimulation: public ProtocolSimulation {
public:
	SimpleProtocolSimulation( Label l ):
		ProtocolSimulation( l )
		{}

	virtual ~SimpleProtocolSimulation()
		{
			vector< Event* >::iterator iter;
			for( iter = messages.begin(); iter != messages.end(); ++iter )
				delete *iter;
		}
	//
	// This function creates the required simulation components
	// such as nodes, layers, and message events. The network's
	// topology is defined here. The topology could easily be
	// changed during simulation by creating special events for
	// this action, or by having another process influence the
	// network status.
	//
	virtual void initSimulation() {

		cout << "Initializing necessary processes..." << endl;
		//
		// Every protocol simulation requires a transmission
		// medium. It is, however, not necessary to derive one's
		// own class from the base TransmissionMedium. An object
		// of this class can be used as default transmission medium.
		//
		createMedium< TransmissionMedium >( "Medium" );
		//
		// Layers are created by a template function that requires
		// the corresponding entity type as parameter. All protocol
		// entities will be initialized automatically upon ProtocolStack
		// construction, which is why all layers must be defined
		// prior to creating any nodes.
		//
		// Furthermore, the layers must be in top-down order because the
		// entities of a node are linked automatically to communicate
		// with the layers above and below.
		//
		createLayer< SoftwareEntity >( "Software" );
		createLayer< HardwareEntity >( "Hardware" );
		//
		// Creation of all nodes. For simplicity, this has also
		// been wrapped into a function, as nodes must be created,
		// and registered with the NetTopology for initialization.
		// The given string is the node's label.
		//
		ProtocolStack* A = initNode( "A" );
		ProtocolStack* B = initNode( "B" );
		ProtocolStack* C = initNode( "C" );
		ProtocolStack* D = initNode( "D" );
		ProtocolStack* E = initNode( "E" );
		ProtocolStack* F = initNode( "F" );
		//
		// Next, the topology is built by connecting the newly
		// created nodes. Links are set by giving a probability
		// for message delivery, which will be modeled internally by
		// a random variable. Both directions can be given separately,
		// as setLink is overloaded with a fourth double parameter.
		//
		getTransmissionMedium() -> setLink( A, B, 0.9 );
		getTransmissionMedium() -> setLink( B, C, 0.9 );
		getTransmissionMedium() -> setLink( A, C, 1.0 );
		getTransmissionMedium() -> setLink( C, D, 1.0 );
		getTransmissionMedium() -> setLink( D, E, 0.8 );
		getTransmissionMedium() -> setLink( D, F, 0.9 );
		getTransmissionMedium() -> setLink( E, F, 0.9 );
		//
		// Show the connectivy matrix with connection quality.
		//
		cout << "----------------------------" << endl
			 << "Current connectivity matrix:" << endl << endl;
		getTransmissionMedium() -> printNodeMatrix();
		//
		// Schedule message events to occur during simulation.
		// This is also done by a wrapper function, which stores
		// messages in a vector for memory management.
		//
		scheduleMessageEvent( A, D, "message from A to D", 10 );
		scheduleMessageEvent( A, F, "message from A to F", 15 );
		scheduleMessageEvent( B, F, "message from B to F", 13 );
		scheduleMessageEvent( E, A, "message from E to A", 9 );

		printExecutionList();
		cout << "Initialization finished." << endl;
	}

	void scheduleMessageEvent( ProtocolStack* src, ProtocolStack* dest, string msg, SimTime t ) {
		messages.push_back( new MessageEvent( this, src, dest, msg ) );
		messages.back() -> scheduleAt( t );
	}

	ProtocolStack* initNode( Label node ) {
		ProtocolStack* newNode = new SimpleNetNode( this, node );
		getNetTopology() -> registerNode( newNode, 0, 0 );
		return newNode;
	}

private:
	vector< Event* > messages;
};

//
// As almost everything is specified in the simulation, the main
// function only handles ODEMx monitoring objects such as the
// HtmlTrace, ProtocolXmlTrace, and a simulation report.
//
int main( int argc, char* argv[] ) {

	SimpleProtocolSimulation sim( "Example Protocol Simulation" );

	HtmlTrace* htmlTrace = new HtmlTrace( &sim, "SimpleProtocolHtmlTrace.html" );

	Time::German german( Time::Base( 1970, 1, 1, 480, Time::minutes ) );
//	XmlTrace* xmlTrace = new XmlTrace( sim, "SimpleProtocolXmlTrace",
//									   Time::defaultFormat, 1000 );

	XmlTrace* xmlTrace = new XmlTrace( &sim, "SimpleProtocolXmlTrace",
									   german, 1000 );

	sim.addConsumer( htmlTrace );
	sim.addConsumer( xmlTrace );
	sim.startTrace();

	cout << endl << "Starting " << sim.getLabel() << endl << endl;
	cout << "============================================" << endl;

	sim.run();

	cout << "============================================" << endl;
	cout << "Finished " << sim.getLabel() << endl << endl;

	HtmlReport* reporter = new HtmlReport( &sim, "SimpleProtocolHtmlReport.html" );
	reporter -> addProducer( sim.getTransmissionMedium() );
	reporter -> generateReport();
	sim.stopTrace();

	delete htmlTrace;
	delete xmlTrace;
	delete reporter;
}
