//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2009 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file basicEvent.cpp

	\author Ronald Kluth

	\date created at 2008/12/08

	\brief Example for basic event simulation

	This example shows the basic techniques for
	realizing an event-based simulation with ODEMx.

	\since 2.1
*/
/** \example basicEvent.cpp
	More basic simulation techniques of ODEMx are introduced.

	An event simulation contains a number of events that describe
	the certain occurrences in a model. An event represents one atomic
	action. These actions are timeless. Time consumption is realised
	with special time-operations (scheduling of events). In this example
	several time-operations are presented together with an example event.
*/

#include <odemx/base/Event.h>
#include <odemx/base/Simulation.h>

#include <iostream>

using namespace std;
using namespace odemx;

//
// This example Event represents simple SimTime ticks.
// It outputs a '*' every full tick in SimTime.
//
class Tick: public Event
{
public:
	Tick(): Event( "Tick" ) {}

	virtual void eventAction()
	{
		cout << " * ";
		scheduleIn( 1 );
	}
};

int main( int argc, char* argv[] )
{
	Simulation* sim = getDefaultSimulation();
	//
	// For this example, the DefaultSimulation is used implicitly.
	// In more complex applications it is recommended to provide a
	// custom-made simulation class.
	//
	Tick ticker;
	//
	// The new event is just created at this point.
	// It won't be executed because it is not scheduled.
	// An event can be scheduled by any of these methods:
	//
	// 'schedule()'
	// 'scheduleIn()'
	// 'scheduleAt()'
	// 'scheduleAppend()'
	// 'scheduleAppendIn()'
	// 'scheduleAppendAt()'
	//
	// 'schedule()' and 'scheduleAppend()' are equal to
	// 'scheduleIn(0)' and scheduleAppendIn(0), repsectively.
	// 'scheduleAt(t)' and 'scheduleAppendAt(t)' are equal to
	// 'scheduleIn(t-now)' and 'scheduleAppendIn(t-now)'.
	//
	ticker.scheduleAt( 1 );
	//
	// There are three ODEMx helper functions to compute a default 
	// simulation:
	//
	// Firstly, you can call 'simulate()', which computes the 
	// simulation until it is finished. A simulation is finished 
	// if there is no active process or event scheduled, or if it
	// is stopped with 'exitSimulation()'.
	//
	// Secondly, you can compute a simulation step-wise with
	// 'simulateStep()'.
	// 
	// Thirdly, you can run a simulation until a given SimTime is
	// reached or passed with 'simulateUntil()'.
	//
	// Because Tick events will reschedule themselves forever, we
	// shouldn't use 'simulate()'. Instead we try both 'simulateStep()'
	// and 'simulateUntil()'.
	//
	cout << "Basic Simulation Example" << endl;
	cout << "========================" << endl;
	
	for( int i = 1; i < 5; ++i )
	{
		simulateStep();
		cout << endl << "Step " << i << ": time = "
			 << sim -> getTime() << endl;
	}

	cout << endl;
	cout << "continue until SimTime 6 is reached or passed:";
	simulateUntil( 6 );
	cout << endl  << "time = "
		 << sim -> getTime() << endl;

	cout << "========================" << endl;

	//
	// To prevent memory leaks, a few clean-up operations are necessary
	//
	sim -> exitSimulation();
	
	return 0;
}
//
//
// Running the program produces the following output:
//
// Basic Simulation Example
// ========================
//  *
// Step 1: time = 1
//  *
// Step 2: time = 2
//  *
// Step 3: time = 3
//  *
// Step 4: time = 4
//
// continue until SimTime 6 is reached or passed: *  *
// time = 6
// ========================
//
//
// In contrast to the basicProcess example there is a '*' at step 1 because
// event actions are atomic. Keep in mind that always all of the statements
// in eventAction() are executed, even if there are scheduling calls. In this
// case, the same Event object is rescheduled, but it still outputs the '*'.
// Once an Event's action is started, it cannot be stopped or postponed.
//
//
