//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2003, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file basicProcess.cpp

	\author Ralf Gerstenberger

	\date created at 2002/08/26

	\brief Example for basic simulation techniques

	This example shows the basic techniques for
	realising a simulation with ODEMx.

	\since 1.0
*/
/** \example basicProcess.cpp
	The basic simulation techniques of ODEMx are
	introduced.

	A process simulation contains a number of processes that describe
	the active elements in a model. A process contains a sequence of
	actions. These actions are timeless. Time consumption is realised
	with special time-operations. In this example several of this
	time-operations are presented together with an example process.
*/

#include <odemx/base/Process.h>

#include <iostream>

using namespace std;
using namespace odemx;

//
// This example Process is a simple SimTime clock.
// It outputs a '*' every full (1.0) tic in SimTime.
//
class Clock: public Process {
public:
	//
	// We label all instances of this class 'Clock'.
	// ODEMx changes the label on its own to prevent
	// confusion. It will add a number to the label
	// if more than one Clock is created in the
	// simulation. This procedure guarantees that
	// every label is unique within the same simulation
	// context.
	//
	// When we create a Clock object we will have to
	// provide a pointer to the Simulation we want it
	// to participate with.
	//
	Clock(Simulation* sim) : Process(sim, "Clock") {}

	//
	// The behaviour of a user process must be defined
	// by an implementation of the pure virtual function
	// 'int main()'. The return value of this function
	// is stored in the Process object and can be accessed
	// via the member function 'getReturnValue()'.
	// As long as 'main()' hasn't finished, the return
	// value is not valid. You can use the Process member
	// function 'hasReturned()' to check this condition.
	//
	virtual int main() {
		//
		// Because our process has a recurring behavior pattern,
		// we implement its 'life cycle' by using a loop.
		//
		while (true) {
			//
			// Since we are implementing a simple clock, we need
			// a means to let the SimTime pass. Here, we use
			// the member function 'holdFor()', which requires
			// a time period as parameter. We could also use
			// 'activateIn()' instead.
			// The difference between these two is seen if
			// there is more than one process scheduled at
			// the same time. 'holdFor' would than schedule
			// the current process as the last while 'activateIn'
			// would schedule the current process as the first one.
			// Therefore, hold... and activate... functions
			// can be used to choose between fifo or lifo scheduling.
			//
			holdFor( 1 );

			cout << " * ";
		}

		return 0;
	}
};

int main(int argc, char* argv[]) {
	//
	// For this example we can use the DefaultSimulation.
	// In more complex applications it is recommended to
	// provide a custom-made simulation class.
	//
	Simulation* sim = getDefaultSimulation();
	Clock* myClock = new Clock(sim);
	//
	// The new process is just created at this point.
	// It won't be executed because it is not activated.
	// A process can be activated by any of these methods:
	//
	// 'hold()'
	// 'holdFor()'
	// 'holdUntil()'
	// 'activate()'
	// 'activateIn()'
	// 'acitvateAt()'
	//
	// 'hold()' and 'activate()' are equal to 'holdFor(0.0)'
	// and 'activateIn(0.0)'. 'holdUntil(t)' and 'activateAt(t)'
	// are equal to 'holdFor(t-now)' and 'activateIn(t-now)'.
	//
	myClock -> activate();
	//
	// There are three ways to compute a simulation.
	// Firstly, you can 'run()' the simulation until it is
	// finished. A simulation is finished if there is no active
	// process scheduled or if it is stopped with 'exitSimulation()'.
	// Secondly, you can compute a simulation 'step()' by step.
	// Thirdly, you can run a simulation until a given SimTime is
	// reached or passed with 'runUntil()'.
	//
	// Because Clock processes are running forever, we shouldn't use
	// 'run()'. Instead we try both 'step()' and 'runUntil()'.
	//
	cout << "Basic Simulation Example" << endl;
	cout << "========================" << endl;

	for (int i=1; i<5; ++i) {
		sim->step();
		cout << endl << "Step " << i << ": time = " << sim->getTime() << endl;
	}

	cout << endl;
	cout << "continue until SimTime 6.0 is reached or passed:";
	sim->runUntil( 6 );
	cout << endl  << "time = " << sim->getTime() << endl;

	cout << "========================" << endl;

	//
	// To prevent memory leaks, a few clean-up opereations are necessary
	//
	sim->exitSimulation();
	delete myClock;

	return 0;
}
//
//
// Running the program produces the following output:
//
// Basic Simulation Example
// ========================
//
// Step 1: time = 0
//  *
// Step 2: time = 1
//  *
// Step 3: time = 2
//  *
// Step 4: time = 3
//
// continue until SimTime 6.0 is reached or passed: *  *  *
// time = 6
// ========================
//
//
// Keep in mind that at step 1 our clock will not output a '*'
// because that is the first step until the 'holdFor(1);' statement
// is reached. Only after that, the clock starts ticking, so to speak.
//
//

