//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file memoPortTimer.cpp

	\author Ronald Kluth
	<!-- [\author <author>]* -->

	\date created at 2007/10/01

	\brief A short usage example of Ports and Timers

	<!-- [detailed description] -->



	\since 2.0
*/



#include <iostream>
#include <string>
using namespace std;

#include <odemx/odemx.h>
using namespace odemx;

struct Msg: public PortData {

	Msg( const string& str ): content( str ) {}
	string content;
};


class Receiver: public Process {
public:

	Receiver():	Process( "Receiver" )
		{
			myTimer = new Timer( "Receiver Message Timer" );
			myPort = new PortHead( "Receiver Input Port" );
		}

	virtual ~Receiver()
		{
			delete myTimer;
			delete myPort;
		}

	virtual int main() {

		SimTime timeLimit = 10;

		myTimer -> setIn( timeLimit );

		while ( true ) {

			cout << getLabel() << " waiting for message with timeout at "
				 << myTimer -> getExecutionTime() << endl;

			Memory* who = wait( myTimer, myPort );

			cout << endl << "Alerter: \"" << who -> getLabel() << "\"" << endl;

			if ( who == myTimer ) {

				cout << getLabel() << " rescheduled by timeout at " << getCurrentTime() << endl;
				break;
			}
			else if ( who == myPort ) {

				myTimer -> reset( timeLimit );

				std::auto_ptr< PortData* > d = myPort -> get();

				Msg* m;

				if( d.get() != 0 )
					m = (Msg*) *d;
				else
					cout << "PortHead::get() returned 0" << endl;

				cout << getLabel() << " received message \""
					 << m -> content << "\" at " << getCurrentTime() << endl;

				delete m; m = 0;
			}
		}

		cout << getLabel() << "-Error: no message received." << endl;
		cout << getLabel() << " terminated at " << getCurrentTime() << endl;

		return 0;
	}

	PortTail* getInputPort() {
		return myPort -> getTail();
	}

	Timer* getTimer() {
		return myTimer;
	}

private:

	Timer* myTimer;
	PortHead* myPort;
};


class MessageArrival: public Event {
public:

	MessageArrival( Receiver* p, Msg* m ):
		Event( "Message Arrival" ),
		myReceiver( p ),
		myMessage( m )
		{}

	virtual void eventAction() {

		cout << "\"" << getLabel() << "\" Event for " << myReceiver -> getLabel() << endl;
		myReceiver -> getInputPort() -> put( myMessage );
		myReceiver -> getInputPort() -> put( new Msg( "1" ) );
		myReceiver -> getInputPort() -> put( new Msg( "2" ) );
		myReceiver -> getInputPort() -> put( new Msg( "3" ) );
		myReceiver -> getInputPort() -> put( new Msg( "4" ) );
	}

private:

	Receiver* myReceiver;
	Msg* myMessage;
};



/************ MAIN **************************************************/

int main( int argc, char* argv[] ) {

	SimTime arrivalTime = 8;

	if ( argc > 1 ) {

		arrivalTime = atof( argv[1] );

		cerr << "Using command line message arrival time: 	" << arrivalTime
			 << endl << endl;
	}


	Simulation* sim = getDefaultSimulation();

	XmlTrace* trace = new XmlTrace( "MemoPortTimer" );

	sim -> addConsumer( trace );
	sim -> startTrace();

	Receiver* receiver = new Receiver();
	Event* msgArr = new MessageArrival( receiver, new Msg("Port-Timer-Example") );

	msgArr -> scheduleAt( arrivalTime );
	receiver -> activate();

	sim -> printExecutionList(); cout << endl;
	sim -> run();

	XmlReport reporter( sim, "MemoPortTimerReport.xml" );
	reporter.addProducer( receiver -> getInputPort() );
	reporter.addProducer( receiver -> getInputPort() -> getHead() );
	reporter.addProducer( receiver -> getTimer() );

	reporter.generateReport();

	/* CLEAN UP */
	delete receiver -> getInputPort();
	delete receiver;
	delete msgArr;

	sim -> stopTrace();
	delete trace;

	return 0;
}

