//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file NetTopology.cpp

	\author Ronald Kluth
	<!-- [\author <author>]* -->

	\date created at 2007/04/04

	\brief Implementation of classes in NetTopology.h

	\sa NetTopology.h

	<!-- [detailed description] -->

	\since 2.0
*/

#include <odemx/protocol/NetTopology.h>
#include <odemx/protocol/ProtocolStack.h>
#include <odemx/util/ErrorHandling.h>
#include <string>
#include <cassert>
#include <stdexcept>
#include <utility>

using namespace odemx;
using namespace std;

NetTopology::NetTopology( ProtocolSimulation* sim, NetTopologyObserver* o ):
	Observable<NetTopologyObserver>( o ),
	env( sim )
	{}

NetTopology::~NetTopology()
	{}

// make a node known to transmission
void NetTopology::registerNode( ProtocolStack* n, double locationX, double locationY )
{
	//trace
	getTrace() -> mark( this, markRegisterNode, n );

	// observer
	_obsForEach( NetTopologyObserver, RegisterNode( this, n, locationX, locationY ) );

	// store new node
	allNodes.insert( map< NodeID, NodeInfo >::value_type(
			n -> getNodeId(),
			NodeInfo( n -> getNodeId(), n, locationX, locationY ) ) );
}


// remove node from nodeLinks vector and the node list
void NetTopology::removeNode( ProtocolStack* n )
{
	assert( dynamic_cast< ProtocolStack*>( n ) );

	NetworkMap::iterator del = allNodes.find( n -> getNodeId() );

	if ( del == allNodes.end() )
	{
		error( this,  "NetTopology::removeNode(): node not stored in list" );
		return;
	}

	//trace
	getTrace() -> mark( this, markRemoveNode, n );

	// observer
	_obsForEach( NetTopologyObserver, RemoveNode( this, n ) );

	// erase node n from all neighbor lists (node maps)
	NetworkMap::iterator iter;
	for( iter = allNodes.begin(); iter != allNodes.end(); ++iter )
	{
		iter -> second.links.erase( n );
	}

	// erase node from node list
	allNodes.erase( del );
}


bool NetTopology::linked( ProtocolStack* node1, ProtocolStack* node2 ) {
	assert( node1 && node2 );

	bool linked = false;

	try
	{
		// test if node2 is in the neighbor list of node1
		linked = getLinks( node1 ).find( node2 ) != getLinks( node1 ).end();
	}
	catch( std::runtime_error)
	{
		return false;
	}
	return linked;
}


// set connectivity for one direction, assuming there was non before
void NetTopology::makeLink( ProtocolStack* n1, ProtocolStack* n2, double p ) {
	assert( n1 && n2 );

	if( !validProbability(p) )
		return;

	//trace
	getTrace() -> beginMark( this, markMakeLink );
	getTrace() -> addTag( tagFrom, n1 );
	getTrace() -> addTag( tagTo, n2 );
	getTrace() -> addTag( tagQuality, p );
	getTrace() -> endMark();

	// observer
	_obsForEach( NetTopologyObserver, MakeLink( this, n1, n2, p ) );

	// set connectivity between two nodes
	// new Draw(..) will automatically receive
	// a new seed within one Simulation context
	getLinks( n1 ).insert( map< ProtocolStack*, LinkInfo >::value_type(
			n2,
			LinkInfo( new Draw( env, "LinkDraw", p), p, 1 ) ) );
}


// change the Link settings between n1 and n2, assuming there was one before
void NetTopology::changeLink( ProtocolStack* n1, ProtocolStack* n2, double p ) {
	assert( n1 && n2 );

	if ( !validProbability(p) )
		return;

	//trace
	getTrace() -> beginMark( this, markChangeLink );
	getTrace() -> addTag( tagFrom, n1 );
	getTrace() -> addTag( tagTo, n2 );
	getTrace() -> addTag( tagQuality, p );
	getTrace() -> endMark();

	// observer
	_obsForEach( NetTopologyObserver, ChangeLink( this, n1, n2, p ) );

	// reset LinkInfo for n2 in n1's neighbor map
	LinkInfo& link = getLinks( n1 ).find( n2 )->second;
	if( link.loss )
	{
		delete ( link.loss );
		link.loss = 0;
		link.p = 0;
	}

	// set connectivity between the two nodes
	// new Draw(..) will automatically receive
	// a new seed within one Simulation context
	link.loss = new Draw ( env, "LinkDraw", p);
	link.p = p;
}



void NetTopology::removeLink( ProtocolStack* n1, ProtocolStack* n2 ) {

	if( ! ( linked( n1, n2 ) && linked( n2, n1 ) ) )
	{
		warning( this,  string( "NetTopology::removeLink(): no previous link between nodes: " )
						+ n1 -> getLabel() + ", " + n2 -> getLabel() );
	}

	getLinks( n1 ).erase( n2 );
	getLinks( n2 ).erase( n1 );

	//trace
	getTrace() -> beginMark( this, markRemoveLink );
	getTrace() -> addTag( tagNode, n1 );
	getTrace() -> addTag( tagNode, n2 );
	getTrace() -> endMark();

	// observer
	_obsForEach( NetTopologyObserver, RemoveLink( this, n1, n2 ) );
}

void NetTopology::removeLink( const Label& l1, const Label& l2 )
{
	ProtocolStack* node1 = 0;
	ProtocolStack* node2 = 0;

	// find node1 and node2 by comparing labels
	NetworkMap::const_iterator i;
	for( i = allNodes.begin(); i != allNodes.end(); ++i )
	{
		if ( node1 && node2 )
			break;

		assert( dynamic_cast< ProtocolStack* >( i -> second.address ) );

		if( ( i -> second.address -> getLabel() ).compare( l1 ) == 0 )
			node1 = i -> second.address;

		if( ( i -> second.address -> getLabel() ).compare( l2 ) == 0 )
			node2 = i -> second.address;
	}

	// if either one not found output error message, else set link
	if( node1 && node2 )
	{
		removeLink( node1, node2 );
	}
	else
	{
		error( this,  "NetTopology::removeLink(): cannot determine node by Label, is it registered?" );
	}
}

// set same connectivity for bi-directional links
void NetTopology::setLink( ProtocolStack* n1, ProtocolStack* n2, double p_both ) {

	if( ! validProbability( p_both ) )
		return;

	// in case of previous link,
	// the old draw object must be destroyed
	if( linked( n1, n2 ) && linked( n2, n1 ) )
	{
		changeLink( n1, n2, p_both );
		changeLink( n2, n1, p_both );
	}
	else
	{
		makeLink( n1, n2, p_both );
		makeLink( n2, n1, p_both );
	}
}


void NetTopology::setLink( const Label& l1, const Label& l2, double p_both ) {

	ProtocolStack* node1 = 0;
	ProtocolStack* node2 = 0;

	// find node1 and node2 by comparing labels
	NetworkMap::const_iterator i;
	for( i = allNodes.begin(); i != allNodes.end(); ++i )
	{
		if ( node1 && node2 )
			break;

		if( ( i -> second.address -> getLabel() ).compare( l1 ) == 0 )
			node1 = i -> second.address;

		if( ( i -> second.address -> getLabel() ).compare( l2 ) == 0 )
			node2 = i -> second.address;
	}

	// if either one not found output error message, else set link
	if( node1 && node2 )
	{
		setLink( node1, node2, p_both );
	}
	else
	{
		error( this,  "NetTopology::setLink(): cannot determine node by Label, is it registered?" );
	}
}


// set connectivity for one direction only
void NetTopology::setLink( ProtocolStack* n1, ProtocolStack* n2, double p_to, double p_from ) {

	if ( !validProbability( p_to ) || !validProbability( p_from ) )
		return;

	// when changing link settings,
	// the old draw object must be destroyed
	if ( linked( n1, n2 ) && linked( n2, n1 ) )
	{
		changeLink( n1, n2, p_to );
		changeLink( n2, n1, p_from );
	}
	else
	{
		makeLink( n1, n2, p_to );
		makeLink( n2, n1, p_from );
	}
}

// set connectivity for one direction only
void NetTopology::setLink( const Label& l1, const Label& l2, double p_to, double p_from ) {

	ProtocolStack* node1 = 0;
	ProtocolStack* node2 = 0;

	// find node1 and node2 by comparing labels
	NetworkMap::const_iterator i;
	for( i = allNodes.begin(); i != allNodes.end(); ++i )
	{
		if ( node1 && node2 )
			break;

		if( ( i -> second.address -> getLabel() ).compare( l1 ) == 0 )
			node1 = i -> second.address;

		if( ( i -> second.address -> getLabel() ).compare( l2 ) == 0 )
			node2 = i -> second.address;
	}

	// if either one not found output error message, else set link
	if( node1 && node2 )
	{
		setLink( node1, node2, p_to, p_from );
	}
	else
	{
		error( this,  "NetTopology::setLink(): cannot determine node by Label, is it registered?" );
	}
}

double NetTopology::getConnectivitySample( ProtocolStack* sendNode, ProtocolStack* receiveNode )
{
	assert( linked( sendNode, receiveNode ) );
	return getLinks( sendNode ).find( receiveNode )->second.loss -> sample();
}


const NetworkMap& NetTopology::getAllNodes() const
{
	return allNodes;
}

ProtocolStack* NetTopology::getNodeById( NodeID nid ) const
{
	NetworkMap::const_iterator found = allNodes.find( nid );
	if( found == allNodes.end() )
	{
		error( this, "NetTopology::getNodeById(): node not found, is it registered?" );
		return 0;
	}
	return found -> second.address;
}

map< ProtocolStack*, LinkInfo >& NetTopology::getLinks( ProtocolStack* n )
{
	NetworkMap::iterator found = allNodes.find( n -> getNodeId() );

	if( found == allNodes.end() )
	{
		error( this, string("NetTopology::getLinks(): ")
					 + n -> getLabel()
					 + " not found, is it registered?" );
		// no other way to prevent the crash and signal an error
		throw std::runtime_error( "NetTopology::getLinks(): node not found." );
	}

	return found -> second.links;
}


bool NetTopology::setNodeLocation( ProtocolStack* n, double locationX, double locationY )
{
	NetworkMap::iterator found = allNodes.find( n -> getNodeId() );

	if( found == allNodes.end() )
	{
		error( this, string("NetTopology::setNodeLocation(): ")
					 + n -> getLabel()
					 + " not found, is it registered?" );
		return false;
	}

	found -> second.location.x = locationX;
	found -> second.location.y = locationY;
	return true;
}

const Location NetTopology::getNodeLocation( ProtocolStack* n ) const
{
	NetworkMap::const_iterator found = allNodes.find( n -> getNodeId() );
	if( found == allNodes.end() )
	{
		error( this, string("NetTopology::getNodeLocation(): ")
					 + n -> getLabel()
					 + " not found, is it registered?" );
		// no other way to prevent the crash and signal an error
		throw std::runtime_error( "NetTopology::getNodeLocation(): node not found." );
	}
	return found -> second.location;
}

Trace* NetTopology::getTrace()
{
	return static_cast< Trace* >( env );
}

ProtocolSimulation* NetTopology::getSimulation() const
{
	return env;
}

// print current Matrix for debug purposes
void NetTopology::printNodeMatrix() {

	NetworkMap::const_iterator i;
	map< ProtocolStack*, LinkInfo >::const_iterator j;    // *j is a pair

	// print matrix entries
	for( i = allNodes.begin(); i != allNodes.end(); ++i )
	{
		const NodeInfo& node = i -> second;
		std::cout << node.address -> getLabel() << std::endl;

		// print matrix line
		for( j = node.links.begin(); j != node.links.end(); ++j)
		{
			std::cout << "  " << j -> first -> getLabel()
					  << setw(10) << fixed << setprecision(6)
					  << node.links.find( j -> first ) -> second.p
					  << std::endl;
		}

		std::cout << std::endl;
	}
	std::cout << resetiosflags(ios_base::fixed) << std::endl;
}


bool NetTopology::validProbability( double p )
{
	if ( p < 0 )
	{
		error( this,  "NetTopology::validProbability(): negative connectivity value" );
		return false;
	}
	if ( p > 1 )
	{
		error( this,  "NetTopology::validProbability(): connectivity greater 1" );
		return false;
	}
	return true;
}


const MarkTypeId NetTopology::baseMarkId = MarkType::getNewBaseMarkId( typeid(NetTopology) );

const MarkType NetTopology::markRegisterNode( "register node", baseMarkId+3, typeid(NetTopology) ); ///< New node was registered with the network, it is now known to the TransmissionMedium
const MarkType NetTopology::markRemoveNode( "remove node", baseMarkId+4, typeid(NetTopology) ); ///< Node was removed from the network
const MarkType NetTopology::markMakeLink( "make link", baseMarkId+5, typeid(NetTopology) ); ///< Unidirectional link between two nodes was set with a given probability
const MarkType NetTopology::markChangeLink( "change link", baseMarkId+6, typeid(NetTopology) ); ///< Unidirectional link probability between two nodes was changed
const MarkType NetTopology::markRemoveLink( "remove link", baseMarkId+7, typeid(NetTopology) ); ///< Entirely removed links between two nodes, the random variable representing the link is deleted

const TagId NetTopology::baseTagId = NetTopology::baseMarkId;

const Tag NetTopology::tagFrom( "from", baseTagId+1 ); ///< Displays detail about the originating node
const Tag NetTopology::tagTo( "to", baseTagId+2 ); ///< Displays detail about the receiving node
const Tag NetTopology::tagQuality( "quality", baseTagId+3 ); ///< Displays detail about the link quality
const Tag NetTopology::tagNode( "node", baseTagId+4 ); ///< Displays details about disconnected nodes
