//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file NetTopology.h

	\author Ronald Kluth
	<!-- [\author <author>]* -->

	\date created at 2007/04/04

	\brief Declaration of odemx::NetTopology

	\sa NetTopology.cpp

	<!-- [detailed description] -->

	\since 2.0
*/

#ifndef ODEMX_NETTOPOLOGY_INCLUDED
#define ODEMX_NETTOPOLOGY_INCLUDED

#include <map>
#include <vector>
#include <iostream>
#include <iomanip>  // for pretty printing

#include <odemx/util/MarkType.h>
#include <odemx/util/Trace.h>
#include <odemx/protocol/ProtocolStack.h>
#include <odemx/protocol/ProtocolSimulation.h>
#include <odemx/random/DiscreteDist.h>

namespace odemx {

	// forward declaration
	class NetTopologyObserver;

	/** \struct Location
		\ingroup protocol
		\author Ronald Kluth
		\brief %Location is a structure for storing network node locations
		\sa ProtocolStack, NetTopology
		\since 2.1
	*/
	struct Location
	{
		Location( double locX, double locY )
		:	x( locX ),
			y( locY )
			{}

		double x;
		double y;
	};

	/** \struct LinkInfo
		\ingroup protocol
		\author Ronald Kluth
		\brief %LinkInfo is a structure for maintaining data about node pair connections
		\sa ProtocolStack, NetTopology

		LinkInfo is used to describe the connection between two nodes
		in terms of quality.

		\since 2.0
	*/
	struct LinkInfo
	{
		/// Construction
		LinkInfo( DiscreteDist* rand, double prob, double distance )
		:	loss( rand ),
			p( prob ),
			nodeDistance( distance )
			{}

		/// Destruction
		~LinkInfo() { /*if ( loss ) delete loss;*/ }

		DiscreteDist* loss;		///< random variable for message loss
		double p;				///< explicit loss probability
		double nodeDistance;	///< distance between two nodes
							    // multiplied with data for transmission delay
	private:
		/// Default Construction
		LinkInfo(): loss( 0 ), p( 0 ), nodeDistance( 1 ) {}
	};

	struct NodeInfo
	{
		NodeInfo( NodeID id, ProtocolStack* p, double locX, double locY )
		:	id( id ),
			address( p ),
			location( locX, locY ),
			links()
			{}

		NodeID id;
		ProtocolStack* address;
		Location location;
		std::map< ProtocolStack*, LinkInfo > links;

	private:
		// a private constructor prevents the NetworkMap from making
		// uninitalized entries, but also disallows use of operator[]
		NodeInfo();
	};

	typedef std::map< NodeID, NodeInfo > NetworkMap;


	/** \class NetTopology
		\ingroup protocol
		\author Ronald Kluth
		\brief %NetTopology is a class for modeling the structure of a network
		\sa ProtocolStack, LinkInfo

		A network's topology is described by nodes and the links with
		their neighbors. This class provides the means to model the structure
		of a communications network and set the link quality between
		nodes and their neighbors.

		\since 2.0
	*/
	class NetTopology: public virtual TraceProducer,
					   public Observable<NetTopologyObserver>
	{
	public:
		/**
			\brief Construction
			\param sim
				pointer to the ProtocolSimulation object
		*/
		NetTopology( ProtocolSimulation* sim, NetTopologyObserver* o = 0 );
		/// Destruction
		virtual ~NetTopology();

		/**
			\name Node handling

			These functions provide access to the node structure of
			a network. Nodes can be registered or removed, and
			a list of all network nodes is available.

			@{
		*/
		/**
			\brief make a node known to the net
			\param n
				new ProtocolStack

			The new node \p n will be added to the LinkMatrix, but by defaut
			it is not connected to any other nodes. The connection quality
			to its neighbors must be set afterwards.
		*/
		void registerNode( ProtocolStack* n, double locationX, double locationY );

		/**
			\brief remove a node from the network structure
			\param n
				ProtocolStack to be removed

			The node \p n will only be removed from the LinkMatrix. It will
			not be deleted by this function.
		*/
		void removeNode( ProtocolStack* n );

		/**
			\brief get a pointer to a node object from its ID
			\param nid
				id of the node object to find
			\return
				pointer to the object at position \p nid in allNodes
		 */
		ProtocolStack* getNodeById( NodeID nid ) const;

		/**
			\brief get a vector listing all nodes currently in the network
			\return
				reference to a NodeVector containing all nodes
		*/
		const NetworkMap& getAllNodes() const;

		/**
			\brief get a map listing all neighbors of \c n and the connection
				quality between n and each one of them
			\return
				reference to a map containing all neighbors and link info
			\throw
				std::runtime_error if \p n  is not registered
		*/
		std::map< ProtocolStack*, LinkInfo >& getLinks( ProtocolStack* n );

		/**
			\brief set a node's location in the network
			\return
				\c true if successful, \c false if node not registered
		*/
		bool setNodeLocation( ProtocolStack* n, double locationX, double locationY );

		/**
			\brief get a node's physical location
			\return
				a Location object for \c n
			\throw
				std::runtime_error if \p n is not registered
		*/
		const Location getNodeLocation( ProtocolStack* n ) const;
		//@}

		/**
			\name Link handling

			These functions manage the link quality (connectivity)
			between nodes. Links can be set bi-directionally or with each
			direction separately. The LinkMatrix stores LinkInfo objects
			for both directions. The default connectivity is 0.

			@{
		*/
		/**
			\brief set connectivity for bi-directional links
			\param n1
				pointer to first Node
			\param n2
				pointer to second Node
			\param p_both
				probability of message arrival

			The connectivity between the two nodes will be set to \p p_both
			and a corresponding random variable will be initialized. Both
			directions will get the same message arrival probability.
		*/
		void setLink( ProtocolStack* n1, ProtocolStack* n2, double p_both );

		/**
			\brief set connectivity for bi-directional links by label
			\param l1
				label of first Node
			\param l2
				label of second Node
			\param p_both
				probability of message arrival

			The connectivity between the two nodes will be set to \p p_both
			and a corresponding random variable will be initialized. Both
			directions will get the same message arrival probability.
		*/
		void setLink( const Label& l1, const Label& l2, double p_both );

		/**
			\brief set connectivity for each direction
			\param n1
				pointer to first Node
			\param n2
				pointer to second Node
			\param p_to
				probability of message arrival from \p n1 to \p n2
			\param p_from
				probability of message arrival from \p n2 to \p n1

			The connectivity between the two nodes will be set to \p p_to
			\p p_from respectively. Corresponding random variables will
			be initialized.
		*/
		void setLink( ProtocolStack* n1, ProtocolStack* n2, double p_to, double p_from );

		/**
			\brief set connectivity for each direction
			\param l1
				pointer to first Node
			\param l2
				pointer to second Node
			\param p_to
				probability of message arrival from \p n1 to \p n2
			\param p_from
				probability of message arrival from \p n2 to \p n1

			The connectivity between the two nodes will be set to \p p_to
			\p p_from respectively. Corresponding random variables will
			be initialized.
		*/
		void setLink( const Label& l1, const Label& l2, double p_to, double p_from );

		/**
			\brief cut link for both directions
			\param n1
				pointer to first Node
			\param n2
				pointer to second Node
			\param both
				set \p false if only the direction \p n1 to \p n2 is to be disabled

			The connection quality between the two nodes will be reset to \p 0,
			and the corresponding random variables will be deleted. After
			calling this function, the two nodes will have no connection
			at all.
		*/
		void removeLink( ProtocolStack* n1, ProtocolStack* n2 );

		/**
			\brief cut link for both directions
			\param n1
				Label of first Node
			\param n2
				Label of second Node
			\param both
				set \p false if only the direction \p n1 to \p n2 is to be disabled

			The connection quality between the two nodes will be reset to \p 0,
			and the corresponding random variables will be deleted. After
			calling this function, the two nodes will have no connection
			at all.
		*/
		void removeLink( const Label& n1, const Label& n2 );

		/**
			\brief check if two nodes are linked
			\param n1
				pointer to first Node
			\param n2
				pointer to second Node
			\return
				true if a link was set between \p n1 and \p n2

			\note This function only checks whether \p n2 is in the listed as
			neighbor of \p n1, and not both directions.
		*/
		bool linked( ProtocolStack* n1, ProtocolStack* n2 );

		///	print current connection matrix for debug purposes
		void printNodeMatrix();

		/// get the trace object
		Trace* getTrace();

		/// get the ProtocolSimulation object
		ProtocolSimulation* getSimulation() const;

	protected:

		/// get a random connectivity value from LinkInfo::loss
		double getConnectivitySample( ProtocolStack* sendNode, ProtocolStack* receiveNode );

	private:

		/**
			\brief check if the given value is between 0 and 1
			\param p
				connection probability
			\return
				\p true if given probability is between 0 and 1
		*/
		bool validProbability( double p );

		/**
			\brief create a link between nodes
			\param n1
				pointer to first Node
			\param n2
				pointer to second Node
			\param p
				connection probability

			This internal function creates a link between two
			given nodes in direction \p n1 to \p n2 only. A random
			variable with the specified probability will be created.
		*/
		void makeLink( ProtocolStack* n1, ProtocolStack* n2, double p );

		/**
			\brief change a link between nodes
			\param n1
				pointer to first Node
			\param n2
				pointer to second Node
			\param p
				connection probability

			This internal function changes an existing connection between two
			given nodes in direction \p n1 to \p n2 only. A random
			variable with the specified probability will be created.
		*/
		void changeLink( ProtocolStack* n1, ProtocolStack* n2, double p );

	private:

		/// list of all nodes in the network, indexed by NodeIDs
		NetworkMap allNodes;
		/// pointer to the ProtocolSimulation object (context)
		ProtocolSimulation* env;

	public:

		/**
			\name Trace MarkTypes

			These MarkTypes and Tags are used to trace NetTopology calls.
			A TraceConsumer can use these constants to identify
			trace events sent by NetTopology.

			@{
		*/
		static const MarkTypeId baseMarkId;

		static const MarkType markRegisterNode;
		static const MarkType markRemoveNode;
		static const MarkType markMakeLink;
		static const MarkType markChangeLink;
		static const MarkType markRemoveLink;

		static const TagId baseTagId;

		static const Tag tagFrom;
		static const Tag tagTo;
		static const Tag tagQuality;
		static const Tag tagNode;
		//@}
	};

	/** \interface NetTopologyObserver

		\author Ronald Kluth
		<!-- [\author <author>]* -->

		\brief Observer for NetTopology specific simulation events.

		<!-- [\note {notes}]* -->

		\sa NetTopology, TransmissionMedium

		<!-- [detailed description] -->

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.0
	*/
	class NetTopologyObserver {

	public:
		/// Destructor
		virtual ~NetTopologyObserver() {}

		/// Register new ProtocolStack
		virtual void onRegisterNode( NetTopology* sender, ProtocolStack* n, double locX, double locY ) {}
		/// Remove an existing ProtocolStack
		virtual void onRemoveNode( NetTopology* sender, ProtocolStack* n ) {}
		/// Create a link between two Nodes
		virtual void onMakeLink( NetTopology* sender, ProtocolStack* n1, ProtocolStack* n2, double p ) {}
		/// Change the link of two Nodes
		virtual void onChangeLink( NetTopology* sender, ProtocolStack* n1, ProtocolStack* n2, double p ) {}
		/// Disable the connection between two NetNodes
		virtual void onRemoveLink( NetTopology* sender, ProtocolStack* n1, ProtocolStack* n2 ) {}
	};
}

#endif /* ODEMX_NETTOPOLOGY_INCLUDED */

