//----------------------------------------------------------------------------
//	Copyright (C) 2007 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ProtocolEntity.cpp

	\author Ronald Kluth
	<!-- [\author <author>]* -->

	\date created at 2007/04/04

	\brief Implementation of classes in ProtocolEntity.h

	\sa ProtocolEntity.h

	<!-- [detailed description] -->

	\since 2.0
*/

#include <odemx/protocol/ProtocolEntity.h>
#include <odemx/protocol/ProtocolLayer.h>
#include <odemx/protocol/ProtocolStack.h>
#include <odemx/protocol/TransmissionMedium.h>

#include <string>

using namespace odemx;

ProtocolEntity::ProtocolEntity( ProtocolSimulation* sim, ProtocolLayer* layer,
								ProtocolStack* owner, ProtocolEntityObserver* o )
	// entity labels automatically get their owner and layer added in front
:	Process( sim, ( owner->getLabel() + " " + layer->getLabel() + " Entity"), o ),
	Observable<ProtocolEntityObserver>( o ),
	node( owner ),
	top( 0 ),
	bottom( 0 )
	{
		topPort = new PortHead( sim, ( Process::getLabel() + " Top Port" ) );
		bottomPort = new PortHead( sim, ( Process::getLabel() + " Bottom Port" ) );
	}

ProtocolEntity::~ProtocolEntity()
	{
		// port destructor will not remove the port
		// if there is another process connected to head/tail
		if ( topPort )
			delete topPort;
		if ( bottomPort )
			delete bottomPort;
	}

int ProtocolEntity::main()
{
	Memory* theAlerter = 0;

	while (true)
	{
		// check input ports and wait for message arrival
		theAlerter = wait( topPortHead(), bottomPortHead() );

		if ( theAlerter == topPortHead() )
		{
			handleUpperLayerInput( readTop() );
		}
		else if ( theAlerter == bottomPortHead() )
		{
			handleLowerLayerInput( readBottom() );
		}
		else if ( theAlerter == 0 )
		{
			if ( isInterrupted() )
			{
				warning( this,  "ProtocolEntity::main(): wait state interrupted by another process" );
			}
			else
			{
				error( this,  "ProtocolEntity::main(): wait state alerter is 0" );
			}
		}
		else
		{
			error( this,  "ProtocolEntity::main(): cannot identify alerter" );
			break;
		}

		theAlerter = 0;
	}

	return 0;
}


void ProtocolEntity::transmit( PortData* e )
{
	// trace
	getTrace() -> mark( this, markTransmit );

	//observer
	_obsForEach( ProtocolEntityObserver, Transmit( this, e ) );

	if ( this == getNode() -> getBottomLayerEntity() )
	{
		getTransmissionMedium() -> transmit( (ProtocolMessage*)e );
	}
	else
	{
		getBottomEntity() -> putTop( e );
	}
}


void ProtocolEntity::passUpward( PortData* e )
{
	// trace
	getTrace() -> mark( this, markPassUpward );

	//observer
	_obsForEach( ProtocolEntityObserver, PassUpward( this, e ) );

	if ( this == getNode() -> getTopLayerEntity() )
	{
		getNode() -> _receivedData( e );
	}
	else
	{
		getTopEntity() -> putBottom( e );
	}
}

/********** port access **************************************/

PortHead* ProtocolEntity::topPortHead() {
	return topPort;
}


PortTail* ProtocolEntity::getTopPortTail() {
	return topPort -> getTail();
}


PortHead* ProtocolEntity::bottomPortHead() {
	return bottomPort;
}


PortTail* ProtocolEntity::getBottomPortTail() {
	return bottomPort -> getTail();
}

/****************** port manipulation ***********************/

// read the ProtocolEntity's ports
PortData* ProtocolEntity::readTop()
{
	// trace
	getTrace() -> mark( this, markReadTop );

	// PortHead::get() returns a pointer to the ElementType
	PortData* e = *(topPort -> get());

	//observer
	_obsForEach( ProtocolEntityObserver, ReadTop( this, e ) );

	return e;
}

PortData* ProtocolEntity::readBottom()
{
	// trace
	getTrace() -> mark( this, markReadBottom );

	// PortHead::get() returns a pointer to the ElementType
	PortData* e = *(bottomPort -> get());

	//observer
	_obsForEach( ProtocolEntityObserver, ReadBottom( this, e ) );

	return e;
}

// call PortTail::put()
void ProtocolEntity::putTop( PortData* e )
{
	getTopPortTail() -> put( e );
}

void ProtocolEntity::putBottom( PortData* e )
{
	getBottomPortTail() -> put( e );
}

/********** setters and getters ******************************/

ProtocolStack* ProtocolEntity::getNode()
{
	return node;
}

TransmissionMedium* ProtocolEntity::getTransmissionMedium()
{
	return ( (ProtocolSimulation*)getSimulation() ) -> getTransmissionMedium();
}

void ProtocolEntity::setTopEntity( ProtocolEntity* e )
{
	top = e;
}

void ProtocolEntity::setBottomEntity( ProtocolEntity* e )
{
	bottom = e;
}

ProtocolEntity* ProtocolEntity::getTopEntity()
{
	return top;
}

ProtocolEntity* ProtocolEntity::getBottomEntity()
{
	return bottom;
}

const MarkTypeId ProtocolEntity::baseMarkId = MarkType::getNewBaseMarkId( typeid(ProtocolEntity) );

const MarkType ProtocolEntity::markReadTop( "handle top input", baseMarkId+1, typeid(ProtocolEntity) ); ///< ProtocolEntity received input from upper layer and handles the given data
const MarkType ProtocolEntity::markReadBottom( "handle bottom input", baseMarkId+2, typeid(ProtocolEntity) ); ///< ProtocolEntity received input from lower layer and handles the given data
const MarkType ProtocolEntity::markTransmit( "transmit", baseMarkId+3, typeid(ProtocolEntity) ); ///< ProtocolEntity passes data to the lower layer (or transmits via TransmissionMedium)
const MarkType ProtocolEntity::markPassUpward( "pass upward", baseMarkId+4, typeid(ProtocolEntity) ); ///< ProtocolEntity passes data to the upper layer
