//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ProtocolLayer.cpp

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Implementation of classes in ProtocolLayer.h

	\sa ProtocolLayer.h

	<!-- [detailed description] -->

	\since 2.0
*/

#include <odemx/protocol/ProtocolLayer.h>
#include <odemx/protocol/ProtocolSimulation.h>
#include <odemx/protocol/ProtocolEntity.h>

#include <iostream>
#include <algorithm> // for_each

template< typename PtrType >
void delete_ptr( PtrType* p )
{
	delete p;
}

using namespace odemx;

ProtocolLayer::ProtocolLayer( ProtocolSimulation* sim, Label l, ProtocolLayerObserver* o ):
	DefLabeledObject( sim, l ),
	Observable<ProtocolLayerObserver>( o ),
	env( sim )
	{
		env -> layerManager.addLayer( this );

		//trace
		getTrace() -> mark( this, markCreate );

		// observer
		_obsForEach( ProtocolLayerObserver, Create( this ) );
	}

ProtocolLayer::~ProtocolLayer()
	{
		env -> layerManager.removeLayer( this );

		// clean up the entities that were created by this layer
		for_each( entityList.begin(), entityList.end(), &delete_ptr<ProtocolEntity> );

		//trace
		getTrace() -> mark( this, markDestroy );

		// observer
		_obsForEach( ProtocolLayerObserver, Destroy( this ) );
	}

void ProtocolLayer::addEntity( ProtocolEntity* e ) {

	//trace
	getTrace() -> mark( this, markAddEntity, e );

	// observer
	_obsForEach( ProtocolLayerObserver, AddEntity( this, e ) );

	entityList.push_back( e );
}

void ProtocolLayer::removeEntity( ProtocolEntity* e ) {

	//trace
	getTrace() -> mark( this, markRemoveEntity, e );

	// observer
	_obsForEach( ProtocolLayerObserver, RemoveEntity( this, e ) );

	ProtocolEntityList::iterator iter;
	iter = find( entityList.begin(), entityList.end(), e );

	if ( iter != entityList.end() )
		entityList.erase( iter );
	else
		error( this, "ProtocolLayer::removeEntity(): protocol entity not in this layer" );
}

void ProtocolLayer::printEntityList() {

	std::cout << "..........................." << std::endl;

	ProtocolEntityList::iterator iter;
	for( iter = entityList.begin(); iter != entityList.end(); ++iter ) {
		std::cout << (*iter) -> getLabel() << std::endl;
	}

	std::cout << "..........................." << std::endl;
}

ProtocolEntityList& ProtocolLayer::getEntityList() {
	return entityList;
}

ProtocolSimulation* ProtocolLayer::getSimulation() const {
	return env;
}

Trace* ProtocolLayer::getTrace() const {
	return static_cast<Trace*>(env);
}

ProtocolLayerManager::ProtocolLayerManager()
:	destructing( false )
	{}

ProtocolLayerManager::~ProtocolLayerManager()
{
	destructing = true;
	for_each( allLayersTopDown.begin(), allLayersTopDown.end(), &delete_ptr<ProtocolLayer> );
}

void ProtocolLayerManager::addLayer( ProtocolLayer* layer )
{
	layerNameMap[ layer -> getLabel() ] = layer;
	allLayersTopDown.push_back( layer );
}

void ProtocolLayerManager::removeLayer( ProtocolLayer* layer )
{
	// iterator crash when this class's destructor is called
	// ugly workaround: destructing layers may not access
	// allLayersTopDown because there is a for_each running on the list
	if( destructing ) return;

	layerNameMap.erase( layer -> getLabel() );
	allLayersTopDown.erase( find( allLayersTopDown.begin(), allLayersTopDown.end(), layer ) );
}

ProtocolLayerList& ProtocolLayerManager::getLayerList()
{
	return allLayersTopDown;
}

ProtocolLayer* ProtocolLayerManager::getLayerByLabel( Label l )
{
	return layerNameMap[ l ];
}

const MarkTypeId ProtocolLayer::baseMarkId = MarkType::getNewBaseMarkId( typeid(ProtocolLayer) );

const MarkType ProtocolLayer::markCreate( "create", baseMarkId+1, typeid(ProtocolLayer) ); ///< Call of ProtocolLayer constructor
const MarkType ProtocolLayer::markDestroy( "destroy", baseMarkId+2, typeid(ProtocolLayer) ); ///< Call of ProtocolLayer destructor
const MarkType ProtocolLayer::markAddEntity( "add entity", baseMarkId+3, typeid(ProtocolLayer) ); ///< New ProtocolEntity of a node is created within a ProtocolLayer
const MarkType ProtocolLayer::markRemoveEntity( "remove entity", baseMarkId+4, typeid(ProtocolLayer) ); ///< Removal of ProtocolEntity from a ProtocolLayer
