//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ProtocolSimulation.h

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Declaration of odemx::ProtocolSimulation

	\sa ProtocolSimulation.cpp

	<!-- [detailed description] -->

	\since 2.0
*/

#ifndef ODEMX_PROTOCOL_SIMULATION_INCLUDED
#define ODEMX_PROTOCOL_SIMULATION_INCLUDED

#include <odemx/base/Simulation.h>
#include <odemx/protocol/ProtocolLayer.h>

namespace odemx {

	// forward declaration
	class NetTopology;
	class TransmissionMedium;
	class TransmissionMediumObserver;
	class ProtocolLayerManager;

	/** \class ProtocolSimulation

		\ingroup protocol

		\author Ronald Kluth

		\brief %ProtocolSimulation is the base class for ODEMx protocol simulations

		\sa ProtocolLayerManager, NetTopology, TransmissionMedium

		ProtocolSimulation is a specialization of Simulation. It is the context
		for protocol model components and offers functions for accessing
		the NetTopology, the TransmissionMedium and the ProtocolLayers.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.0
	*/
	class ProtocolSimulation: public Simulation
	{
	public:

		/**
			\brief Construction
			\param l
				label of this object
			\param o
				initial observer
		*/
		ProtocolSimulation( Label l = "ProtocolSimulation",
							SimulationObserver* o = 0 );

		/**
			\brief Construction
			\param l
				label of this object
			\param o
				initial observer
		*/
		ProtocolSimulation( Label l, SimTime startAt, SimulationObserver* o = 0 );

		/// Destruction
		virtual ~ProtocolSimulation();

		/// get access to the network's topology
		NetTopology* getNetTopology();

		/// get a pointer to the TransmissionMedium
		TransmissionMedium* getTransmissionMedium();

		/// get a pointer to a layer by providing its label
		ProtocolLayer* getLayerByLabel( Label l );

		/**
			\brief get a list of all layers

			This function allows external access to the node's input port.
			The given element \p msg will be put into the node's input port.
		*/
		ProtocolLayerList& getLayerList();

		/**
			\brief template function for medium creation
			\tparam MediumType Type name of a subclass of TransmissionMedium
			\param l
				Label of this object
			\param o
				initial observer
			\return
				pointer to the newly created TransmissionMedium object

			This function must be called with a TransmissionMedium type as
			template parameter. A new TransmissionMedium object will be
			created and automatically registered with the simulation.
		*/
		template < typename MediumType >
		MediumType* createMedium( Label l, TransmissionMediumObserver* o = 0 )
		{
			if( medium )
			{
				error( this, "ProtocolSimulation::createMedium(): TransmissionMedium already set" );
			}
			else
			{
				medium = new MediumType( this, l, o );
			}
			return static_cast< MediumType* >( medium );
		}

		/**
			\brief template function for layer creation
			\tparam EntityType Type name of a subclass of ProtocolEntity
			\param l
				Label of this object
			\param o
				initial observer

			This function must be called with a ProtocolEntity type as
			template parameter. A new ProtocolLayerT object will be
			created and automatically registered with the layer manager.
		*/
		template < typename EntityType >
		void createLayer( Label l, ProtocolLayerObserver* o = 0 )
		{
			new ProtocolLayerT< EntityType >( this, l, o );
		}

		/**
			\brief print all created layers
			\param withEntities
				determine whether to show all entities

			This function prints a protocol simulation's layers. If
			\c withEntities is set to \c true, all entities registered
			with a layer will be displayed, which are as many as there
			are nodes in the network.
		*/
		void printLayers( bool withEntities = false );

	protected:

		/**
			\brief Initialization of a protocol simulation

			Implement this method to initialize a protocol simulation.
			This is the best place for initializing the medium and the layers
			of a ProtocolSimulation. One can also create and activate the
			first processes of a simulation in this function.

			\sa createMedium(), createLayer()
		*/
		virtual void initSimulation() {};

	private:

		friend class ProtocolStack;
		friend class ProtocolLayer;

		/// pointer to the TransmissionMedium
		TransmissionMedium* medium;
		/// pointer to the ProtocolLayerManager
		ProtocolLayerManager layerManager;
	};
}

#endif /* ODEMX_PROTOCOL_SIMULATION_INCLUDED */
