//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ProtocolStack.cpp

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Implementation of classes in ProtocolStack.h

	\sa ProtocolStack.h

	<!-- [detailed description] -->

	\since 2.0
*/

#include <odemx/protocol/ProtocolStack.h>
#include <odemx/protocol/ProtocolSimulation.h>
#include <odemx/protocol/ProtocolMessage.h>

using namespace odemx;

ProtocolStack::ProtocolStack( ProtocolSimulation* sim, Label l, ProtocolStackObserver* o ):
	DefLabeledObject( sim, l ),
	Observable<ProtocolStackObserver>( o ),
	env( sim )
	{
		// trace
		getTrace() -> mark( this, markCreate );

		// observer
		_obsForEach(ProtocolStackObserver, Create(this));

		// initialize the node's entities for all layers
		ProtocolLayerList& layers = sim -> getLayerList();

		if ( layers.size() == 0 )
		{
			error( this, "ProtocolStack(): No protocol layers defined, cannot create entities" );
		}
		else
		{
			initEntities( layers );
		}

		nodeId = getNextId();
	}


ProtocolStack::~ProtocolStack()
	{
		// trace
		getTrace() -> mark( this, markDestroy );

		// observer
		_obsForEach(ProtocolStackObserver, Destroy(this));
	}


// initialize this node's layer entities
void ProtocolStack::initEntities( ProtocolLayerList& layers ) {

	ProtocolEntity* previous = 0;
	ProtocolEntity* created = 0;
	ProtocolLayer* currentLayer = 0;

	ProtocolLayerList::iterator iter;
	for( iter = layers.begin(); iter != layers.end(); ++iter ) {

		currentLayer = (*iter);

		// fill map with entities for this node
		created = currentLayer -> createNewEntity( this );

		// remember entity corresponding to its layer
		entities[ currentLayer ] = created;

		// link this node's entities
		if ( previous ) {
			previous -> setBottomEntity( created );
			created -> setTopEntity( previous );
		}

		// remember the new Entity for the next loop
		previous = created;
		// schedule the new process
		created -> hold();
	}

	topEntity = entities[ layers.front() ];
	bottomEntity = entities[ layers.back() ];
}

void ProtocolStack::inputData( PortData* e ) {

	//trace
	getTrace() -> mark( this, markInputData );
	// observer
	_obsForEach( ProtocolStackObserver, InputData( this, e ) );

	getTopLayerEntity() -> putTop( e );
}

void ProtocolStack::receiveTransmission( ProtocolMessage* msg ) {

	//trace
	getTrace() -> mark( this, markReceiveTransmission, msg -> toString() );

	// observer
	_obsForEach( ProtocolStackObserver, ReceiveTransmission( this, msg ) );

	// retrieve lowest level entity
	getBottomLayerEntity() -> putBottom( (PortData*) msg );
}


void ProtocolStack::_receivedData( PortData* msg ) {

	//trace
	getTrace() -> mark( this, markReceivedData );

	// observer
	_obsForEach( ProtocolStackObserver, ReceivedData( this, msg ) );

	receivedData( msg );
}


unsigned int ProtocolStack::getNextId() {
	static unsigned int currentId = 0;
	return currentId++;
}


NodeID ProtocolStack::getNodeId() {
	return nodeId;
}


ProtocolEntity* ProtocolStack::getTopLayerEntity() {
	return topEntity;
}


ProtocolEntity* ProtocolStack::getBottomLayerEntity() {
	return bottomEntity;
}


ProtocolLayerManager& ProtocolStack::getLayerManager() {
	return env -> layerManager;
}


ProtocolEntity* ProtocolStack::getEntityByLayer( ProtocolLayer* layer ) {

	std::map< ProtocolLayer*, ProtocolEntity* >::iterator
	found = entities.find( layer );

	if ( found != entities.end() )
		return found -> second;

	return 0;
}


ProtocolEntity* ProtocolStack::getEntityByLayerName( Label l ) {

	ProtocolLayer* layer = getLayerManager().getLayerByLabel( l );

	std::map< ProtocolLayer*, ProtocolEntity* >::iterator
	found = entities.find( layer );

	if ( found != entities.end() )
		return found -> second;

	return 0;
}


const MarkTypeId ProtocolStack::baseMarkId = MarkType::getNewBaseMarkId( typeid(ProtocolStack) );

const MarkType ProtocolStack::markCreate( "create", baseMarkId+1, typeid(ProtocolStack) ); ///< Call of ProtocolStack constructor
const MarkType ProtocolStack::markDestroy( "destroy", baseMarkId+2, typeid(ProtocolStack) ); ///< Call of ProtocolStack destructor
const MarkType ProtocolStack::markInputData( "input data", baseMarkId+3, typeid(ProtocolStack) ); ///< Use of the input interface of a node, which is then handled by the top layer entity
const MarkType ProtocolStack::markReceivedData( "received data", baseMarkId+4, typeid(ProtocolStack) ); ///< Use of the output interface of a node, called when the top layer passes data upward
const MarkType ProtocolStack::markReceiveTransmission( "receive transmission", baseMarkId+5, typeid(ProtocolStack) ); ///< Node received a transmission from the physical medium, the data is passed on to the lowest layer entity

