//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ProtocolStack.h

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Declaration of odemx::ProtocolStack

	\sa ProtocolStack.cpp

	<!-- [detailed description] -->

	\since 2.0
*/

#ifndef ODEMX_NET_NODE_INCLUDED
#define ODEMX_NET_NODE_INCLUDED

#include <odemx/util/MarkType.h>
#include <odemx/util/LabeledObject.h>
#include <odemx/synchronization/Port.h>
#include <odemx/protocol/ProtocolEntity.h>
#include <odemx/protocol/ProtocolLayer.h>

#include <map>

namespace odemx {

	// forward declaration
	class TransmissionMedium;
	class ProtocolSimulation;
	class ProtocolMessage;
	class ProtocolStackObserver;

	typedef unsigned int NodeID;

	/** \class ProtocolStack

		\ingroup protocol

		\author Ronald Kluth
		<!-- [\author <author>]* -->

		\brief %ProtocolStack is the base class for nodes in protocol simulations

		\note %ProtocolStack supports Observation.
		\note %ProtocolStack supports Trace.

		\sa NetTopology, ProtocolLayer, ProtocolEntity, ProtocolSimulation

		ProtocolStacks are the basic structural components of network models.
		Following the OSI layer model for network protocols, a node manages
		exactly one ProtocolEntity per protocol layer. All ProtocolEntities
		belonging to a ProtocolStack are automatically created and linked according
		to their layers upon construction of that node.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.0
	*/
	class ProtocolStack: public virtual TraceProducer,
						public DefLabeledObject,
						public Observable<ProtocolStackObserver> {

	public:

		/**
			\brief Construction
			\param sim
				pointer to the Simulation object
			\param l
				label of this object
		*/
		ProtocolStack( ProtocolSimulation* sim, Label l, ProtocolStackObserver* o = 0 );

		/// Destruction
		virtual ~ProtocolStack();

		/// get a pointer to the Trace object
		virtual Trace* getTrace() const { return env; }

		/// get a pointer to the simulation context of this node
		ProtocolSimulation* getSimulation() const { return env; }

		/// get a pointer to the simulation's transmission medium
		TransmissionMedium* getTransmissionMedium() const
		{
			return env -> getTransmissionMedium();
		}

		/**
			\brief get the node's ID, which was set during construction
			\return
				id of this node

			This function returns the node's id. Node-IDs are used to make
			the NetTopology more efficient by storing information about nodes
			in a map indexed by NodeIDs.
		*/
		NodeID getNodeId();

		/**
			\name Node access

			These functions provide access to a node's data input and output.
			@{
		*/

		/**
			\brief node message input from the physical layer
			\param msg
				message the node receives from the network

			This function provides access to a node's physical layer entity.
			The given message \p msg will be put into the bottom layer
			entity's input port.
		*/
		void receiveTransmission( ProtocolMessage* msg );


		/**
			\brief transmit a message
			\param e
				message in form of a port element

			This function can be used to send a message through the network
			via the ProtocolStack's top layer entity. The element \p e will be
			passed on to that entity, which should evaluate it and act
			according to the element's type.
		*/
		void inputData( PortData* e );

	protected:

		/**
			\brief automatically called when node has received data
			\param msg
				message received by this node's topEntity

			This function gets called whenever the topmost entity passes
			data to the upper level. Since there isn't one, the user
			needs to specify what to do with the received data.
		*/
		virtual void receivedData( PortData* msg ) = 0;
		//@}

	public:

		/**
			\name Protocol layers and corresponding entities

			These functions provide access to a node's entities
			which provide different functionality, according
			to the layer of the protocol model.

			@{
		*/
		/**
			\brief get the entity representing the top layer
			\return
				top layer entity
		*/
		ProtocolEntity* getTopLayerEntity();

		/**
			\brief get the entity representing the bottom layer
			\return
				bottom layer entity

			This function returns the lowest layer entity, which
			is connected to the TransmissionMedium.
		*/
		ProtocolEntity* getBottomLayerEntity();

		/**
			\brief get the entity representing \p layer
			\param layer
				pointer to the layer that contains the requested entity
			\return
				a protocol entity in the specified layer
		*/
		ProtocolEntity* getEntityByLayer( ProtocolLayer* layer );

		/**
			\brief get the entity representing layer \p l
			\param l
				label of the layer that contains the requested entity
			\return
				a protocol entity in the specified layer
		*/
		ProtocolEntity* getEntityByLayerName( Label l );
		//@}

	private:

		friend class ProtocolEntity;

		/**
			\brief initialize all of the node's entities
			\param layers
				reference to the simulation's layer list

			This function initializes and links all entities belonging to
			a node. For each layer in the list, one entity will be created
			and linked to its corresponding lower and upper layer entity.
			The function is called automatically during ProtocolStack construction.
		*/
		virtual void initEntities( ProtocolLayerList& layers );

		/// internal function which provides the next node id
		static NodeID getNextId();

		/// wraps the trace and observer calls upon data reception
		void _receivedData( PortData* msg );

		/// access the simulation's layer manager
		ProtocolLayerManager& getLayerManager();

	private:

		/// associative array linking protocol layers and this node's entities
		std::map< ProtocolLayer*, ProtocolEntity* > entities;
		/// topmost entity in the protocol stack
		ProtocolEntity* topEntity;
		/// bottom enetity in the protocol stack
		ProtocolEntity* bottomEntity;
		/// pointer to the simulation context
		ProtocolSimulation* env;
		/// integer id for storing the node in the NetTopology
		unsigned int nodeId;

	public:

		/**
			\name Trace MarkTypes

			These MarkTypes and Tags are used to trace ProtocolStack calls.
			A TraceConsumer can use these constants to identify additional
			trace events (besides those from base class Process) sent by
			ProtocolStack.

			@{
		*/
		static const MarkTypeId baseMarkId;

		static const MarkType markCreate;
		static const MarkType markDestroy;
		static const MarkType markInputData;
		static const MarkType markReceivedData;
		static const MarkType markReceiveTransmission;
		//@}
	};

	/** \interface ProtocolStackObserver

		\author Ronald Kluth

		\brief Observer for ProtocolStack specific simulation events.

		\sa ProtocolStack

		\since 2.0
	*/
	class ProtocolStackObserver
	{

	public:
		/// Destructor
		virtual ~ProtocolStackObserver() {}

		/// Construction of observed node
		virtual void onCreate( ProtocolStack* sender ) {}
		/// Destruction of observed node
		virtual void onDestroy( ProtocolStack* sender ) {}
		/// Output to environment
		virtual void onInputData( ProtocolStack* sender, PortData* message ) {}
		/// transmit a message
		virtual void onReceivedData( ProtocolStack* sender, PortData* message ) {}
		/// Reception from network
		virtual void onReceiveTransmission( ProtocolStack* sender, ProtocolMessage* message ) {}
	};

}

#endif /* ODEMX_NET_NODE_INCLUDED */

