//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file TransmissionEvents.cpp

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Implementation of classes in TransmissionEvents.h

	\sa TransmissionEvents.h

	<!-- [detailed description] -->

	\since 2.0
*/

#include <odemx/protocol/TransmissionMedium.h>
#include <odemx/protocol/TransmissionEvents.h>
#include <odemx/protocol/ProtocolMessage.h>

using namespace odemx;

UnicastTransmission* UnicastTransmission::create( ProtocolSimulation* sim,
		Label l, ProtocolMessage* m, UnicastTransmissionObserver* o )
{
	return new UnicastTransmission( sim, l, m, o );
}

UnicastTransmission::UnicastTransmission( ProtocolSimulation* sim, Label l,
										  ProtocolMessage* m, UnicastTransmissionObserver* o ):
	Event( sim, l ),
	Observable<UnicastTransmissionObserver>( o ),
	message( m )
	{}

void UnicastTransmission::eventAction() {

	TransmissionMedium* medium =
			((ProtocolSimulation*)getSimulation()) -> getTransmissionMedium();

	ProtocolStack* sender = message -> getSendNode();
	ProtocolStack* receiver = message -> getReceiveNode();

	// get random sample for connectivity model
	if ( medium -> badConnectivity( sender, receiver ) ) {

		reportFailure( message, sender, receiver );
	}
	else {

		reportSuccess( message, sender, receiver );

		receiver -> receiveTransmission( message );
	}

	medium -> finishedTransmissions.push_back( this );
}

void UnicastTransmission::reportFailure( ProtocolMessage* m,
										 ProtocolStack* sender,
										 ProtocolStack* receiver ) {

	TransmissionMedium* medium =
		((ProtocolSimulation*)getSimulation()) -> getTransmissionMedium();

	medium -> updateUnicastFailure();

	//trace
	getTrace() -> beginMark( this, markUnicastFailure );
	getTrace() -> addTag( tagFrom, sender );
	getTrace() -> addTag( tagTo, receiver );
	getTrace() -> endMark( m -> toString() ); // msg handed over as comment

	// observer
	_obsForEach( UnicastTransmissionObserver, Failure( this, m, sender, receiver ) );
}

void UnicastTransmission::reportSuccess( ProtocolMessage* m,
										 ProtocolStack* sender,
										 ProtocolStack* receiver ) {

	TransmissionMedium* medium =
		((ProtocolSimulation*)getSimulation()) -> getTransmissionMedium();

	medium -> updateUnicastSuccess();

	//trace
	getTrace() -> beginMark( this, markUnicastSuccess );
	getTrace() -> addTag( tagFrom, sender );
	getTrace() -> addTag( tagTo, receiver );
	getTrace() -> endMark( m -> toString() ); // msg handed over as comment

	// observer
	_obsForEach( UnicastTransmissionObserver, Success( this, m, sender, receiver ) );
}

BroadcastTransmission* BroadcastTransmission::create( ProtocolSimulation* sim,
		Label l, ProtocolMessage* m, BroadcastTransmissionObserver* o )
{
	return new BroadcastTransmission( sim, l, m, o );
}

BroadcastTransmission::BroadcastTransmission( ProtocolSimulation* sim, Label l,
											  ProtocolMessage* m, BroadcastTransmissionObserver* o ):
	Event( sim, l ),
	Observable<BroadcastTransmissionObserver>( o ),
	message( m )
	{}

void BroadcastTransmission::eventAction()
{
	// monitor previous transmissions of the same BROADCAST,
	// must copy route if true because pointer access would cause overwrite
	bool previousDeliveries = false;

	ProtocolStack* sender = message -> getSendNode();
	ProtocolStack* receiver = 0;

	TransmissionMedium* medium =
		((ProtocolSimulation*)getSimulation()) -> getTransmissionMedium();

	// iterator for neighbor / connectivity maps
	map< ProtocolStack*, LinkInfo >::const_iterator j;    // *j is a pair

	const map< ProtocolStack*, LinkInfo >& senderLinks = medium -> getLinks( sender );

	// iterate over all neighbors
	for( j = senderLinks.begin(); j != senderLinks.end(); ++j )
	{
		receiver = (*j).first;

		// attempt to transmit the message
		if( medium -> badConnectivity( sender, receiver ) )
		{
			reportFailure( message, sender, receiver );
		}
		else
		{
			reportSuccess( message, sender, receiver );

			// make sure duplicate BROADCASTS do not access
			// the same member data via pointer
			if ( !previousDeliveries )
			{
				receiver -> receiveTransmission( message );
				previousDeliveries = true;
			}
			else // message was broadcast before by sender, copy message
			{
				ProtocolMessage* messageCopy = message -> deepCopy();
				receiver -> receiveTransmission( messageCopy );
			}
		}
	}

	medium -> finishedTransmissions.push_back( this );
}

void BroadcastTransmission::reportFailure( ProtocolMessage* m,
										   ProtocolStack* sender,
										   ProtocolStack* receiver ) {

	TransmissionMedium* medium =
		((ProtocolSimulation*)getSimulation()) -> getTransmissionMedium();

	medium -> updateBroadcastFailure();

	//trace
	getTrace() -> beginMark( this, markBroadcastFailure );
	getTrace() -> addTag( tagFrom, sender );
	getTrace() -> addTag( tagTo, receiver );
	getTrace() -> endMark( m -> toString() ); // msg handed over as comment

	// observer
	_obsForEach( BroadcastTransmissionObserver, Failure( this, m, sender, receiver ) );
}

void BroadcastTransmission::reportSuccess( ProtocolMessage* m,
										   ProtocolStack* sender,
										   ProtocolStack* receiver ) {

	TransmissionMedium* medium =
		((ProtocolSimulation*)getSimulation()) -> getTransmissionMedium();

	medium -> updateBroadcastSuccess();

	//trace
	getTrace() -> beginMark( this, markBroadcastSuccess );
	getTrace() -> addTag( tagFrom, sender );
	getTrace() -> addTag( tagTo, receiver );
	getTrace() -> endMark( m -> toString() ); // msg handed over as comment

	// observer
	_obsForEach( BroadcastTransmissionObserver, Success( this, m, sender, receiver ) );
}

const MarkTypeId UnicastTransmission::baseMarkId = MarkType::getNewBaseMarkId( typeid(UnicastTransmission) );

const MarkType UnicastTransmission::markUnicastSuccess( "unicast success", baseMarkId+1, typeid(UnicastTransmission) ); ///< Unicast message was delivered successfully
const MarkType UnicastTransmission::markUnicastFailure( "unicast failure", baseMarkId+2, typeid(UnicastTransmission) ); ///< Unicast message delivery failed due to bad connectivity

const TagId UnicastTransmission::baseTagId = UnicastTransmission::baseMarkId;

const Tag UnicastTransmission::tagFrom( "from", baseTagId+1 ); ///< Displays information about the sender of a message
const Tag UnicastTransmission::tagTo( "to", baseTagId+2 ); ///< Displays information about the intended receiver of a message

const MarkTypeId BroadcastTransmission::baseMarkId = MarkType::getNewBaseMarkId( typeid(BroadcastTransmission) );

const MarkType BroadcastTransmission::markBroadcastSuccess( "broadcast success", baseMarkId+1, typeid(BroadcastTransmission) ); ///< Broadcast message was delivered successfully from a node to one of its neighbors
const MarkType BroadcastTransmission::markBroadcastFailure( "broadcast failure", baseMarkId+2, typeid(BroadcastTransmission) ); ///< Broadcast message delivery to one neighbor failed

const TagId BroadcastTransmission::baseTagId = BroadcastTransmission::baseMarkId;

const Tag BroadcastTransmission::tagFrom( "from", baseTagId+1 ); ///< Displays information about the sender of a message
const Tag BroadcastTransmission::tagTo( "to", baseTagId+2 ); ///< Displays information about the intended receiver of a message
