//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file TransmissionMedium.cpp

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Implementation of classes in TransmissionMedium.h

	\sa TransmissionMedium.h

	<!-- [detailed description] -->

	\since 2.0
*/

#include <odemx/protocol/ProtocolMessage.h>
#include <odemx/protocol/TransmissionMedium.h>
#include <odemx/protocol/TransmissionEvents.h>

using namespace odemx;

TransmissionMedium::TransmissionMedium( ProtocolSimulation* sim, Label l,
										TransmissionMediumObserver* o ):
	NetTopology( sim, o ),
	DefLabeledObject( sim, l ),
	StatisticObject( sim ),
	Observable<TransmissionMediumObserver>( o )
	{
		//trace
		getTrace() -> mark( this, markCreate );

		// observer
		_obsForEach( TransmissionMediumObserver, Create( this ) );

		//statistics
		reset();
	}

TransmissionMedium::~TransmissionMedium()
	{
		// memory management
		deleteFinishedTransmissions();

		//trace
		getTrace() -> mark( this, markDestroy );

		// observer
		_obsForEach( TransmissionMediumObserver, Destroy( this ) );
	}

bool TransmissionMedium::transmit( ProtocolMessage* m ) {

	// memory management
	deleteFinishedTransmissions();

	//trace
	getTrace() -> mark( this, markTransmit, m -> toString() );

	// observer
	_obsForEach( TransmissionMediumObserver, Transmit( this, m ) );

	++messageCount;

	m -> computeTransmissionTime( getCurrentTime() );

	// only unicasts can have a connection test here,
	// broadcasts are always transmitted,
	// failure is reported by BroadcastTransmission
	if ( m -> isUnicast() ) {

		ProtocolStack* sender = m -> getSendNode();
		ProtocolStack* receiver = m -> getReceiveNode();

		// error handling
		if ( noLinkSet( sender, receiver ) ) {

			// dist variable between sender and receiver not initialized, no connection
			error( this, string( "TransmissionMedium::transmit(): " )
					+"no connection set for node pair in unicast transmission: "
					+sender -> getLabel() + " -> " + receiver -> getLabel() );
			return false;
		}

		++unicastCount;

		// send message to receiveNode
		scheduleUnicastTransmission( m );

	}
	else { // m is broadcast

		++broadcastCount;

		// send message to all neighbors
		scheduleBroadcastTransmission( m );
	}

	return true;
}

bool TransmissionMedium::noLinkSet(  ProtocolStack* sendNode, ProtocolStack* receiveNode )
{
	return !( NetTopology::linked( sendNode, receiveNode ) );
}

bool TransmissionMedium::badConnectivity(  ProtocolStack* sendNode, ProtocolStack* receiveNode )
{
	return ( NetTopology::getConnectivitySample( sendNode, receiveNode ) == 0 ) ? true : false;
}

void TransmissionMedium::scheduleUnicastTransmission( ProtocolMessage* m )
{
	SimTime t = m -> getTransmissionTime();

	//trace
	getTrace() -> beginMark( this, markScheduleUnicast );
	getTrace() -> addTag( tagFrom, m -> getSendNode() );
	getTrace() -> addTag( tagTo, m -> getReceiveNode() );
	getTrace() -> addTag( tagAt, t );
	getTrace() -> endMark( m -> toString() ); // msg handed over as comment

	// observer
	_obsForEach( TransmissionMediumObserver, ScheduleUnicast( this, m, t ) );

	Event* unicast = UnicastTransmission::create( getSimulation(), "Unicast",  m );

	unicast -> scheduleAppendAt( t );
}

void TransmissionMedium::scheduleBroadcastTransmission( ProtocolMessage* m ) {

	SimTime t = m -> getTransmissionTime();

	//trace
	getTrace() -> beginMark( this, markScheduleBroadcast );
	getTrace() -> addTag( tagFrom, m -> getSendNode() );
	getTrace() -> addTag( tagAt, t );
	getTrace() -> endMark( m -> toString() ); // msg handed over as comment

	// observer
	_obsForEach( TransmissionMediumObserver, ScheduleBroadcast( this, m, t ) );

	Event* broadcast = BroadcastTransmission::create( getSimulation(), "Broadcast",  m );

	broadcast -> scheduleAppendAt( t );
}

Trace* TransmissionMedium::getTrace() const {
	return static_cast< Trace* >( getSimulation() );
}

SimTime TransmissionMedium::getCurrentTime() const {
	return getSimulation() -> getTime();
}

void TransmissionMedium::updateCollisionCount( unsigned int c )
{
	collisionCount += c;
}

void TransmissionMedium::updateUnicastSuccess( unsigned int c )
{
	successfulUnicastCount += c;
}

void TransmissionMedium::updateBroadcastSuccess( unsigned int c )
{
	successfulBroadcastCount += c;
}

void TransmissionMedium::updateUnicastFailure( unsigned int c )
{
	failedUnicastCount += c;
}

void TransmissionMedium::updateBroadcastFailure( unsigned int c )
{
	failedBroadcastCount += c;
}

void TransmissionMedium::report( Report* r )
{
	assert( r );

	static vector< string > labels; 		static vector< Column::Type  > types;
	labels.push_back("Name");				types.push_back(Column::STRING);
	labels.push_back("Reset at");			types.push_back(Column::SIMTIME);
	labels.push_back("Nodes");				types.push_back(Column::INTEGER);
	labels.push_back("Messages");			types.push_back(Column::INTEGER);
	labels.push_back("Broadcasts");			types.push_back(Column::INTEGER);
	labels.push_back("Unicasts");			types.push_back(Column::INTEGER);
	labels.push_back("Successful Broadcasts");	types.push_back(Column::INTEGER);
	labels.push_back("Successful Unicasts");	types.push_back(Column::INTEGER);
	labels.push_back("Failed Broadcasts");	types.push_back(Column::INTEGER);
	labels.push_back("Failed Unicasts");	types.push_back(Column::INTEGER);
	labels.push_back("Collisions");			types.push_back(Column::INTEGER);

	static utTableDef def( labels.size(), labels, types );

	Table* t = r -> createTable( "Transmission Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< NetTopology::getAllNodes().size()
	<< messageCount
	<< broadcastCount
	<< unicastCount
	<< successfulBroadcastCount
	<< successfulUnicastCount
	<< failedBroadcastCount
	<< failedUnicastCount
	<< collisionCount;
}

void TransmissionMedium::reset()
{
	StatisticObject::reset();

	messageCount = 0;
	broadcastCount = 0;
	unicastCount = 0;
	successfulBroadcastCount = 0;
	successfulUnicastCount = 0;
	failedBroadcastCount = 0;
	failedUnicastCount = 0;
	collisionCount = 0;
}

void TransmissionMedium::deleteFinishedTransmissions()
{
	for( std::vector< Event* >::iterator i = finishedTransmissions.begin();
		 i != finishedTransmissions.end(); ++ i )
	{
		if( *i ) delete *i;
	}
	finishedTransmissions.clear();
}

const MarkTypeId TransmissionMedium::baseMarkId = MarkType::getNewBaseMarkId( typeid(TransmissionMedium) );

const MarkType TransmissionMedium::markCreate( "create", baseMarkId+1, typeid(TransmissionMedium) ); ///< Call of TransmissionMedium constructor
const MarkType TransmissionMedium::markDestroy( "destroy", baseMarkId+2, typeid(TransmissionMedium) ); ///< Call of TransmissionMedium destructor
const MarkType TransmissionMedium::markTransmit( "transmit", baseMarkId+3, typeid(TransmissionMedium) ); ///< Attempt to transmit the given ProtocolMessage, coming from lowest layer entity
const MarkType TransmissionMedium::markScheduleUnicast( "schedule unicast", baseMarkId+4, typeid(TransmissionMedium) ); ///< Message is not a broadcast to all neighbors, scheduled a UnicastTransmission event
const MarkType TransmissionMedium::markScheduleBroadcast( "schedule broadcast", baseMarkId+5, typeid(TransmissionMedium) ); ///< Message is a broadcast to all neighbors, scheduled a BroadcastTransmission event

const TagId TransmissionMedium::baseTagId = TransmissionMedium::baseMarkId;

const Tag TransmissionMedium::tagFrom( "from", baseTagId+1 ); ///< Displays information about the sender of a message
const Tag TransmissionMedium::tagTo( "to", baseTagId+2 ); ///< Displays information about the receiver of a message
const Tag TransmissionMedium::tagAt( "at time", baseTagId+3 ); ///< Displays information about the transmission time of a message
