//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2003, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ContinuousDist.cpp

	\author Ralf Gerstenberger

	\date created at 2003/05/09

	\brief Implementation of classes in ContinuousDist.h

	\sa ContinuousDist.h

	<!-- [detailed description] -->

	\since 1.0
*/

#include <cmath>
#include <iostream>

#include <odemx/random/ContinuousDist.h>
#include <odemx/util/ErrorHandling.h>

using namespace odemx;

Normal::Normal( Label title, double na, double nb ):
	ContinuousDist( title )
	{
		if (nb < 0.0) {
			nb = -nb;
			warning( this, "Normal(): given deviation is less than 0; changed to -nb");
		}
		a=na; b=nb; zyqeven=0;
		zyqu=zyqv=0.0;
	}

Normal::Normal(DistContext* c, Label title, double na, double nb):
	ContinuousDist(c, title)
	{
		if (nb < 0.0) {
			nb = -nb;
			warning( this, "Normal(): given deviation is less than 0; changed to -nb");
		}
		a=na; b=nb; zyqeven=0;
		zyqu=zyqv=0.0;
	}

double Normal::sample() {
	double z;
	uses++;
	if (zyqeven) {
		zyqeven=0;
		z= zyqu*cos(zyqv);
	}
	else {
		zyqeven= 1;
		zyqu= sqrt(-2.0*log(getSample()));
		zyqv= 6.28318530717958647693*getSample();
		z= zyqu*sin(zyqv);
	};
	return (z*b + a);
};

void Normal::report(Report* r) {
	Table* t=getTable(r);

	(*t) << "Normal" << long(getUses()) << long(getSeed()) << double(a) << double(b) << TableDefinition::ENDL;
}

Negexp::Negexp( Label title, double na ):
	ContinuousDist( title )
	{
		if (na <= 0.0) {
			na = (na<0.0) ? -na : 0.001;
			warning( this, "Negexp(): expected value 'na' less or equal 0; -na or 0.001 used");
		}
		a=na;
	}

Negexp::Negexp (DistContext* c, Label title, double na):
	ContinuousDist(c, title)
	{
		if (na <= 0.0) {
			na = (na<0.0) ? -na : 0.001;
			warning( this, "Negexp(): expected value 'na' less or equal 0; -na or 0.001 used");
		}
		a=na;
	}

double Negexp::sample() {
	uses++;
	return( -log(getSample())/a);
}

void Negexp::report(Report* r) {
	Table* t=getTable(r);

	(*t) << "Negexp" << long(getUses()) << long(getSeed()) << double(a) << TableDefinition::ENDL;
}

Uniform::Uniform( Label title, double ua, double ub ):
	ContinuousDist( title )
	{
		if (ua>ub) {
			double help= ua; ua= ub; ub= help;
			warning( this, "Uniform(): lower bound greater than upper bound; bounds swapped" );
		}
		a= ua; b= ub;
		zyqspan= b-a;
	}

Uniform::Uniform (DistContext* c, Label title, double ua, double ub):
	ContinuousDist(c, title)
	{
		if (ua>ub) {
			double help= ua; ua= ub; ub= help;
			warning( this, "Uniform(): lower bound greater than upper bound; bounds swapped" );
		}
		a= ua; b= ub;
		zyqspan= b-a;
	}

double Uniform::sample() {
	uses++;
	return( zyqspan*getSample() + a);
}

void Uniform::report(Report* r) {
	Table* t=getTable(r);

	(*t) << "Uniform" << long(getUses()) << long(getSeed()) << double(a) << double(b) << TableDefinition::ENDL;
}

Erlang::Erlang( Label title, double ea, int eb ):
	ContinuousDist( title )
	{
		if (ea<=0.0) {
			ea= (ea<0.0) ? -ea : 0.01;
			warning( this, "Erlang: ea<=0; -ea or 0.01 used");
		}
		if (eb<=0) {
			eb= (eb<0) ? -eb : 1;
			warning( this, "Erlang: eb<=0; -eb or 1 used");
		}
		a=ea; b= eb;
//		zyqab= a/b;
	}

Erlang::Erlang (DistContext* c, Label title, double ea, int eb):
	ContinuousDist(c, title)
	{
		if (ea<=0.0) {
			ea= (ea<0.0) ? -ea : 0.01;
			warning( this, "Erlang: ea<=0; -ea or 0.01 used");
		}
		if (eb<=0) {
			eb= (eb<0) ? -eb : 1;
			warning( this, "Erlang: eb<=0; -eb or 1 used");
		}
		a=ea; b= eb;
//		zyqab= a/b;
	}

double Erlang::sample()
{
	uses++;

	double erlangTmp = 0.0;
	for( int j = 0; j < getShape(); ++j )
	{
		erlangTmp += -( 1 / getRate() ) * log( getSample() );
	}
	return erlangTmp;
}

void Erlang::report(Report* r) {
	Table* t=getTable(r);

	(*t) << "Erlang" << long(getUses()) << long(getSeed()) << double(a) << double(b) << TableDefinition::ENDL;
}

ContinuousConst::ContinuousConst( Label title, double d ):
	ContinuousDist( title ),
	a(d)
	{}

ContinuousConst::ContinuousConst(DistContext* c, Label title, double d):
	ContinuousDist(c, title),
	a(d)
	{}

void ContinuousConst::report(Report* r) {
	Table* t=getTable(r);

	(*t) << "ContinuousConst" << long(getUses()) << long(getSeed()) << double(a) << TableDefinition::ENDL;
}

