//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2003, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Dist.cpp

	\author Ralf Gerstenberger

	\date created at 2003/05/09

	\brief Implementation of random number generator base class

	\sa Dist.h

	Implementation of a random number generator taken from ODEM.

	\since 1.0
*/

#include <odemx/random/Dist.h>
#include <odemx/util/ErrorHandling.h>

#include <cassert>
#include <vector>
#include <string>

using namespace odemx;
using namespace std;

const unsigned long odemx::zyqmodulo = 67099547;

DistContext::DistContext() :
	zyqseed(907)
{
}

DistContext::~DistContext()
{
}

void DistContext::setSeed(int n/*=0*/) {
	if (n<0) n= -n;
	if (n>=(int)zyqmodulo) n=n - ((int)(n/zyqmodulo))*zyqmodulo;
	if (n==0) n=(int)(zyqmodulo/2);
	zyqseed=n;
}

unsigned long DistContext::getSeed() {
	return zyqseed;
}

unsigned long DistContext::getNextSeed() {
	zyq_next_seed();
	return zyqseed;
}

void DistContext::zyq_next_seed() {
	int k;

	k=7; zyqseed= zyqseed*k;
	if (zyqseed >= zyqmodulo)
	    zyqseed = zyqseed - ((int)(zyqseed/zyqmodulo))*zyqmodulo;
	k=13; zyqseed= zyqseed*k;
	if (zyqseed >= zyqmodulo)
	    zyqseed = zyqseed - ((int)(zyqseed/zyqmodulo))*zyqmodulo;
	k=15; zyqseed= zyqseed*k;
	if (zyqseed >= zyqmodulo)
	    zyqseed = zyqseed - ((int)(zyqseed/zyqmodulo))*zyqmodulo;
	k=27; zyqseed= zyqseed*k;
	if (zyqseed >= zyqmodulo)
	    zyqseed = zyqseed - ((int)(zyqseed/zyqmodulo))*zyqmodulo;
}

Dist::Dist(DistContext* c/*=0*/, Label label/*=""*/)
:	DefLabeledObject(c, label),
	StatisticObject(c),
	context(c)
{
	if( context == 0 )
	{
		// error: Dist was created without reference to DistContext.
		fatalError( this, "Dist(); invalid argument: DistContext* c", -1 );
	}

	antithetics = 0;
	u = ustart = c -> getNextSeed();
}

Dist::~Dist()
{
}

void Dist::setSeed(int n/*=0*/) {
	if (n<0) n= -n;
	if (n>=(int)zyqmodulo) n=n - ((int)(n/zyqmodulo))*zyqmodulo;
	if (n==0) n=(int)(zyqmodulo/2);
	u=ustart=n;
}

unsigned long Dist::getSeed() {
	return ustart;
}

double Dist::getSample() {
	int k;
	double zyqsample, help;

	k=32; u=k*u;
	if( u>=zyqmodulo)  u =  u- ((int)(u/zyqmodulo))*zyqmodulo;

	// yes, this is needed twice
	k=32; u=k*u;
	if( u>=zyqmodulo)  u =  u- ((int)(u/zyqmodulo))*zyqmodulo;

	k=8; u=k*u;
	if( u>=zyqmodulo)  u =  u- ((int)(u/zyqmodulo))*zyqmodulo;

	help=(double)u/(double)zyqmodulo;
	zyqsample= antithetics ? 1.0 -  help : help;
	return zyqsample;
}

class RNGTable : public TableDefinition {
public:
	virtual size_t getNumberOfColumns() const {return 8;}

	virtual const string getLabelOfColumn(size_t i) const {

		static vector< string > labels;
		labels.reserve(8);
		labels.push_back("Name");
		labels.push_back("Reset at");
		labels.push_back("Type");
		labels.push_back("Uses");
		labels.push_back("Seed");
		labels.push_back("Parameter 1");
		labels.push_back("Parameter 2");
		labels.push_back("Parameter 3");

		return (i < 8)  ?  labels[i] : 0;
	}

	virtual Column::Type  getTypeOfColumn(size_t i) const {
		static vector< Column::Type > types;
		types.reserve(8);
		types.push_back( Column::STRING);
		types.push_back( Column::SIMTIME);
		types.push_back( Column::STRING);
		types.push_back( Column::INTEGER);
		types.push_back( Column::INTEGER);
		types.push_back( Column::REAL);
		types.push_back( Column::REAL);
		types.push_back( Column::REAL);

		return (i < 8)  ?  types[i] : Column::INVALID;
	}
};

TableDefinition* Dist::getTableDefinition() {
	static RNGTable td;
	return &td;
}

Table* Dist::getTable(Report* r) {
	assert(r!=0);

	Table* t = r -> createTable( "Random Number Generators", getTableDefinition() );
	assert(t!=0);

	*t << getLabel() << SimTime( getResetTime() );

	return t;
}

