//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2003, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Statistics.cpp

	\author Ralf Gerstenberger

	\date created at 2003/05/20

	\brief Implementation of classes in Statistics.h

	\sa Statistics.h

	<!-- [detailed description] -->

	\since 1.0
*/
#ifdef _MSC_VER
#pragma warning (disable : 4786)
#endif

#include <odemx/statistics/Statistics.h>
#include <odemx/util/ErrorHandling.h>

#include <cmath>
#include <cassert>

#include <sstream>
#include <string>
#include <algorithm>

using namespace std;
using namespace odemx;

Tab::Tab( Label label/*=""*/ ):
	DefLabeledObject( getDefaultSimulation(), label ),
	StatisticObject( getDefaultSimulation() )
	{
		reset();
	}

Tab::Tab (Simulation* s, Label label/*=""*/):
	DefLabeledObject( s, label ),
	StatisticObject( s )
	{
		if( s == 0 )
		{
			// error: Tab was created without reference to Simulation.
			error( this, "Tab::Tab(): invalid argument Simulation* s == 0" );
		}

		reset();
	}

Tab::~Tab() {}

Count::Count( Label title/*=""*/ ):
	Tab( title )
	{
		reset();
	}

Count::Count (Simulation* s, Label title/*=""*/):
	Tab(s, title)
	{
		reset();
	}

Count::~Count() {};

void Count::report(Report* r) {

	assert(r);

	static vector< string > labels;		static vector< Column::Type  > types;
	labels.push_back("Name");			types.push_back( Column::STRING);
	labels.push_back("Reset at");		types.push_back( Column::SIMTIME);
	labels.push_back("Uses");			types.push_back( Column::INTEGER);
	labels.push_back("Value");			types.push_back( Column::INTEGER);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("Count Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< getObs()
	<< getValue();
}

Sum::Sum( Label title/*=""*/ ):
	Tab( title )
	{
		reset();
	}

Sum::Sum (Simulation* s, Label title/*=""*/):
	Tab(s, title)
	{
		reset();
	}

Sum::~Sum() {}

void Sum::report(Report* r) {

	assert(r);

	static vector< string > labels;		static vector< Column::Type  > types;
	labels.push_back("Name");			types.push_back( Column::STRING);
	labels.push_back("Reset at");		types.push_back( Column::SIMTIME);
	labels.push_back("Uses");			types.push_back( Column::INTEGER);
	labels.push_back("Value");			types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("Sum Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< getObs()
	<< getValue();
}

Tally::Tally( Label title/*=""*/ ):
	Tab( title )
	{
		reset();
	}

Tally::Tally (Simulation* s, Label title/*=""*/):
	Tab(s, title)
	{
		reset();
	}

Tally::~Tally() {}

void Tally::reset() {
	Tab::reset();
	sum = sumsq = min = max = 0.0;
}

void Tally::update (double v) {
	Tab::update();

	sum += v;
	sumsq += v*v;

	if (getObs() == 1) min = max = v;
	else if (v < min) min = v;
	else if (v > max) max = v;
}

double Tally::getStandardDeviation()
{
	if (getObs()<=1)
		return 0.0;

	double variance = fabs( sumsq/getObs() - getMean()*getMean() );
	return sqrt( variance );
}

void Tally::report(Report* r) {

	assert(r);

	static vector< string > labels;		static vector< Column::Type  > types;
	labels.push_back("Name");			types.push_back( Column::STRING);
	labels.push_back("Reset at");		types.push_back( Column::SIMTIME);
	labels.push_back("Uses");			types.push_back( Column::INTEGER);
	labels.push_back("Size");			types.push_back( Column::INTEGER);
	labels.push_back("Min");			types.push_back( Column::REAL);
	labels.push_back("Max");			types.push_back( Column::REAL);
	labels.push_back("Mean");			types.push_back( Column::REAL);
	labels.push_back("Standard Deviation");		types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("Tally Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< getObs()
	<< getSize()
	<< getMin()
	<< getMax()
	<< getMean()
	<< getStandardDeviation();
}

Accum::Accum( Label title/*=""*/, Approx app/*=stepwise*/ ):
	Tab( title ),
	approx( app )
	{
		reset();
	}

Accum::Accum(Simulation* s, Label title/*=""*/, Approx app/*=stepwise*/):
	Tab(s, title),
	approx(app)
	{
		reset();
	}

Accum::~Accum() {}

void Accum::reset() {
	Tab::reset();
	sumt = sumsqt = min = max = lastv = 0.0;
	lasttime = manager->getTime();
}

void Accum::update (double v) {
	SimTime now, lastSpan;

	Tab::update();

	now     = manager->getTime();
	lastSpan    = now - lasttime;
	lasttime= now;

    if (getObs()>1) {

    	if (approx==linear)
		{
	        double mean = (v+lastv)/2;
	        sumt += mean * lastSpan;
	        sumsqt += mean * mean * lastSpan;
		}
		else // approx == stepwise
		{
	        sumt   += lastv * lastSpan;
	        sumsqt += lastv * lastv * lastSpan;
	    }
    }
	lastv   = v;

	if (getObs() == 1) min = max = v;
	else if (v < min) min = v;
	else if (v > max) max = v;
}

double Accum::getMean() const {
	double avg; 
	SimTime span, t;

	t = manager->getTime(); // current time
	span = t-getResetTime(); // total time span of measurements
	t = t-lasttime; // time span since last update

	if (span == 0)
		return 0.0;

	avg = (sumt + lastv*t) / span;

	return avg;
}

double Accum::getStandardDeviation() const {
	if (getObs()<=1)
		return 0.0;

	SimTime span, t;
	double avg, variance;

	t = manager->getTime(); // current time
	span = t-getResetTime(); // total time span of measurements
	t = t-lasttime; // time span since last update

	avg = getMean();

	if( span == 0.0 )
		return 0.0;

	variance = fabs( (sumsqt+lastv*lastv*t)/span - avg*avg );
	return sqrt( variance );
}

void Accum::report(Report* r) {

	assert(r);

	static vector< string > labels;		static vector< Column::Type  > types;
	labels.push_back("Name");			types.push_back( Column::STRING);
	labels.push_back("Reset at");		types.push_back( Column::SIMTIME);
	labels.push_back("Uses");			types.push_back( Column::INTEGER);
	labels.push_back("Size");			types.push_back( Column::INTEGER);
	labels.push_back("Min");			types.push_back( Column::REAL);
	labels.push_back("Max");			types.push_back( Column::REAL);
	labels.push_back("Mean");			types.push_back( Column::REAL);
	labels.push_back("Standard Deviation");		types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("Accum Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< getObs()
	<< getSize()
	<< getMin()
	<< getMax()
	<< getMean()
	<< getStandardDeviation();
}

Histo::Histo( Label title, double low, double upp, int n):
	Tally( title ),
	data(n+2)
	{
		lower=low;
		upper=upp;
		ncells=n;

		if (upper==lower) {
			upper=lower+100.0;
			warning( this, "Histo(): borders are equal; borders changed");
		}

		if (upper < lower) {
			std::swap(lower, upper);
			warning( this, "Histo(): upper borders < lower border; borders swapped");
		}

		if ((ncells < 1)) {
			warning( this, "Histo(): number of cells < 1; number increased");
			ncells=10;

			data.resize(ncells+2);
		}

		width = (upper-lower)/ncells;
		limit = ncells+1;
	}

Histo::Histo(Simulation* s, Label title, double low, double upp, int n):
	Tally(s, title),
	data(n+2)
	{
		lower=low;
		upper=upp;
		ncells=n;

		if (upper==lower) {
			upper=lower+100.0;
			warning( this, "Histo() borders are equal; borders changed");
		}

		if (upper < lower) {
			std::swap(lower, upper);
			warning( this, "Histo() upper borders < lower border; borders swapped");
		}

		if ((ncells < 1)) {
			warning( this, "Histo() number of cells < 1; number increased");
			ncells=10;

			data.resize(ncells+2);
		}

		width = (upper-lower)/ncells;
		limit = ncells+1;
	}

Histo::~Histo()
	{}

void Histo::reset() {
	Tally::reset();
	for (int k = 0; k <= limit; k++) data[k] = 0;
}

void Histo::update(double v) {
	int cell;

	Tally::update(v);

	v -= lower;
	if (v<0.0) cell = 0;
	else {
		cell = (int)(floor(v/width))+1;
		if (cell>limit) cell = limit;
	}
	data[cell]++;
}

long Histo::round( double d ) {

	long retVal = (long) floor( d );

	if( d - retVal > 0.5 )
		return retVal + 1;

	return retVal;
}

void Histo::report(Report* r) {

	assert(r);

	// Summary (Tally)
	static vector< string > labels;		static vector< Column::Type  > types;
	labels.push_back("Name");			types.push_back( Column::STRING);
	labels.push_back("Reset at");		types.push_back( Column::SIMTIME);
	labels.push_back("Uses");			types.push_back( Column::INTEGER);
	labels.push_back("Low");			types.push_back( Column::REAL);
	labels.push_back("High");			types.push_back( Column::REAL);
	labels.push_back("Cells");			types.push_back( Column::INTEGER);
	labels.push_back("Size");			types.push_back( Column::INTEGER);
	labels.push_back("Min");			types.push_back( Column::REAL);
	labels.push_back("Max");			types.push_back( Column::REAL);
	labels.push_back("Mean");			types.push_back( Column::REAL);
	labels.push_back("Standard Deviation");		types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("Histo Statistics - Summary", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< getObs()
	<< lower
	<< upper
	<< ncells
	<< getSize()
	<< getMin()
	<< getMax()
	<< getMean()
	<< getStandardDeviation();

	// Histogram
	static vector< string > labels2;	static vector< Column::Type  > types2;
	labels2.push_back("Low");			types2.push_back( Column::REAL );
	labels2.push_back("High");			types2.push_back( Column::REAL );
	labels2.push_back("Count");			types2.push_back( Column::INTEGER );
	labels2.push_back("Percentage");	types2.push_back( Column::REAL );
	labels2.push_back("Diagram");		types2.push_back( Column::BAR );

	static utTableDef def2( labels2.size(), labels2, types2 );

	std::ostringstream tableName;
	tableName
	<< "Histogram: " << getLabel()
	<< " [" << getSize() << " samples]";

	Table* t2=r->createTable(tableName.str(), &def2);
	assert(t2);

	// compute multiplier for graphical representation
	std::vector<int>::const_iterator it;
	double maxPercentage = 0;
	for( it = data.begin(); it != data.end(); ++it ) {
		double percentage = ((*it)*100.0)/getObs();

		if( percentage > maxPercentage )
			maxPercentage = percentage;
	}
	// diagram width should be 100%, i.e. the maximum value should fill it
	double multiplier = 100 / maxPercentage;

	// histogram table output with graphical representation
	it=data.begin();
	if( *it != 0 ) {
		double percentage = ((*it)*100.0)/getObs();

		*t2
		<< getMin()
		<< lower
		<< *it
		<< percentage
		<< round( multiplier * percentage );
	}
	++it;

	for( int i = 0; i < ncells; ++i, ++it ) {
		double percentage = ((*it)*100.0)/getObs();

		*t2
		<< (lower + i*width)
		<< (lower + (i+1)*width)
		<< *it
		<< percentage
		<< round( multiplier * percentage );
	}

	if (*it!=0) {
		double percentage = ((*it)*100.0)/getObs();

		*t2
		<< upper
		<< getMax()
		<< *it
		<< percentage
		<< round( multiplier * percentage );
	}
}

Regress::Regress( Label title ):
	Tab( title )
	{
		reset();
	}

Regress::Regress(Simulation* s, Label title):
	Tab(s, title)
	{
		reset();
	}

Regress::~Regress() {}

void Regress::update(double vx, double vy) {
	Tab::update();

	sumX += vx;
	sumY += vy;
	sumSqX += vx*vx;
	sumXMulY += vx*vy;
	sumSqY += vy*vy;
}

void Regress::reset() {
	Tab::reset();

	sumX = sumY = sumSqX = sumXMulY = sumSqY = 0.0;
}

double Regress::getXMean() {
	return getObs() ? sumX/getObs() : 0;
}

double Regress::getYMean() {
	return getObs() ? sumY/getObs() : 0;
}

double Regress::getDx() {
	return fabs(getObs()*sumSqX - sumX*sumX);
}

double Regress::getDy() {
	return fabs(getObs()*sumSqY - sumY*sumY);
}

double Regress::getResidualStandardDeviation() {
	// RES.ST.DEV
	double sd = sqrt((sumSqY - getIntercept()*sumY - getEstimatedRegressionCoefficient()*sumXMulY)/(getObs() - 2));
	return sd;
}

double Regress::getEstimatedRegressionCoefficient() {
	// EST.REG.COEFF
	double a1 = (getObs()*sumXMulY - sumX*sumY)/getDx();
	return a1;
}

double Regress::getIntercept() {
	// INTERCEPT
	double a0 = (sumY*sumSqX - sumX*sumXMulY)/getDx();
	return a0;
}

double Regress::getRegressCoefficientStandardDeviation() {
	// ST.DEV.REG.COEFF
	return (getObs()*getResidualStandardDeviation()/sqrt((getObs()-2)*getDx()));
}

double Regress::getCorrelationCoefficient() {
	// CORR.COEFF
	double r2 = ((getObs()*sumXMulY - sumX*sumY)*(getObs()*sumXMulY - sumX*sumY))/(getDx()*getDy());
	return r2;
}

void Regress::report(Report* r) {

	assert(r);

	static vector< string > labels;			static vector< Column::Type  > types;
	labels.push_back("Name");				types.push_back( Column::STRING);
	labels.push_back("Reset at");			types.push_back( Column::SIMTIME);
	labels.push_back("Uses");				types.push_back( Column::INTEGER);
	labels.push_back("XBar");				types.push_back( Column::REAL);
	labels.push_back("YBar");				types.push_back( Column::REAL);
	labels.push_back("Dx");					types.push_back( Column::REAL);
	labels.push_back("Dy");					types.push_back( Column::REAL);
	labels.push_back("Residual Standard Deviation");		types.push_back( Column::REAL);
	labels.push_back("Estimated Regression Coefficient");	types.push_back( Column::REAL);
	labels.push_back("Intercept");			types.push_back( Column::REAL);
	labels.push_back("Regression Coefficient Standard Deviation");	types.push_back( Column::REAL);
	labels.push_back("Correlation Coefficient");			types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("Regress Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< getObs()
	<< getXMean()
	<< getYMean()
	<< getDx()
	<< getDy()
	<< getResidualStandardDeviation()
	<< getEstimatedRegressionCoefficient()
	<< getIntercept()
	<< getRegressCoefficientStandardDeviation()
	<< getCorrelationCoefficient();
}
