//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2003, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Statistics.h

	\author Ralf Gerstenberger

	\date created at 2003/05/20

	\brief Declaration of statistic-classes

	\sa Statistics.cpp

	<!-- [detailed description] -->

	\since 1.0
*/

#ifndef ODEMX_STATISTICS_INCLUDED
#define ODEMX_STATISTICS_INCLUDED

#include <vector>

#include <odemx/util/Report.h>
#include <odemx/util/LabeledObject.h>
#include <odemx/util/StatisticObject.h>
#include <odemx/base/Simulation.h>

namespace odemx {
	/**
		\defgroup statistics Statistics

		A group of classes which provide tools for statistical analysis of
		simulations.
	*/

	/** \class Tab

		\ingroup statistics

		\author Ralf Gerstenberger

		\brief Base-class for statistics

		\note NOT Tab from ODEM

		<!-- [\sa {references to other classes}]* -->

		<!-- [detailed description] -->

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class Tab : public DefLabeledObject,
				public StatisticObject,
				public virtual ReportProducer
	{
	public:
		/// Construction for DefaultSimulation
		Tab( Label label="" );
		/// Construction for user-defined Simulation
		Tab(Simulation* s, Label label="");
		/// Destruction
		virtual ~Tab();

		/// update usage counter
		void update() {obs++;}
		/// reset usage counter and StatisticObject base
		virtual void reset() {obs=0; StatisticObject::reset();}
		/// get usage counter
		unsigned int getObs() const {return obs;}

	protected:
		unsigned int obs; ///< the usage counter
	};

	/** \class Count

		\ingroup statistics

		\author Ralf Gerstenberger

		\brief Counter

		\note Count from ODEM

		\note Count supports Report

		<!-- [\sa {references to other classes}]* -->

		Count is counting integers. It can be used also to
		calculate an integer sum of all provided numbers.
		To increase the counter use update().

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class Count : public Tab {
	public:
		/// Construction for DefaultSimulation
		Count( Label title="" );
		/// Construction for user-defined Simulation
		Count (Simulation* s, Label title="");
		/// Destruction
		virtual ~Count();

		/// Update counter
		void update (int v=1) {
			Tab::update();
			value += v;
		};

		/// Reset counter and statistics
		virtual void reset() {
			Tab::reset();
			value = 0;
		}

		/// Get current counter value
		int getValue() {return value;}

		/// Report generation
		virtual void report(Report* r);

	private:
		int value;
	};

	/** \class Sum

		\ingroup statistics

		\author Ralf Gerstenberger

		\brief Sum

		\note Sum from ODEM

		\note Sum supports Report

		<!-- [\sa {references to other classes}]* -->

		Sum is used to compute 'double' sums. Use update()
		to change the sum.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class  Sum : public Tab {
	public:
		/// Construction for DefaultSimulation
		Sum ( Label title="" );
		/// Construction for user-defined Simulation
		Sum (Simulation* s, Label title="");
		/// Destruction
		virtual ~Sum();

		/// Update sum
		void update (double v=1.0) {
			Tab::update();
			value += v;
		};

		/// Reset sum and statistics
		virtual void reset() {
			Tab::reset();
			value = 0.0;
		}

		/// Get current value
		double getValue() {return value;}

		/// Report generation
		virtual void report(Report* r);

	private:
		double value;
	};

	/** \class Tally

		\ingroup statistics

		\author Ralf Gerstenberger

		\brief Compute statistics about provided data

		\note Tally from ODEM

		\note Tally supports Report

		<!-- [\sa {references to other classes}]* -->

		Tally computes minimum, maximum, mean-value,
		standard deviation and number of samples (size)
		independent from simulation time (see Accum).

		The average value is computed as follows:
		A value of 1 for 4 time units together with a value
		of 0 for 1 time unit will result in a mean value of
		(1+0)/2 = 0.5.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class Tally : public Tab {
	public:
		/// Construction for DefaultSimulation
		Tally( Label title="" );
		/// Construction for user-defined Simulation
		Tally (Simulation* s, Label title="");
		/// Destruction
		virtual ~Tally();

		/// Update statistics
		virtual void update (double v);
		/// Reset statistics
		virtual void reset();

		/// Get number of samples
		unsigned int getSize() {return getObs();}
		/// Get minimum
		double getMin() {return min;}
		/// Get maximum
		double getMax() {return max;}
		/// Get average
		double getMean() {return getObs() ? sum/getObs() : 0.0;}
		/// Get standard deviation
		double getStandardDeviation();

		/// Report generation
		virtual void report(Report* r);

	protected:
		double sum, sumsq, min, max;
	};

	/** \class Accum

		\ingroup statistics

		\author Ralf Gerstenberger

		\brief Compute statistics about provided data

		\note Accum from ODEM
		\note Accum supports Report

		<!-- [\sa {references to other classes}]* -->

		Accum computes minimum, maximum, mean-value,
		standard deviation and number of samples (size) according to
		the simulation time (see Tally).

		Time weighted mean values are computed according to two
		different approximation stategies:

		\c Accum::stepwise: assumes constant values between samples

			A value of 1 held for 4 time units, followed by a
			value of 0 for 1 time unit will result in a mean
			value of (1*4 + 0*1)/5 = 0.8 .

		\c Accum::linear: assumes linear increase/decrease of values
		between samples

			A value of 1 held for 4 time units, followed by a
			value of 0 for 1 time unit will result in a mean
			value of (1*4/2 + 0*1)/5 = 0.4 .

		\note The approximation strategy can be set with the second constructor
		parameter, which by default is set to \c stepwise.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class Accum : public Tab {
	public:
		enum Approx {stepwise, linear};

		/// Construction for DefaultSimulation
		Accum( Label title="", Approx app = stepwise );
		/// Construction for user-defined Simulation
		Accum (Simulation* s, Label title="", Approx app = stepwise);
		/// Destruction
		virtual ~Accum();

		/// Update statistics for time-discrete variables
		virtual void update (double v);
		/// Reset statistics
		virtual void reset();

		/// Get number of samples
		unsigned int getSize() const {return getObs();}
		/// Get minimum
		double getMin() const {return min;}
		/// Get maximum
		double getMax() const {return max;}
		/// Get average (over time)
		double getMean() const;
		/// Get divergence
		double getStandardDeviation() const;

		/// Report generation
		virtual void report(Report* r);

	protected:
		double sumt, sumsqt, min, max, lastv;
		SimTime lasttime;
		Approx approx;
	};

	/** \class Histo

		\ingroup statistics

		\author Ralf Gerstenberger

		\brief Statistical analysis plus histogram

		\note Histo from ODEM
		\note Histo supports Report

		<!-- [\sa {references to other classes}]* -->

		Histo computes a statistical analysis like Tally
		and also generates a histogram.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class Histo : public Tally {
	public:
		/**
			\brief Construction for DefaultSimulation
			\param title
				label of this object
			\param low
				lowest border
			\param up
				highest border
			\param n
				number of categories
		*/
		Histo( Label title, double low, double up, int n );

		/**
			\brief Construction for user-defined Simulation
			\param s
				pointer to Simulation object
			\param title
				label of this object
			\param low
				lowest border
			\param up
				highest border
			\param n
				number of categories
		*/
		Histo(Simulation* s, Label title, double low, double up, int n);
		/// Destruction
		virtual ~Histo();

		/// Update statistics
		virtual void update(double v);
		/// Reset statistics
		virtual void reset();

		/// Get Tally-statistics
		const Tally* getTally() {return static_cast<const Tally*>(this);}
		/// Get histogram data
		const std::vector<int>& getData() {return data;}

		/// Report generation
		virtual void report(Report* r);

	protected:
		double lower, upper;
		int ncells;
		std::vector<int> data;
		int limit;
		double width;

	private:

		long round( double d );
	};

	/** \class Regress

		\ingroup statistics

		\author Ralf Gerstenberger

		\brief %Regress analysis

		\note Regress from ODEM
		\note Regress supports Report

		<!-- [\sa {references to other classes}]* -->

		Regress computes the linear correlation between
		provided data X and Y by computing two mean values,
		whose intersection is the center of rotation, and a slope.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class Regress : public Tab {
	public:
		/// Construction for DefaultSimulation
		Regress( Label title );
		/// Construction for user-defined Simulation
		Regress(Simulation* s, Label title);
		/// Destruction
		virtual ~Regress();

		/// Update statistics
		void update(double vx, double vy);
		/// Reset statistics
		virtual void reset();

		/// Has sufficient data for a regress analysis?
		bool hasSufficientData() {return getObs()>5;}

		/// get mean of x update values
		double getXMean();
		/// get mean of y update values
		double getYMean();

		///
		double getDx();
		double getDy();

		/**
		 * @brief get standard deviation of all residuals
		 *
		 * @note Residuals describe the distance of the measured value from
		 * the estimated linear regression function.
		 */
		double getResidualStandardDeviation(); // RES.ST.DEV

		/**
		 * @brief get the regression coefficient
		 *
		 * @note The regression coefficient describes the slope of the linear
		 */
		double getEstimatedRegressionCoefficient(); // EST.REG

		/// intercept helper value ???
		double getIntercept(); // INTERCEPT

		/**
		 * @brief get standard deviation of the regression coefficient estimates
		 */
		double getRegressCoefficientStandardDeviation(); // ST.DEV.REG.COEFF

		/**
		 * @brief get the correlation coefficient
		 *
		 * The correlation coefficient is a measure for the relationship of
		 * x and y. \c 0 means no linear correlation exists, while \c 1 means
		 * there is a strong linear correlation between these variables.
		 */
		double getCorrelationCoefficient(); // CORR.COEFF

		/// Report generation
		virtual void report(Report* r);

	protected:
		double sumX, sumY, sumSqX, sumXMulY, sumSqY;
	};
}

#endif

