//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Bin.cpp

	\author Ralf Gerstenberger

	\date created at 2002/03/26

	\brief Implementation of classes in Bin.h

	\sa Bin.h

	<!-- [detailed description] -->

	\since 1.0
*/

#include <odemx/base/InSort.h>
#include <odemx/util/ErrorHandling.h>
#include <odemx/synchronization/Bin.h>

#include <string>
#include <cassert>
#include <algorithm>

using namespace std;
using namespace odemx;

Bin::Bin( Label l, unsigned int startTokenNumber, BinObserver* o/*=0*/ ):
	Observable<BinObserver>(o),
	DefLabeledObject( getDefaultSimulation(), l ),
	StatisticObject( getDefaultSimulation() ),
	env( getDefaultSimulation() ),
	tokenNumber( startTokenNumber ),
	initToken( startTokenNumber ),
	tokenStatistics( l + " token statistics" ),
	takeWait( l + " queue" )
	{
		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEach(BinObserver, Create(this));

		// Init Statistics
		reset();
	}

Bin::Bin( Simulation* s, Label l, unsigned int startTokenNumber, BinObserver* o/*=0*/ ):
	Observable<BinObserver>(o),
	DefLabeledObject(s, l),
	StatisticObject(s),
	env(s),
	tokenNumber(startTokenNumber),
	initToken(startTokenNumber),
	tokenStatistics(s, ( l + " token statistics" )),
	takeWait(s, (l + " queue"))
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEach(BinObserver, Create(this));

		// Init Statistics
		reset();
	}

Bin::~Bin() {
	// trace
	getTrace()->mark(this, markDestroy);

	// observer
	_obsForEach(BinObserver, Destroy(this));
}

// Token
unsigned int Bin::take(unsigned int n) {

	// resource handling only implemented for processes
	if( !env -> getCurrentSched() ||
		env -> getCurrentSched() -> getSchedType() != Sched::PROCESS ) {

		error( this,  "Bin::take(): called by non-Process object" );
		return 0;
	}

	// compute order of service
	takeWait.inSort(getCurrentProcess());

	if (n > tokenNumber ||
		takeWait.getTop()!=getCurrentProcess()) {
		// not enough token or not the
		// first process to serve

		// trace
		traceTSFG(markTakeFail, n);

		// observer
		_obsForEach(BinObserver, TakeFail(this, n));

		// statistics
		SimTime waitTime=env->getTime();

		// block execution
		while (n>tokenNumber || takeWait.getTop()!=getCurrentProcess()) {
			getCurrentProcess()->sleep();

			if (getCurrentProcess()->isInterrupted()) {
				takeWait.remove(getCurrentProcess());
				return 0;
			}
		}

		// block released
		// statistics
		sumWaitTime+=env->getTime()-waitTime;
	}

	// trace
	getTrace()->mark(this, markChangeTokenNumber, tokenNumber-n, tokenNumber);

	// observer
	_obsAForEach(BinObserver, TokenNumber, tokenNumber, tokenNumber-n);

	// take token
	tokenNumber -= n;

	// statistics
	++users;
	tokenStatistics.update(tokenNumber);

	// trace
	traceTSFG(markTakeSucceed, n);

	// observer
	_obsForEach(BinObserver, TakeSucceed(this, n));

	// remove from list
	takeWait.remove(getCurrentProcess());

	// awake next process
	awakeFirst(&takeWait);

	return n;
}

void Bin::give(unsigned int n) {
	// trace
	getTrace()->mark(this, markChangeTokenNumber, tokenNumber+n, tokenNumber);

	// observer
	_obsAForEach(BinObserver, TokenNumber, tokenNumber, tokenNumber+n);

	// releas token
	tokenNumber += n;

	// statistics
	++providers;
	tokenStatistics.update(tokenNumber);

	// trace
	traceTSFG(markGive, n);

	// observer
	_obsForEach(BinObserver, Give(this, n));

	// awake process waiting for give
	awakeFirst(&takeWait);
}

void Bin::reset() {
	// reset time
	StatisticObject::reset();

	// reset queue statistics
	takeWait.reset();

	// Accum reset and re-init
	tokenStatistics.reset();
	tokenStatistics.update(getTokenNumber());

	users=providers=0;
	sumWaitTime=0;
}

double Bin::getAVWaitTime() const {
	return sumWaitTime/getUsers();
}

void Bin::report(Report* r) {

	assert(r);
	takeWait.report(r);

	static vector< string > labels;			static vector< Column::Type  > types;
	labels.push_back("Name");				types.push_back( Column::STRING);
	labels.push_back("Reset at");			types.push_back( Column::SIMTIME);
	labels.push_back("Queue");				types.push_back( Column::STRING);
	labels.push_back("Users");				types.push_back( Column::INTEGER);
	labels.push_back("Providers");			types.push_back( Column::INTEGER);
	labels.push_back("Init token number");	types.push_back( Column::INTEGER);
	labels.push_back("Min token number");	types.push_back( Column::INTEGER);
	labels.push_back("Max token number");	types.push_back( Column::INTEGER);
	labels.push_back("Now token number");	types.push_back( Column::INTEGER);
	labels.push_back("Avg token number");	types.push_back( Column::REAL);
	labels.push_back("Avg waiting time");	types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("Bin Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< takeWait.getLabel()
	<< getUsers()
	<< getProviders()
	<< getInitial()
	<< getMin()
	<< getMax()
	<< getTokenNumber()
	<< getAVFreeToken()
	<< getAVWaitTime();
}

void Bin::traceTSFG(MarkType m, unsigned int n) {
	getTrace()->beginMark(this, m);
	getTrace()->addTag(tagCurrent, getCurrentProcess());
	getTrace()->addTag(tagTokenNumber, n);
	getTrace()->endMark();
}

const MarkTypeId Bin::baseMarkId = MarkType::getNewBaseMarkId( typeid(Bin) );

const MarkType Bin::markCreate( "create", baseMarkId+1, typeid(Bin)); ///< Call of Bin constructor
const MarkType Bin::markDestroy( "destroy", baseMarkId+2, typeid(Bin)); ///< Call of Bin destructor
const MarkType Bin::markTakeFail( "take failed", baseMarkId+3, typeid(Bin)); ///< Bin does not contain enough tokens, caller Process is enqueued and suspended
const MarkType Bin::markTakeSucceed( "take succeeded", baseMarkId+4, typeid(Bin)); ///< Bin contains enough tokens, caller Process succeeded in acquiring the requested number
const MarkType Bin::markGive( "give", baseMarkId+5, typeid(Bin)); ///< Bin was supplied with the number of tokens shown
const MarkType Bin::markChangeTokenNumber( "change token number", baseMarkId+6, typeid(Bin)); ///< Bin token number changed, displays old and new values

const TagId Bin::baseTagId = Bin::baseMarkId;

const Tag Bin::tagCurrent = Tag( "current" , baseTagId+1); ///< Displays details about the currently active Process
const Tag Bin::tagTokenNumber = Tag( "token number", baseTagId+2); ///< Displays details about the Bin token number

