//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file CondQ.cpp

	\author Ralf Gerstenberger

	\date created at 2002/07/11

	\brief Implementation of classes in CondQ.h

	\sa CondQ.h

	<!-- [detailed description] -->

	\since 1.0
*/

#include <odemx/base/InSort.h>
#include <odemx/util/ErrorHandling.h>
#include <odemx/synchronization/CondQ.h>

#include <string>
#include <cassert>

using namespace std;
using namespace odemx;

CondQ::CondQ( Label l, CondQObserver* o/* = 0*/ ):
	Observable<CondQObserver>(o),
	DefLabeledObject( getDefaultSimulation(), l ),
	StatisticObject( getDefaultSimulation() ),
	env( getDefaultSimulation() ),
	processes( l + " queue" )
	{
		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEach(CondQObserver, Create(this));

		// init statistics
		reset();
	}

CondQ::CondQ(Simulation* s, Label l, CondQObserver* o/* = 0*/) :
	Observable<CondQObserver>(o),
	DefLabeledObject(s, l),
	StatisticObject(s),
	env(s),
	processes(s, (l + " queue"))
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEach(CondQObserver, Create(this));

		// init statistics
		reset();
	}

CondQ::~CondQ() {
	// trace
	getTrace()->mark(this, markDestroy);

	// observer
	_obsForEach(CondQObserver, Destroy(this));
}

bool CondQ::wait(Condition cond) {
	// trace
	getTrace()->mark(this, markWait, getCurrentProcess());

	// observer
	_obsForEach(CondQObserver, Wait(this, getCurrentProcess()));

	// wait for cond is true
	processes.inSort(getCurrentProcess());

	// statistics
	SimTime t=env->getTime();

	while (!(getCurrentProcess()->*cond)() ) {
		getCurrentProcess()->sleep();

		if (getCurrentProcess()->isInterrupted()) {
			processes.remove(getCurrentProcess());

			return false;
		}
	}

	processes.remove(getCurrentProcess());

	// statistics
	t = env->getTime() - t;
	sumWaitTime += t;
	if (t==0)
		++zeros;

	++users;

	// trace
	getTrace()->mark(this, markContinue, getCurrentProcess());

	// observer
	_obsForEach(CondQObserver, Continue(this, getCurrentProcess()));

	return true;
}

void CondQ::signal() {
	// trace
	getTrace()->mark(this, markSignal, getCurrentProcess());

	// observer
	_obsForEach(CondQObserver, Signal(this, getCurrentProcess()));

	// statistics
	++signals;

	if (processes.isEmpty())
		return;

	// test conditions
	awakeAll(&processes);
}

void CondQ::reset() {
	StatisticObject::reset();
	processes.reset();

	users = processes.getLength();
	zeros = signals = 0;
	sumWaitTime = 0.0;
}

void CondQ::report(Report* r) {

	assert (r!=0);
	processes.report(r);

	static vector< string > labels;			static vector< Column::Type  > types;
	labels.push_back("Name");				types.push_back( Column::STRING);
	labels.push_back("Reset at");			types.push_back( Column::SIMTIME);
	labels.push_back("Queue");				types.push_back( Column::STRING);
	labels.push_back("Users");				types.push_back( Column::INTEGER);
	labels.push_back("Signals");			types.push_back( Column::INTEGER);
	labels.push_back("Zero wait users");	types.push_back( Column::INTEGER);
	labels.push_back("Avg waiting time");	types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("CondQ Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< processes.getLabel()
	<< getUsers()
	<< getSignals()
	<< getZeroWait()
	<< getAVWaitTime();
}

const MarkTypeId CondQ::baseMarkId = MarkType::getNewBaseMarkId( typeid(CondQ) );

const MarkType CondQ::markCreate("create", baseMarkId+1, typeid(CondQ)); ///< Call of CondQ constructor
const MarkType CondQ::markDestroy( "destroy", baseMarkId+2, typeid(CondQ)); ///< Call of CondQ destructor
const MarkType CondQ::markWait( "wait", baseMarkId+3, typeid(CondQ)); ///< Process enqueued, caller is suspended until the given condition returns true
const MarkType CondQ::markContinue( "continue", baseMarkId+4, typeid(CondQ)); ///< Condition fulfilled, the waiting Process can continue its execution
const MarkType CondQ::markSignal( "signal", baseMarkId+5, typeid(CondQ)); ///< Condition check triggered, all Processes are awakened to check their waiting condition

