//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Memory.cpp

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Implementation of classes in Memory.h

	\sa Memory.h

	<!-- [detailed description] -->

	\since 2.0
*/

#include <odemx/base/Process.h>
#include <odemx/synchronization/Memory.h>
#include <odemx/util/ErrorHandling.h>

#include <algorithm>

using namespace odemx;
using namespace std;

Memory::Memory( Label l, Type t, MemoryObserver* mo ):
	Observable<MemoryObserver>( mo ),
	StatisticObject( getDefaultSimulation() ),
	env( getDefaultSimulation() ),
	type( t ),
	memoryStatistics( l + " statistics" )
	{
		// observer
		_obsForEach(MemoryObserver, Create(this));

		reset();
	}

Memory::Memory( Simulation* sim, Label l, Type t, MemoryObserver* mo ):
	Observable<MemoryObserver>( mo ),
	StatisticObject( sim ),
	env( sim ),
	type( t ),
	memoryStatistics( sim, l + " statistics" )
	{
		// observer
		_obsForEach(MemoryObserver, Create(this));

		reset();
	}

Memory::~Memory()
	{
		// observer
		_obsForEach(MemoryObserver, Destroy(this));
	}


bool Memory::remember( Sched* newObject ) {

	// check if element is already memorized
	std::list< Sched* >::iterator iter
	= find( memoryList.begin(), memoryList.end(), newObject );

	if ( iter != memoryList.end() ) {

		warning( this,  "Memory::remember(): object already stored in memoryList" );
		return false;
	}

	// trace
	getTrace() -> mark( this, markRemember, newObject );

	// observer
	_obsForEach( MemoryObserver, Remember( this, newObject ) );

	memoryList.push_back( newObject );

	memoryStatistics.update(getNow());

	return true;
}


bool Memory::forget( Sched* rememberedObject ) {

	// check if rememberedObject is in memoryList
	std::list<Sched*>::iterator iter;

	iter = find( memoryList.begin(), memoryList.end(), rememberedObject );

	// if found, remove process from memory
	if ( iter != memoryList.end() ) {

		//trace
		getTrace() -> mark( this, markForget, rememberedObject );

		// observer
		_obsForEach( MemoryObserver, Forget( this, rememberedObject ) );

		memoryList.erase(iter);

		memoryStatistics.update(getNow());
	}
	else {

		// error: object not found
		error( this,  "Memory::forget(): object not in memoryList" );
		return false;
	}

	return true;
}


// remove all registered objects from memory
void Memory::eraseMemory() {

	memoryList.clear();

	memoryStatistics.update(getNow());
}


// wake up all CREATED or IDLE processes in memory (i.e. reschedule them)
void Memory::alert() {

	if ( memoryList.empty() )
	{
		// warning: no object to wake
		warning( this,  "Memory::alert(): no process saved" );
		return;
	}

	// trace
	traceAlert();

	// observer
	_obsForEach( MemoryObserver, Alert( this ) );

	// iterate over memoryList and schedule processes
	std::list< Sched* >::iterator iter;

	for ( iter = memoryList.begin(); iter != memoryList.end(); ++iter )
	{
		// check if Sched* really is a sleeping process
		if ( checkSchedForAlert( *iter ) )
		{
			// wake up the process
			static_cast< Process* >( *iter ) -> alertProcess( this );
		}
		else
		{
			warning( this, ( "Memory::alert(): could not alert Sched object: "
					 + ( *iter ) -> getLabel() ) );
		}
	}

	// remove stored objects from memory
	eraseMemory();
}

bool Memory::checkSchedForAlert( Sched* s )
{
	// check if p really is a sleeping process
	if( s -> getSchedType() != Sched::PROCESS )
	{
		// warning: not a process, no action taken
		warning( this,  "Memory::checkSchedForAlert(): remembered object is not a process" );
		return false;
	}

	Process* p = static_cast< Process* >( s );

	// only sleeping Processes are allowed to be alerted
	if ( p -> getProcessState() == Process::IDLE ||
		 p -> getProcessState() == Process::CREATED )
	{
		return true;
	}
	else if( p -> getProcessState() == Process::TERMINATED )
	{
		// warning: process terminated
		error( this,  "Memory::checkSchedForAlert(): process already terminated" );
	}
	else if( p -> getProcessState() == Process::RUNNABLE )
	{
		// warning: process scheduled
		warning( this,  "Memory::checkSchedForAlert(): process already scheduled" );
	}

	return false;
}

bool Memory::waiting() const {
	return ! memoryList.empty();
}


size_t Memory::countWaiting() const {
	return memoryList.size();
}


const std::list< Sched* >& Memory::getList() const {
	return memoryList;
}


Memory::Type Memory::getMemoryType() const {
	return type;
}


void Memory::reset() {
	StatisticObject::reset();
	memoryStatistics.reset();
	memoryStatistics.update( getNow() );
}


void Memory::report( Report* r ) {
	assert(r);

	static vector< string > labels;			static vector< Column::Type  > types;
	labels.push_back("Name");				types.push_back( Column::STRING);
	labels.push_back("Reset at");			types.push_back( Column::SIMTIME);
	labels.push_back("Min Memory use");	types.push_back( Column::INTEGER);
	labels.push_back("Max Memory use");	types.push_back( Column::INTEGER);
	labels.push_back("Now Memory use");	types.push_back( Column::INTEGER);
	labels.push_back("Avg Memory use");	types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("Memory Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< getMin()
	<< getMax()
	<< getNow()
	<< getAVLength();
}

void Memory::traceAlert() {

	getTrace() -> beginMark( this, markAlert );
	std::list< Sched* >::iterator iter;
	for ( iter = memoryList.begin(); iter != memoryList.end(); ++iter ) {
		getTrace() -> addTag( tagPartner, *iter );
	}
	getTrace() -> endMark();
}


const MarkTypeId Memory::baseMarkId = MarkType::getNewBaseMarkId( typeid(Memory) );
const MarkType Memory::markRemember("remember", Memory::baseMarkId+1, typeid(Memory) ); ///< Memory stored given Process in list
const MarkType Memory::markForget("forget", Memory::baseMarkId+2, typeid(Memory) ); ///< Memory removed stored Process from list
const MarkType Memory::markAlert("alert", Memory::baseMarkId+3, typeid(Memory) ); ///< Memory alerted all remembered Process objects, which are then rescheduled at SimTime now

const TagId Memory::baseTagId = Memory::baseMarkId;
const Tag Memory::tagPartner( "partner" ); ///< Displays information about the current partner Process
