//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Memory.h

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Declaration of odemx::Memory

	\sa Memory.cpp

	<!-- [detailed description] -->

	\since 2.0
*/

#ifndef ODEMX_MEMORY_INCLUDED
#define ODEMX_MEMORY_INCLUDED

#include <odemx/base/Sched.h>
#include <odemx/statistics/Statistics.h>
#include <odemx/util/MarkType.h>
#include <vector>
#include <list>

namespace odemx {

	// forward declaration
	class MemoryObserver;

	/** \class Memory

		\ingroup synch

		\author Ronald Kluth

		\brief %Memory is the base class for all structures that need to remember processes.

		\sa Process, Timer, PortHead, PortTail

		%Memory is an abstract base class for all user defined structures in a model
		that need to store a list of suspended processes or other schedulable
		objects in order to wake them when a certain event occurs.
		This can be the case for timers, but also
		for structures like buffered communication ports. %Memory offers a means
		to notify suspended processes of the availability of a resource and
		reschedule them at that point in simulation time when a resource is
		available or a timeout occurs. The recommended usage with processes is
		via their \p wait() function.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.0
	*/
	class Memory: public Observable< MemoryObserver>,
				  public StatisticObject,
				  public virtual TraceProducer,
				  public virtual ReportProducer {

	// types
	public:

	/**
		\brief Memory types

		%Memory-derivations can be different types of structures. When
		a pointer to a %Memory object is returned, it can be useful to have
		a means of distinguishing them simply by asking for their
		type using \p getMemoryType(). Currently timers, ports and an
		arbitrary waiting condition are supported by the library. However,
		users are encouraged to	derive from Memory and create a user-defined
		Memory type.
	*/
	enum Type {
		TIMER,
		PORTHEAD,
		PORTTAIL,
		WAIT_CONDITION,
		USER_DEFINED
	};

	// public interface
	public:

		friend class Process;

		/**
			\brief Construction for DefaultSimulation
			\param t
				type of this object
			\param mo
				initial observer of this object
		*/
		Memory( Label l, Type t, MemoryObserver* mo = 0 );

		/**
			\brief Construction for user-defined Simulation
			\param sim
				pointer to the simulation object
			\param t
				type of this object
			\param mo
				initial observer of this object
		*/
		Memory( Simulation* sim, Label l, Type t, MemoryObserver* mo = 0 );

		/// Destruction
		virtual ~Memory();

		/**
			\brief Process interaction

			These functions can be used by processes to interact with
			%Memory objects.

			@{
		*/
		/**
			\brief Remember a Sched object
			\param newObject
				Sched object to be saved
			\return
				\p true when successfully stored, \p false if \p newObject already stored

			This function is used to add processes or, more generally,
			schedulable objects to the \p memoryList.

			\sa forget()
		*/
		bool remember( Sched* newObject );	// add Process to object's Memory

		/**
			\brief Remove a Sched object
			\param rememberedObject
				Sched object to be removed
			\return
				\p true when successfully removed, \p false if \p rememberedObject not stored

			This function is used to remove processes or, more generally,
			schedulable objects from the \p memoryList.

			\sa remember()
		*/
		bool forget( Sched* rememberedObject );     // remove Process from object's Memory

		/**
			\brief Remove all Sched objects

			This function is used to remove all processes or, more generally,
			schedulable objects from the \p memoryList.

			\sa remember(), forget()
		*/
		void eraseMemory();     // remove Process from object's Memory
		//@}

		/**
			\name Status information

			These functions return information about the Memory object

			@{
		*/
		/**
			\brief Check availability of this Memory object

			This function is required to be implemented by all %Memory-derived
			classes. It is used to	check the availability of a
			resource before suspending the requesting process.

			\sa Process::wait(), PortHead, PortTail, Timer
		*/
		virtual bool available() = 0;

		/**
			\brief Check if this Memory object has waiting processes
		*/
		bool waiting() const;

		/**
			\brief Get the number of waiting processes
		*/
		size_t countWaiting() const;

		/**
			\brief Get a reference to the list of waiting processes
		*/
		const std::list<Sched*>& getList() const;

		/**
			\brief Check type of this Memory object

			The Type is useful when a Memory pointer is returned
			but, depending on the type of this object, different
			reactions are necessary.

			\sa Process::wait()
		*/
		Type getMemoryType() const;
		//@}

	protected:
		/**
			\brief Alert, i.e. reschedule, all stored processes

			By default, this function is designed to wake up suspended
			or newly created processes, i.e. reschedule them at simulation
			time \p now. The function can be redefined in derived classes
			to suit different needs.
		*/
		virtual void alert();

		/**
			\brief Helper for alert() which tests if \p s is a sleeping Process
			\return
				\c true if \p s can be alerted
		*/
		bool checkSchedForAlert( Sched* s );

		/**
			\brief Get pointer to Trace
			\return
				pointer to Trace object
		*/
		virtual Trace* getTrace() const { return env; }

		/// send alert mark with memory list
		void traceAlert();

	public:

		/** \name Memory statistics
		 *
		 * These functions collect data about Memory usage.
		 *
		 * @{
		 */
		/// reset statistics
		virtual void reset();

		/// min queue length
		unsigned int getMin() const {return (unsigned int)memoryStatistics.getMin();}

		/// max queue length
		unsigned int getMax() const {return (unsigned int)memoryStatistics.getMax();}

		/// current queue length
		unsigned int getNow() const {return memoryList.size();}

		/// average queue length
		double getAVLength() const {return memoryStatistics.getMean();}

		//@}

		/// generate report
		virtual void report(Report* r);

	public:
		/**
			\name Trace MarkTypes

			These MarkTypes and Tags are used to trace Memory events.
			A TraceConsumer can use these constants to identify trace
			events sent	by Memory.

			@{
		*/
		static const MarkTypeId baseMarkId;

		static const MarkType markRemember;
		static const MarkType markForget;
		static const MarkType markAlert;

		static const TagId baseTagId;

		static const Tag tagPartner;
		//@}


	private:
		Simulation* env; ///< simulation context
		Type type;   ///< this object's memo type
		std::list<Sched*> memoryList; ///< Sched object storage

		Accum memoryStatistics; ///< statistics
	};

	/// Provided for convenience
	typedef std::vector<Memory*> MemoryVector;

	/** \interface MemoryObserver

		\author Ronald Kluth

		\brief Observer for Memory-specific calls.

		<!-- [\note {notes}]* -->

		\sa Timer, PortHead, PortTail

		<!-- [detailed description] -->

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.0
	*/
	class MemoryObserver {
	public:
		virtual ~MemoryObserver() {}

		virtual void onCreate(Memory* sender) {} ///< Construction
		virtual void onDestroy(Memory* sender) {} ///< Destruction

		virtual void onRemember(Memory* sender, Sched* s) {} ///< Remember \p s
		virtual void onForget(Memory* sender, Sched* s) {} ///< Forget \p s
		virtual void onAlert(Memory* sender) {} ///< Alert all processes
	};
}

#endif /*ODEMX_MEMORY_INCLUDED*/

