//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Res.cpp

	\author Ralf Gerstenberger

	\date created at 2002/03/21

	\brief Implementation of classes in Res.h

	\sa Res.h

	<!-- [detailed description] -->

	\since 1.0
*/

#include <odemx/synchronization/Res.h>
#include <odemx/base/InSort.h>
#include <odemx/util/ErrorHandling.h>

#include <string>
#include <cassert>
#include <algorithm>

using namespace std;
using namespace odemx;

Res::Res( Label l,
		  unsigned int startTokenNumber,
		  unsigned int maxTokenNumber,
		  ResObserver* o/* = 0*/):
	Observable<ResObserver>(o),
	DefLabeledObject( getDefaultSimulation(), l),
   	StatisticObject( getDefaultSimulation() ),
	env( getDefaultSimulation() ),
	tokenNumber( startTokenNumber ),
	maxToken( maxTokenNumber ),
   	initToken( startTokenNumber ),
	tokenStatistics( l + " token statistics" ),
	acquireWait( l + " queue" )
	{
		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEach(ResObserver, Create(this));

		// Init statistics
		reset();
	}

Res::Res( Simulation* s,
		  Label l,
		  unsigned int startTokenNumber,
		  unsigned int maxTokenNumber,
		  ResObserver* o/* = 0*/):
	Observable<ResObserver>(o),
	DefLabeledObject(s, l),
   	StatisticObject(s),
	env(s),
	tokenNumber(startTokenNumber),
	maxToken(maxTokenNumber),
   	initToken(startTokenNumber),
	tokenStatistics(s, (l + " token statistics")),
	acquireWait(s, (l + " queue"))
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEach(ResObserver, Create(this));

		// Init statistics
		reset();
	}

Res::~Res()
	{
		// trace
		getTrace()->mark(this, markDestroy);

		// observer
		_obsForEach(ResObserver, Destroy(this));
	}

// Token
unsigned int Res::acquire(unsigned int n) {

	// resource handling only implemented for processes
	if( !env -> getCurrentSched() ||
		env -> getCurrentSched() -> getSchedType() != Sched::PROCESS ) {

		error( this,  "Res::acquire(): called by non-Process object" );
		return 0;
	}

	if( n > getTokenLimit() ) {
		// warning: will always block
		warning( this, "Res::acquire(): requesting more than max tokens will always block" );
	}

	// compute order of service
	acquireWait.inSort(getCurrentProcess());

	if (n>tokenNumber ||
		acquireWait.getTop()!=getCurrentProcess()) {
		// not enough token or not the
		// first process to serve

		// trace
		traceARSF(markAcquireFail, n);

		// observer
		_obsForEach(ResObserver, AcquireFail(this, n));

		// statistics
		SimTime waitTime=env->getTime();

		// block execution
		while (n>tokenNumber || acquireWait.getTop()!=getCurrentProcess()) {

			getCurrentProcess()->sleep();

			if (getCurrentProcess()->isInterrupted()) {
				acquireWait.remove(getCurrentProcess());
				return 0;
			}
		}

		// block released
		// statistics
		sumWaitTime+=env->getTime()-waitTime;
	}

	// trace
	getTrace()->mark(this, markChangeTokenNumber, tokenNumber-n, tokenNumber);

	// observer
	_obsAForEach(ResObserver, TokenNumber, tokenNumber, tokenNumber-n);

	// acquire token
	tokenNumber -= n;

	// statistics
	++users;
	tokenStatistics.update(tokenNumber);

	// trace
	traceARSF(markAcquireSucceed, n);

	// observer
	_obsForEach(ResObserver, AcquireSucceed(this, n));

	// remove from list
	acquireWait.remove(getCurrentProcess());

	// awake next process
	awakeFirst(&acquireWait);

	return n;
}

void Res::release(unsigned int n) {

	if ((tokenNumber+n)>maxToken) {
		// Not enough room left.
		// trace
		traceARSF(markReleaseFail, n);

		// observer
		_obsForEach(ResObserver, ReleaseFail(this, n));

		// Error:
		error( this, "Res::release(): limit reached, could not release all tokens");

		// change token number
		n=maxToken-tokenNumber;
	}

	// trace
	getTrace()->mark(this, markChangeTokenNumber, tokenNumber+n, tokenNumber);

	// observer
	_obsAForEach(ResObserver, TokenNumber, tokenNumber, tokenNumber+n);

	// release token
	tokenNumber += n;

	// statistics
	++providers;
	tokenStatistics.update(tokenNumber);

	// trace
	traceARSF(markReleaseSucceed, n);

	// observer
	_obsForEach(ResObserver, ReleaseSucceed(this, n));

	// awake process waiting for release
	awakeFirst(&acquireWait);
}

void Res::control(unsigned int n) {
	// change max token limit
	maxToken+=n;

	// add token
	release(n);
}

unsigned int Res::unControl(unsigned int n) {
	unsigned int tmp;
	// secure enough token
	tmp=acquire(n);

	// remove token
	maxToken-=tmp;

	return tmp;
}

void Res::reset() {
	// reset time
	StatisticObject::reset();

	// reset queue statistics
	acquireWait.reset();

	// Accum reset and re-init
	tokenStatistics.reset();
	tokenStatistics.update(getTokenNumber());

	users=providers=0;
	sumWaitTime=0;
}

double Res::getAVWaitTime() const {
	return sumWaitTime/getAcquires();
}

void Res::report(Report* r) {

	assert(r);
	acquireWait.report(r);

	static vector< string > labels;			static vector< Column::Type  > types;
	labels.push_back("Name");				types.push_back( Column::STRING);
	labels.push_back("Reset at");			types.push_back( Column::SIMTIME);
	labels.push_back("Queue");				types.push_back( Column::STRING);
	labels.push_back("Acquires");			types.push_back( Column::INTEGER);
	labels.push_back("Releases");			types.push_back( Column::INTEGER);
	labels.push_back("Init token number");	types.push_back( Column::INTEGER);
	labels.push_back("Limit token number");	types.push_back( Column::INTEGER);
	labels.push_back("Min token number");	types.push_back( Column::INTEGER);
	labels.push_back("Max token number");	types.push_back( Column::INTEGER);
	labels.push_back("Now token number");	types.push_back( Column::INTEGER);
	labels.push_back("Avg token number");	types.push_back( Column::REAL);
	labels.push_back("Avg waiting time");	types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("Res Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< acquireWait.getLabel()
	<< getAcquires()
	<< getReleases()
	<< getInitial()
	<< getTokenLimit()
	<< getMin()
	<< getMax()
	<< getTokenNumber()
	<< getAVFreeToken()
	<< getAVWaitTime();
}

void Res::traceARSF(MarkType m, unsigned int n) {
	getTrace()->beginMark(this, m);
	getTrace()->addTag(tagCurrent, getCurrentProcess());
	getTrace()->addTag(tagTokenNumber, n);
	getTrace()->endMark();
}

const MarkTypeId Res::baseMarkId = MarkType::getNewBaseMarkId( typeid(Res) );

const MarkType Res::markCreate( "create", baseMarkId+1, typeid(Res)); ///< Call of Res constructor
const MarkType Res::markDestroy( "destroy", baseMarkId+2, typeid(Res)); ///< Call of Res destructor
const MarkType Res::markAcquireFail( "acquire fail", baseMarkId+3, typeid(Res)); ///< Res does not contain a sufficient number of tokens, caller Process is enqueued and suspended
const MarkType Res::markAcquireSucceed( "acquire succeed", baseMarkId+4, typeid(Res)); ///< Res contained enough tokens, caller Process can now continue with its execution
const MarkType Res::markReleaseFail( "release fail", baseMarkId+5, typeid(Res)); ///< Res could not contain all released tokens, only releasing tokens till limit reached
const MarkType Res::markReleaseSucceed( "release succeed", baseMarkId+6, typeid(Res)); ///< Res capacity high enough, Process successfully returned its used tokens
const MarkType Res::markChangeTokenNumber( "change token number", baseMarkId+7, typeid(Res)); ///< Res token number changed, displays old and new values

const TagId Res::baseTagId = Res::baseMarkId;

const Tag Res::tagCurrent = Tag( "current", baseTagId+1); ///< Displays details about the currently active Process
const Tag Res::tagTokenNumber = Tag( "token number", baseTagId+2); ///< Displays details about the Res token number

