//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2003, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Wait.cpp

	\author Ralf Gerstenberger

	\date created at 2003/06/06

	\brief Implementation of classes in Wait.h

	\sa Wait.h

	<!-- [detailed description] -->

	\since 1.0
*/

#include <odemx/synchronization/Wait.h>

using namespace std;
using namespace odemx;

Wait::Wait( Label l ):
	DefLabeledObject( getDefaultSimulation(), l),
	env( getDefaultSimulation() ),
	waitForAll( true )
	{
		// trace
		getTrace()->mark(this, markCreate);
	}

Wait::Wait(Simulation* s, Label l):
	DefLabeledObject(s, l),
	env(s),
	waitForAll(true)
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);
	}

Wait::Wait( Label l, Process* oP1 ):
	DefLabeledObject( getDefaultSimulation(), l),
	env( getDefaultSimulation() ),
	waitForAll( true )
	{
		// trace
		getTrace()->mark(this, markCreate);

		Process** ppP=&oP1;
		initObservedList(1, ppP);
	}

Wait::Wait(Simulation* s, Label l, Process* oP1):
	DefLabeledObject(s, l),
	env(s),
	waitForAll(true)
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);

		Process** ppP=&oP1;
		initObservedList(1, ppP);
	}

Wait::Wait( Label l, Process* oP1, Process* oP2, bool all ):
	DefLabeledObject( getDefaultSimulation(), l ),
	env( getDefaultSimulation() ),
	waitForAll( all )
	{
		// trace
		getTrace()->mark(this, markCreate);

		Process* ppP[2]={oP1, oP2};
		initObservedList(2, ppP);
	}

Wait::Wait(Simulation* s, Label l, Process* oP1, Process* oP2, bool all) :
	DefLabeledObject(s, l),
	env(s),
	waitForAll(all)
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);

		Process* ppP[2]={oP1, oP2};
		initObservedList(2, ppP);
	}

Wait::Wait( Label l, Process* oP1, Process* oP2, Process* oP3, bool all ):
	DefLabeledObject( getDefaultSimulation(), l ),
	env( getDefaultSimulation() ),
	waitForAll( all )
	{
		// trace
		getTrace()->mark(this, markCreate);

		Process* ppP[3]={oP1, oP2, oP3};
		initObservedList(3, ppP);
	}

Wait::Wait(Simulation* s, Label l, Process* oP1, Process* oP2, Process* oP3, bool all) :
	DefLabeledObject(s, l),
	env(s),
	waitForAll(all)
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);

		Process* ppP[3]={oP1, oP2, oP3};
		initObservedList(3, ppP);
	}

Wait::Wait( Label l, int size, Process* oP[], bool all ):
	DefLabeledObject( getDefaultSimulation(), l ),
	env( getDefaultSimulation() ),
	waitForAll( all )
	{
		// trace
		getTrace()->mark(this, markCreate);

		initObservedList(size, oP);
	}

Wait::Wait(Simulation* s, Label l, int size, Process* oP[], bool all) :
	DefLabeledObject(s, l),
	env(s),
	waitForAll(all)
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);

		initObservedList(size, oP);
	}

Wait::~Wait() {
	// trace
	getTrace()->mark(this, markDestroy);
}

void Wait::addProcess(Process* p) {
	if (p==0)
		return;

	observed.push_back(p);
}

void Wait::removeProcess(Process* p) {
	if (p==0)
		return;

	observed.remove(p);
}

const std::list< Process* >& Wait::getList() const
{
	return observed;
}

void Wait::setCondition(bool all/*=true*/) {
	waitForAll = all;
}

void Wait::onDestroy(Process* sender) {
	// Remove sender from observed list
	observed.remove(sender);

	// Kick condition check
	waitP->activate();
}

void Wait::onChangeProcessState(Process* sender, Process::ProcessState oldState, Process::ProcessState newState) {
	// Kick condition check
	if( newState == Process::TERMINATED )
		waitP->activate();
}

bool Wait::wait() {
	waitP=env->getCurrentProcess();

	list<Process*>::iterator i;

	// trace
	getTrace()->mark(this, markWait, waitP);

	// First check condition
	if (!checkObserved()) {
		// Add this as observer
		for (i=observed.begin(); i!=observed.end(); ++i)
			(*i)->Observable<ProcessObserver>::addObserver(this);
	}
	else {
		// trace
		getTrace()->mark(this, markContinue, waitP);

		return true;
	}

	// Wait for condition
	while (!checkObserved()) {
		waitP->sleep();

		// Handle interrupt
		if (waitP->isInterrupted()) {
			// Remove this as observer
			for (i=observed.begin(); i!=observed.end(); ++i)
				(*i)->Observable<ProcessObserver>::removeObserver(this);

			// trace
			getTrace()->mark(this, markContinue, waitP);

			return false;
		}
	}

	// Remove this as observer
	for (i=observed.begin(); i!=observed.end(); ++i)
		(*i)->Observable<ProcessObserver>::removeObserver(this);

	// trace
	getTrace()->mark(this, markContinue, waitP);

	return true;
}

void Wait::initObservedList(int size, Process* oP[]) {
	Process* p = 0;

	// Add entries from oP in observed
	for (int i=0; i<size; ++i) {
		p=oP[i];
		if (p!=0) addProcess(p);
	}
}

bool Wait::checkObserved() {
	// Check Condition: one(all) observed process TERMINATED
	list<Process*>::iterator i;

	if (waitForAll) {
		for (i=observed.begin(); i!=observed.end(); ++i)
			if ( (*i)->getProcessState()!=Process::TERMINATED ) return false;

		return true;
	} else {
		for (i=observed.begin(); i!=observed.end(); ++i)
			if ( (*i)->getProcessState()==Process::TERMINATED ) return true;

		return false;
	}
}

const MarkTypeId Wait::baseMarkId = MarkType::getNewBaseMarkId( typeid(Wait) );

const MarkType Wait::markCreate( "create", baseMarkId+1, typeid(Wait)); ///< Call of Wait constructor
const MarkType Wait::markDestroy( "destroy", baseMarkId+2, typeid(Wait)); ///< Call of Wait destructor
const MarkType Wait::markWait( "wait", baseMarkId+3, typeid(Wait)); ///< Caller Process is suspended until all observed processes have finished
const MarkType Wait::markContinue( "continue", baseMarkId+4, typeid(Wait)); ///< Observed Process have finished, caller Process can continue its execution

