//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file WaitQ.cpp

	\author Ralf Gerstenberger

	\date created at 2002/03/26

	\brief Implementation of classes in WaitQ.h

	\sa WaitQ.h

	<!-- [detailed description] -->

	\since 1.0
*/

#include <odemx/base/InSort.h>
#include <odemx/util/ErrorHandling.h>
#include <odemx/synchronization/WaitQ.h>

#include <string>
#include <cassert>

using namespace std;
using namespace odemx;

WaitQ::WaitQ( Label l, WaitQObserver* o/* = 0*/ ):
	Observable<WaitQObserver>(o),
	DefLabeledObject( getDefaultSimulation(), l ),
	StatisticObject( getDefaultSimulation() ),
	env( getDefaultSimulation() ),
	masterWait( l + " master queue" ),
	slaveWait( l + " slave queue" )
	{
		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEach(WaitQObserver, Create(this));

		// init statistics
		reset();
	}


WaitQ::WaitQ(Simulation* s, Label l, WaitQObserver* o/* = 0*/) :
	Observable<WaitQObserver>(o),
	DefLabeledObject(s, l),
	StatisticObject(s),
	env(s),
	masterWait(s, (l + " master queue")),
	slaveWait(s, (l + " slave queue"))
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEach(WaitQObserver, Create(this));

		// init statistics
		reset();
	}

WaitQ::~WaitQ() {
	// trace
	getTrace()->mark(this, markDestroy);

	// observer
	_obsForEach(WaitQObserver, Destroy(this));
}

bool WaitQ::wait() {

	// WaitQ synchronization only implemented for processes
	if( ! env -> getCurrentSched() ||
		env -> getCurrentSched() -> getSchedType() != Sched::PROCESS )
	{
		error( this,  "WaitQ::wait(): called by non-Process object" );
		return false;
	}

	Process* slave=getCurrentProcess();

	// trace
	getTrace()->mark(this, markWait, slave);

	// observer
	_obsForEach(WaitQObserver, Wait(this, slave));

	// awake masters
	awakeAll(&masterWait);

	// slave waits for master
	slaveWait.inSort(slave);

	// entrance time in queue is stored equal to execution time
	// sleep
	slave->sleep();

	if (slave->isInterrupted()) {

		slaveWait.remove(slave);
		return false;
	}

	return true;
}

bool WaitQ::wait( Weight weightFct ) {

	// WaitQ synchronization only implemented for processes
	if( ! env -> getCurrentSched() ||
		env -> getCurrentSched() -> getSchedType() != Sched::PROCESS )
	{
		error( this,  "WaitQ::wait(): called by non-Process object" );
		return false;
	}

	Process* slave=getCurrentProcess();

	// trace
	getTrace()->mark(this, markWaitWeight, slave);

	// observer
	_obsForEach(WaitQObserver, WaitWeight(this, slave));

	if( ! masterWait.isEmpty() ) {

		// evaluate master weight function for this slave
		std::list< Process* >::const_iterator i = masterWait.getList().begin();
		std::list< Process* >::const_iterator end = masterWait.getList().end();

		// initialize max weighted master variable
		Process* maxWeightMaster = *i;
		// get weight of slave in relation to first master
		double maxWeight = ( maxWeightMaster ->* weightFct )( slave );

		// iterate over master queue to get the master
		// where this slave's weight is highest
		for( ++i; i != end; ++i ) {

			// get the slave's weight for the current master
			double currentWeight = ( (*i) ->* weightFct )( slave );

			// if current master's slave weight is highest so far, remember it
			if( currentWeight > maxWeight ) {
				maxWeight = currentWeight;
				maxWeightMaster = *i;
			}
		}
		// awaken the master process with the highest weight evaluation
		maxWeightMaster -> activateAfter( slave );
	}

	// slave waits for master
	slaveWait.inSort(slave);

	// entrance time in queue is stored equal to execution time
	// sleep
	slave->sleep();

	if (slave->isInterrupted()) {

		slaveWait.remove(slave);
		return false;
	}

	return true;
}

Process* WaitQ::coopt( Weight weightFct ) {

	// WaitQ synchronization only implemented for processes
	if( ! env -> getCurrentSched() ||
		env -> getCurrentSched() -> getSchedType() != Sched::PROCESS )
	{
		error( this,  "WaitQ::coopt(): called by non-Process object" );
		return 0;
	}

	Process* master = getCurrentProcess();
	Process* slave = getWeightedSlave( master, weightFct );

	// no slave available
	if (slave==0) {

		// trace
		getTrace()->mark(this, markCooptWeightFail, master);

		// observer
		_obsForEach(WaitQObserver, CooptWeightFail(this, master));

		// master is waiting for slave
		masterWait.inSort( master );

		do {
			master -> sleep();

			if( master -> isInterrupted() ) {

				masterWait.remove( master );
				return 0;
			}

			slave = getWeightedSlave( master, weightFct );
		} while( slave == 0 );

		// block released
		masterWait.remove( master );
	}

	// remove chosen slave from queue
	slaveWait.remove( slave );

	// statistics
	updateStatistics( master, slave );

	// trace
	traceSUCCEED( markCooptWeightSucceed, master, slave );

	// observer
	_obsForEach( WaitQObserver, CooptWeightSucceed( this, master, slave ) )

	return slave;
}

Process* WaitQ::coopt(Selection sel/*=0*/) {

	// WaitQ synchronization only implemented for processes
	if( ! env -> getCurrentSched() ||
		env -> getCurrentSched() -> getSchedType() != Sched::PROCESS )
	{
		error( this,  "WaitQ::coopt(): called by non-Process object" );
		return 0;
	}

	return synch(true, sel);
}

Process* WaitQ::avail(Selection sel/*=0*/) {

	// WaitQ synchronization only implemented for processes
	if( ! env -> getCurrentSched() ||
		env -> getCurrentSched() -> getSchedType() != Sched::PROCESS )
	{
		error( this,  "WaitQ::avail(): called by non-Process object" );
		return 0;
	}

	return synch(false, sel);
}

Process* WaitQ::synch(bool wait, Selection sel) {
	Process* master=getCurrentProcess();
	Process* slave=getSlave(master, sel);

	if (slave==0) {
		// trace
		if (wait) getTrace()->mark(this, (sel ? markCooptSelFail : markCooptFail), master);
		else getTrace()->mark(this, (sel ? markAvailSelFail : markAvailFail) , master);

		// observer
		if (wait) {
			if (sel==0)	_obsForEach(WaitQObserver, CooptFail(this, master))
			else _obsForEach(WaitQObserver, CooptSelFail(this, master));
		} else {
			if (sel==0)	_obsForEach(WaitQObserver, AvailFail(this, master))
			else _obsForEach(WaitQObserver, AvailSelFail(this, master));
		}

		if (wait) {
			// master is waiting for slave
			masterWait.inSort(master);

			do {
				master->sleep();

				if (master->isInterrupted()) {
					masterWait.remove(master);

					return 0;
				}

				slave=getSlave(master, sel);
			} while (slave==0);

			// block released
			masterWait.remove(master);
		} else
			return 0;
	}

	// get slave
	slaveWait.remove(slave);

	// statistics
	updateStatistics(master, slave);

	// trace
	if (wait) traceSUCCEED( (sel ? markCooptSelSucceed : markCooptSucceed), master, slave);
	else traceSUCCEED( (sel ? markAvailSelSucceed : markAvailSucceed), master, slave);

	// observer
	if (wait) {
		if (sel==0)	_obsForEach(WaitQObserver, CooptSucceed(this, master, slave))
		else _obsForEach(WaitQObserver, CooptSelSucceed(this, master, slave));
	} else {
		if (sel==0)	_obsForEach(WaitQObserver, AvailSucceed(this, master))
		else _obsForEach(WaitQObserver, AvailSelSucceed(this, master));
	}

	return slave;
}

Process* WaitQ::getSlave(Process* master/*=0*/, Selection sel/*=0*/) {
	if (slaveWait.isEmpty() || (master==0))
		return 0;

	if (sel==0)
		return slaveWait.getTop();

	std::list<Process*>::const_iterator i;
	std::list<Process*>::const_iterator end = slaveWait.getList().end();

	for (i=slaveWait.getList().begin(); i!=end; ++i) {
		if ((master->*sel)(*i))
			return *i;
	}

	return 0;
}

Process* WaitQ::getWeightedSlave(Process* master/*=0*/, Weight weightFct ) {

	// nothing to synchronize
	if( slaveWait.isEmpty() || (master == 0) )
		return 0;

	// no weighting to evaluate, take first in queue
	if( weightFct == 0 )
		return slaveWait.getTop();

	// evaluate master weight function
	std::list< Process* >::const_iterator i = slaveWait.getList().begin();
	std::list< Process* >::const_iterator end = slaveWait.getList().end();

	// initialize max weighted slave variable
	Process* maxWeightSlave = *i;
	// get weight of first slave
	double maxWeight = ( master ->* weightFct )( maxWeightSlave );

	// iterate over slave queue to get the maximum weighted slave
	for( ++i; i != end; ++i ) {

		// get the current slave's weight
		double currentWeight = ( master ->* weightFct )( *i );

		// if current slave's weight is highest so far, remember it
		if( currentWeight > maxWeight ) {
			maxWeight = currentWeight;
			maxWeightSlave = *i;
		}
	}

	return maxWeightSlave;
}

void WaitQ::reset() {
	StatisticObject::reset();

	slaveWait.reset();
	masterWait.reset();

	synchs = zeroMasters = zeroSlaves = 0;
	sumMasterWait = sumSlaveWait = 0.0;
}

void WaitQ::report(Report* r) {

	assert (r!=0);
	masterWait.report(r);
	slaveWait.report(r);

	static vector< string > labels;			static vector< Column::Type  > types;
	labels.push_back("Name");				types.push_back( Column::STRING );
	labels.push_back("Reset at");			types.push_back( Column::SIMTIME );
	labels.push_back("Master Queue");		types.push_back( Column::STRING );
	labels.push_back("Slave Queue");		types.push_back( Column::STRING );
	labels.push_back("Number of Synchs");	types.push_back( Column::INTEGER );
	labels.push_back("Zero wait masters");	types.push_back( Column::INTEGER );
	labels.push_back("Avg master wait time");	types.push_back( Column::REAL );
	labels.push_back("Zero wait slaves");	types.push_back( Column::INTEGER );
	labels.push_back("Avg slave wait time");types.push_back( Column::REAL );

	static utTableDef def( labels.size(), labels, types );

	Table* t=r->createTable("Waitq Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< getResetTime()
	<< masterWait.getLabel()
	<< slaveWait.getLabel()
	<< getNumberOfSynch()
	<< getZeroWaitMasters()
	<< getAVMasterWaitTime()
	<< getZeroWaitSlaves()
	<< getAVSlaveWaitTime();
}

void WaitQ::updateStatistics(Process* master, Process* slave) {
	SimTime now = env->getTime();

	// master and slave synchronised (served)
	++synchs;

	// calculate waiting times
	SimTime masterWaitTime = master->getDequeueTime()-master->getEnqueueTime();
	SimTime slaveWaitTime = slave->getDequeueTime()-slave->getEnqueueTime();

	// update waiting time statistics
	if (masterWaitTime<=0.0 || master->getDequeueTime()!=now)
		++zeroMasters;
	else
		sumMasterWait += masterWaitTime;

	if (slaveWaitTime == 0.0)
		++zeroSlaves;
	else
		sumSlaveWait+=slaveWaitTime;
}

void WaitQ::traceSUCCEED(MarkType m, Process* master, Process* slave) {
	getTrace()->beginMark(this, m);
	getTrace()->addTag(tagMaster, master);
	getTrace()->addTag(tagSlave, slave);
	getTrace()->endMark();
}

const MarkTypeId WaitQ::baseMarkId = MarkType::getNewBaseMarkId( typeid(WaitQ) );

const MarkType WaitQ::markCreate( "create", baseMarkId+1, typeid(WaitQ)); ///< Call of WaitQ constructor
const MarkType WaitQ::markDestroy( "destroy", baseMarkId+2, typeid(WaitQ)); ///< Call of WaitQ destructor

const MarkType WaitQ::markWait( "wait", baseMarkId+3, typeid(WaitQ)); ///< Caller Process enqueued as slave, suspended and waiting to be activated by a master Process
const MarkType WaitQ::markWaitWeight( "weighted wait", baseMarkId+3, typeid(WaitQ)); ///< Caller Process enqueued as slave, but with consideration of a Weight function

const MarkType WaitQ::markCooptFail( "coopt fail", baseMarkId+4, typeid(WaitQ)); ///< Master Process could not synchronize with a slave, is suspended and enters master queue
const MarkType WaitQ::markCooptSucceed( "coopt succeed", baseMarkId+5, typeid(WaitQ)); ///< Master Process sucessfully received control over a slave Process
const MarkType WaitQ::markAvailFail( "avail fail", baseMarkId+6, typeid(WaitQ)); ///< Master Process requested a slave Process without success, but is not suspended
const MarkType WaitQ::markAvailSucceed( "avail succeed", baseMarkId+7, typeid(WaitQ)); ///< Master Process successfully received control over a slave Process

const MarkType WaitQ::markCooptSelFail( "coopt select fail", baseMarkId+8, typeid(WaitQ)); ///< Master Process could not synchronize with a slave matching the selection, must wait in master queue
const MarkType WaitQ::markCooptSelSucceed( "coopt select succeed", baseMarkId+9, typeid(WaitQ)); ///< Master Process sucessfully received control over a selected slave Process
const MarkType WaitQ::markAvailSelFail( "avail select fail", baseMarkId+10, typeid(WaitQ)); ///< Master Process could not find a slave matching the selection, but is not suspended
const MarkType WaitQ::markAvailSelSucceed( "avail select succeed", baseMarkId+11, typeid(WaitQ)); ///< Master Process successfully received control over a selected slave Process

const MarkType WaitQ::markCooptWeightFail( "coopt weight fail", baseMarkId+8, typeid(WaitQ)); ///< Master Process with weight function for slave choice could not synchronize with a slave, must wait in master queue
const MarkType WaitQ::markCooptWeightSucceed( "coopt weight succeed", baseMarkId+9, typeid(WaitQ)); ///< Master Process with weight function for slave choice sucessfully received control over the slave Process with the highest weight

const TagId WaitQ::baseTagId = WaitQ::baseMarkId;

const Tag WaitQ::tagMaster = Tag( "master", baseTagId+1); ///< Displays details about the master Process
const Tag WaitQ::tagSlave = Tag( "slave", baseTagId+2); ///< Displays details about the slave Process

