//----------------------------------------------------------------------------
//	Copyright (C) 2002-2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file BinT.cpp

	\author Ralf Gerstenberger

	\date created at 2002/04/22

	\brief Definition of odemx::BinT

	\sa BinT.h

	<!-- [detailed description] -->

	\since 2.1
*/

using namespace odemx;

template <typename Token>
BinT<Token>::BinT( Label l, BinTObserver<Token>* o/*=0*/ ):
	Observable<BinTObserver<Token> >(o),
	DefLabeledObject( getDefaultSimulation(), l ),
	StatisticObject( getDefaultSimulation() ),
	env( getDefaultSimulation() ),
	takeWait( l + " queue" ),
	initToken( 0 ),
	tokenNumber( 0 ),
	tokenStatistics( l + " token statistics" )
	{
		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEachT(BinTObserver<Token>, Create(this));

		// statistics
		reset();
	}

template <typename Token>
BinT<Token>::BinT( Simulation* s, Label l,
				   BinTObserver<Token>* o/*=0*/ ):
	Observable<BinTObserver<Token> >( o ),
	DefLabeledObject( s, l ),
	StatisticObject( s ),
	env( s ),
	takeWait( l + " queue" ),
	initToken( 0 ),
	tokenNumber( 0 ),
	tokenStatistics( l + " token statistics" )
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEachT(BinTObserver<Token>, Create(this));

		// statistics
		reset();
	}

template <typename Token>
BinT<Token>::BinT( Label l, std::vector<Token>& initialTokens, BinTObserver<Token>* o/*=0*/ ):
	Observable<BinTObserver<Token> >(o),
	DefLabeledObject( getDefaultSimulation(), l ),
	StatisticObject( getDefaultSimulation() ),
	env( getDefaultSimulation() ),
	takeWait( l + " queue" ),
	initToken( initialTokens.size() ),
	tokenNumber( initToken ),
	tokenStatistics( l + " token statistics" )
	{
		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEachT(BinTObserver<Token>, Create(this));

		// transfer token to tokenStore
		tokenStore.insert( tokenStore.end(), initialTokens.begin(), initialTokens.end() );

		// statistics
		reset();
	}

template <typename Token>
BinT<Token>::BinT( Simulation* s, Label l,
				   std::vector<Token>& initialTokens, BinTObserver<Token>* o/*=0*/ ):
	Observable<BinTObserver<Token> >( o ),
	DefLabeledObject( s, l ),
	StatisticObject( s ),
	env( s ),
	takeWait( l + " queue" ),
	initToken( initialTokens.size() ),
	tokenNumber( initToken ),
	tokenStatistics( l + " token statistics" )
	{
		assert(s!=0);

		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEachT(BinTObserver<Token>, Create(this));

		// transfer token to tokenStore
		tokenStore.insert( tokenStore.end(), initialTokens.begin(), initialTokens.end() );

		// statistics
		reset();
	}

template <typename Token>
BinT<Token>::~BinT() {
	// trace
	getTrace()->mark(this, markDestroy);

	// observer
	_obsForEachT(BinTObserver<Token>, Destroy(this));
}

template <typename Token>
std::vector<Token>* BinT<Token>::take( unsigned int n ) {

	// resource handling only implemented for processes
	if( !env -> getCurrentSched() ||
		env -> getCurrentSched() -> getSchedType() != Sched::PROCESS ) {

		error( this, "BinT::take(): called by non-Process object" );
		return 0;
	}

	// compute order of service
	takeWait.inSort(getCurrentProcess());

	if( tokenStore.size() < n ||
		takeWait.getTop() != getCurrentProcess() ) {
		// not enough tokens or
		// not first process

		// trace
		traceTSFG( markTakeFail, n );

		// observer
		_obsForEachT(BinTObserver<Token>, TakeFail(this, n));

		// statistics
		SimTime waitTime = env -> getTime();

		// block execution
		while( tokenStore.size() < n ||
			   takeWait.getTop() != getCurrentProcess() ) {

			getCurrentProcess() -> sleep();

			if( getCurrentProcess()->isInterrupted() ) {
				takeWait.remove( getCurrentProcess() );
				return 0;
			}
		}
		// block released
		// statistics
		sumWaitTime += env -> getTime() - waitTime;
	}

	// create return vector
	std::vector<Token>* tokenReturn =
		new std::vector<Token>( tokenStore.begin(), (tokenStore.begin()+n) );

	// trace
	getTrace()->mark(this, markChangeTokenNumber, tokenStore.size()-n, tokenStore.size());

	// observer
	_obsAForEachT(BinTObserver<Token>, TokenNumber, tokenStore.size(), tokenStore.size()-n);

	// remove token from store
	tokenStore.erase( tokenStore.begin(), (tokenStore.begin()+n) );
	tokenNumber -= n;

	// statistics
	++users;
	tokenStatistics.update(tokenNumber);

	// trace
	traceTSFG( markTakeSucceed, n );

	// observer
	_obsForEachT(BinTObserver<Token>, TakeSucceed(this, n));

	// remove process from list
	takeWait.remove( getCurrentProcess() );

	// awake next
	awakeFirst( &takeWait );

	// return token
	return tokenReturn;
}

template <typename Token>
void BinT<Token>::give( std::vector<Token>& t ) {

	if ( t.empty() ) {
		// Error: no token provided in give()
		error( "BinT::give(): no tokens provided in parameter t" );
		return;
	}

	unsigned int n = t.size();

	// trace
	getTrace() -> mark(this, markChangeTokenNumber, (tokenStore.size() + n), tokenStore.size());

	// observer
	_obsAForEachT(BinTObserver<Token>, TokenNumber, tokenStore.size(), (tokenStore.size() + n) );

	// trace
	traceTSFG( markGive, n );

	// observer
	_obsForEachT(BinTObserver<Token>, Give(this, n));

	// transfer token to tokenStore
	tokenStore.insert( tokenStore.end(), t.begin(), t.end() );
	// release token
	tokenNumber += t.size();

	// statistics
	++providers;
	tokenStatistics.update(tokenNumber);

	// wake up wating process objects
	awakeFirst( &takeWait );
}

template <typename Token>
void BinT<Token>::reset() {
	// reset time
	StatisticObject::reset();

	// reset queue statistics
	takeWait.reset();

	// Accum reset and re-init
	tokenStatistics.reset();
	tokenStatistics.update( getTokenNumber() );

	users=providers=0;
	sumWaitTime=0;
}

template <typename Token>
double BinT<Token>::getAVWaitTime() const {
	return sumWaitTime/getUsers();
}

template <typename Token>
void BinT<Token>::report(Report* r) {

	assert(r);
	takeWait.report(r);

	static std::vector< std::string > labels; static std::vector< Column::Type  > types;
	labels.push_back("Name");				types.push_back( Column::STRING);
	labels.push_back("Token type");			types.push_back( Column::STRING);
	labels.push_back("Reset at");			types.push_back( Column::SIMTIME);
	labels.push_back("Queue");				types.push_back( Column::STRING);
	labels.push_back("Users");				types.push_back( Column::INTEGER);
	labels.push_back("Providers");			types.push_back( Column::INTEGER);
	labels.push_back("Init token number");	types.push_back( Column::INTEGER);
	labels.push_back("Min token number");	types.push_back( Column::INTEGER);
	labels.push_back("Max token number");	types.push_back( Column::INTEGER);
	labels.push_back("Now token number");	types.push_back( Column::INTEGER);
	labels.push_back("Avg token number");	types.push_back( Column::REAL);
	labels.push_back("Avg waiting time");	types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t=r -> createTable( "BinT Statistics", &def );
	assert(t);

	*t
	<< getLabel()
	<< typeToString( typeid(Token) )
	<< getResetTime()
	<< takeWait.getLabel()
	<< getUsers()
	<< getProviders()
	<< getInitial()
	<< getMin()
	<< getMax()
	<< getTokenNumber()
	<< getAVFreeToken()
	<< getAVWaitTime();
}

template <typename Token>
void BinT<Token>::traceTSFG(MarkType m, unsigned int n) {
	getTrace()->beginMark(this, m);
	getTrace()->addTag(tagCurrent, getCurrentProcess());
	getTrace()->addTag(tagTokenNumber, n);
	getTrace()->endMark();
}

template <typename Token>
const MarkTypeId BinT<Token>::baseMarkId = MarkType::getNewBaseMarkId( typeid(BinT<Token>) );

template <typename Token>
const MarkType BinT<Token>::markCreate( "create", baseMarkId+1, typeid(BinT<Token>));///< Call of BinT constructor
template <typename Token>
const MarkType BinT<Token>::markDestroy( "destroy", baseMarkId+2, typeid(BinT<Token>));///< Call of BinT destructor
template <typename Token>
const MarkType BinT<Token>::markTakeFail( "take failed", baseMarkId+3, typeid(BinT<Token>));///< BinT does not contain enough tokens, caller Process is enqueued and suspended
template <typename Token>
const MarkType BinT<Token>::markTakeSucceed( "take succeeded", baseMarkId+4, typeid(BinT<Token>));///< BinT contains enough tokens, caller Process succeeded in acquiring the requested number
template <typename Token>
const MarkType BinT<Token>::markGive( "give", baseMarkId+5, typeid(BinT<Token>));///< BinT was supplied with the number of tokens shown
template <typename Token>
const MarkType BinT<Token>::markChangeTokenNumber( "change token number", baseMarkId+6, typeid(BinT<Token>));///< BinT token number changed, displays old and new values

template <typename Token>
const TagId BinT<Token>::baseTagId = BinT<Token>::baseMarkId;

template <typename Token>
const Tag BinT<Token>::tagCurrent = Tag( "current" , baseTagId+1); ///< Displays details about the currently active Process
template <typename Token>
const Tag BinT<Token>::tagTokenNumber = Tag( "token number", baseTagId+2); ///< Displays details about the BinT token number
