//----------------------------------------------------------------------------
//	Copyright (C) 2002-2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file BinT.h

	\author Ronald Kluth

	\date created at 2008/02/18

	\brief Declaration of odemx::BinT and observer

	\sa BinT.cpp

	<!-- [detailed description] -->

	\since 2.1
*/

#ifndef ODEMX_BIN_T_INCLUDED
#define ODEMX_BIN_T_INCLUDED

#include <odemx/base/Process.h>
#include <odemx/base/Simulation.h>
#include <odemx/synchronization/Queue.h>
#include <odemx/util/Observable.h>
#include <odemx/util/ErrorHandling.h>

#include <vector>
#include <string>
#include <deque>

namespace odemx {

	// forward declaration
	template <typename Token>
	class BinTObserver;

	/** \class BinT

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief BinT is a resoucre holding objects of type T.

		\note BinT supports Observation

		\note BinT supports Trace

		<!-- [\sa {references to other classes}]* -->

		BinT is a low bounded (token number >= 0) resource for synchronizing
		Process objects. A process is blocked in take()	if the requested number
		of tokens is greater than the available tokens. It is reactivated
		when enough tokens are given (back) to the resource. If multiple process
		objects are waiting for reactivation process-priority and FIFO-strategy
		are used to choose the process.	BinT is generally used for producer and
		consumer synchronization.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.1
	*/
	template <typename Token>
	class BinT: public Observable<BinTObserver<Token> >,
				public DefLabeledObject,
				public StatisticObject,
				public virtual TraceProducer,
				public virtual ReportProducer {
	public:

		/**
			\brief Construction for use with DefaultSimulation, no initial tokens
			\param l
				label of BinT object
			\param o
				observer of BinT object
		*/
		BinT( Label l, BinTObserver<Token>* o = 0 );

		/**
			\brief Construction for user-defined Simulation, no initial tokens
			\param s
				pointer to simulation
			\param l
				label of BinT object
			\param o
				observer of BinT object
		*/
		BinT( Simulation* s, Label l, BinTObserver<Token>* o = 0 );

		/**
			\brief Construction for use with DefaultSimulation
			\param l
				label of BinT object
			\param o
				observer of BinT object
		*/
		BinT( Label l, std::vector<Token>& initialTokens, BinTObserver<Token>* o = 0 );

		/**
			\brief Construction for user-defined Simulation
			\param s
				pointer to simulation
			\param l
				label of BinT object
			\param o
				observer of BinT object
		*/
		BinT( Simulation* s, Label l, std::vector<Token>& initialTokens, BinTObserver<Token>* o = 0 );

		/// Destruction
		virtual ~BinT();

		/**
			\name Token management

			@{
		*/
		/**
			\brief take \p n tokens

			Takes n tokens from BinT if possible. Otherwise, the
			current process is blocked until enough tokens become
			available.

			\note The returned vector is dynamically allocated with \c new,
			which means that the caller has to call delete on it.
		*/
		std::vector<Token>* take( unsigned int n );

		/**
			\brief give \p n token back

			Gives n tokens to BinT. Blocked process objects
			could be activated by this call.
		*/
		void give( std::vector<Token>& t );

		/// Number of tokens available
		unsigned int getTokenNumber() const { return tokenNumber; }

		/// Get a reference to the stored tokens
		const std::deque<Token>& getTokenStore() const { return tokenStore; }
		//@}

		/// get list of blocked  process objects
		const std::list<Process*>& getWaitingProcessList() const { return takeWait.getList(); }

		/**
			\name Statistics

			@{
		*/
		virtual void reset();

		/// Number of takes
		unsigned int getUsers() const {return users;}

		/// Number of gives
		unsigned int getProviders() const {return providers;}

		/// Initial number of token
		unsigned int getInitial() const {return initToken;}

		/// Min number of token
		unsigned int getMin() const {return (unsigned int)tokenStatistics.getMin();}

		/// Max number of token
		unsigned int getMax() const {return (unsigned int)tokenStatistics.getMax();}

		/// Average free token
		double getAVFreeToken() const {return tokenStatistics.getMean();}

		/// Average waiting time
		double getAVWaitTime() const;
		//@}

		/// get pointer to trace object
		virtual Trace* getTrace() const { return env; }

		/**
			\name Trace MarkTypes

			BinT specific trace marks.

			@{
		*/
		static const MarkTypeId baseMarkId;

		static const MarkType markCreate;
		static const MarkType markDestroy;

		static const MarkType markTakeFail;
		static const MarkType markTakeSucceed;
		static const MarkType markGive;

		static const MarkType markChangeTokenNumber;

		static const TagId baseTagId;

		static const Tag tagCurrent;
		static const Tag tagTokenNumber;
		//@}

		/// Generate report
		virtual void report(Report* r);

		// Implementation
	private:
		/// simulation context
		Simulation* env;
		/// token storage
		std::deque<Token> tokenStore;
		/// process management queue
		Queue takeWait;

		// statistics
		/// initialise token number
		unsigned int initToken;

		/// available number of token
		unsigned int tokenNumber;

		/// token statistics
		Accum tokenStatistics;

		/// user statistic
		unsigned int users;
		unsigned int providers;
		double sumWaitTime;

		// help methods
		/// current process
		Process* getCurrentProcess() { return env->getCurrentProcess(); }
		/// trace marks
		void traceTSFG( MarkType m, unsigned int n );
	};

	/** \interface BinTObserver

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief Observer for BinT specific events

		<!-- [\note {notes}]* -->

		\sa BinT

		<!-- [detailed description] -->

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		<!-- [\since {since when in project}] -->
	*/
	template < typename Token >
	class BinTObserver {
	public:
		virtual ~BinTObserver() {}

		/// Creation
		virtual void onCreate( BinT<Token>* sender ) {}
		/// Destruction
		virtual void onDestroy( BinT<Token>* sender ) {}

		/// Failed attempt to take n token
		virtual void onTakeFail( BinT<Token>* sender, unsigned int n ) {}
		/// Successfull attempt to take n token
		virtual void onTakeSucceed( BinT<Token>* sender, unsigned int n ) {}
		/// Return of n token
		virtual void onGive( BinT<Token>* sender, unsigned int n ) {}

		/// Change of token number
		virtual void onChangeTokenNumber( BinT<Token>* sender, unsigned int oldTokenNumber,
			unsigned int newTokenNumber ) {}
	};

}

#endif /* ODEMX_BIN_T_INCLUDED */

