using namespace odemx;
using namespace std;

/********************* PortHead Template Implementation ***********************/

template < typename ElementType >
PortHeadT< ElementType >::PortHeadT( Label portName,
									 PortMode m /* = WAITING_MODE */,
									 unsigned int max /* = 10000 */,
									 PortHeadTObserver< ElementType >* pho /* = 0 */ ):
	Memory( portName + " (head)", Memory::PORTHEAD ),
	DefLabeledObject( getDefaultSimulation(), portName + " (head)" ),
	Observable< PortHeadTObserver< ElementType > >( pho ),
	mode( m ),
	env( getDefaultSimulation() )
	{
		// trace
		getTrace() -> mark( this, markCreate );

		// observer
		_obsForEachT(PortHeadTObserver< ElementType >, Create(this));

		// create internal port for this head but w/o tail
		if ( max != 0 )
		{
			internalPort = new PortT< ElementType >( env, portName, max, this, 0 );
		}
		else
		{
			error( this, "PortHeadT<>(): attempt to create Port with capacity 0, using default" );
			internalPort = new PortT< ElementType >( env, portName, 10000, this, 0 );
		}
	}


template < typename ElementType >
PortHeadT< ElementType >::PortHeadT( Simulation* sim,
									 Label portName,
									 PortMode m,
									 unsigned int max,
									 PortHeadTObserver< ElementType >* pho ):
	Memory( sim, portName + " (head)", Memory::PORTHEAD ),
	DefLabeledObject( sim, (portName + " (head)") ),
	Observable< PortHeadTObserver< ElementType > >( pho ),
	mode( m ),
	env( sim )
	{
		// trace
		getTrace() -> mark( this, markCreate );

		// observer
		_obsForEachT(PortHeadTObserver< ElementType >, Create(this));

		// create internal port for this head but w/o tail
		if ( max != 0 )
		{
			internalPort = new PortT< ElementType >( env, portName, max, this, 0 );
		}
		else
		{
			error( this, "PortHeadT<>(): attempt to create Port with capacity 0, using default" );
			internalPort = new PortT< ElementType >( env, portName, 10000, this, 0 );
		}
	}

template < typename ElementType >
PortHeadT< ElementType >::~PortHeadT()
	{
		// trace
		getTrace() -> mark( this, markDestroy );

		// observer
		_obsForEachT(PortHeadTObserver< ElementType >, Destroy(this));

		// delete the PortHead and also the Port
		// if it is not also pointed to by a PortTail
		if ( internalPort && internalPort -> portTail )
			internalPort -> portHead = 0;
		else
			delete internalPort;
	}

template < typename ElementType >
std::auto_ptr< ElementType > PortHeadT< ElementType >::get() {

	// trace
	getTrace() -> mark( this, markGet, getPartner() );

	// observer
	_obsForEachT(PortHeadTObserver< ElementType >, Get(this));

	while ( internalPort -> elementList.empty() ) {

		Sched* caller = env -> getCurrentSched();

		switch ( mode ) {
		case WAITING_MODE:

			if ( caller == 0 )
			{
				error( this, "PortHeadT::get(): port in waiting mode called from environment" );
				return std::auto_ptr< ElementType >(0);
			}

			if ( caller -> getSchedType() == Sched::PROCESS ) {

				Memory::remember( caller );
				dynamic_cast<Process*>(caller) -> sleep();

				// check for interruption of waiting period
				if( dynamic_cast<Process*>(caller) -> isInterrupted() ) {

					Memory::forget( caller );
					return std::auto_ptr< ElementType >(0);
				}
			}
			else if ( caller -> getSchedType() == Sched::EVENT ) {
				// error: cannot send an event to sleep
				error( this,  "PortHeadT<>::get(): waiting mode not allowed for events" );
				return std::auto_ptr< ElementType >(0);
			}
			break;

		case ERROR_MODE:
			error( this,  "PortHeadT<>::get(): called on empty port" );
			return std::auto_ptr< ElementType >(0);

		case ZERO_MODE:
			return std::auto_ptr< ElementType >(0);
		}
	}

	// the Port was filled to the limit
	bool wasFull = ( internalPort -> elementList.size() == internalPort -> maxCapacity );

	// remove first element from Port
	std::auto_ptr< ElementType > first( new ElementType( internalPort -> elementList.front() ) );
	internalPort -> elementList.pop_front();

	// update stats
	internalPort -> getCalls++;
	internalPort -> portAccum.update( internalPort -> getNowCount() );

	// alert waiting sender
	if ( wasFull && internalPort -> portTail && internalPort -> portTail -> waiting() )
		internalPort -> portTail -> alert();

	return first;
}



// insert PortData at the beginning of the internal Port queue
template < typename ElementType >
bool PortHeadT< ElementType >::unGet( ElementType newElement ) {

	// trace
	getTrace() -> mark( this, markUnGet, getPartner() );

	// observer
	_obsForEachT(PortHeadTObserver< ElementType >, UnGet(this, newElement));

	if ( internalPort -> elementList.size() >= internalPort -> maxCapacity )
	{
		switch ( mode )
		{
		case WAITING_MODE:
		case ERROR_MODE:
			error( this,  "PortHeadT<>::unGet(): called on full port" );
			/* FALL TROUGH */
		case ZERO_MODE:	return false;
		}
	}

	// insert PortData at the front of the Port queue
	internalPort -> elementList.push_front( newElement );

	// update stats
	internalPort -> ungetCalls++;
	internalPort -> portAccum.update( internalPort -> getNowCount() );

	return true;
}


template < typename ElementType >
PortTailT< ElementType >* PortHeadT< ElementType >::getTail() {

	PortTailT< ElementType >* tail = internalPort -> portTail;

	// tail still zero?
	if ( !tail ) {

		// create new Tail
		tail = new PortTailT< ElementType >( env, internalPort -> getLabel(),
				mode, internalPort -> maxCapacity );

		// register new Tail with Head's Port
		internalPort -> portTail = tail;

		// register Head's internal Port with the Tail
		tail -> internalPort = this -> internalPort;
	}

	return tail;
}


template < typename ElementType >
const std::list< ElementType >& PortHeadT< ElementType >::getList() const {
	return internalPort -> elementList;
}


// change
//      old_Port_head -- old_Port -- old_Port_tail
// into
//      old_Port_head -- new_Port (___get new tail___)
//      new_Port_head -- old_Port -- old_Port_tail
// where
//      old_Port_head == this
// return
//      new_Port_head.
template < typename ElementType >
PortHeadT< ElementType >* PortHeadT< ElementType >::cut() {

	// trace
	getTrace() -> mark( this, markCut, getPartner() );

	// observer
	_obsForEachT(PortHeadTObserver< ElementType >, Cut(this));

	// remember old Port
	PortT< ElementType >* oldPort = internalPort;

	// create a new Head and thereby a new Port
	PortHeadT< ElementType >* newHead =
		new PortHeadT< ElementType >( env, ("Cut " + internalPort -> getLabel()),
									  mode, oldPort -> maxCapacity );

	// get the new Port
	PortT< ElementType >* newPort = newHead -> internalPort;

	// put the new head on the old Port
	oldPort -> portHead  = newHead;
	newHead -> internalPort = oldPort;

	// put the old head (== this) on the new Port
	internalPort = newPort;
	newPort -> portHead = this;

	return newHead;
}


// this PortHead is supposed to be upstream to the portTail oldTail:
// >> headsTail -- headsPort -- thisHead >> oldTail -- tailsPort -- tailsHead >>
//   add the contents of headsPort to tailsPort
//   destroy thisHead and oldTail
//   alert the spliced PortHead if a significant state change happened
template < typename ElementType >
void PortHeadT< ElementType >::splice( PortTailT< ElementType >* oldTail,
		bool deleteHead /*=true*/, bool deleteOldTail /*=true*/) {

	// trace
	getTrace() -> mark( this, markSplice, getPartner() );

	// observer
	_obsForEachT(PortHeadTObserver< ElementType >, Splice(this, oldTail));

	// remember ports
	PortT< ElementType >* tailsPort   = oldTail -> internalPort;
	PortT< ElementType >* headsPort   = this -> internalPort;

	// and their element counts
	int tailsCount = tailsPort -> elementList.size();
	int headsCount = headsPort -> elementList.size();

	// append head's buffer to tail's buffer via std::list::splice()
	if ( headsCount > 0 ) {
		tailsPort -> elementList.splice( tailsPort -> elementList.end(), headsPort -> elementList );
	}

	// patch the head's tail onto tail's Port, which has it's own head,
	// this PortHead becomes superfluous, requires pointers in both directions!
	tailsPort -> portTail					= headsPort -> portTail;
	headsPort -> portTail -> internalPort	= tailsPort;
	// this should not point anywhere anymore, as the object belongs to the tail's port now
	headsPort -> portTail = 0;

	// delete the throwaways
	oldTail -> internalPort = 0; // so the tail port won't be deleted in destructor
	if( deleteOldTail ) delete oldTail;
	if( deleteHead ) delete this;

	// alert the new head if its Port becomes non-empty
	if ( (tailsCount == 0) && (headsCount > 0) && (tailsPort -> portHead) )
	{
		if ( tailsPort -> portHead -> waiting() )
		{
			tailsPort -> portHead -> alert();
		}
	}
}


template < typename ElementType >
void PortHeadT< ElementType >::cppSplice( PortTailT< ElementType >* tail,
										  bool append /*= true*/ )
{
	// trace
	getTrace() -> mark( this, markCppSplice, getPartner() );

	// observer
	_obsForEachT(PortHeadTObserver< ElementType >, CppSplice(this, tail, append));

	list< ElementType >& headList = internalPort -> elementList;
	list< ElementType >& tailList = tail -> internalPort -> elementList;

	if( append )
		headList.splice( headList.end(), tailList );
	else
		headList.splice( headList.begin(), tailList );
}


template < typename ElementType >
unsigned int PortHeadT< ElementType >::getMaxCapacity() {
      return internalPort -> maxCapacity;
}


template < typename ElementType >
void PortHeadT< ElementType >::setMaxCapacity( unsigned int newMaximum ) {

	// trace
	getTrace() -> mark( this, markChangeMaxCapacity, newMaximum, internalPort -> maxCapacity );

	// observer
	_obsAForEachT(PortHeadTObserver< ElementType >, MaxCapacity, internalPort -> maxCapacity, newMaximum);

	internalPort -> maxCapacity = newMaximum;
}


template < typename ElementType >
PortMode PortHeadT< ElementType >::getMode() {
	return mode;
}


template < typename ElementType >
void PortHeadT< ElementType >::setMode( PortMode newMode ) {

	// trace
	getTrace() -> mark( this, markChangeMode, newMode, mode );

	// observer
	_obsAForEachT(PortHeadTObserver< ElementType >, Mode, mode, newMode);

	mode = newMode;
}


// return current number of PortElements
template < typename ElementType >
unsigned int PortHeadT< ElementType >::count() {
	return internalPort -> elementList.size();
}


// does PortHead contain a PortData?
template < typename ElementType >
bool PortHeadT< ElementType >::available () {
	return (count() > 0);
}

// simpler implementation than in Memory because get() does
// some checking already and only the first waiting Process is alerted
template < typename ElementType >
void PortHeadT< ElementType >::alert()
{
	if( ! Memory::waiting() )
	{
		// warning: no object to wake
		if ( mode != ZERO_MODE ){
			warning( this, "PortHeadT<>::alert(): no process saved" );
		}
		return;
	}

	// get() allows WAITING_MODE only for Processes, so cast is okay here
	Process* rememberedObject = static_cast< Process* >( Memory::getList().front() );

	// trace
	getTrace() -> mark( this, markAlert, rememberedObject );

	// observer
	_obsForEachT( PortHeadTObserver< ElementType >, Alert( this, rememberedObject ) );

	// wake up the process
	rememberedObject -> alertProcess( this );

	// update stats
	internalPort -> headAlerts++;

	Memory::forget( rememberedObject );
}


template < typename ElementType >
TraceProducer* PortHeadT< ElementType >::getPartner() {

	// return the current process or event, if one is active
	if ( env -> getCurrentSched() != 0 ) {

		return env -> getCurrentSched();
	}
	else { // or return the simulation

		return env;
	}
}

template < typename ElementType >
const MarkTypeId PortHeadT< ElementType >::baseMarkId = MarkType::getNewBaseMarkId( typeid(PortHeadT< ElementType >) );

template < typename ElementType >
const MarkType PortHeadT< ElementType >::markCreate( "create", baseMarkId+1, typeid(PortHeadT< ElementType >) ); ///< Call of PortHeadT constructor
template < typename ElementType >
const MarkType PortHeadT< ElementType >::markDestroy( "destroy", baseMarkId+2, typeid(PortHeadT< ElementType >) ); ///< Call of PortHeadT destructor

template < typename ElementType >
const MarkType PortHeadT< ElementType >::markGet( "get", baseMarkId+3, typeid(PortHeadT< ElementType >) ); ///< Data request, may lead to suspension of the calling Process, depending on PortMode
template < typename ElementType >
const MarkType PortHeadT< ElementType >::markUnGet( "unGet", baseMarkId+4, typeid(PortHeadT< ElementType >) ); ///< Data was put back to the front of the PortT
template < typename ElementType >
const MarkType PortHeadT< ElementType >::markCut( "cut", baseMarkId+5, typeid(PortHeadT< ElementType >) ); ///< PortHeadT was split into two ports, making the old PortHeadT the head of the new port
template < typename ElementType >
const MarkType PortHeadT< ElementType >::markSplice( "splice", baseMarkId+6, typeid(PortHeadT< ElementType >) ); ///< Two ports are joined into one, appending the contents of the upstream port to the downstream one
template < typename ElementType >
const MarkType PortHeadT< ElementType >::markCppSplice( "cpp splice", baseMarkId+7, typeid(PortHeadT< ElementType >) ); ///< The contents of another port is copied to this port in the same way std::list::splice() does
template < typename ElementType >
const MarkType PortHeadT< ElementType >::markAlert( "alert", baseMarkId+8, typeid(PortHeadT< ElementType >) ); ///< PortHeadT alerts the first Process waiting for Data, triggered by PortTailT::put()

template < typename ElementType >
const MarkType PortHeadT< ElementType >::markChangeMode( "change mode", baseMarkId+9, typeid(PortHeadT< ElementType >) ); ///< Change of the PortMode which controls the get-when-empty behavior of a PortHeadT
template < typename ElementType >
const MarkType PortHeadT< ElementType >::markChangeMaxCapacity( "change maximum capacity", baseMarkId+10, typeid(PortHeadT< ElementType >) ); ///< Change of a PortHeadT's internal PortT capacity

