//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file PortHeadT.h

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Declaration of odemx::PortT, odemx::PortHeadT and odemx::PortTailT

	\sa PortT.cpp, PortHeadT.cpp, PortTailT.cpp

	<!-- [detailed description] -->

	\since 2.0
*/

#ifndef PORT_HEAD_T_INCLUDED
#define PORT_HEAD_T_INCLUDED

#include <memory> // auto_ptr

namespace odemx {

	// forward declaration
	template < typename ElementType >
	class PortT;

	template < typename ElementType >
	class PortTailT;

	template < typename ElementType >
	class PortHeadTObserver;

	/** \class PortHeadT

		\ingroup synch

		\author Ronald Kluth

		\brief %PortHeadT provides read access to a port

		\sa PortTailT, PortT

		This class manages read access to a limited internal buffer.
		It can be run in three different error modes according to
		PortMode. Get on an empty port will either be handled by
		making the reader wait (and alerting him upon availability),
		reporting an error, or returning zero.

		\since 2.0
	*/

	template < typename ElementType >
	class PortHeadT: public Memory,	// is a trace and report producer
					 public DefLabeledObject,
					 public Observable< PortHeadTObserver< ElementType > > {
	public:

		/**
			\brief Construction for DefaultSimulation
			\param portName
				label of the internal port, this object's label will be "portName (head)"
			\param m
				port head's mode, default is WAITING_MODE
			\param max
				maximum number of elements
			\param pho
				initial observer
		*/
		PortHeadT( Label portName,
				   PortMode m = WAITING_MODE,
				   unsigned int max = 10000,
				   PortHeadTObserver< ElementType >* pho = 0 );

		/**
			\brief Construction for user-defined Simulation
			\param sim
				pointer to the simulation object
			\param portName
				label of the internal port, this object's label will be "portName (head)"
			\param m
				port head's mode, default is WAITING_MODE
			\param max
				maximum number of elements
			\param pho
				initial observer
		*/
		PortHeadT( Simulation* sim, Label portName,
				   PortMode m = WAITING_MODE,
				   unsigned int max = 10000,
				   PortHeadTObserver< ElementType >* pho = 0 );

		/// Destruction
		virtual ~PortHeadT ();

		/**
			\brief Port read access
			\return
				the next port element from the internal buffer

			This function is used to read information from a port. It also
			responds according to the object's PortMode. In WAIT_MODE this
			leads to the suspension if the calling Process. If the caller
			is then interrupted before receiving an element, it is
			reactivated and the function returns a null pointer. In ZERO_MODE
			a miss only returns a null pointer, while in ERROR_MODE an error
			will be sent additionally.

			\sa unGet()

			\note The returned object is always dynamically allocated, which
			is why an \c auto_ptr is used. This way, the caller never needs
			to delete the return value. Checking for null pointer is done in
			the following manner:
			\code

			std::auto_ptr< int > ptr = port.get();
			if( ptr.get() == 0 )
			{
				// handle null return value
			}
			else
			{
				// use ptr like any other pointer
			}
			\endcode
		*/
		std::auto_ptr< ElementType > get();        // provides the first PortData

		/**
			\brief Port access, put element back at front
			\param newElement
				the element to put back into the buffer

			\return
				\p true if the port element was successfully added at the front

			This function is used to put an element back at the front
			of the internal port.

			\sa get()
		*/
		bool unGet( ElementType newElement );

		/**
			\brief Get corresponding PortTail
			\return
				corresponding PortTail to this PortHead

			This function provides access to the tail of a port head.
			If it does not exist yet, it is created.
		*/
		PortTailT< ElementType >* getTail();

		/**
			\brief Get a reference to the internal list
			\return
				reference to the buffer's internal list

			This function provides access to the internal buffer list
			of the head's PortT.
		*/
		const std::list< ElementType >& getList() const;

		/**
			\brief Get maximum capacity of the internal buffer
			\return
				capacity of the port

			This function is used to determine the maximum number of
			elements that can be buffered by the internal port.

			\sa setMaxCapacity()
		*/
		unsigned int getMaxCapacity();

		/**
			\brief Change maximum capacity of the port
			\param newMaximum
				the new maximum capacity of the internal port

			This function allows to change the internal buffer limit.

			\sa getMaxCapacity()
		*/
		void setMaxCapacity( unsigned int newMaximum );

		/**
			\brief Get the port head's mode
			\return
				the currently set mode

			This function is used to determine the port head's mode.

			\sa setMode()
		*/
		PortMode getMode();

		/**
			\brief Change the port head's mode
			\param newMode
				new mode to be set

			This function is used to change the port head's mode.

			\sa getMode()
		*/
		void setMode( PortMode newMode );

		/**
			\brief Get current number of port elements
			\return
				current number of elements

			This function returns the current number of buffered elements.
		*/
		unsigned int count();

		/**
			\brief Check for availability of an element
			\return
				\p true if the port is not empty

			This function is used to determine whether an element is
			available from the internal port.
		*/
		virtual bool available ();

		/**
			\brief Cut this PortHead from its tail
			\return
				new port head

			This function will create a	new internal port, which
			is attached to this head. A new tail for that port must then be
			requested via getTail(). Furthermore, a new port head will
			be attached to the old internal port. That new head will be
			returned.

			This provides the ability to insert a process into
			the middle of a port connection, for example as a filter.

			\sa splice()
		*/
		PortHeadT< ElementType >* cut();

		/**
			\brief Splice two separate ports together
			\param oldTail
				the tail of the port to which this head's buffer will be attached
			\param deleteHead
				specifies whether this function should delete this head object
			\param deleteOldTail
				specifies whether this function should delete the oldTail object

			This function is the reverse operation of cut(). It is used to join
			two ports together. This head must be upstream of, i.e. before,
			oldTail's port, which this buffer will be joined with.
			The contents of oldTail's Port is added to this Port. If
			the new head's buffer becomes non-empty, it is alerted.

			\warning Some objects become superflous with this operation!
			This head, oldTail, and oldTail's Port should be dynamically
			allocated, in which case they can be deleted by this function.

			\note This function does not check the maximum Port capacity before
			splicing lists.

			\sa cut()
		*/
		void splice( PortTailT< ElementType >* oldTail, bool deleteHead = true,
					 bool deleteOldTail = true );

		/**
			\brief Splice internal lists of two separate ports together
			\param tail
				the tail of the port whose buffer will be added to this
			\param append
				determines whether to append or prepend tail's contents

			This function is used to join two ports' buffer contents together
			in the same way as std::list::splice(). The contents of the \p
			tail PortT will be moved to \p this port, where parameter \p append
			can	be used to specify whether the data will be appended or prepended
			to this port. Default is append.

			\note This function does not check the maximum Port capacity before
			splicing lists.
		*/
		void cppSplice( PortTailT< ElementType >* tail, bool append = true );

		/**
			\brief Report producing capability
			\param r
				Report object that receives accumulated data

			This function is called by a Report object during report
			generation to provide its data.
		*/
		virtual void report(Report* r) {
			internalPort -> report( r );
			Memory::report( r );
		}

		/// Reset object statistics to \c 0
		virtual void reset()
		{
			Memory::reset();
			internalPort -> reset();
		}

	private:

		friend class PortTailT< ElementType >;

		/**
			\brief Alert the first stored process

			This function is designed to wake up suspended processes,
			i.e. reschedule them at simulation time \p now when they
			are waiting for an empty port to receive an element. Only
			the first remembered process is alerted by the %PortHead.
		*/
		virtual void alert();

		PortT< ElementType >* internalPort;	///< internal buffer for port elements
		PortMode mode;						///< mode for read requests
		Simulation* env;					///< simulation context

		// ========================= Trace Marks ===============================

		/**
			\brief Get currently running Sched object or simulation
			\return
				pointer to simulation or current Sched object
		*/
		TraceProducer* getPartner();

	public:
		/**
			\name Trace MarkTypes

			These MarkTypes and Tags are used to trace PortHead calls.
			A TraceConsumer can use these constants to identify trace
			events sent	by PortHead.

			@{
		*/
		static const MarkTypeId baseMarkId;

		static const MarkType markCreate;
		static const MarkType markDestroy;

		static const MarkType markGet;
		static const MarkType markUnGet;
		static const MarkType markCut;
		static const MarkType markSplice;
		static const MarkType markCppSplice;
		static const MarkType markAlert;

		static const MarkType markChangeMode;
		static const MarkType markChangeMaxCapacity;
		//@}

	};

	/** \interface PortHeadTObserver

		\author Ronald Kluth

		\brief Observer for PortHeadT-specific calls.

		\sa PortHeadT

		<!-- [detailed description] -->

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.0
	*/

	template < typename ElementType >
	class PortHeadTObserver: public MemoryObserver {
	public:
		virtual ~PortHeadTObserver() {}

		virtual void onCreate(PortHeadT< ElementType >* sender) {} ///< Construction
		virtual void onDestroy(PortHeadT< ElementType >* sender) {} ///< Destruction

		virtual void onGet(PortHeadT< ElementType >* sender) {}
		virtual void onUnGet(PortHeadT< ElementType >* sender, ElementType newElement) {}
		virtual void onCut(PortHeadT< ElementType >* sender) {}
		virtual void onSplice(PortHeadT< ElementType >* sender, PortTailT< ElementType >* oldTail) {}
		virtual void onCppSplice(PortHeadT< ElementType >* sender, PortTailT< ElementType >* oldTail, bool append) {}
		virtual void onAlert(PortHeadT< ElementType >* sender, Sched* alerted) {}

		/// PortHead mode change
		virtual void onChangeMode(PortHeadT< ElementType >* sender, PortMode oldMode, PortMode newMode) {}
		/// PortHead capacity change
		virtual void onChangeMaxCapacity(PortHeadT< ElementType >* sender, unsigned int oldMax, unsigned int newMax) {}
	};
}

#endif /* PORT_HEAD_T_INCLUDED */
