//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file PortT.h

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Declaration of odemx::PortT, odemx::PortHeadT and odemx::PortTailT

	\sa PortT.cpp, PortHeadT.cpp, PortTailT.cpp

	<!-- [detailed description] -->

	\since 2.0
*/

#include <list>
#include <string>
#include <odemx/base/Simulation.h>
#include <odemx/base/Process.h>
#include <odemx/synchronization/Memory.h>
#include <odemx/util/ErrorHandling.h>
#include <odemx/util/Report.h>
#include <odemx/util/LabeledObject.h>
#include <odemx/statistics/Statistics.h>

namespace odemx {

	// forward declaration
	template < typename ElementType >
	class PortHeadT;
	template < typename ElementType >
	class PortTailT;

	template < typename ElementType >
	class PortHeadTObserver;
	template < typename ElementType >
	class PortTailTObserver;

	/**
		\brief Port modes

		Since ports implement a limited buffer, a behavior needs to be
		defined when a port is full, but an element is to be inserted,
		or when a port is empty and some process is trying to read from it.
		Modes for port head and corresponding port tail can differ.
	*/
	enum PortMode {
		ERROR_MODE,		///< error when full/empty
		WAITING_MODE,	///< context switch when full/empty
		ZERO_MODE		///< ignore objects when full/empty, return 0
	};

	/** \class PortT

		\ingroup synch

		\author Ronald Kluth

		\brief %Port is an internal class that implements the internal buffer.

		\sa PortHeadT, PortTailT

		This internal class manages the limited buffer that is the base
		for a communications port. It maintains references to its port tail,
		which describes the input interface, and its port head, which is the
		output interface of a port.

		\note In order to use and access ports, you should create either a
		PortHeadT or PortTailT object, which provide the reading and writing
		interface.

		\since 2.0
	*/
	template < typename ElementType >
	class PortT: public DefLabeledObject,
				 public StatisticObject {
	// interface
	public:

		friend class PortHeadT< ElementType >;
		friend class PortTailT< ElementType >;

		/// Construction for DefaultSimulation
		PortT( Label l, unsigned int max,
			   PortHeadT< ElementType >* ph = 0,
			   PortTailT< ElementType >* pt = 0 );

		/// Construction for user-defined Simulation
		PortT( Simulation* sim, Label l, unsigned int max,
			   PortHeadT< ElementType >* ph = 0,
			   PortTailT< ElementType >* pt = 0 );

		/// Destruction
		virtual ~PortT();


		/**
			\brief report port statistics
			\param r
				the report object which receives the data

			This function is used to send information collected from a port
			to a report object which displays the data.
		*/
		void report( Report* r );

		/// reset statistics
		virtual void reset();

		/// min buffer length
		unsigned int getMinCount() const {
			return (unsigned int)portAccum.getMin();
		}

		/// max buffer length
		unsigned int getMaxCount() const {
			return (unsigned int)portAccum.getMax();
		}

		/// current buffer length
		unsigned int getNowCount() const {
			return elementList.size();
		}

		/// average buffer length
		double getAvgCount() const {
			return portAccum.getMean();
		}

	private:

		std::list< ElementType > elementList;	///< internal buffer for PortElements
		unsigned int maxCapacity;				///< Port capacity
		PortHeadT< ElementType >* portHead;		///< corresponding Head interface
		PortTailT< ElementType >* portTail;		///< corresponding Tail interface
		Simulation*	env;						///< Simulation context
		std::string elementTypeName;			///< string representation of element type

		// statistics
		Accum portAccum;
		unsigned int putCalls;
		unsigned int getCalls;
		unsigned int ungetCalls;
		unsigned int headAlerts;
		unsigned int tailAlerts;

	};
}

