using namespace odemx;
using namespace std;

/********************* PortTail Template Implementation ***********************/

template < typename ElementType >
PortTailT< ElementType >::PortTailT( Label portName,
									 PortMode m,
									 unsigned int max,
									 PortTailTObserver< ElementType >* pto ):
	Memory( portName + " (tail)", Memory::PORTTAIL ),
	DefLabeledObject( getDefaultSimulation(), portName + " (tail)" ),
	Observable< PortTailTObserver< ElementType > >( pto ),
	mode( m ),
	env( getDefaultSimulation() )
	{
		// trace
		getTrace() -> mark( this, markCreate );

		// observer
		_obsForEachT(PortTailTObserver< ElementType >, Create(this));

		// create internal port for this tail but w/o head
		if ( max != 0 )
		{
			internalPort = new PortT< ElementType >( env, portName, max, 0, this );
		}
		else
		{
			error( this, "PortTailT<>(): attempt to create Port with capacity 0, using default" );
			internalPort = new PortT< ElementType >( env, portName, 10000, 0, this );
		}
	}

template < typename ElementType >
PortTailT< ElementType >::PortTailT( Simulation* sim,
									 Label portName,
									 PortMode m,
									 unsigned int max,
									 PortTailTObserver< ElementType >* pto ):
	Memory( sim, portName + " (tail)", Memory::PORTTAIL ),
	DefLabeledObject( sim, portName + " (tail)" ),
	Observable< PortTailTObserver< ElementType > >( pto ),
	mode( m ),
	env( sim )
	{
		assert( sim );

		// trace
		getTrace() -> mark( this, markCreate );

		// observer
		_obsForEachT(PortTailTObserver< ElementType >, Create(this));

		// create internal port for this tail but w/o head
		if ( max != 0 )
		{
			internalPort = new PortT< ElementType >( sim, portName, max, 0, this );
		}
		else
		{
			error( this, "PortTailT<>(): attempt to create Port with capacity 0, using default" );
			internalPort = new PortT< ElementType >( sim, portName, 10000, 0, this );
		}
	}

template < typename ElementType >
PortTailT< ElementType >::~PortTailT()
	{
		// trace
		getTrace() -> mark( this, markDestroy );

		// observer
		_obsForEachT(PortTailTObserver< ElementType >, Destroy(this));

		// delete the PortTail and also the Port
		// if it is not also pointed to by a PortHead
		if ( internalPort && internalPort -> portHead )
			internalPort -> portTail = 0;
		else
			delete internalPort;
	}


template < typename ElementType >
int PortTailT< ElementType >::put( ElementType newElement ) {

	// trace
	getTrace() -> mark( this, markPut, getPartner() );

	// observer
	_obsForEachT(PortTailTObserver< ElementType >, Put(this, newElement));

	while ( internalPort -> elementList.size() >= internalPort -> maxCapacity ) {

		Sched* caller = env -> getCurrentSched();

		switch ( mode ) {
		case WAITING_MODE:

			if ( caller == 0 )
			{
				error( this, "PortTailT::put(): port in waiting mode called from environment" );
				return -1;
			}

			if ( caller -> getSchedType() == Sched::PROCESS ) {

				Memory::remember( caller );
				dynamic_cast<Process*>(caller) -> sleep();

				// check for interruption of waiting period
				if( dynamic_cast<Process*>(caller) -> isInterrupted() ) {

					Memory::forget( caller );
					return 0;
				}
			}
			else if ( caller -> getSchedType() == Sched::EVENT ) {
				// error: cannot send an event to sleep
				error( this,  "PortHead::put(): waiting mode not allowed for events" );
				return -1;
			}
			break;

		case ERROR_MODE:
			error( this,  "PortTailT<>::put(): cannot add element to full port" );
			return -1;

		case ZERO_MODE:
			return 0;
		}
	}

	// remember if port was empty
	bool wasEmpty = internalPort -> elementList.empty();

	// insert new element
	internalPort -> elementList.push_back( newElement );

	// update stats
	internalPort -> putCalls++;
	internalPort -> portAccum.update( internalPort -> getNowCount() );

	// notify waiting Processes of newly inserted element
	if ( wasEmpty && internalPort -> portHead && internalPort -> portHead -> waiting() ) {
		internalPort -> portHead -> alert();
	}

	return 1;
}


// provides the corresponding PortHead
// if it does not exit, create it
template < typename ElementType >
PortHeadT< ElementType >* PortTailT< ElementType >::getHead() {

	PortHeadT< ElementType >* head = internalPort -> portHead;

	// head still zero?
	if ( !head ) {

		// create new Head
		head = new PortHeadT< ElementType >( env, internalPort -> getLabel(),
				mode, internalPort -> maxCapacity );

		// register new Head with Tail's Port
		internalPort -> portHead = head;

		// register Tail's internal Port with the Head
		head -> internalPort = internalPort;
	}

	return head;
}


template < typename ElementType >
const std::list< ElementType >& PortTailT< ElementType >::getList() const {
	return internalPort -> elementList;
}


// change
//      old_Port_head -- old_Port -- old_Port_tail
// into
//      old_Port_head -- old_Port -- new_Port_tail
//      (__getHead()___) -- new_Port -- old_Port_tail
// where
//      old_Port_tail == this
// return
//      new_Port_tail.
template < typename ElementType >
PortTailT< ElementType >* PortTailT< ElementType >::cut() {

	// trace
	getTrace() -> mark( this, markCut, getPartner() );

	// observer
	_obsForEachT(PortTailTObserver< ElementType >, Cut(this));

	// remember old Port
	PortT< ElementType >* oldPort = internalPort;

	// create a new Tail and thereby a new Port
	PortTailT< ElementType >* newTail =
		new PortTailT< ElementType >( env, ("Cut " + internalPort -> getLabel()),
									  mode, oldPort -> maxCapacity );

	// get the new Port
	PortT< ElementType >* newPort = newTail -> internalPort;

	// put the new tail on the old Port
	newTail -> internalPort = oldPort;
	oldPort -> portTail = newTail;

	// put the old tail on the new Port
	newPort -> portTail = this;
	internalPort = newPort;

	return newTail;
}


// this PortTail is supposed to be downstream to the PortHead oldHead:
// >> headsTail -- headsPort -- oldHead >> thisTail -- tailsPort -- tailsHead >>

template < typename ElementType >
void PortTailT< ElementType >::splice( PortHeadT< ElementType >* oldHead ) {

	// trace
	getTrace() -> mark( this, markSplice, getPartner() );

	// observer
	_obsForEachT(PortTailTObserver< ElementType >, Splice(this, oldHead));

	oldHead -> splice( this );
}


template < typename ElementType >
void PortTailT< ElementType >::cppSplice( PortHeadT< ElementType >* head,
										  bool append /*= true*/ )
{
	// trace
	getTrace() -> mark( this, markCppSplice, getPartner() );

	// observer
	_obsForEachT(PortTailTObserver< ElementType >, CppSplice(this, head, append));

	list< ElementType >& tailList = internalPort -> elementList;
	list< ElementType >& headList = head -> internalPort -> elementList;

	if( append )
		tailList.splice( tailList.end(), headList );
	else
		tailList.splice( tailList.begin(), headList );
}


template < typename ElementType >
unsigned int PortTailT< ElementType >::getMaxCapacity() {
	return internalPort -> maxCapacity;
}


template < typename ElementType >
void PortTailT< ElementType >::setMaxCapacity( unsigned int newMaximum ) {

	unsigned int oldCapacity = internalPort -> maxCapacity;

	// trace
	getTrace() -> mark( this, markChangeMaxCapacity, newMaximum, oldCapacity );

	// observer
	_obsAForEachT(PortTailTObserver< ElementType >, MaxCapacity, oldCapacity, newMaximum);

	internalPort -> maxCapacity = newMaximum;
}

template < typename ElementType >
unsigned int PortTailT< ElementType >::count() {
	return internalPort -> elementList.size();
}

template < typename ElementType >
unsigned int PortTailT< ElementType >::getAvailableSpace() {
	return getMaxCapacity() - count();
}


template < typename ElementType >
PortMode PortTailT< ElementType >::getMode() {
	return mode;
}


template < typename ElementType >
void PortTailT< ElementType >::setMode( PortMode newMode ) {

	PortMode oldMode = mode;

	// trace
	getTrace() -> mark( this, markChangeMode, newMode, oldMode );

	// observer
	_obsAForEachT(PortTailTObserver< ElementType >, Mode, oldMode, newMode);

	mode = newMode;
}


// does PortTail have room?
template < typename ElementType >
bool PortTailT< ElementType >::available () {
	return ( getAvailableSpace() > 0 );
}


template < typename ElementType >
void PortTailT< ElementType >::alert() {

	if ( ! Memory::waiting() ) {

		// warning: no object to wake
		if ( mode != ZERO_MODE )
			warning( this,  "PortTailT<>::alert(): no process saved" );
		return;
	}

	Sched* rememberedObject = Memory::getList().front();

	// trace
	getTrace() -> mark( this, markAlert, rememberedObject );

	// observer
	_obsForEachT( PortTailTObserver< ElementType >, Alert( this, rememberedObject ) );

	// check if rememberedObject really is a sleeping process
	if ( rememberedObject -> getSchedType() == Sched::PROCESS ) {

		Process* p = (Process*)rememberedObject;

		// allow only unscheduled processes to be alerted
		if ( p -> getProcessState() == Process::IDLE ||
			 p -> getProcessState() == Process::CREATED ) {

			// wake up the process
			p -> alertProcess( this );

			// update stats
			internalPort -> tailAlerts++;
		}
		else if ( p -> getProcessState() == Process::TERMINATED ) {

			// warning: process terminated
			error( this,  "PortTailT<>::alert(): process already terminated" );
		}
		else if ( p -> getProcessState() == Process::RUNNABLE ) {

			// warning: process scheduled
			warning( this,  "PortTailT<>::alert(): process already scheduled" );
		}
		// no warning for the CURRENT process,
		// events might be running on that stack
	}
	else {

		// warning: not a process, no action taken, but continue
		warning( this,  "PortTailT<>::alert(): remembered object is not a process" );
	}

	Memory::forget( rememberedObject );
}


template < typename ElementType >
TraceProducer* PortTailT< ElementType >::getPartner() {

	// return the current process or event, if one is active
	if ( env -> getCurrentSched() != 0 ) {

		return env -> getCurrentSched();
	}
	else { // or return the simulation

		return env;
	}
}


template < typename ElementType >
const MarkTypeId PortTailT< ElementType >::baseMarkId = MarkType::getNewBaseMarkId( typeid(PortTailT< ElementType >) );

template < typename ElementType >
const MarkType PortTailT< ElementType >::markCreate( "create", baseMarkId+1, typeid(PortTailT< ElementType >) ); ///< Call of PortTailT constructor
template < typename ElementType >
const MarkType PortTailT< ElementType >::markDestroy( "destroy", baseMarkId+2, typeid(PortTailT< ElementType >) ); ///< Call of PortTailT destructor

template < typename ElementType >
const MarkType PortTailT< ElementType >::markPut( "put", baseMarkId+3, typeid(PortTailT< ElementType >) ); ///< Attempt to put data into port, may lead to suspension of the calling Process, depending on PortMode
template < typename ElementType >
const MarkType PortTailT< ElementType >::markCut( "cut", baseMarkId+4, typeid(PortTailT< ElementType >) ); ///< Port was split into two ports, making the calling PortTailT the tail of the new port
template < typename ElementType >
const MarkType PortTailT< ElementType >::markSplice( "splice", baseMarkId+5, typeid(PortTailT< ElementType >) ); ///< Two Ports are joined together, whereby the contents of the upstream port is appended to the downstream one
template < typename ElementType >
const MarkType PortTailT< ElementType >::markCppSplice( "cpp splice", baseMarkId+6, typeid(PortTailT< ElementType >) ); ///< The contents of another port is copied to this port in the same way std::list::splice() does
template < typename ElementType >
const MarkType PortTailT< ElementType >::markAlert( "alert", baseMarkId+7, typeid(PortTailT< ElementType >) ); ///< PortTailT alerts the first Process waiting to insert data into full port, triggered by PortHeadT::get()

template < typename ElementType >
const MarkType PortTailT< ElementType >::markChangeMode( "change mode", baseMarkId+8, typeid(PortTailT< ElementType >) ); ///< Change of the PortMode which controls the put-when-full behavior of a PortHeadT
template < typename ElementType >
const MarkType PortTailT< ElementType >::markChangeMaxCapacity( "change maximum capacity", baseMarkId+9, typeid(PortTailT< ElementType >) ); ///< Change of a PortTailT's internal PortT capacity

