//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file PortTailT.h

	\author Ronald Kluth

	\date created at 2007/04/04

	\brief Declaration of odemx::PortT, odemx::PortHeadT and odemx::PortTailT

	\sa PortT.cpp, PortHeadT.cpp, PortTailT.cpp

	<!-- [detailed description] -->

	\since 2.0
*/

#ifndef PORT_TAIL_T_INCLUDED
#define PORT_TAIL_T_INCLUDED

namespace odemx {

	// forward declaration
	template < typename ElementType >
	class PortT;

	template < typename ElementType >
	class PortHeadT;

	template < typename ElementType >
	class PortTailTObserver;

	/** \class PortTailT

		\ingroup synch

		\author Ronald Kluth

		\brief %PortTailT provides write access to a port

		\sa PortHeadT, PortT

		This class manages write access to a limited internal buffer.
		It can be run in three different error modes according to
		PortMode. Put on a full port will either be handled by
		making the sender wait (and alerting him upon availability),
		reporting an error, or returning zero.

		\since 2.0
	*/
	template < typename ElementType >
	class PortTailT: public Memory,	// is a trace and report producer
					 public DefLabeledObject,
					 public Observable< PortTailTObserver< ElementType > > {

	// public interface
	public:

		/**
			\brief Construction for DefaultSimulation
			\param portName
				label of the internal port, this object's label will be "portName_tail"
			\param m
				port tail's mode, default is WAITING_MODE
			\param max
				maximum number of elements, default is 10000
			\param pto
				initial observer
		*/
		PortTailT( Label portName,
				   PortMode m = WAITING_MODE,
				   unsigned int max = 10000,
				   PortTailTObserver< ElementType >* pto = 0 );

		/**
			\brief Construction for user-defined Simulation
			\param sim
				pointer to the simulation object
			\param portName
				label of the internal port, this object's label will be "portName_tail"
			\param m
				port tail's mode, default is WAITING_MODE
			\param max
				maximum number of elements, default is 10000
			\param pto
				initial observer
		*/
		PortTailT( Simulation* sim, Label portName,
				   PortMode m = WAITING_MODE,
				   unsigned int max = 10000,
				   PortTailTObserver< ElementType >* pto = 0 );

		/// Destruction
		virtual ~PortTailT ();

		/**
			\brief Port write access
			\param newElement
				the element to insert into the port
			\return
				\p 1 if insertion successful, \p 0 in ZERO_MODE or interrupt(), else \p -1

			This function is used to insert an element into a port. It also
			responds according to the object's PortMode. In WAIT_MODE, the
			calling Process will be suspended if the Port is already full. When
			space becomes available, the callers will be alerted in a FIFO-manner.
			The caller can be interrupted, in which case that Process will be
			activated immediately and put() returns 0.
		*/
		int put( ElementType newElement );

		/**
			\brief Get corresponding PortHead
			\return
				corresponding PortHead to this PortTail

			This function provides access to the head of this port tail.
			If it does not exist yet, it is created.
		*/
		PortHeadT< ElementType >* getHead();

		/**
			\brief Get a reference to the internal list
			\return
				reference to the buffer's internal list

			This function provides access to the internal buffer list
			of the head's PortT.
		*/
		const std::list< ElementType >& getList() const;

		/**
			\brief Cut this PortTail from its head
			\return
				new port tail

			This function will create a	new internal port, which
			is attached to this tail. A new head for that port must then be
			requested via getHead(). Furthermore, a new port tail will
			be attached to the old internal port. That new tail will be
			returned.

			This provides the ability to insert a process into
			the middle of a port connection, for example as a filter.

			\sa splice()
		*/
		PortTailT< ElementType >* cut();

		/**
			\brief Splice two separate ports together
			\param oldHead
				the tail of the port to which this head's buffer will be attached

			This function is used to join two ports together. This tail must
			be downstream of, i.e. positioned after, oldHead's port, which
			this buffer will be joined with. The contents of this Port
			is added to oldHead's Port. This tail, oldHead, and this Port are
			deleted.

			\note This function does not check the maximum Port capacity before
			splicing lists.

			\sa cut()
		*/
		void splice( PortHeadT< ElementType >* oldHead );

		/**
			\brief Splice internal lists of two separate ports together
			\param head
				the head of the port whose buffer will be added to this
			\param append
				determines whether to append or prepend head's contents

			This function is used to join two ports' buffer contents together
			in the same way as std::list::splice(). The contents of the \p
			tail PortT will be moved to \p this port, where parameter \p append
			can	be used to specify whether the data will appended or prepended
			to this port. Default is append.

			\note This function does not check the maximum Port capacity before
			splicing lists.
		*/
		void cppSplice( PortHeadT< ElementType >* head, bool append = true );

		/**
			\brief Get maximum capacity of the internal buffer
			\return
				capacity of the port

			This function is used to determine the maximum number of
			elements that can be buffered by the internal port.

			\sa setMaxCapacity()
		*/
	 	unsigned int getMaxCapacity();

	 	/**
			\brief Change maximum capacity of the port
			\param newMaximum
				the new maximum capacity of the internal port

			This function allows to change the internal buffer limit.

			\sa getMaxCapacity()
		*/
		void setMaxCapacity( unsigned int newMaximum );

		/**
			\brief Get current number of port elements
			\return
				current number of elements

			This function returns the current number of buffered elements.
		*/
		unsigned int count();

		/**
			\brief Determine free space of the port
			\return
				number of elements that the port can still take in

			This function returns the remaining available space
			of the internal buffer.
		*/
		unsigned int getAvailableSpace();

		/**
			\brief Get the port tail's mode
			\return
				the currently set mode

			This function is used to determine the port tail's mode.

			\sa setMode()
		*/
		PortMode getMode();

		/**
			\brief Change the port tail's mode
			\param newMode
				new mode to be set

			This function is used to change the port tail's mode.

			\sa getMode()
		*/
		void setMode( PortMode newMode );

		/**
			\brief Check for availability of the buffer
			\return
				\p true if the port is not full

			This function is used to determine whether a port
			still has room to contain new elements.
		*/
		virtual bool available();

		/**
			\brief Report producing capability
			\param r
				Report object that receives accumulated data

			This function is called by a Report object during report
			generation to provide its data.
		*/
		virtual void report(Report* r) {
			internalPort -> report( r );
			Memory::report( r );
		}

		/// Reset object statistics to \c 0
		virtual void reset()
		{
			StatisticObject::reset();
			Memory::reset();
			internalPort -> reset();
		}
	private:

		friend class PortHeadT< ElementType >;

		/**
			\brief Alert the first stored process

			This function is designed to wake up suspended processes,
			i.e. reschedule them at simulation time \p now when they
			are waiting for a full port to become available. Only
			the first remembered process is alerted by the %PortTail.
		*/
		virtual void alert();

		PortT< ElementType >* internalPort;	///< internal buffer of PortElements
		PortMode	mode;	///< mode for read requests
		Simulation* env;	///< simulation context

		// ========================= Trace Marks ===============================

		/**
			\brief Get currently running Sched object or simulation
			\return
				pointer to simulation or current Sched object
		*/
		TraceProducer* getPartner();

	public:
		/**
			\name Trace MarkTypes

			These MarkTypes and Tags are used to trace PortHead calls.
			A TraceConsumer can use these constants to identify trace
			events sent	by PortHead.

			@{
		*/
		static const MarkTypeId baseMarkId;

		static const MarkType markCreate;
		static const MarkType markDestroy;

		static const MarkType markPut;
		static const MarkType markCut;
		static const MarkType markSplice;
		static const MarkType markCppSplice;
		static const MarkType markAlert;

		static const MarkType markChangeMode;
		static const MarkType markChangeMaxCapacity;
		//@}

	};


	/** \interface PortTailTObserver

		\author Ronald Kluth

		\brief Observer for PortTailT-specific calls.

		\sa PortTail

		<!-- [detailed description] -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.0
	*/

	template < typename ElementType >
	class PortTailTObserver: public MemoryObserver {
	public:
		virtual ~PortTailTObserver() {}

		virtual void onCreate(PortTailT< ElementType >* sender) {} ///< Construction
		virtual void onDestroy(PortTailT< ElementType >* sender) {} ///< Destruction

		virtual void onPut(PortTailT< ElementType >* sender, ElementType newElement) {}
		virtual void onCut(PortTailT< ElementType >* sender) {}
		virtual void onSplice(PortTailT< ElementType >* sender, PortHeadT< ElementType >* oldHead) {}
		virtual void onCppSplice(PortTailT< ElementType >* sender, PortHeadT< ElementType >* oldHead, bool append) {}
		virtual void onAlert(PortTailT< ElementType >* sender, Sched* alerted) {}

		/// PortTail mode change
		virtual void onChangeMode(PortTailT< ElementType >* sender, PortMode oldMode, PortMode newMode) {}
		/// PortTail capacity change
		virtual void onChangeMaxCapacity(PortTailT< ElementType >* sender, unsigned int oldMax, unsigned int newMax) {}
	};

}

#endif /* PORT_TAIL_T_INCLUDED */
