//----------------------------------------------------------------------------
//	Copyright (C) 2002-2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ResT.cpp

	\author Ronald Kluth

	\date created at 2008/02/18

	\brief Definition of odemx::ResT

	\sa ResT.h

	<!-- [detailed description] -->

	\since 2.1
*/

using namespace odemx;
using namespace std;

template < typename Token >
ResT< Token >::ResT( Label l,
					 vector< Token >& initialResources,
					 ResTObserver< Token >* o /* = 0*/ ):
	ResTBase< Token >( l, initialResources, o ),
	Observable< ResTObserver< Token > >( o )
	{
		// observer
		_obsForEachT(ResTObserver<Token>, Create(this));
	}

template < typename Token >
ResT< Token >::ResT( Simulation* sim,
					 Label l,
					 vector< Token >& initialResources,
					 ResTObserver< Token >* o /* = 0*/ ):
	ResTBase< Token >( sim, l, initialResources, o ),
	Observable< ResTObserver< Token > >( o )
	{
		// observer
		_obsForEachT(ResTObserver<Token>, Create(this));
	}

template < typename Token >
ResT<Token>::~ResT() {
	// observer
	_obsForEachT(ResTObserver<Token>, Destroy(this));
}

template < typename Token >
Token* ResT< Token >::acquire() {

	// get a return vector using acquire( n ) below
	vector< Token* > v = acquire( 1 );

	// return the first element of the vector
	if( ! v.empty() )
		return v.front();

	// or 0, if nothing was returned (i.e. error or interrupt)
	return 0;
}

template < typename Token >
const vector< Token* > ResT< Token >::acquire( unsigned int n ) {

	if( n > ResTBase< Token >::tokenStore.size() ) {
		// warning: will always block
		warning( this, "ResT::acquire(): requesting more than max resources will always block" );
	}

	if( ! getCurrentSched() ||
		getCurrentSched() -> getSchedType() != Sched::PROCESS ) {
		// error: resource handling only implemented for processes
		error( this, "ResT::acquire(): called by non-Process object" );
		return vector< Token* >();
	}

	// compute order of service
	ResTBase< Token >::acquireWait.inSort( getCurrentProcess() );

	if( n > ResTBase< Token >::tokenNumber ||
		ResTBase< Token >::acquireWait.getTop() != getCurrentProcess() ) {
		// not enough tokens or not the first process in queue

		// trace
		getTrace()->beginMark(this, ResTBase<Token>::markAcquireFail);
		getTrace()->addTag(ResTBase<Token>::tagCurrent, getCurrentProcess());
		getTrace()->addTag(ResTBase<Token>::tagTokenNumber, n);
		getTrace()->endMark();

		// observer
		_obsForEachT(ResTObserver<Token>, AcquireFail(this, n));

		// statistics
		SimTime waitTime = getTime();

		// block execution
		while( n > ResTBase< Token >::tokenNumber ||
			   ResTBase< Token >::acquireWait.getTop() != getCurrentProcess() ) {

			getCurrentProcess() -> sleep();

			// handle interrupt
			if( getCurrentProcess() -> isInterrupted() ) {

				ResTBase< Token >::acquireWait.remove( getCurrentProcess() );
				return vector< Token* >();
			}
		}

		// block released
		// statistics
		ResTBase<Token>::sumWaitTime += getTime() - waitTime;
	}

	// trace
	getTrace()->mark(this, ResTBase<Token>::markChangeTokenNumber, ResTBase<Token>::tokenNumber-n, ResTBase<Token>::tokenNumber);

	// observer
	_obsAForEachT(ResTObserver<Token>, TokenNumber, ResTBase< Token >::tokenNumber, ResTBase< Token >::tokenNumber-n);

	// acquire token
	ResTBase< Token >::tokenNumber -= n;

	// fill return vector with pointers to free resource objects
	typename map< Token*, Process* >::iterator i;
	vector< Token* > availableTokens;

	unsigned int acquired = 0;
	for( i = ResTBase< Token >::tokenInfo.begin();
		 i != ResTBase< Token >::tokenInfo.end() && acquired < n;
		 ++i ) {
		// if resource available, store the current user in tokenInfo
		// and add the resource to the return vector
		if( i -> second == 0 ) {
			availableTokens.push_back( i -> first );
			i -> second = getCurrentProcess();
			++acquired;
		}
	}

	// trace
	getTrace()->beginMark(this, ResTBase<Token>::markAcquireSucceed);
	getTrace()->addTag(ResTBase<Token>::tagCurrent, getCurrentProcess());
	getTrace()->addTag(ResTBase<Token>::tagTokenNumber, n);
	getTrace()->endMark();

	// observer
	_obsForEachT(ResTObserver<Token>, AcquireSucceed(this, n));

	// statistics
	++ResTBase<Token>::users;
	ResTBase<Token>::tokenStatistics.update(ResTBase<Token>::tokenNumber);

	// remove process from queue
	ResTBase< Token >::acquireWait.remove( getCurrentProcess() );

	// awake next waiting process
	awakeFirst( ResTBase< Token >::getAcquireWait() );

	// return tokens
	return availableTokens;
}

