//----------------------------------------------------------------------------
//	Copyright (C) 2002-2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ResT.h

	\author Ronald Kluth

	\date created at 2008/02/18

	\brief Declaration of odemx::ResT and observer

	\sa ResT.cpp

	<!-- [detailed description] -->

	\since 2.1
*/

#ifndef ODEMX_RES_T_INCLUDED
#define ODEMX_RES_T_INCLUDED

#include <odemx/synchronization/res/ResTBase.h>

namespace odemx {
	template < typename Token >
	class ResTObserver;

	/** \class ResT

		\author Ronald Kluth

		\brief ResT is a resource holding objects of type Token

		\note ResT supports Observation
		\note ResT supports Trace

		<!-- [\sa {references to other classes}]* -->

		ResT is a resource for synchronizing Process objects. A process is blocked
		in acquire() if the requested number of tokens is greater than the available
		number of tokens. It is reactivated when enough tokens are released to the resource.
		If multiple process	objects are waiting for reactivation, process-priority and
		FIFO-strategy are used to choose the process.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		<!-- [\since {since when in project}] -->
	*/
	template < typename Token >
	class ResT : public ResTBase< Token >,
				 public Observable< ResTObserver< Token > > {
	public:

		/**
		 * \copydoc ResTBase(Label,std::vector<Token>&,ResTBaseObserver<Token>*)
		 */
		ResT( Label l, std::vector< Token >& initialResources,
			  ResTObserver< Token >* o = 0 );

		/**
		 * \copydoc ResTBase(Simulation*,Label,std::vector<Token>&,ResTBaseObserver<Token>*)
		 */
		ResT( Simulation* sim, Label l, std::vector< Token >& initialResources,
			  ResTObserver< Token >* o = 0 );

		/// Destruction
		virtual ~ResT();

		/**
			\name Token usage

			These functions provade access for resource usage by Processes.
			Tokens can be requested and thus acquired when then are
			available. When the tokens are not needed anymore, the owning
			Process	can release them.

			@{
		*/
		/**
			\brief acquire a token

			If there are no tokens in ResT the current process is blocked.
		*/
		virtual Token* acquire();

		/**
			\brief acquire \p n token

			If there are not enough tokens in ResT the current
			process is blocked.
		*/
		const std::vector< Token* > acquire( unsigned int n );

	private:

		/// get pointer to the trace object
		virtual Trace* getTrace() const {
			return (Trace*) ResTBase< Token >::env;
		}

		/// current simulation time
		SimTime getTime() {
			return ResTBase< Token >::env -> getTime();
		}
		/// current process
		Process* getCurrentProcess() {
			return ResTBase< Token >::env -> getCurrentProcess();
		}
		/// current scheduled object
		Sched* getCurrentSched() {
			return ResTBase< Token >::env -> getCurrentSched();
		}

	};

	/** \interface ResTObserver

		\author Ronald Kluth

		\brief Observer for ResT specific events

		<!-- [\note {notes}]* -->

		\sa ResT

		<!-- [detailed description] -->

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		<!-- [\since {since when in project}] -->
	*/
	template < typename Token >
	class ResTObserver: public ResTBaseObserver< Token > {
	public:
		virtual ~ResTObserver() {}

		/// Creation
		virtual void onCreate(ResTBase< Token >* sender) {}
		/// Destruction
		virtual void onDestroy(ResTBase< Token >* sender) {}

		/// Failed attempt to acquire n token
		virtual void onAcquireFail(ResTBase< Token >* sender, unsigned int n) {}
		/// Successful attempt to acquire n token
		virtual void onAcquireSucceed(ResTBase< Token >* sender, unsigned int n) {}
		/// Failed attempt to release n token
		virtual void onReleaseFail(ResTBase< Token >* sender, unsigned int n) {}
		/// Release of n token
		virtual void onReleaseSucceed(ResTBase< Token >* sender, unsigned int n) {}

		/// Change of token number
		virtual void onChangeTokenNumber(ResTBase< Token >* sender, unsigned int oldTokenNumber,
			unsigned int newTokenNumber) {}
	};
}

#endif /* ODEMX_RES_T_INCLUDED */
