//----------------------------------------------------------------------------
//	Copyright (C) 2002-2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ResTBase.cpp

	\author Ronald Kluth

	\date created at 2008/02/18

	\brief Definition of odemx::ResTBase

	\sa ResTBase.h

	<!-- [detailed description] -->

	\since 2.1
*/

using namespace odemx;
using namespace std;

template < typename Token >
ResTBase< Token >::ResTBase( Label l,
							 vector< Token >& initialResources,
							 ResTBaseObserver< Token >* o /* = 0*/ ):
	Observable< ResTBaseObserver< Token > >( o ),
	DefLabeledObject( getDefaultSimulation(), l ),
	StatisticObject( getDefaultSimulation() ),
	env( getDefaultSimulation() ),
	tokenLimit( initialResources.size() ),
	tokenStatistics( l + "token statistics" ),
	tokenStore( initialResources.begin(), initialResources.end() ),
	acquireWait( l + " queue" )
	{
		// tokenStore already filled in initializer list

		// set initial tokenStore
		tokenNumber = tokenStore.size();
		initToken = tokenNumber;

		// fill map tokenInfo with addresses of tokens and 0's for owner Processes
		typename list< Token >::iterator i;
		for( i = tokenStore.begin(); i != tokenStore.end(); ++i )
			tokenInfo.insert( make_pair( &(*i), static_cast< Process* >(0) ) );

		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEachT(ResTBaseObserver<Token>, Create(this));

		// statistics
		reset();
	}

template < typename Token >
ResTBase< Token >::ResTBase( Simulation* sim,
					 Label l,
					 vector< Token >& initialResources,
					 ResTBaseObserver< Token >* o /* = 0*/ ):
	Observable< ResTBaseObserver< Token > >( o ),
	DefLabeledObject( sim, l ),
	StatisticObject( sim ),
	env( sim ),
	tokenLimit( initialResources.size() ),
	tokenStatistics( l + "token statistics" ),
	tokenStore( initialResources.begin(), initialResources.end() ),
	acquireWait( sim, l + " acquire queue" )
	{
		assert( sim != 0 );
		// tokenStore already filled in initializer list

		// set initial tokenNumber
		tokenNumber = tokenStore.size();
		initToken = tokenNumber;

		// fill map tokenInfo with addresses of tokens and 0's for owner Processes
		typename list< Token >::iterator i;
		for( i = tokenStore.begin(); i != tokenStore.end(); ++i )
			tokenInfo.insert( make_pair( &(*i), static_cast< Process* >(0) ) );

		// trace
		getTrace()->mark(this, markCreate);

		// observer
		_obsForEachT(ResTBaseObserver<Token>, Create(this));

		// statistics
		reset();
	}

template < typename Token >
ResTBase< Token >::~ResTBase() {
	// trace
	getTrace()->mark(this, markDestroy);

	// observer
	_obsForEachT(ResTBaseObserver<Token>, Destroy(this));
}

template < typename Token >
void ResTBase< Token >::release( Token* t ) {

	// put free resource into a vector
	vector< Token* > vec;
	vec.push_back( t );

	// handle that with release() below
	return release( vec );
}

template < typename Token >
void ResTBase< Token >::release( vector< Token* >& releaseVec ) {

	Process* current = getCurrentProcess();
	unsigned int n = releaseVec.size();

	// for each released resource, update the tokenInfo map
	typename vector< Token* >::iterator released;
	typename map< Token*, Process* >::iterator stored;

	unsigned int countedReleases = 0;
	for( released = releaseVec.begin(); released != releaseVec.end(); ++released )
	{
		// try to get iterator to released token in tokenInfo
		stored = tokenInfo.find( *released );

		// check if the iterator is valid and whether
		// the current Process really owns that resource
		if( stored != tokenInfo.end() && stored -> second == current )
		{
			// released resource found, set the resource user to 0
			stored -> second = static_cast< Process* >(0);
			++countedReleases;
			continue;
		}

		if( stored == tokenInfo.end() )
		{
			error( this, "ResTBase::release(): token not found in token info" );
		}
		else if( stored -> second != current )
		{
			error( this, "ResTBase::release(): token returned by wrong user" );
		}
	}

	// trace
	getTrace()->mark(this, markChangeTokenNumber, tokenNumber+countedReleases, tokenNumber);

	// observer
	_obsAForEachT(ResTBaseObserver<Token>, TokenNumber, tokenNumber, tokenNumber+countedReleases);

	// release token
	tokenNumber += countedReleases;

	// check for failure
	if ( countedReleases < n )
	{
		warning( this, "ResTBase::release(): could not release all tokens" );

		//trace
		getTrace()->beginMark(this, markReleaseFail);
		getTrace()->addTag(tagCurrent, getCurrentProcess());
		getTrace()->addTag(tagTokenNumber, n - countedReleases);
		getTrace()->endMark();

		// observer
		_obsForEachT(ResTBaseObserver<Token>, ReleaseFail(this, n - countedReleases));
	}

	//trace
	getTrace()->beginMark(this, markReleaseSucceed);
	getTrace()->addTag(tagCurrent, getCurrentProcess());
	getTrace()->addTag(tagTokenNumber, countedReleases);
	getTrace()->endMark();

	// observer
	_obsForEachT(ResTBaseObserver<Token>, ReleaseSucceed(this, countedReleases));

	// statistics
	++providers;
	tokenStatistics.update(tokenNumber);

	// wake up waiting process objects
	awakeFirst( &acquireWait );
}


template < typename Token >
void ResTBase< Token >::transfer( Token* t, Process* oldOwner, Process* newOwner ) {
	vector< Token* > v;
	v.push_back( t );
	return transfer( v, oldOwner, newOwner );
}

template < typename Token >
void ResTBase< Token >::transfer( vector< Token* >& v, Process* oldOwner, Process* newOwner ) {

	typename vector< Token* >::iterator transferToken;
	typename map< Token*, Process* >::iterator found;

	// set a new owner for the given tokens
	for( transferToken = v.begin(); transferToken != v.end(); ++transferToken )
	{
		found = tokenInfo.find( *transferToken );

		// token not found ?
		if( found == tokenInfo.end() )
		{
			error( this, "ResTBase::transfer(): transfer token not found in token info" );
			continue;
		}

		// correct owner ?
		if( found -> second == oldOwner )
		{
			found -> second = newOwner;
		}
		else // oldOwner is different from owner stored in tokenInfo
		{
			error( this, "ResTBase::transfer(): transfer token owned by another Process" );
		}
	}
}

template < typename Token >
void ResTBase< Token >::control( Token t ) {

	// increase max by one
	++tokenLimit;

	// add token to tokenStore and tokenInfo
	tokenStore.push_back( t );
	tokenInfo.insert( make_pair( &(tokenStore.back()), static_cast< Process* >(0) ) );
	++tokenNumber;

	// wake up waiting process objects
	awakeFirst( &acquireWait );
}


template < typename Token >
void ResTBase< Token >::control( vector< Token >& addVec ) {

	// increase by number of added tokens
	tokenLimit += addVec.size();

	// fill tokenStore and tokenInfo with tokens from addVec
	// as long as the tokenLimit is not reached yet
	typename vector< Token >::iterator i;
	for( i = addVec.begin();
		 i != addVec.end() && tokenStore.size() < tokenLimit;
		 ++i ) {

		tokenStore.push_back( *i );
		tokenInfo.insert( make_pair( &(tokenStore.back()), static_cast< Process* >(0) ) );
		++tokenNumber;
	}

	// wake up waiting process objects
	awakeFirst( &acquireWait );
}


template < typename Token >
vector< Token > ResTBase< Token >::unControl( unsigned int n /* = 1 */) {

	if( n > tokenStore.size() ) {
		error( this, "ResTBase::unControl(): removing more than max resources not allowed" );
		return vector< Token >();
	}

	if( ! getCurrentSched() ||
		getCurrentSched() -> getSchedType() != Sched::PROCESS ) {
		// error: resource handling only implemented for processes
		error( this, "ResTBase::unControl(): called by non-Process object" );
		return vector< Token >();
	}

	// acquire enough token
	// compute order of service
	acquireWait.inSort( getCurrentProcess() );

	if( n > tokenNumber ||
		acquireWait.getTop() != getCurrentProcess() ) {
		// not enough tokens or not the first process in queue

		// block execution
		while( n > tokenNumber ||
			   acquireWait.getTop() != getCurrentProcess() ) {

			getCurrentProcess() -> sleep();

			// handle interrupt
			if( getCurrentProcess() -> isInterrupted() ) {

				acquireWait.remove( getCurrentProcess() );
				return vector< Token >();
			}
		}

		// block released
	}

	// observer
	_obsAForEachT(ResTBaseObserver<Token>, TokenNumber, tokenNumber, tokenNumber-n);

	// fill return vector with the uncontrolled token objects
	typename map< Token*, Process* >::iterator tInfo = tokenInfo.begin();
	vector< Token > unControlTokens;

	unsigned int acquired = 0;
	while( tInfo != tokenInfo.end() && acquired < n )
	{
		if( tInfo -> second == 0 ) {

			unControlTokens.push_back( *(tInfo -> first) );
			++acquired;

			// remove the uncontrolled token from store
			tokenStore.remove( *(tInfo -> first) );

			// remove the uncontrolled token from managing info map
			// by using a copy of the non-incremented iterator returned by op++
			tokenInfo.erase( tInfo++ );
		}
		else
		{
			++tInfo;
		}
	}

	// acquire token
	tokenNumber -= acquired;

	// adjust token limit
	tokenLimit -= acquired;

	// remove process from queue
	acquireWait.remove( getCurrentProcess() );

	// awake next waiting process
	awakeFirst( &acquireWait );

	// return tokens
	return unControlTokens;
}


template < typename Token >
unsigned int ResTBase< Token >::getTokenNumber() const {
	return tokenNumber;
}

template < typename Token >
unsigned int ResTBase< Token >::getTokenLimit() const {
	return tokenLimit;
}

template < typename Token >
const map< Token*, Process* >& ResTBase< Token >::getTokenInfo() const {
	return tokenInfo;
}

template < typename Token >
const std::list< Process* >& ResTBase< Token >::getWaitingProcessList() const {
	return acquireWait.getList();
}

template < typename Token >
Trace* ResTBase< Token >::getTrace() const {
	return env;
}

template < typename Token >
void ResTBase< Token >::reset() {
	// reset time
	StatisticObject::reset();

	// reset queue statistics
	acquireWait.reset();

	// Accum reset and re-init
	tokenStatistics.reset();
	tokenStatistics.update(getTokenNumber());

	users=providers=0;
	sumWaitTime=0;
}

template < typename Token >
double ResTBase< Token >::getAVWaitTime() const {
	return sumWaitTime/getAcquires();
}

template < typename Token >
void ResTBase< Token >::report(Report* r) {

	assert(r);
	acquireWait.report(r);

	static vector< string > labels;			static vector< Column::Type  > types;
	labels.push_back("Name");				types.push_back( Column::STRING);
	labels.push_back("Token type");			types.push_back( Column::STRING);
	labels.push_back("Reset at");			types.push_back( Column::SIMTIME);
	labels.push_back("Queue");				types.push_back( Column::STRING);
	labels.push_back("Acquires");			types.push_back( Column::INTEGER);
	labels.push_back("Releases");			types.push_back( Column::INTEGER);
	labels.push_back("Init token number");	types.push_back( Column::INTEGER);
	labels.push_back("Limit token number");	types.push_back( Column::INTEGER);
	labels.push_back("Min token number");	types.push_back( Column::INTEGER);
	labels.push_back("Max token number");	types.push_back( Column::INTEGER);
	labels.push_back("Now token number");	types.push_back( Column::INTEGER);
	labels.push_back("Avg token number");	types.push_back( Column::REAL);
	labels.push_back("Avg waiting time");	types.push_back( Column::REAL);

	static utTableDef def( labels.size(), labels, types );

	Table* t = r -> createTable("ResT Statistics", &def);
	assert(t);

	*t
	<< getLabel()
	<< typeToString( typeid(Token) )
	<< StatisticObject::getResetTime()
	<< acquireWait.getLabel()
	<< getAcquires()
	<< getReleases()
	<< getInitial()
	<< getTokenLimit()
	<< getMin()
	<< getMax()
	<< getTokenNumber()
	<< getAVFreeToken()
	<< getAVWaitTime();
}

template < typename Token >
const MarkTypeId ResTBase< Token >::baseMarkId = MarkType::getNewBaseMarkId( typeid(ResTBase< Token >) );

template < typename Token >
const MarkType ResTBase< Token >::markCreate( "create", baseMarkId+1, typeid(ResTBase< Token >)); ///< Call of ResTconstructor
template < typename Token >
const MarkType ResTBase< Token >::markDestroy( "destroy", baseMarkId+2, typeid(ResTBase< Token >)); ///< Call of ResT destructor
template < typename Token >
const MarkType ResTBase< Token >::markAcquireFail( "acquire fail", baseMarkId+3, typeid(ResTBase< Token >)); ///< ResT does not contain a sufficient number of tokens, caller Process is enqueued and suspended
template < typename Token >
const MarkType ResTBase< Token >::markAcquireSucceed( "acquire succeed", baseMarkId+4, typeid(ResTBase< Token >)); ///< ResT contained enough tokens, caller Process can now continue with its execution
template < typename Token >
const MarkType ResTBase< Token >::markReleaseFail( "release fail", baseMarkId+5, typeid(ResTBase< Token >)); ///< ResT could not contain all released tokens, only releasing tokens till limit reached
template < typename Token >
const MarkType ResTBase< Token >::markReleaseSucceed( "release succeed", baseMarkId+6, typeid(ResTBase< Token >)); ///< ResT capacity high enough, Process successfully returned its used tokens
template < typename Token >
const MarkType ResTBase< Token >::markChangeTokenNumber( "change token number", baseMarkId+7, typeid(ResTBase< Token >)); ///< ResT token number changed, displays old and new values

template < typename Token >
const TagId ResTBase< Token >::baseTagId = ResTBase< Token >::baseMarkId;

template < typename Token >
const Tag ResTBase< Token >::tagCurrent = Tag( "current", baseTagId+1); ///< Displays details about the currently active Process
template < typename Token >
const Tag ResTBase< Token >::tagTokenNumber = Tag( "token number", baseTagId+2); ///< Displays details about the ResT token number

