//----------------------------------------------------------------------------
//	Copyright (C) 2002-2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ResT.h

	\author Ronald Kluth

	\date created at 2008/02/18

	\brief Declaration of odemx::ResTBase and observer

	\sa ResT.cpp

	<!-- [detailed description] -->

	\since 2.1
*/

#ifndef ODEMX_RES_T_BASE_INCLUDED
#define ODEMX_RES_T_BASE_INCLUDED

#include <odemx/base/Process.h>
#include <odemx/base/Simulation.h>
#include <odemx/synchronization/Queue.h>
#include <odemx/util/Observable.h>
#include <odemx/util/ErrorHandling.h>

#include <set>
#include <map>

namespace odemx {

	//forward declaration
	template <typename Token>
	class ResTBaseObserver;

	/** \class ResTBase

		\author Ronald Kluth

		\brief ResTBase is an abstract base class for resources holding
		objects of type Token

		\note ResTBase supports Observation
		\note ResTBase supports Trace

		\sa Res, ResT, ResTChoice

		ResTBase provides many functions needed in resource classes
		for synchronizing Process objects.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		<!-- [\since {since when in project}] -->
	*/
	template < typename Token >
	class ResTBase: public Observable< ResTBaseObserver< Token > >,
					public DefLabeledObject,
					public StatisticObject,
					public virtual TraceProducer,
					public virtual ReportProducer {
	public:

		/**
			\brief Construction for use with DefaultSimulation
			\param l
				label of ResT object
			\param initialResources
				vector containing the initial tokens of the resource
			\param o
				observer of ResT object

			\note The Token type must be copyable! The initial resources will
			be copied into the internal token store of the object.

			\note The token limit is set by the size of the \c initialResources
			vector. The change it, use control(), or unControl().
		*/
		ResTBase( Label l, std::vector< Token >& initialResources,
				  ResTBaseObserver< Token >* o = 0 );

		/**
			\brief Construction for user-defined Simulation
			\param s
				pointer to simulation
			\param l
				label of ResT object
			\param initialResources
				vector containing the initial tokens of the resource
			\param o
				observer of ResT object

			\note The Token type must be copyable! The initial resources will
			be copied into the internal token store of the object.

			\note The token limit is set by the size of the \c initialResources
			vector. The change it, use control(), or unControl().
		*/
		ResTBase( Simulation* sim, Label l, std::vector< Token >& initialResources,
				  ResTBaseObserver< Token >* o = 0 );

		/// Destruction
		virtual ~ResTBase();

		/**
			\name Token usage

			These functions provide access for resource usage by Processes.
			Tokens can be requested and thus acquired when then are
			available. When the tokens are not needed anymore, the owning
			Process	can release them.

			\note There are different ways to select tokens,
			this base class only provides a declaration for acquire.

			\note Use the derived classes ResT or ResTChoice
			@{
		*/
		/**
			\brief forces implementation of acquire

			Pure virtual function which must be implemented in
			derived classes.
		*/
		virtual Token* acquire() = 0;

		/**
			\brief release a token

			Returns a previously acquired token to the resource
		*/
		void release( Token* e );

		/**
			\brief release token

			Returns previously acquired tokens to the resource
		*/
		void release( std::vector< Token* >& s );

		/**
			\brief transfer control of an acquired token to \p newOwner

			The \p newOwner is registered as holder of token \p e, meaning
			this info is changed in the internal token info map.
		*/
		void transfer( Token* t, Process* oldOwner, Process* newOwner );

		/**
			\brief transfer control of acquired token to \p newOwner

			The \p newOwner is registered as holder of the tokens in \p s.
		*/
		void transfer( std::vector< Token* >& v, Process* oldOwner, Process* newOwner);

		/**
			\brief add token \p e to ressource

			Add token \p e to the managed token set.
		*/
		void control( Token t );

		/**
			\brief add token to ressource

			Add token to the managed token set.
		*/
		void control( std::vector< Token >& addVec );

		/**
			\brief remove \p n token from ressource

			Remove token from the managed token set. If there are not enough
			token left in the ressource, the current process is blocked.
		*/
		std::vector< Token > unControl( unsigned int n = 1 );
		//@}

		/** \name ResT status information

			The following functions provide access to the ResT object's
			status during a simulation.

			@{
		*/
		/**
			\brief get number of available token

			Get number of currently avaliable token in ressource
		*/
		unsigned int getTokenNumber() const;

		/**
			\brief get number of managed token

			Get maximum number of tokens managed by this ressource
		*/
		unsigned int getTokenLimit() const;

		/**
			\brief Get token map

			Get info about all tokens and their owning processes.
		*/
		const std::map< Token*, Process* >& getTokenInfo() const;

		/// Get a reference to the stored tokens
		const std::list<Token>& getTokenStore() const { return tokenStore; }

		/// get list of blocked  process objects
		const std::list< Process* >& getWaitingProcessList() const;
		//@}

		/// Generate report
		virtual void report(Report* r);

		/**
			\name Statistics

			Statistics

			@{
		*/
		/// Reset statistics
		virtual void reset();

		/// Number of takes
		unsigned int getAcquires() const {return users;}

		/// Number of gives
		unsigned int getReleases() const {return providers;}

		/// Init number of token
		unsigned int getInitial() const {return initToken;}

		/// Min number of token
		unsigned int getMin() const {return (unsigned int)tokenStatistics.getMin();}

		/// Max number of token
		unsigned int getMax() const {return (unsigned int)tokenStatistics.getMax();}

		/// Average free token
		double getAVFreeToken() const {return tokenStatistics.getMean();}

		/// Average waiting time
		double getAVWaitTime() const;
		//@}

		/// get pointer to trace object
		virtual Trace* getTrace() const;

		/**
			\name ResT specific marks
			@{
		*/
		static const MarkTypeId baseMarkId;

		static const MarkType markCreate;
		static const MarkType markDestroy;

		static const MarkType markAcquireFail;
		static const MarkType markAcquireSucceed;
		static const MarkType markReleaseFail;
		static const MarkType markReleaseSucceed;

		static const MarkType markChangeTokenNumber;

		static const TagId baseTagId;

		static const Tag tagCurrent;
		static const Tag tagTokenNumber;
		//@}

	protected:

		/// current process
		Process* getCurrentProcess() { return env -> getCurrentProcess(); }
		/// current scheduled object
		Sched* getCurrentSched() { return env -> getCurrentSched(); }

		ProcessQueue* getAcquireWait() { return &acquireWait; }

		/// simulation
		Simulation* env;
		/// current number of tokens
		unsigned int tokenNumber;
		/// maximum number of tokens
		unsigned int tokenLimit;

		// Statistics
		unsigned int initToken; ///< init token number
		Accum tokenStatistics; ///< token statistics

		unsigned int users;
		unsigned int providers;
		double sumWaitTime;

		/// token storage
		std::list< Token > tokenStore;
		/// token management
		std::map< Token*, Process* > tokenInfo;
		/// process management
		Queue acquireWait;
	};

	/** \interface ResTBaseObserver

		\author Ronald Kluth

		\brief Observer for ResT specific events

		<!-- [\note {notes}]* -->

		\sa ResT

		<!-- [detailed description] -->

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		<!-- [\since {since when in project}] -->
	*/
	template <typename Token>
	class ResTBaseObserver {
	public:
		virtual ~ResTBaseObserver() {}

		/// Creation
		virtual void onCreate(ResTBase< Token >* sender) = 0;
		/// Destruction
		virtual void onDestroy(ResTBase< Token >* sender) = 0;

		/// Failed attempt to acquire n token
		virtual void onAcquireFail(ResTBase< Token >* sender, unsigned int n) = 0;
		/// Successful attempt to acquire n token
		virtual void onAcquireSucceed(ResTBase< Token >* sender, unsigned int n) = 0;
		/// Failed attempt to release n token
		virtual void onReleaseFail(ResTBase< Token >* sender, unsigned int n) = 0;
		/// Release of n token
		virtual void onReleaseSucceed(ResTBase< Token >* sender, unsigned int n) = 0;

		/// Change of token number
		virtual void onChangeTokenNumber(ResTBase< Token >* sender, unsigned int oldTokenNumber,
			unsigned int newTokenNumber) = 0;
	};

}

#endif /*ODEMX_RES_T_BASE_INCLUDED*/
