//----------------------------------------------------------------------------
//	Copyright (C) 2002-2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ResTChoice.cpp

	\author Ronald Kluth

	\date created at 2008/02/18

	\brief Definition of odemx::ResTChoice

	\sa ResTChoice.h

	<!-- [detailed description] -->

	\since 2.1
*/

using namespace odemx;
using namespace std;

template < typename Token >
ResTChoice< Token >::ResTChoice( Label l,
					 vector< Token >& initialResources,
					 ResTChoiceObserver< Token >* o /* = 0*/ ):
	ResTBase< Token >( l, initialResources, o ),
	Observable< ResTChoiceObserver< Token > >( o )
	{
		// observer
		_obsForEachT(ResTChoiceObserver<Token>, Create(this));
	}


template < typename Token >
ResTChoice< Token >::ResTChoice( Simulation* sim,
					 Label l,
					 vector< Token >& initialResources,
					 ResTChoiceObserver< Token >* o /* = 0*/ ):
	ResTBase< Token >( sim, l, initialResources, o ),
	Observable< ResTChoiceObserver< Token > >( o )
	{
		// observer
		_obsForEachT(ResTChoiceObserver<Token>, Create(this));
	}

template < typename Token >
ResTChoice<Token>::~ResTChoice() {
	// observer
	_obsForEachT(ResTChoiceObserver<Token>, Destroy(this));
}

template < typename Token >
Token* ResTChoice< Token >::acquire( typename ResTChoice< Token >::Selection select ) {

	vector< Token* > found = acquire( select, 1 );

	if( ! found.empty() )
		return found.front();

	return 0;
}

template < typename Token >
const vector< Token* > ResTChoice< Token >::acquire( typename ResTChoice< Token >::Selection select, unsigned int n ) {

	if( n > ResTBase< Token >::tokenLimit ) {
		// warning: will always block
		warning( this, "ResTChoice::acquire(): requesting more than max resources will always block" );
	}

	if( ! getCurrentSched() ||
		getCurrentSched() -> getSchedType() != Sched::PROCESS ) {
		// error: resource handling only implemented for processes
		error( this, "ResTChoice::acquire(): called by non-Process object" );
		return vector< Token* >();
	}

	// compute order of service
	ResTBase< Token >::acquireWait.inSort( getCurrentProcess() );

	// the return vector to be filled by getMatchingTokens()
	vector< Token* > matches;

	if( ResTBase< Token >::acquireWait.getTop() != getCurrentProcess() ||
		! getMatchingTokens( select, n, matches ) ) {
		// not enough matching tokens or
		// not the first process in queue

		// trace
		getTrace()->beginMark(this, ResTBase<Token>::markAcquireFail);
		getTrace()->addTag(ResTBase<Token>::tagCurrent, getCurrentProcess());
		getTrace()->addTag(ResTBase<Token>::tagTokenNumber, n);
		getTrace()->endMark();

		// observer
		_obsForEachT(ResTChoiceObserver<Token>, AcquireFail(this, n));

		// statistics
		SimTime waitTime = getTime();

		do {

			getCurrentProcess() -> sleep();

			// handle interrupt
			if( getCurrentProcess() -> isInterrupted() ) {

				ResTBase< Token >::acquireWait.remove( getCurrentProcess() );
				return vector< Token* >();
			}
		}
		while( ResTBase< Token >::acquireWait.getTop() != getCurrentProcess() ||
				 ! getMatchingTokens( select, n, matches ) );

		// block released
		// statistics
		ResTBase<Token>::sumWaitTime += getTime() - waitTime;
	}


	// trace
	getTrace()->mark(this, ResTBase<Token>::markChangeTokenNumber, ResTBase<Token>::tokenNumber-n, ResTBase<Token>::tokenNumber);

	// observer
	_obsAForEachT(ResTChoiceObserver<Token>, TokenNumber, ResTBase< Token >::tokenNumber, ResTBase< Token >::tokenNumber-n);

	// acquire token
	ResTBase< Token >::tokenNumber -= n;

	// fill return vector with pointers to free resource objects
	typename vector< Token* >::iterator i;

	for( i = matches.begin(); i != matches.end(); ++i ) {
		// store the current user of selected resources in ResTBase< Token >::tokenInfo

		if( ResTBase< Token >::tokenInfo[ *i ] == 0 ) {
			ResTBase< Token >::tokenInfo[ *i ] = getCurrentProcess();
		}
		else {
			// this error should never happen
			error( this, "ResTChoice::acquire(): selected token is owned by another Process" );
		}
	}

	// trace
	getTrace()->beginMark(this, ResTBase<Token>::markAcquireSucceed);
	getTrace()->addTag(ResTBase<Token>::tagCurrent, getCurrentProcess());
	getTrace()->addTag(ResTBase<Token>::tagTokenNumber, n);
	getTrace()->endMark();

	// observer
	_obsForEachT(ResTChoiceObserver<Token>, AcquireSucceed(this, n));

	// statistics
	++ResTBase<Token>::users;
	ResTBase<Token>::tokenStatistics.update(ResTBase<Token>::tokenNumber);

	// remove from list
	ResTBase< Token >::acquireWait.remove( getCurrentProcess() );

	// awake next
	awakeFirst( ResTBase< Token >::getAcquireWait() );

	return matches;
}


template < typename Token >
bool ResTChoice< Token >::getMatchingTokens( typename ResTChoice< Token >::Selection select,
									   unsigned int n,
									   vector< Token* >& result )
{
	unsigned int count = 0;
	result.clear();

	// find matching tokens
	typename std::map< Token*, Process* >::iterator i;
	for( i = ResTBase< Token >::tokenInfo.begin(); i != ResTBase< Token >::tokenInfo.end(); ++i ) {

		// count all matches
		if( select( getCurrentProcess(), i -> first ) ) {
			count++;

			// store free matches in result vector
			if( i -> second == 0 ) {
				result.push_back( i -> first );

				if( result.size() == n )
					break;
			}
		}
	}

	if( count < n ) {
		warning( this, "ResTChoice::getMatchingTokens(): too few matching tokens in resource, will always block" );
	}

	return result.size() == n;
}

