//----------------------------------------------------------------------------
//	Copyright (C) 2002-2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ResTChoice.h

	\author Ronald Kluth

	\date created at 2008/02/18

	\brief Declaration of odemx::ResTChoice and observer

	\sa ResTChoice.cpp

	<!-- [detailed description] -->

	\since 2.1
*/

#ifndef ODEMX_RES_T_CHOICE_INCLUDED
#define ODEMX_RES_T_CHOICE_INCLUDED

namespace odemx {
	//forward declaration
	template < typename Token >
	class ResTChoiceObserver;

	/** \class ResTChoice

		\author Ronald Kluth

		\brief ResTChoice is an abstract base class for resources holding
		objects of type Token

		ResTChoice is a resource class for synchronizing Process objects based
		on the availability of selected resources. A process is blocked in
		acquire() if not enough tokens matching the selection criteria
		are available. It is reactivated when enough tokens are released
		to the resource. If multiple process objects are waiting for
		reactivation, process queue priority and FIFO-strategy are used to
		choose the process.

		\note ResTChoice supports Observation
		\note ResTChoice supports Trace

		\sa Res, ResT

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		<!-- [\since {since when in project}] -->
	*/
	template < typename Token >
	class ResTChoice: public ResTBase< Token >,
					  public Observable< ResTChoiceObserver< Token > > {
	public:
		/// function type for coding selections
		typedef bool( *Selection )( Process* owner, Token* t );

		/**
		 * \copydoc ResTBase(Label,std::vector<Token>&,ResTBaseObserver<Token>*)
		 */
		ResTChoice( Label l, std::vector< Token >& initialResources,
					ResTChoiceObserver< Token >* o = 0 );

		/**
		 * \copydoc ResTBase(Simulation*,Label,std::vector<Token>&,ResTBaseObserver<Token>*)
		 */
		ResTChoice( Simulation* sim, Label l, std::vector< Token >& initialResources,
					ResTChoiceObserver< Token >* o = 0 );

		/// Destruction
		virtual ~ResTChoice();

		/**
			\name Token usage

			These functions provade access for resource usage by Processes.
			Tokens can be requested and thus acquired when then are
			available. When the tokens are not needed anymore, the owning
			Process	can release them.

			@{
		*/
		/**
			\brief acquire a token

			Required implementation of acquire, will output an error
			to
		*/
		virtual Token* acquire() {

			error( this, "ResTChoice::acquire(): parameter selection required, use overload" );
			return 0;
		}


		/**
			\brief find and acquire a token

			If no token matching the Selection condition is available,
			the current process is blocked
		*/
		Token* acquire( typename ResTChoice< Token >::Selection select );

		/**
			\brief find and acquire token

			If there are not enough tokens matching the Selection condition
			available, the current process is blocked
		*/
		const std::vector< Token* > acquire( typename ResTChoice< Token >::Selection select, unsigned int n );

	private:

		/// get pointer to the trace object
		virtual Trace* getTrace() const {
			return (Trace*) ResTBase< Token >::env;
		}

		/// current simulation time
		SimTime getTime() {
			return ResTBase< Token >::env -> getTime();
		}

		/// currently executed Process
		Process* getCurrentProcess() {
			return ResTBase< Token>::env -> getCurrentProcess();
		}

		/// currently executed Event
		Sched* getCurrentSched() {
			return ResTBase< Token>::env -> getCurrentSched();
		}

		/// help method to find tokens matching the selection
		bool getMatchingTokens( typename ResTChoice<Token>::Selection select,
								unsigned int n,
								std::vector< Token* >& result );
	};

	/** \interface ResTChoiceObserver

		\author Ronald Kluth

		\brief Observer for ResT specific events

		<!-- [\note {notes}]* -->

		\sa ResT

		<!-- [detailed description] -->

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		<!-- [\since {since when in project}] -->
	*/
	template < typename Token >
	class ResTChoiceObserver: public ResTBaseObserver< Token > {
	public:
		virtual ~ResTChoiceObserver() {}

		/// Creation
		virtual void onCreate(ResTBase< Token >* sender) {}
		/// Destruction
		virtual void onDestroy(ResTBase< Token >* sender) {}

		/// Failed attempt to acquire n token
		virtual void onAcquireFail(ResTBase< Token >* sender, unsigned int n) {}
		/// Successful attempt to acquire n token
		virtual void onAcquireSucceed(ResTBase< Token >* sender, unsigned int n) {}
		/// Failed to release of n token
		virtual void onReleaseFail(ResTBase< Token >* sender, unsigned int n) {}
		/// Successful release of n token
		virtual void onReleaseSucceed(ResTBase< Token >* sender, unsigned int n) {}

		/// Change of token number
		virtual void onChangeTokenNumber(ResTBase< Token >* sender, unsigned int oldTokenNumber,
			unsigned int newTokenNumber) {}
	};

}


#endif /*ODEMX_RES_T_CHOICE_INCLUDED*/
