/**
 * @file Globals.h
 *
 * @brief Declarations of global helper functions and classes
 *
 * This file contains declarations of global helper functions which assist
 * testing. These function are implemented in ODEMxUnitTest.cpp.
 *
 */

#ifndef GLOBALS_H_
#define GLOBALS_H_

#include <odemx/util/LabeledObject.h>
#include <odemx/util/MarkType.h>
#include <odemx/util/TypedObject.h>
#include <iostream>
#include <ostream>
#include <string>
#include <vector>

using odemx::LabeledObject;
using odemx::MarkType;
using odemx::typeToString;

/**
 * @brief Get error stream to examine error and warning output.
 *
 * For testing purposes, the error stream \c cerr is redirected into an
 * \c ostringstream, which can be examined after running a test to see
 * how a function call responded to erroneous input. It is used to
 * check ODEMx warnings and error messages.
 *
 * @return Reference to a static error stream object
 */
std::ostringstream& errorStream();

/**
 * @brief Clear the error stream buffer
 *
 * This function is used to empty the error stream's buffer before running
 * tests which produce warnings or error messages, it allows for using
 * string comparisons after tests to check for the expected output message.
 */
void clearErrorStream();

/**
 * @brief Helper to transform any given data into a string
 * @param any data to be transformed
 * @return string representation of \c any
 * @note only works if T defines op<<() for string streams
 */
template< typename T >
inline std::string toString( const T& any )
{
	std::ostringstream convert;
	convert << any;
	return convert.str();
}

/**
 * @struct Observation Globals.h
 *
 * @brief Type holding information about observed simulation events
 *
 * This type stores four public string members describing a simulation event.
 * If the event does not require any descriptive details, these two strings
 * will be empty. The members of this struct may need to be updated if
 * ODEMx observer functions have more than 4 parameters.
 */
struct Observation
{
	std::string event,		///< observed event name
				sender,		///< event sender
				detail1,	///< event detail 1
				detail2;	///< event detail 2
	/**
	 * @brief Constructor
	 * @param e event name, usually the name of the called Observer function
	 * @param s sender name, the label of the observed object
	 * @param d1 event detail, such as partner objects, time, attributes
	 * @param d2 event detail, such as partner objects, time, attributes
	 */
	Observation( const std::string& e, const std::string& s,
				 const std::string& d1, const std::string& d2 ):
		event( e ),
		sender( s ),
		detail1( d1 ),
		detail2( d2 )
		{}
};

/**
 * @class ObserverBase Globals.h
 *
 * @brief Provides basic functionality for all test observer classes
 *
 * This class provides basic functionality for all observer classes used during
 * unit testing. Its primary purpose is to store all observations in a history,
 * making it possible to verify the occurrence of all expected events during
 * calls of tested functions.
 */
class ObserverBase
{
public:
	/// History vector storing Observation objects
	std::vector< Observation > history;
	/// Boolean value to switch console output on and off
	bool writeToConsole;

	/**
	 * @brief Constructor
	 * @param output set to \c true to output observed events to the console
	 */
	ObserverBase( bool output ):
		writeToConsole( output )
		{}

	/**
	 * @brief Get the last observation
	 * @return pointer to the last Observation, or \c 0 if history is empty
	 */
	Observation* lastObservation()
	{
		return history.empty() ? 0 : &( history.back() );
	}

	/// Write last observation to stdout
	void print()
	{
		if( lastObservation() )
		{
			std::cout << lastObservation() -> sender
				 << ": " << lastObservation() -> event;

			if( ! lastObservation() -> detail1.empty() )
				std::cout << " -- " << lastObservation() -> detail1;

			if( ! lastObservation() -> detail2.empty() )
				std::cout << " -- " << lastObservation() -> detail2;

			std::cout << std::endl;
		}
		else
		{
			std::cout << "???" << std::endl;
		}
	}

	/// Add new Observation to history
	void saw( const std::string& event, const LabeledObject* sender,
			  const std::string& det1 = "", const std::string& det2 = "" )
	{
		history.push_back( Observation( event,
				sender ? sender -> getLabel() : "0",
				det1, det2 ) );
		if( writeToConsole ) print();
	}

	/// Add new Observation to history
	void saw( const std::string& event, const LabeledObject* sender,
			  const LabeledObject* partner1, const LabeledObject* partner2 = 0 )
	{
		history.push_back( Observation( event,
				sender ? sender -> getLabel() : "0",
				partner1 ? partner1 -> getLabel() : "",
				partner2 ? partner2 -> getLabel() : "" ) );
		if( writeToConsole ) print();
	}

	/**
	 * @brief Determine whether an expected event was observed
	 * @return \c true if found in history, \c false otherwise
	 */
	bool observed( const std::string& event,
				   const std::string& sender,
				   const std::string& detail1 = "",
				   const std::string& detail2 = "" )
	{
		std::vector< Observation >::reverse_iterator i;
		for( i = history.rbegin(); i != history.rend(); ++i )
		{
			// check for matching event
			if( ( *i ).event == event && ( *i ).sender == sender )
			{
				// skip this event if the details do not match
				if( ! detail1.empty() &&
					( *i ).detail1 != "" &&
					( *i ).detail1 != detail1 )
					continue;

				if( ! detail2.empty() &&
					( *i ).detail2 != "" &&
					( *i ).detail2 != detail2 )
					continue;

				// only reached if the details match too
				return true;
			}
		}
		return false;
	}
};

/**
 *
 */
struct TraceMark
{
	std::string sender,
				markType,
				markId,
				markScope,
				detail1,
				detail2,
				comment;

	TraceMark( const std::string& s, const MarkType& m,
			   const std::string& det1, const std::string& det2,
			   const std::string& com )
	:	sender( s ),
		markType( m.getName() ),
		markId( toString( m.getId() ) ),
		markScope( typeToString( m.getScope() ) ),
		detail1( det1 ),
		detail2( det2 ),
		comment( com )
		{}
};

/**
 * @class TraceBase Globals.h
 *
 * @brief Provides basic functionality for test trace consumer classes
 *
 * This class provides basic functionality for all trace consumer classes used
 * during unit testing. Its primary purpose is to store all trace marks in a
 * history, making it possible to verify the occurrence of all expected events
 * during calls of tested functions.
 */
class TraceBase
{
public:
	/// History vector storing TraceMark objects
	std::vector< TraceMark > history;
	/// Boolean value to switch console output on and off
	bool writeToConsole;

	/**
	 * @brief Constructor
	 * @param output set to \c true to output observed events to the console
	 */
	TraceBase( bool output ):
		writeToConsole( output )
		{}

	/**
	 * @brief Get the last mark
	 * @return pointer to the last TraceMark, or \c 0 if history is empty
	 */
	TraceMark* lastMark()
	{
		return history.empty() ? 0 : &( history.back() );
	}

	/// Write last mark to stdout
	void print()
	{
		if( lastMark() )
		{
			std::cout
			<< lastMark() -> sender
			<< " (" << lastMark() -> markType
			<< ", " << lastMark() -> markId
			<< ", " << lastMark() -> markScope
			<< ")";

			if( ! lastMark() -> detail1.empty() )
				std::cout << " -- " << lastMark() -> detail1;

			if( ! lastMark() -> detail2.empty() )
				std::cout << " -- " << lastMark() -> detail2;

			if( ! lastMark() -> comment.empty() )
				std::cout << " -- " << lastMark() -> comment;

			std::cout << std::endl;
		}
		else
		{
			std::cout << "???" << std::endl;
		}
	}

	/// Add new TraceMark to history
	void mark( TraceMark tm )
	{
		history.push_back( tm );
		if( writeToConsole ) print();
	}

	/**
	 * @brief Determine whether an expected trace mark was received
	 * @return \c true if found in history, \c false otherwise
	 */
	bool received( const std::string& sender,
				   const std::string& markType,
				   const std::string& markId,
				   const std::string& markScope,
				   const std::string& detail1 = "",
				   const std::string& detail2 = "",
				   const std::string& comment = "" )
	{
		std::vector< TraceMark >::reverse_iterator i;
		for( i = history.rbegin(); i != history.rend(); ++i )
		{
			const TraceMark& current = *i;

			if( current.sender != sender )
				continue;
			if( current.markType != markType )
				continue;
			if( current.markId != markId )
				continue;
			if( current.markScope != markScope )
				continue;
			if( current.detail1 != "" && current.detail1 != detail1 )
				continue;
			if( current.detail2 != "" && current.detail2 != detail2 )
				continue;
			if( current.comment != "" && current.comment != comment )
				continue;

			// only reached if the details match too
			return true;
		}
		return false;
	}

};

#endif /*GLOBALS_H_*/
