/**
 * @file TestEvent.cpp
 * @brief Tests for ODEMx class Event
 */

#include "TestBase.h"

/// @cond DOXYGEN_SKIP
SUITE( Base )
{
/// @endcond

	/**
	 * @struct EventFixture
	 * @brief Helper struct providing set-up/tear-down of Event tests
	 *
	 * Structs like this are called fixtures. They are used to provide a
	 * mechanism for quickly setting up multiple tests that use the same
	 * variables. Since these variables are declared as public members
	 * and the UnitTest++ TEST_FIXTURE macro creates classes derived from
	 * their fixture struct, it is possible to use these variables directly
	 * in every test case.
	 *
	 * The constructor provides the set-up of a test's variables while the
	 * destructor handles the tear-down phase. This means that for every
	 * test case, a new object is created, which encapsulates all objects
	 * used during the test.
	 *
	 * For ODEMx tests most of these fixtures contain their own Simulation
	 * object because the DefaultSimulation is a static object and only
	 * exists once for the whole test program, i.e. all suites and tests.
	 * To avoid likely naming conflicts, scheduling problems and interference
	 * from different test simulations, a new context is advisable for each
	 * test.
	 */
	struct EventFixture
	{
		EventFixture():
			sim( "SimulationTest" ),
			observer( false ),
			eventObj( "EventTestObject" ),
			eventPtr( new EventTest( &sim, "EventTestPointer", &observer ) )
			{}

		~EventFixture()
		{
			if( eventPtr ) delete eventPtr;
		}

		SimulationTest sim; ///< a user-defined simulation
		EventTestObserver observer; ///< an observer for Event-specific calls
		EventTest eventObj; ///< an Event object
		EventTest* eventPtr; ///< pointer to a dynamically allocated Event object
	};

	/**
	 * @test odemx::Event construction and destruction
	 *
	 * Tests correct Event member initialization:
	 * @li Simulation context
	 * @li SchedType
	 * @li Label
	 * @li execution time
	 * @li scheduling priority
	 * @li observation calls of \c onCreate() and \c onDestroy()
	 *
	 * @note One object uses the constructor for DefaultSimulation, while
	 * the other one uses a given Simulation object.
	 */
	TEST_FIXTURE( EventFixture, ConstructionDestruction )
	{
		CHECK( eventObj.getSimulation() == getDefaultSimulation() );
		CHECK( eventObj.getSchedType() == Sched::EVENT );
		CHECK( eventObj.getLabel() == "EventTestObject" );
		CHECK( eventObj.getExecutionTime() == 0.0 );
		CHECK( eventObj.getPriority() == 0.0 );

		CHECK( eventPtr -> getSimulation() == &sim );
		CHECK( eventPtr -> getSchedType() == Sched::EVENT );
		CHECK( eventPtr -> getLabel() == "EventTestPointer" );
		CHECK( eventPtr -> getExecutionTime() == 0.0 );
		CHECK( eventPtr -> getPriority() == 0.0 );

		// only the eventPtr is being observed here
		CHECK( observer.observed( "onCreate", eventPtr -> getLabel() ) );

		EventTest* e = new EventTest( "EventTestCreationDestructionObservation", &observer );
		CHECK( observer.lastObservation() -> event == "onCreate" );

		delete e; e = 0;
		CHECK( observer.lastObservation() -> event == "onDestroy" );
	}

	/**
	 * @test odemx::Event::schedule()
	 *
	 * Expected function call effects:
	 * @li Simulation is not empty, i.e. an object is scheduled in ExecutionList
	 * @li the Event object knows it is scheduled
	 * @li the Event execution time is equal to the current SimTime
	 * @li observation calls of \c onSchedule() and \c onChangeExecutionTime()
	 */
	TEST_FIXTURE( EventFixture, Schedule )
	{
		CHECK( sim.isEmpty() );

		// scheduling events always sets/changes the execution time
		eventPtr -> schedule();
		CHECK( ! sim.isEmpty() );
		CHECK( eventPtr -> isScheduled() );
		CHECK( eventPtr -> getExecutionTime() == sim.getTime() );
		CHECK( observer.observed( "onSchedule", eventPtr -> getLabel() ) );
		CHECK( observer.observed( "onChangeExecutionTime", eventPtr -> getLabel() ) );
	}

	/**
	 * @test odemx::Event::scheduleIn(SimTime)
	 *
	 * Expected function call effects:
	 * @li Simulation is not empty, i.e. an object is scheduled in ExecutionList
	 * @li the Event object knows it is scheduled
	 * @li the Event execution time is equal to the current SimTime + delay
	 * @li observation calls of \c onScheduleIn() and \c onChangeExecutionTime()
	 */
	TEST_FIXTURE( EventFixture, ScheduleIn )
	{
		// set and execute an event to advance simulation time
		// to be able to check the correct process delay
		EventTest e( &sim, "SimAdvanceEvent" );
		SimTime advanceTime = 15;
		e.scheduleAt( advanceTime );
		sim.step();
		CHECK( sim.getTime() == advanceTime );

		SimTime delay = 512;
		SimTime oldTime = eventPtr -> getExecutionTime();
		CHECK( sim.isEmpty() );

		// scheduling events always sets/changes the execution time
		eventPtr -> scheduleIn( delay );
		CHECK( ! sim.isEmpty() );
		CHECK( eventPtr -> isScheduled() );
		CHECK_EQUAL( sim.getTime() + delay, eventPtr -> getExecutionTime() );
		CHECK( observer.observed( "onScheduleIn", eventPtr -> getLabel() ) );
		CHECK( observer.observed( "onChangeExecutionTime", eventPtr -> getLabel(),\
				toString( oldTime ), toString( sim.getTime() + delay ) ) );
	}

	/**
	 * @test odemx::Event::scheduleAt(SimTime)
	 *
	 * Expected function call effects:
	 * @li Simulation is not empty, i.e. an object is scheduled in ExecutionList
	 * @li the Event object knows it is scheduled
	 * @li the Event execution time is equal to the given absolute time
	 * @li observation calls of \c onScheduleAt() and \c onChangeExecutionTime()
	 */
	TEST_FIXTURE( EventFixture, ScheduleAt )
	{
		SimTime absoluteTime = 27;
		SimTime oldTime = eventPtr -> getExecutionTime();
		CHECK( sim.isEmpty() );

		// scheduling events always sets/changes the execution time
		eventPtr -> scheduleAt( absoluteTime );
		CHECK( ! sim.isEmpty() );
		CHECK( eventPtr -> isScheduled() );
		CHECK_EQUAL( absoluteTime, eventPtr -> getExecutionTime() );
		CHECK( observer.observed( "onScheduleAt", eventPtr -> getLabel() ) );
		CHECK( observer.observed( "onChangeExecutionTime", eventPtr -> getLabel(),\
				toString( oldTime ), toString( absoluteTime ) ) );
	}

	/**
	 * @test odemx::Event::scheduleAppend()
	 *
	 * Expected function call effects:
	 * @li the Event object knows it is scheduled
	 * @li the Event execution time is equal to the current SimTime
	 * @li the Event is scheduled after another Event
	 * @li observation calls of \c onScheduleAppend() and \c onChangeExecutionTime()
	 */
	TEST_FIXTURE( EventFixture, ScheduleAppend )
	{
		EventTest e( &sim, "ScheduleAppendTestEvent" );
		e.schedule();

		// scheduling events always sets/changes the execution time
		eventPtr -> scheduleAppend();
		CHECK( eventPtr -> isScheduled() );
		CHECK( eventPtr -> getExecutionTime() == sim.getTime() );
		CHECK( observer.observed( "onScheduleAppend", eventPtr -> getLabel() ) );
		CHECK( observer.observed( "onChangeExecutionTime", eventPtr -> getLabel() ) );

		// check correct order
		CHECK_EQUAL( sim.getNextSched(), &e );
		sim.step();
		CHECK( !e.isScheduled() );
		CHECK_EQUAL( sim.getNextSched(), eventPtr );
	}

	/**
	 * @test odemx::Event::scheduleAppendIn(SimTime)
	 *
	 * Expected function call effects:
	 * @li the Event object knows it is scheduled
	 * @li the Event execution time is equal to the current SimTime + delay
	 * @li the Event is scheduled after another Event
	 * @li observation calls of \c onScheduleAppendIn() and \c onChangeExecutionTime()
	 */
	TEST_FIXTURE( EventFixture, ScheduleAppendIn )
	{
		EventTest e( &sim, "ScheduleAppendInTestEvent" );
		e.schedule();

		SimTime delay = 512;
		SimTime oldTime = eventPtr -> getExecutionTime();

		// scheduling events always sets/changes the execution time
		eventPtr -> scheduleAppendIn( delay );
		CHECK( eventPtr -> isScheduled() );
		CHECK_EQUAL( sim.getTime() + delay, eventPtr -> getExecutionTime() );
		CHECK( observer.observed( "onScheduleAppendIn", eventPtr -> getLabel() ) );
		CHECK( observer.observed( "onChangeExecutionTime", eventPtr -> getLabel(),\
				toString( oldTime ), toString( sim.getTime() + delay ) ) );

		// check correct order
		CHECK_EQUAL( sim.getNextSched(), &e );
		sim.step();
		CHECK( !e.isScheduled() );
		CHECK_EQUAL( sim.getNextSched(), eventPtr );
	}

	/**
	 * @test odemx::Event::scheduleAppendAt(SimTime)
	 *
	 * Expected function call effects:
	 * @li the Event object knows it is scheduled
	 * @li the Event execution time is equal to the given absolute time
	 * @li the Event is scheduled after another Event
	 * @li observation calls of \c onScheduleAppendAt() and \c onChangeExecutionTime()
	 */
	TEST_FIXTURE( EventFixture, ScheduleAppendAt )
	{
		EventTest e( &sim, "ScheduleAppendInTestEvent" );
		e.schedule();

		SimTime absoluteTime = 15;
		SimTime oldTime = eventPtr -> getExecutionTime();

		// scheduling events always sets/changes the execution time
		eventPtr -> scheduleAt( absoluteTime );
		CHECK( eventPtr -> isScheduled() );
		CHECK_EQUAL( absoluteTime, eventPtr -> getExecutionTime() );
		CHECK( observer.observed( "onScheduleAt", eventPtr -> getLabel() ) );
		CHECK( observer.observed( "onChangeExecutionTime", eventPtr -> getLabel(),\
				toString( oldTime ), toString( absoluteTime ) ) );

		// check correct order
		CHECK_EQUAL( sim.getNextSched(), &e );
		sim.step();
		CHECK( !e.isScheduled() );
		CHECK_EQUAL( sim.getNextSched(), eventPtr );
	}

	/**
	 * @test odemx::Event::removeFromSchedule()
	 *
	 * Expected function call effects:
	 * @li Simulation gets empty, i.e. the scheduled object is removed from the
	 * ExecutionList
	 * @li the Event object knows it is not scheduled
	 * @li observation call of \c onRemoveFromSchedule()
	 * @li error output when called on non-scheduled event
	 */
	TEST_FIXTURE( EventFixture, RemoveFromSchedule )
	{
		eventPtr -> schedule();
		CHECK( ! sim.isEmpty() );

		eventPtr -> removeFromSchedule();
		CHECK( sim.isEmpty() );
		CHECK( ! eventPtr -> isScheduled() );
		CHECK( observer.observed( "onRemoveFromSchedule", eventPtr -> getLabel() ) );

		clearErrorStream();
		EventTest e( "RemoveFromScheduleErrorEvent" );
		e.removeFromSchedule();
		CHECK_EQUAL( "ODEMx ERROR: Event::removeFromSchedule(): Event not scheduled -- Object: RemoveFromScheduleErrorEvent\n", \
				errorStream().str() );
	}

	/**
	 * @test odemx::Event::execute()
	 *
	 * Expected function call effects:
	 * @li \c eventAction() was executed
	 * @li the Event is removed from the schedule (ExecutionList)
	 * @li the Event knows it is not scheduled
	 * @li observation call of \c onExecuteEvent()
	 */
	TEST_FIXTURE( EventFixture, Execute )
	{
		eventPtr -> schedule(); // must be scheduled for execution, else fatal error
		eventPtr -> execute();
		CHECK( eventPtr -> executed );
		CHECK( sim.isEmpty() );
		CHECK( ! eventPtr -> isScheduled() );
		CHECK( observer.observed( "onExecuteEvent", eventPtr -> getLabel() ) );
	}

	/**
	 * @test odemx::Event::setPriority()
	 *
	 * Expected function call effects:
	 * @li return of old Event priority
	 * @li observation call of \c onChangePriority()
	 */
	TEST_FIXTURE( EventFixture, Priority )
	{
		Priority newPriority = 2;
		Priority oldPriority = eventPtr -> setPriority( newPriority );
		CHECK_EQUAL( newPriority, eventPtr -> getPriority() );
		CHECK( observer.observed( "onChangePriority", eventPtr -> getLabel(),\
				toString( oldPriority ), toString( newPriority) ) );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
