/**
 * @file TestSimulation.cpp
 * @brief Tests for ODEMx class Simulation
 */

#include "TestBase.h"

/// @cond DOXYGEN_SKIP
SUITE( Base )
{
/// @endcond

	/**
	 * @struct SimulationFixture
	 * @brief Helper struct providing set-up/tear-down of Simulation tests
	 *
	 * @copydetails EventFixture
	 */
	struct SimulationFixture
	{
		SimulationFixture():
			observer( false ),
			sim( "SimulationTest", &observer ),
			event( &sim, "EventTest" ),
			process( &sim, "ProcessTest" ),
			process2( &sim, "ProcessTest" ),
			process3( &sim, "ProcessTest" )
			{}

		SimulationTestObserver observer; ///< an onserver for monitoring Simulation calls
		SimulationTest sim; ///< a Simulation test object
		EventTest event; ///< a helper Event object
		ProcessTest process; ///< a helper Process object
		ProcessTest process2; ///< a helper Process object
		ProcessTest process3; ///< a helper Process object
	};

	/**
	 * @test odemx::Simulation construction and destruction
	 *
	 * Tests correct member initialization:
	 * @li no current process
	 * @li no current Sched
	 * @li start time
	 * @li current time equals start time
	 * @li observation calls of \c onCreate() and \c onDestroy()
	 */
	TEST_FIXTURE( SimulationFixture, ConstructionDestruction )
	{
		CHECK( sim.getCurrentProcess() == 0 );
		CHECK( sim.getCurrentSched() == 0 );
		CHECK( sim.getStartTime() == 0 );
		CHECK_EQUAL( sim.getTime(), sim.getStartTime() );

		SimTime startAt = 10000;
		SimulationTestObserver* o = new SimulationTestObserver();
		Simulation* s = new SimulationTest( "SimulationTest2", startAt, o );
		CHECK_EQUAL( startAt, s -> getStartTime() );
		CHECK_EQUAL( sim.getTime(), sim.getStartTime() );
		CHECK( o -> lastObservation() -> event == "onCreate" );
		CHECK( o -> lastObservation() -> sender == "SimulationTest2" );

		delete s;
		CHECK( o -> lastObservation() -> event == "onDestroy" );
		CHECK( o -> lastObservation() -> sender == "SimulationTest2" );
		delete o;
	}

	/**
	 * @test odemx::Simulation::setProcessAsXXX, management of Process lists
	 *
	 * A Simulation keeps track of all Processes created in its context.
	 * These objects are managed in lists according to their ProcessState,
	 * which can be CREATED, RUNNABLE, IDLE, TERMINATED, or CURRENT. Since
	 * there is only one current Process at a certain point in time, only
	 * the other four states are managed.
	 *
	 * In this test, a Process is walked through its states by calling the
	 * corresponding member functions, which in turn will move the object
	 * from one list to the next.
	 *
	 * Expected effect after each call:
	 * @li initially, \c process is in the list of created processes
	 * @li after \c activate(), it is in the list of runnable processes
	 * @li after \c sleep(), it is in the list of idle processes
	 * @li after \c interrupt(), it is in the list of runnable processes
	 * @li after running the Simulation, it is in the list of terminated
	 * processes
	 */
	TEST_FIXTURE( SimulationFixture, SetProcessAsXXX )
	{
		// process should be in the list of created processes
		CHECK( observer.observed( "onChangeProcessList", sim.getLabel(), "ProcessTest_2", "CREATED" ) );

		// move to list of runnable (scheduled) processes
		process.activate();
		CHECK( observer.observed( "onChangeProcessList", sim.getLabel(), "ProcessTest", "RUNNABLE" ) );

		// move to list of idle (not scheduled) processes
		process.sleep();
		CHECK( observer.observed( "onChangeProcessList", sim.getLabel(), "ProcessTest", "IDLE" ) );

		// move to list of runnable (scheduled) processes
		process.interrupt();
		CHECK( observer.observed( "onChangeProcessList", sim.getLabel(), "ProcessTest", "RUNNABLE" ) );

		// move to list of terminated processes by execution
		sim.run();
		CHECK( process.executed );
		CHECK( observer.observed( "onChangeProcessList", sim.getLabel(), "ProcessTest", "TERMINATED" ) );
	}

	/**
	 * @test odemx::Simulation::step()
	 *
	 * A Process and an Event are scheduled to be executed by the Simulation.
	 * Step is called twice to execute both objects.
	 *
	 * Expected function call effects after first call:
	 * @li the Simulation called the user-defined initialization function
	 * @li the Process was executed
	 * @li observation calls of \c onStep(), \c onInitialization(),
	 * \c onChangeTime(), \c onExecuteProcess(), \c onChangeCurrentProcess(),
	 * and two calls to \c onChangeProcessList(), i.e. RUNNABLE and TERMINATED
	 *
	 * Expected function call effects after second call:
	 * @li the Event was executed
	 * @li observation calls of \c onStep(),\c onChangeTime(), and
	 * \c onExecuteEvent()
	 * @li no calls of \c onInitialization(),\c onChangeCurrentProcess(),
	 * and \c onChangeProcessList()
	 */
	TEST_FIXTURE( SimulationFixture, Step )
	{
		observer.history.clear();
		process.activate();
		event.scheduleAppend();

		// check process execution and initialization
		sim.step();
		CHECK( sim.calledInit );
		CHECK( process.executed );
		CHECK( observer.observed( "onStep", sim.getLabel() ) );
		CHECK( observer.observed( "onInitialization", sim.getLabel() ) );
		CHECK( observer.observed( "onChangeTime", sim.getLabel(), "0", "0" ) );
		CHECK( observer.observed( "onExecuteProcess", sim.getLabel(), process.getLabel() ) );
		CHECK( observer.observed( "onChangeCurrentProcess", sim.getLabel(), "", process.getLabel() ) );
		CHECK( observer.observed( "onChangeProcessList", sim.getLabel(), process.getLabel(), "RUNNABLE" ) );
		CHECK( observer.observed( "onChangeProcessList", sim.getLabel(), process.getLabel(), "TERMINATED" ) );
		observer.history.clear();

		// check event execution
		sim.step();
		CHECK( event.executed );
		CHECK( ! observer.observed( "onInitialization", sim.getLabel() ) );
		CHECK( ! observer.observed( "onChangeCurrentProcess", sim.getLabel() ) );
		CHECK( ! observer.observed( "onChangeProcessList", sim.getLabel() ) );
		CHECK( observer.observed( "onStep", sim.getLabel() ) );
		CHECK( observer.observed( "onChangeTime", sim.getLabel(), "0", "0" ) );
		CHECK( observer.observed( "onExecuteEvent", sim.getLabel(), event.getLabel() ) );
	}

	/**
	 * @test odemx::Simulation::run()
	 *
	 * A Process is scheduled to be executed by the Simulation.
	 *
	 * Expected function call effects:
	 * @li the Simulation called the user-defined initialization function
	 * @li the Process was executed
	 * @li observation calls of \c onRun(), \c onInitialization(),
	 * \c onChangeTime(), \c onExecuteProcess(), \c onChangeCurrentProcess(),
	 * and two calls to \c onChangeProcessList(), i.e. RUNNABLE and TERMINATED
	 */
	TEST_FIXTURE( SimulationFixture, Run )
	{
		observer.history.clear();
		process.activateIn( 1 );
		sim.run();
		CHECK( sim.calledInit );
		CHECK( process.executed );
		CHECK( observer.observed( "onRun", sim.getLabel() ) );
		CHECK( observer.observed( "onInitialization", sim.getLabel() ) );
		CHECK( observer.observed( "onChangeTime", sim.getLabel(), "0", "1" ) );
		CHECK( observer.observed( "onExecuteProcess", sim.getLabel(), process.getLabel() ) );
		CHECK( observer.observed( "onChangeCurrentProcess", sim.getLabel(), "", process.getLabel() ) );
		CHECK( observer.observed( "onChangeProcessList", sim.getLabel(), process.getLabel(), "RUNNABLE" ) );
		CHECK( observer.observed( "onChangeProcessList", sim.getLabel(), process.getLabel(), "TERMINATED" ) );
	}

	/**
	 * @test odemx::Simulation::runUntil()
	 *
	 * Objects are scheduled to be executed by the Simulation at three
	 * different times:
	 * @li 499: \c process
	 * @li 500: \c event, process2
	 * @li 501: \c process3
	 *
	 * Expected function call effects after calling \c runUntil(499):
	 * @li the Simulation called the user-defined initialization function
	 * @li \c process was executed
	 * @li neither <tt>event, process2</tt> nor \c process3 were executed
	 * @li observation calls of \c onRunUntil(), \c onInitialization(),
	 * \c onChangeTime(), \c onExecuteProcess(), \c onChangeCurrentProcess(),
	 * and two calls to \c onChangeProcessList(), i.e. RUNNABLE and TERMINATED
	 *
	 * Expected function call effects after calling \c runUntil(500):
	 * @li \c event and \c process 2 were executed
	 * @li \c process3 was not executed
	 *
	 * Expected function call effects after calling \c runUntil(600):
	 * @li \c process3 was executed
	 * @li the Simulation is finished
	 */
	TEST_FIXTURE( SimulationFixture, RunUntil )
	{
		process.activateAt( 499 );
		event.scheduleAt( 500 );
		process2.activateAt( 500 );
		process3.activateAt( 501 );
		observer.history.clear();

		SimTime untilTime = 499;
		sim.runUntil( untilTime );
		CHECK( sim.calledInit );
		CHECK( process.executed );
		CHECK( ! process2.executed );
		CHECK( ! event.executed );
		CHECK( ! process3.executed );
		CHECK( observer.observed( "onRunUntil", sim.getLabel(), toString( untilTime ) ) );
		CHECK( observer.observed( "onInitialization", sim.getLabel() ) );
		CHECK( observer.observed( "onChangeTime", sim.getLabel(), "0", toString( untilTime ) ) );
		CHECK( observer.observed( "onExecuteProcess", sim.getLabel(), process.getLabel() ) );
		CHECK( observer.observed( "onChangeCurrentProcess", sim.getLabel(), "", process.getLabel() ) );
		CHECK( observer.observed( "onChangeProcessList", sim.getLabel(), process.getLabel(), "TERMINATED" ) );

		sim.runUntil( 500 );
		CHECK( process2.executed );
		CHECK( event.executed );
		CHECK( ! process3.executed );

		sim.runUntil( 600 );
		CHECK( process3.executed );
		CHECK( sim.isFinished() );
	}

	/**
	 * @test odemx::Simulation::exitSimulation()
	 *
	 * Precondition: a Process is scheduled.
	 *
	 * Expected function call effects:
	 * @li the Simulation is finished despite non-empty schedule
	 * @li observation call of \c onExitSimulation()
	 */
	TEST_FIXTURE( SimulationFixture, ExitSimulation )
	{
		observer.history.clear();
		CHECK( ! sim.isFinished() );

		process.activate();
		CHECK( ! sim.isEmpty() );

		sim.exitSimulation();
		CHECK( observer.observed( "onExitSimulation", sim.getLabel() ) );
		CHECK( sim.isFinished() );
	}

	/**
	 * @test odemx::DefaultSimulation-related functions
	 *
	 * This test calls functions which operate on the DefaultSimulation.
	 * Three Processes and an Event are created with their DefaultSimulation
	 * constructor and then executed by calling global ODEMx functions.
	 *
	 * Expected observations:
	 * @li getDefaultSimulation() returns a pointer to a DefaultSimulation object
	 * @li \c simulateStep() only executes one simulation step of the
	 * DefaultSimulation
	 * @li \c simulateUntil( untilTime ) executes all objects scheduled up to
	 * \c untilTime
	 * @li \c simulate() executes the remaining scheduled objects and
	 * finishes the DefaultSimulation
	 */
	TEST( DefaultSimulation )
	{
		CHECK( dynamic_cast< DefaultSimulation* >( getDefaultSimulation() ) );

		SimTime untilTime = 150;
		EventTest e1( "E1" );
		e1.scheduleAt( untilTime );
		ProcessTest p1( "P1" ), p2( "P2" ), p3( "P3" );
		p1.activate();
		p2.hold();
		p3.holdFor( untilTime + 75 );

		simulateStep();
		CHECK( p1.executed );
		CHECK( ! e1.executed );
		CHECK( ! p2.executed );
		CHECK( ! p3.executed );

		simulateUntil( untilTime );
		CHECK( p2.executed );
		CHECK( e1.executed );
		CHECK( ! p3.executed );

		simulate();
		CHECK( p3.executed );
		CHECK( getDefaultSimulation() -> isFinished() );
	}
}
