/**
 * @file TestHtmlReport.cpp
 * @date Aug 15, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class HtmlReport
 */

#include "TestDisplay.h"
#include "../TestBase/TestBase.h"

/// @cond DOXYGEN_SKIP
SUITE( Display )
{
/// @endcond

	/**
	 * @struct HtmlReportFixture
	 * @brief Helper struct providing set-up/tear-down of HtmlReport tests
	 *
	 * @copydetails EventFixture
	 */
	struct HtmlReportFixture
	{
		Base::SimulationTest sim;
		ostringstream reportStream;
		HtmlReport streamReporter;
		string fileName;
		HtmlReport fileReporter;
		Count count;

		HtmlReportFixture()
		:	sim( "HtmlReportTestSim" ),
			reportStream(),
			streamReporter( &sim, reportStream ),
			fileName( "HtmlReportTest.html" ),
			fileReporter( &sim, fileName ),
			count( &sim, "CountTest" )
			{}
	};

	/**
	 * @test odemx::HtmlReport construction
	 *
	 * Expected effects:
	 * @li empty file names are handled
	 * @li HtmlReport can be constructed with streams or files
	 */
	TEST_FIXTURE( HtmlReportFixture, ConstructionDestruction )
	{
		Time::German germanTime;
		ostringstream stream;
		HtmlReport report1( &sim, "", germanTime );
		HtmlReport report2( "", germanTime );
		HtmlReport report3( &sim, "HtmlReportTestConstructionFile1.html", germanTime );
		HtmlReport report4( "HtmlReportTestConstructionFile2.html", germanTime );
		HtmlReport report5( &sim, stream, germanTime );
		HtmlReport report6( stream, germanTime );
	}

	/**
	 * @test odemx::HtmlReport stream output
	 *
	 * Expected function call effects:
	 * @li the stream contains the expected xml output
	 */
	TEST_FIXTURE( HtmlReportFixture, StreamOutput )
	{
		SimTime time = 25;
		Base::EventTest simAdvanceEvent( &sim, "HtmlReportTestStreamAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		count.reset(); // non-0 reset time
		unsigned int inc = 17;
		int updateCount = 9;
		for( int i = 0; i != updateCount; ++i )
		{
			count.update( inc );
		}

		streamReporter.addProducer( &count );
		streamReporter.generateReport();

		CHECK( reportStream.str().find( "Count Statistics" ) != string::npos );
		CHECK( reportStream.str().find( "Name" ) != string::npos );
		CHECK( reportStream.str().find( "Reset at" ) != string::npos );
		CHECK( reportStream.str().find( "Uses" ) != string::npos );
		CHECK( reportStream.str().find( "Value" ) != string::npos );
		CHECK( reportStream.str().find( count.getLabel() ) != string::npos );
		CHECK( reportStream.str().find( toString( time ) ) != string::npos );
		CHECK( reportStream.str().find( toString( updateCount ) ) != string::npos );
		CHECK( reportStream.str().find( toString( updateCount * inc ) ) != string::npos );
	}

	/**
	 * @test odemx::HtmlReport stream output
	 *
	 * Expected function call effects:
	 * @li the file contains the expected xml output
	 */
	TEST_FIXTURE( HtmlReportFixture, FileOutput )
	{
		SimTime time = 36;
		Base::EventTest simAdvanceEvent( &sim, "HtmlReportTestFileAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		count.reset(); // non-0 reset time
		unsigned int inc = 7;
		int updateCount = 27;
		for( int i = 0; i != updateCount; ++i )
		{
			count.update( inc );
		}

		fileReporter.addProducer( &count );
		fileReporter.generateReport();

		ostringstream fileStream;
		ifstream inFile( fileName.c_str() );
		string tmp;
		while( inFile )
		{
			getline( inFile, tmp );
			fileStream << tmp << endl;
		}

		CHECK( fileStream.str().find( "Count Statistics" ) != string::npos );
		CHECK( fileStream.str().find( "Name" ) != string::npos );
		CHECK( fileStream.str().find( "Reset at" ) != string::npos );
		CHECK( fileStream.str().find( "Uses" ) != string::npos );
		CHECK( fileStream.str().find( "Value" ) != string::npos );
		CHECK( fileStream.str().find( count.getLabel() ) != string::npos );
		CHECK( fileStream.str().find( toString( time ) ) != string::npos );
		CHECK( fileStream.str().find( toString( updateCount ) ) != string::npos );
		CHECK( fileStream.str().find( toString( updateCount * inc ) ) != string::npos );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
