/**
 * @file TestHtmlTrace.cpp
 * @date Aug 21, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class HtmlTrace
 */

#include "TestDisplay.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Display )
{
/// @endcond

	/**
	 * @struct HtmlTraceFixture
	 * @brief Helper struct providing set-up/tear-down of HtmlTrace tests
	 *
	 * @copydetails EventFixture
	 */
	struct HtmlTraceFixture
	{
		Base::SimulationTest trace;
		Util::TraceProducerTest producer;
		std::string fileName;
		ostringstream htmlStream;
		Time::German timeFormat;
		HtmlTrace htmlStreamTrace;
		HtmlTrace htmlFileTrace;

		HtmlTraceFixture()
		:	trace( "HtmlTraceTestSim" ),
			producer( trace, "HtmlTraceProducerTest" ),
			fileName( "HtmlTraceTestFile.html" ),
			htmlStream(),
			timeFormat( Time::Base( 1970, 1, 1, 0, Time::minutes ) ),
			htmlStreamTrace( &trace, htmlStream, timeFormat ),
			htmlFileTrace( &trace, fileName, Time::defaultFormat )
			{
				trace.addConsumer( &htmlStreamTrace );
				trace.addConsumer( &htmlFileTrace );
			}

		static SimTime advanceSim( Simulation& sim, SimTime t )
		{
			Base::EventTest simAdvanceEvent( &sim, "HtmlTraceTestAdvanceEvent" );
			simAdvanceEvent.scheduleIn( t );
			sim.step(); // event execution advances sim time
			return t;
		}
	};

	string readHtmlFile( const string& fileName )
	{
		ostringstream fileStream;
		ifstream inFile( fileName.c_str() );

		if( ! inFile.is_open() )
		{
			cerr << "HtmlTraceTest: cannot open HTML file";
		}

		string tmp;
		while( inFile )
		{
			getline( inFile, tmp );
			fileStream << tmp << endl;
		}

		return fileStream.str();
	}

	/**
	 * @test odemx::HtmlTrace construction and destruction
	 *
	 * Expected effects:
	 * @li empty file names are handled
	 * @li HtmlTrace can be constructed with streams or files
	 */
	TEST_FIXTURE( HtmlTraceFixture, ConstructionDestruction )
	{
		HtmlTrace fileTrace1( "HtmlTraceTestConstructionDefSimFile1.html" );
		HtmlTrace fileTrace2( "HtmlTraceTestConstructionDefSimFile2.html", timeFormat );
		HtmlTrace fileTrace3( &trace, "HtmlTraceTestConstructionUserSimFile1.html" );
		HtmlTrace fileTrace4( &trace, "HtmlTraceTestConstructionUserSimFile2.html", timeFormat );

		HtmlTrace streamTrace1( htmlStream );
		HtmlTrace streamTrace2( htmlStream, timeFormat );
		HtmlTrace streamTrace3( &trace, htmlStream );
		HtmlTrace streamTrace4( &trace, htmlStream, timeFormat );
	}

	/**
	 * @test odemx::HtmlTrace::start/stopTrace()
	 *
	 * Expected function call effects:
	 * @li stream and file contain the html trace root with simulation label
	 * and start time
	 */
	TEST_FIXTURE( HtmlTraceFixture, StartAndStop )
	{
		SimTime time = advanceSim( trace, 36231 );

		trace.startTrace();
		trace.stopTrace();

		CHECK( htmlStream.str().find( "HtmlTrace for simulation" ) != string::npos );
		CHECK( htmlStream.str().find( trace.getLabel() ) != string::npos );
		CHECK( htmlStream.str().find( simTimeToString( time, timeFormat ) ) != string::npos );
		CHECK( htmlStream.str().find( htmlStreamTrace.getFilter() ) != string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "HtmlTrace for simulation" ) != string::npos );
		CHECK( fileStr.find( trace.getLabel() ) != string::npos );
		CHECK( fileStr.find( simTimeToString( time, Time::defaultFormat ) ) != string::npos );
		CHECK( fileStr.find( htmlFileTrace.getFilter() ) != string::npos );
	}

	/**
	 * @test odemx::HtmlTrace::pause/continueTrace()
	 *
	 * Expected function call effects:
	 * @li stream and file contain a pause mark sent by the Simulation
	 * with pause start time and reactivation time
	 */
	TEST_FIXTURE( HtmlTraceFixture, PauseAndContinue )
	{
		trace.startTrace();
		trace.pauseTrace();

		SimTime reactivationTime = advanceSim( trace, 42424 );

		trace.continueTrace();
		trace.stopTrace();

		CHECK( htmlStream.str().find( "BREAK" ) != string::npos );
		CHECK( htmlStream.str().find( simTimeToString( reactivationTime, timeFormat ) ) != string::npos );
		CHECK( htmlStream.str().find( "CONTINUE" ) != string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "BREAK" ) != string::npos );
		CHECK( fileStr.find( simTimeToString( reactivationTime, Time::defaultFormat ) ) != string::npos );
		CHECK( fileStr.find( "CONTINUE" ) != string::npos );
	}

	/**
	 * @test odemx::HtmlTrace::begin/endMark()
	 *
	 * Expected function call effects:
	 * @li the composed mark contains SimTime, mark name, sender details and
	 * the closing comment
	 */
	TEST_FIXTURE( HtmlTraceFixture, BeginAndEndMark )
	{
		SimTime sendTime = advanceSim( trace, 55439 );
		trace.startTrace();

		// open composed mark
		trace.beginMark( &producer, Util::TraceProducerTest::markComposed );
		trace.endMark( "composed mark comment" );

		CHECK( htmlStream.str().find( "markComposed" ) != string::npos );
		CHECK( htmlStream.str().find( simTimeToString( sendTime, timeFormat ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( producer.getLabel() )	!= string::npos );
		CHECK( htmlStream.str().find( "composed mark comment" ) != string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "markComposed" ) != string::npos );
		CHECK( fileStr.find( simTimeToString( sendTime, Time::defaultFormat ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( producer.getLabel() )	!= string::npos );
		CHECK( fileStr.find( "composed mark comment" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::HtmlTrace::mark(const TraceProducer*,MarkType,const string&)
	 *
	 * Expected function call effects:
	 * @li the simple mark contains SimTime, mark name, sender details and
	 * the closing comment
	 */
	TEST_FIXTURE( HtmlTraceFixture, SimpleMark )
	{
		SimTime sendTime = advanceSim( trace, 4021 );
		trace.startTrace();

		trace.mark( &producer, Util::TraceProducerTest::markEvent, "test comment" );

		CHECK( htmlStream.str().find( "markEvent" ) != string::npos );
		CHECK( htmlStream.str().find( simTimeToString( sendTime, timeFormat ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( producer.getLabel() )	!= string::npos );
		CHECK( htmlStream.str().find( "test comment" ) != string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "markEvent" ) != string::npos );
		CHECK( fileStr.find( simTimeToString( sendTime, Time::defaultFormat ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( producer.getLabel() )	!= string::npos );
		CHECK( fileStr.find( "test comment" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::HtmlTrace::mark(const TraceProducer*,MarkType,const TraceProducer*,const string&)
	 *
	 * Expected function call effects:
	 * @li the simple mark contains SimTime, mark name, sender details, partner
	 * details and the closing comment
	 */
	TEST_FIXTURE( HtmlTraceFixture, SimpleMarkWithPartner )
	{
		Util::TraceProducerTest dummyPartner( trace, "HtmlTraceProducerDummyPartner" );
		SimTime sendTime = advanceSim( trace, 4021 );
		trace.startTrace();

		trace.mark( &producer, Util::TraceProducerTest::markPartner, &dummyPartner, "test partner comment" );

		CHECK( htmlStream.str().find( "markPartner" ) != string::npos );
		CHECK( htmlStream.str().find( simTimeToString( sendTime, timeFormat ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( producer.getLabel() ) != string::npos );
		CHECK( htmlStream.str().find( dummyPartner.getLabel() ) != string::npos );
		CHECK( htmlStream.str().find( "test partner comment" ) != string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "markPartner" ) != string::npos );
		CHECK( fileStr.find( simTimeToString( sendTime, Time::defaultFormat ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( producer.getLabel() ) != string::npos );
		CHECK( fileStr.find( dummyPartner.getLabel() ) != string::npos );
		CHECK( fileStr.find( "test partner comment" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::HtmlTrace::mark(const TraceProducer*,MarkType,const TraceProducer*,const TraceProducer*,const string&)
	 *
	 * Expected function call effects:
	 * @li the simple mark contains SimTime, mark name, sender details, both
	 * partners and the closing comment
	 */
	TEST_FIXTURE( HtmlTraceFixture, SimpleMarkWithTwoPartners )
	{
		Util::TraceProducerTest dummyPartner1( trace, "HtmlTraceProducerDummyPartner1" );
		Util::TraceProducerTest dummyPartner2( trace, "HtmlTraceProducerDummyPartner2" );
		SimTime sendTime = advanceSim( trace, 120621 );
		trace.startTrace();

		trace.mark( &producer, Util::TraceProducerTest::markTwoPartners, &dummyPartner1, &dummyPartner2, "test two partners comment" );

		CHECK( htmlStream.str().find( "markTwoPartners" ) != string::npos );
		CHECK( htmlStream.str().find( simTimeToString( sendTime, timeFormat ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( producer.getLabel() ) != string::npos );
		CHECK( htmlStream.str().find( dummyPartner1.getLabel() ) != string::npos );
		CHECK( htmlStream.str().find( dummyPartner2.getLabel() ) != string::npos );
		CHECK( htmlStream.str().find( "test two partners comment" ) != string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "markTwoPartners" ) != string::npos );
		CHECK( fileStr.find( simTimeToString( sendTime, Time::defaultFormat ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( producer.getLabel() ) != string::npos );
		CHECK( fileStr.find( dummyPartner1.getLabel() ) != string::npos );
		CHECK( fileStr.find( dummyPartner2.getLabel() ) != string::npos );
		CHECK( fileStr.find( "test two partners comment" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::HtmlTrace::mark(const TraceProducer*,MarkType,AttributeType,AttributeType,const string&)
	 *
	 * Expected function call effects:
	 * @li the simple mark contains SimTime, mark name, sender details, tags
	 * for the old and new values, and the closing comment
	 */
	TEST_FIXTURE( HtmlTraceFixture, SimpleMarkAttributeChange )
	{
		SimTime sendTime = advanceSim( trace, 220621 );
		trace.startTrace();

		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, true, false, "test attribute comment" );

		CHECK( htmlStream.str().find( "markAttributeChange" ) != string::npos );
		CHECK( htmlStream.str().find( simTimeToString( sendTime, timeFormat ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( producer.getLabel() ) != string::npos );
		CHECK( htmlStream.str().find( "old Value" ) != string::npos );
		CHECK( htmlStream.str().find( "false" ) != string::npos );
		CHECK( htmlStream.str().find( "new Value" ) != string::npos );
		CHECK( htmlStream.str().find( "true" ) != string::npos );
		CHECK( htmlStream.str().find( "test attribute comment" ) != string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "markAttributeChange" ) != string::npos );
		CHECK( fileStr.find( simTimeToString( sendTime, Time::defaultFormat ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( producer.getLabel() ) != string::npos );
		CHECK( fileStr.find( "old Value" ) != string::npos );
		CHECK( fileStr.find( "false" ) != string::npos );
		CHECK( fileStr.find( "new Value" ) != string::npos );
		CHECK( fileStr.find( "true" ) != string::npos );
		CHECK( fileStr.find( "test attribute comment" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::HtmlTrace::mark(const TraceProducer*,MarkType,AttributeType,AttributeType,const string&)
	 *
	 * Expected function call effects:
	 * @li attribute changes for built-in types are correctly transformed and
	 * stored in HTML table cells
	 */
	TEST_FIXTURE( HtmlTraceFixture, SimpleMarkAttributeTypes )
	{
		trace.startTrace();

		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, true, false );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, 'a', 'b' );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(short), MAX_VAL(short) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(int), MAX_VAL(int) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(long), MAX_VAL(long) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(long long), MAX_VAL(long long) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(float), MAX_VAL(float) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(double), MAX_VAL(double) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, "const char*", "const std::string&" );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, (unsigned char)'^', (unsigned char)'#' );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned short), MAX_VAL(unsigned short) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned int), MAX_VAL(unsigned int) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned long), MAX_VAL(unsigned long) );

		CHECK( htmlStream.str().find( "true</td>" ) != string::npos );
		CHECK( htmlStream.str().find( "false</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( 'a' ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( 'b' ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MIN_VAL(short) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MAX_VAL(short) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MIN_VAL(int) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MAX_VAL(int) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MIN_VAL(long) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MAX_VAL(long) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MIN_VAL(long long) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MAX_VAL(long long) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MIN_VAL(float) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MAX_VAL(float) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MIN_VAL(double) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MAX_VAL(double) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString(	"const char*" ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( "const std::string&" ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString(	"^" ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( "#" ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MIN_VAL(unsigned short) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MAX_VAL(unsigned short) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MIN_VAL(unsigned int) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MAX_VAL(unsigned int) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MIN_VAL(unsigned long) ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( MAX_VAL(unsigned long) ) + "</td>" ) != string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "true</td>" ) != string::npos );
		CHECK( fileStr.find( "false</td>" ) != string::npos );
		CHECK( fileStr.find( toString( 'a' ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( 'b' ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MIN_VAL(short) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MAX_VAL(short) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MIN_VAL(int) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MAX_VAL(int) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MIN_VAL(long) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MAX_VAL(long) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MIN_VAL(long long) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MAX_VAL(long long) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MIN_VAL(float) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MAX_VAL(float) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MIN_VAL(double) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MAX_VAL(double) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString(	"const char*" ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( "const std::string&" ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString(	"^" ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( "#" ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MIN_VAL(unsigned short) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MAX_VAL(unsigned short) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MIN_VAL(unsigned int) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MAX_VAL(unsigned int) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MIN_VAL(unsigned long) ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( MAX_VAL(unsigned long) ) + "</td>" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::HtmlTrace::begin/endTag(Tag)
	 *
	 * Expected function call effects:
	 * @li the composed tag is correctly represented in XML
	 */
	TEST_FIXTURE( HtmlTraceFixture, BeginAndEndTag )
	{
		trace.startTrace();

		trace.beginMark( &producer, Util::TraceProducerTest::markComposed );
		trace.beginTag( Util::TraceProducerTest::tagComposed );
		trace.endTag( Util::TraceProducerTest::tagComposed );
		trace.endMark();

		CHECK( htmlStream.str().find( "tagComposed" ) != string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "tagComposed" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::HtmlTrace::addTag(Tag, TagValueType)
	 *
	 * Expected function call effects:
	 * @li tags for built-in types are correctly transformed and
	 * stored in HTML tables
	 */
	TEST_FIXTURE( HtmlTraceFixture, AddTagValueTypes )
	{
		trace.startTrace();
		trace.beginMark( &producer, Util::TraceProducerTest::markComposed );

		bool value1 = true;
		char value2 = '?';
		short value3 = 1;
		int value4 = 1;
		long value5 = 1;
		long long value6 = 1;
		float value7 = 1.1f;
		double value8 = 1.1;
		const char* value9 = "value";
		const TraceProducer* value10 = &producer;
		unsigned char value11 = '1';
		unsigned short value12 = 1;
		unsigned int value13 = 1;
		unsigned long value14 = 1;

		trace.addTag( Util::TraceProducerTest::tagAttribute, value1 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value2 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value3 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value4 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value5 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value6 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value7 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value8 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value9 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value10 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value11 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value12 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value13 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value14 );

		trace.endMark();
		trace.stopTrace();

		CHECK( htmlStream.str().find( "true</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value2 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value3 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value4 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value5 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value6 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value7 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value8 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value9 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( producer.getLabel() + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value11 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value12 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value13 ) + "</td>" ) != string::npos );
		CHECK( htmlStream.str().find( toString( value14 ) + "</td>" ) != string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "true</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value2 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value3 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value4 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value5 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value6 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value7 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value8 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value9 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( producer.getLabel() + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value11 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value12 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value13 ) + "</td>" ) != string::npos );
		CHECK( fileStr.find( toString( value14 ) + "</td>" ) != string::npos );
	}

	/**
	 * @test odemx::HtmlTrace mark blocking
	 *
	 * The Filter is set to block all marks. Expected effects:
	 * @li no marks are written to XML stream or file
	 */
	TEST_FIXTURE( HtmlTraceFixture, PassNone )
	{
		Util::TraceProducerTest dummyPartner1( trace, "XmlTraceProducerDummyPartner1" );
		Util::TraceProducerTest dummyPartner2( trace, "XmlTraceProducerDummyPartner2" );

		trace.startTrace();
		htmlStreamTrace.passNone();
		htmlFileTrace.passNone();

		trace.mark( &producer, Util::TraceProducerTest::markEvent, "test comment" );
		trace.mark( &producer, Util::TraceProducerTest::markPartner, &dummyPartner1, "test partner comment" );
		trace.mark( &producer, Util::TraceProducerTest::markTwoPartners, &dummyPartner1, &dummyPartner2, "test two partners comment" );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, true, false );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, 'a', 'b' );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(short), MAX_VAL(short) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(int), MAX_VAL(int) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(long), MAX_VAL(long) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(long long), MAX_VAL(long long) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(float), MAX_VAL(float) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(double), MAX_VAL(double) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, "const char*", "const std::string&" );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, (unsigned char)'^', (unsigned char)'#' );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned short), MAX_VAL(unsigned short) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned int), MAX_VAL(unsigned int) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned long), MAX_VAL(unsigned long) );

		CHECK( htmlStream.str().find( "mark" ) == string::npos );
		string fileStr = readHtmlFile( fileName );
		CHECK( fileStr.find( "mark" ) == string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::HtmlTrace::printState()
	 *
	 * Expected call effects:
	 * @li all possible Process states are printed as strings in the trace
	 * @li there is no error string in the trace as this statement should never
	 * be reached
	 */
	TEST_FIXTURE( HtmlTraceFixture, PrintProcessState )
	{
		trace.startTrace();

		Base::ProcessTest process( &trace, "ProcessTest" );
		process.activate();
		process.sleep();
		process.interrupt();
		trace.run();

		CHECK( htmlStream.str().find( "CREATED" ) != string::npos );
		CHECK( htmlStream.str().find( "IDLE" ) != string::npos );
		CHECK( htmlStream.str().find( "RUNNABLE" ) != string::npos );
		CHECK( htmlStream.str().find( "CURRENT" ) != string::npos );
		CHECK( htmlStream.str().find( "TERMINATED" ) != string::npos );
		CHECK( htmlStream.str().find( "Trace ERROR" ) == string::npos );

		string fileStr = readHtmlFile( fileName );

		CHECK( fileStr.find( "CREATED" ) != string::npos );
		CHECK( fileStr.find( "IDLE" ) != string::npos );
		CHECK( fileStr.find( "RUNNABLE" ) != string::npos );
		CHECK( fileStr.find( "CURRENT" ) != string::npos );
		CHECK( fileStr.find( "TERMINATED" ) != string::npos );
		CHECK( fileStr.find( "Trace ERROR" ) == string::npos );

		trace.stopTrace();
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
