/**
 * @file TestTimeFormat.cpp
 * @date Aug 16, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx classes Time::Base, Time::Format, Time::German,
 * Time::ISO8601, Time::Zone
 */

#include "TestDisplay.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Display )
{
/// @endcond

	/// helper to convert enum Time::Unit to string
	string unitToString( const Time::Unit unit )
	{
		switch( unit )
		{
		case Time::milliseconds:
			return "milliseconds";
		case Time::seconds:
			return "seconds";
		case Time::minutes:
			return "minutes";
		case Time::hours:
			return "hours";
		default:
			return "no_unit";
		}
		return "error: Time::unit could not be determined";
	}

	/// helper for converting a Time::Base to string
	string baseToString( const Time::Base& base )
	{
		ostringstream stream;

		stream
		<< "Year: " << base.year
		<< ", Month: " << base.month
		<< ", Day: " << base.day
		<< ", Offset: " << base.offset
		<< ", Unit: " << unitToString( base.unit );

		return stream.str();
	}

	/**
	 * @struct TraceFilterFixture
	 * @brief Helper struct providing set-up/tear-down of TraceFilter tests
	 *
	 * @copydetails EventFixture
	 */
	struct TimeFixture
	{
		TimeFormatTest format;
		Time::German german;
		Time::Base base;
		Time::ISO8601 iso;

		TimeFixture()
		:	format(),
			german(),
			base( 2000, 01, 01, 1000000, Time::milliseconds ),
			iso( base )
			{}
	};

	/**
	 * @test odemx::Time::Base construction
	 *
	 * Expected effects:
	 * @li year, month, day, offset, and unit are initialized correctly
	 */
	TEST( TimeBase )
	{
		Time::Base base( 2008, 8, 16, 7244, Time::minutes );
		CHECK_EQUAL( "Year: 2008, Month: 8, Day: 16, Offset: 7244, Unit: minutes",
				baseToString( base ) );
	}

	/**
	 * @test odemx::Time::Zone construction
	 *
	 * Expected effects:
	 * @li hour parameter out of permitted range causes an error message
	 * @li minute parameter out of permitted range causes an error message
	 * @li otherwise, hour and minute are initialized correctly
	 */
	TEST( TimeZone )
	{
		clearErrorStream();
		Time::Zone zone1( -13, 0 );
		CHECK_EQUAL( "ODEMx ERROR: Time::Zone(): hour value range is -12 through 14\n",
				errorStream().str() );

		clearErrorStream();
		Time::Zone zone2( 15, 0 );
		CHECK_EQUAL( "ODEMx ERROR: Time::Zone(): hour value range is -12 through 14\n",
				errorStream().str() );

		clearErrorStream();
		Time::Zone zone3( 0, 1 );
		CHECK_EQUAL( "ODEMx ERROR: Time::Zone(): minute requires values 0, 15, 30, or 45\n",
				errorStream().str() );

		unsigned short minute[] = { 0, 15, 30, 45 };
		short hour = -1;

		clearErrorStream();
		for( unsigned short i = 0; i != 4; ++i )
		{
			Time::Zone zone( hour, minute[ i ] );
			CHECK_EQUAL( hour, zone.hour );
			CHECK_EQUAL( minute[ i ], zone.minute );
		}
		CHECK( errorStream().str().empty() );
	}

	/**
	 * @test odemx::Time::Format construction
	 *
	 * Expected effects:
	 * @li the object can be default-constructed with 0-Time::Base
	 * @li the object can be constructed with a reference to a Time::Base
	 * @li the Time::Base is set correctly
	 */
	TEST_FIXTURE( TimeFixture, FormatConstruction )
	{
		TimeFormatTest testFormat1;
		CHECK_EQUAL( "Year: 0, Month: 0, Day: 0, Offset: 0, Unit: no_unit",
				baseToString( testFormat1.getTimeBase() ) );

		TimeFormatTest testFormat2( Time::Base( 2008, 8, 16, 7244, Time::minutes ) );
		CHECK_EQUAL( "Year: 2008, Month: 8, Day: 16, Offset: 7244, Unit: minutes",
				baseToString( testFormat2.getTimeBase() ) );
	}

	/**
	 * @test odemx::Time::German construction
	 *
	 * Expected function call effects:
	 * @li the object can be default-constructed with 0-Time::Base
	 * @li the object can be constructed with a reference to a Time::Base
	 * @li the Time::Base is set correctly
	 */
	TEST_FIXTURE( TimeFixture, GermanConstruction )
	{
		Time::German testFormat1;
		CHECK_EQUAL( "Year: 0, Month: 0, Day: 0, Offset: 0, Unit: no_unit",
				baseToString( testFormat1.getTimeBase() ) );

		Time::Base base( 2008, 8, 16, 7244, Time::minutes );
		Time::German testFormat2( base );
		CHECK_EQUAL( "Year: 2008, Month: 8, Day: 16, Offset: 7244, Unit: minutes",
				baseToString( testFormat2.getTimeBase() ) );
	}

	/**
	 * @test odemx::Time::Format::setTimeBase()
	 *
	 * Expected function call effects:
	 * @li time base can be set by a reference and by giving separate parameters
	 */
	TEST_FIXTURE( TimeFixture, FormatSetTimeBase )
	{
		Time::Base base( 2008, 8, 16, 7244, Time::minutes );
		format.setTimeBase( base );
		CHECK_EQUAL( baseToString( base ), baseToString( format.getTimeBase() ) );

		format.setTimeBase( 2008, 8, 16, 7244, Time::minutes );
		CHECK_EQUAL( baseToString( base ), baseToString( format.getTimeBase() ) );
	}

	/**
	 * @test odemx::Time::Format::timeToString(SimTime)
	 *
	 * Expected function call effects:
	 * @li if no Time::Unit is set, \c computeTimeValues() produces an error
	 * @li time display is computed correctly for units in milliseconds,
	 * seconds, minutes, and hours
	 */
	TEST_FIXTURE( TimeFixture, FormatTimeToString )
	{
		// default initialization without suitable Time::Base
		TimeFormatTest test;
		clearErrorStream();
		test.timeToString( 0 );
		CHECK_EQUAL( "ODEMx ERROR: Time::Format::computeTimeValues(): cannot compute time, no base unit given\n",
				errorStream().str() );

		// set suitable Time::Base
		format.setTimeBase( 0,0,0,0,Time::milliseconds );
		CHECK_EQUAL( "time: 0/3/24-17:46:40.123", format.timeToString( 10000000123ll ) );
		format.setTimeBase( 0,0,0,0,Time::seconds );
		CHECK_EQUAL( "time: 0/3/24-17:46:40", format.timeToString( 10000000 ) );
		// the following two tests might produce a different result
		// if SimTime is not an integral type
		format.setTimeBase( 0,0,0,1439,Time::minutes );
		CHECK_EQUAL( "time: 0/3/25-17:45:0", format.timeToString( 10000000 / 60 ) );
		format.setTimeBase( 0,0,0,0,Time::hours );
		CHECK_EQUAL( "time: 0/3/24-17:0:0", format.timeToString( 10000000 / 60 / 60 ) );
	}

	/**
	 * @test odemx::Time::German::timeToString(SimTime)
	 *
	 * Expected function call effects:
	 * @li time display is computed correctly for units in milliseconds,
	 * seconds, minutes, and hours
	 */
	TEST_FIXTURE( TimeFixture, GermanTimeToString )
	{
		// set suitable Time::Base
		german.setTimeBase( 0,0,0,0,Time::milliseconds );
		CHECK_EQUAL( "00-03-24 / 17:46:40.123", german.timeToString( 10000000123ll ) );
		german.setTimeBase( 0,0,0,0,Time::seconds );
		CHECK_EQUAL( "00-03-24 / 17:46:40", german.timeToString( 10000000 ) );
		// the following two tests might produce a different result
		// if SimTime is not an integral type
		german.setTimeBase( 0,0,0,1439,Time::minutes );
		CHECK_EQUAL( "00-03-25 / 17:45:00", german.timeToString( 10000000 / 60 ) );
		german.setTimeBase( 0,0,0,0,Time::hours );
		CHECK_EQUAL( "00-03-24 / 17:00:00", german.timeToString( 10000000 / 60 / 60 ) );
	}

	/**
	 * @test odemx::Time::ISO8601::timeToString(SimTime)
	 *
	 * Expected function call effects:
	 * @li time display is computed correctly for units in milliseconds,
	 * seconds, minutes, and hours
	 */
	TEST_FIXTURE( TimeFixture, ISO8601TimeToString )
	{
		// set suitable Time::Base
		iso.setTimeBase( 0,0,0,0,Time::milliseconds );
		iso.setPrecision( Time::precision_milli );
		// default zone is UTC-0
		CHECK_EQUAL( "0000-03-24T17:46:40.123Z", iso.timeToString( 10000000123ll ) );
		iso.setTimeBase( 0,0,0,0,Time::seconds );
		iso.setPrecision( Time::precision_sec );
		iso.setZone( -1, 0 );
		CHECK_EQUAL( "0000-03-24T17:46:40-01:00", iso.timeToString( 10000000 ) );
		// the following two tests might produce a different result
		// if SimTime is not an integral type
		iso.setTimeBase( 0,0,0,1439,Time::minutes );
		iso.setPrecision( Time::precision_min );
		iso.setZone( 12, 30 );
		CHECK_EQUAL( "0000-03-25T17:45+12:30", iso.timeToString( 10000000 / 60 ) );
		iso.setTimeBase( 0,0,0,0,Time::hours );
		iso.setZone( 4, 45 );
		CHECK_EQUAL( "0000-03-24T17:00+04:45", iso.timeToString( 10000000 / 60 / 60 ) );
	}

	/**
	 * @test odemx::simTimeToString(SimTime,Time::Format&)
	 *
	 * Expected function call effects:
	 * @li the given SimTime is returned in string representation
	 * @li the resulting string is formatted using the given format parameter
	 */
	TEST_FIXTURE( TimeFixture, SimTimeToString )
	{
		CHECK_EQUAL( toString( 12345 ), simTimeToString( 12345, Time::defaultFormat ) );
		format.setTimeBase( 0,0,0,0,Time::seconds );
		CHECK_EQUAL( "time: 0/3/24-17:46:40", simTimeToString( 10000000, format ) );
	}


/// @cond DOXYGEN_SKIP
}
/// @endcond
