/**
 * @file TestTraceFilter.cpp
 * @date Aug 16, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class TraceFilter
 */

#include "TestDisplay.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Display )
{
/// @endcond

	/**
	 * @struct TraceFilterFixture
	 * @brief Helper struct providing set-up/tear-down of TraceFilter tests
	 *
	 * @copydetails EventFixture
	 */
	struct TraceFilterFixture
	{
		Trace trace;
		TraceFilterTest filter;
		Util::TraceProducerTest producer1;
		Util::TraceProducerTest producer2;

		TraceFilterFixture()
		:	trace(),
			filter(),
			producer1( trace, "TraceFilterTestProducer1" ),
			producer2( trace, "TraceFilterTestProducer2" )
			{}
	};

	/**
	 * @test odemx::TraceFilter construction
	 *
	 * Expected effects:
	 * @li default filter setting is to let all marks pass
	 */
	TEST_FIXTURE( TraceFilterFixture, ConstructionDestruction )
	{
		TraceFilterTest f;
		CHECK( f.getFilter().find( "FilterMode: PASS ALL" ) != string::npos );
	}

	/**
	 * @test odemx::TraceFilter::setFilter(const string&)
	 *
	 * Expected function call effects:
	 * @li all filter commands are correctly interpreted and
	 * inserted into the internal comparison sets
	 * @li empty string causes no changes
	 */
	TEST_FIXTURE( TraceFilterFixture, SetFilter )
	{
		filter.setFilter("none; \
				mti: 101, 222, 604; \
				mtn: create, destroy, change state; \
				mts: odemx::Simulation, odemx::Process; \
				sl: My Process, My Event; \
				st: odemx::Timer");

		CHECK_EQUAL( string( "FilterMode: PASS NONE " )
				+ "| MarkTypeIds: 101, 222, 604 "
				+ "| MarkTypeNames: change state, create, destroy "
				+ "| MarkTypeScopes: odemx::Process, odemx::Simulation "
				+ "| SenderLabels: My Event, My Process "
				+ "| SenderTypes: odemx::Timer",
				filter.getFilter() );

		string previousFilter = filter.getFilter();
		filter.setFilter( "" );
		CHECK_EQUAL( previousFilter, filter.getFilter() );
	}

	/**
	 * @test odemx::TraceFilter::setFilter(const string&) errors
	 *
	 * Expected function call effects:
	 * @li in case of unknown mode string, an error message is sent to
	 * \c errorStream()
	 * @li in case of NaN for MarkTypeId, an error message is sent to
	 * \c errorStream()
	 */
	TEST_FIXTURE( TraceFilterFixture, SetFilterErrors )
	{
		clearErrorStream();
		filter.setFilter( "abc" );
		CHECK_EQUAL( "ODEMx ERROR: TraceFilter::setFilter(): error in filter pattern, mode expected\n",
				errorStream().str() );

		clearErrorStream();
		filter.setFilter( "all; mti: abc;" );
		CHECK_EQUAL( "ODEMx ERROR: TraceFilter::setFilter(): error in filter pattern, number expected\n",
				errorStream().str() );
	}

	/**
	 * @test odemx::TraceFilter::filterXXX() with input operator
	 *
	 * Expected function call effects:
	 * @li all filter values are correctly inserted into the internal
	 * comparison sets
	 */
	TEST_FIXTURE( TraceFilterFixture, FilterWithOperator )
	{
		filter.passNone();
		filter.filterMarkById()		<< 101 << 222 << 604;
		filter.filterMarkByName()	<< "create" << "destroy" << "change state";
		filter.filterMarkByScope()	<< typeid(Simulation) << "odemx::Process";
		filter.filterSenderByLabel()<< "My Process"	<< "My Event";
		filter.filterSenderByType()	<< typeid(Timer);

		string filterValue = filter.getFilter(); 
		CHECK( filterValue.find( "FilterMode: PASS NONE" ) != string::npos );
		CHECK( filterValue.find( "MarkTypeIds: 101, 222, 604" ) != string::npos );
		CHECK( filterValue.find( "MarkTypeNames: change state, create, destroy" ) != string::npos );
		CHECK( filterValue.find( "MarkTypeScopes:" ) != string::npos );
		CHECK( filterValue.find( "odemx::Process" ) != string::npos );
		CHECK( filterValue.find( typeToString( typeid(Simulation) ) ) != string::npos );
		CHECK( filterValue.find( "SenderLabels: My Event, My Process" ) != string::npos );
		CHECK( filterValue.find( "SenderTypes:" ) != string::npos );
		CHECK( filterValue.find( typeToString( typeid(Timer) ) ) != string::npos );
	}

	/**
	 * @test odemx::TraceFilter::passAll/None() filter mode
	 *
	 * Expected function call effects:
	 * @li the default mode is PASSALL
	 * @li the filter is set accordingly
	 */
	TEST_FIXTURE( TraceFilterFixture, PassMode )
	{
		// default
		CHECK_EQUAL( string( "FilterMode: PASS ALL" ),	filter.getFilter() );

		filter.passNone();
		CHECK_EQUAL( string( "FilterMode: PASS NONE" ),	filter.getFilter() );

		filter.passAll();
		CHECK_EQUAL( string( "FilterMode: PASS ALL" ),	filter.getFilter() );

		filter.setFilter( "none" );
		CHECK_EQUAL( string( "FilterMode: PASS NONE" ),	filter.getFilter() );

		filter.setFilter( "all" );
		CHECK_EQUAL( string( "FilterMode: PASS ALL" ),	filter.getFilter() );
	}

	/**
	 * @test odemx::TraceFilter::pass() all except the filtered
	 *
	 * Expected function call effects:
	 * @li all marks pass unless they have been added to the filter
	 * @li correct filtering by mark id
	 * @li correct filtering by mark name
	 * @li correct filtering by mark scope
	 * @li correct filtering by sender label
	 * @li correct filtering by sender type
	 */
	TEST_FIXTURE( TraceFilterFixture, PassAllExceptFiltered )
	{
		// filter not set, should always pass
		CHECK_EQUAL( true, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );

		filter.filterMarkById() << Util::MarkTypeTest::mark1.getId();
		CHECK_EQUAL( false, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );
		CHECK_EQUAL( true, filter.pass( &Util::MarkTypeTest::mark2, &producer1 ) );

		filter.resetFilter();
		filter.filterMarkByName() << Util::MarkTypeTest::mark1.getName();
		CHECK_EQUAL( false, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );
		CHECK_EQUAL( true, filter.pass( &Util::MarkTypeTest::mark2, &producer1 ) );

		filter.resetFilter();
		filter.filterMarkByScope() << Util::MarkTypeTest::mark1.getScope();
		CHECK_EQUAL( false, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );
		CHECK_EQUAL( true, filter.pass( &Util::TraceProducerTest::markEvent, &producer1 ) );

		filter.resetFilter();
		filter.filterSenderByLabel() << producer1.getLabel();
		CHECK_EQUAL( false, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );
		CHECK_EQUAL( true, filter.pass( &Util::MarkTypeTest::mark1, &producer2 ) );

		filter.resetFilter();
		filter.filterSenderByType() << typeid( producer1 );
		// need different sender type
		Base::SimulationTest sim( "TraceFilterTestPassAllSim" );
		CHECK_EQUAL( false, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );
		CHECK_EQUAL( true, filter.pass( &Util::MarkTypeTest::mark1, &sim ) );
	}

	/**
	 * @test odemx::TraceFilter::pass() none except the filtered
	 *
	 * Expected function call effects:
	 * @li all marks are filtered out unless they have been added to the filter
	 * @li correct filtering by mark id
	 * @li correct filtering by mark name
	 * @li correct filtering by mark scope
	 * @li correct filtering by sender label
	 * @li correct filtering by sender type
	 */
	TEST_FIXTURE( TraceFilterFixture, PassNoneExceptFiltered )
	{
		filter.passNone();
		filter.filterMarkById() << Util::MarkTypeTest::mark1.getId();
		CHECK_EQUAL( true, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );
		CHECK_EQUAL( false, filter.pass( &Util::MarkTypeTest::mark2, &producer1 ) );

		filter.resetFilter();
		filter.passNone();
		filter.filterMarkByName() << Util::MarkTypeTest::mark1.getName();
		CHECK_EQUAL( true, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );
		CHECK_EQUAL( false, filter.pass( &Util::MarkTypeTest::mark2, &producer1 ) );

		filter.resetFilter();
		filter.passNone();
		filter.filterMarkByScope() << Util::MarkTypeTest::mark1.getScope();
		CHECK_EQUAL( true, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );
		CHECK_EQUAL( false, filter.pass( &Util::TraceProducerTest::markEvent, &producer1 ) );

		filter.resetFilter();
		filter.passNone();
		filter.filterSenderByLabel() << producer1.getLabel();
		CHECK_EQUAL( true, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );
		CHECK_EQUAL( false, filter.pass( &Util::MarkTypeTest::mark1, &producer2 ) );

		filter.resetFilter();
		filter.passNone();
		filter.filterSenderByType() << typeid( producer1 );
		// need different sender type
		Base::SimulationTest sim( "TraceFilterTestPassAllSim" );
		CHECK_EQUAL( true, filter.pass( &Util::MarkTypeTest::mark1, &producer1 ) );
		CHECK_EQUAL( false, filter.pass( &Util::MarkTypeTest::mark1, &sim ) );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
