/**
 * @file TestXmlReport.cpp
 * @date Aug 15, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class XmlReport
 */

#include "TestDisplay.h"
#include "../TestBase/TestBase.h"

/// @cond DOXYGEN_SKIP
SUITE( Display )
{
/// @endcond

	/**
	 * @struct XmlReportFixture
	 * @brief Helper struct providing set-up/tear-down of XmlReport tests
	 *
	 * @copydetails EventFixture
	 */
	struct XmlReportFixture
	{
		Base::SimulationTest sim;
		ostringstream reportStream;
		XmlReport streamReporter;
		string fileName;
		XmlReport fileReporter;
		Count count;

		XmlReportFixture()
		:	sim( "XmlReportTestSim" ),
			reportStream(),
			streamReporter( &sim, reportStream ),
			fileName( "XmlReportTest.xml" ),
			fileReporter( &sim, fileName ),
			count( &sim, "CountTest" )
			{}
	};

	/**
	 * @test odemx::XmlReport construction
	 *
	 * Expected effects:
	 * @li empty file names are handled
	 * @li XmlReport can be constructed with streams or files
	 */
	TEST_FIXTURE( XmlReportFixture, ConstructionDestruction )
	{
		Time::German germanTime;
		ostringstream stream;
		XmlReport report1( &sim, "", germanTime );
		XmlReport report2( "", germanTime );
		XmlReport report3( &sim, "XmlReportTestConstructionFile1.xml", germanTime );
		XmlReport report4( "XmlReportTestConstructionFile2.xml", germanTime );
		XmlReport report5( &sim, stream, germanTime );
		XmlReport report6( stream, germanTime );
	}

	/**
	 * @test odemx::XmlReport stream output
	 *
	 * Expected function call effects:
	 * @li the stream contains the expected xml output
	 */
	TEST_FIXTURE( XmlReportFixture, StreamOutput )
	{
		SimTime time = 25;
		Base::EventTest simAdvanceEvent( &sim, "XmlReportTestStreamAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		unsigned int inc = 13;
		count.reset(); // non-0 reset time
		count.update( inc ); // two updates
		count.update( inc ); // result in value inc * 2

		streamReporter.addProducer( &count );
		streamReporter.generateReport();

		CHECK( reportStream.str().find( "<reporttable label=\"Count Statistics\" columns=\"4\" lines=\"1\">" ) != string::npos );
		CHECK( reportStream.str().find( "<column number=\"1\" label=\"Name\" type=\"STRING\" />" ) != string::npos );
		CHECK( reportStream.str().find( "<column number=\"2\" label=\"Reset at\" type=\"SIMTIME\" />" ) != string::npos );
		CHECK( reportStream.str().find( "<column number=\"3\" label=\"Uses\" type=\"INTEGER\" />" ) != string::npos );
		CHECK( reportStream.str().find( "<column number=\"4\" label=\"Value\" type=\"INTEGER\" />" ) != string::npos );
		CHECK( reportStream.str().find( "<line number=\"1\">" ) != string::npos );
		CHECK( reportStream.str().find( "<cell col=\"1\">CountTest</cell>" ) != string::npos );
		CHECK( reportStream.str().find( "<cell col=\"2\">25</cell>" ) != string::npos );
		CHECK( reportStream.str().find( "<cell col=\"3\">2</cell>" ) != string::npos );
		CHECK( reportStream.str().find( "<cell col=\"4\">26</cell>" ) != string::npos );
		CHECK( reportStream.str().find( "</line>" ) != string::npos );
		CHECK( reportStream.str().find( "</reporttable>" ) != string::npos );
	}

	/**
	 * @test odemx::XmlReport stream output
	 *
	 * Expected function call effects:
	 * @li the file contains the expected xml output
	 */
	TEST_FIXTURE( XmlReportFixture, FileOutput )
	{
		SimTime time = 36;
		Base::EventTest simAdvanceEvent( &sim, "XmlReportTestFileAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		unsigned int inc = 13;
		count.reset(); // non-0 reset time
		count.update( inc ); // two updates
		count.update( inc ); // result in value inc * 2

		fileReporter.addProducer( &count );
		fileReporter.generateReport();

		ostringstream fileStream;
		ifstream inFile( fileName.c_str() );
		string tmp;
		while( inFile )
		{
			getline( inFile, tmp );
			fileStream << tmp << endl;
		}

		CHECK( fileStream.str().find( "<reporttable label=\"Count Statistics\" columns=\"4\" lines=\"1\">" ) != string::npos );
		CHECK( fileStream.str().find( "<column number=\"1\" label=\"Name\" type=\"STRING\" />" ) != string::npos );
		CHECK( fileStream.str().find( "<column number=\"2\" label=\"Reset at\" type=\"SIMTIME\" />" ) != string::npos );
		CHECK( fileStream.str().find( "<column number=\"3\" label=\"Uses\" type=\"INTEGER\" />" ) != string::npos );
		CHECK( fileStream.str().find( "<column number=\"4\" label=\"Value\" type=\"INTEGER\" />" ) != string::npos );
		CHECK( fileStream.str().find( "<line number=\"1\">" ) != string::npos );
		CHECK( fileStream.str().find( "<cell col=\"1\">CountTest</cell>" ) != string::npos );
		CHECK( fileStream.str().find( "<cell col=\"2\">36</cell>" ) != string::npos );
		CHECK( fileStream.str().find( "<cell col=\"3\">2</cell>" ) != string::npos );
		CHECK( fileStream.str().find( "<cell col=\"4\">26</cell>" ) != string::npos );
		CHECK( fileStream.str().find( "</line>" ) != string::npos );
		CHECK( fileStream.str().find( "</reporttable>" ) != string::npos );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
