/**
 * @file TestXmlTrace.cpp
 * @date Aug 18, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class XmlTrace
 */

#include "TestDisplay.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Display )
{
/// @endcond

	/**
	 * @struct XmlTraceFixture
	 * @brief Helper struct providing set-up/tear-down of XmlTrace tests
	 *
	 * @copydetails EventFixture
	 */
	struct XmlTraceFixture
	{
		Base::SimulationTest trace;
		Util::TraceProducerTest producer;
		std::string fileName;
		ostringstream xmlStream;
		Time::German timeFormat;
		XmlTrace xmlStreamTrace;
		XmlTrace xmlFileTrace;

		XmlTraceFixture()
		:	trace( "XmlTraceTestSim" ),
			producer( trace, "XmlTraceProducerTest" ),
			fileName( "XmlTraceTestFile" ),
			xmlStream(),
			timeFormat( Time::Base( 1970, 1, 1, 0, Time::minutes ) ),
			xmlStreamTrace( &trace, xmlStream, timeFormat ),
			xmlFileTrace( &trace, fileName, Time::defaultFormat )
			{
				trace.addConsumer( &xmlStreamTrace );
				trace.addConsumer( &xmlFileTrace );
			}
	};

	string readXmlFile( const string& fileName )
	{
		ostringstream fileStream;
		ifstream inFile( ( fileName + "_0.xml" ).c_str() );

		if( ! inFile.is_open() )
		{
			cerr << "XmlTraceTest: cannot open XML file";
		}

		string tmp;
		while( inFile )
		{
			getline( inFile, tmp );
			fileStream << tmp << endl;
		}

		return fileStream.str();
	}

	SimTime advanceSim( Simulation& sim, SimTime t )
	{
		Base::EventTest simAdvanceEvent( &sim, "XmlTraceTestAdvanceEvent" );
		simAdvanceEvent.scheduleIn( t );
		sim.step(); // event execution advances sim time
		return t;
	}

	/**
	 * @test odemx::XmlTrace construction and destruction
	 *
	 * Expected effects:
	 * @li empty file names are handled
	 * @li XmlTrace can be constructed with streams or files
	 */
	TEST_FIXTURE( XmlTraceFixture, ConstructionDestruction )
	{
		unsigned int markEntriesPerFile = 1000;
		XmlTrace fileTrace1; // stdout
		XmlTrace fileTrace2( "XmlTraceTestConstructionDefSimFile" );
		XmlTrace fileTrace3( "XmlTraceTestConstructionDefSimFile", timeFormat );
		XmlTrace fileTrace4( "XmlTraceTestConstructionDefSimFile", timeFormat, markEntriesPerFile );

		XmlTrace fileTrace5( &trace ); // stdout
		XmlTrace fileTrace6( &trace, "XmlTraceTestConstructionUserSimFile" );
		XmlTrace fileTrace7( &trace, "XmlTraceTestConstructionUserSimFile", timeFormat );
		XmlTrace fileTrace8( &trace, "XmlTraceTestConstructionUserSimFile", timeFormat, markEntriesPerFile );

		XmlTrace streamTrace1( xmlStream );
		XmlTrace streamTrace2( xmlStream, timeFormat );
		XmlTrace streamTrace3( &trace, xmlStream );
		XmlTrace streamTrace4( &trace, xmlStream, timeFormat );
	}

	/**
	 * @test odemx::XmlTrace::start/stopTrace()
	 *
	 * Expected function call effects:
	 * @li stream and file contain the xml trace root with simulation label
	 * and start time
	 */
	TEST_FIXTURE( XmlTraceFixture, StartAndStop )
	{
		SimTime time = advanceSim( trace, 36231 );

		trace.startTrace();
		trace.stopTrace();

		CHECK( xmlStream.str().find( "<xmltrace" ) != string::npos );
		CHECK( xmlStream.str().find( trace.getLabel() ) != string::npos );
		CHECK( xmlStream.str().find( simTimeToString( time, timeFormat ) ) != string::npos );
		CHECK( xmlStream.str().find( "</xmltrace>" ) != string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( "<xmltrace" ) != string::npos );
		CHECK( fileStr.find( trace.getLabel() ) != string::npos );
		CHECK( fileStr.find( simTimeToString( time, Time::defaultFormat ) ) != string::npos );
		CHECK( fileStr.find( "</xmltrace>" ) != string::npos );
	}

	/**
	 * @test odemx::XmlTrace::pause/continueTrace()
	 *
	 * Expected function call effects:
	 * @li stream and file contain a pause mark sent by the Simulation
	 * with start time \c 0 and end time \c reactivationTime
	 */
	TEST_FIXTURE( XmlTraceFixture, PauseAndContinue )
	{
		trace.startTrace();
		trace.pauseTrace();

		SimTime reactivationTime = advanceSim( trace, 42424 );

		trace.continueTrace();
		trace.stopTrace();

		CHECK( xmlStream.str().find( "pause" ) != string::npos );
		CHECK( xmlStream.str().find( string("<sender type=\"") + typeToString( typeid(Base::SimulationTest) ) + "\">XmlTraceTestSim</sender>" ) != string::npos );
		CHECK( xmlStream.str().find( ( string("<tag name=\"start time\">")
				+ simTimeToString( 0, timeFormat ) + "</tag>" ) ) != string::npos );
		CHECK( xmlStream.str().find( ( string("<tag name=\"end time\">")
				+ simTimeToString( reactivationTime, timeFormat ) + "</tag>" ) ) != string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( "pause" ) != string::npos );
		CHECK( fileStr.find( string("<sender type=\"") + typeToString( typeid(Base::SimulationTest) ) + "\">XmlTraceTestSim</sender>" ) != string::npos );
		CHECK( fileStr.find( ( string("<tag name=\"start time\">")
				+ simTimeToString( 0, Time::defaultFormat ) + "</tag>" ) ) != string::npos );
		CHECK( fileStr.find( ( string("<tag name=\"end time\">")
				+ simTimeToString( reactivationTime, Time::defaultFormat ) + "</tag>" ) ) != string::npos );
	}

	/**
	 * @test odemx::XmlTrace::begin/endMark()
	 *
	 * Expected function call effects:
	 * @li the composed mark contains SimTime, mark name, sender details and
	 * the closing comment
	 */
	TEST_FIXTURE( XmlTraceFixture, BeginAndEndMark )
	{
		SimTime sendTime = advanceSim( trace, 55439 );
		trace.startTrace();

		// open composed mark
		trace.beginMark( &producer, Util::TraceProducerTest::markComposed );
		trace.endMark( "composed mark comment" );

		CHECK( xmlStream.str().find( string("time=\"") + simTimeToString( sendTime, timeFormat ) ) != string::npos );
		CHECK( xmlStream.str().find( "name=\"markComposed\"" ) != string::npos );
		CHECK( xmlStream.str().find( string("<sender type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerTest</sender>" )
				!= string::npos );
		CHECK( xmlStream.str().find( "<tag name=\"comment\">composed mark comment</tag>" ) != string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( string("time=\"") + simTimeToString( sendTime, Time::defaultFormat ) ) != string::npos );
		CHECK( fileStr.find( "name=\"markComposed\"" ) != string::npos );
		CHECK( fileStr.find( string("<sender type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerTest</sender>" )
				!= string::npos );
		CHECK( fileStr.find( "<tag name=\"comment\">composed mark comment</tag>" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::XmlTrace::mark(const TraceProducer*,MarkType,const string&)
	 *
	 * Expected function call effects:
	 * @li the simple mark contains SimTime, mark name, sender details and
	 * the closing comment
	 */
	TEST_FIXTURE( XmlTraceFixture, SimpleMark )
	{
		SimTime sendTime = advanceSim( trace, 4021 );
		trace.startTrace();

		trace.mark( &producer, Util::TraceProducerTest::markEvent, "test comment" );

		CHECK( xmlStream.str().find( string("time=\"") + simTimeToString( sendTime, timeFormat ) ) != string::npos );
		CHECK( xmlStream.str().find( "name=\"markEvent\"" ) != string::npos );
		CHECK( xmlStream.str().find( string("<sender type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerTest</sender>" )
				!= string::npos );
		CHECK( xmlStream.str().find( "<tag name=\"comment\">test comment</tag>" ) != string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( string("time=\"") + simTimeToString( sendTime, Time::defaultFormat ) ) != string::npos );
		CHECK( fileStr.find( "name=\"markEvent\"" ) != string::npos );
		CHECK( fileStr.find( string("<sender type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerTest</sender>" )
				!= string::npos );
		CHECK( fileStr.find( "<tag name=\"comment\">test comment</tag>" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::XmlTrace::mark(const TraceProducer*,MarkType,const TraceProducer*,const string&)
	 *
	 * Expected function call effects:
	 * @li the simple mark contains SimTime, mark name, sender details, partner
	 * details and the closing comment
	 */
	TEST_FIXTURE( XmlTraceFixture, SimpleMarkWithPartner )
	{
		Util::TraceProducerTest dummyPartner( trace, "XmlTraceProducerDummyPartner" );
		SimTime sendTime = advanceSim( trace, 4021 );
		trace.startTrace();

		trace.mark( &producer, Util::TraceProducerTest::markPartner, &dummyPartner, "test partner comment" );

		CHECK( xmlStream.str().find( string("time=\"") + simTimeToString( sendTime, timeFormat ) ) != string::npos );
		CHECK( xmlStream.str().find( "name=\"markPartner\"" ) != string::npos );
		CHECK( xmlStream.str().find( string("<sender type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerTest</sender>" )
				!= string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"partner\" type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerDummyPartner</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( "<tag name=\"comment\">test partner comment</tag>" ) != string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( string("time=\"") + simTimeToString( sendTime, Time::defaultFormat ) ) != string::npos );
		CHECK( fileStr.find( "name=\"markPartner\"" ) != string::npos );
		CHECK( fileStr.find( string("<sender type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerTest</sender>" )
				!= string::npos );
		CHECK( fileStr.find( string("<tag name=\"partner\" type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerDummyPartner</tag>" ) != string::npos );
		CHECK( fileStr.find( "<tag name=\"comment\">test partner comment</tag>" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::XmlTrace::mark(const TraceProducer*,MarkType,const TraceProducer*,const TraceProducer*,const string&)
	 *
	 * Expected function call effects:
	 * @li the simple mark contains SimTime, mark name, sender details, both
	 * partners and the closing comment
	 */
	TEST_FIXTURE( XmlTraceFixture, SimpleMarkWithTwoPartners )
	{
		Util::TraceProducerTest dummyPartner1( trace, "XmlTraceProducerDummyPartner1" );
		Util::TraceProducerTest dummyPartner2( trace, "XmlTraceProducerDummyPartner2" );
		SimTime sendTime = advanceSim( trace, 120621 );
		trace.startTrace();

		trace.mark( &producer, Util::TraceProducerTest::markTwoPartners, &dummyPartner1, &dummyPartner2, "test two partners comment" );

		CHECK( xmlStream.str().find( string("time=\"") + simTimeToString( sendTime, timeFormat ) ) != string::npos );
		CHECK( xmlStream.str().find( "name=\"markTwoPartners\"" ) != string::npos );
		CHECK( xmlStream.str().find( string("<sender type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerTest</sender>" )
				!= string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"first partner\" type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerDummyPartner1</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"second partner\" type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerDummyPartner2</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( "<tag name=\"comment\">test two partners comment</tag>" ) != string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( string("time=\"") + simTimeToString( sendTime, Time::defaultFormat ) ) != string::npos );
		CHECK( fileStr.find( "name=\"markTwoPartners\"" ) != string::npos );
		CHECK( fileStr.find( string("<sender type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerTest</sender>" )
				!= string::npos );
		CHECK( fileStr.find( string("<tag name=\"first partner\" type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerDummyPartner1</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"second partner\" type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerDummyPartner2</tag>" ) != string::npos );
		CHECK( fileStr.find( "<tag name=\"comment\">test two partners comment</tag>" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::XmlTrace::mark(const TraceProducer*,MarkType,AttributeType,AttributeType,const string&)
	 *
	 * Expected function call effects:
	 * @li the simple mark contains SimTime, mark name, sender details, tags
	 * for the old and new values, and the closing comment
	 */
	TEST_FIXTURE( XmlTraceFixture, SimpleMarkAttributeChange )
	{
		SimTime sendTime = advanceSim( trace, 220621 );
		trace.startTrace();

		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, true, false, "test attribute comment" );

		CHECK( xmlStream.str().find( string("time=\"") + simTimeToString( sendTime, timeFormat ) ) != string::npos );
		CHECK( xmlStream.str().find( "name=\"markAttributeChange\"" ) != string::npos );
		CHECK( xmlStream.str().find( string("<sender type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerTest</sender>" )
				!= string::npos );
		CHECK( xmlStream.str().find( "<tag name=\"new value\">1</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( "<tag name=\"old value\">0</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( "<tag name=\"comment\">test attribute comment</tag>" ) != string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( string("time=\"") + simTimeToString( sendTime, Time::defaultFormat ) ) != string::npos );
		CHECK( fileStr.find( "name=\"markAttributeChange\"" ) != string::npos );
		CHECK( fileStr.find( string("<sender type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">XmlTraceProducerTest</sender>" )
				!= string::npos );
		CHECK( fileStr.find( "<tag name=\"new value\">1</tag>" ) != string::npos );
		CHECK( fileStr.find( "<tag name=\"old value\">0</tag>" ) != string::npos );
		CHECK( fileStr.find( "<tag name=\"comment\">test attribute comment</tag>" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::XmlTrace::mark(const TraceProducer*,MarkType,AttributeType,AttributeType,const string&)
	 *
	 * Expected function call effects:
	 * @li attribute changes for built-in types are correctly transformed and
	 * stored in XML tags
	 */
	TEST_FIXTURE( XmlTraceFixture, SimpleMarkAttributeTypes )
	{
		trace.startTrace();

		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, true, false );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, 'a', 'b' );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(short), MAX_VAL(short) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(int), MAX_VAL(int) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(long), MAX_VAL(long) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(long long), MAX_VAL(long long) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(float), MAX_VAL(float) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(double), MAX_VAL(double) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, "const char*", "const std::string&" );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, (unsigned char)'^', (unsigned char)'#' );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned short), MAX_VAL(unsigned short) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned int), MAX_VAL(unsigned int) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned long), MAX_VAL(unsigned long) );

		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( true ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( false ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( 'a' ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( 'b' ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( MIN_VAL(short) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( MAX_VAL(short) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( MIN_VAL(int) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( MAX_VAL(int) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( MIN_VAL(long) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( MAX_VAL(long) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( MIN_VAL(long long) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( MAX_VAL(long long) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( MIN_VAL(float) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( MAX_VAL(float) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( MIN_VAL(double) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( MAX_VAL(double) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString(	"const char*" ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( "const std::string&" ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString(	"^" ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( "#" ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( MIN_VAL(unsigned short) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( MAX_VAL(unsigned short) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( MIN_VAL(unsigned int) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( MAX_VAL(unsigned int) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"new value\">") + toString( MIN_VAL(unsigned long) ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"old value\">") + toString( MAX_VAL(unsigned long) ) + "</tag>" ) != string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( true ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( false ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( 'a' ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( 'b' ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( MIN_VAL(short) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( MAX_VAL(short) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( MIN_VAL(int) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( MAX_VAL(int) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( MIN_VAL(long) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( MAX_VAL(long) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( MIN_VAL(long long) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( MAX_VAL(long long) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( MIN_VAL(float) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( MAX_VAL(float) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( MIN_VAL(double) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( MAX_VAL(double) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString(	"const char*" ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( "const std::string&" ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString(	"^" ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( "#" ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( MIN_VAL(unsigned short) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( MAX_VAL(unsigned short) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( MIN_VAL(unsigned int) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( MAX_VAL(unsigned int) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"new value\">") + toString( MIN_VAL(unsigned long) ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"old value\">") + toString( MAX_VAL(unsigned long) ) + "</tag>" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::XmlTrace::begin/endTag(Tag)
	 *
	 * Expected function call effects:
	 * @li the composed tag is correctly represented in XML
	 */
	TEST_FIXTURE( XmlTraceFixture, BeginAndEndTag )
	{
		trace.startTrace();

		trace.beginMark( &producer, Util::TraceProducerTest::markComposed );
		trace.beginTag( Util::TraceProducerTest::tagComposed );
		trace.endTag( Util::TraceProducerTest::tagComposed );
		trace.endMark();

		CHECK( xmlStream.str().find( "<tag name=\"tagComposed\"></tag>" ) != string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( "<tag name=\"tagComposed\"></tag>" ) != string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::XmlTrace::addTag(Tag, TagValueType)
	 *
	 * Expected function call effects:
	 * @li tags for built-in types are correctly transformed and
	 * stored in XML tags
	 */
	TEST_FIXTURE( XmlTraceFixture, AddTagValueTypes )
	{
		trace.startTrace();
		trace.beginMark( &producer, Util::TraceProducerTest::markComposed );

		bool value1 = true;
		char value2 = '?';
		short value3 = 1;
		int value4 = 1;
		long value5 = 1;
		long long value6 = 1;
		float value7 = 1.1f;
		double value8 = 1.1;
		const char* value9 = "value";
		const TraceProducer* value10 = &producer;
		unsigned char value11 = '1';
		unsigned short value12 = 1;
		unsigned int value13 = 1;
		unsigned long value14 = 1;

		trace.addTag( Util::TraceProducerTest::tagAttribute, value1 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value2 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value3 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value4 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value5 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value6 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value7 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value8 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value9 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value10 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value11 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value12 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value13 );
		trace.addTag( Util::TraceProducerTest::tagAttribute, value14 );

		trace.endMark();
		trace.stopTrace();

		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + "true" + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value2 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value3 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value4 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value5 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value6 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value7 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value8 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value9 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\" type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">" + producer.getLabel() + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value11 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value12 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value13 ) + "</tag>" ) != string::npos );
		CHECK( xmlStream.str().find( string("<tag name=\"tagAttribute\">") + toString( value14 ) + "</tag>" ) != string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + "true" + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value2 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value3 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value4 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value5 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value6 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value7 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value8 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value9 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\" type=\"") + typeToString( typeid(Util::TraceProducerTest) ) + "\">" + producer.getLabel() + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value11 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value12 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value13 ) + "</tag>" ) != string::npos );
		CHECK( fileStr.find( string("<tag name=\"tagAttribute\">") + toString( value14 ) + "</tag>" ) != string::npos );
	}

	/**
	 * @test odemx::XmlTrace mark blocking
	 *
	 * The Filter is set to block all marks. Expected effects:
	 * @li no marks are written to XML stream or file
	 */
	TEST_FIXTURE( XmlTraceFixture, PassNone )
	{
		Util::TraceProducerTest dummyPartner1( trace, "XmlTraceProducerDummyPartner1" );
		Util::TraceProducerTest dummyPartner2( trace, "XmlTraceProducerDummyPartner2" );

		trace.startTrace();
		xmlStreamTrace.passNone();
		xmlFileTrace.passNone();

		trace.mark( &producer, Util::TraceProducerTest::markEvent, "test comment" );
		trace.mark( &producer, Util::TraceProducerTest::markPartner, &dummyPartner1, "test partner comment" );
		trace.mark( &producer, Util::TraceProducerTest::markTwoPartners, &dummyPartner1, &dummyPartner2, "test two partners comment" );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, true, false );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, 'a', 'b' );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(short), MAX_VAL(short) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(int), MAX_VAL(int) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(long), MAX_VAL(long) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(long long), MAX_VAL(long long) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(float), MAX_VAL(float) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(double), MAX_VAL(double) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, "const char*", "const std::string&" );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, (unsigned char)'^', (unsigned char)'#' );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned short), MAX_VAL(unsigned short) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned int), MAX_VAL(unsigned int) );
		trace.mark( &producer, Util::TraceProducerTest::markAttributeChange, MIN_VAL(unsigned long), MAX_VAL(unsigned long) );

		CHECK( xmlStream.str().find( "mark" ) == string::npos );
		string fileStr = readXmlFile( fileName );
		CHECK( fileStr.find( "mark" ) == string::npos );

		trace.stopTrace();
	}

	/**
	 * @test odemx::XmlTrace::printState()
	 *
	 * Expected call effects:
	 * @li all possible Process states are printed as strings in the trace
	 * @li there is no error string in the trace as this statement should never
	 * be reached
	 */
	TEST_FIXTURE( XmlTraceFixture, PrintProcessState )
	{
		trace.startTrace();

		Base::ProcessTest process( &trace, "ProcessTest" );
		process.activate();
		process.sleep();
		process.interrupt();
		trace.run();

		CHECK( xmlStream.str().find( "CREATED" ) != string::npos );
		CHECK( xmlStream.str().find( "IDLE" ) != string::npos );
		CHECK( xmlStream.str().find( "RUNNABLE" ) != string::npos );
		CHECK( xmlStream.str().find( "CURRENT" ) != string::npos );
		CHECK( xmlStream.str().find( "TERMINATED" ) != string::npos );
		CHECK( xmlStream.str().find( "Trace ERROR" ) == string::npos );

		string fileStr = readXmlFile( fileName );

		CHECK( fileStr.find( "CREATED" ) != string::npos );
		CHECK( fileStr.find( "IDLE" ) != string::npos );
		CHECK( fileStr.find( "RUNNABLE" ) != string::npos );
		CHECK( fileStr.find( "CURRENT" ) != string::npos );
		CHECK( fileStr.find( "TERMINATED" ) != string::npos );
		CHECK( fileStr.find( "Trace ERROR" ) == string::npos );

		trace.stopTrace();
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
