/**
 * @file TestNetTopology.cpp
 * @date Jul 16, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class NetTopology
 */

#include "TestProtocol.h"
#include <stdexcept>

/// @cond DOXYGEN_SKIP
SUITE( Protocol )
{
/// @endcond

	/**
	 * @struct NetTopologyFixture
	 * @brief Helper struct providing set-up/tear-down of NetTopology tests
	 *
	 * @copydetails EventFixture
	 */
	struct NetTopologyFixture
	{
		ProtocolSimulation sim;
		NetTopologyTestObserver observer;
		NetTopologyTest topology;
		ProtocolStackTest node1;
		ProtocolStackTest node2;
		ProtocolStackTest node3;

		NetTopologyFixture()
		:	observer( false ),
			topology( &sim, "NetTopologyTest", &observer ),
			node1( &sim, "NetTopologyTestNode1" ),
			node2( &sim, "NetTopologyTestNode2" ),
			node3( &sim, "NetTopologyTestNode3" )
			{}
	};

	/**
	 * @test odemx::NetTopology construction
	 *
	 * Test correct member initialization
	 * @li simulation
	 * @li observer
	 */
	TEST_FIXTURE( NetTopologyFixture, ConstructionDestruction )
	{
		CHECK_EQUAL( &sim, topology.getSimulation() );
		CHECK_EQUAL( &observer, topology.getObservers().front() );
	}

	/**
	 * @test odemx::NetTopology::registerNode(ProtocolStack*,double,double)
	 *
	 * Expected function call effects:
	 * @li the function call is observed
	 * @li the node is found in the vector containing all nodes
	 * @li the node's location is stored correctly
	 */
	TEST_FIXTURE( NetTopologyFixture, RegisterNode )
	{
		double locX = 1.5;
		double locY = 2.7;
		observer.history.clear();

		topology.registerNode( &node1, locX, locY );
		CHECK( observer.observed( "onRegisterNode", topology.getLabel(), node1.getLabel(), "(1.5, 2.7)" ) );

		const NetworkMap& nodes = topology.getAllNodes();
		CHECK( nodes.find( node1.getNodeId() ) != nodes.end() );
		Location loc = topology.getNodeLocation( &node1 );
		CHECK_EQUAL( locX, loc.x );
		CHECK_EQUAL( locY, loc.y );
	}

	/**
	 * @test odemx::NetTopology::removeNode(ProtocolStack*)
	 *
	 * Expected function call effects:
	 * @li removing a non-registered node sends an error message to
	 * \c errorStream()
	 * @li the function call is observed by a NetTopologyObserver
	 * @li the node is not found in the vector containing all nodes
	 * @li the node's link map is empty
	 * @li the node was removed from all other nodes' link maps
	 */
	TEST_FIXTURE( NetTopologyFixture, RemoveNode )
	{
		clearErrorStream();
		topology.removeNode( &node1 );
		CHECK_EQUAL( "ODEMx ERROR: NetTopology::removeNode(): node not stored in list -- Object: NetTopologyTest\n" , \
				errorStream().str() );

		const NetworkMap& nodes = topology.getAllNodes();
		topology.registerNode( &node1, 1, 1 );
		topology.registerNode( &node2, 1, 1 );
		topology.registerNode( &node3, 1, 1 );
		topology.setLink( &node2, &node1, 1 );
		topology.setLink( &node3, &node1, 1 );
		const map< ProtocolStack*, LinkInfo >& node2links = topology.getLinks( &node2 );
		const map< ProtocolStack*, LinkInfo >& node3links = topology.getLinks( &node3 );
		CHECK( node2links.find( &node1 ) != node2links.end() );
				CHECK( node3links.find( &node1 ) != node3links.end() );

		topology.removeNode( &node1 );
		CHECK( observer.observed( "onRemoveNode", topology.getLabel(), node1.getLabel() ) );
		CHECK( nodes.find( node1.getNodeId() ) == nodes.end() );
		CHECK_THROW( topology.getLinks( &node1 ), std::runtime_error );
		CHECK( node2links.find( &node1 ) == node2links.end() );
		CHECK( node3links.find( &node1 ) == node3links.end() );
	}

	/**
	 * @test odemx::NetTopology::linked(ProtocolStack*,ProtocolStack*)
	 *
	 * Precondition: \c setLink() was called with \c node1 and \c node2.
	 *
	 * Expected function call effects:
	 * @li returns true for check with \c node1 and \c node2
	 * @li returns false for check with \c node1 and \c node3
	 */
	TEST_FIXTURE( NetTopologyFixture, Linked )
	{
		topology.registerNode( &node1, 1, 1 );
		topology.registerNode( &node2, 1, 1 );
		topology.setLink( &node1, &node2, 1 );

		CHECK( topology.linked( &node1, &node2 ) );
		CHECK( ! topology.linked( &node1, &node3 ) );
	}

	/**
	 * @test odemx::NetTopology::removeLink(ProtocolStack*,ProtocolStack*)
	 *
	 * Precondition: \c setLink() was called with \c node1 and \c node2.
	 *
	 * Expected function call effects:
	 * @li throws a runtime error when one of the nodes is not registered
	 * @li call with nodes that aren't linked produces warnings
	 * @li the call was observed
	 * @li there is no link between \c node1 and \c node2
	 */
	TEST_FIXTURE( NetTopologyFixture, RemoveLink )
	{
		topology.registerNode( &node1, 3, 3 );
		clearErrorStream();
		CHECK_THROW( topology.removeLink( &node1, &node2 ), std::runtime_error );
		CHECK_EQUAL( string("ODEMx WARNING: NetTopology::removeLink(): no previous link between nodes: ")
					 + "NetTopologyTestNode1, NetTopologyTestNode2 -- Object: NetTopologyTest\n"
					 + "ODEMx ERROR: NetTopology::getLinks(): NetTopologyTestNode2 not found, "
					 + "is it registered? -- Object: NetTopologyTest\n", \
				errorStream().str() );

		topology.registerNode( &node2, 4, 5 );
		topology.setLink( &node1, &node2, 1 );
		topology.removeLink( &node1, &node2 );

		CHECK( observer.observed( "onRemoveLink", topology.getLabel(), node1.getLabel(), node2.getLabel() ) );
		CHECK( ! topology.linked( &node1, &node2 ) );
		CHECK( ! topology.linked( &node2, &node1 ) );
	}

	/**
	 * @test odemx::NetTopology::removeLink(Label,Label)
	 *
	 * This function calls removeLink(ProtocolStack*,ProtocolStack*). Only
	 * the Label part is tested here. Expected function call effects:
	 * @li invalid or unregistered Labels cause error messages to be sent to
	 * \c errorStream()
	 * @li there is a link between \c node1 and \c node2
	 */
	TEST_FIXTURE( NetTopologyFixture, RemoveLinkLabeled )
	{
		topology.registerNode( &node1, 1, 1 );
		topology.registerNode( &node2, 1, 1 );
		topology.registerNode( &node3, 1, 1 );
		double connectivity = 1;

		topology.setLink( &node1, &node2, connectivity );
		CHECK( topology.linked( &node1, &node2 ) );
		CHECK( topology.linked( &node2, &node1 ) );

		clearErrorStream();
		topology.removeLink( node3.getLabel(), node2.getLabel() );
		CHECK_EQUAL( string("ODEMx WARNING: NetTopology::removeLink(): no previous ")
				+ "link between nodes: NetTopologyTestNode3, NetTopologyTestNode2 -- "
				+ "Object: NetTopologyTest\n",
				errorStream().str() );
		CHECK( ! topology.linked( &node3, &node2 ) );

		clearErrorStream();
		topology.removeLink( node1.getLabel(), node3.getLabel() );
		CHECK_EQUAL( string("ODEMx WARNING: NetTopology::removeLink(): no previous ")
				+ "link between nodes: NetTopologyTestNode1, NetTopologyTestNode3 -- "
				+ "Object: NetTopologyTest\n",
				errorStream().str() );
		CHECK( ! topology.linked( &node1, &node3 ) );

		topology.removeLink( node1.getLabel(), node2.getLabel() );
		CHECK( ! topology.linked( &node1, &node2 ) );
		CHECK( ! topology.linked( &node2, &node1 ) );
	}

	/**
	 * @test odemx::NetTopology::setLink(ProtocolStack*,ProtocolStack*,double)
	 *
	 * Expected function call effects:
	 * @li invalid probability values cause error messages to be sent to
	 * \c errorStream()
	 * @li with proper arguments, two calls to \c makeLink() are observed
	 * @li calling again with the same pair of nodes causes
	 * two calls to \c changeLink(), which can be observed
	 * @li there exists a link between \c node1 and \c node2
	 */
	TEST_FIXTURE( NetTopologyFixture, SetLink )
	{
		topology.registerNode( &node1, 1, 1 );
		topology.registerNode( &node2, 1, 1 );
		double connectivity;

		clearErrorStream();
		connectivity = 1.01;
		topology.setLink( &node1, &node2, connectivity );
		CHECK_EQUAL( "ODEMx ERROR: NetTopology::validProbability(): connectivity greater 1 -- Object: NetTopologyTest\n", \
				errorStream().str() );

		clearErrorStream();
		connectivity = -0.01;
		topology.setLink( &node1, &node2, connectivity );
		CHECK_EQUAL( "ODEMx ERROR: NetTopology::validProbability(): negative connectivity value -- Object: NetTopologyTest\n", \
				errorStream().str() );

		CHECK( ! topology.linked( &node1, &node2 ) );
		CHECK( ! topology.linked( &node2, &node1 ) );

		connectivity = 1;
		topology.setLink( &node1, &node2, connectivity );
		CHECK( observer.observed( "onMakeLink", topology.getLabel(), node1.getLabel() + "->" + node2.getLabel(), toString( connectivity ) ) );
		CHECK( observer.observed( "onMakeLink", topology.getLabel(), node2.getLabel() + "->" + node1.getLabel(), toString( connectivity ) ) );
		CHECK( topology.linked( &node1, &node2 ) );
		CHECK( topology.linked( &node2, &node1 ) );

		connectivity = 0.5;
		topology.setLink( &node1, &node2, connectivity );
		CHECK( observer.observed( "onChangeLink", topology.getLabel(), node1.getLabel() + "->" + node2.getLabel(), toString( connectivity ) ) );
		CHECK( observer.observed( "onChangeLink", topology.getLabel(), node2.getLabel() + "->" + node1.getLabel(), toString( connectivity ) ) );
		CHECK( topology.linked( &node1, &node2 ) );
		CHECK( topology.linked( &node2, &node1 ) );
	}

	/**
	 * @test odemx::NetTopology::setLink(Label,Label,double)
	 *
	 * This function calls setLink(ProtocolStack*,ProtocolStack*,double). Only
	 * the Label part is tested here. Expected function call effects:
	 * @li invalid or unregistered Labels cause error messages to be sent to
	 * \c errorStream()
	 * @li there is a link between \c node1 and \c node2
	 */
	TEST_FIXTURE( NetTopologyFixture, SetLinkLabeled )
	{
		topology.registerNode( &node1, 1, 1 );
		topology.registerNode( &node2, 1, 1 );
		//topology.registerNode( &node3, 1, 1 );
		double connectivity = 1;

		clearErrorStream();
		topology.setLink( node3.getLabel(), node2.getLabel(), connectivity );
		CHECK_EQUAL( "ODEMx ERROR: NetTopology::setLink(): cannot determine node by Label, is it registered? -- Object: NetTopologyTest\n", \
				errorStream().str() );
		CHECK( ! topology.linked( &node2, &node3 ) );

		clearErrorStream();
		topology.setLink( node1.getLabel(), node3.getLabel(), connectivity );
		CHECK_EQUAL( "ODEMx ERROR: NetTopology::setLink(): cannot determine node by Label, is it registered? -- Object: NetTopologyTest\n", \
				errorStream().str() );
		CHECK( ! topology.linked( &node1, &node3 ) );

		topology.setLink( node1.getLabel(), node2.getLabel(), connectivity );
		CHECK( topology.linked( &node1, &node2 ) );
		CHECK( topology.linked( &node2, &node1 ) );
	}

	/**
	 * @test odemx::NetTopology::setLink(ProtocolStack*,ProtocolStack*,double,double)
	 *
	 * Expected function call effects:
	 * @li invalid probability values cause error messages to be sent to
	 * \c errorStream()
	 * @li with proper arguments, two calls to \c makeLink() are observed
	 * with the proper connectivity values
	 * @li calling again with the same pair of nodes causes
	 * two calls to \c changeLink(), which can be observed
	 * @li there exists a link between \c node1 and \c node2
	 */
	TEST_FIXTURE( NetTopologyFixture, SetLinkEach )
	{
		topology.registerNode( &node1, 1, 1 );
		topology.registerNode( &node2, 1, 1 );
		double connectivity;

		clearErrorStream();
		connectivity = 1.01;
		topology.setLink( &node1, &node2, connectivity, 1 );
		CHECK_EQUAL( "ODEMx ERROR: NetTopology::validProbability(): connectivity greater 1 -- Object: NetTopologyTest\n", \
				errorStream().str() );

		clearErrorStream();
		connectivity = -0.01;
		topology.setLink( &node1, &node2, 1, connectivity );
		CHECK_EQUAL( "ODEMx ERROR: NetTopology::validProbability(): negative connectivity value -- Object: NetTopologyTest\n", \
				errorStream().str() );

		CHECK( ! topology.linked( &node1, &node2 ) );
		CHECK( ! topology.linked( &node2, &node1 ) );

		topology.setLink( &node1, &node2, 0.1, 0.2 );
		CHECK( observer.observed( "onMakeLink", topology.getLabel(), node1.getLabel() + "->" + node2.getLabel(), toString( 0.1 ) ) );
		CHECK( observer.observed( "onMakeLink", topology.getLabel(), node2.getLabel() + "->" + node1.getLabel(), toString( 0.2 ) ) );
		CHECK( topology.linked( &node1, &node2 ) );
		CHECK( topology.linked( &node2, &node1 ) );

		topology.setLink( &node1, &node2, 0.3, 0.4 );
		CHECK( observer.observed( "onChangeLink", topology.getLabel(), node1.getLabel() + "->" + node2.getLabel(), toString( 0.3 ) ) );
		CHECK( observer.observed( "onChangeLink", topology.getLabel(), node2.getLabel() + "->" + node1.getLabel(), toString( 0.4 ) ) );
		CHECK( topology.linked( &node1, &node2 ) );
		CHECK( topology.linked( &node2, &node1 ) );
	}

	/**
	 * @test odemx::NetTopology::setLink(Label,Label,double,double)
	 *
	 * This function calls setLink(ProtocolStack*,ProtocolStack*,double,double). Only
	 * the Label part is tested here. Expected function call effects:
	 * @li invalid or unregistered Labels cause error messages to be sent to
	 * \c errorStream()
	 * @li there is a link between \c node1 and \c node2
	 */
	TEST_FIXTURE( NetTopologyFixture, SetLinkEachLabeled )
	{
		topology.registerNode( &node1, 1, 1 );
		topology.registerNode( &node2, 1, 1 );
		double connectivity = 1;

		clearErrorStream();
		topology.setLink( node3.getLabel(), node2.getLabel(), connectivity, connectivity );
		CHECK_EQUAL( "ODEMx ERROR: NetTopology::setLink(): cannot determine node by Label, is it registered? -- Object: NetTopologyTest\n", \
				errorStream().str() );
		CHECK( ! topology.linked( &node2, &node3 ) );

		clearErrorStream();
		topology.setLink( node1.getLabel(), node3.getLabel(), connectivity, connectivity );
		CHECK_EQUAL( "ODEMx ERROR: NetTopology::setLink(): cannot determine node by Label, is it registered? -- Object: NetTopologyTest\n", \
				errorStream().str() );
		CHECK( ! topology.linked( &node1, &node3 ) );

		topology.setLink( node1.getLabel(), node2.getLabel(), connectivity, connectivity );
		CHECK( topology.linked( &node1, &node2 ) );
		CHECK( topology.linked( &node2, &node1 ) );
	}

	/**
	 * @test odemx::NetTopology::setNodeLocation(ProtocolStack*,double.double)
	 *
	 * Expected function call effects:
	 * @li the same location is returned by \c getNodeLocation()
	 */
	TEST_FIXTURE( NetTopologyFixture, SetNodeLocation )
	{
		topology.registerNode( &node1, 1, 1 );
		double locationX = 2.5,
			   locationY = 3.6;

		topology.setNodeLocation( &node1, locationX, locationY );
		Location loc = topology.getNodeLocation( &node1 );
		CHECK_EQUAL( locationX, loc.x );
		CHECK_EQUAL( locationY, loc.y );
	}

	/**
	 * @test odemx::NetTopology::setNodeLocation(ProtocolStack*,double.double)
	 *
	 * Expected function call effects:
	 * @li
	 */
	TEST_FIXTURE( NetTopologyFixture, GetNodeById )
	{
		topology.registerNode( &node1, 1, 1 );
		double locationX = 2.5,
			   locationY = 3.6;

		topology.setNodeLocation( &node1, locationX, locationY );
		Location loc = topology.getNodeLocation( &node1 );
		CHECK_EQUAL( locationX, loc.x );
		CHECK_EQUAL( locationY, loc.y );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
