/**
 * @file TestProtocolEntity.cpp
 * @date Jul 18, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class ProtocolEntity
 */

#include "TestProtocol.h"

/// @cond DOXYGEN_SKIP
SUITE( Protocol )
{
/// @endcond

	/**
	 * @struct ProtocolEntityFixture
	 * @brief Helper struct providing set-up/tear-down of ProtocolEntity tests
	 *
	 * @copydetails EventFixture
	 */
	struct ProtocolEntityFixture
	{
		// make sure the observer is deleted last because many
		// destructors still send observations
		ProtocolEntityTestObserver observer;
		ProtocolSimulation sim;
		ProtocolStackTest* node;
		ProtocolEntityTest* topEntity;
		ProtocolEntityTest* bottomEntity;
		ProtocolLayer* topLayer;
		ProtocolLayer* bottomLayer;

		ProtocolEntityFixture()
		:	observer( false ),
			sim( "ProtocolEntityTestSim" ),
			node( 0 ),
			topEntity( 0 ),
			bottomEntity( 0 ),
			topLayer( 0 ),
			bottomLayer( 0 )
		{
			sim.createMedium< TransmissionMediumTest >( "ProtocolEntityTestMedium" );
			sim.createLayer< ProtocolEntityTest >( "TestTopLayer" );
			sim.createLayer< ProtocolEntityTest >( "TestBottomLayer" );

			topLayer = sim.getLayerList().front();
			bottomLayer = sim.getLayerList().back();

			// this creates two entities of type ProtocolEntityTest,
			// one for each layer, and schedules them with hold()
			node = new ProtocolStackTest( &sim, "ProtocolEntityTestNode" );

			topEntity = dynamic_cast< ProtocolEntityTest* >( node -> getEntityByLayer( topLayer ) );
			topEntity -> Observable<ProcessObserver>::addObserver( &observer );
			topEntity -> Observable<ProtocolEntityObserver>::addObserver( &observer );

			bottomEntity = dynamic_cast< ProtocolEntityTest* >( node -> getEntityByLayer( bottomLayer ) );
			bottomEntity -> Observable<ProcessObserver>::addObserver( &observer );
			bottomEntity -> Observable<ProtocolEntityObserver>::addObserver( &observer );
		}

		~ProtocolEntityFixture()
		{
			if ( node ) delete node;
		}
	};

	/**
	 * @test odemx::ProtocolEntity construction
	 *
	 * Checks correct member initialization for both entities:
	 * @li owner node
	 * @li label
	 * @li simulation context
	 * @li transmission medium
	 */
	TEST_FIXTURE( ProtocolEntityFixture, ConstructionDestruction )
	{
		CHECK_EQUAL( node, topEntity -> getNode() );
		CHECK_EQUAL( node -> getLabel() + " " + topLayer -> getLabel() + " Entity",\
				topEntity -> getLabel() );
		CHECK_EQUAL( &sim, topEntity -> getSimulation() );
		CHECK_EQUAL( sim.getTransmissionMedium(), topEntity -> getTransmissionMedium() );

		CHECK_EQUAL( node, bottomEntity -> getNode() );
		CHECK_EQUAL( node -> getLabel() + " " + bottomLayer -> getLabel() + " Entity",\
				bottomEntity -> getLabel() );
		CHECK_EQUAL( &sim, bottomEntity -> getSimulation() );
		CHECK_EQUAL( sim.getTransmissionMedium(), bottomEntity -> getTransmissionMedium() );
	}

	/**
	 * @test odemx::ProtocolEntity::main()
	 *
	 * Precondition: the entity is scheduled. Simulation will execute \c main().
	 *
	 * Expected function call effects:
	 * @li once the entity starts, it calls \c wait() to receive
	 * messages from its ports
	 * @li when data is put into one of its ports, it immediately responds
	 * by calling the appropriate handler
	 * @li after calling the handler, the entity calls \c wait() again
	 */
	TEST_FIXTURE( ProtocolEntityFixture, Main )
	{
		// remove bottomEntity from the schedule so it doesn't interfere
		bottomEntity -> sleep();
		observer.history.clear();

		ProtocolEntityTestData data1( "TestTopData" ),
							   data2( "TestBottomData" );

		sim.step();
		CHECK( observer.observed( "onWait", topEntity -> getLabel(), \
				"ProtocolEntityTestNode TestTopLayer Entity Top Port (head)+ProtocolEntityTestNode TestTopLayer Entity Bottom Port (head)" ) );
		observer.history.clear();

		topEntity -> putTop( &data1 );
		sim.step();
		CHECK( observer.observed( "onReadTop", topEntity -> getLabel(), "TestTopData" ) );
		CHECK( topEntity -> handledTopInput );
		CHECK( observer.observed( "onWait", topEntity -> getLabel(), \
				"ProtocolEntityTestNode TestTopLayer Entity Top Port (head)+ProtocolEntityTestNode TestTopLayer Entity Bottom Port (head)" ) );
		observer.history.clear();

		topEntity -> putBottom( &data2 );
		sim.step();
		CHECK( observer.observed( "onReadBottom", topEntity -> getLabel(), "TestBottomData" ) );
		CHECK( topEntity -> handledBottomInput );
		CHECK( observer.observed( "onWait", topEntity -> getLabel(), \
				"ProtocolEntityTestNode TestTopLayer Entity Top Port (head)+ProtocolEntityTestNode TestTopLayer Entity Bottom Port (head)" ) );
	}

	/**
	 * @test odemx::ProtocolEntity::transmit(PortData*) to the next layer below
	 *
	 * Expected function call effects:
	 * @li the \c transmit() call was observed
	 * @li the next lower entity called \c readTop() because this entity is not
	 * the lowest layer entity
	 */
	TEST_FIXTURE( ProtocolEntityFixture, TransmitToNextLayer )
	{
		// move bottomEntity to front in schedule
		bottomEntity -> activate();

		TransmissionMediumTest* medium =
			dynamic_cast< TransmissionMediumTest* >( sim.getTransmissionMedium() );

		ProtocolEntityTestData data( "TransmitTestData" );

		topEntity -> transmit( &data );
		CHECK( observer.observed( "onTransmit", topEntity -> getLabel(), data.toString() ) );
		CHECK( ! medium -> transmitCalled );

		sim.step();
		CHECK( observer.observed( "onReadTop", bottomEntity -> getLabel(), data.toString() ) );
	}

	/**
	 * @test odemx::ProtocolEntity::transmit(PortData*) to TransmissionMedium
	 *
	 * Expected function call effects:
	 * @li the call was observed
	 * @li \c TransmissionMedium::transmit() was called because this entity is
	 * the lowest layer entity
	 */
	TEST_FIXTURE( ProtocolEntityFixture, TransmitToMedium )
	{
		TransmissionMediumTest* medium =
			dynamic_cast< TransmissionMediumTest* >( sim.getTransmissionMedium() );

		ProtocolEntityTestData data( "TransmitTestData" );

		bottomEntity -> transmit( &data );
		CHECK( observer.observed( "onTransmit", bottomEntity -> getLabel(), data.toString() ) );
		CHECK( medium -> transmitCalled );
	}

	/**
	 * @test odemx::ProtocolEntity::passUpward(PortData*) to the next layer above
	 *
	 * Expected function call effects:
	 * @li the call was observed
	 * @li the next higher entity called \c readBottom() because this entity
	 * is not the top layer entity
	 */
	TEST_FIXTURE( ProtocolEntityFixture, PassUpwardToNextLayer )
	{
		ProtocolEntityTestData data( "PassUpwardTestData" );

		bottomEntity -> passUpward( &data );
		CHECK( observer.observed( "onPassUpward", bottomEntity -> getLabel(), data.toString() ) );

		sim.step();
		CHECK( observer.observed( "onReadBottom", topEntity -> getLabel(), data.toString() ) );
	}

	/**
	 * @test odemx::ProtocolEntity::passUpward(PortData*) to the node
	 *
	 * Expected function call effects:
	 * @li the call was observed
	 * @li ProtocolStack::receivedData() was called because this entity
	 * is the top layer entity
	 */
	TEST_FIXTURE( ProtocolEntityFixture, PassUpwardToNode )
	{
		ProtocolEntityTestData data( "PassUpwardTestData" );

		topEntity -> passUpward( &data );
		CHECK( observer.observed( "onPassUpward", topEntity -> getLabel(), data.toString() ) );
		CHECK( node -> receivedDataFromTopEntity );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond

