/**
 * @file TestProtocolLayer.cpp
 * @date Jul 19, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class ProtocolLayer
 */

#include "TestProtocol.h"

/// @cond DOXYGEN_SKIP
SUITE( Protocol )
{
/// @endcond

	/**
	 * @struct ProtocolLayerFixture
	 * @brief Helper struct providing set-up/tear-down of ProtocolLayer tests
	 *
	 * @copydetails EventFixture
	 */
	struct ProtocolLayerFixture
	{
		ProtocolSimulation sim;
		ProtocolLayerTestObserver observer;
		ProtocolLayerT< ProtocolEntityTest > layer;
		ProtocolStackTest node;

		ProtocolLayerFixture()
		:	sim( "ProtocolLayerTestSim" ),
			observer( false ),
			layer( &sim, "ProtocolLayerTest", &observer ),
			node( &sim, "ProtocolLayerTestNode" )
			{}
	};

	/**
	 * @test odemx::ProtocolLayer construction and destruction
	 *
	 * Expected construction effects:
	 * @li a pointer to the simulation is stored
	 * @li the layer is added to the simulation's layer list
	 * @li the label is set correctly
	 * @li the layer can be retrieved by its label
	 * @li the observer is added correctly
	 * @li construction and destruction can be observed
	 * @li upon destruction, the layer is removed from the layer list
	 */
	TEST_FIXTURE( ProtocolLayerFixture, ConstructionDestruction )
	{
		observer.history.clear();

		ProtocolSimulation testSim( "ProtocolLayerTestSim" );
		Label layerLabel = "ProtocolLayerTestConstructionDestructionLayer";
		ProtocolLayerT< ProtocolEntityTest >* layer2 =
			new ProtocolLayerT< ProtocolEntityTest >( &testSim, layerLabel, &observer );

		CHECK_EQUAL( &testSim, layer2 -> getSimulation() );
		CHECK_EQUAL( (size_t)1, testSim.getLayerList().size() );
		CHECK( find( testSim.getLayerList().begin(), testSim.getLayerList().end(), layer2 )
				!= testSim.getLayerList().end() );
		CHECK_EQUAL( layerLabel, layer2 -> getLabel() );
		CHECK_EQUAL( layer2, testSim.getLayerByLabel( layer2 -> getLabel() ) );
		CHECK_EQUAL( &observer, layer2 -> getObservers().front() );
		CHECK( observer.observed( "onCreate", layerLabel ) );

		delete layer2;
		CHECK( observer.observed( "onDestroy", layerLabel ) );
		CHECK_EQUAL( (size_t)0, testSim.getLayerList().size() );
		CHECK( find( testSim.getLayerList().begin(), testSim.getLayerList().end(), layer2 )
				== testSim.getLayerList().end() );
	}

	/**
	 * @test odemx::ProtocolLayer::createEntity()
	 *
	 * Expected function call effects:
	 * @li the entity has the correct type, as specified for ProtocolLayerT
	 * @li the entity was added to this layer's list of all its entities
	 */
	TEST_FIXTURE( ProtocolLayerFixture, CreateEntity )
	{
		size_t layerCount = layer.getEntityList().size();
		ProtocolEntity* entity = layer.createNewEntity( &node );
		CHECK( dynamic_cast< ProtocolEntityTest* >( entity ) != 0 );
		CHECK_EQUAL( layerCount + 1, layer.getEntityList().size() );
		CHECK( find( layer.getEntityList().begin(), layer.getEntityList().end(), entity )
				!= layer.getEntityList().end() );
	}

	/**
	 * @test odemx::ProtocolLayer::addEntity()
	 *
	 * Expected function call effects:
	 * @li the call is observed
	 * @li the entity was added to this layer's list of all its entities
	 */
	TEST_FIXTURE( ProtocolLayerFixture, AddEntity )
	{
		size_t layerCount = layer.getEntityList().size();

		ProtocolEntity* entity =
			new ProtocolEntityTest( &sim, &layer, &node );
		layer.addEntity( entity );

		CHECK( observer.observed( "onAddEntity", layer.getLabel() ) );
		CHECK_EQUAL( layerCount + 1, layer.getEntityList().size() );
	}

	/**
	 * @test odemx::ProtocolLayer::removeEntity()
	 *
	 * Expected function call effects:
	 * @li the call is observed
	 * @li the entity was added to this layer's list of all its entities
	 */
	TEST_FIXTURE( ProtocolLayerFixture, RemoveEntity )
	{
		size_t layerCount = layer.getEntityList().size();
		ProtocolEntity* entity =
			new ProtocolEntityTest( &sim, &layer, &node );
		layer.addEntity( entity );
		CHECK_EQUAL( layerCount + 1, layer.getEntityList().size() );

		layer.removeEntity( entity );
		CHECK( observer.observed( "onRemoveEntity", layer.getLabel() ) );
		CHECK_EQUAL( layerCount, layer.getEntityList().size() );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
