/**
 * @file TestProtocolMessage.cpp
 * @date Jul 19, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class ProtocolMessage
 */

#include "TestProtocol.h"

/// @cond DOXYGEN_SKIP
SUITE( Protocol )
{
/// @endcond

	/**
	 * @struct ProtocolMessageFixture
	 * @brief Helper struct providing set-up/tear-down of ProtocolMessage tests
	 *
	 * @copydetails EventFixture
	 */
	struct ProtocolMessageFixture
	{
		ProtocolSimulation sim;
		ProtocolStackTest node1;
		ProtocolStackTest node2;

		ProtocolMessageFixture()
		:	sim( "ProtocolMessageTestSim" ),
			node1( &sim, "ProtocolMessageTestNode1" ),
			node2( &sim, "ProtocolMessageTestNode2" )
			{}
	};

	/**
	 * @test odemx::ProtocolMessage transmission time computation
	 *
	 * Expected function call effects:
	 * @li transmission time is current time plus duration
	 */
	TEST_FIXTURE( ProtocolMessageFixture, TransmissionTime )
	{
		SimTime duration = 999;
		ProtocolMessageTest message( "MessageContent", &node1, &node2, duration );

		message.computeTransmissionTime( sim.getTime() );
		CHECK_EQUAL( sim.getTime() + duration, message.getTransmissionTime() );
	}

	/**
	 * @test odemx::ProtocolMessage virtual functions
	 *
	 * Ensure that the ProtocolMessage interface is still intact:
	 * @li \c isUnicast()
	 * @li \c getSource()
	 * @li \c getSendNode()
	 * @li \c getDestination()
	 * @li \c getReceiveNode()
	 * @li \c getTransmissionDuration()
	 * @li \c deepCopy()
	 * @li \c toString()
	 */
	TEST_FIXTURE( ProtocolMessageFixture, VirtualFunctions )
	{
		SimTime duration = 187;
		ProtocolMessage* message = new ProtocolMessageTest( "MessageContent", &node1, &node2, duration );

		CHECK( message -> isUnicast() );
		CHECK_EQUAL( &node1, message -> getSource() );
		CHECK_EQUAL( &node1, message -> getSendNode() );
		CHECK_EQUAL( &node2, message -> getDestination() );
		CHECK_EQUAL( &node2, message -> getReceiveNode() );
		CHECK_EQUAL( duration, message -> getTransmissionDuration() );
		CHECK( dynamic_cast< ProtocolMessageTest* >( message -> deepCopy() ) );
		CHECK_EQUAL( string("Message '") + dynamic_cast< ProtocolMessageTest* >( message ) -> content
				+ "' from "	+ node1.getLabel() + " to " + node2.getLabel(),
				message -> toString() );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
