/**
 * @file TestProtocolSimulation.cpp
 * @date Jul 19, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class ProtocolSimulation
 */

#include "TestProtocol.h"

/// @cond DOXYGEN_SKIP
SUITE( Protocol )
{
/// @endcond

	/**
	 * @struct ProtocolSimulationFixture
	 * @brief Helper struct providing set-up/tear-down of ProtocolSimulation tests
	 *
	 * @copydetails EventFixture
	 */
	struct ProtocolSimulationFixture
	{
		ProtocolSimulation sim;

		ProtocolSimulationFixture()
		:	sim( "ProtocolSimulationTestSim" )
			{}
	};

	/**
	 * @test odemx::ProtocolSimulation::getNetTopology()
	 *
	 * Expected function call effects:
	 * @li if no TransmissionMedium exists, it returns 0 and sends an
	 * error message to \c errorStream()
	 * @li once the TransmissionMedium exists, a pointer to the
	 * NetTopology is returned
	 */
	TEST_FIXTURE( ProtocolSimulationFixture, GetNetTopology )
	{
		clearErrorStream();
		NetTopology* topology = sim.getNetTopology();
		CHECK_EQUAL( static_cast< NetTopology* >( 0 ), topology );
		CHECK_EQUAL( "ODEMx ERROR: ProtocolSimulation::getNetTopology(): NetTopology not available without TransmissionMedium -- Object: ProtocolSimulationTestSim\n",\
				errorStream().str() );

		sim.createMedium< TransmissionMediumTest >( "ProtocolSimulationTestMedium" );
		topology = sim.getNetTopology();
		CHECK( dynamic_cast< NetTopology* >( topology ) );
	}

	/**
	 * @test odemx::ProtocolSimulation::getTransmissionMedium()
	 *
	 * Expected function call effects:
	 * @li if no TransmissionMedium exists, it returns 0 and sends an
	 * error message to \c errorStream()
	 * @li once the TransmissionMedium exists, a pointer to the
	 * correct type of TransmissionMedium is returned
	 */
	TEST_FIXTURE( ProtocolSimulationFixture, GetTransmissionMedium )
	{
		clearErrorStream();
		TransmissionMedium* medium = sim.getTransmissionMedium();
		CHECK_EQUAL( static_cast< TransmissionMedium* >( 0 ), medium );
		CHECK_EQUAL( "ODEMx ERROR: ProtocolSimulation::getTransmissionMedium(): TransmissionMedium not created, Simulation not initialized? -- Object: ProtocolSimulationTestSim\n",
				errorStream().str() );

		medium = sim.createMedium< TransmissionMediumTest >( "ProtocolSimulationTestMedium" );
		CHECK( dynamic_cast< TransmissionMediumTest* >( medium ) );
	}

	/**
	 * @test odemx::ProtocolSimulation::createMedium()
	 *
	 * Expected function call effects:
	 * @li the created TransmissionMedium is that which the Simulation returns
	 * @li the medium's Simulation is set correctly
	 * @li the medium's label is set correctly
	 * @li the medium's observer is set correctly
	 */
	TEST_FIXTURE( ProtocolSimulationFixture, CreateMedium )
	{
		TransmissionMediumObserver observer;
		Label mediumLabel = "ProtocolSimulationTestMedium";
		TransmissionMediumTest* medium =
			sim.createMedium< TransmissionMediumTest >( mediumLabel, &observer );
		CHECK_EQUAL( medium, sim.getTransmissionMedium() );
		CHECK_EQUAL( &sim, medium -> getSimulation() );
		CHECK_EQUAL( mediumLabel, medium -> getLabel() );
		CHECK_EQUAL( &observer, medium -> Observable< TransmissionMediumObserver >::getObservers().front() );

		// need to remove the observer, or else the test crashes due to
		// order of destructor calls: observer gone, but observable sends onDestroy...
		medium -> Observable< TransmissionMediumObserver >::removeObserver( &observer );
	}

	/**
	 * @test odemx::ProtocolSimulation::createLayer()
	 *
	 * Expected function call effects:
	 * @li the created layer is added to the layer list
	 * @li the layer's Simulation is set correctly
	 * @li the layer's label is set correctly
	 * @li the layer can be retrieved by its label
	 * @li the layer's observer is set correctly
	 */
	TEST_FIXTURE( ProtocolSimulationFixture, CreateLayer )
	{
		ProtocolLayerTestObserver observer( false );
		Label layerLabel = "ProtocolSimulationTestLayer";
		CHECK( sim.getLayerList().empty() );

		sim.createLayer< ProtocolEntityTest >( layerLabel, &observer );
		CHECK_EQUAL( (size_t)1, sim.getLayerList().size() );

		ProtocolLayer* layer = sim.getLayerList().front();
		CHECK_EQUAL( &sim, layer -> getSimulation() );
		CHECK_EQUAL( layerLabel, layer -> getLabel() );
		CHECK_EQUAL( &observer, layer -> getObservers().front() );
		CHECK_EQUAL( layer, sim.getLayerByLabel( layerLabel ) );
		// safely remove observer
		layer -> removeObserver( &observer );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond

