/**
 * @file TestProtocolStack.cpp
 * @date Jul 20, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class ProtocolStack
 */

#include "TestProtocol.h"

/// @cond DOXYGEN_SKIP
SUITE( Protocol )
{
/// @endcond

	/**
	 * @struct ProtocolStackFixture
	 * @brief Helper struct providing set-up/tear-down of ProtocolStack tests
	 *
	 * @copydetails EventFixture
	 */
	struct ProtocolStackFixture
	{
		ProtocolSimulation sim;
		ProtocolStackTestObserver observer;
		ProtocolStackTest* node;
		ProtocolEntityTest* topEntity;
		ProtocolEntityTest* middleEntity;
		ProtocolEntityTest* bottomEntity;

		ProtocolStackFixture()
		:	sim( "ProtocolStackTestSim" ),
			observer( false ),
			node( 0 ),
			topEntity( 0 ),
			middleEntity( 0 ),
			bottomEntity( 0 )
			{
				sim.createMedium< TransmissionMediumTest >( "ProtocolStackTestMedium" );
				sim.createLayer< ProtocolEntityTest >( "TestTopLayer" );
				sim.createLayer< ProtocolEntityTest >( "TestMiddleLayer" );
				sim.createLayer< ProtocolEntityTest >( "TestBottomLayer" );

				// this creates three entities of type ProtocolEntityTest,
				// one for each layer, and schedules them with hold()
				node = new ProtocolStackTest( &sim, "ProtocolStackTestNode", &observer );
			}

		~ProtocolStackFixture()
			{
				if ( node ) delete node;
			}
	};

	/**
	 * @test odemx::ProtocolStack construction and destruction
	 *
	 * Expected construction effects:
	 * @li an error message is sent to \c errorStream() if no layers are defined
	 * @li the creation of a node can be observed
	 * @li the entities for all layers are initialized and scheduled properly
	 * @li highest and lowest layer entities are set properly
	 * @li IDs are numbered consecutively, and the node has an ID
	 *
	 * Expected destruction effects:
	 * @li the destruction of a node is observed
	 */
	TEST_FIXTURE( ProtocolStackFixture, ConstructionDestruction )
	{
		// create an error-producing node
		clearErrorStream();
		ProtocolSimulation testSim( "ProtocolStackTestConstructionDestructionSim" );
		ProtocolStack* failNode = new ProtocolStackTest( &testSim, "ProtocolStackTestConstructionDestructionNode" );
		CHECK_EQUAL( "ODEMx ERROR: ProtocolStack(): No protocol layers defined, cannot create entities -- Object: ProtocolStackTestConstructionDestructionNode\n",\
				errorStream().str() );
		NodeID previousNodeId = failNode -> getNodeId();
		delete failNode;

		// create a correctly set up node
		observer.history.clear();
		Label nodeLabel = "ProtocolStackTestConstructionDestructionNode";
		ProtocolStack* theNode =
			new ProtocolStackTest( &sim, nodeLabel, &observer );
		// get the node's entities
		topEntity = dynamic_cast< ProtocolEntityTest* >( theNode -> getEntityByLayerName( "TestTopLayer" ) );
		middleEntity = dynamic_cast< ProtocolEntityTest* >( theNode -> getEntityByLayerName( "TestMiddleLayer" ) );
		bottomEntity = dynamic_cast< ProtocolEntityTest* >( theNode -> getEntityByLayerName( "TestBottomLayer" ) );

		CHECK( observer.observed( "onCreate", theNode -> getLabel() ) );
		CHECK_EQUAL( topEntity, theNode -> getTopLayerEntity() );
		CHECK_EQUAL( bottomEntity, theNode -> getBottomLayerEntity() );
		CHECK( topEntity -> isScheduled() );
		CHECK( middleEntity -> isScheduled() );
		CHECK( bottomEntity -> isScheduled() );
		CHECK_EQUAL( previousNodeId + 1, theNode -> getNodeId() );

		// check destruction
		delete theNode;
		CHECK( observer.observed( "onDestroy", nodeLabel ) );
	}

	/**
	 * @test odemx::ProtocolStack::inputData(PortData*)
	 *
	 * Expected function call effects:
	 * @li the call is observed
	 * @li the top-layer entity is awakened to handle the data
	 */
	TEST_FIXTURE( ProtocolStackFixture, InputData )
	{
		ProtocolEntityTestData data( "ProtocolStackTestData" );
		node -> inputData( &data );
		CHECK( observer.observed( "onInputData", node -> getLabel(), data.toString() ) );
		sim.step();

		topEntity = dynamic_cast< ProtocolEntityTest* >( node -> getTopLayerEntity() );
		CHECK( topEntity -> handledTopInput );
	}

	/**
	 * @test odemx::ProtocolStack::receiveTransmission(ProtocolMessage*)
	 *
	 * Expected function call effects:
	 * @li the call is observed
	 * @li the bottom-layer entity is awakened to handle the data
	 */
	TEST_FIXTURE( ProtocolStackFixture, ReceiveTransmission )
	{
		ProtocolStackTest node2( &sim, "ProtocolStackTestReceiveNode" );
		ProtocolMessageTest data( "ProtocolStackTestData", &node2, node, 0 );
		node -> receiveTransmission( &data );
		CHECK( observer.observed( "onReceiveTransmission", node -> getLabel(), data.toString() ) );
		sim.run();

		bottomEntity = dynamic_cast< ProtocolEntityTest* >( node -> getBottomLayerEntity() );
		CHECK( bottomEntity -> handledBottomInput );
	}

	/**
	 * @test odemx::ProtocolStack::_receivedData(PortData*)
	 *
	 * The function is tested indirectly by having the top entity passing data.
	 *
	 * Expected function call effects:
	 * @li the call is observed
	 * @li the virtual function \c receivedData() is called
	 */
	TEST_FIXTURE( ProtocolStackFixture, _ReceivedData )
	{
		ProtocolEntityTestData data( "ProtocolStackTestData" );
		node -> getTopLayerEntity() -> passUpward( &data );
		CHECK( observer.observed( "onReceivedData", node -> getLabel(), data.toString() ) );
		CHECK( node -> receivedDataFromTopEntity );
	}

	/**
	 * @test odemx::ProtocolStack::getEntityByLayer(ProtocolLayer*)
	 *
	 * Expected function call effects:
	 * @li returns the correct entity if the layer exists
	 * @li returns 0 if Layer not found
	 */
	TEST_FIXTURE( ProtocolStackFixture, GetEntityByLayer )
	{
		ProtocolLayerList& layers = sim.getLayerList();
		CHECK_EQUAL( node -> getTopLayerEntity(), node -> getEntityByLayer( layers.front() ) );
		CHECK_EQUAL( node -> getBottomLayerEntity(), node -> getEntityByLayer( layers.back() ) );

		// must be checked here because it would otherwise become the last layer
		// the check above would not work like this
		ProtocolLayerT< ProtocolEntityTest > unknownLayer( &sim, "ProtocolStackTestLayer" );
		CHECK_EQUAL( static_cast< ProtocolEntity* >( 0 ), node -> getEntityByLayer( &unknownLayer ) );

	}

	/**
	 * @test odemx::ProtocolStack::getEntityByLayerName(Label)
	 *
	 * Expected function call effects:
	 * @li returns the correct entity if the layer exists
	 * @li returns 0 if Layer not found
	 */
	TEST_FIXTURE( ProtocolStackFixture, GetEntityByLayerName )
	{
		ProtocolLayerList& layers = sim.getLayerList();
		CHECK_EQUAL( node -> getTopLayerEntity(), node -> getEntityByLayerName( layers.front() -> getLabel() ) );
		CHECK_EQUAL( node -> getBottomLayerEntity(), node -> getEntityByLayerName( layers.back() -> getLabel() ) );

		// must be checked here because it would otherwise become the last layer
		// the check above would not work like this
		ProtocolLayerT< ProtocolEntityTest > unknownLayer( &sim, "ProtocolStackTestLayer" );
		CHECK_EQUAL( static_cast< ProtocolEntity* >( 0 ), node -> getEntityByLayerName( unknownLayer.getLabel() ) );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
