/**
 * @file TestTransmissionEvents.cpp
 * @date Jul 21, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx classes BroadcastTransmission and UnicastTransmission
 */

#include "TestProtocol.h"

/// @cond DOXYGEN_SKIP
SUITE( Protocol )
{
/// @endcond

	/**
	 * @struct TransmissionEventFixture
	 * @brief Helper struct providing set-up/tear-down of transmission event tests
	 *
	 * @copydetails EventFixture
	 */
	struct TransmissionEventFixture
	{
		ProtocolSimulation sim;
		BroadcastTransmissionTestObserver broadcastObserver;
		UnicastTransmissionTestObserver unicastObserver;
		ProtocolStackTestObserver nodeObserver;
		ProtocolStackTest* node1;
		ProtocolStackTest* node2;
		ProtocolStackTest* node3;
		ProtocolMessageTest* broadcastMessage;
		ProtocolMessageTest* unicastMessage;

		TransmissionEventFixture()
		:	sim( "TransmissionEventTestSim" ),
			broadcastObserver( false ),
			unicastObserver( false ),
			nodeObserver( false ),
			node1( 0 ),
			node2( 0 ),
			node3( 0 ),
			broadcastMessage( 0 ),
			unicastMessage( 0 )
			{
				sim.createMedium< TransmissionMedium >( "TransmissionEventTestMedium" );
				sim.createLayer< ProtocolEntityTest >( "TransmissionEventTestLayer" );

				node1 = new ProtocolStackTest( &sim, "TransmissionEventTestNode1", &nodeObserver );
				node2 = new ProtocolStackTest( &sim, "TransmissionEventTestNode2", &nodeObserver );
				node3 = new ProtocolStackTest( &sim, "TransmissionEventTestNode3", &nodeObserver );

				sim.getNetTopology() -> registerNode( node1, 0, 0 );
				sim.getNetTopology() -> registerNode( node2, 1, 1 );
				sim.getNetTopology() -> registerNode( node3, 2, 2 );

				broadcastMessage = new ProtocolMessageTest( "BroadcastEventTestMessage",
						node1, 0, 20 );
				unicastMessage = new ProtocolMessageTest( "UnicastEventTestMessage",
						node1, node2, 10 );
			}

		~TransmissionEventFixture()
			{
				if( node1 ) delete node1;
				if( node2 ) delete node2;
				if( node3 ) delete node3;
				if( unicastMessage ) delete unicastMessage;
				if( broadcastMessage ) delete broadcastMessage;
			}
	};

	/**
	 * @test odemx::UnicastTransmission::eventAction()
	 *
	 * Expected function call effects:
	 * @li with bad connectivity, transmission failure can be observed
	 * @li with good connectivity, transmission success can be observed
	 * and the receiver gets the message
	 */
	TEST_FIXTURE( TransmissionEventFixture, UnicastEventAction )
	{
		CHECK( unicastMessage -> isUnicast() );

		// force bad connectivity between nodes
		sim.getNetTopology() -> setLink( node1, node2, 0 );
		UnicastTransmission* unicastFail = UnicastTransmission::create(
				&sim, "UnicastEventFailureTest", unicastMessage, &unicastObserver );
		unicastFail -> schedule();

		sim.step();
		CHECK( unicastObserver.observed( "onFailure", unicastFail -> getLabel(), \
				node1 -> getLabel()+"->"+node2 -> getLabel(), unicastMessage -> toString() ) );
		CHECK( ! nodeObserver.observed( "onReceiveTransmission", node2 -> getLabel(), unicastMessage -> toString() ) );

		// force good connectivity between nodes
		sim.getNetTopology() -> setLink( node1, node2, 1 );
		UnicastTransmission* unicast = UnicastTransmission::create(
				&sim, "UnicastEventTest", unicastMessage, &unicastObserver );
		unicast -> schedule();

		sim.step();
		CHECK( unicastObserver.observed( "onSuccess", unicast -> getLabel(), \
				node1 -> getLabel()+"->"+node2 -> getLabel(), unicastMessage -> toString() ) );
		CHECK( nodeObserver.observed( "onReceiveTransmission", node2 -> getLabel(), unicastMessage -> toString() ) );
	}

	/**
	 * @test odemx::BroadcastTransmission::eventAction()
	 *
	 * Expected function call effects:
	 * @li with bad connectivity, transmission failure can be observed
	 * @li with good connectivity, transmission success can be observed
	 * and the receiver gets the message
	 */
	TEST_FIXTURE( TransmissionEventFixture, BroadcastEventAction )
	{
		CHECK( ! broadcastMessage -> isUnicast() );

		// force different connectivity between nodes
		sim.getNetTopology() -> setLink( node1, node2, 0 );
		sim.getNetTopology() -> setLink( node1, node3, 1 );
		BroadcastTransmission* broadcastFail = BroadcastTransmission::create(
				&sim, "BroadcastEventFailureTest", broadcastMessage, &broadcastObserver );
		broadcastFail -> schedule();

		sim.step();
		// check node2
		CHECK( broadcastObserver.observed( "onFailure", broadcastFail -> getLabel(), \
				node1 -> getLabel()+"->"+node2 -> getLabel(), broadcastMessage -> toString() ) );
		CHECK( ! nodeObserver.observed( "onReceiveTransmission", node2 -> getLabel(), broadcastMessage -> toString() ) );

		// check node3
		CHECK( broadcastObserver.observed( "onSuccess", broadcastFail -> getLabel(), \
				node1 -> getLabel()+"->"+node3 -> getLabel(), broadcastMessage -> toString() ) );
		CHECK( nodeObserver.observed( "onReceiveTransmission", node3 -> getLabel(), broadcastMessage -> toString() ) );

		// force good connectivity between nodes
		sim.getNetTopology() -> setLink( node1, node2, 1 );
		sim.getNetTopology() -> setLink( node1, node3, 1 );
		BroadcastTransmission* broadcast = BroadcastTransmission::create(
				&sim, "BroadcastEventTest", broadcastMessage, &broadcastObserver );
		broadcast -> schedule();

		sim.step();
		CHECK( broadcastObserver.observed( "onSuccess", broadcast -> getLabel(), \
				node1 -> getLabel()+"->"+node2 -> getLabel(), broadcastMessage -> toString() ) );
		CHECK( nodeObserver.observed( "onReceiveTransmission", node2 -> getLabel(), broadcastMessage -> toString() ) );
		CHECK( broadcastObserver.observed( "onSuccess", broadcast -> getLabel(), \
				node1 -> getLabel()+"->"+node3 -> getLabel(), broadcastMessage -> toString() ) );
		CHECK( nodeObserver.observed( "onReceiveTransmission", node3 -> getLabel(), broadcastMessage -> toString() ) );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
