/**
 * @file TestTransmissionMedium.cpp
 * @date Jul 21, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class TransmissionMedium
 */

#include "TestProtocol.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Protocol )
{
/// @endcond

	/**
	 * @struct TransmissionMediumFixture
	 * @brief Helper struct providing set-up/tear-down of TransmissionMedium tests
	 *
	 * @copydetails EventFixture
	 */
	struct TransmissionMediumFixture
	{
		ProtocolSimulation sim;
		Util::ReportTest reporter;
		TransmissionMediumTestObserver observer;
		TransmissionMedium medium;
		ProtocolStack* node1;
		ProtocolStack* node2;

		TransmissionMediumFixture()
		:	sim( "TransmissionMediumTestSim" ),
			reporter(),
			observer( false ),
			medium( &sim, "TransmissionMediumTest", &observer ),
			node1( 0 ),
			node2( 0 )
			{
				node1 = new ProtocolStackTest( &sim, "TransmissionMediumTestNode1" );
				node2 =	new ProtocolStackTest( &sim, "TransmissionMediumTestNode2" );
				medium.registerNode( node1, 0, 0 );
				medium.registerNode( node2, 0, 0 );
			}

		~TransmissionMediumFixture()
		{
			if( node1 ) delete node1;
			if( node2 ) delete node2;
		}
	};

	/**
	 * @test odemx::TransmissionMedium construction and destruction
	 *
	 * Tests member initialization:
	 * @li simulation
	 * @li label
	 * @li observer
	 *
	 * Expected function call effects:
	 * @li construction can be observed
	 * @li destruction can be observed
	 */
	TEST_FIXTURE( TransmissionMediumFixture, ConstructionDestruction )
	{
		Label label = "TransmissionMediumConstructionDestructionTest";
		TransmissionMedium* tm
			= new TransmissionMedium( &sim, label, &observer );

		CHECK_EQUAL( &sim, tm -> getSimulation() );
		CHECK_EQUAL( label, tm -> getLabel() );
		CHECK_EQUAL( &observer, tm -> Observable< TransmissionMediumObserver >::getObservers().front() );
		CHECK( observer.observed( "onCreate", tm -> getLabel() ) );

		delete tm;
		CHECK( observer.observed( "onDestroy", label ) );
	}

	/**
	 * @test odemx::TransmissionMedium::transmit(ProtocolMessage*) unicast
	 *
	 * Expected function call effects:
	 * @li attempting to send a message between nodes without link fails and
	 * sends an error message to \c errorStream()
	 * @li the transmit call can be observed
	 * @li attempting to send a unicast message between linked nodes
	 * leads to the scheduling of a UnicastTransmission event at current time
	 * + message transmission duration
	 */
	TEST_FIXTURE( TransmissionMediumFixture, TransmitUnicast )
	{
		SimTime duration = 5;
		ProtocolMessageTest unicastMessage( "TransmissionMediumTransmitTest",
				node1, node2, duration );

		clearErrorStream();
		bool transmitSuccess = medium.transmit( &unicastMessage );
		CHECK_EQUAL( string( "ODEMx ERROR: TransmissionMedium::transmit(): " )
				+ "no connection set for node pair in unicast transmission: "
				+ "TransmissionMediumTestNode1 -> TransmissionMediumTestNode2 "
				+ "-- Object: TransmissionMediumTest\n",
				errorStream().str() );
		CHECK( ! transmitSuccess );

		medium.setLink( node1, node2, 1 );
		transmitSuccess = medium.transmit( &unicastMessage );
		CHECK( transmitSuccess );
		CHECK( observer.observed( "onTransmit", medium.getLabel(), unicastMessage.toString() ) );
		CHECK( observer.observed( "onScheduleUnicast", medium.getLabel(),
				unicastMessage.toString(), toString( medium.getCurrentTime() + duration ) ) );
	}

	/**
	 * @test odemx::TransmissionMedium::transmit(ProtocolMessage*) broadcast
	 *
	 * Expected function call effects:
	 * @li the transmit call can be observed
	 * @li attempting to send a broadcast message leads to the scheduling of a
	 * BroadcastTransmission event at current time + message transmission
	 * duration
	 */
	TEST_FIXTURE( TransmissionMediumFixture, TransmitBroadcast )
	{
		medium.setLink( node1, node2, 1 );
		SimTime duration = 5;
		ProtocolMessageTest broadcastMessage( "TransmissionMediumTransmitTest",
				node1, 0, duration );

		bool transmitSuccess = medium.transmit( &broadcastMessage );
		CHECK( transmitSuccess );
		CHECK( observer.observed( "onTransmit", medium.getLabel(), broadcastMessage.toString() ) );
		CHECK( observer.observed( "onScheduleBroadcast", medium.getLabel(),
				broadcastMessage.toString(), toString( medium.getCurrentTime() + duration ) ) );
	}

	/**
	 * @test odemx::TransmissionMedium::scheduleUnicastTransmission(ProtocolMessage*)
	 *
	 * Expected function call effects:
	 * @li the call can be observed
	 * @li an object of type UnicastTransmission is schedduled for execution
	 * at current time + transmission duration
	 */
	TEST_FIXTURE( TransmissionMediumFixture, ScheduleUnicast )
	{
		SimTime duration = 5;
		ProtocolMessageTest unicastMessage( "TransmissionMediumScheduleUnicastTest",
				node1, 0, duration );
		unicastMessage.computeTransmissionTime( medium.getCurrentTime() );

		medium.scheduleUnicastTransmission( &unicastMessage );
		CHECK( observer.observed( "onScheduleUnicast", medium.getLabel(),
				unicastMessage.toString(), toString( medium.getCurrentTime() + duration ) ) );

		CHECK( dynamic_cast< UnicastTransmission* >( sim.getNextSched() ) );
		CHECK_EQUAL( medium.getCurrentTime() + duration, sim.getNextSched() -> getExecutionTime() );
	}

	/**
	 * @test odemx::TransmissionMedium::scheduleBroadcastTransmission(ProtocolMessage*)
	 *
	 * Expected function call effects:
	 * @li the call can be observed
	 * @li an object of type BroadcastTransmission is scheduled for
	 * execution at current time + transmission duration
	 */
	TEST_FIXTURE( TransmissionMediumFixture, ScheduleBroadcast )
	{
		SimTime duration = 5;
		ProtocolMessageTest broadcastMessage( "TransmissionMediumScheduleBroadcastTest",
				node1, 0, duration );
		broadcastMessage.computeTransmissionTime( medium.getCurrentTime() );

		medium.scheduleBroadcastTransmission( &broadcastMessage );
		CHECK( observer.observed( "onScheduleBroadcast", medium.getLabel(),
				broadcastMessage.toString(), toString( medium.getCurrentTime() + duration ) ) );

		CHECK( dynamic_cast< BroadcastTransmission* >( sim.getNextSched() ) );
		CHECK_EQUAL( medium.getCurrentTime() + duration, sim.getNextSched() -> getExecutionTime() );
	}

	/**
	 * @test odemx::TransmissionMedium::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( TransmissionMediumFixture, Report )
	{
		ProtocolSimulation simulation( "TransmissionMediumTestReportSim" );
		TransmissionMedium* reportMedium =
			simulation.createMedium< TransmissionMedium >( "TransmissionMediumTestReport" );

		CHECK( reportMedium == simulation.getTransmissionMedium() );
		CHECK( &simulation == reportMedium -> getSimulation() );

		simulation.createLayer< ProtocolEntityTest >( "TransmissionMediumTestReportLayer" );

		ProtocolStackTest* testNode1 = new ProtocolStackTest( &simulation, "TransmissionMediumTestReportNode1" );
		ProtocolStackTest* testNode2 =	new ProtocolStackTest( &simulation, "TransmissionMediumTestReportNode2" );
		simulation.runUntil( 1 ); // start entities and remove them from schedule
		reportMedium -> reset();

		reportMedium -> registerNode( testNode1, 0, 0 );
		reportMedium -> registerNode( testNode2, 0, 0 );
		reportMedium -> setLink( testNode1, testNode2, 1 );

		CHECK( simulation.isEmpty() );

		ProtocolMessageTest unicastMessage( "TransmissionMediumTestReportUnicast", testNode1, testNode2, 5 );
		ProtocolMessageTest broadcastMessage( "TransmissionMediumTestReportUnicast", testNode1, 0, 1 );
		reportMedium -> transmit( &unicastMessage );
		reportMedium -> transmit( &broadcastMessage );
		simulation.step(); // broadcast
		simulation.step(); // execute entity scheduled by broadcast
		simulation.step(); // unicast
		simulation.step(); // execute entity scheduled by unicast

		reporter.addProducer( reportMedium );
		reporter.generateReport();
		CHECK( reporter.processedTables );
		CHECK_EQUAL( (size_t) 1, reporter.allTableContent.size() );

		// only check second table, the first one is the queue's
		CHECK_EQUAL( reportMedium -> getLabel(), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( reportMedium -> getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( toString( reportMedium -> getAllNodes().size() ), reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( 2 ), reporter.allTableContent[0][3] ); // messageCount
		CHECK_EQUAL( toString( 1 ), reporter.allTableContent[0][4] ); // broadcastCount
		CHECK_EQUAL( toString( 1 ), reporter.allTableContent[0][5] ); // unicastCount
		CHECK_EQUAL( toString( 1 ), reporter.allTableContent[0][6] ); // successfulBroadcastCount
		CHECK_EQUAL( toString( 1 ), reporter.allTableContent[0][7] ); // successfulUnicastCount
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][8] ); // failedBroadcastCount
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][9] ); // failedUnicastCount
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][10] ); // collisionCount

		delete testNode2;
		delete testNode1;
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond

