/**
 * @file TestContinuousConst.cpp
 * @date Aug 9, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class ContinuousConst
 */

#include "TestRandom.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Random )
{
/// @endcond

	/**
	 * @struct ContinuousConstFixture
	 * @brief Helper struct providing set-up/tear-down of ContinuousConst tests
	 *
	 * @copydetails EventFixture
	 */
	struct ContinuousConstFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		double constValue;
		ContinuousConst continuousConst;

		ContinuousConstFixture()
		:	sim( "ContinuousConstTestSim" ),
			reporter(),
			constValue( 1.23456789 ),
			continuousConst( &sim, "ContinuousConstTest", constValue )
			{}
	};

	/**
	 * @test odemx::ContinuousConst construction
	 *
	 * Expected function call effects:
	 * @li label is set
	 * @li dist context (simulation) is set
	 * @li const value is set
	 */
	TEST_FIXTURE( ContinuousConstFixture, ConstructionDestruction )
	{
		clearErrorStream();
		Label l = "ContinuousConstTestDefSimConstruction";
		ContinuousConst continuousConst1( l, constValue );
		CHECK_EQUAL( l, continuousConst1.getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), continuousConst1.getStatisticManager() );
		CHECK_EQUAL( constValue, continuousConst1.getA() );

		clearErrorStream();
		l = "ContinuousConstTestUserSimConstruction";
		ContinuousConst continuousConst2( &sim, l, constValue );
		CHECK_EQUAL( l, continuousConst2.getLabel() );
		CHECK_EQUAL( &sim, continuousConst2.getStatisticManager() );
		CHECK_EQUAL( constValue, continuousConst2.getA() );
	}

	/**
	 * @test odemx::ContinuousConst::sample()
	 *
	 * Expected function call effects:
	 * @li samples equal the const value
	 */
	TEST_FIXTURE( ContinuousConstFixture, Sample )
	{
		CHECK_EQUAL( continuousConst.getA(), continuousConst.sample() );
	}

	/**
	 * @test odemx::ContinuousConst::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( ContinuousConstFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "ContinuousConstTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		continuousConst.reset();

		for( int i = 0; i < 23; ++i )
		{
			continuousConst.sample();
		}

		reporter.addProducer( &continuousConst );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		CHECK_EQUAL( toString( continuousConst.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( continuousConst.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( "ContinuousConst", reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( long( continuousConst.getUses() ) ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( long( continuousConst.getSeed() ) ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( double( continuousConst.getA() ) ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][7] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
