/**
 * @file TestDiscreteConst.cpp
 * @date Aug 9, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class DiscreteConst
 */

#include "TestRandom.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Random )
{
/// @endcond

	/**
	 * @struct DiscreteConstFixture
	 * @brief Helper struct providing set-up/tear-down of DiscreteConst tests
	 *
	 * @copydetails EventFixture
	 */
	struct DiscreteConstFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		int constValue;
		DiscreteConst discreteConst;

		DiscreteConstFixture()
		:	sim( "DiscreteConstTestSim" ),
			reporter(),
			constValue( 7 ),
			discreteConst( &sim, "DiscreteConstTest", constValue )
			{}
	};

	/**
	 * @test odemx::DiscreteConst construction
	 *
	 * Expected function call effects:
	 * @li label is set
	 * @li dist context (simulation) is set
	 * @li const value is set
	 */
	TEST_FIXTURE( DiscreteConstFixture, ConstructionDestruction )
	{
		clearErrorStream();
		Label l = "DiscreteConstTestDefSimConstruction";
		DiscreteConst discreteConst1( l, constValue );
		CHECK_EQUAL( l, discreteConst1.getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), discreteConst1.getStatisticManager() );
		CHECK_EQUAL( constValue, discreteConst1.getA() );

		clearErrorStream();
		l = "DiscreteConstTestUserSimConstruction";
		DiscreteConst discreteConst2( &sim, l, constValue );
		CHECK_EQUAL( l, discreteConst2.getLabel() );
		CHECK_EQUAL( &sim, discreteConst2.getStatisticManager() );
		CHECK_EQUAL( constValue, discreteConst2.getA() );
	}

	/**
	 * @test odemx::DiscreteConst::sample()
	 *
	 * Expected function call effects:
	 * @li samples equal the const value
	 */
	TEST_FIXTURE( DiscreteConstFixture, Sample )
	{
		CHECK_EQUAL( discreteConst.getA(), discreteConst.sample() );
	}

	/**
	 * @test odemx::DiscreteConst::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( DiscreteConstFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "DiscreteConstTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		discreteConst.reset();

		for( int i = 0; i < 23; ++i )
		{
			discreteConst.sample();
		}

		reporter.addProducer( &discreteConst );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		CHECK_EQUAL( toString( discreteConst.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( discreteConst.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( "DiscreteConst", reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( long( discreteConst.getUses() ) ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( long( discreteConst.getSeed() ) ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( double( discreteConst.getA() ) ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][7] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
