/**
 * @file TestDraw.cpp
 * @date Aug 9, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Draw
 */

#include "TestRandom.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Random )
{
/// @endcond

	/**
	 * @struct DrawFixture
	 * @brief Helper struct providing set-up/tear-down of Draw tests
	 *
	 * @copydetails EventFixture
	 */
	struct DrawFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		double probability;
		int cells;
		Draw draw;
		Histo histogram;

		DrawFixture()
		:	sim( "DrawTestSim" ),
			reporter(),
			probability( double(2) / 3 ),
			cells( 1 ),
			draw( &sim, "DrawTest", probability ),
			histogram( &sim, "DrawTestHistogram", 0, 1, cells )
			{}
	};

	/**
	 * @test odemx::Draw construction
	 *
	 * Expected function call effects:
	 * @li label is set
	 * @li dist context (simulation) is set
	 * @li probability for drawing a \c 1 is set
	 */
	TEST_FIXTURE( DrawFixture, ConstructionDestruction )
	{
		clearErrorStream();
		Label l = "DrawTestDefSimConstruction";
		Draw draw1( l, probability );
		CHECK_EQUAL( l, draw1.getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), draw1.getStatisticManager() );
		CHECK_EQUAL( probability, draw1.getA() );

		clearErrorStream();
		l = "DrawTestUserSimConstruction";
		Draw draw2( &sim, l, probability );
		CHECK_EQUAL( l, draw2.getLabel() );
		CHECK_EQUAL( &sim, draw2.getStatisticManager() );
		CHECK_EQUAL( probability, draw2.getA() );
	}

	/**
	 * @test odemx::Draw::sample()
	 *
	 * Expected function call effects:
	 * @li samples are always within the interval [0, 1]
	 * @li the probability is close to the mean value
	 */
	TEST_FIXTURE( DrawFixture, Sample )
	{
		for( int i = 0; i < 1000000; ++i )
		{
			int value = draw.sample();
			CHECK( 0 <= value );
			CHECK( 1 >= value );
			histogram.update( value );
		}

		CHECK_CLOSE( draw.getA(), histogram.getMean(), 0.0005  );

		// print histogram
//		reporter.addProducer( &histogram );
//		reporter.generateReport();
//		CHECK( reporter.processedTables );
//
//		vector< string >& table = reporter.allTableContent[1];
//		for( size_t i = 0; i < table.size(); ++i )
//		{
//			cout << table[i] << " ";
//			if( ( i % 5 ) == 4 ) cout << endl;
//		}
//		cout << endl;
	}

	/**
	 * @test odemx::Draw::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( DrawFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "DrawTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		draw.reset();

		for( int i = 0; i < 42; ++i )
		{
			draw.sample();
		}

		reporter.addProducer( &draw );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		CHECK_EQUAL( toString( draw.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( draw.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( "Draw", reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( long( draw.getUses() ) ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( long( draw.getSeed() ) ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( double( draw.getA() ) ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][7] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
