/**
 * @file TestErlang.cpp
 * @date Aug 9, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Erlang
 */

#include "TestRandom.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"
#include <cstdlib> // atof()
#include <cmath> // floor()

#include <fstream>

/// @cond DOXYGEN_SKIP
SUITE( Random )
{
/// @endcond

	/**
	 * @struct ErlangFixture
	 * @brief Helper struct providing set-up/tear-down of Erlang tests
	 *
	 * @copydetails EventFixture
	 */
	struct ErlangFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		double rate;
		int shape;
		int cells;
		Erlang erlang;
		Histo histogram;

		ErlangFixture()
		:	sim( "ErlangTestSim" ),
			reporter(),
			rate( 2.0 ),
			shape( 3 ),
			cells( 100 ),
			erlang( &sim, "ErlangTest", rate, shape ),
			histogram( &sim, "ErlangTestHistogram", 0, 20, cells )
			{}
	};

	/**
	 * @test odemx::Erlang construction
	 *
	 * Expected function call effects:
	 * @li lower bound greater upper bound produces a warning and switches
	 * the two values
	 * @li label is set
	 * @li dist context (simulation) is set
	 * @li lower and upper bounds are set
	 */
	TEST_FIXTURE( ErlangFixture, ConstructionDestruction )
	{
		clearErrorStream();
		Label l = "ErlangTestDefSimConstruction";
		Erlang erlang1( l, -rate, -shape );
		CHECK_EQUAL( string( "ODEMx WARNING: Erlang: ea<=0; -ea or 0.01 used -- Object: ErlangTestDefSimConstruction\n" )
					+"ODEMx WARNING: Erlang: eb<=0; -eb or 1 used -- Object: ErlangTestDefSimConstruction\n",
					errorStream().str() );

		CHECK_EQUAL( l, erlang1.getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), erlang1.getStatisticManager() );
		CHECK_EQUAL( rate, erlang1.getRate() );
		CHECK_EQUAL( shape, erlang1.getShape() );

		clearErrorStream();
		l = "ErlangTestUserSimConstruction";
		Erlang erlang2( &sim, l, -rate, -shape );
		CHECK_EQUAL( string( "ODEMx WARNING: Erlang: ea<=0; -ea or 0.01 used -- Object: ErlangTestUserSimConstruction\n" )
					+"ODEMx WARNING: Erlang: eb<=0; -eb or 1 used -- Object: ErlangTestUserSimConstruction\n",
					errorStream().str() );

		CHECK_EQUAL( l, erlang2.getLabel() );
		CHECK_EQUAL( &sim, erlang2.getStatisticManager() );
		CHECK_EQUAL( rate, erlang2.getRate() );
		CHECK_EQUAL( shape, erlang2.getShape() );
	}

	/**
	 * @test odemx::Erlang::sample()
	 *
	 * Expected function call effects:
	 * @li samples are always within the interval [lower, upper]
	 * @li all values have roughly the same hit percentage
	 */
	TEST_FIXTURE( ErlangFixture, Sample )
	{
		for( int i = 0; i < 1000000; ++i )
		{
			double value = erlang.sample();
			histogram.update( value );
		}

		CHECK_CLOSE( erlang.getShape() / erlang.getRate(), histogram.getMean(), 0.002  );
		CHECK_CLOSE( erlang.getShape() / std::pow( erlang.getRate(), 2 ), std::pow( histogram.getStandardDeviation(), 2 ), 0.0002 );

		// print histogram
//		reporter.addProducer( &histogram );
//		reporter.generateReport();
//		CHECK( reporter.processedTables );
//
//		vector< string >& table = reporter.allTableContent[1];
//		for( size_t i = 0; i < table.size(); ++i )
//		{
//			cout << table[i] << " ";
//			if( ( i % 5 ) == 4 ) cout << endl;
//		}
//		cout << endl;
	}

	/**
	 * @test odemx::Erlang::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( ErlangFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "ErlangTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 42 );
		sim.step(); // event execution advances sim time
		erlang.reset();

		for( int i = 0; i < 666; ++i )
		{
			erlang.sample();
		}

		reporter.addProducer( &erlang );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		CHECK_EQUAL( toString( erlang.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( erlang.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( "Erlang", reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( long( erlang.getUses() ) ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( long( erlang.getSeed() ) ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( double( erlang.getRate() ) ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( double( erlang.getShape() ) ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][7] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
