/**
 * @file TestNegexp.cpp
 * @date Aug 9, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Negexp
 */

#include "TestRandom.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"
#include <cmath>

/// @cond DOXYGEN_SKIP
SUITE( Random )
{
/// @endcond

	/**
	 * @struct NegexpFixture
	 * @brief Helper struct providing set-up/tear-down of Negexp tests
	 *
	 * @copydetails EventFixture
	 */
	struct NegexpFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		double inverseMean;
		int cells;
		Negexp negExp;
		Histo histogram;

		NegexpFixture()
		:	sim( "NegexpTestSim" ),
			reporter(),
			inverseMean( 3 ),
			cells( (int)inverseMean * 10 ),
			negExp( &sim, "NegexpTest", inverseMean ),
			histogram( &sim, "NegexpTestHistogram", 0, 5, cells )
			{}
	};

	/**
	 * @test odemx::Negexp construction
	 *
	 * Expected function call effects:
	 * @li negative inverseMean value produces a warning and inverts it
	 * @li label is set
	 * @li dist context (simulation) is set
	 * @li inverseMean value is set
	 */
	TEST_FIXTURE( NegexpFixture, ConstructionDestruction )
	{
		clearErrorStream();
		Label l = "NegexpTestDefSimConstruction";
		Negexp negExp1( l, -inverseMean );
		CHECK_EQUAL( "ODEMx WARNING: Negexp(): expected value 'na' less or equal 0; -na or 0.001 used -- Object: NegexpTestDefSimConstruction\n",
				errorStream().str() );

		CHECK_EQUAL( l, negExp1.getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), negExp1.getStatisticManager() );
		CHECK_EQUAL( inverseMean, negExp1.getA() );

		clearErrorStream();
		l = "NegexpTestUserSimConstruction";
		Negexp negExp2( &sim, l, -inverseMean );
		CHECK_EQUAL( "ODEMx WARNING: Negexp(): expected value 'na' less or equal 0; -na or 0.001 used -- Object: NegexpTestUserSimConstruction\n",
				errorStream().str() );

		CHECK_EQUAL( l, negExp2.getLabel() );
		CHECK_EQUAL( &sim, negExp2.getStatisticManager() );
		CHECK_EQUAL( inverseMean, negExp2.getA() );
	}

	/**
	 * @test odemx::Negexp::sample()
	 *
	 * Expected function call effects:
	 * @li samples are always greater or equal \c 0
	 * @li the given inverseMean value is very close to the computed one
	 * @li the given inverseMean value is very close to the variance
	 */
	TEST_FIXTURE( NegexpFixture, Sample )
	{
		for( int i = 0; i < 1000000; ++i )
		{
			double value = negExp.sample();
			CHECK( 0 <= value );
			histogram.update( value );
		}

		CHECK_CLOSE( 1 / negExp.getA(), histogram.getMean(), 0.001  );
		CHECK_CLOSE( 1 / std::pow( negExp.getA(), 2 ), std::pow( histogram.getStandardDeviation(), 2 ), 0.001  );

		// print histogram
//		reporter.addProducer( &histogram );
//		reporter.generateReport();
//		CHECK( reporter.processedTables );
//
//		vector< string >& table = reporter.allTableContent[1];
//		for( size_t i = 0; i < table.size(); ++i )
//		{
//			cout << table[i] << " ";
//			if( ( i % 5 ) == 4 ) cout << endl;
//		}
//		cout << endl;
	}

	/**
	 * @test odemx::Negexp::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( NegexpFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "NegexpTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		negExp.reset();

		for( int i = 0; i < 123; ++i )
		{
			negExp.sample();
		}

		reporter.addProducer( &negExp );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		CHECK_EQUAL( toString( negExp.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( negExp.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( "Negexp", reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( long( negExp.getUses() ) ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( long( negExp.getSeed() ) ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( double( negExp.getA() ) ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][7] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
