/**
 * @file TestNormal.cpp
 * @date Aug 9, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Normal
 */

#include "TestRandom.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Random )
{
/// @endcond

	/**
	 * @struct NormalFixture
	 * @brief Helper struct providing set-up/tear-down of Normal tests
	 *
	 * @copydetails EventFixture
	 */
	struct NormalFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		double mean;
		double deviation;
		int cells;
		Normal normal;
		Histo histogram;

		NormalFixture()
		:	sim( "NormalTestSim" ),
			reporter(),
			mean( 10 ),
			deviation( 5 ),
			cells( (int)mean * 2 ),
			normal( &sim, "NormalTest", mean, deviation ),
			histogram( &sim, "NormalTestHistogram", 0, mean * 2, cells )
			{}
	};

	/**
	 * @test odemx::Normal construction
	 *
	 * Expected function call effects:
	 * @li negative mean value produces a warning and inverts it
	 * @li label is set
	 * @li dist context (simulation) is set
	 * @li mean value is set
	 */
	TEST_FIXTURE( NormalFixture, ConstructionDestruction )
	{
		clearErrorStream();
		Label l = "NormalTestDefSimConstruction";
		Normal normal1( l, mean, -deviation );
		CHECK_EQUAL( "ODEMx WARNING: Normal(): given deviation is less than 0; changed to -nb -- Object: NormalTestDefSimConstruction\n",
				errorStream().str() );

		CHECK_EQUAL( l, normal1.getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), normal1.getStatisticManager() );
		CHECK_EQUAL( mean, normal1.getA() );
		CHECK_EQUAL( deviation, normal1.getB() );

		clearErrorStream();
		l = "NormalTestUserSimConstruction";
		Normal normal2( &sim, l, mean, -deviation );
		CHECK_EQUAL( "ODEMx WARNING: Normal(): given deviation is less than 0; changed to -nb -- Object: NormalTestUserSimConstruction\n",
				errorStream().str() );

		CHECK_EQUAL( l, normal2.getLabel() );
		CHECK_EQUAL( &sim, normal2.getStatisticManager() );
		CHECK_EQUAL( mean, normal2.getA() );
		CHECK_EQUAL( deviation, normal2.getB() );
	}

	/**
	 * @test odemx::Normal::sample()
	 *
	 * Expected function call effects:
	 * @li samples are always greater or equal \c 0
	 * @li the given mean value is very close to the computed one
	 * @li the given mean value is very close to the variance
	 *
	 * @warning The computed mean value for this test differs from the given
	 * one by almost \c 0.5.
	 */
	TEST_FIXTURE( NormalFixture, Sample )
	{
		for( int i = 0; i < 1000000; ++i )
		{
			double value = normal.sample();
			histogram.update( value );
		}

		CHECK_CLOSE( normal.getA(), histogram.getMean(), 0.5  );
		CHECK_CLOSE( normal.getB(), histogram.getStandardDeviation(), 0.05  );

		// print histogram
//		reporter.addProducer( &histogram );
//		reporter.generateReport();
//		CHECK( reporter.processedTables );
//
//		vector< string >& table = reporter.allTableContent[1];
//		for( size_t i = 0; i < table.size(); ++i )
//		{
//			cout << table[i] << " ";
//			if( ( i % 5 ) == 4 ) cout << endl;
//		}
//		cout << endl;
	}

	/**
	 * @test odemx::Normal::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( NormalFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "NormalTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		normal.reset();

		for( int i = 0; i < 123; ++i )
		{
			normal.sample();
		}

		reporter.addProducer( &normal );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		CHECK_EQUAL( toString( normal.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( normal.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( "Normal", reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( long( normal.getUses() ) ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( long( normal.getSeed() ) ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( double( normal.getA() ) ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( double( normal.getB() ) ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][7] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
