/**
 * @file TestPoisson.cpp
 * @date Aug 9, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Poisson
 */

#include "TestRandom.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"
#include <cmath>

/// @cond DOXYGEN_SKIP
SUITE( Random )
{
/// @endcond

	/**
	 * @struct PoissonFixture
	 * @brief Helper struct providing set-up/tear-down of Poisson tests
	 *
	 * @copydetails EventFixture
	 */
	struct PoissonFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		double mean;
		int cells;
		Poisson poisson;
		Histo histogram;

		PoissonFixture()
		:	sim( "PoissonTestSim" ),
			reporter(),
			mean( 100 ),
			cells( (int)mean * 2 ),
			poisson( &sim, "PoissonTest", mean ),
			histogram( &sim, "PoissonTestHistogram", 0, mean * 2, cells )
			{}
	};

	/**
	 * @test odemx::Poisson construction
	 *
	 * Expected function call effects:
	 * @li negative mean value produces a warning and inverts it
	 * @li label is set
	 * @li dist context (simulation) is set
	 * @li mean value is set
	 */
	TEST_FIXTURE( PoissonFixture, ConstructionDestruction )
	{
		clearErrorStream();
		Label l = "PoissonTestDefSimConstruction";
		Poisson poisson1( l, -mean );
		CHECK_EQUAL( "ODEMx WARNING: Poisson(): mean a is <= 0.0; changed to -a -- Object: PoissonTestDefSimConstruction\n",
				errorStream().str() );

		CHECK_EQUAL( l, poisson1.getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), poisson1.getStatisticManager() );
		CHECK_EQUAL( mean, poisson1.getA() );

		clearErrorStream();
		l = "PoissonTestUserSimConstruction";
		Poisson poisson2( &sim, l, -mean );
		CHECK_EQUAL( "ODEMx WARNING: Poisson(): mean a is <= 0.0; changed to -a -- Object: PoissonTestUserSimConstruction\n",
				errorStream().str() );

		CHECK_EQUAL( l, poisson2.getLabel() );
		CHECK_EQUAL( &sim, poisson2.getStatisticManager() );
		CHECK_EQUAL( mean, poisson2.getA() );
	}

	/**
	 * @test odemx::Poisson::sample()
	 *
	 * Expected function call effects:
	 * @li samples are always greater or equal \c 0
	 * @li the given mean value is very close to the computed one
	 * @li the given mean value is very close to the variance
	 */
	TEST_FIXTURE( PoissonFixture, Sample )
	{
		for( int i = 0; i < 1000000; ++i )
		{
			int value = poisson.sample();
			CHECK( 0 <= value );
			histogram.update( value );
		}

		CHECK_CLOSE( poisson.getA(), histogram.getMean(), 0.01  );
		CHECK_CLOSE( poisson.getA(), std::pow( histogram.getStandardDeviation(), 2 ), 0.1  );

		// print histogram
//		reporter.addProducer( &histogram );
//		reporter.generateReport();
//		CHECK( reporter.processedTables );
//
//		vector< string >& table = reporter.allTableContent[1];
//		for( size_t i = 0; i < table.size(); ++i )
//		{
//			cout << table[i] << " ";
//			if( ( i % 5 ) == 4 ) cout << endl;
//		}
//		cout << endl;
	}

	/**
	 * @test odemx::Poisson::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( PoissonFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "PoissonTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		poisson.reset();

		for( int i = 0; i < 123; ++i )
		{
			poisson.sample();
		}

		reporter.addProducer( &poisson );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		CHECK_EQUAL( toString( poisson.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( poisson.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( "Poisson", reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( long( poisson.getUses() ) ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( long( poisson.getSeed() ) ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( double( poisson.getA() ) ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][7] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
