/**
 * @file TestRandint.cpp
 * @date Aug 9, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Randint
 */

#include "TestRandom.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"
#include <cstdlib> // atof()

/// @cond DOXYGEN_SKIP
SUITE( Random )
{
/// @endcond

	/**
	 * @struct RandintFixture
	 * @brief Helper struct providing set-up/tear-down of Randint tests
	 *
	 * @copydetails EventFixture
	 */
	struct RandintFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		int lower;
		int upper;
		int cells;
		Randint rand;
		Histo histogram;

		RandintFixture()
		:	sim( "RandintTestSim" ),
			reporter(),
			lower( 0 ),
			upper( 4 ),
			cells( upper - lower ),
			rand( &sim, "RandintTest", lower, upper ),
			histogram( &sim, "RandintTestHistogram", lower, upper, cells )
			{}
	};

	/**
	 * @test odemx::Randint construction
	 *
	 * Expected function call effects:
	 * @li lower bound greater upper bound produces a warning and switches
	 * the two values
	 * @li label is set
	 * @li dist context (simulation) is set
	 * @li lower and upper bounds are set
	 */
	TEST_FIXTURE( RandintFixture, ConstructionDestruction )
	{
		clearErrorStream();
		Label l = "RandintTestDefSimConstruction";
		Randint rand1( l, upper, lower );
		CHECK_EQUAL( "ODEMx WARNING: Randint(): lower bound greater than upper bound; bounds swapped -- Object: RandintTestDefSimConstruction\n",
				errorStream().str() );

		CHECK_EQUAL( l, rand1.getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), rand1.getStatisticManager() );
		CHECK_EQUAL( lower, rand1.getA() );
		CHECK_EQUAL( upper, rand1.getB() );

		clearErrorStream();
		l = "RandintTestUserSimConstruction";
		Randint rand2( &sim, l, upper, lower );
		CHECK_EQUAL( "ODEMx WARNING: Randint(): lower bound greater than upper bound; bounds swapped -- Object: RandintTestUserSimConstruction\n",
				errorStream().str() );

		CHECK_EQUAL( l, rand2.getLabel() );
		CHECK_EQUAL( &sim, rand2.getStatisticManager() );
		CHECK_EQUAL( lower, rand2.getA() );
		CHECK_EQUAL( upper, rand2.getB() );
	}

	/**
	 * @test odemx::Randint::sample()
	 *
	 * Expected function call effects:
	 * @li samples are always within the interval [lower, upper]
	 * @li all values have roughly the same hit percentage
	 */
	TEST_FIXTURE( RandintFixture, Sample )
	{
		for( int i = 0; i < 1000000; ++i )
		{
			int value = rand.sample();
			CHECK( rand.getA() <= value );
			CHECK( rand.getB() >= value );
			histogram.update( value );
		}

		reporter.addProducer( &histogram );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		vector< string >& table = reporter.allTableContent[1];
		for( int i = 0; i <= cells; ++i )
		{
			CHECK_CLOSE( 100.0 / (cells + 1), atof( table[ i * 5 + 3 ].c_str() ), 0.1  );
		}

		// print histogram
//		for( size_t i = 0; i < table.size(); ++i )
//		{
//			cout << table[i] << " ";
//			if( ( i % 5 ) == 4 ) cout << endl;
//		}
//		cout << endl;
	}

	/**
	 * @test odemx::Randint::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( RandintFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "RandintTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 42 );
		sim.step(); // event execution advances sim time
		rand.reset();

		for( int i = 0; i < 666; ++i )
		{
			rand.sample();
		}

		reporter.addProducer( &rand );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		CHECK_EQUAL( toString( rand.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( rand.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( "Randint", reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( long( rand.getUses() ) ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( long( rand.getSeed() ) ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( double( rand.getA() ) ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( double( rand.getB() ) ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][7] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
