/**
 * @file TestUniform.cpp
 * @date Aug 9, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Uniform
 */

#include "TestRandom.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"
#include <cstdlib> // atof()
#include <cmath> // floor()

/// @cond DOXYGEN_SKIP
SUITE( Random )
{
/// @endcond

	/**
	 * @struct UniformFixture
	 * @brief Helper struct providing set-up/tear-down of Uniform tests
	 *
	 * @copydetails EventFixture
	 */
	struct UniformFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		double lower;
		double upper;
		int cells;
		Uniform uniform;
		Histo histogram;

		UniformFixture()
		:	sim( "UniformTestSim" ),
			reporter(),
			lower( 0.0 ),
			upper( 4.0 ),
			cells( (int)floor(upper - lower) * 5 ),
			uniform( &sim, "UniformTest", lower, upper ),
			histogram( &sim, "UniformTestHistogram", lower, upper, cells )
			{}
	};

	/**
	 * @test odemx::Uniform construction
	 *
	 * Expected function call effects:
	 * @li lower bound greater upper bound produces a warning and switches
	 * the two values
	 * @li label is set
	 * @li dist context (simulation) is set
	 * @li lower and upper bounds are set
	 */
	TEST_FIXTURE( UniformFixture, ConstructionDestruction )
	{
		clearErrorStream();
		Label l = "UniformTestDefSimConstruction";
		Uniform uniform1( l, upper, lower );
		CHECK_EQUAL( "ODEMx WARNING: Uniform(): lower bound greater than upper bound; bounds swapped -- Object: UniformTestDefSimConstruction\n",
				errorStream().str() );

		CHECK_EQUAL( l, uniform1.getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), uniform1.getStatisticManager() );
		CHECK_EQUAL( lower, uniform1.getA() );
		CHECK_EQUAL( upper, uniform1.getB() );

		clearErrorStream();
		l = "UniformTestUserSimConstruction";
		Uniform uniform2( &sim, l, upper, lower );
		CHECK_EQUAL( "ODEMx WARNING: Uniform(): lower bound greater than upper bound; bounds swapped -- Object: UniformTestUserSimConstruction\n",
				errorStream().str() );

		CHECK_EQUAL( l, uniform2.getLabel() );
		CHECK_EQUAL( &sim, uniform2.getStatisticManager() );
		CHECK_EQUAL( lower, uniform2.getA() );
		CHECK_EQUAL( upper, uniform2.getB() );
	}

	/**
	 * @test odemx::Uniform::sample()
	 *
	 * Expected function call effects:
	 * @li samples are always within the interval [lower, upper]
	 * @li all values have roughly the same hit percentage
	 */
	TEST_FIXTURE( UniformFixture, Sample )
	{
		for( int i = 0; i < 5000000; ++i )
		{
			double value = uniform.sample();
			CHECK( uniform.getA() <= value );
			CHECK( uniform.getB() >= value );
			histogram.update( value );
		}

		reporter.addProducer( &histogram );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		vector< string >& table = reporter.allTableContent[1];
		for( int i = 0; i < cells; ++i )
		{
			CHECK_CLOSE( 100.0 / cells, atof( table[ i * 5 + 3 ].c_str() ), 0.02  );
		}

		// print histogram
//		for( size_t i = 0; i < table.size(); ++i )
//		{
//			cout << table[i] << " ";
//			if( ( i % 5 ) == 4 ) cout << endl;
//		}
//		cout << endl;
	}

	/**
	 * @test odemx::Uniform::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( UniformFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "UniformTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 42 );
		sim.step(); // event execution advances sim time
		uniform.reset();

		for( int i = 0; i < 666; ++i )
		{
			uniform.sample();
		}

		reporter.addProducer( &uniform );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		CHECK_EQUAL( toString( uniform.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( uniform.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( "Uniform", reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( long( uniform.getUses() ) ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( long( uniform.getSeed() ) ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( double( uniform.getA() ) ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( double( uniform.getB() ) ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( 0.0 ), reporter.allTableContent[0][7] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
