/**
 * @file TestAccum.cpp
 * @date Aug 6, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Accum
 */

#include "TestStatistics.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

#include <cmath>

/// @cond DOXYGEN_SKIP
SUITE( Statistics )
{
/// @endcond

	/**
	 * @struct AccumFixture
	 * @brief Helper struct providing set-up/tear-down of Accum tests
	 *
	 * @copydetails EventFixture
	 */
	struct AccumFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		Accum accum, accumLin;

		AccumFixture()
		:	sim( "AccumTestSim" ),
			reporter(),
			accum( &sim, "AccumTestStepwise", Accum::stepwise ),
			accumLin( &sim, "AccumTestLinear", Accum::linear )
			{}
	};

	/**
	 * @test odemx::Accum construction
	 *
	 * Expected effects:
	 * @li label is set correctly
	 * @li statistics manager (Simulation) is set correctly
	 * @li accum values are 0-initialized
	 * @li last reset time is current SimTime
	 */
	TEST_FIXTURE( AccumFixture, ConstructionDestruction )
	{
		Label label = "TestAccumConstructionDefSim";
		std::auto_ptr< Accum > accum1( new Accum( label ) );
		CHECK_EQUAL( label, accum1 -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), accum1 -> getStatisticManager() );
		CHECK_EQUAL( (unsigned int) 0, accum1 -> getSize() );
		CHECK_EQUAL( 0.0, accum1 -> getMin() );
		CHECK_EQUAL( 0.0, accum1 -> getMax() );
		CHECK_EQUAL( 0.0, accum1 -> getMean() );
		CHECK_EQUAL( 0.0, accum1 -> getStandardDeviation() );
		CHECK_EQUAL( getDefaultSimulation() -> getTime(), accum1 -> getResetTime() );

		label = "TestAccumConstructionUserSim";
		std::auto_ptr< Accum > accum2( new Accum( &sim, label ) );
		CHECK_EQUAL( label, accum2 -> getLabel() );
		CHECK_EQUAL( &sim, accum2 -> getStatisticManager() );
		CHECK_EQUAL( (unsigned int) 0, accum2 -> getSize() );
		CHECK_EQUAL( 0.0, accum2 -> getMin() );
		CHECK_EQUAL( 0.0, accum2 -> getMax() );
		CHECK_EQUAL( 0.0, accum2 -> getMean() );
		CHECK_EQUAL( 0.0, accum2 -> getStandardDeviation() );
		CHECK_EQUAL( sim.getTime(), accum2 -> getResetTime() );
	}

	/**
	 * @test odemx::Accum::update()
	 *
	 * Expected function call effects:
	 * @li usage counter is increased by the number of \c update() calls
	 * @li \c min and \c max are set correctly
	 * @li in stepwise mode, constant value between steps is assumed
	 * @li
	 */
	TEST_FIXTURE( AccumFixture, Update )
	{
		double min = 0.0;
		double max = 1.0;
		accum.update( max );
		accumLin.update( max );

		Base::EventTest simAdvanceEvent( &sim, "AccumTestUpdateAdvanceEvent" );
		simAdvanceEvent.scheduleIn( 4 );
		sim.step(); // event execution advances sim time

		accum.update( min );
		accumLin.update( min );
		simAdvanceEvent.scheduleIn( 1 );
		sim.step(); // event execution advances sim time

		CHECK_EQUAL( (unsigned int) 2, accum.getSize() );
		CHECK_EQUAL( min, accum.getMin() );
		CHECK_EQUAL( max, accum.getMax() );
		CHECK_EQUAL( 0.8, accum.getMean() );

		CHECK_EQUAL( 0.4, accumLin.getMean() );
	}

	/**
	 * @test odemx::Accum::reset()
	 *
	 * Expected function call effects:
	 * @li the accum values are set to 0
	 * @li the last reset time is set to the sim's time
	 */
	TEST_FIXTURE( AccumFixture, Reset )
	{
		SimTime time = 25;
		Base::EventTest simAdvanceEvent( &sim, "AccumTestResetAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		accum.update( 16 );
		accum.update( 11 );

		accum.reset();
		CHECK_EQUAL( (unsigned int) 0, accum.getSize() );
		CHECK_EQUAL( 0.0, accum.getMin() );
		CHECK_EQUAL( 0.0, accum.getMax() );
		CHECK_EQUAL( 0.0, accum.getMean() );
		CHECK_EQUAL( 0.0, accum.getStandardDeviation() );
		CHECK_EQUAL( time, accum.getResetTime() );
	}

	/**
	 * @test odemx::Accum::report(Report*)
	 *
	 * Expected function call effects:
	 * @li the report table contains correct values
	 */
	TEST_FIXTURE( AccumFixture, Report )
	{
		SimTime time = 25;
		Base::EventTest simAdvanceEvent( &sim, "AccumTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		accum.reset(); // non-0 reset time
		double min = 1.2;
		double max = 3.4;
		accum.update( min );
		simAdvanceEvent.scheduleIn( 1 );
		sim.step();
		accum.update( max );
		simAdvanceEvent.scheduleIn( 1 );
		sim.step();

		double mean = accum.getMean();
		double deviation = std::sqrt(
				( ( min - mean ) * ( min - mean )
				+ ( max - mean ) * ( max - mean ) ) / 2 );

		CHECK_EQUAL( (unsigned int) 2, accum.getSize() );
		CHECK_EQUAL( min, accum.getMin() );
		CHECK_EQUAL( max, accum.getMax() );
		CHECK_EQUAL( ( min + max ) / 2, mean );
		CHECK_CLOSE( deviation, accum.getStandardDeviation(), 0.000000000000001 );

		reporter.addProducer( &accum );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		// check the table line
		CHECK_EQUAL( accum.getLabel(), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( accum.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( toString( accum.getObs() ), reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( accum.getSize() ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( accum.getMin() ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( accum.getMax() ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( accum.getMean() ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( accum.getStandardDeviation() ), reporter.allTableContent[0][7] );
	}

	/**
	 * @test odemx::Accum::getMean()
	 *
	 * Expected function call effects:
	 * @li the mean is computed correctly
	 */
	TEST_FIXTURE( AccumFixture, GetMean )
	{
		Base::EventTest simAdvanceEvent( &sim, "AccumTestGetMeanAdvanceEvent" );
		double sum = 0;
		int values[] = { 1, 5, 2, 7, 10, 5 };
		size_t count = ( sizeof( values ) / sizeof( int ) );

		for( size_t i = 0; i < count; ++i )
		{
			sum += values[i];
			accum.update( values[i] );
			simAdvanceEvent.scheduleIn( 1 );
			sim.step(); // event execution advances sim time
		}

		double mean = sum / count;

		CHECK_EQUAL( mean, accum.getMean() );
	}

	/**
	 * @test odemx::Accum::getStandardDeviation()
	 *
	 * Expected function call effects:
	 * @li the standard deviation is computed correctly
	 */
	TEST_FIXTURE( AccumFixture, GetStandardDeviation )
	{
		Base::EventTest simAdvanceEvent( &sim, "AccumTestGetStandardDevianceAdvanceEvent" );
		double sum = 0;
		int values[] = { 1, 5, 2, 7, 10, 5 };
		size_t count = ( sizeof( values ) / sizeof( int ) );

		for( size_t i = 0; i < count; ++i )
		{
			sum += values[i];
			accum.update( values[i] );
			simAdvanceEvent.scheduleIn( 1 );
			sim.step(); // event execution advances sim time
		}

		double mean = sum / count;
		double variance = 0;
		for( size_t i = 0; i < count; ++i )
		{
			variance += ( values[i] - mean ) * ( values[i] - mean );
		}
		variance /= count;

		double standardDeviation = std::sqrt( variance );

		CHECK_EQUAL( standardDeviation, accum.getStandardDeviation() );
		CHECK_CLOSE( standardDeviation, accum.getStandardDeviation(), 0.001 );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
