/**
 * @file TestCount.cpp
 * @date Aug 5, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Count
 */

#include "TestStatistics.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Statistics )
{
/// @endcond

	/**
	 * @struct CountFixture
	 * @brief Helper struct providing set-up/tear-down of Count tests
	 *
	 * @copydetails EventFixture
	 */
	struct CountFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		Count count;

		CountFixture()
		:	sim( "CountTestSim" ),
			reporter(),
			count( &sim, "CountTest")
			{}
	};

	/**
	 * @test odemx::Count construction
	 *
	 * Expected effects:
	 * @li label is set correctly
	 * @li statistics manager (Simulation) is set correctly
	 * @li counter value is 0
	 * @li last reset time is current SimTime
	 */
	TEST_FIXTURE( CountFixture, ConstructionDestruction )
	{
		Label label = "TestCountConstructionDefSim";
		std::auto_ptr< Count > count1( new Count( label ) );
		CHECK_EQUAL( label, count1 -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), count1 -> getStatisticManager() );
		CHECK_EQUAL( (int) 0, count1 -> getValue() );
		CHECK_EQUAL( getDefaultSimulation() -> getTime(), count1 -> getResetTime() );

		label = "TestCountConstructionUserSim";
		std::auto_ptr< Count > count2( new Count( &sim, label ) );
		CHECK_EQUAL( label, count2 -> getLabel() );
		CHECK_EQUAL( &sim, count2 -> getStatisticManager() );
		CHECK_EQUAL( (int) 0, count2 -> getValue() );
		CHECK_EQUAL( sim.getTime(), count2 -> getResetTime() );
	}

	/**
	 * @test odemx::Count::update()
	 *
	 * Expected function call effects:
	 * @li usage counter is increased by one
	 */
	TEST_FIXTURE( CountFixture, Update )
	{
		count.update();
		CHECK_EQUAL( 1, count.getValue() );

		int increase = 32;
		count.update( increase );
		CHECK_EQUAL( 1 + increase, count.getValue() );
	}

	/**
	 * @test odemx::Count::reset()
	 *
	 * Expected function call effects:
	 * @li the counter value is set to 0
	 * @li the last reset time is set to the sim's time
	 */
	TEST_FIXTURE( CountFixture, Reset )
	{
		SimTime time = 25;
		Base::EventTest simAdvanceEvent( &sim, "CountTestResetAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		count.update( 16 );
		count.update( 11 );
		CHECK( count.getValue() > 0 );

		count.reset();
		CHECK_EQUAL( (int) 0, count.getValue() );
		CHECK_EQUAL( time, count.getResetTime() );
	}

	/**
	 * @test odemx::Count::report(Report*)
	 *
	 * Expected function call effects:
	 * @li the report table contains correct values for label, reset time,
	 * number of uses, and the counter value
	 */
	TEST_FIXTURE( CountFixture, Report )
	{
		SimTime time = 25;
		Base::EventTest simAdvanceEvent( &sim, "CountTestResetAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		unsigned int inc = 13;
		count.reset(); // non-0 reset time
		count.update( inc ); // two updates
		count.update( inc ); // result in value inc * 2

		reporter.addProducer( &count );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		// check the table line
		CHECK_EQUAL( count.getLabel(), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( count.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( toString( count.getObs() ), reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( count.getValue() ), reporter.allTableContent[0][3] );
	}


/// @cond DOXYGEN_SKIP
}
/// @endcond
