/**
 * @file TestHisto.cpp
 * @date Aug 6, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Histo
 */

#include "TestStatistics.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

#include <cmath>

/// @cond DOXYGEN_SKIP
SUITE( Statistics )
{
/// @endcond

	/**
	 * @struct HistoFixture
	 * @brief Helper struct providing set-up/tear-down of Histo tests
	 * @note Base class Tally-specific properties are not tested with Histo.
	 * @copydetails EventFixture
	 */
	struct HistoFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		double lower;
		double upper;
		int cells;
		Histo histogram;

		HistoFixture()
		:	sim( "HistoTestSim" ),
			reporter(),
			lower( 0 ),
			upper( 5 ),
			cells( 5 ),
			histogram( &sim, "HistoTest", lower, upper, cells )
			{}
	};

	/**
	 * @test odemx::Histo construction
	 *
	 * Expected effects:
	 * @li label is set correctly
	 * @li statistics manager (Simulation) is set correctly
	 * @li equal borders produce a warning
	 * @li lower border greater than upper border produces a warning
	 * @li number of cells less than 1 produces a warning
	 */
	TEST_FIXTURE( HistoFixture, ConstructionDestruction )
	{
		Label label = "TestHistoConstructionDefSim";
		std::auto_ptr< Histo > histogram1( new Histo( label, 0, 1, 1 ) );
		CHECK_EQUAL( label, histogram1 -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), histogram1 -> getStatisticManager() );

		label = "TestHistoConstructionUserSim";
		std::auto_ptr< Histo > histogram2( new Histo( &sim, label, 0, 1, 1 ) );
		CHECK_EQUAL( label, histogram2 -> getLabel() );
		CHECK_EQUAL( &sim, histogram2 -> getStatisticManager() );

		clearErrorStream();
		std::auto_ptr< Histo > warn1( new Histo( &sim, "HistoWarning1", 1, 1, 1 ) );
		CHECK_EQUAL( "ODEMx WARNING: Histo() borders are equal; borders changed -- Object: HistoWarning1\n",
				errorStream().str() );

		clearErrorStream();
		std::auto_ptr< Histo > warn2( new Histo( &sim, "HistoWarning2", 2, 1, 1 ) );
		CHECK_EQUAL( "ODEMx WARNING: Histo() upper borders < lower border; borders swapped -- Object: HistoWarning2\n",
				errorStream().str() );

		clearErrorStream();
		std::auto_ptr< Histo > warn3( new Histo( &sim, "HistoWarning3", 1, 2, 0 ) );
		CHECK_EQUAL( "ODEMx WARNING: Histo() number of cells < 1; number increased -- Object: HistoWarning3\n",
				errorStream().str() );
	}

	/**
	 * @test odemx::Histo::reset()
	 *
	 * Expected function call effects:
	 * @li all data cell counters are reset to \c 0
	 */
	TEST_FIXTURE( HistoFixture, Reset )
	{
		histogram.update( 1 );
		histogram.update( 3 );
		histogram.update( 5 );

		const std::vector<int>& data = histogram.getData();
		histogram.reset();

		std::vector< int >::const_iterator i;
		for( i = data.begin(); i!= data.end(); ++i )
		{
			CHECK_EQUAL( 0, *i );
		}
	}

	/**
	 * @test odemx::Histo::update(double)
	 *
	 * Expected function call effects:
	 * @li the cell counter corresponding to the update-value is incremented
	 */
	TEST_FIXTURE( HistoFixture, Update )
	{
		histogram.update( -0.1 ); // less than lower limit
		histogram.update( 0.5 );
		histogram.update( 1.5 );
		histogram.update( 2.5 );
		histogram.update( 3.5 );
		histogram.update( 4.5 );
		histogram.update( 5.5 ); // greater than upper limit

		const std::vector<int>& data = histogram.getData();

		std::vector< int >::const_iterator i;
		for( i = data.begin(); i!= data.end(); ++i )
		{
			CHECK_EQUAL( 1, *i );
		}
	}

	/**
	 * @test odemx::Histo::report(Report*)
	 *
	 * Expected function call effects:
	 * @li the report table contains correct values
	 */
	TEST_FIXTURE( HistoFixture, Report )
	{
		SimTime time = 25;
		Base::EventTest simAdvanceEvent( &sim, "HistoTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		histogram.reset(); // non-0 reset time
		double min = 1.2;
		double max = 3.4;
		histogram.update( min );
		histogram.update( max );

		double mean = histogram.getMean();
		double deviation = std::sqrt(
				( ( min - mean ) * ( min - mean )
				+ ( max - mean ) * ( max - mean ) ) / 2 );

		CHECK_EQUAL( (unsigned int) 2, histogram.getSize() );
		CHECK_EQUAL( min, histogram.getMin() );
		CHECK_EQUAL( max, histogram.getMax() );
		CHECK_EQUAL( ( min + max ) / 2, mean );
		CHECK_CLOSE( deviation, histogram.getStandardDeviation(), 0.000000000000001 );

		reporter.addProducer( &histogram );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		// check the first table
		CHECK_EQUAL( histogram.getLabel(), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( histogram.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( toString( histogram.getObs() ), reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( lower ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( upper ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( cells ), reporter.allTableContent[0][5] );
		CHECK_EQUAL( toString( histogram.getSize() ), reporter.allTableContent[0][6] );
		CHECK_EQUAL( toString( histogram.getMin() ), reporter.allTableContent[0][7] );
		CHECK_EQUAL( toString( histogram.getMax() ), reporter.allTableContent[0][8] );
		CHECK_EQUAL( toString( histogram.getMean() ), reporter.allTableContent[0][9] );
		CHECK_EQUAL( toString( histogram.getStandardDeviation() ), reporter.allTableContent[0][10] );

		// check second table, line 1
		CHECK_EQUAL( toString( lower ), reporter.allTableContent[1][0] ); // lower cell limit
		CHECK_EQUAL( toString( lower + 1 ), reporter.allTableContent[1][1] ); // upper cell limit
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[1][2] ); // cell hit counter
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[1][3] ); // hit percentage
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[1][4] ); // bar display value
		// line 2
		CHECK_EQUAL( toString( lower + 1), reporter.allTableContent[1][5] ); // lower cell limit
		CHECK_EQUAL( toString( lower + 2 ), reporter.allTableContent[1][6] ); // upper cell limit
		CHECK_EQUAL( toString( 1 ), reporter.allTableContent[1][7] ); // cell hit counter
		CHECK_EQUAL( toString( 50 ), reporter.allTableContent[1][8] ); // hit percentage
		CHECK_EQUAL( toString( 100 ), reporter.allTableContent[1][9] ); // bar display value
		// line 3
		CHECK_EQUAL( toString( lower + 2), reporter.allTableContent[1][10] ); // lower cell limit
		CHECK_EQUAL( toString( lower + 3 ), reporter.allTableContent[1][11] ); // upper cell limit
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[1][12] ); // cell hit counter
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[1][13] ); // hit percentage
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[1][14] ); // bar display value
		// line 4
		CHECK_EQUAL( toString( lower + 3), reporter.allTableContent[1][15] ); // lower cell limit
		CHECK_EQUAL( toString( lower + 4 ), reporter.allTableContent[1][16] ); // upper cell limit
		CHECK_EQUAL( toString( 1 ), reporter.allTableContent[1][17] ); // cell hit counter
		CHECK_EQUAL( toString( 50 ), reporter.allTableContent[1][18] ); // hit percentage
		CHECK_EQUAL( toString( 100 ), reporter.allTableContent[1][19] ); // bar display value
		// line 5
		CHECK_EQUAL( toString( lower + 4), reporter.allTableContent[1][20] ); // lower cell limit
		CHECK_EQUAL( toString( lower + 5 ), reporter.allTableContent[1][21] ); // upper cell limit
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[1][22] ); // cell hit counter
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[1][23] ); // hit percentage
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[1][24] ); // bar display value
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
