/**
 * @file TestSum.cpp
 * @date Aug 6, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Sum
 */

#include "TestStatistics.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Statistics )
{
/// @endcond

	/**
	 * @struct SumFixture
	 * @brief Helper struct providing set-up/tear-down of Sum tests
	 *
	 * @copydetails EventFixture
	 */
	struct SumFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		Sum sum;

		SumFixture()
		:	sim( "SumTestSim" ),
			reporter(),
			sum( &sim, "SumTest" )
			{}
	};

	/**
	 * @test odemx::Sum construction
	 *
	 * Expected effects:
	 * @li label is set correctly
	 * @li statistics manager (Simulation) is set correctly
	 * @li sum value is 0
	 * @li last reset time is current SimTime
	 */
	TEST_FIXTURE( SumFixture, ConstructionDestruction )
	{
		Label label = "TestSumConstructionDefSim";
		std::auto_ptr< Sum > sum1( new Sum( label ) );
		CHECK_EQUAL( label, sum1 -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), sum1 -> getStatisticManager() );
		CHECK_EQUAL( (int) 0, sum1 -> getValue() );
		CHECK_EQUAL( getDefaultSimulation() -> getTime(), sum1 -> getResetTime() );

		label = "TestSumConstructionUserSim";
		std::auto_ptr< Sum > sum2( new Sum( &sim, label ) );
		CHECK_EQUAL( label, sum2 -> getLabel() );
		CHECK_EQUAL( &sim, sum2 -> getStatisticManager() );
		CHECK_EQUAL( (int) 0, sum2 -> getValue() );
		CHECK_EQUAL( sim.getTime(), sum2 -> getResetTime() );
	}

	/**
	 * @test odemx::Sum::update()
	 *
	 * Expected function call effects:
	 * @li usage counter is increased by one
	 */
	TEST_FIXTURE( SumFixture, Update )
	{
		sum.update();
		CHECK_EQUAL( (unsigned int) 1, sum.getValue() );

		unsigned int increase = 32;
		sum.update( increase );
		CHECK_EQUAL( 1 + increase, sum.getValue() );
	}

	/**
	 * @test odemx::Sum::reset()
	 *
	 * Expected function call effects:
	 * @li the sum value is set to 0
	 * @li the last reset time is set to the sim's time
	 */
	TEST_FIXTURE( SumFixture, Reset )
	{
		SimTime time = 25;
		Base::EventTest simAdvanceEvent( &sim, "SumTestResetAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		sum.update( 16 );
		sum.update( 11 );
		CHECK( sum.getValue() > 0 );

		sum.reset();
		CHECK_EQUAL( (unsigned int) 0, sum.getValue() );
		CHECK_EQUAL( time, sum.getResetTime() );
	}

	/**
	 * @test odemx::Sum::report(Report*)
	 *
	 * Expected function call effects:
	 * @li the report table contains correct values for label, reset time,
	 * number of uses, and the counter value
	 */
	TEST_FIXTURE( SumFixture, Report )
	{
		SimTime time = 25;
		Base::EventTest simAdvanceEvent( &sim, "SumTestResetAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time

		unsigned int inc = 13;
		sum.reset(); // non-0 reset time
		sum.update( inc ); // two updates
		sum.update( inc ); // result in value inc * 2

		reporter.addProducer( &sum );
		reporter.generateReport();
		CHECK( reporter.processedTables );

		// check the table line
		CHECK_EQUAL( sum.getLabel(), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( sum.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( toString( sum.getObs() ), reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( sum.getValue() ), reporter.allTableContent[0][3] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
