/**
 * @file TestTab.cpp
 * @date Aug 5, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Tab
 */

#include "TestStatistics.h"
#include "../TestBase/TestBase.h"

/// @cond DOXYGEN_SKIP
SUITE( Statistics )
{
/// @endcond

	/**
	 * @struct TabFixture
	 * @brief Helper struct providing set-up/tear-down of Tab tests
	 *
	 * @copydetails EventFixture
	 */
	struct TabFixture
	{
		Base::SimulationTest sim;
		TabTest tab;

		TabFixture()
		:	sim( "TabTestSim" ),
			tab( &sim, "TabTest" )
			{}
	};

	/**
	 * @test odemx::Tab construction
	 *
	 * Expected effects:
	 * @li label is set correctly
	 * @li statistics manager (Simulation) is set correctly
	 * @li usage count is 0
	 * @li last reset time is current SimTime
	 */
	TEST_FIXTURE( TabFixture, ConstructionDestruction )
	{
		Label label = "TestTabConstructionDefSim";
		std::auto_ptr< Tab > tab( new TabTest( label ) );
		CHECK_EQUAL( label, tab -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), tab -> getStatisticManager() );
		CHECK_EQUAL( (unsigned int) 0, tab -> getObs() );
		CHECK_EQUAL( getDefaultSimulation() -> getTime(), tab -> getResetTime() );

		label = "TestTabConstructionUserSim";
		std::auto_ptr< Tab > tab2( new TabTest( &sim, label ) );
		CHECK_EQUAL( label, tab2 -> getLabel() );
		CHECK_EQUAL( &sim, tab2 -> getStatisticManager() );
		CHECK_EQUAL( (unsigned int) 0, tab2 -> getObs() );
		CHECK_EQUAL( sim.getTime(), tab2 -> getResetTime() );
	}

	/**
	 * @test odemx::Tab::update()
	 *
	 * Expected function call effects:
	 * @li usage counter is increased by one
	 */
	TEST_FIXTURE( TabFixture, Update )
	{
		tab.update();
		CHECK_EQUAL( (unsigned int) 1, tab.getObs() );
	}

	/**
	 * @test odemx::Tab::reset()
	 *
	 * Expected function call effects:
	 * @li the usage counter is set to 0
	 * @li the last reset time is set to the sim's time
	 */
	TEST_FIXTURE( TabFixture, Reset )
	{
		SimTime time = 25;
		Base::EventTest simAdvanceEvent( &sim, "TabTestResetAdvanceEvent" );
		simAdvanceEvent.scheduleAt( time );
		sim.step(); // event execution advances sim time
		tab.update();
		tab.update();
		CHECK( tab.getObs() > 0 );

		tab.reset();
		CHECK_EQUAL( (unsigned int) 0, tab.getObs() );
		CHECK_EQUAL( time, tab.getResetTime() );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
