/**
 * @file TestBin.cpp
 * @date Jul 23, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Bin
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	/**
	 * @struct BinFixture
	 * @brief Helper struct providing set-up/tear-down of Bin tests
	 *
	 * @copydetails EventFixture
	 */
	struct BinFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		BinTestObserver observer;
		odemx::Bin bin;

		BinFixture()
		:	sim( "BinTestSim" ),
			reporter(),
			observer( false ),
			bin( &sim, "BinTest", 25, &observer )
			{}
	};

	/**
	 * @test odemx::Bin construction and destruction
	 *
	 * Expected effects:
	 * @li observer is set correctly
	 * @li initial tokens are set
	 * @li label is set
	 * @li simulation environment is set for DefaultSimulation or user-definied
	 * simulation
	 * @li construction and destruction can be observed
	 */
	TEST_FIXTURE( BinFixture, ConstructionDestruction )
	{
		observer.history.clear();
		unsigned int tokens = 10;
		Label label = "BinTestConstructionWithDefaultSim";
		Bin* testBin = new Bin( label, tokens, &observer );
		CHECK_EQUAL( &observer, testBin -> getObservers().front() );
		CHECK_EQUAL( tokens, testBin -> getTokenNumber() );
		CHECK_EQUAL( tokens, testBin -> getInitial() );
		CHECK_EQUAL( label, testBin -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), testBin -> getTrace() );
		CHECK( observer.observed("onCreate", testBin -> getLabel() ) );

		delete testBin;
		CHECK( observer.observed("onDestroy", label ) );

		tokens = 99;
		label = "BinTestConstructionWithUserSim";
		Bin* testBin2 = new Bin( &sim, label, tokens, &observer );
		CHECK_EQUAL( &observer, testBin2 -> getObservers().front() );
		CHECK_EQUAL( tokens, testBin2 -> getTokenNumber() );
		CHECK_EQUAL( tokens, testBin2 -> getInitial() );
		CHECK_EQUAL( label, testBin2 -> getLabel() );
		CHECK_EQUAL( &sim, testBin2 -> getTrace() );
		CHECK( observer.observed("onCreate", testBin2 -> getLabel() ) );
	}

	/**
	 * @test odemx::Bin::take(unsigned int) success
	 *
	 * Expected function call effects:
	 * @li non-Process objects cause an error message
	 * @li the token number is decreased by the amount taken
	 * @li token number change and take success can be observed
	 * @li on success, the function returns the amount taken
	 * @li usage is registered in statistics
	 */
	TEST_FIXTURE( BinFixture, TakeSucceed )
	{
		unsigned int oldTokenNumber = bin.getTokenNumber();
		unsigned int takeCount = oldTokenNumber / 2 + 1;

		clearErrorStream();
		bin.take( takeCount );
		CHECK_EQUAL( "ODEMx ERROR: Bin::take(): called by non-Process object -- Object: BinTest\n", \
				errorStream().str() );

		BinTaker taker( &sim, "BinTestTakeProcess", bin, takeCount );
		taker.activate();
		sim.step();
		CHECK_EQUAL( oldTokenNumber - takeCount, bin.getTokenNumber() );
		CHECK( observer.observed( "onChangeTokenNumber", bin.getLabel(), \
				toString( oldTokenNumber ), toString( bin.getTokenNumber() ) ) );
		CHECK( observer.observed( "onTakeSucceed", bin.getLabel(), toString( takeCount ) ) );
		CHECK_EQUAL( takeCount, taker.received );
		CHECK_EQUAL( (unsigned int)1, bin.getUsers() );
	}

	/**
	 * @test odemx::Bin::take(unsigned int) failure
	 *
	 * Expected function call effects:
	 * @li a failing call can be observed
	 * @li the Process is inserted into a queue
	 * @li the Process is suspended, i.e. in state IDLE
	 * @li on Process interrupt, the function returns 0 and the Process is
	 * removed from the waiting queue
	 * @li waiting processes are awakened properly upon token increase
	 */
	TEST_FIXTURE( BinFixture, TakeFail )
	{
		unsigned int oldTokenNumber = bin.getTokenNumber();
		unsigned int takeCount = oldTokenNumber + 1;

		BinTaker taker( &sim, "BinTestTakeProcess", bin, takeCount );
		taker.activate();
		sim.step();
		CHECK( observer.observed( "onTakeFail", bin.getLabel(), toString( takeCount ) ) );
		CHECK_EQUAL( (size_t)1, bin.getWaitingProcessList().size() );
		CHECK_EQUAL( &taker, bin.getWaitingProcessList().front() );
		CHECK_EQUAL( Process::IDLE, taker.getProcessState() );

		Base::InterruptEvent interrupt( &sim, &taker );
		interrupt.schedule();
		sim.step();
		CHECK_EQUAL( Process::RUNNABLE, taker.getProcessState() );
		sim.step();
		CHECK_EQUAL( (unsigned int)0, taker.received );
		CHECK_EQUAL( (size_t)0, bin.getWaitingProcessList().size() );

		BinTaker taker2( &sim, "BinTestTakeProcess", bin, takeCount );
		taker2.hold();
		BinTaker taker3( &sim, "BinTestTakeProcess", bin, takeCount );
		taker3.hold();
		sim.step();
		sim.step();
		CHECK_EQUAL( (size_t)2, bin.getWaitingProcessList().size() );

		bin.give( takeCount * 2 );
		sim.run();
		CHECK_EQUAL( takeCount, taker2.received );
		CHECK_EQUAL( takeCount, taker3.received );
		CHECK_EQUAL( (size_t)0, bin.getWaitingProcessList().size() );
	}

	/**
	 * @test odemx::Bin::give(unsigned int)
	 *
	 * Expected function call effects:
	 * @li token number change and give call can be observed
	 * @li waiting processes are awakened properly upon token increase
	 */
	TEST_FIXTURE( BinFixture, Give )
	{
		unsigned int oldTokenNumber = bin.getTokenNumber();
		unsigned int giveCount = oldTokenNumber + 1;

		BinTaker taker( &sim, "BinTestTakeProcess", bin, oldTokenNumber + 1 );
		taker.hold();
		sim.step();
		CHECK_EQUAL( (size_t)1, bin.getWaitingProcessList().size() );

		bin.give( giveCount );
		CHECK( observer.observed( "onChangeTokenNumber", bin.getLabel(), \
				toString( oldTokenNumber ), toString( bin.getTokenNumber() ) ) );
		CHECK( observer.observed( "onGive", bin.getLabel(), toString( giveCount ) ) );

		sim.run();
		CHECK_EQUAL( oldTokenNumber + 1, taker.received );
	}

	/**
	 * @test odemx::Bin::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( BinFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "BinTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		bin.reset();

		BinTaker taker( &sim, "BinTestReportTakerProcess", bin, 15 );
		taker.activate();
		sim.step();
		bin.give( 25 );

		reporter.addProducer( &bin );
		reporter.generateReport();
		CHECK( reporter.processedTables );
		CHECK_EQUAL( (size_t) 2, reporter.allTableContent.size() );

		// only check second table, the first one is the queue's
		CHECK_EQUAL( toString( bin.getLabel() ), reporter.allTableContent[1][0] );
		CHECK_EQUAL( toString( bin.getResetTime() ), reporter.allTableContent[1][1] );
		CHECK_EQUAL( bin.getLabel() + " queue", reporter.allTableContent[1][2] );
		CHECK_EQUAL( toString( bin.getUsers() ), reporter.allTableContent[1][3] );
		CHECK_EQUAL( toString( bin.getProviders() ), reporter.allTableContent[1][4] );
		CHECK_EQUAL( toString( bin.getInitial() ), reporter.allTableContent[1][5] );
		CHECK_EQUAL( toString( bin.getMin() ), reporter.allTableContent[1][6] );
		CHECK_EQUAL( toString( bin.getMax() ), reporter.allTableContent[1][7] );
		CHECK_EQUAL( toString( bin.getTokenNumber() ), reporter.allTableContent[1][8] );
		CHECK_EQUAL( toString( bin.getAVFreeToken() ), reporter.allTableContent[1][9] );
		CHECK_EQUAL( toString( bin.getAVWaitTime() ), reporter.allTableContent[1][10] );
	}


/// @cond DOXYGEN_SKIP
}
/// @endcond

