/**
 * @file TestBinT.cpp
 * @date Jul 26, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class template BinT
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	// compare TokenTest objects
	bool operator== ( const TokenTest& t1, const TokenTest& t2 )
	{
		return ( t1.id == t2.id ) && ( t1.label == t2.label );
	}

	// make TokenTest objects streamable
	std::ostream& operator<< ( std::ostream& stream, const TokenTest& t )
	{
		stream << "(" << t.id << ", " << t.label << ")";
		return stream;
	}

	/**
	 * @struct BinTFixture
	 * @brief Helper struct providing set-up/tear-down of BinT tests
	 *
	 * @copydetails EventFixture
	 */
	struct BinTFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		BinTTestObserver observer;
		odemx::BinT< TokenTest > bin;

		BinTFixture()
		:	sim( "BinTTestSim" ),
			reporter(),
			observer( false ),
			bin( &sim, "BinTTest", &observer )
			{
				std::vector< TokenTest > tokens;
				for( int i = 0; i != 25; ++i )
				{
					tokens.push_back( TokenTest( i, string("BinTToken") + toString( i ) ) );
				}
				// add the tokens to the bin
				bin.give( tokens );
			}
	};

	/**
	 * @test odemx::BinT construction and destruction
	 *
	 * Expected effects:
	 * @li observer is set correctly
	 * @li label is set
	 * @li simulation environment is set for DefaultSimulation or user-definied
	 * simulation
	 * @li initial tokens are set
	 * @li construction and destruction can be observed
	 */
	TEST_FIXTURE( BinTFixture, ConstructionDestruction )
	{
		observer.history.clear();

		BinT< TokenTest >* bin1 =
			new BinT< TokenTest >( "BinTConstructionTest1", &observer );
		CHECK_EQUAL( &observer, bin1 -> getObservers().front() );
		CHECK_EQUAL( "BinTConstructionTest1", bin1 -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), bin1 -> getTrace() );
		CHECK_EQUAL( (unsigned int)0, bin1 -> getInitial() );
		CHECK_EQUAL( (unsigned int)0, bin1 -> getTokenNumber() );
		CHECK( observer.observed( "onCreate", bin1 -> getLabel() ) );

		BinT< TokenTest >* bin2 =
			new BinT< TokenTest >( &sim, "BinTConstructionTest2", &observer );
		CHECK_EQUAL( &observer, bin2 -> getObservers().front() );
		CHECK_EQUAL( "BinTConstructionTest2", bin2 -> getLabel() );
		CHECK_EQUAL( &sim, bin2 -> getTrace() );
		CHECK_EQUAL( (unsigned int)0, bin2 -> getInitial() );
		CHECK_EQUAL( (unsigned int)0, bin2 -> getTokenNumber() );
		CHECK( observer.observed( "onCreate", bin2 -> getLabel() ) );

		std::vector< TokenTest > initialTokens;
		initialTokens.push_back( TokenTest( 1, "InitToken1" ) );
		initialTokens.push_back( TokenTest( 2, "InitToken2" ) );

		BinT< TokenTest >* bin3 =
			new BinT< TokenTest >( "BinTConstructionTest3", initialTokens, &observer );
		CHECK_EQUAL( &observer, bin3 -> getObservers().front() );
		CHECK_EQUAL( "BinTConstructionTest3", bin3 -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), bin3 -> getTrace() );
		CHECK_EQUAL( initialTokens.size(), bin3 -> getInitial() );
		CHECK_EQUAL( initialTokens.size(), bin3 -> getTokenNumber() );
		CHECK_EQUAL( initialTokens.front(), bin3 -> getTokenStore().front() );
		CHECK_EQUAL( initialTokens.back(), bin3 -> getTokenStore().back() );
		CHECK( observer.observed( "onCreate", bin3 -> getLabel() ) );

		Label label = "BinTConstructionTest4";
		BinT< TokenTest >* bin4 =
			new BinT< TokenTest >( label, initialTokens, &observer );
		CHECK_EQUAL( &observer, bin4 -> getObservers().front() );
		CHECK_EQUAL( label, bin4 -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), bin4 -> getTrace() );
		CHECK_EQUAL( initialTokens.size(), bin4 -> getInitial() );
		CHECK_EQUAL( initialTokens.size(), bin4 -> getTokenNumber() );
		CHECK_EQUAL( initialTokens.front(), bin4 -> getTokenStore().front() );
		CHECK_EQUAL( initialTokens.back(), bin4 -> getTokenStore().back() );
		CHECK( observer.observed( "onCreate", bin4 -> getLabel() ) );

		delete bin4;
		CHECK( observer.observed( "onDestroy", label ) );
		delete bin3;
		delete bin2;
		delete bin1;
	}

	/**
	 * @test odemx::BinT::take(unsigned int) success
	 *
	 * Expected function call effects:
	 * @li non-Process objects cause an error message
	 * @li the token number is decreased by the amount taken
	 * @li token number change and take success can be observed
	 * @li on success, the function returns the amount taken
	 * @li usage is registered in statistics
	 * @li the tokens returned are the first \c takeCount elements of the
	 * tokenStore
	 */
	TEST_FIXTURE( BinTFixture, TakeSucceed )
	{
		unsigned int oldTokenNumber = bin.getTokenNumber();
		unsigned int takeCount = oldTokenNumber / 2 + 1;

		clearErrorStream();
		bin.take( takeCount );
		CHECK_EQUAL( "ODEMx ERROR: BinT::take(): called by non-Process object -- Object: BinTTest\n", \
				errorStream().str() );

		BinTTaker taker( &sim, "BinTTestTakeSucceedProcess", bin, takeCount );
		taker.activate();
		sim.step();
		CHECK_EQUAL( oldTokenNumber - takeCount, bin.getTokenNumber() );
		CHECK_EQUAL( oldTokenNumber - takeCount, bin.getTokenStore().size() );
		CHECK( observer.observed( "onChangeTokenNumber", bin.getLabel(), \
				toString( oldTokenNumber ), toString( bin.getTokenNumber() ) ) );
		CHECK( observer.observed( "onTakeSucceed", bin.getLabel(), toString( takeCount ) ) );
		CHECK_EQUAL( (unsigned int)1, bin.getUsers() );

		for( unsigned int i = 0; i != takeCount; ++i )
		{
			CHECK_EQUAL( i, (unsigned int)taker.received -> at( i ).id );
		}
	}

	/**
	 * @test odemx::BinT::take(unsigned int) failure
	 *
	 * Expected function call effects:
	 * @li a failing call can be observed
	 * @li the Process is inserted into a queue
	 * @li the Process is suspended, i.e. in state IDLE
	 * @li on interrupt the function returns 0
	 * @li waiting processes are awakened properly upon token increase
	 */
	TEST_FIXTURE( BinTFixture, TakeFail )
	{
		unsigned int oldTokenNumber = bin.getTokenNumber();
		unsigned int takeCount = oldTokenNumber + 1;

		BinTTaker taker( &sim, "BinTTestTakeFailProcess1", bin, takeCount );
		taker.activate();
		sim.step();
		CHECK( observer.observed( "onTakeFail", bin.getLabel(), toString( takeCount ) ) );
		CHECK_EQUAL( (size_t)1, bin.getWaitingProcessList().size() );
		CHECK_EQUAL( &taker, bin.getWaitingProcessList().front() );
		CHECK_EQUAL( Process::IDLE, taker.getProcessState() );

		Base::InterruptEvent interrupt( &sim, &taker );
		interrupt.schedule();
		sim.step();
		CHECK_EQUAL( Process::RUNNABLE, taker.getProcessState() );
		sim.step();
		CHECK_EQUAL( static_cast< std::vector< TokenTest >* >( 0 ), taker.received );

		BinTTaker taker2( &sim, "BinTTestTakeFailProcess2", bin, takeCount );
		taker2.hold();
		BinTTaker taker3( &sim, "BinTTestTakeFailProcess3", bin, takeCount );
		taker3.hold();
		sim.step();
		sim.step();
		CHECK_EQUAL( (size_t)2, bin.getWaitingProcessList().size() );

		std::vector< TokenTest > tokens;
		int last = bin.getTokenStore().back().id;
		for( unsigned int i = last; i != last + takeCount * 2; ++i )
		{
			tokens.push_back( TokenTest( i, string("BinTToken") + toString( i ) ) );
		}
		bin.give( tokens );
		sim.run();
		CHECK_EQUAL( takeCount, taker2.received -> size() );
		CHECK_EQUAL( takeCount, taker3.received -> size() );
		CHECK_EQUAL( (size_t)0, bin.getWaitingProcessList().size() );
	}

	/**
	 * @test odemx::BinT::give(unsigned int)
	 *
	 * Expected function call effects:
	 * @li token number change and give call can be observed
	 * @li waiting processes are awakened properly upon token increase
	 */
	TEST_FIXTURE( BinTFixture, Give )
	{
		unsigned int oldTokenNumber = bin.getTokenNumber();
		unsigned int giveCount = oldTokenNumber + 1;

		BinTTaker taker( &sim, "BinTTestGiveProcess", bin, oldTokenNumber + 1 );
		taker.hold();
		sim.step();
		CHECK_EQUAL( (size_t)1, bin.getWaitingProcessList().size() );

		std::vector< TokenTest > tokens;
		int last = bin.getTokenStore().back().id;
		for( unsigned int i = last; i != last + giveCount; ++i )
		{
			tokens.push_back( TokenTest( i, string("BinTToken") + toString( i ) ) );
		}
		bin.give( tokens );
		CHECK( observer.observed( "onChangeTokenNumber", bin.getLabel(), \
				toString( oldTokenNumber ), toString( bin.getTokenNumber() ) ) );
		CHECK( observer.observed( "onGive", bin.getLabel(), toString( giveCount ) ) );

		sim.run();
		CHECK_EQUAL( oldTokenNumber + 1, taker.received -> size() );
	}

	/**
	 * @test odemx::BinT::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( BinTFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "BinTTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		bin.reset();

		BinTTaker taker( &sim, "BinTTestReportTakerProcess", bin, 15 );
		taker.activate();
		sim.step();

		std::vector< TokenTest > tokens;
		unsigned int last = bin.getTokenStore().back().id;
		for( unsigned int i = last; i != last + 25; ++i )
		{
			tokens.push_back( TokenTest( i, string("BinTToken") + toString( i ) ) );
		}
		bin.give( tokens );

		reporter.addProducer( &bin );
		reporter.generateReport();
		CHECK( reporter.processedTables );
		CHECK_EQUAL( (size_t) 2, reporter.allTableContent.size() );

		// only check second table, the first one is the queue's
		size_t len = string( bin.getLabel() + " queue" ).length();
		CHECK_EQUAL( toString( bin.getLabel() ), reporter.allTableContent[1][0] );
		CHECK_EQUAL( typeToString( typeid(TokenTest) ), reporter.allTableContent[1][1] );
		CHECK_EQUAL( toString( bin.getResetTime() ), reporter.allTableContent[1][2] );
		// this test uses substr() because it returns Label "BinTTest queue_4" -- why???
		CHECK_EQUAL( bin.getLabel() + " queue", reporter.allTableContent[1][3].substr( 0, len ) );
		CHECK_EQUAL( toString( bin.getUsers() ), reporter.allTableContent[1][4] );
		CHECK_EQUAL( toString( bin.getProviders() ), reporter.allTableContent[1][5] );
		CHECK_EQUAL( toString( bin.getInitial() ), reporter.allTableContent[1][6] );
		CHECK_EQUAL( toString( bin.getMin() ), reporter.allTableContent[1][7] );
		CHECK_EQUAL( toString( bin.getMax() ), reporter.allTableContent[1][8] );
		CHECK_EQUAL( toString( bin.getTokenNumber() ), reporter.allTableContent[1][9] );
		CHECK_EQUAL( toString( bin.getAVFreeToken() ), reporter.allTableContent[1][10] );
		CHECK_EQUAL( toString( bin.getAVWaitTime() ), reporter.allTableContent[1][11] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
