/**
 * @file TestCondQ.cpp
 * @date Jul 26, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class CondQ
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	/**
	 * @struct CondQFixture
	 * @brief Helper struct providing set-up/tear-down of CondQ tests
	 *
	 * @copydetails EventFixture
	 */
	struct CondQFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		CondQTestObserver observer;
		odemx::CondQ condQ;

		CondQFixture()
		:	sim( "CondQTestSim" ),
			reporter(),
			observer( false ),
			condQ( &sim, "CondQTest", &observer )
			{}
	};

	/**
	 * @test odemx::CondQ construction and destruction
	 *
	 * Expected effects:
	 * @li observer is set correctly
	 * @li label is set
	 * @li simulation environment is set for DefaultSimulation or user-definied
	 * simulation
	 * @li construction and destruction can be observed
	 */
	TEST_FIXTURE( CondQFixture, ConstructionDestruction )
	{
		observer.history.clear();

		Label label( "CondQConstructionTest1" );
		CondQ* testCondQ = new CondQ( label, &observer );
		CHECK_EQUAL( &observer, testCondQ -> getObservers().front() );
		CHECK_EQUAL( label, testCondQ -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), testCondQ -> getTrace() );
		CHECK( observer.observed( "onCreate", testCondQ -> getLabel() ) );

		label = "CondQConstructionTest2";
		CondQ* testCondQ2 = new CondQ( &sim, label, &observer );
		CHECK_EQUAL( &observer, testCondQ2 -> getObservers().front() );
		CHECK_EQUAL( label, testCondQ2 -> getLabel() );
		CHECK_EQUAL( &sim, testCondQ2 -> getTrace() );
		CHECK( observer.observed( "onCreate", testCondQ2 -> getLabel() ) );

		delete testCondQ2;
		CHECK( observer.observed( "onDestroy", label ) );
		delete testCondQ;
	}

	/**
	 * @test odemx::CondQ::wait(Condition*)
	 *
	 * Expected function call effects:
	 * @li the call can be observed
	 * @li both waiting processes are added to the list
	 * @li a process with fulfilled condition is reactivated and the
	 * function returns \c true
	 * @li successful continuation of a process can be observed
	 * @li the statistics are updated
	 * @li for interrupted processes, the function returns \c false
	 * @li processes are removed from the list
	 */
	TEST_FIXTURE( CondQFixture, Wait )
	{
		unsigned int uses = condQ.getUsers();

		CondQTestProcess p1( &sim, "CondQTestWaitProcess1", condQ );
		CondQTestProcess p2( &sim, "CondQTestWaitProcess2", condQ );
		p1.hold();
		p2.hold();
		sim.step();
		sim.step();
		CHECK( p1.running );
		CHECK( observer.observed( "onWait", condQ.getLabel(), p1.getLabel() ) );
		CHECK( ! p1.reactivated );
		CHECK( p2.running );
		CHECK( observer.observed( "onWait", condQ.getLabel(), p2.getLabel() ) );
		CHECK( ! p2.reactivated );
		CHECK_EQUAL( (size_t)2, condQ.getProcessList().size() );
		CHECK( find( condQ.getProcessList().begin(), condQ.getProcessList().end(), &p1 )
				!= condQ.getProcessList().end() );
		CHECK( find( condQ.getProcessList().begin(), condQ.getProcessList().end(), &p2 )
				!= condQ.getProcessList().end() );

		// make the condition check return true
		p2.conditionValue = true;
		condQ.signal();
		sim.step();
		sim.step();
		CHECK( ! p1.reactivated );
		CHECK( p2.reactivated );
		CHECK( p2.waitReturnValue == true );
		CHECK( observer.observed( "onContinue", condQ.getLabel(), p2.getLabel() ) );
		CHECK_EQUAL( (unsigned int) uses + 1, condQ.getUsers() );
		CHECK( find( condQ.getProcessList().begin(), condQ.getProcessList().end(), &p2 )
				== condQ.getProcessList().end() );
		p1.interrupt();
		CHECK( p1.isInterrupted() );
		sim.step();
		CHECK( p1.reactivated );
		CHECK( p1.waitReturnValue == false );
		CHECK( find( condQ.getProcessList().begin(), condQ.getProcessList().end(), &p1 )
				== condQ.getProcessList().end() );
	}

	/**
	 * @test odemx::CondQ::signal()
	 *
	 * Expected function call effects:
	 * @li the call can be observed
	 * @li the statistics are updated
	 * @li all enqueued processes are rescheduled to check their
	 * condition function again
	 */
	TEST_FIXTURE( CondQFixture, Signal )
	{
		unsigned int signalCount = condQ.getSignals();

		CondQTestProcess p1( &sim, "CondQTestSignalProcess1", condQ );
		CondQTestProcess p2( &sim, "CondQTestSignalProcess2", condQ );
		p1.hold();
		p2.hold();
		sim.step();
		sim.step();

		CondQSignal signalInvoker( &sim, "CondQTestSignalInvoker", condQ );
		signalInvoker.activate();
		sim.step();

		//condQ.signal();

		// signal marks the current process, which should be 0,
		// since this is a call from outside the simulation
		CHECK( observer.observed( "onSignal", condQ.getLabel(), signalInvoker.getLabel() ) );
		CHECK_EQUAL( signalCount + 1, condQ.getSignals() );
		CHECK( p1.isScheduled() );
		CHECK( p2.isScheduled() );
	}

	/**
	 * @test odemx::CondQ::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( CondQFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "CondQTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		condQ.reset();

		CondQTestProcess p1( &sim, "CondQTestWaitProcess1", condQ );
		CondQTestProcess p2( &sim, "CondQTestWaitProcess2", condQ );
		p1.hold();
		p2.hold();
		sim.step(); // enqueue p1
		sim.step(); // enqueue p2

		simAdvanceEvent.scheduleIn( 5 );
		sim.step(); // event execution advances sim time

		reporter.addProducer( &condQ );
		reporter.generateReport();
		CHECK( reporter.processedTables );
		CHECK_EQUAL( (size_t) 2, reporter.allTableContent.size() );

		// only check second table, the first one is the condQ's
		CHECK_EQUAL( condQ.getLabel(), reporter.allTableContent[1][0] );
		CHECK_EQUAL( toString( condQ.getResetTime() ), reporter.allTableContent[1][1] );
		CHECK_EQUAL( condQ.getLabel() + " queue", reporter.allTableContent[1][2] );
		CHECK_EQUAL( toString( condQ.getUsers() ), reporter.allTableContent[1][3] );
		CHECK_EQUAL( toString( condQ.getSignals() ), reporter.allTableContent[1][4] );
		CHECK_EQUAL( toString( condQ.getZeroWait() ), reporter.allTableContent[1][5] );
		CHECK_EQUAL( toString( condQ.getAVWaitTime() ), reporter.allTableContent[1][6] );
	}
/// @cond DOXYGEN_SKIP
}
/// @endcond
