/**
 * @file TestMemory.cpp
 * @date Jul 22, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Memory
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	/**
	 * @struct MemoryFixture
	 * @brief Helper struct providing set-up/tear-down of Memory tests
	 *
	 * @copydetails EventFixture
	 */
	struct MemoryFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		MemoryTestObserver observer;
		MemoryTestProcess* p1;
		MemoryTestProcess* p2;
		MemoryTestProcess* p3;
		MemoryTest memo;

		MemoryFixture()
		:	sim( "MemoryTestSim" ),
			reporter(),
			observer( false ),
			p1( 0 ),
			p2( 0 ),
			p3( 0 ),
			memo( &sim, "MemoryTest", Memory::USER_DEFINED, &observer )
			{
				p1 = new MemoryTestProcess( &sim, "MemoryTestProcess1" );
				p2 = new MemoryTestProcess( &sim, "MemoryTestProcess2" );
				p3 = new MemoryTestProcess( &sim, "MemoryTestProcess3" );
			}

		~MemoryFixture()
		{
			if( p1 ) delete p1;
			if( p2 ) delete p2;
			if( p3 ) delete p3;
		}
	};

	/**
	 * @test odemx::Memory construction and destruction
	 *
	 * Expected effects:
	 * @li construction can be observed
	 * @li members are correctly initialized
	 * @li destruction can be observed
	 */
	TEST_FIXTURE( MemoryFixture, ConstructionDestruction )
	{
		MemoryTest* defMemory = new MemoryTest( "MemoryTestDefaultSimConstruction", Memory::USER_DEFINED );
		CHECK_EQUAL( getDefaultSimulation(), defMemory -> getStatisticManager() );
		delete defMemory;

		observer.clear();
		Label l = "MemoryTestConstructionDestruction";
		MemoryTest* memory = new MemoryTest( &sim, l, Memory::USER_DEFINED, &observer );
		CHECK( observer.onCreateMemory );
		CHECK_EQUAL( memory, observer.createdMemoryObject );
		CHECK_EQUAL( &sim, memory -> getStatisticManager() );
		CHECK_EQUAL( l, memory -> getLabel() );
		CHECK_EQUAL( Memory::USER_DEFINED, memory -> getMemoryType() );
		CHECK_EQUAL( &observer, memory -> getObservers().front() );

		delete memory;
		CHECK( observer.onDestroyMemory );
		CHECK_EQUAL( memory, observer.destroyedMemoryObject );
	}

	/**
	 * @test odemx::Memory::remember(Sched*)
	 *
	 * Expected function call effects:
	 * @li a successful call returns \c true
	 * @li the given object is added to the memory list
	 * @li the call can be observed
	 * @li calling the function twice with the same object sends a warning to
	 * \c errorStream()
	 * @li failed calls return \c false
	 */
	TEST_FIXTURE( MemoryFixture, Remember )
	{
		bool success = memo.remember( p1 );
		CHECK( success );
		CHECK( memo.waiting() == true );
		CHECK_EQUAL( (size_t)1, memo.countWaiting() );
		CHECK( find( memo.getList().begin(), memo.getList().end(), p1 ) != memo.getList().end() );
		CHECK( observer.observed( "onRemember", memo.getLabel(), p1 -> getLabel() ) );

		clearErrorStream();
		success = memo.remember( p1 );
		CHECK( ! success );
		CHECK_EQUAL( "ODEMx WARNING: Memory::remember(): object already stored in memoryList -- Object: MemoryTest\n", \
				errorStream().str() );
	}

	/**
	 * @test odemx::Memory::forget(Sched*)
	 *
	 * Expected function call effects:
	 * @li the object was removed from the list of waiting processes
	 * @li trying to remove an object that's not in the list causes an error
	 * message to be sent to \c errorStream()
	 * @li the call can be observed
	 */
	TEST_FIXTURE( MemoryFixture, Forget )
	{
		memo.remember( p1 );
		memo.remember( p2 );
		memo.remember( p3 );

		memo.forget( p2 );
		CHECK_EQUAL( (size_t)2, memo.countWaiting() );
		CHECK_EQUAL( p1, memo.getList().front() );
		CHECK_EQUAL( p3, memo.getList().back() );
		CHECK( observer.observed( "onForget", memo.getLabel(), p2 -> getLabel() ) );

		clearErrorStream();
		memo.forget( p2 );
		CHECK_EQUAL( "ODEMx ERROR: Memory::forget(): object not in memoryList -- Object: MemoryTest\n", \
				errorStream().str() );
	}

	/**
	 * @test odemx::eraseMemory()
	 *
	 * Expected function call effects:
	 * @li the list of remembered objects is cleared
	 */
	TEST_FIXTURE( MemoryFixture, EraseMemory )
	{
		memo.remember( p1 );
		memo.remember( p2 );
		memo.remember( p3 );

		memo.eraseMemory();
		CHECK( memo.waiting() == false );
		CHECK( memo.getList().empty() );
	}

	/**
	 * @test odemx::Memory::alert()
	 *
	 * Expected function call effects:
	 * @li attempting to alert non-process objects causes a warning to be sent
	 * to \c errorStream()
	 * @li the call can be observed
	 * @li CREATED or IDLE processes are alerted
	 * @li the memory list is empty
	 * @li attempting to alert a terminated process causes a warning to be sent
	 * to \c errorStream()
	 * @li attempting to alert a previously scheduled process causes a warning
	 * to be sent to \c errorStream()
	 */
	TEST_FIXTURE( MemoryFixture, Alert )
	{
		MemoryTestEvent e1( &sim, "MemoryTestAlertEvent" );
		memo.remember( &e1 );

		clearErrorStream();
		memo.alert();
		CHECK_EQUAL( string( "ODEMx WARNING: Memory::checkSchedForAlert(): remembered object is not a process -- Object: MemoryTest\n" )
							+"ODEMx WARNING: Memory::alert(): could not alert Sched object: MemoryTestAlertEvent -- Object: MemoryTest\n", \
				errorStream().str() );

		p1 -> sleep();
		CHECK_EQUAL( Process::IDLE, p1 -> getProcessState() );
		CHECK_EQUAL( Process::CREATED, p2 -> getProcessState() );
		memo.remember( p1 );
		memo.remember( p2 );
		memo.alert();
		CHECK( observer.observed( "onAlert", memo.getLabel() ) );
		CHECK( p1 -> isAlerted() );
		CHECK( p2 -> isAlerted() );
		CHECK( memo.getList().empty() );

		p1 -> cancel();
		CHECK_EQUAL( Process::TERMINATED, p1 -> getProcessState() );
		memo.remember( p1 );
		clearErrorStream();
		memo.alert();
		CHECK_EQUAL( string( "ODEMx ERROR: Memory::checkSchedForAlert(): process already terminated -- Object: MemoryTest\n" )
							+"ODEMx WARNING: Memory::alert(): could not alert Sched object: MemoryTestProcess1 -- Object: MemoryTest\n", \
				errorStream().str() );

		p2 -> activate();
		CHECK_EQUAL( Process::RUNNABLE, p2 -> getProcessState() );
		memo.remember( p2 );
		clearErrorStream();
		memo.alert();
		CHECK_EQUAL( string( "ODEMx WARNING: Memory::checkSchedForAlert(): process already scheduled -- Object: MemoryTest\n" )
							+"ODEMx WARNING: Memory::alert(): could not alert Sched object: MemoryTestProcess2 -- Object: MemoryTest\n", \
				errorStream().str() );
	}

	/**
	 * @test odemx::Memory::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( MemoryFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "QueueTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		memo.reset();

		memo.remember( p1 );
		memo.remember( p2 );
		memo.remember( p3 );

		simAdvanceEvent.scheduleIn( 5 );
		sim.step(); // event execution advances sim time
		memo.forget( p2 );

		reporter.addProducer( &memo );
		reporter.generateReport();
		CHECK( reporter.processedTables );
		CHECK_EQUAL( (size_t) 1, reporter.allTableContent.size() );

		// only check second table, the first one is the memo's
		CHECK_EQUAL( toString( memo.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( memo.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( toString( memo.getMin() ), reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( memo.getMax() ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( memo.getNow() ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( memo.getAVLength() ), reporter.allTableContent[0][5] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
