/**
 * @file TestPortHead.cpp
 * @date Jul 27, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class template PortHeadT
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	/// make PortTestElement objects streamable
	std::ostream& operator<< ( std::ostream& stream, const PortTestElement& e )
	{
		return stream << e.id;
	}

	/// compare PortTestElement objects
	bool operator== ( const PortTestElement& e1, const PortTestElement& e2 )
	{
		return e1.id == e2.id;
	}

	/**
	 * @struct PortHeadTFixture
	 * @brief Helper struct providing set-up/tear-down of PortHeadT tests
	 *
	 * @copydetails EventFixture
	 */
	struct PortHeadTFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		PortHeadTTestObserver observer;

		PortHeadTFixture()
		:	sim( "PortHeadTTestSim" ),
			reporter(),
			observer( false )
			{}
	};

	/**
	 * @test odemx::PortHeadT construction and destruction
	 *
	 * Expected effects:
	 * @li label is set correctly
	 * @li mode is set correctly
	 * @li capacity is set correctly
	 * @li the Memory type is set correctly
	 * @li observer is set correctly
	 * @li construction and destruction can be observed
	 * @li attempt to create a 0-size port causes an error message and
	 * leads to the creation of a port with default capacity
	 */
	TEST_FIXTURE( PortHeadTFixture, ConstructionDestruction )
	{
		observer.history.clear();

		Label label = "PortHeadTTestConstruction1";
		PortHeadT< PortTestElement >* port1 =
			new PortHeadT< PortTestElement >( label );
		CHECK_EQUAL( label + " (head)", port1 -> getLabel() );
		CHECK_EQUAL( WAITING_MODE, port1 -> getMode() );
		CHECK_EQUAL( (unsigned int) 10000, port1 -> getMaxCapacity() );
		CHECK_EQUAL( Memory::PORTHEAD, port1 -> getMemoryType() );

		label = "PortHeadTTestConstruction2";
		PortHeadT< PortTestElement >* port2 =
			new PortHeadT< PortTestElement >( label, ERROR_MODE, 1000, &observer );
		CHECK_EQUAL( label + " (head)", port2 -> getLabel() );
		CHECK_EQUAL( ERROR_MODE, port2 -> getMode() );
		CHECK_EQUAL( (unsigned int) 1000, port2 -> getMaxCapacity() );
		CHECK_EQUAL( &observer, port2 -> Observable< PortHeadTObserver< PortTestElement > >::getObservers().front() );
		CHECK( observer.observed( "onCreate", port2 -> getLabel() ) );

		label = "PortHeadTTestConstruction3";
		PortHeadT< PortTestElement >* port3 =
			new PortHeadT< PortTestElement >( &sim, label );
		CHECK_EQUAL( label + " (head)", port3 -> getLabel() );
		CHECK_EQUAL( WAITING_MODE, port3 -> getMode() );
		CHECK_EQUAL( (unsigned int) 10000, port3 -> getMaxCapacity() );
		CHECK_EQUAL( Memory::PORTHEAD, port3 -> getMemoryType() );

		label = "PortHeadTTestConstruction4";
		PortHeadT< PortTestElement >* port4 =
			new PortHeadT< PortTestElement >( &sim, label, ERROR_MODE, 1000, &observer );
		CHECK_EQUAL( label + " (head)", port4 -> getLabel() );
		CHECK_EQUAL( ERROR_MODE, port4 -> getMode() );
		CHECK_EQUAL( (unsigned int) 1000, port4 -> getMaxCapacity() );
		CHECK_EQUAL( &observer, port4 -> Observable< PortHeadTObserver< PortTestElement > >::getObservers().front() );
		CHECK( observer.observed( "onCreate", port4 -> getLabel() ) );

		clearErrorStream();
		PortHeadT< PortTestElement >* zeroPort1 =
			new PortHeadT< PortTestElement >( "PortHeadTTestConstruction5", ZERO_MODE, 0 );
		CHECK_EQUAL( "ODEMx ERROR: PortHeadT<>(): attempt to create Port with capacity 0, using default -- Object: PortHeadTTestConstruction5 (head)\n",
				errorStream().str() );
		CHECK_EQUAL( (unsigned int) 10000, zeroPort1 -> getMaxCapacity() );

		clearErrorStream();
		PortHeadT< PortTestElement >* zeroPort2 =
			new PortHeadT< PortTestElement >( &sim, "PortHeadTTestConstruction6", ZERO_MODE, 0 );
		CHECK_EQUAL( "ODEMx ERROR: PortHeadT<>(): attempt to create Port with capacity 0, using default -- Object: PortHeadTTestConstruction6 (head)\n",
				errorStream().str() );
		CHECK_EQUAL( (unsigned int) 10000, zeroPort2 -> getMaxCapacity() );

		label = port4 -> getLabel();
		delete port4;
		CHECK( observer.observed( "onDestroy", label ) );

		delete zeroPort2;
		delete zeroPort1;
		delete port3;
		delete port2;
		delete port1;
	}

	/**
	 * @test odemx::PortHeadT::get() with ZERO_MODE
	 *
	 * Expected function call effects:
	 * @li empty port returns a null pointer without warning or error
	 * @li when not empty, the port returns the first element
	 * and removes it from the buffer
	 */
	TEST_FIXTURE( PortHeadTFixture, ZeroModeGet )
	{
		PortHeadT< int* > head( &sim, "PortHeadTTestZeroModeGet", ZERO_MODE, 100 );
		CHECK( 0 == head.get().get() );

		// get old number of elements and put some into the port
		const std::list< int* >& elementList = head.getList();
		int* element1 = new int( 1 );
		int* element2 = new int( 2 );
		head.getTail() -> put( element1 );
		head.getTail() -> put( element2 );
		size_t elementCount = elementList.size();

		// get the first element
		std::auto_ptr< int* > elementPtr = head.get();
		CHECK( 0 != elementPtr.get() );
		CHECK_EQUAL( element1, *elementPtr );
		CHECK_EQUAL( elementCount - 1, elementList.size() );
		CHECK( find( elementList.begin(), elementList.end(), element2 ) != elementList.end() );

		delete element1;
		delete element2;
		// required to delete the whole port
		delete head.getTail();
	}

	/**
	 * @test odemx::PortHeadT::get() with ERROR_MODE
	 *
	 * Expected function call effects:
	 * @li empty port returns a null pointer and sends an error message
	 * to \c errorStream()
	 * @li when not empty, the port returns the first element
	 * and removes it from the buffer
	 */
	TEST_FIXTURE( PortHeadTFixture, ErrorModeGet )
	{
		clearErrorStream();
		PortHeadT< string > head( &sim, "PortHeadTTestErrorModeGet", ERROR_MODE, 100 );
		CHECK( 0 == head.get().get() );
		CHECK_EQUAL( "ODEMx ERROR: PortHeadT<>::get(): called on empty port -- Object: PortHeadTTestErrorModeGet (head)\n",
				errorStream().str() );

		// get old number of elements and put some into the port
		const std::list< string >& elementList = head.getList();
		string element1( "1" );
		string element2( "2" );
		head.getTail() -> put( element1 );
		head.getTail() -> put( element2 );
		size_t elementCount = elementList.size();

		// get the first element
		std::auto_ptr< string > elementPtr = head.get();
		CHECK( 0 != elementPtr.get() );
		CHECK_EQUAL( element1, *elementPtr );
		CHECK_EQUAL( elementCount - 1, elementList.size() );
		CHECK( find( elementList.begin(), elementList.end(), element2 ) != elementList.end() );

		// required to delete the whole port
		delete head.getTail();
	}

	/**
	 * @test odemx::PortHeadT::get() with WAITING_MODE
	 *
	 * Expected function call effects:
	 * @li calling on empty port from environment produces an error
	 * @li calling on empty port from event produces an error
	 * @li interrupt of a waiting Process causes return of a null pointer
	 * @li a waiting Process is alerted and gets the first element
	 * as soon as one becomes available
	 * @li the port returns the first element and removes it from the buffer
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortHeadTFixture, WaitModeGet )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestWaitModeGet", WAITING_MODE, 100, &observer );

		// call from environment
		clearErrorStream();
		CHECK( 0 == head.get().get() );
		CHECK_EQUAL( "ODEMx ERROR: PortHeadT::get(): port in waiting mode called from environment -- Object: PortHeadTTestWaitModeGet (head)\n",
				errorStream().str() );

		// call from Event
		PortHeadTTestGetEvent getterEvent( &sim, "PortHeadTTestWaitGetFailEvent", head );
		getterEvent.schedule();
		clearErrorStream();
		sim.step();
		CHECK( getterEvent.received == 0 );
		CHECK_EQUAL( "ODEMx ERROR: PortHeadT<>::get(): waiting mode not allowed for events -- Object: PortHeadTTestWaitModeGet (head)\n",
				errorStream().str() );

		// call from Process, which gets interrupted
		PortHeadTTestGetProcess getterProcess1( &sim, "PortHeadTTestWaitGetProcess", head );
		getterProcess1.activate();
		sim.step();
		CHECK_EQUAL( Process::IDLE, getterProcess1.getProcessState() );

		// interrupt the Process
		getterProcess1.interrupt();
		CHECK( getterProcess1.isScheduled() );
		sim.step();
		// 0 means that the auto_ptr was 0
		CHECK( 0 == getterProcess1.received.id );

		// successful call from Process
		PortHeadTTestGetProcess getterProcess2( &sim, "PortHeadTTestWaitGetProcess", head );
		getterProcess2.activate();
		sim.step();

		// get old number of elements and put some into the port
		const std::list< PortTestElement >& elementList = head.getList();
		// make an element available
		PortTestElement element1 = { 1 };
		head.getTail() -> put( element1 );
		PortTestElement element2 = { 2 };
		head.getTail() -> put( element2 );
		size_t elementCount = elementList.size();

		// let the Process resume and get the first element
		sim.step();
		CHECK_EQUAL( element1, getterProcess2.received );
		CHECK_EQUAL( elementCount - 1, elementList.size() );
		CHECK( find( elementList.begin(), elementList.end(), element2 ) != elementList.end() );
		CHECK( observer.observed( "onGet", head.getLabel() ) );

		// required to delete the whole port
		delete head.getTail();
	}

	/**
	 * @test odemx::PortHeadT::get() PortTailT alert call
	 *
	 * Expected function call effects:
	 * @li Processes waiting at the PortTail of a full Port are alerted as
	 * soon as space becomes available
	 * @li the new element of the waiting Process is successfully put into
	 * the buffer
	 */
	TEST_FIXTURE( PortHeadTFixture, GetAlertTail )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestGetAlertTail", WAITING_MODE, 1, &observer );
		PortTestElement element1 = { 1 };
		std::auto_ptr< PortTailT< PortTestElement > > tail( head.getTail() );
		tail -> put( element1 );

		unsigned int elementId = 9;
		PortTailTTestPutProcess putProcess( &sim, "PortHeadTTestGetAlertTailPutter", *tail, elementId );
		putProcess.activate();
		sim.step();
		CHECK_EQUAL( Process::IDLE, putProcess.getProcessState() );

		// taking one element from the port should alert the waiting putProcess
		head.get();
		CHECK( putProcess.isAlerted() );
		sim.step();
		// check if the new element was put into the port
		CHECK_EQUAL( elementId, head.getList().front().id );
	}

	/**
	 * @test odemx::PortHeadT::unGet(ElementType)
	 *
	 * Expected function call effects:
	 * @li successful call returns true
	 * @li the element is added to the front of the buffer
	 * @li the call can be observed
	 * @li if the port is full and in WAITING_MODE or ERROR_MODE, it sends
	 * an error message to \c errorStream() and returns false
	 * @li in ZERO_MODE unsuccessful calls only return false without error
	 */
	TEST_FIXTURE( PortHeadTFixture, UnGet )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestUnGet", WAITING_MODE, 2, &observer );
		PortTestElement element1 = { 1 };
		PortTestElement element2 = { 2 };
		PortTestElement element3 = { 3 };
		std::auto_ptr< PortTailT< PortTestElement > > tail( head.getTail() );
		tail -> put( element1 );

		bool returnValue = head.unGet( element2 );
		CHECK_EQUAL( true, returnValue );
		CHECK_EQUAL( element1, head.getList().back() );
		CHECK_EQUAL( element2, head.getList().front() );
		CHECK( observer.observed( "onUnGet", head.getLabel(), toString( element2 ) ) );

		clearErrorStream();
		returnValue = head.unGet( element3 );
		CHECK_EQUAL( false, returnValue );
		CHECK_EQUAL( "ODEMx ERROR: PortHeadT<>::unGet(): called on full port -- Object: PortHeadTTestUnGet (head)\n",
				errorStream().str() );

		head.setMode( ERROR_MODE );
		clearErrorStream();
		returnValue = head.unGet( element3 );
		CHECK_EQUAL( false, returnValue );
		CHECK_EQUAL( "ODEMx ERROR: PortHeadT<>::unGet(): called on full port -- Object: PortHeadTTestUnGet (head)\n",
				errorStream().str() );

		head.setMode( ZERO_MODE );
		clearErrorStream();
		returnValue = head.unGet( element3 );
		CHECK_EQUAL( false, returnValue );
		CHECK( errorStream().str().empty() );
	}

	/**
	 * @test odemx::PortHeadT::getTail()
	 *
	 * Expected function call effects:
	 * @li mode and maximum capacity values of head and tail are the same
	 */
	TEST_FIXTURE( PortHeadTFixture, GetTail )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTestGetTail", WAITING_MODE, 1 );
		std::auto_ptr< PortTailT< PortTestElement > > tail( head.getTail() );
		CHECK_EQUAL( tail -> getMaxCapacity(), head.getMaxCapacity() );
		CHECK_EQUAL( tail -> getMode(), head.getMode() );
	}

	/**
	 * @test odemx::PortHeadT::cut()
	 *
	 * Expected function call effects:
	 * @li there are now two Ports, with the upstream one still holding the data
	 * @li a new head to the old buffer is returned
	 * @li the original tail now has a new internal buffer
	 * @li mode and capacity of the new Port are the same
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortHeadTFixture, Cut )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestCut", WAITING_MODE, 2, &observer );
		PortTestElement element1 = { 1 };
		PortTestElement element2 = { 2 };
		std::auto_ptr< PortTailT< PortTestElement > > tail( head.getTail() );
		tail -> put( element1 );
		tail -> put( element2 );
		CHECK_EQUAL( (size_t) 2, head.count() );

		// an intermediary Process could do something with the data
		// from old buffer and put it into the new buffer, to which
		// the original head is then attached
		std::auto_ptr< PortHeadT< PortTestElement > > oldBufferHead( head.cut() );
		std::auto_ptr< PortTailT< PortTestElement > > newBufferTail( head.getTail() );
		CHECK_EQUAL( &head, newBufferTail -> getHead() );
		CHECK_EQUAL( tail.get(), oldBufferHead -> getTail() );
		CHECK_EQUAL( (size_t) 0, newBufferTail -> count() );
		CHECK_EQUAL( (size_t) 2, oldBufferHead -> count() );
		CHECK_EQUAL( oldBufferHead -> getMode(), newBufferTail -> getMode() );
		CHECK_EQUAL( oldBufferHead -> getMaxCapacity(), newBufferTail -> getMaxCapacity() );
		CHECK( observer.observed( "onCut", head.getLabel() ) );
	}

	/**
	 * @test odemx::PortHeadT::splice(PortTailT*)
	 *
	 * Expected function call effects:
	 * @li two ports with separate buffers are joined into one
	 * @li the upstream port's contents is appended to the downstream port's buffer
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortHeadTFixture, Splice )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestSplice", WAITING_MODE, 2, &observer );
		PortTestElement element1 = { 1 };
		PortTestElement element2 = { 2 };
		PortTestElement element3 = { 3 };
		std::auto_ptr< PortTailT< PortTestElement > > tail( head.getTail() );
		PortHeadT< PortTestElement >* oldBufferHead( head.cut() );
		PortTailT< PortTestElement >* newBufferTail( head.getTail() );
		tail -> put( element1 );
		tail -> put( element2 );
		CHECK_EQUAL( (size_t) 2, tail -> count() );
		newBufferTail -> put( element3 );
		CHECK_EQUAL( (size_t) 1, head.count() );

		// this will join the ports again and delete the two superflous objects
		Label label = oldBufferHead -> getLabel();
		oldBufferHead -> Observable< PortHeadTObserver< PortTestElement > >::addObserver( &observer );
		oldBufferHead -> splice( newBufferTail );
		CHECK_EQUAL( (size_t) 3, head.count() );
		CHECK_EQUAL( element3, head.getList().front() );
		CHECK_EQUAL( element2, head.getList().back() );
		CHECK( observer.observed( "onSplice", label ) );
	}

	/**
	 * @test odemx::PortHeadT::splice(PortTailT*) with head alert
	 *
	 * Expected function call effects:
	 * @li when the upstream port gets the appended elements and if it has
	 * waiting processes, alert() is called
	 */
	TEST_FIXTURE( PortHeadTFixture, SpliceAlert )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestSplice", WAITING_MODE, 2, &observer );
		PortHeadTTestGetProcess getter( &sim, "PortHeadTTestSpliceGetter", head );
		getter.activate();
		sim.step();
		CHECK_EQUAL( Process::IDLE, getter.getProcessState() );

		PortTestElement element1 = { 1 };
		std::auto_ptr< PortTailT< PortTestElement > > tail( head.getTail() );
		PortHeadT< PortTestElement >* oldBufferHead( head.cut() );
		PortTailT< PortTestElement >* newBufferTail( head.getTail() );
		tail -> put( element1 );
		CHECK_EQUAL( (size_t) 1, tail -> count() );
		CHECK_EQUAL( (size_t) 0, head.count() );

		// this will join the ports again and delete the two superflous objects
		oldBufferHead -> splice( newBufferTail );
		CHECK_EQUAL( (size_t) 1, head.count() );
		CHECK( getter.isAlerted() );
	}

	/**
	 * @test odemx::PortHeadT::cppSplice(PortTailT*) append contents
	 *
	 * Expected function call effects:
	 * @li the contents of the tail's port is appended to head's port
	 * @li the tail's buffer is empty
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortHeadTFixture, CppSpliceAppend )
	{
		PortHeadT< PortTestElement > port1Head( &sim, "PortHeadTTestCppSpliceAppend", WAITING_MODE, 2, &observer );
		std::auto_ptr< PortTailT< PortTestElement > > port1Tail( port1Head.getTail() );
		PortTestElement element1 = { 1 };
		port1Tail -> put( element1 );

		PortTailT< PortTestElement > port2Tail( &sim, "PortHeadTTestCppSpliceAppendTail", WAITING_MODE, 2 );
		PortTestElement element2 = { 2 };
		port2Tail.put( element2 );

		// default is appending of elements
		bool append = true;
		port1Head.cppSplice( &port2Tail, append );
		CHECK_EQUAL( element1, port1Head.getList().front() );
		CHECK_EQUAL( element2, port1Head.getList().back() );
		CHECK( port2Tail.getList().empty() );
		CHECK( observer.observed( "onCppSplice", port1Head.getLabel(), port2Tail.getLabel(), toString( append ) ) );
	}

	/**
	 * @test odemx::PortHeadT::cppSplice(PortTailT*) prepend contents
	 *
	 * Expected function call effects:
	 * @li the contents of the tail's port is prepended to head's port
	 * @li the tail's buffer is empty
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortHeadTFixture, CppSplicePrepend )
	{
		PortHeadT< PortTestElement > port1Head( &sim, "PortHeadTTestCppSplicePrepend", WAITING_MODE, 2, &observer );
		std::auto_ptr< PortTailT< PortTestElement > > port1Tail( port1Head.getTail() );
		PortTestElement element1 = { 1 };
		port1Tail -> put( element1 );

		PortTailT< PortTestElement > port2Tail( &sim, "PortHeadTTestCppSplicePrependTail", WAITING_MODE, 2 );
		PortTestElement element2 = { 2 };
		port2Tail.put( element2 );

		// false says: prepend instead
		bool append = false;
		port1Head.cppSplice( &port2Tail, append );
		CHECK_EQUAL( element2, port1Head.getList().front() );
		CHECK_EQUAL( element1, port1Head.getList().back() );
		CHECK( port2Tail.getList().empty() );
		CHECK( observer.observed( "onCppSplice", port1Head.getLabel(), port2Tail.getLabel(), toString( append ) ) );
	}

	/**
	 * @test odemx::PortHeadT::setMode(PortMode)
	 *
	 * Expected function call effects:
	 * @li the mode is set correctly
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortHeadTFixture, SetMode )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestSetMode", WAITING_MODE, 2, &observer );
		head.setMode( ERROR_MODE );
		CHECK_EQUAL( ERROR_MODE, head.getMode() );
		CHECK( observer.observed( "onChangeMode", head.getLabel(), "WAITING_MODE", "ERROR_MODE" ) );
	}

	/**
	 * @test odemx::PortHeadT::setMaxCapacity(unsigned int)
	 *
	 * Expected function call effects:
	 * @li the new maximum capacity is set correctly
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortHeadTFixture, SetMaxCapacity )
	{
		unsigned int oldMax = 2;
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestSetMaxCapacity", WAITING_MODE, oldMax, &observer );
		unsigned int newMax = 125;
		head.setMaxCapacity( newMax );
		CHECK_EQUAL( newMax, head.getMaxCapacity() );
		CHECK( observer.observed( "onChangeMaxCapacity", head.getLabel(), toString( oldMax ), toString( newMax ) ) );
	}

	/**
	 * @test odemx::PortHeadT::alert()
	 *
	 * Expected function call effects:
	 * @li the waiting process is alerted
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortHeadTFixture, Alert )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestAlert", WAITING_MODE, 1, &observer );
		std::auto_ptr< PortTailT< PortTestElement > > tail( head.getTail() );

		// the Process that gets to wait
		PortHeadTTestGetProcess getterProcess( &sim, "PortHeadTTestAlertProcess", head );
		getterProcess.activate();
		sim.step();

		// make an element available
		PortTestElement element1 = { 1 };
		tail -> put( element1 );
		CHECK( getterProcess.isAlerted() );
		CHECK( observer.observed( "onAlert", head.getLabel(), getterProcess.getLabel() ) );
	}

	/**
	 * @test odemx::PortHeadT::available()
	 *
	 * Expected function call effects:
	 * @li the function returns \c true when the buffer is not empty
	 */
	TEST_FIXTURE( PortHeadTFixture, Available )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestAvailable", WAITING_MODE, 1, &observer );
		std::auto_ptr< PortTailT< PortTestElement > > tail( head.getTail() );
		CHECK( ! head.available() );

		// make an element available
		PortTestElement element1 = { 1 };
		tail -> put( element1 );
		CHECK( head.available() );
	}

	/**
	 * @test odemx::PortHeadT::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( PortHeadTFixture, Report )
	{
		PortHeadT< PortTestElement > head( &sim, "PortHeadTTestReport", ZERO_MODE, 10, &observer );
		std::auto_ptr< PortTailT< PortTestElement > > tail( head.getTail() );

		Base::EventTest simAdvanceEvent( &sim, "PortHeadTTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		head.reset();

		PortTestElement element1 = { 1 };
		tail -> put( element1 );
		tail -> put( element1 );
		tail -> put( element1 );

		simAdvanceEvent.scheduleIn( 5 );
		sim.step(); // event execution advances sim time

		// remove one element to change the average
		head.get();

		simAdvanceEvent.scheduleIn( 5 );
		sim.step(); // event execution advances sim time

		reporter.addProducer( &head );
		reporter.generateReport();
		CHECK( reporter.processedTables );
		CHECK_EQUAL( (size_t) 2, reporter.allTableContent.size() );

		// only check first table, the second one is the Memory's
		Label portLabel = head.getLabel().substr( 0, (head.getLabel().length() - 7) );
		CHECK_EQUAL( portLabel, reporter.allTableContent[0][0] );
		CHECK_EQUAL( typeToString( typeid(PortTestElement) ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( toString( head.getMaxCapacity() ), reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( head.getResetTime() ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( 3 ), reporter.allTableContent[0][4] ); // put
		CHECK_EQUAL( toString( 1 ), reporter.allTableContent[0][5] ); // get
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][6] ); // unGet
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][7] ); // min
		CHECK_EQUAL( toString( 3 ), reporter.allTableContent[0][8] ); // max
		CHECK_EQUAL( toString( 2 ), reporter.allTableContent[0][9] ); // now
		CHECK_EQUAL( toString( 2.5 ), reporter.allTableContent[0][10] ); // avg
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][11] ); // head alert
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][12] ); // head alert
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond

