/**
 * @file TestPortTail.cpp
 * @date Jul 29, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class template PortTailT
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	/**
	 * @struct PortTailTFixture
	 * @brief Helper struct providing set-up/tear-down of PortTailT tests
	 *
	 * @copydetails EventFixture
	 */
	struct PortTailTFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		PortTailTTestObserver observer;

		PortTailTFixture()
		:	sim( "PortTailTTestSim" ),
			reporter(),
			observer( false )
			{}
	};

	/**
	 * @test odemx::PortTailT construction and destruction
	 *
	 * Expected effects:
	 * @li label is set correctly
	 * @li mode is set correctly
	 * @li capacity is set correctly
	 * @li the Memory type is set correctly
	 * @li observer is set correctly
	 * @li construction and destruction can be observed
	 * @li attempt to create a 0-size port causes an error message and
	 * leads to the creation of a port with default capacity
	 */
	TEST_FIXTURE( PortTailTFixture, ConstructionDestruction )
	{
		observer.history.clear();

		Label label = "PortTailTTestConstruction1";
		PortTailT< PortTestElement >* port1 =
			new PortTailT< PortTestElement >( label );
		CHECK_EQUAL( label + " (tail)", port1 -> getLabel() );
		CHECK_EQUAL( WAITING_MODE, port1 -> getMode() );
		CHECK_EQUAL( (unsigned int) 10000, port1 -> getMaxCapacity() );
		CHECK_EQUAL( Memory::PORTTAIL, port1 -> getMemoryType() );

		label = "PortTailTTestConstruction2";
		PortTailT< PortTestElement >* port2 =
			new PortTailT< PortTestElement >( label, ERROR_MODE, 1000, &observer );
		CHECK_EQUAL( label + " (tail)", port2 -> getLabel() );
		CHECK_EQUAL( ERROR_MODE, port2 -> getMode() );
		CHECK_EQUAL( (unsigned int) 1000, port2 -> getMaxCapacity() );
		CHECK_EQUAL( &observer, port2 -> Observable< PortTailTObserver< PortTestElement > >::getObservers().front() );
		CHECK( observer.observed( "onCreate", port2 -> getLabel() ) );

		label = "PortTailTTestConstruction3";
		PortTailT< PortTestElement >* port3 =
			new PortTailT< PortTestElement >( &sim, label );
		CHECK_EQUAL( label + " (tail)", port3 -> getLabel() );
		CHECK_EQUAL( WAITING_MODE, port3 -> getMode() );
		CHECK_EQUAL( (unsigned int) 10000, port3 -> getMaxCapacity() );
		CHECK_EQUAL( Memory::PORTTAIL, port3 -> getMemoryType() );

		label = "PortTailTTestConstruction4";
		PortTailT< PortTestElement >* port4 =
			new PortTailT< PortTestElement >( &sim, label, ERROR_MODE, 1000, &observer );
		CHECK_EQUAL( label + " (tail)", port4 -> getLabel() );
		CHECK_EQUAL( ERROR_MODE, port4 -> getMode() );
		CHECK_EQUAL( (unsigned int) 1000, port4 -> getMaxCapacity() );
		CHECK_EQUAL( &observer, port4 -> Observable< PortTailTObserver< PortTestElement > >::getObservers().front() );
		CHECK( observer.observed( "onCreate", port4 -> getLabel() ) );

		clearErrorStream();
		PortTailT< PortTestElement >* zeroPort1 =
			new PortTailT< PortTestElement >( "PortTailTTestConstruction5", ZERO_MODE, 0 );
		CHECK_EQUAL( "ODEMx ERROR: PortTailT<>(): attempt to create Port with capacity 0, using default -- Object: PortTailTTestConstruction5 (tail)\n",
				errorStream().str() );
		CHECK_EQUAL( (unsigned int) 10000, zeroPort1 -> getMaxCapacity() );

		clearErrorStream();
		PortTailT< PortTestElement >* zeroPort2 =
			new PortTailT< PortTestElement >( &sim, "PortTailTTestConstruction6", ZERO_MODE, 0 );
		CHECK_EQUAL( "ODEMx ERROR: PortTailT<>(): attempt to create Port with capacity 0, using default -- Object: PortTailTTestConstruction6 (tail)\n",
				errorStream().str() );
		CHECK_EQUAL( (unsigned int) 10000, zeroPort2 -> getMaxCapacity() );

		label = port4 -> getLabel();
		delete port4;
		CHECK( observer.observed( "onDestroy", label ) );

		delete zeroPort2;
		delete zeroPort1;
		delete port3;
		delete port2;
		delete port1;
	}

	/**
	 * @test odemx::PortTailT::put() with ZERO_MODE
	 *
	 * Expected function call effects:
	 * @li successful call returns 1
	 * @li the new element is put into the buffer
	 * @li when the buffer is full, the call returns 0
	 */
	TEST_FIXTURE( PortTailTFixture, ZeroModePut )
	{
		PortTailT< int* > tail( &sim, "PortTailTTestZeroModePut", ZERO_MODE, 1 );
		const std::list< int* >& elementList = tail.getList();
		int* element1 = new int( 1 );
		unsigned int oldCount = tail.count();

		int returnValue = tail.put( element1 );
		CHECK_EQUAL( 1, returnValue );
		CHECK( find( elementList.begin(), elementList.end(), element1 ) != elementList.end() );
		CHECK_EQUAL( oldCount + 1, tail.count() );

		CHECK_EQUAL( (unsigned int)0, tail.getAvailableSpace() );
		int* element2 = new int( 2 );
		returnValue = tail.put( element2 );
		CHECK_EQUAL( 0, returnValue );

		delete element1;
		delete element2;
	}

	/**
	 * @test odemx::PortTailT::put() with ERROR_MODE
	 *
	 * Expected function call effects:
	 * @li successful call returns 1
	 * @li the new element is put into the buffer
	 * @li when the buffer is full, the call returns -1 and sends an error
	 * message to \c errorStream()
	 */
	TEST_FIXTURE( PortTailTFixture, ErrorModePut )
	{
		PortTailT< int* > tail( &sim, "PortTailTTestErrorModePut", ERROR_MODE, 1 );
		const std::list< int* >& elementList = tail.getList();

		int* element1 = new int( 1 );
		unsigned int oldCount = tail.count();

		int returnValue = tail.put( element1 );
		CHECK_EQUAL( 1, returnValue );
		CHECK( find( elementList.begin(), elementList.end(), element1 ) != elementList.end() );
		CHECK_EQUAL( oldCount + 1, tail.count() );

		clearErrorStream();
		CHECK_EQUAL( (unsigned int)0, tail.getAvailableSpace() );
		int* element2 = new int( 2 );
		returnValue = tail.put( element2 );
		CHECK_EQUAL( -1, returnValue );
		CHECK_EQUAL( "ODEMx ERROR: PortTailT<>::put(): cannot add element to full port -- Object: PortTailTTestErrorModePut (tail)\n",
				errorStream().str() );

		delete element1;
		delete element2;
	}

	/**
	 * @test odemx::PortTailT::put() with WAITING_MODE
	 *
	 * Expected function call effects:
	 * @li the call can be observed
	 * @li calling on full port from environment produces an error
	 * @li calling on full port from event produces an error
	 * @li interrupt of a waiting Process causes return value 0
	 * @li a waiting Process is alerted and puts its element into the port as
	 * soon as space becomes available
	 * @li the element is appended to the buffer
	 */
	TEST_FIXTURE( PortTailTFixture, WaitModePut )
	{
		PortTailT< PortTestElement > tail( &sim, "PortTailTTestWaitModePut", WAITING_MODE, 1, &observer );

		// get old number of elements and put some into the port
		const std::list< PortTestElement >& elementList = tail.getList();
		// make an element available
		PortTestElement element1 = { 1 };
		tail.put( element1 );
		CHECK( observer.observed( "onPut", tail.getLabel(), toString( element1 ) ) );

		// call from environment
		clearErrorStream();
		CHECK_EQUAL( (unsigned int)0, tail.getAvailableSpace() );
		PortTestElement element2 = { 2 };
		int returnValue = tail.put( element2 );
		CHECK_EQUAL( -1, returnValue );
		CHECK_EQUAL( "ODEMx ERROR: PortTailT::put(): port in waiting mode called from environment -- Object: PortTailTTestWaitModePut (tail)\n",
				errorStream().str() );

		// call from Event
		unsigned int id = 42;
		PortTailTTestPutEvent putEvent( &sim, "PortTailTTestWaitPutFailEvent", tail, id );
		putEvent.schedule();
		clearErrorStream();
		sim.step();
		CHECK( putEvent.putReturnValue == -1 );
		CHECK_EQUAL( "ODEMx ERROR: PortHead::put(): waiting mode not allowed for events -- Object: PortTailTTestWaitModePut (tail)\n",
				errorStream().str() );

		// call from Process, which gets interrupted
		PortTailTTestPutProcess putProcess1( &sim, "PortTailTTestWaitPutProcess", tail, id );
		putProcess1.activate();
		sim.step();
		CHECK_EQUAL( Process::IDLE, putProcess1.getProcessState() );

		// interrupt the Process
		putProcess1.interrupt();
		CHECK( putProcess1.isScheduled() );
		sim.step();
		// 0 means that the put()-caller was interrupted
		CHECK( 0 == putProcess1.putReturnValue );

		// successful call from Process
		PortTailTTestPutProcess putProcess2( &sim, "PortTailTTestWaitPutProcess", tail, id );
		putProcess2.activate();
		sim.step();

		std::auto_ptr< PortHeadT< PortTestElement > > head( tail.getHead() );
		// make space available
		head -> get();
		CHECK_EQUAL( (unsigned int)1, tail.getAvailableSpace() );
		CHECK( putProcess2.isAlerted() );
		sim.step();
		// 1 means that the put()-call was successful
		CHECK( 1 == putProcess2.putReturnValue );
		CHECK_EQUAL( id, elementList.back().id );
	}

	/**
	 * @test odemx::PortTailT::put() PortHeadT alert call
	 *
	 * Expected function call effects:
	 * @li Processes waiting at the PortHead of an empty Port are alerted as
	 * soon as data becomes available
	 * @li the new element is immediately taken by the waiting Process
	 */
	TEST_FIXTURE( PortTailTFixture, PutAlertHead )
	{
		PortTailT< PortTestElement > tail( &sim, "PortTailTTestPutAlertHead", WAITING_MODE, 1, &observer );
		std::auto_ptr< PortHeadT< PortTestElement > > head( tail.getHead() );

		PortHeadTTestGetProcess getterProcess1( &sim, "PortTailTTestPutAlertHeadProcess", *head );
		getterProcess1.activate();
		sim.step();
		CHECK_EQUAL( Process::IDLE, getterProcess1.getProcessState() );
		CHECK( head -> waiting() );

		PortTestElement element1 = { 1 };
		tail.put( element1 );
		CHECK( getterProcess1.isAlerted() );
		sim.step();
		CHECK_EQUAL( element1, getterProcess1.received );
	}

	/**
	 * @test odemx::PortTailT::getHead()
	 *
	 * Expected function call effects:
	 * @li mode and maximum capacity values of head and tail are the same
	 */
	TEST_FIXTURE( PortTailTFixture, GetHead )
	{
		PortTailT< PortTestElement > tail( &sim, "PortTailTTestGetHead", WAITING_MODE, 1 );
		std::auto_ptr< PortHeadT< PortTestElement > > head( tail.getHead() );
		CHECK_EQUAL( tail.getMaxCapacity(), head -> getMaxCapacity() );
		CHECK_EQUAL( tail.getMode(), head -> getMode() );
	}

	/**
	 * @test odemx::PortTailT::cut()
	 *
	 * Expected function call effects:
	 * @li there are now two Ports, with the downstream one holding the data
	 * @li a new tail to the old buffer is returned
	 * @li the original tail now has a new internal buffer
	 * @li mode and capacity of the new Port are the same
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortTailTFixture, Cut )
	{
		PortTailT< PortTestElement > tail( &sim, "PortTailTTestCut", WAITING_MODE, 2, &observer );
		std::auto_ptr< PortHeadT< PortTestElement > > head( tail.getHead() );

		PortTestElement element1 = { 1 };
		PortTestElement element2 = { 2 };
		tail.put( element1 );
		tail.put( element2 );
		CHECK_EQUAL( (size_t) 2, head -> count() );

		// an intermediary Process could filter data by cutting in between ports
		std::auto_ptr< PortTailT< PortTestElement > > oldBufferTail( tail.cut() );
		std::auto_ptr< PortHeadT< PortTestElement > > newBufferHead( tail.getHead() );
		CHECK_EQUAL( &tail, newBufferHead -> getTail() );
		CHECK_EQUAL( head.get(), oldBufferTail -> getHead() );
		CHECK_EQUAL( (size_t) 0, newBufferHead -> count() );
		CHECK_EQUAL( (size_t) 2, oldBufferTail -> count() );
		CHECK_EQUAL( oldBufferTail -> getMode(), newBufferHead -> getMode() );
		CHECK_EQUAL( oldBufferTail -> getMaxCapacity(), newBufferHead -> getMaxCapacity() );
		CHECK( observer.observed( "onCut", tail.getLabel() ) );
	}

	/**
	 * @test odemx::PortTailT::splice(PortHeadT*)
	 *
	 * Expected function call effects:
	 * @li two ports with separate buffers are joined into one
	 * @li the upstream port's contents is appended to the downstream port's buffer
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortTailTFixture, Splice )
	{
		PortTailT< PortTestElement > tail( &sim, "PortTailTTestSplice", WAITING_MODE, 2, &observer );
		std::auto_ptr< PortHeadT< PortTestElement > > head( tail.getHead() );
		PortTestElement element1 = { 1 };
		PortTestElement element2 = { 2 };
		PortTestElement element3 = { 3 };
		PortTailT< PortTestElement >* oldBufferTail = tail.cut();
		PortHeadT< PortTestElement >* newBufferHead = tail.getHead();
		tail.put( element1 );
		tail.put( element2 );
		CHECK_EQUAL( (size_t) 2, newBufferHead -> count() );
		oldBufferTail -> put( element3 );
		CHECK_EQUAL( (size_t) 1, head -> count() );

		// this will join the ports again and delete the two superflous objects
		Label tailLabel = oldBufferTail -> getLabel();
		oldBufferTail -> Observable< PortTailTObserver< PortTestElement > >::addObserver( &observer );
		oldBufferTail -> splice( newBufferHead );
		CHECK_EQUAL( (size_t) 3, head -> count() );
		CHECK_EQUAL( element3, head -> getList().front() );
		CHECK_EQUAL( element2, head -> getList().back() );
		CHECK( observer.observed( "onSplice", tailLabel ) );
	}

	/**
	 * @test odemx::PortTailT::cppSplice(PortHeadT*) append contents
	 *
	 * Expected function call effects:
	 * @li the contents of the head's port is appended to tail's port
	 * @li the head's buffer is empty
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortTailTFixture, CppSpliceAppend )
	{
		PortTailT< PortTestElement > port1Tail( &sim, "PortTailTTestCppSpliceAppend", WAITING_MODE, 2, &observer );
		PortHeadT< PortTestElement > port2Head( &sim, "PortTailTTestCppSpliceAppendHead", WAITING_MODE, 2 );
		PortTestElement element1 = { 1 };
		port1Tail.put( element1 );
		std::auto_ptr< PortTailT< PortTestElement > > port2Tail( port2Head.getTail() );

		PortTestElement element2 = { 2 };
		port2Tail -> put( element2 );

		// default is appending of elements
		bool append = true;
		port1Tail.cppSplice( &port2Head, append );
		CHECK_EQUAL( element1, port1Tail.getList().front() );
		CHECK_EQUAL( element2, port1Tail.getList().back() );
		CHECK( port2Head.getList().empty() );
		CHECK( observer.observed( "onCppSplice", port1Tail.getLabel(), port2Head.getLabel(), toString( append ) ) );
	}

	/**
	 * @test odemx::PortTailT::cppSplice(PortHeadT*) prepend contents
	 *
	 * Expected function call effects:
	 * @li the contents of the head's port is prepended to tail's port
	 * @li the head's buffer is empty
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortTailTFixture, CppSplicePrepend )
	{
		PortTailT< PortTestElement > port1Tail( &sim, "PortTailTTestCppSpliceAppend", WAITING_MODE, 2, &observer );
		PortHeadT< PortTestElement > port2Head( &sim, "PortTailTTestCppSpliceAppendHead", WAITING_MODE, 2 );
		PortTestElement element1 = { 1 };
		port1Tail.put( element1 );
		std::auto_ptr< PortTailT< PortTestElement > > port2Tail( port2Head.getTail() );

		PortTestElement element2 = { 2 };
		port2Tail -> put( element2 );

		// false says: prepend instead
		bool append = false;
		port1Tail.cppSplice( &port2Head, append );
		CHECK_EQUAL( element2, port1Tail.getList().front() );
		CHECK_EQUAL( element1, port1Tail.getList().back() );
		CHECK( port2Head.getList().empty() );
		CHECK( observer.observed( "onCppSplice", port1Tail.getLabel(), port2Head.getLabel(), toString( append ) ) );
	}

	/**
	 * @test odemx::PortTailT::setMode(PortMode)
	 *
	 * Expected function call effects:
	 * @li the mode is set correctly
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortTailTFixture, SetMode )
	{
		PortTailT< PortTestElement > tail( &sim, "PortTailTTestSetMode", WAITING_MODE, 2, &observer );
		tail.setMode( ERROR_MODE );
		CHECK_EQUAL( ERROR_MODE, tail.getMode() );
		CHECK( observer.observed( "onChangeMode", tail.getLabel(), "WAITING_MODE", "ERROR_MODE" ) );
	}

	/**
	 * @test odemx::PortTailT::setMaxCapacity(unsigned int)
	 *
	 * Expected function call effects:
	 * @li the new maximum capacity is set correctly
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortTailTFixture, SetMaxCapacity )
	{
		unsigned int oldMax = 2;
		PortTailT< PortTestElement > tail( &sim, "PortTailTTestSetMaxCapacity", WAITING_MODE, oldMax, &observer );
		unsigned int newMax = 125;
		tail.setMaxCapacity( newMax );
		CHECK_EQUAL( newMax, tail.getMaxCapacity() );
		CHECK( observer.observed( "onChangeMaxCapacity", tail.getLabel(), toString( oldMax ), toString( newMax ) ) );
	}

	/**
	 * @test odemx::PortTailT::alert()
	 *
	 * Expected function call effects:
	 * @li the waiting process is alerted
	 * @li the call can be observed
	 */
	TEST_FIXTURE( PortTailTFixture, Alert )
	{
		PortTailT< PortTestElement > tail( &sim, "PortTailTTestAlert", WAITING_MODE, 1, &observer );
		std::auto_ptr< PortHeadT< PortTestElement > > head( tail.getHead() );

		// fill the Port
		PortTestElement element1 = { 1 };
		tail.put( element1 );

		// create Process that gets to wait
		unsigned int id = 17;
		PortTailTTestPutProcess putProcess( &sim, "PortTailTTestAlertPutProcess", tail, id );
		putProcess.activate();
		sim.step();

		// make space available
		head -> get();
		CHECK_EQUAL( (unsigned int)1, tail.getAvailableSpace() );
		CHECK( putProcess.isAlerted() );
		sim.step();
		// 1 means that the put()-call was successful
		CHECK( 1 == putProcess.putReturnValue );
		CHECK_EQUAL( id, tail.getList().back().id );
		CHECK( observer.observed( "onAlert", tail.getLabel(), putProcess.getLabel() ) );
	}

	/**
	 * @test odemx::PortTailT::available()
	 *
	 * Expected function call effects:
	 * @li the function returns \c true when the buffer is not full
	 */
	TEST_FIXTURE( PortTailTFixture, Available )
	{
		PortTailT< PortTestElement > tail( &sim, "PortTailTTestAvailable", WAITING_MODE, 1, &observer );
		CHECK( tail.available() );

		PortTestElement element1 = { 1 };
		tail.put( element1 );
		CHECK( ! tail.available() );
	}

	/**
	 * @test odemx::PortTailT::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( PortTailTFixture, Report )
	{
		PortTailT< PortTestElement > tail( &sim, "PortTailTTestReport", ZERO_MODE, 10, &observer );

		Base::EventTest simAdvanceEvent( &sim, "PortTailTTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		tail.reset();

		PortTestElement element1 = { 1 };
		tail.put( element1 );
		tail.put( element1 );
		tail.put( element1 );

		simAdvanceEvent.scheduleIn( 5 );
		sim.step(); // event execution advances sim time

		// remove one element to change the average
		std::auto_ptr< PortHeadT< PortTestElement > > head( tail.getHead() );
		head -> get();

		simAdvanceEvent.scheduleIn( 5 );
		sim.step(); // event execution advances sim time

		reporter.addProducer( &tail );
		reporter.generateReport();
		CHECK( reporter.processedTables );
		CHECK_EQUAL( (size_t) 2, reporter.allTableContent.size() );

		// only check first table, the second one is the Memory's
		Label portLabel = tail.getLabel().substr( 0, (tail.getLabel().length() - 7) );
		CHECK_EQUAL( portLabel, reporter.allTableContent[0][0] );
		CHECK_EQUAL( typeToString( typeid(PortTestElement) ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( toString( tail.getMaxCapacity() ), reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( tail.getResetTime() ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( 3 ), reporter.allTableContent[0][4] ); // put
		CHECK_EQUAL( toString( 1 ), reporter.allTableContent[0][5] ); // get
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][6] ); // unGet
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][7] ); // min
		CHECK_EQUAL( toString( 3 ), reporter.allTableContent[0][8] ); // max
		CHECK_EQUAL( toString( 2 ), reporter.allTableContent[0][9] ); // now
		CHECK_EQUAL( toString( 2.5 ), reporter.allTableContent[0][10] ); // avg
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][11] ); // head alert
		CHECK_EQUAL( toString( 0 ), reporter.allTableContent[0][12] ); // tail alert
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
