/**
 * @file TestQueue.cpp
 * @date Jul 30, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Queue
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	/**
	 * @struct QueueFixture
	 * @brief Helper struct providing set-up/tear-down of Queue tests
	 *
	 * @copydetails EventFixture
	 */
	struct QueueFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		odemx::Queue queue;
		Base::ProcessTest process1;
		Base::ProcessTest process2;
		Base::ProcessTest process3;

		QueueFixture()
		:	sim( "QueueTestSim" ),
			reporter(),
			queue( &sim, "QueueTest" ),
			process1( &sim, "QueueTestProcess1" ),
			process2( &sim, "QueueTestProcess2" ),
			process3( &sim, "QueueTestProcess3" )
			{}
	};

	/**
	 * @test odemx::Queue construction
	 *
	 * Expected effects:
	 * @li the simulation (statistics manager) is set correctly
	 * @li the label is set correctly
	 * @li the statistics are initialized
	 * @li the length of the queue is \c 0
	 */
	TEST_FIXTURE( QueueFixture, ConstructionDestruction )
	{
		Label label = "QueueTestConstructionDefSim";
		Queue queue1( label );
		CHECK_EQUAL( getDefaultSimulation(), queue1.getStatisticManager() );
		CHECK_EQUAL( label, queue1.getLabel() );
		CHECK_EQUAL( (unsigned int)0, queue1.getMin() );
		CHECK_EQUAL( (unsigned int)0, queue1.getMax() );
		CHECK_EQUAL( (unsigned int)0, queue1.getNow() );
		CHECK_EQUAL( (unsigned int)0, queue1.getLength() );

		label = "QueueTestConstructionUserSim";
		Queue queue2( &sim, label );
		CHECK_EQUAL( &sim, queue2.getStatisticManager() );
		CHECK_EQUAL( label, queue2.getLabel() );
		CHECK_EQUAL( (unsigned int)0, queue2.getMin() );
		CHECK_EQUAL( (unsigned int)0, queue2.getMax() );
		CHECK_EQUAL( (unsigned int)0, queue2.getNow() );
		CHECK_EQUAL( (unsigned int)0, queue2.getLength() );
	}

	/**
	 * @test odemx::Queue::inSort(Process*,bool)
	 *
	 * Expected function call effects:
	 * @li all process pointers are added to the queue
	 * @li in FIFO mode, the process is appended to the list
	 * @li in LIFO mode, the process is prepended to the list
	 * @li the processes are in the expected order
	 */
	TEST_FIXTURE( QueueFixture, InSort )
	{
		queue.inSort( &process2 );
		CHECK_EQUAL( (unsigned int) 1, queue.getLength() );

		bool fifo = true; // is the default
		queue.inSort( &process3, fifo );
		CHECK_EQUAL( (unsigned int) 2, queue.getLength() );

		fifo = false;
		queue.inSort( &process1, fifo );
		CHECK_EQUAL( (unsigned int) 3, queue.getLength() );

		CHECK_EQUAL( (unsigned int) 0, queue.getMin() );
		CHECK_EQUAL( (unsigned int) 3, queue.getMax() );

		CHECK_EQUAL( &process1, queue.getTop() ); queue.popQueue();
		CHECK_EQUAL( &process2, queue.getTop() ); queue.popQueue();
		CHECK_EQUAL( &process3, queue.getTop() ); queue.popQueue();
		CHECK( queue.isEmpty() );
	}

	/**
	 * @test odemx::Queue::remove(Process*)
	 *
	 * Expected function call effects:
	 * @li the specified Process is removed from the queue
	 * @li the statistics are updated
	 * @li the order of the other processes is not affected
	 */
	TEST_FIXTURE( QueueFixture, Remove )
	{
		queue.inSort( &process1 );
		queue.inSort( &process2 );
		queue.inSort( &process3 );
		queue.reset();

		CHECK_EQUAL( (unsigned int) 3, queue.getLength() );
		CHECK_EQUAL( (unsigned int) 3, queue.getMin() );
		queue.remove( &process2 );
		// 0 means the process is not in the queue
		CHECK_EQUAL( (unsigned int) 0, queue.getPosition( &process2 ) );
		CHECK_EQUAL( (unsigned int) 2, queue.getMin() );
		CHECK_EQUAL( (unsigned int) 2, queue.getLength() );
		CHECK_EQUAL( (unsigned int) 3, queue.getMax() );

		CHECK_EQUAL( &process1, queue.getTop() ); queue.popQueue();
		CHECK_EQUAL( &process3, queue.getTop() ); queue.popQueue();
		CHECK( queue.isEmpty() );
	}

	/**
	 * @test odemx::Queue::popQueue()
	 *
	 * Expected function call effects:
	 * @li the first Process is removed from the queue
	 * @li the statistics are updated
	 * @li the order of the other processes is not affected
	 */
	TEST_FIXTURE( QueueFixture, PopQueue )
	{
		queue.inSort( &process1 );
		queue.inSort( &process2 );
		queue.inSort( &process3 );
		queue.reset();

		Process* removed = queue.getTop();
		CHECK_EQUAL( (unsigned int) 3, queue.getLength() );
		CHECK_EQUAL( (unsigned int) 3, queue.getMin() );
		queue.popQueue();
		// 0 means the process is not in the queue
		CHECK_EQUAL( (unsigned int) 0, queue.getPosition( removed ) );
		CHECK_EQUAL( (unsigned int) 2, queue.getMin() );
		CHECK_EQUAL( (unsigned int) 2, queue.getLength() );
		CHECK_EQUAL( (unsigned int) 3, queue.getMax() );

		CHECK_EQUAL( &process2, queue.getTop() ); queue.popQueue();
		CHECK_EQUAL( &process3, queue.getTop() ); queue.popQueue();
		CHECK( queue.isEmpty() );
	}

	/**
	 * @test odemx::Queue::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( QueueFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "QueueTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		queue.reset();

		queue.inSort( &process1 );
		queue.inSort( &process2 );
		queue.inSort( &process3 );
		queue.remove( &process2 );

		simAdvanceEvent.scheduleIn( 5 );
		sim.step(); // event execution advances sim time

		reporter.addProducer( &queue );
		reporter.generateReport();
		CHECK( reporter.processedTables );
		CHECK_EQUAL( (size_t) 1, reporter.allTableContent.size() );

		// only check second table, the first one is the queue's
		CHECK_EQUAL( toString( queue.getLabel() ), reporter.allTableContent[0][0] );
		CHECK_EQUAL( toString( queue.getResetTime() ), reporter.allTableContent[0][1] );
		CHECK_EQUAL( toString( queue.getMin() ), reporter.allTableContent[0][2] );
		CHECK_EQUAL( toString( queue.getMax() ), reporter.allTableContent[0][3] );
		CHECK_EQUAL( toString( queue.getNow() ), reporter.allTableContent[0][4] );
		CHECK_EQUAL( toString( queue.getAVLength() ), reporter.allTableContent[0][5] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
