/**
 * @file TestRes.cpp
 * @date Jul 30, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Res
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	/**
	 * @struct ResFixture
	 * @brief Helper struct providing set-up/tear-down of Res tests
	 *
	 * @copydetails EventFixture
	 */
	struct ResFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		ResTestObserver observer;
		odemx::Res res;

		ResFixture()
		:	sim( "ResTestSim" ),
			reporter(),
			observer( false ),
			res( &sim, "ResTest", 5, 20, &observer )
			{}
	};

	/**
	 * @test odemx::Res construction and destruction
	 *
	 * Expected effects:
	 * @li simulation environment is set for DefaultSimulation or user-definied
	 * simulation
	 * @li label is set
	 * @li initial tokens are set
	 * @li maximum tokens are set
	 * @li observer is set correctly
	 * @li construction and destruction can be observed
	 */
	TEST_FIXTURE( ResFixture, ConstructionDestruction )
	{
		observer.history.clear();
		unsigned int startTokens = 2;
		unsigned int maxTokens = 4;
		Label label = "ResTestConstructionWithDefaultSim";
		Res* testRes = new Res( label, startTokens, maxTokens, &observer );
		CHECK_EQUAL( getDefaultSimulation(), testRes -> getTrace() );
		CHECK_EQUAL( label, testRes -> getLabel() );
		CHECK_EQUAL( startTokens, testRes -> getTokenNumber() );
		CHECK_EQUAL( startTokens, testRes -> getInitial() );
		CHECK_EQUAL( maxTokens, testRes -> getTokenLimit() );
		CHECK_EQUAL( &observer, testRes -> getObservers().front() );
		CHECK( observer.observed( "onCreate", testRes -> getLabel() ) );

		delete testRes;
		CHECK( observer.observed( "onDestroy", label ) );

		startTokens = 42;
		maxTokens = 100;
		label = "ResTestConstructionWithUserSim";
		Res* testRes2 = new Res( &sim, label, startTokens, maxTokens, &observer );
		CHECK_EQUAL( &sim, testRes2 -> getTrace() );
		CHECK_EQUAL( label, testRes2 -> getLabel() );
		CHECK_EQUAL( startTokens, testRes2 -> getTokenNumber() );
		CHECK_EQUAL( startTokens, testRes2 -> getInitial() );
		CHECK_EQUAL( maxTokens, testRes2 -> getTokenLimit() );
		CHECK_EQUAL( &observer, testRes2 -> getObservers().front() );
		CHECK( observer.observed( "onCreate", testRes2 -> getLabel() ) );
	}

	/**
	 * @test odemx::Res::acquire(unsigned int) success
	 *
	 * Expected function call effects:
	 * @li non-Process objects cause an error message
	 * @li the token number is decreased by the amount taken
	 * @li token number change and take success can be observed
	 * @li on success, the function returns the amount taken
	 * @li usage is registered in statistics
	 */
	TEST_FIXTURE( ResFixture, AcquireSucceed )
	{
		unsigned int oldTokenNumber = res.getTokenNumber();
		unsigned int takeCount = oldTokenNumber / 2 + 1;

		clearErrorStream();
		res.acquire( takeCount );
		CHECK_EQUAL( "ODEMx ERROR: Res::acquire(): called by non-Process object -- Object: ResTest\n", \
				errorStream().str() );

		ResTaker taker( &sim, "ResTestTakeProcess", res, takeCount );
		taker.activate();
		sim.step();
		CHECK_EQUAL( oldTokenNumber - takeCount, res.getTokenNumber() );
		CHECK( observer.observed( "onChangeTokenNumber", res.getLabel(), \
				toString( oldTokenNumber ), toString( res.getTokenNumber() ) ) );
		CHECK( observer.observed( "onAcquireSucceed", res.getLabel(), toString( takeCount ) ) );
		CHECK_EQUAL( takeCount, taker.received );
		CHECK_EQUAL( (unsigned int)1, res.getAcquires() );
	}

	/**
	 * @test odemx::Res::acquire(unsigned int) failure
	 *
	 * Expected function call effects:
	 * @li a failing call can be observed
	 * @li the Process is inserted into a queue
	 * @li the Process is suspended, i.e. in state IDLE
	 * @li on Process interrupt, the function returns 0 and the Process is
	 * removed from the waiting queue
	 * @li waiting processes are awakened properly upon token increase
	 */
	TEST_FIXTURE( ResFixture, AcquireFail )
	{
		unsigned int oldTokenNumber = res.getTokenNumber();
		unsigned int takeCount = oldTokenNumber + 1;

		ResTaker taker( &sim, "ResTestTakeProcess", res, takeCount );
		taker.activate();
		sim.step();

		CHECK( observer.observed( "onAcquireFail", res.getLabel(), toString( takeCount ) ) );
		CHECK_EQUAL( (size_t)1, res.getWaitingForAcquire().size() );
		CHECK_EQUAL( &taker, res.getWaitingForAcquire().front() );
		CHECK_EQUAL( Process::IDLE, taker.getProcessState() );

		Base::InterruptEvent interrupt( &sim, &taker );
		interrupt.schedule();
		sim.step();
		CHECK_EQUAL( Process::RUNNABLE, taker.getProcessState() );
		sim.step();
		CHECK_EQUAL( (unsigned int)0, taker.received );
		CHECK_EQUAL( (size_t)0, res.getWaitingForAcquire().size() );

		ResTaker taker2( &sim, "ResTestTakeProcess", res, takeCount );
		taker2.hold();
		ResTaker taker3( &sim, "ResTestTakeProcess", res, takeCount );
		taker3.hold();
		sim.step();
		sim.step();
		CHECK_EQUAL( (size_t)2, res.getWaitingForAcquire().size() );

		res.release( takeCount * 2 );
		sim.run();
		CHECK_EQUAL( takeCount, taker2.received );
		CHECK_EQUAL( takeCount, taker3.received );
		CHECK_EQUAL( (size_t)0, res.getWaitingForAcquire().size() );
	}

	/**
	 * @test odemx::Res::acquire(unsigned int) block warning
	 *
	 * Expected function call effects:
	 * @li non-Process objects cause an error message
	 * @li the token number is decreased by the amount taken
	 * @li token number change and take success can be observed
	 * @li on success, the function returns the amount taken
	 * @li usage is registered in statistics
	 */
	TEST_FIXTURE( ResFixture, AcquireBlockWarning )
	{
		unsigned int takeCount = res.getTokenLimit() + 1;

		clearErrorStream();
		ResTaker taker( &sim, "ResTestTakeProcess", res, takeCount );
		taker.activate();
		sim.step();
		CHECK_EQUAL( "ODEMx WARNING: Res::acquire(): requesting more than max tokens will always block -- Object: ResTest\n", \
				errorStream().str() );
	}

	/**
	 * @test odemx::Res::release(unsigned int)
	 *
	 * Expected function call effects:
	 * @li token number change and release call can be observed
	 * @li waiting processes are awakened properly upon token increase
	 */
	TEST_FIXTURE( ResFixture, Release )
	{
		unsigned int oldTokenNumber = res.getTokenNumber();
		unsigned int releaseCount = res.getTokenLimit() + 1;
		unsigned int actuallyReleased = res.getTokenLimit() - res.getTokenNumber();

		ResTaker taker( &sim, "ResTestTakeProcess", res, oldTokenNumber + 1 );
		taker.hold();
		sim.step();
		CHECK_EQUAL( (size_t)1, res.getWaitingForAcquire().size() );

		clearErrorStream();
		res.release( releaseCount );
		CHECK_EQUAL( "ODEMx ERROR: Res::release(): limit reached, could not release all tokens -- Object: ResTest\n",
				errorStream().str() );
		CHECK( observer.observed( "onReleaseFail", res.getLabel(), toString( releaseCount ) ) );
		CHECK( observer.observed( "onChangeTokenNumber", res.getLabel(), \
				toString( oldTokenNumber ), toString( res.getTokenNumber() ) ) );
		CHECK( observer.observed( "onReleaseSucceed", res.getLabel(), toString( actuallyReleased ) ) );

		sim.run();
		CHECK_EQUAL( oldTokenNumber + 1, taker.received );
	}

	/**
	 * @test odemx::Res::control(unsigned int)
	 *
	 * Expected function call effects:
	 * @li the token limit is increased by the given amount
	 * @li the number of given tokens is successfully released to the resource
	 */
	TEST_FIXTURE( ResFixture, Control )
	{
		unsigned int increase = 23;
		unsigned int oldLimit = res.getTokenLimit();

		res.control( increase );
		CHECK_EQUAL( oldLimit + increase, res.getTokenLimit() );
		CHECK( observer.observed( "onReleaseSucceed", res.getLabel(), toString( increase ) ) );
	}

	/**
	 * @test odemx::Res::unControl(unsigned int)
	 *
	 * Expected function call effects:
	 * @li the token limit is decreased by the given amount
	 * @li the number of given tokens is successfully acquired from the resource
	 */
	TEST_FIXTURE( ResFixture, UnControl )
	{
		int decrease = res.getTokenNumber();
		unsigned int oldLimit = res.getTokenLimit();

		// make sure this Process must wait
		ResTestUnControlProcess interruptedUnController( &sim, "ResTestUnControlProcess", res, res.getTokenNumber() + 1 );
		interruptedUnController.activate();
		sim.step();

		Base::InterruptEvent interrupt( &sim, &interruptedUnController );
		interrupt.schedule();
		sim.step(); // event activation
		sim.step(); // process activation
		CHECK_EQUAL( 0, interruptedUnController.unControlReturnValue );

		ResTestUnControlProcess unController( &sim, "ResTestUnControlProcess", res, decrease );
		unController.activate();
		sim.step();

		CHECK_EQUAL( decrease, unController.unControlReturnValue );
		CHECK_EQUAL( oldLimit - decrease, res.getTokenLimit() );
		CHECK( observer.observed( "onAcquireSucceed", res.getLabel(), toString( decrease ) ) );
	}

	/**
	 * @test odemx::Res::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( ResFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "ResTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		res.reset();

		ResTaker taker( &sim, "ResTestReportTakerProcess", res, 15 );
		taker.activate();
		sim.step();
		res.release( 25 );

		reporter.addProducer( &res );
		reporter.generateReport();
		CHECK( reporter.processedTables );
		CHECK_EQUAL( (size_t) 2, reporter.allTableContent.size() );

		// only check second table, the first one is the queue's
		CHECK_EQUAL( toString( res.getLabel() ), reporter.allTableContent[1][0] );
		CHECK_EQUAL( toString( res.getResetTime() ), reporter.allTableContent[1][1] );
		CHECK_EQUAL( res.getLabel() + " queue", reporter.allTableContent[1][2] );
		CHECK_EQUAL( toString( res.getAcquires() ), reporter.allTableContent[1][3] );
		CHECK_EQUAL( toString( res.getReleases() ), reporter.allTableContent[1][4] );
		CHECK_EQUAL( toString( res.getInitial() ), reporter.allTableContent[1][5] );
		CHECK_EQUAL( toString( res.getTokenLimit() ), reporter.allTableContent[1][6] );
		CHECK_EQUAL( toString( res.getMin() ), reporter.allTableContent[1][7] );
		CHECK_EQUAL( toString( res.getMax() ), reporter.allTableContent[1][8] );
		CHECK_EQUAL( toString( res.getTokenNumber() ), reporter.allTableContent[1][9] );
		CHECK_EQUAL( toString( res.getAVFreeToken() ), reporter.allTableContent[1][10] );
		CHECK_EQUAL( toString( res.getAVWaitTime() ), reporter.allTableContent[1][11] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
