/**
 * @file TestResTChoice.cpp
 * @date Aug 1, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class template ResTChoice
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	// select tokens by their id value
	bool tokenSelect( Process* resTaker, TokenTest* t )
	{
		return (t -> id) <= ( static_cast< ResTChoiceTaker* >(resTaker) ->  tokenSelectId );
	}

	/**
	 * @struct ResTChoiceFixture
	 * @brief Helper struct providing set-up/tear-down of ResTChoice tests
	 *
	 * @copydetails EventFixture
	 */
	struct ResTChoiceFixture
	{
		Base::SimulationTest sim;
		ResTChoiceTestObserver observer;
		std::auto_ptr< odemx::ResTChoice< TokenTest > > res;

		ResTChoiceFixture()
		:	sim( "ResTChoiceTestSim" ),
			observer( false ),
			res( 0 )
			{
				std::vector< TokenTest > tokens;
				for( int i = 0; i != 25; ++i )
				{
					tokens.push_back( TokenTest( i + 1, string("ResTChoiceToken") + toString( i + 1 ) ) );
				}

				// set the auto_ptr and add the initial tokens to the res
				res.reset( new ResTChoice< TokenTest >( &sim, "ResTChoiceTest", tokens, &observer ) );
			}
	};

	/**
	 * @test odemx::ResTChoice construction and destruction
	 *
	 * Expected effects:
	 * @li observer is set correctly
	 * @li label is set
	 * @li simulation environment is set for DefaultSimulation or user-definied
	 * simulation
	 * @li initial tokens are set and transferred to the token store
	 * @li construction and destruction can be observed
	 */
	TEST_FIXTURE( ResTChoiceFixture, ConstructionDestruction )
	{
		observer.history.clear();
		std::vector< TokenTest > initialTokens;
		initialTokens.push_back( TokenTest( 1, "InitToken1" ) );
		initialTokens.push_back( TokenTest( 2, "InitToken2" ) );
		initialTokens.push_back( TokenTest( 3, "InitToken3" ) );

		ResTChoice< TokenTest >* res1 =
			new ResTChoice< TokenTest >( "ResTChoiceConstructionTest1", initialTokens, &observer );
		CHECK_EQUAL( &observer, res1 -> Observable< ResTChoiceObserver< TokenTest > >::getObservers().front() );
		CHECK_EQUAL( &observer, res1 -> Observable< ResTBaseObserver< TokenTest > >::getObservers().front() );
		CHECK_EQUAL( "ResTChoiceConstructionTest1", res1 -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), res1 -> getStatisticManager() );
		CHECK_EQUAL( initialTokens.size(), res1 -> getInitial() );
		CHECK_EQUAL( initialTokens.size(), res1 -> getTokenNumber() );
		CHECK_EQUAL( initialTokens.size(), res1 -> getTokenLimit() );
		CHECK_EQUAL( initialTokens.front(), res1 -> getTokenStore().front() );
		CHECK_EQUAL( initialTokens.back(), res1 -> getTokenStore().back() );
		CHECK( observer.observed( "onCreate", res1 -> getLabel() ) );

		std::vector< TokenTest >::iterator i;
		const std::list< TokenTest >& store = res1 -> getTokenStore();
		for( i = initialTokens.begin(); i != initialTokens.end(); ++i )
		{
			CHECK( find( store.begin(), store.end(), *i ) != store.end() );
		}

		Label label = "ResTChoiceConstructionTest2";
		ResTChoice< TokenTest >* res2 =
			new ResTChoice< TokenTest >( &sim, label, initialTokens, &observer );
		CHECK_EQUAL( &observer, res2 -> Observable< ResTChoiceObserver< TokenTest > >::getObservers().front() );
		CHECK_EQUAL( &observer, res2 -> Observable< ResTBaseObserver< TokenTest > >::getObservers().front() );
		CHECK_EQUAL( label, res2 -> getLabel() );
		CHECK_EQUAL( &sim, res2 -> getStatisticManager() );
		CHECK_EQUAL( initialTokens.size(), res2 -> getInitial() );
		CHECK_EQUAL( initialTokens.size(), res2 -> getTokenNumber() );
		CHECK_EQUAL( initialTokens.size(), res2 -> getTokenLimit() );
		CHECK_EQUAL( initialTokens.front(), res2 -> getTokenStore().front() );
		CHECK_EQUAL( initialTokens.back(), res2 -> getTokenStore().back() );
		CHECK( observer.observed( "onCreate", res2 -> getLabel() ) );

		const std::list< TokenTest >& store2 = res2 -> getTokenStore();
		for( i = initialTokens.begin(); i != initialTokens.end(); ++i )
		{
			CHECK( find( store2.begin(), store2.end(), *i ) != store2.end() );
		}

		delete res2;
		CHECK( observer.observed( "onDestroy", label ) );
		delete res1;
	}

	/**
	 * @test odemx::ResTChoice::acquire(ResTChoice::Selection)
	 *
	 * Expected function call effects:
	 * @li a token that satisfies the selection function is returned
	 * @li the acquiring process is set as owner of the token
	 * @li if the acquiring process is interrupted, the function returns 0
	 */
	TEST_FIXTURE( ResTChoiceFixture, AcquireSingleToken )
	{
		// create ResT with one token
		std::vector< TokenTest > init;
		init.push_back( TokenTest( 1, "ResTChoiceTestSingleToken" ) );
		ResTChoice< TokenTest > singleRes( &sim, "ResTChoiceTestSingle", init, &observer );

		ResTChoiceTaker acquirer( &sim, "ResTChoiceTestSingleTaker1", singleRes, 1, &Synchronization::tokenSelect, 1 );
		// ensure the token satisfies the selection function

		acquirer.activate();
		sim.step(); // enqueue
		CHECK_EQUAL( init.front(), *(acquirer.token) );
		CHECK( tokenSelect( &acquirer, acquirer.token ) );
		CHECK_EQUAL( &acquirer, singleRes.getTokenInfo().find( acquirer.token )->second );
		CHECK_EQUAL( 1, acquirer.getReturnValue() );

		// check 0-return for interrupted Process
		CHECK_EQUAL( (unsigned int) 0, singleRes.getTokenNumber() );
		ResTChoiceTaker acquirer2( &sim, "ResTChoiceTestSingleTaker2", singleRes, 1, &Synchronization::tokenSelect, 1 );
		acquirer2.activate();
		sim.step();

		Base::InterruptEvent interrupt( &sim, &acquirer2 );
		interrupt.schedule();
		sim.step(); // interrupt
		sim.step(); // acquirer2
		CHECK( acquirer2.hasReturned() );
		CHECK_EQUAL( static_cast< TokenTest* >( 0 ), acquirer2.token );
	}

	/**
	 * @test odemx::ResTChoice::acquire(ResTChoice::Selection,unsigned int) success
	 *
	 * Expected function call effects:
	 * @li a vector with pointers to available tokens is returned
	 * @li the number of returned tokens equals the number of requested tokens
	 * @li the acquiring process is stored as owner in the token info map
	 * @li all returned tokens fulfill the selection condition
	 * @li a successful call can be observed
	 * @li change of the available number of tokens can be observed
	 */
	TEST_FIXTURE( ResTChoiceFixture, AcquireMultipleSuccess )
	{
		unsigned int oldTokenCount = res -> getTokenNumber();
		unsigned int takeCount = 10;
		ResTChoiceTaker acquirer( &sim, "ResTChoiceTestMultipleTaker", *res, takeCount, &Synchronization::tokenSelect, takeCount );
		acquirer.activate();
		sim.step(); // take
		CHECK( acquirer.hasReturned() );
		CHECK_EQUAL( (int)takeCount, acquirer.getReturnValue() );

		// check that the acquirer is registered with each token entry in info map
		const std::map< TokenTest*, Process* >& info = res -> getTokenInfo();
		const std::vector< TokenTest* >& tokens = acquirer.receivedTokens;
		std::vector< TokenTest* >::const_iterator i;
		for( i = tokens.begin(); i != tokens.end(); ++i )
		{
			CHECK( tokenSelect( &acquirer, *i ) );
			CHECK_EQUAL( &acquirer, info.find( *i )->second );
		}
		CHECK( observer.observed( "onAcquireSucceed", res -> getLabel(), toString( takeCount ) ) );
		CHECK( observer.observed( "onChangeTokenNumber", res -> getLabel(), toString( oldTokenCount ), toString( res -> getTokenNumber() ) ) );
	}

	/**
	 * @test odemx::ResTChoice::acquire(ResTChoice::Selection,unsigned int) failure
	 *
	 * Expected function call effects:
	 * @li the call fails when requesting more than the available tokens
	 * @li the Process is inserted into a queue to wait for available tokens
	 * @li a failing call can be observed
	 * @li in case of interrupt, an empty vector is returned
	 */
	TEST_FIXTURE( ResTChoiceFixture, AcquireMultipleFailure )
	{
		// ensure that this amount cannot be acquired
		unsigned int takeCount = res -> getTokenNumber() + 1;
		ResTChoiceTaker acquirer( &sim, "ResTChoiceTestMultipleTakerFail", *res, takeCount, &Synchronization::tokenSelect, takeCount );
		acquirer.activate();
		sim.step();

		const std::list< Process* >& waiting = res -> getWaitingProcessList();
		CHECK( find( waiting.begin(), waiting.end(), &acquirer ) != waiting.end() );
		CHECK( observer.observed( "onAcquireFail", res -> getLabel(), toString( takeCount ) ) );

		Base::InterruptEvent interrupt( &sim, &acquirer );
		interrupt.schedule();
		sim.step(); // execute Event
		sim.step(); // execute Process
		CHECK( acquirer.hasReturned() );
		CHECK_EQUAL( (size_t) 0, acquirer.receivedTokens.size() );
	}

	/**
	 * @test odemx::ResTChoice::acquire(ResTChoice::Selection,unsigned int) wait for matching
	 *
	 * Expected function call effects:
	 * @li the call fails when requesting non-available tokens
	 * @li the Process is inserted into a queue to wait for available tokens
	 * @li a failing call can be observed
	 * @li in case of release, the waiting process is reactivated to check its
	 * condition
	 * @li if enough matching tokens have become available, they are acquired
	 * successfully and the process is removed from the queue
	 */
	TEST_FIXTURE( ResTChoiceFixture, AcquireMatchingTokens )
	{
		// acquire all tokens that match id less or equal takeCount
		// will release them after the next process was executed
		unsigned int takeCount = res -> getTokenLimit() / 2;
		ResTChoiceTaker dummy( &sim, "ResTChoiceTestMatchingDummy", *res, takeCount, &Synchronization::tokenSelect, takeCount, true );
		dummy.activate();
		sim.step(); // take

		// attempt to take one token with id less or equal takeCount - 2
		// exists but is already taken...
		ResTChoiceTaker acquirer( &sim, "ResTChoiceTestMatchingTaker", *res, 1, &Synchronization::tokenSelect, 1 );
		acquirer.activate();
		sim.step(); // enqueue

		const std::list< Process* >& waiting = res -> getWaitingProcessList();
		CHECK( find( waiting.begin(), waiting.end(), &acquirer ) != waiting.end() );
		CHECK( observer.observed( "onAcquireFail", res -> getLabel(), toString( 1 ) ) );
		sim.step(); // dummy token release, schedules waiting after dummy
		sim.step(); // take
		CHECK( find( waiting.begin(), waiting.end(), &acquirer ) == waiting.end() );
		CHECK( observer.observed( "onAcquireSucceed", res -> getLabel(), toString( 1 ) ) );
		CHECK( tokenSelect( &acquirer, acquirer.token ) );
		CHECK_EQUAL( &acquirer, res -> getTokenInfo().find( acquirer.token )->second );
	}

	/**
	 * @test odemx::ResTChoice::acquire() warning and error
	 *
	 * Expected function call effects:
	 * @li calls from non-Process objects cause an error message
	 * @li attempting to acquire more than max tokens causes a warning message
	 * @li attempting to acquire more matching tokens than the resource contains
	 * causes a warning message
	 */
	TEST_FIXTURE( ResTChoiceFixture, AcquireError )
	{
		clearErrorStream();
		res -> acquire( &Synchronization::tokenSelect, 10 );
		CHECK_EQUAL( "ODEMx ERROR: ResTChoice::acquire(): called by non-Process object -- Object: ResTChoiceTest\n",
				errorStream().str() );

		clearErrorStream();
		ResTChoiceTaker acquirer( &sim, "ResTChoiceTestFailTaker", *res, res -> getTokenLimit() + 1, &Synchronization::tokenSelect, res -> getTokenLimit()  );
		acquirer.activate();
		sim.step();
		CHECK_EQUAL( string("ODEMx WARNING: ResTChoice::acquire(): requesting more than max resources will always block -- Object: ResTChoiceTest\n")
					+ "ODEMx WARNING: ResTChoice::getMatchingTokens(): too few matching tokens in resource, will always block -- Object: ResTChoiceTest\n",
					errorStream().str() );

		std::vector< TokenTest > init;
		init.push_back( TokenTest( 1, "ResTChoiceTestAcquireErrorToken" ) );
		ResTChoice< TokenTest > failRes( &sim, "ResTChoiceTestAcquireError", init, &observer );

		clearErrorStream();
		ResTChoiceTaker acquirerFail( &sim, "ResTTestTooFewMatchingTaker", failRes, 1, &Synchronization::tokenSelect, 0 );
		acquirerFail.activate();
		sim.step();
		CHECK_EQUAL( "ODEMx WARNING: ResTChoice::getMatchingTokens(): too few matching tokens in resource, will always block -- Object: ResTChoiceTestAcquireError\n",
				errorStream().str() );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
