/**
 * @file TestTimer.cpp
 * @date Aug 1, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class Timer
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	/**
	 * @struct TimerFixture
	 * @brief Helper struct providing set-up/tear-down of Timer tests
	 *
	 * @copydetails EventFixture
	 */
	struct TimerFixture
	{
		Base::SimulationTest sim;
		TimerTestObserver observer;
		Timer timer;

		TimerFixture():
			sim( "TimerTestSim" ),
			observer( false ),
			timer( &sim, "TimerTest", &observer )
			{}
	};

	/**
	 * @test odemx::Timer construction and destruction
	 *
	 * Expected effects:
	 * @li simulation context is set correctly
	 * @li Memory type is TIMER
	 * @li observer is set correctly
	 * @li construction and destruction can be observed
	 */
	TEST_FIXTURE( TimerFixture, ConstructionDestruction )
	{
		Timer* timer1 = new Timer( "TimerTestConstruction1", &observer );
		CHECK_EQUAL( getDefaultSimulation(), timer1 -> getTrace() );
		CHECK_EQUAL( Memory::TIMER, timer1 -> getMemoryType() );
		CHECK_EQUAL( &observer, timer1 -> Observable<EventObserver>::getObservers().front() );
		CHECK_EQUAL( &observer, timer1 -> Observable<MemoryObserver>::getObservers().front() );
 		CHECK_EQUAL( &observer, timer1 -> Observable<TimerObserver>::getObservers().front() );
		CHECK( observer.observed( "onCreate", timer1 -> getLabel() ) );

		Timer* timer2 = new Timer( &sim, "TimerTestConstruction2", &observer );
		CHECK_EQUAL( &sim, timer2 -> getTrace() );
		CHECK_EQUAL( Memory::TIMER, timer2 -> getMemoryType() );
		CHECK_EQUAL( &observer, timer2 -> Observable<EventObserver>::getObservers().front() );
		CHECK_EQUAL( &observer, timer2 -> Observable<MemoryObserver>::getObservers().front() );
 		CHECK_EQUAL( &observer, timer2 -> Observable<TimerObserver>::getObservers().front() );
		CHECK( observer.observed( "onCreate", timer2 -> getLabel() ) );

		Label timer2Label = timer2 -> getLabel();
		delete timer2;
		CHECK( observer.observed( "onDestroy", timer2Label ) );
		delete timer1;
	}

	/**
	 * @test odemx::Timer::setIn(SimTime)
	 *
	 * Expected function call effects:
	 * @li an invalid SimTime value causes an error message and immediate return
	 * @li the Timer is scheduled at \c now + \c delay
	 * @li the call can be observed
	 * @li the timer is set and not available for use
	 */
	TEST_FIXTURE( TimerFixture, SetIn )
	{
		clearErrorStream();
		timer.setIn( -1 );
		CHECK_EQUAL( "ODEMx ERROR: Timer::setIn(): cannot schedule Timer, invalid SimTime < 0 -- Object: TimerTest\n",
				errorStream().str() );

		SimTime delay = 5;
		timer.setIn( delay );
		CHECK_EQUAL( sim.getTime() + delay, timer.getExecutionTime() );
		CHECK( observer.observed( "onSet", timer.getLabel(), toString( delay ) ) );
		CHECK( timer.isSet() );
		CHECK( ! timer.available() );
	}

	/**
	 * @test odemx::Timer::setAt(SimTime)
	 *
	 * Expected function call effects:
	 * @li an invalid SimTime value causes an error message and immediate return
	 * @li the Timer is scheduled at \c absoluteTime
	 * @li the call can be observed
	 * @li the timer is set and not available for use
	 */
	TEST_FIXTURE( TimerFixture, SetAt )
	{
		SimTime delay = 5;
		timer.setIn( delay );
		sim.step();
		CHECK_EQUAL( delay, sim.getTime() );

		clearErrorStream();
		timer.setAt( delay - 1 );
		CHECK_EQUAL( "ODEMx ERROR: Timer::setAt(): cannot schedule Timer, SimTime already passed -- Object: TimerTest\n",
				errorStream().str() );

		SimTime absoluteTime = 99;
		timer.setAt( absoluteTime );
		CHECK_EQUAL( absoluteTime, timer.getExecutionTime() );
		CHECK( observer.observed( "onSet", timer.getLabel(), toString( absoluteTime - sim.getTime() ) ) );
		CHECK( timer.isSet() );
		CHECK( ! timer.available() );
	}

	/**
	 * @test odemx::Timer::reset(SimTime)
	 *
	 * Expected function call effects:
	 * @li resetting an unset timer causes a warning message
	 * @li an invalid SimTime value causes an error message and immediate return
	 * @li the Timer is scheduled at \c now + \c delay
	 * @li the call can be observed
	 * @li the timer is set and not available for use
	 */
	TEST_FIXTURE( TimerFixture, Reset )
	{
		clearErrorStream();
		timer.reset( 0 );
		CHECK_EQUAL( "ODEMx WARNING: Timer::reset(): scheduling idle Timer by reset -- Object: TimerTest\n",
				errorStream().str() );

		clearErrorStream();
		timer.reset( -1 );
		CHECK_EQUAL( "ODEMx ERROR: Timer::reset(): cannot schedule Timer, invalid SimTime < 0 -- Object: TimerTest\n",
				errorStream().str() );

		SimTime delay = 5;
		SimTime oldTimeout = timer.getExecutionTime();
		timer.reset( delay );
		CHECK_EQUAL( sim.getTime() + delay, timer.getExecutionTime() );
		CHECK( observer.observed( "onReset", timer.getLabel(), toString( oldTimeout ), toString( delay ) ) );
		CHECK( timer.isSet() );
		CHECK( ! timer.available() );
	}

	/**
	 * @test odemx::Timer::eventAction()
	 *
	 * This function is tested indirectly by scheduling the Timer and
	 * running the Simulation.
	 *
	 * Expected function call effects:
	 * @li the call can be observed
	 * @li empty Memory causes a warning because Timers are supposed to wake
	 * processes and thus prevent deadlocks
	 * @li the timer is not set and thus available for use
	 * @li registered processes are alerted
	 */
	TEST_FIXTURE( TimerFixture, EventAction )
	{
		clearErrorStream();
		timer.setIn( 0 );
		sim.step();
		CHECK( observer.observed( "onTimeout", timer.getLabel() ) );
		CHECK_EQUAL( "ODEMx WARNING: Timer::eventAction(): Memory empty, cannot alert any Process -- Object: TimerTest\n",
				errorStream().str() );

		Base::ProcessTest process( &sim, "TimerTestProcess" );
		SimTime delay = 5;
		timer.setIn( delay );
		timer.registerProcess( &process );
		sim.step();
		CHECK( process.isAlerted() );
		CHECK( ! timer.isSet() );
		CHECK( timer.available() );
	}

	/**
	 * @test odemx::Timer::stop()
	 *
	 * Expected function call effects:
	 * @li the call can be observed
	 * @li the timer is not set and thus available for use
	 * @li registered processes are removed from Memory
	 */
	TEST_FIXTURE( TimerFixture, Stop )
	{
		Base::ProcessTest process( &sim, "TimerTestProcess" );
		timer.registerProcess( &process );
		timer.setIn( 0 );

		timer.stop();
		CHECK( observer.observed( "onStop", timer.getLabel() ) );
		CHECK( ! timer.isSet() );
		CHECK( timer.available() );
		CHECK( ! timer.waiting() );
	}

	/**
	 * @test odemx::Timer::registerProcess(Process*)
	 *
	 * Expected function call effects:
	 * @li the call can be observed
	 * @li the process is found in the list of waiting processes
	 */
	TEST_FIXTURE( TimerFixture, RegisterProcess )
	{
		Base::ProcessTest process( &sim, "TimerTestProcess" );
		timer.registerProcess( &process );
		CHECK( observer.observed( "onRegisterProcess", timer.getLabel(), process.getLabel() ) );
		CHECK( find( timer.getList().begin(), timer.getList().end(), &process ) != timer.getList().end() );
	}

	/**
	 * @test odemx::Timer::removeProcess(Process*)
	 *
	 * Expected function call effects:
	 * @li the call can be observed
	 * @li the process is not found in the list of waiting processes
	 */
	TEST_FIXTURE( TimerFixture, RemoveProcess )
	{
		Base::ProcessTest process( &sim, "TimerTestProcess" );
		timer.registerProcess( &process );

		timer.removeProcess( &process );
		CHECK( observer.observed( "onRemoveProcess", timer.getLabel(), process.getLabel() ) );
		CHECK( find( timer.getList().begin(), timer.getList().end(), &process ) == timer.getList().end() );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
