/**
 * @file TestWaitQ.cpp
 * @date Aug 2, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx class WaitQ
 */

#include "TestSynchronization.h"
#include "../TestBase/TestBase.h"
#include "../TestUtil/TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Synchronization )
{
/// @endcond

	/**
	 * @struct WaitQFixture
	 * @brief Helper struct providing set-up/tear-down of WaitQ tests
	 *
	 * @copydetails EventFixture
	 */
	struct WaitQFixture
	{
		Base::SimulationTest sim;
		Util::ReportTest reporter;
		WaitQTestObserver observer;
		WaitQ waitQ;

		WaitQFixture():
			sim( "WaitQTestSim" ),
			reporter(),
			observer( false ),
			waitQ( &sim, "WaitQTest", &observer )
			{}
	};

	/**
	 * @test odemx::WaitQ construction and destruction
	 *
	 * Expected effects:
	 * @li observer is set correctly
	 * @li label is set
	 * @li simulation environment is set for DefaultSimulation or user-definied
	 * simulation
	 * @li construction and destruction can be observed
	 */
	TEST_FIXTURE( WaitQFixture, ConstructionDestruction )
	{
		observer.history.clear();

		Label label( "WaitQConstructionTest1" );
		WaitQ* testWaitQ = new WaitQ( label, &observer );
		CHECK_EQUAL( &observer, testWaitQ -> getObservers().front() );
		CHECK_EQUAL( label, testWaitQ -> getLabel() );
		CHECK_EQUAL( getDefaultSimulation(), testWaitQ -> getTrace() );
		CHECK( observer.observed( "onCreate", testWaitQ -> getLabel() ) );

		label = "WaitQConstructionTest2";
		WaitQ* testWaitQ2 = new WaitQ( &sim, label, &observer );
		CHECK_EQUAL( &observer, testWaitQ2 -> getObservers().front() );
		CHECK_EQUAL( label, testWaitQ2 -> getLabel() );
		CHECK_EQUAL( &sim, testWaitQ2 -> getTrace() );
		CHECK( observer.observed( "onCreate", testWaitQ2 -> getLabel() ) );

		delete testWaitQ2;
		CHECK( observer.observed( "onDestroy", label ) );
		delete testWaitQ;
	}

	/**
	 * @test odemx::WaitQ::wait() slave wait call
	 *
	 * Precondition: two processes are waiting as masters.
	 *
	 * Expected function call effects:
	 * @li calling from non-Process objects causes an error message
	 * @li the call can be observed
	 * @li all masters are re-scheduled to check for suitable slave
	 * @li the slave is put into a queue to wait for a master Process
	 * @li on slave interrupt, the function returns false
	 */
	TEST_FIXTURE( WaitQFixture, Wait )
	{
		WaitQTestMaster master1( &sim, "WaitQTestWaitMaster1", waitQ );
		WaitQTestMaster master2( &sim, "WaitQTestWaitMaster2", waitQ );
		master1.activate();
		master2.activate();
		sim.step(); // enqueue
		sim.step(); // enqueue
		const std::list< Process* >& masters = waitQ.getWaitingMasters();
		CHECK( find( masters.begin(), masters.end(), &master1 ) != masters.end() );
		CHECK( find( masters.begin(), masters.end(), &master2 ) != masters.end() );

		clearErrorStream();
		waitQ.wait();
		CHECK_EQUAL( "ODEMx ERROR: WaitQ::wait(): called by non-Process object -- Object: WaitQTest\n",
				errorStream().str() );

		WaitQTestSlave slave( &sim, "WaitQTestWaitSlave", waitQ );
		slave.activate();
		sim.step(); // enqueue and wake masters to check for slaves

		CHECK( observer.observed( "onWait", waitQ.getLabel(), slave.getLabel() ) );
		CHECK( master1.isScheduled() );
		CHECK( master2.isScheduled() );
		const std::list< Process* >& slaves = waitQ.getWaitingSlaves();
		CHECK( find( slaves.begin(), slaves.end(), &slave ) != slaves.end() );
		CHECK_EQUAL( Process::IDLE, slave.getProcessState() );

		Base::InterruptEvent interrupt( &sim, &slave );
		interrupt.schedule();
		sim.step();
		CHECK( slave.isScheduled() );
		sim.step();
		CHECK_EQUAL( false, slave.waitReturnValue );
	}


	/**
	 * @test odemx::WaitQ::wait(Weight) slave wait call without waiting masters
	 *
	 * Precondition: two processes are waiting as masters.
	 *
	 * Expected function call effects:
	 * @li calling from non-Process objects causes an error message
	 * @li the call can be observed
	 * @li the slave is put into a queue to wait for a master Process
	 * @li on slave interrupt, the function returns false
	 */
	TEST_FIXTURE( WaitQFixture, WaitWeightEmpty )
	{
		clearErrorStream();
		waitQ.wait( (Weight) &WaitQTestMaster::weight );
		CHECK_EQUAL( "ODEMx ERROR: WaitQ::wait(): called by non-Process object -- Object: WaitQTest\n",
				errorStream().str() );

		WaitQTestSlave slave( &sim, "WaitQTestWaitWeightEmptySlave", waitQ, 0, (Weight) &WaitQTestMaster::weight );
		slave.activate();
		sim.step(); // enqueue and wake masters to check for slaves

		CHECK( observer.observed( "onWaitWeight", waitQ.getLabel(), slave.getLabel() ) );
		const std::list< Process* >& slaves = waitQ.getWaitingSlaves();
		CHECK( find( slaves.begin(), slaves.end(), &slave ) != slaves.end() );
		CHECK_EQUAL( Process::IDLE, slave.getProcessState() );

		Base::InterruptEvent interrupt( &sim, &slave );
		interrupt.schedule();
		sim.step();
		CHECK( slave.isScheduled() );
		sim.step();
		CHECK_EQUAL( false, slave.waitReturnValue );
	}

	/**
	 * @test odemx::WaitQ::wait(Weight) slave wait call without waiting masters
	 *
	 * Precondition: two processes are waiting as masters.
	 *
	 * Expected function call effects:
	 * @li the slave is weighted against all masters and the master with the
	 * highest weight is scheduled in order to acquire this slave
	 */
	TEST_FIXTURE( WaitQFixture, WaitWeight )
	{
		WaitQTestMaster master1( &sim, "WaitQTestWaitWeightMaster1", waitQ, 0, true );
		WaitQTestMaster master2( &sim, "WaitQTestWaitWeightMaster2", waitQ, 0, true );
		master1.activate();
		master2.activate();
		sim.step(); // enqueue
		sim.step(); // enqueue
		const std::list< Process* >& masters = waitQ.getWaitingMasters();
		CHECK( find( masters.begin(), masters.end(), &master1 ) != masters.end() );
		CHECK( find( masters.begin(), masters.end(), &master2 ) != masters.end() );

		// make sure this master will be chosen
		master1.weightValue = 10;
		WaitQTestSlave slave( &sim, "WaitQTestWaitWeightSlave", waitQ, 0, (Weight) &WaitQTestMaster::weight );
		slave.activate();
		sim.step(); // enqueue and check masters for best weight

		CHECK( ! master2.isScheduled() );
		CHECK( master1.isScheduled() );
		sim.step();

		CHECK_EQUAL( &slave, master1.slave );
	}

	/**
	 * @test odemx::WaitQ::coopt() success
	 *
	 * Precondition: a process is waiting as slave.
	 *
	 * Expected function call effects:
	 * @li calling from non-Process objects causes an error message
	 * @li the successful call can be observed
	 * @li the statistics are updated
	 * @li the waiting slave is taken by the master
	 */
	TEST_FIXTURE( WaitQFixture, Coopt )
	{
		clearErrorStream();
		waitQ.coopt();
		CHECK_EQUAL( "ODEMx ERROR: WaitQ::coopt(): called by non-Process object -- Object: WaitQTest\n",
				errorStream().str() );

		WaitQTestSlave slave1( &sim, "WaitQTestCooptSlave1", waitQ );
		slave1.activate();
		sim.step(); // enqueue

		unsigned int synchs = waitQ.getNumberOfSynch();
		WaitQTestMaster master( &sim, "WaitQTestCooptMaster", waitQ );
		master.activate();
		sim.step(); // get slave
		CHECK( observer.observed( "onCooptSucceed", waitQ.getLabel(), master.getLabel(), slave1.getLabel() ) );
		CHECK_EQUAL( synchs + 1, waitQ.getNumberOfSynch() );
		CHECK_EQUAL( &slave1, master.slave );
	}

	/**
	 * @test odemx::WaitQ::coopt() without waiting slave
	 *
	 * Precondition: no process is waiting as slave.
	 *
	 * Expected function call effects:
	 * @li the calling process is suspended
	 * @li the caller is added to the master waiting queue
	 * @li the failing call can be observed
	 * @li in case of interrupt the return value is a 0-pointer
	 */
	TEST_FIXTURE( WaitQFixture, CooptFail )
	{
		WaitQTestMaster master( &sim, "WaitQTestCooptFailMaster", waitQ );
		master.activate();
		sim.step(); // coopt fails
		CHECK( observer.observed( "onCooptFail", waitQ.getLabel(), master.getLabel() ) );
		const std::list< Process* >& masters = waitQ.getWaitingMasters();
		CHECK( find( masters.begin(), masters.end(), &master ) != masters.end() );

		Base::InterruptEvent interrupt( &sim, &master );
		interrupt.schedule();
		sim.step();
		CHECK( master.isScheduled() );
		sim.step();
		CHECK_EQUAL( static_cast< Process* >( 0 ), master.slave );
	}

	/**
	 * @test odemx::WaitQ::coopt(Selection) success
	 *
	 * Precondition: two processes are waiting as slave, one fits.
	 *
	 * Expected function call effects:
	 * @li the successful call can be observed
	 * @li the statistics are updated
	 * @li the waiting slave is taken by the master
	 */
	TEST_FIXTURE( WaitQFixture, CooptSelect )
	{
		int selectValue = 125;
		WaitQTestSlave slave1( &sim, "WaitQTestCooptSelSlave1", waitQ, selectValue );
		WaitQTestSlave slave2( &sim, "WaitQTestCooptSelSlave2", waitQ, 5 );
		slave1.activate();
		slave2.activate();
		sim.step(); // enqueue
		sim.step(); // enqueue

		// make sure slave1 is chosen
		unsigned int synchs = waitQ.getNumberOfSynch();
		WaitQTestMaster cooptTester( &sim, "WaitQTestCooptMaster", waitQ, selectValue );
		cooptTester.activate();
		sim.step(); // get slave
		CHECK( observer.observed( "onCooptSelSucceed", waitQ.getLabel(), cooptTester.getLabel(), slave1.getLabel() ) );
		CHECK_EQUAL( synchs + 1, waitQ.getNumberOfSynch() );
		CHECK_EQUAL( &slave1, cooptTester.slave );
	}

	/**
	 * @test odemx::WaitQ::coopt(Selection) failure with waiting
	 *
	 * Precondition: two processes are waiting as slave, but neither fits the selection.
	 *
	 * Expected function call effects:
	 * @li the failing call can be observed
	 * @li the calling process is added to the master waiting queue
	 * @li in case of master interrupt, the return value is a 0-pointer
	 */
	TEST_FIXTURE( WaitQFixture, CooptSelectFail )
	{
		int selectValue = 125;
		WaitQTestSlave slave1( &sim, "WaitQTestCooptSelFailSlave1", waitQ, 1 );
		WaitQTestSlave slave2( &sim, "WaitQTestCooptSelFailSlave2", waitQ, 2 );
		slave1.activate();
		slave2.activate();
		sim.step(); // enqueue
		sim.step(); // enqueue

		WaitQTestMaster cooptTester( &sim, "WaitQTestCooptSelMaster", waitQ, selectValue );
		cooptTester.activate();
		sim.step(); // get slave
		CHECK( observer.observed( "onCooptSelFail", waitQ.getLabel(), cooptTester.getLabel(), slave1.getLabel() ) );
		const std::list< Process* >& masters = waitQ.getWaitingMasters();
		CHECK( find( masters.begin(), masters.end(), &cooptTester ) != masters.end() );

		Base::InterruptEvent interrupt( &sim, &cooptTester );
		interrupt.schedule();
		sim.step();
		CHECK( cooptTester.isScheduled() );
		sim.step();
		CHECK_EQUAL( static_cast< Process* >( 0 ), cooptTester.slave );
	}

	/**
	 * @test odemx::WaitQ::coopt(Weight) master coopt call with waiting slaves
	 *
	 * Precondition: two processes are waiting as slaves.
	 *
	 * Expected function call effects:
	 * @li the call can be observed
	 * @li the statistics are updated
	 * @li the master received the highest weighing slave
	 */
	TEST_FIXTURE( WaitQFixture, CooptWeight )
	{
		WaitQTestSlave slave1( &sim, "WaitQTestCooptWeightSlave1", waitQ, 0, (Weight) &WaitQTestMaster::weight );
		WaitQTestSlave slave2( &sim, "WaitQTestCooptWeightSlave2", waitQ, 0, (Weight) &WaitQTestMaster::weight );
		slave1.activate();
		slave2.activate();
		sim.step(); // enqueue
		sim.step(); // enqueue

		// make sure slave1 is chosen
		slave1.weightValue = 10;
		unsigned int synchs = waitQ.getNumberOfSynch();
		WaitQTestMaster master1( &sim, "WaitQTestCooptWeightMaster1", waitQ, 0, true );
		master1.activate();
		sim.step(); // enqueue
		CHECK( observer.observed( "onCooptWeightSucceed", waitQ.getLabel(), master1.getLabel(), slave1.getLabel() ) );
		CHECK_EQUAL( synchs + 1, waitQ.getNumberOfSynch() );
		CHECK_EQUAL( &slave1, master1.slave );
	}

	/**
	 * @test odemx::WaitQ::coopt(Weight) master coopt call without waiting slaves
	 *
	 * Precondition: two processes are waiting as masters.
	 *
	 * Expected function call effects:
	 * @li calling from non-Process objects causes an error message
	 * @li the calling process is added to a waiting queue
	 * @li the calling process is suspended
	 * @li the failing call can be observed
	 */
	TEST_FIXTURE( WaitQFixture, CooptWeightFail )
	{
		clearErrorStream();
		waitQ.coopt( (Weight) &WaitQTestMaster::weight );
		CHECK_EQUAL( "ODEMx ERROR: WaitQ::coopt(): called by non-Process object -- Object: WaitQTest\n",
				errorStream().str() );

		WaitQTestMaster master1( &sim, "WaitQTestCooptWeightMaster1", waitQ, 0, true );
		master1.activate();
		sim.step(); // enqueue
		const std::list< Process* >& masters = waitQ.getWaitingMasters();
		CHECK( find( masters.begin(), masters.end(), &master1 ) != masters.end() );
		CHECK_EQUAL( Process::IDLE, master1.getProcessState() );
		CHECK( observer.observed( "onCooptWeightFail", waitQ.getLabel(), master1.getLabel() ) );

		Base::InterruptEvent interrupt( &sim, &master1 );
		interrupt.schedule();
		sim.step();
		CHECK( master1.isScheduled() );
		sim.step();
		CHECK_EQUAL( static_cast< Process* >( 0 ), master1.slave );
	}

	/**
	 * @test odemx::WaitQ::avail() get slave if available without waiting
	 *
	 * Precondition: a process is waiting as slave.
	 *
	 * Expected function call effects:
	 * @li calling from non-Process objects causes an error message
	 * @li the successful call can be observed
	 * @li the statistics are updated
	 * @li the waiting slave is taken by the master
	 */
	TEST_FIXTURE( WaitQFixture, Avail )
	{
		clearErrorStream();
		waitQ.avail();
		CHECK_EQUAL( "ODEMx ERROR: WaitQ::avail(): called by non-Process object -- Object: WaitQTest\n",
				errorStream().str() );

		WaitQTestSlave slave1( &sim, "WaitQTestAvailSlave1", waitQ );
		slave1.activate();
		sim.step(); // enqueue

		unsigned int synchs = waitQ.getNumberOfSynch();
		WaitQTestAvail availTester( &sim, "WaitQTestAvailMaster", waitQ );
		availTester.activate();
		sim.step(); // get slave
		CHECK( observer.observed( "onAvailSucceed", waitQ.getLabel(), availTester.getLabel(), slave1.getLabel() ) );
		CHECK_EQUAL( synchs + 1, waitQ.getNumberOfSynch() );
		CHECK_EQUAL( &slave1, availTester.slave );
	}

	/**
	 * @test odemx::WaitQ::avail() without waiting slave
	 *
	 * Precondition: no process is waiting as slave.
	 *
	 * Expected function call effects:
	 * @li the calling process is not suspended
	 * @li the failing call can be observed
	 * @li the return value is a 0-pointer
	 */
	TEST_FIXTURE( WaitQFixture, AvailFail )
	{
		WaitQTestAvail availTester( &sim, "WaitQTestAvailEmptyMaster", waitQ );
		availTester.activate();
		sim.step(); // avail fails
		CHECK( observer.observed( "onAvailFail", waitQ.getLabel(), availTester.getLabel() ) );
		CHECK( availTester.hasReturned() );
		CHECK_EQUAL( static_cast< Process* >( 0 ), availTester.slave );
	}

	/**
	 * @test odemx::WaitQ::avail(Selection) get selected slave if available without waiting
	 *
	 * Precondition: two processes are waiting as slave.
	 *
	 * Expected function call effects:
	 * @li the successful call can be observed
	 * @li the statistics are updated
	 * @li the waiting slave is taken by the master
	 */
	TEST_FIXTURE( WaitQFixture, AvailSelect )
	{
		int selectValue = 125;
		WaitQTestSlave slave1( &sim, "WaitQTestAvailSelSlave1", waitQ, selectValue );
		WaitQTestSlave slave2( &sim, "WaitQTestAvailSelSlave2", waitQ, 5 );
		slave1.activate();
		slave2.activate();
		sim.step(); // enqueue
		sim.step(); // enqueue

		// make sure slave1 is chosen
		unsigned int synchs = waitQ.getNumberOfSynch();
		WaitQTestAvail availTester( &sim, "WaitQTestAvailMaster", waitQ, selectValue );
		availTester.activate();
		sim.step(); // get slave
		CHECK( observer.observed( "onAvailSelSucceed", waitQ.getLabel(), availTester.getLabel(), slave1.getLabel() ) );
		CHECK_EQUAL( synchs + 1, waitQ.getNumberOfSynch() );
		CHECK_EQUAL( &slave1, availTester.slave );
	}

	/**
	 * @test odemx::WaitQ::avail(Selection) get selected slave if available without waiting
	 *
	 * Precondition: two processes are waiting as slave, but neither fits the selection.
	 *
	 * Expected function call effects:
	 * @li the failing call can be observed
	 * @li the return value is a 0-pointer
	 */
	TEST_FIXTURE( WaitQFixture, AvailSelectFail )
	{
		int selectValue = 125;
		WaitQTestSlave slave1( &sim, "WaitQTestAvailSelFailSlave1", waitQ, 1 );
		WaitQTestSlave slave2( &sim, "WaitQTestAvailSelFailSlave2", waitQ, 2 );
		slave1.activate();
		slave2.activate();
		sim.step(); // enqueue
		sim.step(); // enqueue

		WaitQTestAvail availTester( &sim, "WaitQTestAvailSelFailMaster", waitQ, selectValue );
		availTester.activate();
		sim.step(); // get slave
		CHECK( observer.observed( "onAvailSelFail", waitQ.getLabel(), availTester.getLabel(), slave1.getLabel() ) );
		CHECK( availTester.hasReturned() );
		CHECK_EQUAL( static_cast< Process* >( 0 ), availTester.slave );
	}

	/**
	 * @test odemx::WaitQ::report()
	 *
	 * Expected function call effects:
	 * @li all table columns are filled with correct values
	 */
	TEST_FIXTURE( WaitQFixture, Report )
	{
		Base::EventTest simAdvanceEvent( &sim, "QueueTestReportAdvanceEvent" );
		simAdvanceEvent.scheduleAt( 14 );
		sim.step(); // event execution advances sim time
		waitQ.reset();

		WaitQTestSlave slave1( &sim, "WaitQTestCooptSlave1", waitQ );
		slave1.activate();
		sim.step(); // enqueue

		simAdvanceEvent.scheduleIn( 5 );
		sim.step(); // event execution advances sim time

		WaitQTestMaster master( &sim, "WaitQTestCooptMaster", waitQ );
		master.activate();
		sim.step(); // get slave

		reporter.addProducer( &waitQ );
		reporter.generateReport();
		CHECK( reporter.processedTables );
		CHECK_EQUAL( (size_t) 2, reporter.allTableContent.size() );

		// only check second table, the first one holds the master and slave queue statistics
		CHECK_EQUAL( waitQ.getLabel(), reporter.allTableContent[1][0] );
		CHECK_EQUAL( toString( waitQ.getResetTime() ), reporter.allTableContent[1][1] );
		CHECK_EQUAL( waitQ.getLabel() + " master queue", reporter.allTableContent[1][2] );
		CHECK_EQUAL( waitQ.getLabel() + " slave queue", reporter.allTableContent[1][3] );
		CHECK_EQUAL( toString( waitQ.getNumberOfSynch() ), reporter.allTableContent[1][4] );
		CHECK_EQUAL( toString( waitQ.getZeroWaitMasters() ), reporter.allTableContent[1][5] );
		CHECK_EQUAL( toString( waitQ.getAVMasterWaitTime() ), reporter.allTableContent[1][6] );
		CHECK_EQUAL( toString( waitQ.getZeroWaitSlaves() ), reporter.allTableContent[1][7] );
		CHECK_EQUAL( toString( waitQ.getAVSlaveWaitTime() ), reporter.allTableContent[1][8] );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
