/**
 * @file TestErrorHandling.cpp
 * @brief Tests for ODEMx debug, warning and error messages
 */

#include "TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Util )
{
/// @endcond

	/**
	 * @struct ErrorHandlingFixture
	 * @brief Helper struct providing set-up/tear-down of error handling tests
	 *
	 * @copydetails EventFixture
	 */
	struct ErrorHandlingFixture
	{
		LabelScopeTest labelScope; ///< a user-defined LabelScope
		DefLabeledObject labeledObject; ///< a labeled object

		ErrorHandlingFixture():
			labelScope(),
			labeledObject( &labelScope, "TestErrorHandlingLabeledObject" )
			{}
	};

	/**
	 * @test odemx::debugOut(string)
	 *
	 * When the macro _DEBUG is defined, ODEMx provides debug message output
	 * by defining the macro debugOut, which calls odemx::debug(). This test is
	 * simply a call to debugOut(). It should print the given message to stderr,
	 * which can be validated by checking errorStream().
	 */
	TEST( Debug )
	{
		clearErrorStream();
		debugOut( "debug message" );
		CHECK( errorStream().str() == "ODEMx DEBUG: debug message\n" );
	}

	/**
	 * @test odemx::debugOut(LabelObject*,string)
	 *
	 * When the macro _DEBUG is defined, ODEMx provides debug message output
	 * by defining the macro debugOut, which calls odemx::debug(). This test
	 * calls debugOut() with a labeled object and a message. It should
	 * print the message and the object's label to stderr, which can be
	 * validated by checking errorStream().
	 */
	TEST_FIXTURE( ErrorHandlingFixture, LabeledDebug )
	{
		clearErrorStream();
		debugOutLabel( &labeledObject, "debug message" );
		CHECK( errorStream().str() == "ODEMx DEBUG: debug message -- Object: TestErrorHandlingLabeledObject\n" );
	}

	/**
	 * @test odemx::warning(string)
	 *
	 * This test calls warning() with a string message. It should print
	 * the message to stderr, which can be validated by checking errorStream().
	 */
	TEST( Warning )
	{
		clearErrorStream();
		warning( "warning message" );
		CHECK( errorStream().str() == "ODEMx WARNING: warning message\n" );
	}

	/**
	 * @test odemx::warning(LabeledObject*,string)
	 *
	 * This test calls warning() with a labeled object and a message. It should
	 * print the message and the object's label to stderr, which can be
	 * validated by checking errorStream().
	 */
	TEST_FIXTURE( ErrorHandlingFixture, LabeledWarning )
	{
		clearErrorStream();
		warning( &labeledObject, "warning message" );
		CHECK( errorStream().str() == "ODEMx WARNING: warning message -- Object: TestErrorHandlingLabeledObject\n" );
	}

	/**
	 * @test odemx::error(string)
	 *
	 * This test calls error() with a string message. It should print
	 * the message to stderr, which can be validated by checking errorStream().
	 */
	TEST( Error )
	{
		clearErrorStream();
		error( "error message" );
		CHECK( errorStream().str() == "ODEMx ERROR: error message\n" );
	}

	/**
	 * @test odemx::error(LabeledObject*,string)
	 *
	 * This test calls error() with a labeled object and a message. It should
	 * print the message and the object's label to stderr, which can be
	 * validated by checking errorStream().
	 */
	TEST_FIXTURE( ErrorHandlingFixture, LabeledError )
	{
		clearErrorStream();
		error( &labeledObject, "error message" );
		CHECK( errorStream().str() == "ODEMx ERROR: error message -- Object: TestErrorHandlingLabeledObject\n" );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
