/**
 * @file TestLabeledObject.cpp
 * @brief Tests for ODEMx LabelScope and LabeledObject
 */

#include "TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Util )
{
/// @endcond

	/**
	 * @test LabelScope construction
	 *
	 * This test first checks the default values after construction of a
	 * LabelScope object. A second test object is constructed with user-defined
	 * values.
	 */
	TEST( LabelScopeConstructionDestruction )
	{
		LabelScopeTest defLabelScope;
		CHECK_EQUAL( defLabelScope.getSpace(), '_' );
		CHECK_EQUAL( defLabelScope.getDefLabel(), "NoName" );

		LabelScopeTest labelScope( "DefaultLabel", ' ' );
		CHECK_EQUAL( labelScope.getSpace(), ' ' );
		CHECK_EQUAL( labelScope.getDefLabel(), "DefaultLabel" );
	}

	/**
	 * @test odemx::LabelScope::createLabel(string&)
	 *
	 * Expected function call effects:
	 * @li calling without parameter creates a default label
	 * @li calling with unique string paramter returns that string as label
	 * @li calling with a previously used string parameter creates a label
	 * with appended default space character and a number
	 */
	TEST( LabelScopeCreateLabel )
	{
		LabelScopeTest labelScope( "DefaultLabel", ' ' );

		Label defLabel = labelScope.createLabel();
		CHECK_EQUAL( labelScope.getDefLabel(), defLabel );

		Label testLabel1 = labelScope.createLabel( "LabelTest" );
		CHECK_EQUAL( "LabelTest", testLabel1 );

		Label testLabel2 = labelScope.createLabel( "LabelTest" );
		std::ostringstream testLabelStream;
		testLabelStream << "LabelTest" << labelScope.getSpace() << 1;
		CHECK_EQUAL( testLabelStream.str(), testLabel2 );
	}

	/**
	 * @test odemx::DefLabeledObject construction
	 *
	 * Test label initialization:
	 * @li default constructor creates label "???"
	 * @li constructor without scope creates just the plain given label
	 * @li constructor with scope creates unique counted labels
	 */
	TEST( DefLabeledObjectConstructionDestruction )
	{
		DefLabeledObject dlo;
		CHECK_EQUAL( "???", dlo.getLabel() );

		DefLabeledObject dlo2( 0 , "LabelTestWithoutScope" );
		CHECK_EQUAL( "LabelTestWithoutScope", dlo2.getLabel() );

		LabelScopeTest labelScope( "DefaultLabel", ' ' );
		Label scopedLabel = "LabelTestWithScope";
		DefLabeledObject dlo3( &labelScope, scopedLabel );
		CHECK_EQUAL( scopedLabel, dlo3.getLabel() );

		// check for appended number
		DefLabeledObject dlo4( &labelScope, scopedLabel );
		std::ostringstream testLabelStream;
		testLabelStream << scopedLabel << labelScope.getSpace() << 1;
		CHECK_EQUAL( testLabelStream.str(), dlo4.getLabel() );
	}

	/**
	 * @test odemx::DefLabeledObject::setLabel(LabelScope*,string&)
	 *
	 * Test label initialization:
	 * @li \c setLabel() without scope creates just the plain given label
	 * @li \c setLabel() with scope creates unique counted labels
	 */
	TEST( DefLabeledObjectSetLabel )
	{
		DefLabeledObjectTest dlo;
		dlo.setLabel( 0 , "LabelTestWithoutScope" );
		CHECK_EQUAL( "LabelTestWithoutScope", dlo.getLabel() );

		LabelScopeTest labelScope( "DefaultLabel", ' ' );
		Label scopedLabel = "LabelTestWithScope";
		DefLabeledObjectTest dlo2;
		dlo2.setLabel( &labelScope, scopedLabel );
		CHECK_EQUAL( scopedLabel, dlo2.getLabel() );

		// check for appended number
		DefLabeledObjectTest dlo3;
		dlo3.setLabel( &labelScope, scopedLabel );
		std::ostringstream testLabelStream;
		testLabelStream << scopedLabel << labelScope.getSpace() << 1;
		CHECK_EQUAL( testLabelStream.str(), dlo3.getLabel() );
	}


/// @cond DOXYGEN_SKIP
}
/// @endcond
