/**
 * @file TestMarkType.cpp
 * @date Jul 6, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx Tag and MarkType
 */

#include "TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Util )
{
/// @endcond

	/**
	 * @test odemx::MarkType construction and destruction
	 *
	 * Tests member initialization:
	 * @li name
	 * @li id
	 * @li scope
	 * @li addition of new marks to a list of marks upon construction
	 * @li removal from the upon destruction
	 */
	TEST( MarkTypeConstructionDestruction )
	{
		// this would add another mark to MarkTypeTest's list
		MarkType testMark( "testMark", 12345, typeid(MarkTypeTest) );
		MarkTypeList& markList = MarkType::getList();

		CHECK_EQUAL( "testMark", testMark.getName() );
		CHECK_EQUAL( static_cast< MarkTypeId >( 12345 ), testMark.getId() );
		CHECK_EQUAL( typeToString( typeid( MarkTypeTest ) ), typeToString( testMark.getScope() ) );
		CHECK( find( markList.begin(), markList.end(), &testMark ) != markList.end() );

		// this would add another mark to MarkTypeTest's list
		MarkType* testMark2 = new MarkType( "listTestMark", 12345, typeid(MarkTypeTest) );
		CHECK( find( markList.begin(), markList.end(), testMark2 ) != markList.end() );

		delete testMark2;
		CHECK( find( markList.begin(), markList.end(), testMark2 ) == markList.end() );
	}

	/**
	 * @test odemx::MarkType static mark creation and registration
	 *
	 * This test checks the creation and correct registration of static
	 * MarkType class members. MarkTypeTest is used for this purpose.
	 * there are two static marks declared for this class.
	 *
	 * Expected observations:
	 * @li \c mark1: correct initialization of name, id, and scope
	 * @li \c mark2: correct initialization of name, id, and scope
	 * @li the static list of all marks contains pointers to \c mark1 and
	 * \c mark2
	 * @li the scope-id map contains MarkTypeTest as a scope name and
	 * associates it with its \c baseMarkId
	 */
	TEST( MarkTypeStaticMarks )
	{
		CHECK_EQUAL( "mark1", MarkTypeTest::mark1.getName() );
		CHECK_EQUAL( MarkTypeTest::baseMarkId+1, MarkTypeTest::mark1.getId() );
		CHECK_EQUAL( typeToString( typeid( MarkTypeTest ) ), typeToString( MarkTypeTest::mark1.getScope() ) );

		CHECK_EQUAL( "mark2", MarkTypeTest::mark2.getName() );
		CHECK_EQUAL( MarkTypeTest::baseMarkId+2, MarkTypeTest::mark2.getId() );
		CHECK_EQUAL( typeToString( typeid( MarkTypeTest ) ), typeToString( MarkTypeTest::mark2.getScope() ) );

		MarkTypeList& markList = MarkType::getList();
		CHECK( find( markList.begin(), markList.end(), &MarkTypeTest::mark1 ) != markList.end() );
		CHECK( find( markList.begin(), markList.end(), &MarkTypeTest::mark2 ) != markList.end() );

		// check if the type was entered as scope in a map, with its baseMarkId
		MarkScopeIdMap& scopeIdMap = MarkType::getScopeIdMap();
		CHECK( scopeIdMap.find( typeToString( typeid( MarkTypeTest ) ) ) != scopeIdMap.end() );
		CHECK_EQUAL( MarkTypeTest::baseMarkId, scopeIdMap[ typeToString( typeid( MarkTypeTest ) ) ] );
	}

	/**
	 * @test odemx::Tag construction
	 *
	 * Tests member initialization:
	 * @li name
	 * @li id
	 * @li static tag members
	 */
	TEST( TagConstructionDestruction )
	{
		Tag testTag( "testTag", 67890 );
		CHECK_EQUAL( "testTag", testTag.getName() );
		CHECK_EQUAL( static_cast< TagId >( 67890 ), testTag.getId() );

		CHECK_EQUAL( "MarkTypeTest additional info tag", MarkTypeTest::tag1.getName() );
		CHECK_EQUAL( MarkTypeTest::baseTagId+1, MarkTypeTest::tag1.getId() );
	}

	// required definitions of the static MarkType and baseMarkId constants
	const MarkTypeId Util::MarkTypeTest::baseMarkId =
		MarkType::getNewBaseMarkId( typeid(MarkTypeTest) );

	const MarkType
	Util::MarkTypeTest::mark1( "mark1", baseMarkId+1, typeid(MarkTypeTest));

	const MarkType
	Util::MarkTypeTest::mark2( "mark2", baseMarkId+2, typeid(MarkTypeTest));

	const TagId
	Util::MarkTypeTest::baseTagId = MarkTypeTest::baseMarkId;

	const Tag
	Util::MarkTypeTest::tag1( "MarkTypeTest additional info tag", baseTagId+1 );

/// @cond DOXYGEN_SKIP
}
/// @endcond
