/**
 * @file TestReport.cpp
 * @date Jul 12, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx classes Report, ReportProducer
 */

#include "TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Util )
{
/// @endcond

	/**
	 * @struct ReportFixture
	 * @brief Helper struct providing set-up/tear-down of Report tests
	 *
	 * @copydetails EventFixture
	 */
	struct ReportFixture
	{
		ReportProducerTest producer1;
		ReportProducerTest producer2;
		ReportTest reporter;
	};

	/**
	 * @test odemx::Report::addProducer(ReportProducer*)
	 *
	 * Expected function call effects:
	 * @li the producers are added to the internal list
	 */
	TEST_FIXTURE( ReportFixture, AddProducer )
	{
		std::list< ReportProducer* >& producers = reporter.getReportProducers();
		CHECK( producers.empty() );

		reporter.addProducer( &producer1 );
		reporter.addProducer( &producer2 );

		CHECK_EQUAL( size_t(2), producers.size() );
		CHECK( find( producers.begin(), producers.end(), &producer1 ) != producers.end() );
		CHECK( find( producers.begin(), producers.end(), &producer2 ) != producers.end() );
	}

	/**
	 * @test odemx::Report::removeProducer(ReportProducer*)
	 *
	 * Expected function call effects:
	 * @li the producers are removed from the internal list
	 */
	TEST_FIXTURE( ReportFixture, RemoveProducer )
	{
		std::list< ReportProducer* >& producers = reporter.getReportProducers();
		CHECK( producers.empty() );

		reporter.addProducer( &producer1 );
		reporter.addProducer( &producer2 );
		reporter.removeProducer( &producer1 );

		CHECK_EQUAL( size_t(1), producers.size() );
		CHECK( find( producers.begin(), producers.end(), &producer1 ) == producers.end() );
		CHECK( find( producers.begin(), producers.end(), &producer2 ) != producers.end() );
	}

	/**
	 * @test odemx::Report::createTable(const string&,TableDefinition*)
	 *
	 * Expected function call effects:
	 * @li with the first call, a new Table is created
	 * @li the table is added to the internal table vector
	 * @li with the same call, the previously created table is returned
	 */
	TEST_FIXTURE( ReportFixture, CreateTable )
	{
		std::vector< Table* >& tables = reporter.getReportTables();
		CHECK( tables.empty() );

		DynamicTableDefinition tableDef;

		Table* table = reporter.createTable( "Test Table", &tableDef );

		CHECK_EQUAL( size_t(1), tables.size() );
		CHECK( find( tables.begin(), tables.end(), table ) != tables.end() );

		Table* table2 = reporter.createTable( "Test Table", &tableDef );
		CHECK_EQUAL( table, table2 );
	}

	/**
	 * @test odemx::Report::findTable(const string&,TableDefinition*)
	 *
	 * Expected function call effects:
	 * @li returns 0 if table name is empty or table definition is 0
	 * @li tables with different definitions but the same labels are found
	 * @li tables with different labels but the same definitions are found
	 */
	TEST_FIXTURE( ReportFixture, FindTable )
	{
		std::vector< Table* >& tables = reporter.getReportTables();
		CHECK( tables.empty() );

		DynamicTableDefinition tableDef;
		DynamicTableDefinition tableDef2;
		tableDef2.addColumn( "int", Column::INTEGER );

		Table* table1 = reporter.createTable( "Test Table", &tableDef );
		Table* table2 = reporter.createTable( "Test Table 2", &tableDef );
		Table* table3 = reporter.createTable( "Test Table", &tableDef2 );

		CHECK( 0 == reporter.findTable( "Test Table", 0 ) );
		CHECK( 0 == reporter.findTable( "", &tableDef ) );

		Table* found = 0;
		found = reporter.findTable( "Test Table", &tableDef );
		CHECK_EQUAL( table1, found );

		found = reporter.findTable( "Test Table 2", &tableDef );
		CHECK_EQUAL( table2, found );

		found = reporter.findTable( "Test Table", &tableDef2 );
		CHECK_EQUAL( table3, found );
	}

	/**
	 * @test odemx::Report::generateReport()
	 *
	 * Expected function call effects:
	 * @li all registered producers called \c report()
	 * @li the Report object called \c processTables()
	 */
	TEST_FIXTURE( ReportFixture, GenerateReport )
	{
		std::list< ReportProducer* >& producers = reporter.getReportProducers();
		CHECK( producers.empty() );

		reporter.addProducer( &producer1 );
		reporter.addProducer( &producer2 );
		CHECK( ! producer1.sentReport );
		CHECK( ! producer2.sentReport );

		reporter.generateReport();

		CHECK( producer1.sentReport );
		CHECK( producer2.sentReport );
		CHECK( reporter.processedTables );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
