/**
 * @file TestStatisticObject.cpp
 * @date Jul 11, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx classes StatisticObject and StatisticManager
 */

#include "TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Util )
{
/// @endcond

	/**
	 * @struct StatisticObjectFixture
	 * @brief Helper struct providing set-up/tear-down of StatisticObject tests
	 *
	 * @copydetails EventFixture
	 */
	struct StatisticObjectFixture
	{
		StatisticManagerTest manager1; ///< a statistics manager for testing
		StatisticManagerTest manager2; ///< another statistics manager for switching
		StatisticObject statsObject1; ///< a statistics object
		StatisticObject statsObject2; ///< a statistic object to test for-each calls

		StatisticObjectFixture()
		:	manager1(),
			manager2(),
			statsObject1( &manager1 ),
			statsObject2( &manager1 )
			{}

	};

	/**
	 * @test odemx::StatisticObject construction and destruction
	 *
	 * Construction effects:
	 * @li correct reset time initialization
	 * @li registration with the given StatisticManager
	 *
	 * Destruction effects:
	 * @li removal from the StatisticManager's list
	 */
	TEST_FIXTURE( StatisticObjectFixture, ConstructionDestruction )
	{
		const std::list<StatisticObject*>& statsList = manager1.getStatisticObjects();

		StatisticObject* stats = new StatisticObject( &manager1 );
		CHECK_EQUAL( &manager1, stats -> getStatisticManager() );
		CHECK_EQUAL( manager1.getTime(), stats -> getResetTime() );
		CHECK( find(statsList.begin(), statsList.end(), stats) != statsList.end() );

		delete stats;
		CHECK( find(statsList.begin(), statsList.end(), stats) == statsList.end() );
	}

	/**
	 * @test odemx::StatisticObject::setStatisticManager(StatisticManager*)
	 *
	 * Test switching of the manager. Expected function call effects:
	 * @li a pointer to the old manager is returned
	 * @li successful removal from the previous StatisticManager
	 * @li successful registration with the given StatisticManager
	 */
	TEST_FIXTURE( StatisticObjectFixture, SetStatisticManager )
	{
		const std::list<StatisticObject*>& statsList = manager1.getStatisticObjects();
		const std::list<StatisticObject*>& newList = manager2.getStatisticObjects();

		StatisticManager* oldManager = statsObject1.setStatisticManager( &manager2 );
		CHECK_EQUAL( oldManager, &manager1 );
		CHECK( find( newList.begin(), newList.end(), &statsObject1 ) != newList.end() );
		CHECK( find( statsList.begin(), statsList.end(), &statsObject1 ) == statsList.end() );
	}

	/**
	 * @test odemx::StatisticObject::reset()
	 *
	 * Expected function call effects:
	 * @li the reset time of the StatisticObject is set to the manager's time
	 */
	TEST_FIXTURE( StatisticObjectFixture, Reset )
	{
		manager1.setTime( 15 );
		statsObject1.reset();
		CHECK_EQUAL( manager1.getTime(), statsObject1.getResetTime() );
	}

	/**
	 * @test odemx::StatisticManager::reset()
	 *
	 * Expected function call effects:
	 * @li the reset time of all registered StatisticObjects is set to
	 * the manager's time
	 */
	TEST_FIXTURE( StatisticObjectFixture, ManagerReset )
	{
		manager1.setTime( 25 );
		manager1.reset();
		CHECK_EQUAL( manager1.getTime(), statsObject1.getResetTime() );
		CHECK_EQUAL( manager1.getTime(), statsObject2.getResetTime() );
		CHECK_EQUAL( statsObject1.getResetTime(), statsObject2.getResetTime() );
	}

	/**
	 * @test odemx::StatisticManager::add(StatisticObject*)
	 *
	 * Expected function call effects:
	 * @li the StatisticObject is now registered with this manager
	 */
	TEST_FIXTURE( StatisticObjectFixture, ManagerAdd )
	{
		const std::list<StatisticObject*>& statsList = manager2.getStatisticObjects();
		manager2.add( &statsObject1 );
		CHECK( find( statsList.begin(), statsList.end(), &statsObject1 ) != statsList.end() );
	}

	/**
	 * @test odemx::StatisticManager::remove(StatisticObject*)
	 *
	 * Expected function call effects:
	 * @li the StatisticObject is not registered with this manager anymore
	 */
	TEST_FIXTURE( StatisticObjectFixture, ManagerRemove )
	{
		const std::list<StatisticObject*>& statsList = manager1.getStatisticObjects();
		manager1.remove( &statsObject1 );
		CHECK( find( statsList.begin(), statsList.end(), &statsObject1 ) == statsList.end() );
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
