/**
 * @file TestTrace.cpp
 * @date Jul 9, 2008
 * @author Ronald Kluth
 *
 * @brief Tests for ODEMx classes Trace, TraceProducer, TraceConsumer
 */

#include "TestUtil.h"

/// @cond DOXYGEN_SKIP
SUITE( Util )
{
/// @endcond

	/**
	 * @struct TraceFixture
	 * @brief Helper struct providing set-up/tear-down of Trace tests
	 *
	 * @copydetails EventFixture
	 */
	struct TraceFixture
	{
		Trace trace;
		TraceProducerTest producer;
		TraceConsumerTest consumer1;
		TraceConsumerTest consumer2;

		TraceFixture()
		:	trace(),
			producer( trace, "TraceProducerTest" ),
			consumer1( false ),
			consumer2( false )
			{}

		void clearConsumerHistory()
		{
			consumer1.history.clear();
			consumer2.history.clear();
		}

		bool consumerHistoryIsEmpty()
		{
			return consumer1.history.empty() && consumer2.history.empty();
		}
	};

	// definitions of static class members
	const MarkType Util::TraceProducerTest::markEvent( "markEvent", 1, typeid( TraceProducerTest ) );
	const MarkType Util::TraceProducerTest::markPartner( "markPartner", 2, typeid( TraceProducerTest ) );
	const MarkType Util::TraceProducerTest::markTwoPartners( "markTwoPartners", 3, typeid( TraceProducerTest ) );
	const MarkType Util::TraceProducerTest::markAttributeChange( "markAttributeChange", 4, typeid( TraceProducerTest ) );
	const MarkType Util::TraceProducerTest::markComposed( "markComposed", 5, typeid( TraceProducerTest ) );
	const Tag Util::TraceProducerTest::tagAttribute( "tagAttribute" );
	const Tag Util::TraceProducerTest::tagComposed( "tagComposed" );

	/**
	 * @test odemx::Trace warning when destroying a running trace
	 *
	 * This test checks for the correct warning when the destructor is
	 * called but the trace has not been stopped with \c stopTrace().
	 *
	 * Expected effect:
	 * @li a warning message is sent to \c errorStream()
	 */
	TEST( TraceConstructionDestruction )
	{
		Trace* tracePtr = new Trace();
		tracePtr -> startTrace();

		clearErrorStream();
		delete tracePtr;
		CHECK_EQUAL( "ODEMx WARNING: ~Trace(): trace was not stopped properly - call stopTrace()\n", \
				errorStream().str() );
	}

	/**
	 * @test odemx::Trace::addConsumer(TraceConsumer*)
	 *
	 * Expected function call effects:
	 * @li error message when adding an object of the wrong type
	 * @li consumer list is not empty
	 * @li the added consumer is found in the list
	 * @li adding consumers to a running trace calls \c startTrace() on the
	 * consumer
	 * @li adding consumers to a paused trace calls \c startTrace() and
	 * \c pauseTrace() on the consumer
	 */
	TEST_FIXTURE( TraceFixture, TraceAddConsumer )
	{
		clearErrorStream();
		trace.addConsumer( static_cast< TraceConsumer* >( 0 ) );
		CHECK_EQUAL( "ODEMx ERROR: Trace::addConsumer(): failed to add TraceConsumer: wrong type\n", \
				errorStream().str() );

		const std::list< TraceConsumer* >& consumerList = trace.getConsumers();
		CHECK( consumerList.empty() );

		TraceConsumerTest dummyConsumer( false );
		trace.addConsumer( &dummyConsumer );

		CHECK( ! consumerList.empty() );
		CHECK( find( consumerList.begin(), consumerList.end(), &dummyConsumer ) \
				!= consumerList.end() );

		trace.startTrace();
		trace.addConsumer( &consumer1 );
		CHECK( consumer1.received( "Start", "Trace", "0", typeToString( typeid( consumer1 ) ) ) );

		trace.pauseTrace();
		trace.addConsumer( &consumer2 );
		CHECK( consumer2.received( "Start", "Trace", "0", typeToString( typeid( consumer2 ) ) ) );
		CHECK( consumer2.received( "Pause", "Trace", "0", typeToString( typeid( consumer2 ) ) ) );
	}

	/**
	 * @test odemx::Trace::removeConsumer(TraceConsumer*)
	 *
	 * Expected function call effects:
	 * @li error message when removing an object of the wrong type
	 * @li consumer list size decreased by one
	 * @li the removed consumer is not found in the list
	 */
	TEST_FIXTURE( TraceFixture, TraceRemoveConsumer )
	{
		clearErrorStream();
		trace.removeConsumer( static_cast< TraceConsumer* >( 0 ) );
		CHECK_EQUAL( "ODEMx ERROR: Trace::removeConsumer(): failed to remove TraceConsumer: wrong type\n", \
				errorStream().str() );

		const std::list< TraceConsumer* >& consumerList = trace.getConsumers();
		CHECK( consumerList.empty() );

		TraceConsumerTest dummyConsumer1( false ),
						  dummyConsumer2( false ),
						  dummyConsumer3( false );

		trace.addConsumer( &dummyConsumer1 );
		trace.addConsumer( &dummyConsumer2 );
		trace.addConsumer( &dummyConsumer3 );
		CHECK_EQUAL( static_cast< size_t >( 3 ), consumerList.size() );

		trace.removeConsumer( &dummyConsumer2 );
		CHECK_EQUAL( static_cast< size_t >( 2 ), consumerList.size() );
		CHECK( find( consumerList.begin(), consumerList.end(), &dummyConsumer2 ) \
				== consumerList.end() );
		CHECK( find( consumerList.begin(), consumerList.end(), &dummyConsumer1 ) \
				!= consumerList.end() );
		CHECK( find( consumerList.begin(), consumerList.end(), &dummyConsumer3 ) \
				!= consumerList.end() );
	}

	/**
	 * @test odemx::Trace::startTrace()
	 *
	 * Expected function call effects:
	 * @li \c startTrace() was called on all registered consumers
	 */
	TEST_FIXTURE( TraceFixture, StartTrace )
	{
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );
		trace.startTrace();
		CHECK( consumer1.received( "Start", "Trace", "0", typeToString( typeid( consumer1 ) ) ) );
		CHECK( consumer2.received( "Start", "Trace", "0", typeToString( typeid( consumer2 ) ) ) );
	}

	/**
	 * @test odemx::Trace::stopTrace()
	 *
	 * Expected function call effects:
	 * @li when trace is not running, the call is not forwarded to registered
	 * consumers
	 * @li a paused trace causes a call to continueTrace() before stopping
	 * @li an open mark will be closed
	 * @li all consumers receive a call to \c stopTrace()
	 */
	TEST_FIXTURE( TraceFixture, StopTrace )
	{
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );
		trace.stopTrace();
		CHECK( consumerHistoryIsEmpty() );

		trace.startTrace();
		trace.beginMark( &producer, MarkType( "", 0, typeid(producer) ) );
		trace.pauseTrace();
		trace.stopTrace();
		CHECK( consumer1.received( "Continue", "Trace", "0", typeToString( typeid( consumer1 ) ) ) );
		CHECK( consumer2.received( "Continue", "Trace", "0", typeToString( typeid( consumer2 ) ) ) );
		CHECK( consumer1.received( "Last Sender", "End Open Mark", "0", typeToString( typeid( consumer1 ) ) ) );
		CHECK( consumer2.received( "Last Sender", "End Open Mark", "0", typeToString( typeid( consumer2 ) ) ) );
		CHECK( consumer1.received( "Stop", "Trace", "0", typeToString( typeid( consumer1 ) ) ) );
		CHECK( consumer2.received( "Stop", "Trace", "0", typeToString( typeid( consumer2 ) ) ) );
	}

	/**
	 * @test odemx::Trace::pauseTrace()
	 *
	 * Expected function call effects:
	 * @li an open mark will be closed
	 * @li all consumers receive a call to \c pauseTrace()
	 * @li calling the function on a non-running trace has no effect
	 */
	TEST_FIXTURE( TraceFixture, PauseTrace )
	{
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		trace.pauseTrace();
		CHECK( consumerHistoryIsEmpty() );

		trace.startTrace();
		trace.beginMark( &producer, MarkType( "", 0, typeid(producer) ) );
		trace.pauseTrace();
		CHECK( consumer1.received( "Pause", "Trace", "0", typeToString( typeid( consumer1 ) ) ) );
		CHECK( consumer2.received( "Pause", "Trace", "0", typeToString( typeid( consumer2 ) ) ) );
		CHECK( consumer1.received( "Last Sender", "End Open Mark", "0", typeToString( typeid( consumer1 ) ) ) );
		CHECK( consumer2.received( "Last Sender", "End Open Mark", "0", typeToString( typeid( consumer2 ) ) ) );

		// check with paused trace
		clearConsumerHistory();
		trace.pauseTrace();
		CHECK( consumerHistoryIsEmpty() );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.pauseTrace();
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test odemx::Trace::continueTrace()
	 *
	 * Expected function call effects:
	 * @li all consumers receive a call to \c continueTrace()
	 * @li calling the function on a non-paused trace has no effect
	 */
	TEST_FIXTURE( TraceFixture, ContinueTrace )
	{
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		// check with created trace
		trace.continueTrace();
		CHECK( consumerHistoryIsEmpty() );

		// check with running trace
		trace.startTrace();
		clearConsumerHistory();
		trace.continueTrace();
		CHECK( consumerHistoryIsEmpty() );

		// check with paused trace
		trace.pauseTrace();
		trace.continueTrace();
		CHECK( consumer1.received( "Continue", "Trace", "0", typeToString( typeid( consumer1 ) ) ) );
		CHECK( consumer2.received( "Continue", "Trace", "0", typeToString( typeid( consumer2 ) ) ) );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.continueTrace();
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test odemx::Trace::beginMark(const TraceProducer*,MarkType)
	 *
	 * A test to check the creation of composed marks.
	 * Expected function call effects:
	 * @li only a running trace forwards the call
	 * @li a previous open mark is closed automatically
	 * @li all consumers receive this call
	 */
	TEST_FIXTURE( TraceFixture, BeginMark )
	{
		// check with created trace
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		trace.beginMark( &producer, TraceProducerTest::markComposed );
		CHECK( consumerHistoryIsEmpty() );

		// check with running trace
		trace.startTrace();
		// open another mark, which should be closed automatically
		trace.beginMark( &producer, MarkType( "", 0, typeid(producer) ) );
		// this is the actual testing mark
		trace.beginMark( &producer, TraceProducerTest::markComposed );
		// check closure of old mark
		CHECK( consumer1.received( "Last Sender", "End Open Mark", "0", typeToString( typeid( consumer1 ) ) ) );
		CHECK( consumer2.received( "Last Sender", "End Open Mark", "0", typeToString( typeid( consumer2 ) ) ) );
		// check reception of new mark
		CHECK( consumer1.received( producer.getLabel(), "markComposed", "5", typeToString( typeid( producer ) ), \
				"Begin Open Mark", "" , "" ) );
		CHECK( consumer2.received( producer.getLabel(), "markComposed", "5", typeToString( typeid( producer ) ), \
				"Begin Open Mark", "" , "" ) );

		// check with paused trace
		trace.pauseTrace();
		clearConsumerHistory();
		trace.beginMark( &producer, TraceProducerTest::markComposed );
		CHECK( consumerHistoryIsEmpty() );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.beginMark( &producer, TraceProducerTest::markComposed );
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test odemx::Trace::endMark(const string&)
	 *
	 * A test to check the closing of composed marks.
	 * Expected function call effects:
	 * @li only a running trace forwards the call
	 * @li if there is no open mark, an error is sent to \c errorStream()
	 * @li the previously open mark is closed
	 * @li all consumers receive this call
	 */
	TEST_FIXTURE( TraceFixture, EndMark )
	{
		// check with created trace
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		trace.endMark( "composed mark comment" );
		CHECK( consumerHistoryIsEmpty() );

		// check with running trace
		trace.startTrace();

		// call without open mark
		clearErrorStream();
		trace.endMark( "composed mark comment" );
		CHECK_EQUAL( "ODEMx ERROR: Trace::endMark(): called without a corresponding beginMark()\n", \
				errorStream().str() );

		// open another mark, which should be closed
		trace.beginMark( &producer, TraceProducerTest::markComposed );
		trace.endMark( "composed mark comment" );

		// check closure of old mark
		CHECK( consumer1.received( "Last Sender", "End Open Mark", "0", typeToString( typeid( consumer1 ) ), \
				"", "", "composed mark comment" ) );
		CHECK( consumer2.received( "Last Sender", "End Open Mark", "0", typeToString( typeid( consumer2 ) ), \
				"", "", "composed mark comment" ) );

		// check with paused trace
		trace.pauseTrace();
		clearConsumerHistory();
		trace.endMark( "composed mark comment" );
		CHECK( consumerHistoryIsEmpty() );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.endMark( "composed mark comment" );
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test odemx::Trace::mark(const TraceProducer*,MarkType,const string&)
	 *
	 * A simple trace mark test. Expected function call effects:
	 * @li no effects on non-running trace
	 * @li when trace is running, the mark is forwarded to all consumers
	 */
	TEST_FIXTURE( TraceFixture, SimpleMark )
	{
		// check with created trace
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		// check with created trace
		trace.mark( &producer, TraceProducerTest::markEvent, "test comment" );
		CHECK( consumerHistoryIsEmpty() );

		// check with running trace
		trace.startTrace();
		trace.mark( &producer, TraceProducerTest::markEvent, "test comment" );
		CHECK( consumer1.received( producer.getLabel(), "markEvent", "1", typeToString( typeid( producer ) ), \
				"", "", "test comment" ) );
		CHECK( consumer2.received( producer.getLabel(), "markEvent", "1", typeToString( typeid( producer ) ), \
				"", "", "test comment" ) );

		// check with paused trace
		trace.pauseTrace();
		clearConsumerHistory();
		trace.mark( &producer, TraceProducerTest::markEvent, "test comment" );
		CHECK( consumerHistoryIsEmpty() );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.mark( &producer, TraceProducerTest::markEvent, "test comment" );
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test odemx::Trace::mark(const TraceProducer*,MarkType,const TraceProducer*,const string&)
	 *
	 * A simple trace mark test. Expected function call effects:
	 * @li no effects on non-running trace
	 * @li when trace is running, the mark is forwarded to all consumers
	 */
	TEST_FIXTURE( TraceFixture, SimpleMarkWithPartner )
	{
		TraceProducerTest dummyPartner( trace, "TraceProducerDummyPartner" );
		// check with created trace
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		// check with created trace
		trace.mark( &producer, TraceProducerTest::markPartner, &dummyPartner, "test comment" );
		CHECK( consumerHistoryIsEmpty() );

		// check with running trace
		trace.startTrace();
		trace.mark( &producer, TraceProducerTest::markPartner, &dummyPartner, "test comment" );
		CHECK( consumer1.received( producer.getLabel(), "markPartner", "2", typeToString( typeid( producer ) ), \
				dummyPartner.getLabel(), "", "test comment" ) );
		CHECK( consumer2.received( producer.getLabel(), "markPartner", "2", typeToString( typeid( producer ) ), \
				dummyPartner.getLabel(), "", "test comment" ) );

		// check with paused trace
		trace.pauseTrace();
		clearConsumerHistory();
		trace.mark( &producer, TraceProducerTest::markPartner, &dummyPartner, "test comment" );
		CHECK( consumerHistoryIsEmpty() );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.mark( &producer, TraceProducerTest::markPartner, &dummyPartner, "test comment" );
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test odemx::Trace::mark(const TraceProducer*,MarkType,const TraceProducer*,const TraceProducer*,const string&)
	 *
	 * A simple trace mark test. Expected function call effects:
	 * @li no effects on non-running trace
	 * @li when trace is running, the mark is forwarded to all consumers
	 */
	TEST_FIXTURE( TraceFixture, SimpleMarkWithTwoPartners )
	{
		TraceProducerTest dummyPartner1( trace, "TraceProducerDummyPartner1" );
		TraceProducerTest dummyPartner2( trace, "TraceProducerDummyPartner2" );
		// check with created trace
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		// check with created trace
		trace.mark( &producer, TraceProducerTest::markTwoPartners, &dummyPartner1, &dummyPartner2, "test comment" );
		CHECK( consumerHistoryIsEmpty() );

		// check with running trace
		trace.startTrace();
		trace.mark( &producer, TraceProducerTest::markTwoPartners, &dummyPartner1, &dummyPartner2, "test comment" );
		CHECK( consumer1.received( producer.getLabel(), "markTwoPartners", "3", typeToString( typeid( producer ) ), \
				dummyPartner1.getLabel(), dummyPartner2.getLabel(), "test comment" ) );
		CHECK( consumer2.received( producer.getLabel(), "markTwoPartners", "3", typeToString( typeid( producer ) ), \
				dummyPartner1.getLabel(), dummyPartner2.getLabel(), "test comment" ) );

		// check with paused trace
		trace.pauseTrace();
		clearConsumerHistory();
		trace.mark( &producer, TraceProducerTest::markTwoPartners, &dummyPartner1, &dummyPartner2, "test comment" );
		CHECK( consumerHistoryIsEmpty() );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.mark( &producer, TraceProducerTest::markTwoPartners, &dummyPartner1, &dummyPartner2, "test comment" );
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test template odemx::Trace::mark(const TraceProducer*,MarkType,AttributeType,AttributeType,const string&)
	 *
	 * A simple trace mark test. Expected function call effects:
	 * @li no effects on non-running trace
	 * @li when trace is running, the mark is forwarded to all consumers
	 */
	TEST_FIXTURE( TraceFixture, SimpleMarkAttributeChange )
	{
		// check with created trace
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		// check with created trace
		trace.mark( &producer, TraceProducerTest::markAttributeChange, true, false );
		CHECK( consumerHistoryIsEmpty() );

		// check with running trace
		trace.startTrace();
		trace.mark( &producer, TraceProducerTest::markAttributeChange, true, false );
		CHECK( consumer1.received( producer.getLabel(), "markAttributeChange", "4", typeToString( typeid( producer ) ), \
				toString( true ), toString( false ), typeToString( typeid(bool) ) ) );
		CHECK( consumer2.received( producer.getLabel(), "markAttributeChange", "4", typeToString( typeid( producer ) ), \
				toString( true ), toString( false ), typeToString( typeid(bool) ) ) );

		// check with paused trace
		trace.pauseTrace();
		clearConsumerHistory();
		trace.mark( &producer, TraceProducerTest::markAttributeChange, true, false );
		CHECK( consumerHistoryIsEmpty() );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.mark( &producer, TraceProducerTest::markAttributeChange, true, false );
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test template odemx::Trace::mark(const TraceProducer*,MarkType,AttributeType,AttributeType,const string&)
	 *
	 * A simple trace mark test which checks TraceConsumer functions.
	 * Expected function call effects:
	 * @li when trace is running, the mark is forwarded to all consumers
	 * @li marks of all C++ built-in types are received by the TraceConsumers
	 */
	TEST_FIXTURE( TraceFixture, SimpleMarkAttributeTypes )
	{
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		Label sender = producer.getLabel();
		const std::string markType = "markAttributeChange";
		const std::string id = "4";
		const std::string scope = typeToString( typeid( producer ) );
		std::string type;

		trace.startTrace();
		type = typeToString( typeid(bool) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, true, false );
		CHECK( consumer1.received( sender, markType, id, scope,	toString( true ), toString( false ), type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	toString( true ), toString( false ), type ) );

		type = typeToString( typeid(char) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, 'a', 'b' );
		CHECK( consumer1.received( sender, markType, id, scope,	"a", "b", type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	"a", "b", type ) );

		type = typeToString( typeid(short) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, MIN_VAL(short), MAX_VAL(short) );
		CHECK( consumer1.received( sender, markType, id, scope,	toString( MIN_VAL(short) ), toString( MAX_VAL(short) ), type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	toString( MIN_VAL(short) ), toString( MAX_VAL(short) ), type ) );

		type = typeToString( typeid(int) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, MIN_VAL(int), MAX_VAL(int) );
		CHECK( consumer1.received( sender, markType, id, scope,	toString( MIN_VAL(int) ), toString( MAX_VAL(int) ), type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	toString( MIN_VAL(int) ), toString( MAX_VAL(int) ), type ) );

		type = typeToString( typeid(long) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, MIN_VAL(long), MAX_VAL(long) );
		CHECK( consumer1.received( sender, markType, id, scope,	toString( MIN_VAL(long) ), toString( MAX_VAL(long) ), type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	toString( MIN_VAL(long) ), toString( MAX_VAL(long) ), type ) );

		type = typeToString( typeid(long long) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, MIN_VAL(long long), MAX_VAL(long long) );
		CHECK( consumer1.received( sender, markType, id, scope,	toString( MIN_VAL(long long) ), toString( MAX_VAL(long long) ), type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	toString( MIN_VAL(long long) ), toString( MAX_VAL(long long) ), type ) );

		type = typeToString( typeid(float) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, MIN_VAL(float), MAX_VAL(float) );
		CHECK( consumer1.received( sender, markType, id, scope,	toString( MIN_VAL(float) ), toString( MAX_VAL(float) ), type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	toString( MIN_VAL(float) ), toString( MAX_VAL(float) ), type ) );

		type = typeToString( typeid(double) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, MIN_VAL(double), MAX_VAL(double) );
		CHECK( consumer1.received( sender, markType, id, scope,	toString( MIN_VAL(double) ), toString( MAX_VAL(double) ), type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	toString( MIN_VAL(double) ), toString( MAX_VAL(double) ), type ) );

		type = typeToString( typeid(const char*) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, "const char*", "const std::string&" );
		CHECK( consumer1.received( sender, markType, id, scope,	"const char*", "const std::string&", type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	"const char*", "const std::string&", type ) );

		type = typeToString( typeid(unsigned char) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, (unsigned char)'^', (unsigned char)'#' );
		CHECK( consumer1.received( sender, markType, id, scope,	"^", "#", type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	"^", "#", type ) );

		type = typeToString( typeid(unsigned short) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, MIN_VAL(unsigned short), MAX_VAL(unsigned short) );
		CHECK( consumer1.received( sender, markType, id, scope,	toString( MIN_VAL(unsigned short) ), toString( MAX_VAL(unsigned short) ), type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	toString( MIN_VAL(unsigned short) ), toString( MAX_VAL(unsigned short) ), type ) );

		type = typeToString( typeid(unsigned int) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, MIN_VAL(unsigned int), MAX_VAL(unsigned int) );
		CHECK( consumer1.received( sender, markType, id, scope,	toString( MIN_VAL(unsigned int) ), toString( MAX_VAL(unsigned int) ), type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	toString( MIN_VAL(unsigned int) ), toString( MAX_VAL(unsigned int) ), type ) );

		type = typeToString( typeid(unsigned long) );
		trace.mark( &producer, TraceProducerTest::markAttributeChange, MIN_VAL(unsigned long), MAX_VAL(unsigned long) );
		CHECK( consumer1.received( sender, markType, id, scope,	toString( MIN_VAL(unsigned long) ), toString( MAX_VAL(unsigned long) ), type ) );
		CHECK( consumer2.received( sender, markType, id, scope,	toString( MIN_VAL(unsigned long) ), toString( MAX_VAL(unsigned long) ), type ) );
	}

	/**
	 * @test odemx::Trace::beginTag(Tag)
	 *
	 * A test to check the creation of composed tags.
	 * Expected function call effects:
	 * @li only a running trace forwards the call
	 * @li if there is no open mark, an error is sent to \c errorStream()
	 * @li all consumers receive this call
	 */
	TEST_FIXTURE( TraceFixture, BeginTag )
	{
		// check with created trace
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		trace.beginTag( TraceProducerTest::tagComposed );
		CHECK( consumerHistoryIsEmpty() );

		// check with running trace
		trace.startTrace();

		// call without open mark
		clearErrorStream();
		trace.beginTag( TraceProducerTest::tagComposed );
		CHECK_EQUAL( "ODEMx ERROR: Trace::beginTag(): called without an open composed mark\n", \
				errorStream().str() );

		// open a mark
		trace.beginMark( &producer, TraceProducerTest::markComposed );
		trace.beginTag( TraceProducerTest::tagComposed );

		// check opening of new tag
		CHECK( consumer1.received( "Begin", "Composed Tag", "0", typeToString( typeid( consumer1 ) ), \
				"", "" , "" ) );
		CHECK( consumer2.received( "Begin", "Composed Tag", "0", typeToString( typeid( consumer2 ) ), \
				"", "" , "" ) );

		// check with paused trace
		trace.pauseTrace();
		clearConsumerHistory();
		trace.beginTag( TraceProducerTest::tagComposed );
		CHECK( consumerHistoryIsEmpty() );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.beginTag( TraceProducerTest::tagComposed );
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test odemx::Trace::endTag(Tag)
	 *
	 * A test to check the creation of composed tags.
	 * Expected function call effects:
	 * @li only a running trace forwards the call
	 * @li if there is no open mark, an error is sent to \c errorStream()
	 * @li all consumers receive this call
	 */
	TEST_FIXTURE( TraceFixture, EndTag )
	{
		// check with created trace
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		trace.endTag( TraceProducerTest::tagComposed );
		CHECK( consumerHistoryIsEmpty() );

		// check with running trace
		trace.startTrace();

		// call without open mark
		clearErrorStream();
		trace.endTag( TraceProducerTest::tagComposed );
		CHECK_EQUAL( "ODEMx ERROR: Trace::endTag(): called without an open composed mark\n", \
				errorStream().str() );

		// open a mark
		trace.beginMark( &producer, TraceProducerTest::markComposed );
		trace.endTag( TraceProducerTest::tagComposed );

		// check closure of old tag
		CHECK( consumer1.received( "End", "Composed Tag", "0", typeToString( typeid( consumer1 ) ), \
				"", "" , "" ) );
		CHECK( consumer2.received( "End", "Composed Tag", "0", typeToString( typeid( consumer2 ) ), \
				"", "" , "" ) );

		// check with paused trace
		trace.pauseTrace();
		clearConsumerHistory();
		trace.endTag( TraceProducerTest::tagComposed );
		CHECK( consumerHistoryIsEmpty() );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.endTag( TraceProducerTest::tagComposed );
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test template odemx::Trace::addTag(Tag, TagValueType)
	 *
	 * A test to check the creation of composed tags.
	 * Expected function call effects:
	 * @li only a running trace forwards the call
	 * @li if there is no open mark, an error is sent to \c errorStream()
	 * @li all consumers receive this call
	 */
	TEST_FIXTURE( TraceFixture, AddTag )
	{
		// check with created trace
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );

		trace.addTag( TraceProducerTest::tagAttribute, true );
		CHECK( consumerHistoryIsEmpty() );

		// check with running trace
		trace.startTrace();

		// call without open mark
		clearErrorStream();
		trace.addTag( TraceProducerTest::tagAttribute, true );
		CHECK_EQUAL( "ODEMx ERROR: Trace::addTag<>(): called without an open composed mark\n", \
				errorStream().str() );

		// open a mark
		trace.beginMark( &producer, TraceProducerTest::markComposed );
		trace.addTag( TraceProducerTest::tagAttribute, true );

		// check reception of new tag
		CHECK( consumer1.received( "Tag", "Value", "0", typeToString( typeid( true ) ), \
				toString( true ) ) );
		CHECK( consumer2.received( "Tag", "Value", "0", typeToString( typeid( true ) ), \
				toString( true ) ) );

		// check with paused trace
		trace.pauseTrace();
		clearConsumerHistory();
		trace.endTag( TraceProducerTest::tagComposed );
		CHECK( consumerHistoryIsEmpty() );

		// check with stopped trace
		trace.stopTrace();
		clearConsumerHistory();
		trace.endTag( TraceProducerTest::tagComposed );
		CHECK( consumerHistoryIsEmpty() );
	}

	/**
	 * @test template odemx::Trace::addTag(Tag, TagValueType)
	 *
	 * A simple trace mark test which checks TraceConsumer functions.
	 * Expected function call effects:
	 * @li when trace is running, the mark is forwarded to all consumers
	 * @li tags of all C++ built-in types are received by the TraceConsumers
	 */
	TEST_FIXTURE( TraceFixture, AddTagValueTypes )
	{
		trace.addConsumer( &consumer1 );
		trace.addConsumer( &consumer2 );
		std::string type;
		trace.startTrace();
		trace.beginMark( &producer, TraceProducerTest::markComposed );

		{
			bool value = true;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			char value = '?';
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			short value = 1;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			int value = 1;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			long value = 1;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			long long value = 1;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			float value = 1.1f;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			double value = 1.1;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			const char* value = "value";
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			const TraceProducer* value = &producer;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			unsigned char value = '1';
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			unsigned short value = 1;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			unsigned int value = 1;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}

		{
			unsigned long value = 1;
			type = typeToString( typeid( value ) );
			trace.addTag( TraceProducerTest::tagAttribute, value );
			CHECK( consumer1.received( "Tag", "Value", "0", type, toString( value ) ) );
			CHECK( consumer2.received( "Tag", "Value", "0", type, toString( value ) ) );
		}
	}

/// @cond DOXYGEN_SKIP
}
/// @endcond
