/**
 * @file TestUtil.h
 *
 * @brief Contains all helper classes required for testing ODEMx module \b %Util.
 */

#ifndef TESTUTIL_H_
#define TESTUTIL_H_

#include <UnitTest++.h>
#include "../Main/Globals.h"

#ifndef _DEBUG
#define _DEBUG
#include <odemx/util/ErrorHandling.h>
#undef _DEBUG
#endif

#include <odemx/base/Simulation.h>
#include <odemx/util/LabeledObject.h>
#include <odemx/util/MarkType.h>
#include <odemx/util/Observable.h>
#include <odemx/util/StatisticObject.h>
#include <odemx/util/Report.h>
#include <odemx/util/Trace.h>
#include <odemx/util/Translate.h>
#include <odemx/util/TypedObject.h>
#include <odemx/util/Version.h>

#include <algorithm>
#include <iostream>
#include <map>
#include <string>

#if defined(_MSC_VER)
#undef min
#undef max
#endif
#include <limits>

#define MIN_VAL(Type) ((Type)std::numeric_limits<Type>::min())
#define MAX_VAL(Type) ((Type)std::numeric_limits<Type>::max())


/**
 * @namespace Util
 *
 * @brief Contains helpers and tests for module ODEMx %Util

 * The UnitTest++ macro SUITE( Util ) declares namespace %Util.
 * Consequently, all helpers for this test suite should be placed in
 * the same namespace.
 */
namespace Util {

	using namespace odemx;
	using namespace std;

	/*********************** LABELED OBJECT ***********************************/

	/**
	 * @class LabelScopeTest
	 * @brief LabelScope subclass providing a public interface for data members
	 */
	class LabelScopeTest:
		public LabelScope
	{
	public:
		/// Constructor for default initialization
		LabelScopeTest(): LabelScope() {}
		/// Constructor for user-defined initialization
		LabelScopeTest( const std::string& defLabel, char defSpace ):
			LabelScope( defLabel, defSpace )
			{}

		/// get the space character
		char getSpace() { return spc; }
		/// get the default label for unnamed objects
		const string& getDefLabel() { return def; }
		/// get the lookup map which contains all the scope's labels, with counter
		const map< string, unsigned int >& getLookupMap() { return lookup; }
	};

	/**
	 * @class DefLabeledObjectTest
	 * @brief DefLabeledObject subclass providing a public interface for setting the label
	 */
	class DefLabeledObjectTest:
		public DefLabeledObject
	{
	public:
		/// Constructor for default initialization
		DefLabeledObjectTest(): DefLabeledObject() {}
		/// Constructor for user-defined initialization
		DefLabeledObjectTest(LabelScope* scope, const std::string& label)
		:	DefLabeledObject( scope, label )
		{}

		/// makes the protected base class method public for testing
		void setLabel(LabelScope* scope, const std::string& label)
		{
			DefLabeledObject::setLabel( scope, label );
		}
	};

	/************************* MARKTYPE ***************************************/

	/**
	 * @class MarkTypeTest
	 * @brief A class for testing static const MarkType and Tag class members
	 */
	class MarkTypeTest
	{
	public:
		static const MarkTypeId baseMarkId; ///< base ID for mark types
		static const MarkType mark1; ///< a test mark type
		static const MarkType mark2; ///< a test mark type
		static const TagId baseTagId; ///< base ID for tag types
		static const Tag tag1; ///< a test tag type
	};

	/************************* OBSERVABLE *************************************/

	/**
	 * @class ObservableTestObserver
	 * @brief An observer class for testing template Observable
	 */
	template < typename T = int >
	class ObservableTestObserver:
		public ObserverBase
	{
	public:
		/**
		 * @brief Constructor
		 * @param output determines whether observations are sent to stdout
		 */
		ObservableTestObserver( bool output )
		:	ObserverBase( output )
			{}
		/// the method to be called by the event macro _obsForEach
		void onExecute( DefLabeledObject* sender )
		{
			ObserverBase::saw( "onExecute", sender );
		}
		/// the method to be called by the attribute change macro _obsAForEach
		void onChangeMember( DefLabeledObject* sender, T oldValue, T newValue )
		{
			ObserverBase::saw( "onChangeMember", sender, toString( oldValue ), toString( newValue ) );
		}
	};

	/**
	 * @class ObservableTest
	 * @brief A subclass for testing ODEMx observation macros
	 */
	class ObservableTest:
		public Observable< ObservableTestObserver<> >,
		public DefLabeledObject
	{
	public:
		int member; ///< data member monitored for attribute change

		/// Constructor
		ObservableTest( Label l, int m, ObservableTestObserver<>* o )
		:	Observable< ObservableTestObserver<> >( o ),
			DefLabeledObject( 0, l ),
			member( m )
			{}
		/// a method to test the observer event macro _obsForEach
		virtual void execute()
		{
			_obsForEach( ObservableTestObserver<>, Execute( this ) );
		}
		/// a method to test the attribute change macro _obsAForEach
		void setMember( int newValue )
		{
			_obsAForEach( ObservableTestObserver<>, Member, member, newValue );
			member = newValue;
		}
	};

	/**
	 * @class ObservableTestT
	 * @brief A templated subclass for testing ODEMx template observation macros
	 *
	 * @tparam T the type of the data member \c t_member
	 */
	template < typename T >
	class ObservableTestT:
		public Observable< ObservableTestObserver< T > >,
		public DefLabeledObject
	{
	public:
		T t_member; ///< data member monitored for attribute change

		/// Constructor
		ObservableTestT( Label l, T t, ObservableTestObserver< T >* o )
		:	Observable< ObservableTestObserver< T > >( o ),
			DefLabeledObject( 0, l ),
			t_member( t )
		{}
		/// a method to test the template observer event macro _obsForEachT
		virtual void execute()
		{
			_obsForEachT( ObservableTestObserver< T >, Execute( this ) );
		}
		/// a method to test the template attribute change macro _obsAForEachT
		void setMember( T newValue )
		{
			_obsAForEachT( ObservableTestObserver< T >, Member, t_member, newValue );
			t_member = newValue;
		}
	};

	/************************* TABLE ******************************************/

	/**
	 * @class ColumnTest
	 * @brief A subclass for testing ODEMx table columns
	 */
	template < typename DataType >
	class ColumnTest:
		public ColumnT< DataType >
	{
	public:
		/// Constructor
		ColumnTest( const Label& l, Column::Type t )
		:	ColumnT< DataType >( l, t )
		{}
		/// append data to column, i.e. add content of next line in table
		void addData( DataType d ) { ColumnT< DataType >::addData( d ); }
		/// read data from column, the given line is range-checked
		DataType getData( size_t line ) { return ColumnT< DataType >::getData( line ); }
	};

	/************************* REPORT *****************************************/

	/**
	 * @class ReportProducerTest
	 * @brief A subclass for testing the ODEMx report mechanism
	 */
	class ReportProducerTest:
		public ReportProducer
	{
	public:
		bool sentReport; ///< monitors whether report() was called

		/// Constructor
		ReportProducerTest()
		:	sentReport( false )
		{}

		/**
		 * @brief implementation of pure virtual function, sets \c sentReport
		 * to \c true
		 * @param r the Report object to send the data to
		 */
		virtual void report( Report* r )
		{
			sentReport = true;
		}
	};

	/**
	 * @class ReportTest
	 * @brief A subclass for testing the ODEMx report mechanism
	 */
	class ReportTest:
		public Report
	{
	public:
		bool processedTables; ///< monitors whether processTables() was called
		std::vector< std::vector< std::string > > allTableContent;

		/// Constructor
		ReportTest()
		:	processedTables( false )
		{}
		/**
		 * @brief get a list of the associated ReportProducers
		 * @return a reference to the internal list
		 */
		std::list< ReportProducer* >& getReportProducers()
		{
			return Report::rs;
		}
		/**
		 * @brief get a list of the created tables
		 * @return a reference to the internal list
		 */
		std::vector< Table* >& getReportTables()
		{
			return Report::ts;
		}
		/**
		 * @brief implementation of pure virtual function, sets \c
		 * processedTables to \c true
		 *
		 * Aside from testing odemx::Report, this function is also used for
		 * checking all ReportProducer::report() results
		 */
		virtual void processTables()
		{
			processedTables = true;

			Table* tab = 0;
			std::vector< Table* >::iterator it;

			// iterate over the tables of all ReportProducers
			for( it = ts.begin(); it != ts.end(); ++it )
			{
				std::vector< std::string > tableContent;

				tab = *it;

				if( tab == 0 )
				{
					error( "ReportTest::processTables(): Table pointer is 0" );
					continue;
				}

				// write cells to XML
				for ( size_t line = 0; line < tab -> getNumberOfLines(); ++line )
				{
					for( size_t col = 0; col < tab -> getNumberOfColumns(); ++col )
					{
						tableContent.push_back( translate( tab -> getSTRING( col, line ) ) );
					}
				}

				allTableContent.push_back( tableContent );
			}
		}
	};

	/************************ STATISTIC OBJECT ********************************/

	/**
	 * @class StatisticManagerTest
	 * @brief A class used for testing ODEMx base classes for statistics
	 */
	class StatisticManagerTest:
		public StatisticManager
	{
	public:
		SimTime time; ///< time value that is used as reset time

		/// Constructor
		StatisticManagerTest()
		:	time( 0 )
		{}

		/// get a list of the registered StatisticObjects
		const std::list<StatisticObject*>& getStatisticObjects() { return so; }
		/// set time, used for testing different reset times
		void setTime( SimTime t ) { time = t; }
		/// get time, used to provide the rest time to StatisticObjects
		virtual SimTime getTime() const { return time; }
	};

	/*************************** TRACE ****************************************/

	/**
	 * @class TraceConsumerTest
	 * @brief A class used for testing ODEMx trace
	 */
	class TraceProducerTest:
		public TraceProducer
	{
	public:
		Trace& trace; ///< reference to trace object
		Label label; ///< label of this object

		/// Constructor
		TraceProducerTest( Trace& t, Label l )
		:	trace( t ),
			label( l )
		{}

		/// required implementation of pure virtual function
		virtual Trace* getTrace() const { return &trace; }
		/// required implementation of pure virtual function
		virtual Label getLabel() const { return label; }

		static const MarkType markEvent; ///< MarkType for simple mark without details
		static const MarkType markPartner; ///< MarkType for simple mark with partner
		static const MarkType markTwoPartners; ///< MarkType for simple mark with two partners
		static const MarkType markAttributeChange; ///< MarkType for simple mark with attribute change
		static const MarkType markComposed; ///< MarkType for composed marks
		static const Tag tagAttribute;
		static const Tag tagComposed;
	};

	/**
	 * @class TraceConsumerTest
	 * @brief A class for testing ODEMx trace consumers
	 */
	class TraceConsumerTest:
		public TraceConsumer,
		public TraceBase
	{
	public:
		/**
		 * @brief Constructor
		 * @param output set to \c true to output observed events to the console
		 */
		TraceConsumerTest( bool output )
		:	TraceConsumer(),
			TraceBase( output )
		{}

		/**
		 * @name Mark consumption
		 *
		 * All of the following functions are required to implemented because
		 * of the TraceConsumer interface. They all work the same way in this
		 * class: the function call is logged in the history of TraceBase in
		 * the form of a TraceMark, which is composed of strings containing all
		 * information to uniquely identify each function call during testing.
		 *
		 * @{
		 */

		// some of the following implementations are hacks using TraceMarks
		// to store tags as these are composed of strings anyway
		virtual void startTrace()
		{ TraceBase::mark( TraceMark( "Start", MarkType( "Trace", 0, typeid( *this ) ), "", "" , "" ) ); }
		virtual void stopTrace()
		{ TraceBase::mark( TraceMark( "Stop", MarkType( "Trace", 0, typeid( *this ) ), "", "" , "" ) ); }
		virtual void pauseTrace()
		{ TraceBase::mark( TraceMark( "Pause", MarkType( "Trace", 0, typeid( *this ) ), "", "" , "" ) ); }
		virtual void continueTrace()
		{ TraceBase::mark( TraceMark( "Continue", MarkType( "Trace", 0, typeid( *this ) ), "", "" , "" ) ); }

		virtual void beginMark(const TraceProducer* sender, const MarkType* m)
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, "Begin Open Mark", "" , "" ) ); }
		virtual void endMark(const std::string& comment = "")
		{ TraceBase::mark( TraceMark( "Last Sender", MarkType( "End Open Mark", 0, typeid( *this ) ), "", "" , comment ) ); }

		virtual void mark(const TraceProducer* sender, const MarkType* m, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, "", "" , comment ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, const TraceProducer* partner, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, partner -> getLabel(), "" , comment ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, const TraceProducer* firstPartner, const TraceProducer* secondPartner, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, firstPartner -> getLabel(), secondPartner -> getLabel() , comment ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, bool newValue, bool oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(bool) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, char newValue, char oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(char) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, short newValue, short oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(short) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, int newValue, int oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(int) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, long newValue, long oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(long) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, long long newValue, long long oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(long long) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, float newValue, float oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(float) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, double newValue, double oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(double) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, const char* newValue, const char* oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(const char*) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned char newValue, unsigned char oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(unsigned char) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned short newValue, unsigned short oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(unsigned short) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned int newValue, unsigned int oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(unsigned int) ) ) ); }
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned long newValue, unsigned long oldValue, const std::string& comment = "")
		{ TraceBase::mark( TraceMark( sender -> getLabel(), *m, toString( newValue ), toString( oldValue ) , typeToString( typeid(unsigned long) ) ) ); }

		virtual void beginTag(Tag t)
		{ TraceBase::mark( TraceMark( "Begin", MarkType( "Composed Tag", 0, typeid( *this ) ), "", "" , "" ) ); }
		virtual void endTag(Tag t)
		{ TraceBase::mark( TraceMark( "End", MarkType( "Composed Tag", 0, typeid( *this ) ), "", "" , "" ) ); }

		virtual void addTag(Tag t, bool value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, char value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, short value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, int value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, long value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, long long value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, float value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, double value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, const char* value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, const TraceProducer* value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, unsigned char value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, unsigned short value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, unsigned int value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		virtual void addTag(Tag t, unsigned long value)
		{ TraceBase::mark( TraceMark( "Tag", MarkType( "Value", 0, typeid( value ) ), toString( value ), "" , "" ) ); }
		//@}
	};

} // namespace Util

#endif /* TESTUTIL_H_ */
