//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file ErrorHandling.h

	\author Ralf Gerstenberger
	<!-- [\author <author>]* -->

	\date created at 2002/01/22

	\brief Declaration of error handler functions

	\sa ErrorHandling.cpp

	<!-- [detailed description] -->

	

	\since 1.0
*/

#ifndef ODEMX_ERRORHANDLING_INCLUDED
#define ODEMX_ERRORHANDLING_INCLUDED

#include <string>

namespace odemx {

	// forward declaration
	class LabeledObject;

	/**
		\ingroup util

		\brief %debug() writes debug messages in debug mode
		\param message
			debug message as const string&

		\note use the macro debugOut(message) instead,
		to skip the function call in the release build
	*/
	void debug( const std::string& message ) ;

	/**
		\ingroup util

		\brief %debug() writes debug messages in debug mode
		\param sender
			the object that sent this message
		\param message
			debug message as const string&

		\note use the macro debugOutLabel(obj,message) instead,
		to skip the function call in the release build
	*/
	void debug( const LabeledObject* sender, const std::string& message ) ;

#ifdef _DEBUG
#define debugOut(message) debug(message)
#define debugOutLabel(obj,message) debug(obj,message)
#else
#define debugOut(message)
#define debugOutLabel(obj,message)
#endif

	/**
		\ingroup util

		\brief %warning() writes the message to \c stderr and continues;
		\param message
			warning description as const string&

		This function is used to inform the user about unusual events,
		that indicate a user error.
	*/
	void warning( const std::string& message ) ;

	/**
		\ingroup util

		\brief %warning() writes the message to \c stderr and continues;
		\param sender
			the object that caused the warning
		\param message
			warning description as const string&

		This function is used to inform the user about unusual events,
		that indicate a user error.
	*/
	void warning( const LabeledObject* sender, const std::string& message ) ;

	/** 
		\ingroup util

		\brief %error() writes the message to \c stderr and continues;
		\param message
			error description as const string&

		This function reports errors, caused by wrong usage and internal errors,
		the system could handle. The behaviour of the simulation could be wrong
		after an error() is reported.
	*/
	void error( const std::string& message ) ;

	/** 
		\ingroup util

		\brief %error() writes the message to \c stderr and continues;
		\param sender
			the object that caused the warning
		\param message
			error description as const string&

		This function reports errors, caused by wrong usage and internal errors
		the system could handle. The behaviour of the simulation could be wrong
		after an error() is reported.
	*/
	void error( const LabeledObject* sender, const std::string& message ) ;

	/**
		\ingroup util

		\brief %fatalError() writes the message to \c stderr and terminates the program;

		This function reports errors, caused by wrong usage and internal errors,
		the system can't handle.

		\param message
			error description as const string&
		\param exitValue
			exit value of program
	*/
	void fatalError( const std::string& message, int exitValue ) ;
	
	/**
		\ingroup util

		\brief %fatalError() writes the message to \c stderr and terminates the program;

		This function reports errors, caused by wrong usage and internal errors,
		the system can't handle.

		\param message
			error description as const string&
		\param exitValue
			exit value of program
	*/
	void fatalError( const LabeledObject* sender, const std::string& message, int exitValue ) ;
}

#endif

