//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004, 2007 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file MarkType.h

	\brief Declaration of trace mark types

	\copydetails Trace.h
*/

#ifndef ODEMX_MARKTYPE_INCLUDED
#define ODEMX_MARKTYPE_INCLUDED

#include <string>
#include <map>
#include <list>
#include <cassert>

namespace odemx {

	// forward declaration
	class MarkType;

	/// Mark type information
	typedef unsigned long MarkTypeId;

	/// Mark type list
	typedef std::list< MarkType* > MarkTypeList;

	/// Mark type information
	typedef std::map< std::string, MarkTypeId > MarkScopeIdMap;

	/// Tag identification
	typedef unsigned long TagId;

	/** \class MarkType

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief MarkType describes a trace mark.

		<!-- [\note {notes}]* -->

		\sa Trace TraceProducer TraceConsumer Tag

		A mark is defined by its MarkType.
		The MarkType contains a string (name) and a number (id).
		Furthermore, a MarkType is always associated to a definition scope.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class MarkType {
	public:

		/**
			\brief Constructor

			\param n
				name of MarkType
			\param i
				id of MarkType
			\param s
				scope of MarkType
		*/
		MarkType(const std::string& n, MarkTypeId i, const std::type_info& s);

		/// Destructor
		~MarkType();

		/// get MarkType name
		const std::string getName() const;

		/// get MarkType id
		MarkTypeId getId() const;

		/// get MarkType definition scope
		const std::type_info& getScope() const;

		/// get new base mark ID for the next mark type scope
		static const MarkTypeId getNewBaseMarkId( const std::type_info& newType );

		/// get a reference to a list containing all constructed MarkTypes
		static MarkTypeList& getList();

		/// get a reference to a map containing MarkType scopes with BaseMarkId
		static MarkScopeIdMap& getScopeIdMap();

		/// print a simple list of all mark types to stdout
		static void printAll();

		/// print a simple list of mark type scopes with BaseMarkId to stdout
		static void printScopeIds();

	private:

		const std::string name; 	///< MarkType name
		MarkTypeId id; 				///< MarkType id
		const std::type_info& scope;///< MarkType definition scope
	};

	/** \class Tag

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief A Tag is used to build composed marks.

		<!-- [\note {notes}]* -->

		\sa Trace TraceProducer TraceConsumer MarkType

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class Tag {
	public:
		/**
			\brief Constructor
			\param n
				Tag name
			\param i
				Tag id
		*/
		Tag(const std::string& n, TagId i = NOID);

		/// get Tag name
		const std::string getName() const;

		/// get Tag id
		TagId getId() const;

		static const TagId NOID; ///< allows tags without id

	private:

		const std::string name; ///< Tag name
		TagId id; ///< Tag id
	};

}

#endif /*ODEMX_MARKTYPE_INCLUDED*/
