//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Report.h

	\author Ralf Gerstenberger
	<!-- [\author <author>]* -->

	\date created at 2002/06/21

	\brief Declaration of Report mechanism.

	\sa Report.cpp

	The Report mechanism is used to report accumulated data. All data
	are provided in tables. User defined Report-Objects transform the
	data into some output format (for example XML or HTML).

	\since 1.0
*/

#ifndef ODEMX_REPORT_INCLUDED
#define ODEMX_REPORT_INCLUDED

#include <odemx/util/SimTime.h>
#include <odemx/util/TypedObject.h>
#include <odemx/util/Table.h>
#include <odemx/util/LabeledObject.h>

#include <list>
#include <vector>
#include <string>
#include <typeinfo>

namespace odemx {

	/** \interface ReportProducer

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief A ReportProducer generates data.

		<!-- [\note {notes}]* -->

		\sa Report Table TableDefinition

		A ReportProducer implementation generates data for reports.
		It has to implement virtual void report(Report* r).

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class ReportProducer {
	public:
		virtual ~ReportProducer() {}
		/**
			\brief report generation

			This function is called by a Report object during report generation.
			A ReportProducer implementation implements this method to
			provide its data. ReportProducer can be associated to one or
			more Report objects. A ReportProducer will contribute to each
			Report it is associated to.
		*/
		virtual void report( Report* r ) = 0;
	};

	/** \class Report

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief Base class for special report objects

		<!-- [\note {notes}]* -->

		\sa Table TableDefinition ReportProducer

		Report is the base class for user defined report objects.
		Report manages Tables, ReportProducer and the report generation.
		Each Report object represents an individual report. There can be
		multiple reports in a simulation.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class Report {
	public:

		Report() {} ///< Construction
		virtual ~Report(); ///< Destruction

		/// \name ReportProducer handling
		//@{
		/// add ReportProducer \p rp to this Report
		void addProducer( ReportProducer* rp ) { rs.push_back(rp); }

		/// remove ReportProducer \p rp from this Report
		void removeProducer( ReportProducer* rp ) { rs.remove(rp); }
		//@}

		/// \name Table management
		//@{

		/**
			\brief create a Table

			\param name
				name of new Table
			\param def
				TableDefinition of new Table

			\return new Table

			\note If there is already a Table with an equal name and
			definition it will be returned instead of creating a new Table.
		*/
		Table* createTable( const std::string& name, TableDefinition* def );

		/**
			\brief find a Table

			\param name
				name of Table
			\param def
				TableDefinition of Table

			\return found Table or 0
		*/
		Table* findTable( const std::string& name, TableDefinition* def );
		//@}

		/**
			\brief generate report by getting data from ReportProducers

			Report generation:
			@li call all registered ReportProducer objects
			@li call processTables
		*/
		void generateReport();

	protected:
		/// contained Tables
		std::vector< Table* > ts;

		/**
			\brief process Tables in this Report

			Report objects have to provide a meaningful processTables() function.
		*/
		virtual void processTables() = 0;

		/// associated ReportProducer
		std::list< ReportProducer* > rs;
	};

	/** \class utTableDef

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief utTableDef simplifies report table definitions

		<!-- [\note {notes}]* -->

		<!-- [\sa {references to other classes}]* -->

		utTableDef can be used as a TableDefinition in Report::createTable or
		Report::findTable. It can be used to create a TableDefinition without
		defining a new class.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class utTableDef : public TableDefinition {
	private:
		size_t sz;
		const std::vector< std::string >& columnLabels;
		const std::vector< Column::Type >& columnTypes;
	public:

		utTableDef( size_t size,
				    const std::vector< std::string >& labels,
				    const std::vector< Column::Type >& types);

		virtual ~utTableDef() {}

		/**
			\name Implementation of TableDefinition interface
			\sa TableDefinition
			@{
		*/
		virtual size_t getNumberOfColumns() const;
		virtual const std::string getLabelOfColumn( size_t i ) const;
		virtual Column::Type getTypeOfColumn( size_t i ) const;
		//@}
	};
}

#endif

