//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file StatisticObject.h

	\author Ralf Gerstenberger

	\date created at 2003/05/24

	\brief Base classes for statistic objects

	\sa StatisticObject.cpp

	<!-- [detailed description] -->

	\since 1.0
*/

#ifndef ODEMX_STATISTICOBJECT_INCLUDED
#define ODEMX_STATISTICOBJECT_INCLUDED

#include <odemx/util/SimTime.h>
#include <list>

namespace odemx {
	class StatisticManager;

	/** \class StatisticObject

		\ingroup util

		\author Ralf Gerstenberger

		\brief Base class for statistics-generating classes

		<!-- [\note {notes}]* -->

		\sa StatisticManager

		StatisticObject declares a method for resetting statistics.
		This is required to skip misleading start and initialization
		phases of simulations. It also handles registration of
		statistics-generating objects.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class StatisticObject {
	public:
		/**
			\brief Construction

			Add statistic object to StatisticManager.
		*/
		StatisticObject(StatisticManager* manager);

		/**
			\brief Destruction

			Remove statistic object from StatisticManager.
		*/
		virtual ~StatisticObject();

		/// reset statistics
		virtual void reset();

		/// reset time
		virtual SimTime getResetTime() const {return resetAt;}

		/// get statistic manager
		StatisticManager* getStatisticManager() {return manager;}

		/// change statistic manager
		StatisticManager* setStatisticManager(StatisticManager* nm);

	protected:
		StatisticManager* manager; ///< pointer to the statistics manager
		SimTime resetAt; ///< the last reset time
	};

	/** \class StatisticManager

		\ingroup util

		\author Ralf Gerstenberger

		\brief Management of StatisticObjects

		<!-- [\note {notes}]* -->

		\sa StatisticObject

		StatisticManager provides a reset method for resetting all
		registered StatisticObjects. It also provides methods for
		registration.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class StatisticManager {
	public:
		/// Contruction
		StatisticManager();

		/// Destruction
		virtual ~StatisticManager();

		/// reset statistics
		virtual void reset();

		/// register StatisticObject
		virtual void add(StatisticObject* o);

		/// unregister StatisticObject
		virtual void remove(StatisticObject* o);

		/// get SimTime for reset time
		// making this pure virtual leads to problems in Dist construction
		virtual SimTime getTime() const { return static_cast< SimTime >(0); }

		// Implementation
	protected:
		std::list<StatisticObject*> so;
	};
}

#endif

