//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Table.h

	\author Ronald Kluth

	\date created at 2008/03/10

	\brief Implementation of odemx::Table - related functions

	\sa Table.cpp

	\since 2.1
*/

#include <odemx/util/Table.h>
using namespace odemx;
using namespace std;

/********************** Column *********************************************/

const string Column::typeToString( Type t )
{
	switch( t )
	{
	case Column::BAR: 		return string( "BAR" );
	case Column::SIMTIME: 	return string( "SIMTIME" );
	case Column::INTEGER: 	return string( "INTEGER" );
	case Column::REAL: 		return string( "REAL" );
	case Column::STRING: 	return string( "STRING" );
	default: 				return string( "INVALID" );
	}
}

/********************* TableDefinition ***********************************/

// compare table definitions, used to determine if two tables are the same
bool odemx::operator ==( const TableDefinition& f, const TableDefinition& s )
{
	size_t columns = f.getNumberOfColumns();
	string strF, strS;

	// check for same column number
	if( columns != s.getNumberOfColumns() )
		return false;

	// compare each column by type and label
	for( size_t i = 0; i < columns; ++i ) {
		if( f.getTypeOfColumn(i) != s.getTypeOfColumn(i) )
			return false;

		strF = f.getLabelOfColumn(i);
		strS = s.getLabelOfColumn(i);
		if( strF != strS )
			return false;
	}
	// all checks passed
	return true;
}

size_t DynamicTableDefinition::getNumberOfColumns() const
{
	return columnLabels.size();
}

const std::string DynamicTableDefinition::getLabelOfColumn(size_t i) const
{
	return (i < columnLabels.size())  ?  columnLabels[i] : "ERROR";
}

Column::Type DynamicTableDefinition::getTypeOfColumn(size_t i) const
{
	return (i < columnTypes.size())  ?  columnTypes[i] : Column::INVALID;
}

void DynamicTableDefinition::addColumn(const string& label, Column::Type type)
{
	if ( label.empty() || type == Column::INVALID )
		return;

	columnLabels.push_back(label);
	columnTypes.push_back(type);
}

/********************* Table Functions ***********************************/

/// Construction is managed by Report
Table::Table( Label l, TableDefinition* d )
:	label( l ),
	def( d ),
	lineCount( 0 )
{
	// create vectors for columns according to TableDefinition
	for( size_t i = 0; i < def -> getNumberOfColumns(); ++i ) {

		Column* newColumn = 0;

		// create each column according to the given type
		switch( def -> getTypeOfColumn(i) ) {

		case Column::BAR:
			newColumn = new ColumnT< long >( def -> getLabelOfColumn( i ), Column::BAR );
			break;

		case Column::INTEGER:
			newColumn = new ColumnT< long >( def -> getLabelOfColumn( i ), Column::INTEGER );
			break;

		case Column::REAL:
			newColumn = new ColumnT< double >( def -> getLabelOfColumn( i ), Column::REAL );
			break;

		case Column::STRING:
			newColumn = new ColumnT< std::string >( def -> getLabelOfColumn( i ), Column::STRING );
			break;

		case Column::SIMTIME:
			newColumn = new ColumnT< SimTime >( def -> getLabelOfColumn( i ), Column::SIMTIME );
			break;

		default: // INVALID
			// error
			error( "Table::Table(): invalid column type detected" );
			break;
		}

		if( newColumn )
			columnVec.push_back( newColumn );
		else
			error( "Table::Table(): could not create column" );
	}

}

Table::~Table()
{
	// clean up columns
	std::vector< Column* >::const_iterator i;
	for( i = columnVec.begin(); i != columnVec.end(); ++i )
	{
		delete *i;
	}
}

/********************** Input Streaming **********************************/

// only four functions actually add data, the others are merely
// wrappers using the three operator << implementations

// this function deals with INTEGER and with BAR column types
Table& odemx::operator <<( Table& t, long d ) {

	// try to get the current column according to inputPosition
	// this also checks for the correct type of column, returns 0 if wrong
	ColumnT<long>* currentColumn =
		dynamic_cast< ColumnT<long>* >( t.getCurrentInputColumn() );

	// test column type
	if ( currentColumn != 0 ) {
		currentColumn -> addData( d );
	}
	else {
		error( "Table operator <<(long): wrong column type at input position; cannot enter data into Table" );
		return t;
	}

	// set input pointer to next position for streaming
	t.advanceInputPosition();

	return t;
}


Table& odemx::operator <<( Table& t, long long d ) {

	// try to get the current column according to inputPosition
	// this also checks for the correct type of column, returns 0 if wrong
	ColumnT<long long>* currentColumn =
		dynamic_cast< ColumnT<long long>* >( t.getCurrentInputColumn() );

	// test column type
	if ( currentColumn != 0 ) {
		currentColumn -> addData( d );
	}
	else {
		error( "Table operator <<(long long): wrong column type at input position; cannot enter data into Table" );
		return t;
	}

	// set input pointer to next position for streaming
	t.advanceInputPosition();

	return t;
}


// the above function forces use of this one:
Table& odemx::operator <<( Table& t, double d ) {

	// try to get the current column according to inputPosition
	// this also checks for the correct type of column, returns 0 if wrong
	ColumnT<double>* currentColumn =
		dynamic_cast< ColumnT<double>* >( t.getCurrentInputColumn() );

	// test column type
	if ( currentColumn != 0 ) {
		currentColumn -> addData( d );
	}
	else {
		error( "Table operator <<(double): wrong column type at input position; cannot enter data into Table" );
		return t;
	}

	// set input pointer to next position for streaming
	t.advanceInputPosition();

	return t;
}

Table& odemx::operator <<( Table& t, const string s ) {

	// try to get the current column according to inputPosition
	// this also checks for the correct type of column, returns 0 if wrong
	ColumnT<string>* currentColumn =
		dynamic_cast< ColumnT<string>* >( t.getCurrentInputColumn() );

	// test column type
	if ( currentColumn != 0 ) {
		currentColumn -> addData( s );
	}
	else {
		error( "Table operator <<(string): wrong column type at input position; cannot enter data into Table" );
		return t;
	}

	// set input pointer to next position for streaming
	t.advanceInputPosition();

	return t;
}

Table& odemx::operator <<( Table& t, int i ) {
	return t << long( i );
}

Table& odemx::operator <<( Table& t, unsigned int i ) {
	return t << long( i );
}

Table& odemx::operator <<( Table& t, unsigned long i ) {
	return t << long( i );
}

Table& odemx::operator <<( Table& t, float f ) {
	return t << double( f );
}

Table& odemx::operator <<( Table& t, const char* s ) {
	return t << string( s );
}


// this function is used to skip table cells or the rest of the line
Table& odemx::operator <<( Table& t, TableDefinition::ControlCode cc ) {

	size_t line;

	// control codes may end a line early
	// or cause a jump to the next table cell
	switch( cc )
	{
	case TableDefinition::ENDL:
		// fill line with default values
		line = t.lineCount;

		while( line == t.lineCount ) {
			// enter default value
			t.addDefaultValue();
		}
		break;

	case TableDefinition::TAB:
		// enter default value
		t.addDefaultValue();
		break;

	default:
		// error
		return t;
	}

	return t;
}

void Table::addDefaultValue() {

	Column* currentCol = getCurrentInputColumn();

	// enter default value
	switch( getCurrentInputColumn() -> getType() )
	{
	case Column::SIMTIME:
		dynamic_cast< ColumnT<SimTime>* >( currentCol ) -> addData( 0 );
		break;

	case Column::INTEGER:
		dynamic_cast< ColumnT<long>* >( currentCol ) -> addData( 0 );
		break;

	case Column::REAL:
		dynamic_cast< ColumnT<double>* >( currentCol ) -> addData( 0.0 );
		break;

	case Column::STRING:
		dynamic_cast< ColumnT<string>* >( currentCol ) -> addData( "" );
		break;

	case Column::BAR:
		dynamic_cast< ColumnT<long>* >( currentCol ) -> addData( 0 );
		break;

	default:
		/* MPi no text for this from Ralf */
		error( string( "Table::addDefaultValue(): invalid column type, cannot add data -- Object:" ) + getLabel() );
	}

	// update inputColumnPointer
	advanceInputPosition();
}


/******************** Output streaming ************************************/

Table& odemx::operator >>( Table& t, long long& to ) {

	// try to get the current column according to outputPosition
	// this also checks for the correct type of column, returns 0 if wrong
	ColumnT<long long>* currentColumn =
		dynamic_cast< ColumnT<long long>* >( t.getCurrentOutputColumn() );

	// test column type
	if ( currentColumn != 0 ) {
		to = currentColumn -> getData( t.getOutputLine() );
	}
	else {
		error( string( "Table::operator >>(long long): wrong column type -- Object:" ) + t.getLabel() );
		return t;
	}

	// set output pointer to next position for streaming
	t.advanceOutputPosition();

	return t;
}

Table& odemx::operator >>( Table& t, long& to ) {

	// try to get the current column according to outputPosition
	// this also checks for the correct type of column, returns 0 if wrong
	ColumnT<long>* currentColumn =
		dynamic_cast< ColumnT<long>* >( t.getCurrentOutputColumn() );

	// test column type
	if ( currentColumn != 0 ) {
		to = currentColumn -> getData( t.getOutputLine() );
	}
	else {
		error( string( "Table::operator >>(long): wrong column type -- Object:" ) + t.getLabel() );
		return t;
	}

	// set output pointer to next position for streaming
	t.advanceOutputPosition();

	return t;
}

Table& odemx::operator >>( Table& t, double& to ) {

	// try to get the current column according to outputPosition
	// this also checks for the correct type of column, is 0 if wrong
	ColumnT<double>* currentColumn =
		dynamic_cast< ColumnT<double>* >( t.getCurrentOutputColumn() );

	// test column type
	if ( currentColumn != 0 ) {
		to = currentColumn -> getData( t.getOutputLine() );
	}
	else {

		ColumnT<long>* currentColumn =
			dynamic_cast< ColumnT<long>* >( t.getCurrentOutputColumn() );

		if ( currentColumn != 0 ) {
			to = (double) currentColumn -> getData( t.getOutputLine() );
		}
		else {
			error( string( "Table::operator >>(double): wrong column type -- Object:" ) + t.getLabel() );
			return t;
		}
	}

	// set output pointer to next position for streaming
	t.advanceOutputPosition();

	return t;
}

Table& odemx::operator >>( Table& t, string& to ) {

	// any column type accepted, conversion with ostringstream
	ostringstream convert;

	Column* currentColumn = t.getCurrentOutputColumn();

	// try to get the current column according to inputPosition
	// this also checks for the correct type of column, returns 0 if wrong
	switch( currentColumn -> getType() ) {

	case Column::BAR:
	case Column::INTEGER:
		convert << dynamic_cast< ColumnT<long>* >( currentColumn ) -> getData( t.getOutputLine() );
		break;

	case Column::REAL:
		convert << dynamic_cast< ColumnT<double>* >( currentColumn ) -> getData( t.getOutputLine() );
		break;

	case Column::STRING:
		convert << dynamic_cast< ColumnT<string>* >( currentColumn ) -> getData( t.getOutputLine() );
		break;

	case Column::SIMTIME:
		convert << dynamic_cast< ColumnT<SimTime>* >( currentColumn ) -> getData( t.getOutputLine() );
		break;

	default:
		error( string( "Table::operator >>(string): invalid column type -- Object:" ) + t.getLabel() );
		return t;
	}

	// write string to stream
	to = convert.str();

	// update outputColumnPointer
	t.advanceOutputPosition();

	return t;
}


/****************** Random Access Output **********************************/

SimTime Table::getSIMTIME( size_t col, size_t line )
{
	if( col >= columnVec.size() )
	{
		error( string("Table::getSIMTIME(): column out of range -- Object:") + getLabel() );
		return 0;
	}

	if( line >= lineCount )
	{
		error( string("Table::getSIMTIME(): line out of range -- Object:") + getLabel() );
		return 0;
	}

	// try to get the column col, check for the correct type, is 0 if wrong
	ColumnT<SimTime>* currentColumn =
		dynamic_cast< ColumnT<SimTime>* >( columnVec[ col ] );

	// test column type
	if ( currentColumn != 0 && currentColumn -> getType() == Column::SIMTIME )
	{
		return currentColumn -> getData( line );
	}
	else
	{
		error( string("Table::getSIMTIME(): wrong column type; cannot read data -- Object:") + getLabel() );
		return 0;
	}
}


long Table::getBAR( size_t col, size_t line ) {

	if( col >= columnVec.size() )
	{
		error( string("Table::getBAR(): column out of range -- Object:") + getLabel() );
		return 0;
	}

	if( line >= lineCount )
	{
		error( string("Table::getBAR(): line out of range -- Object:") + getLabel() );
		return 0;
	}

	// try to get the column col, check for the correct type, is 0 if wrong
	ColumnT<long>* currentColumn =
		dynamic_cast< ColumnT<long>* >( columnVec[ col ] );

	// test column type
	if ( currentColumn != 0 && currentColumn -> getType() == Column::BAR ) {
		return currentColumn -> getData( line );
	}
	else {
		error( string("Table::getBAR(): wrong column type; cannot read data -- Object:") + getLabel() );
		return 0;
	}
}

long Table::getINTEGER( size_t col, size_t line ) {

	if( col >= columnVec.size() )
	{
		error( string("Table::getINTEGER(): column out of range -- Object:") + getLabel() );
		return 0;
	}

	if( line >= lineCount )
	{
		error( string("Table::getINTEGER(): line out of range -- Object:") + getLabel() );
		return 0;
	}

	// try to get the column col, check for the correct type, is 0 if wrong
	ColumnT<long>* currentColumn =
		dynamic_cast< ColumnT<long>* >( columnVec[ col ] );

	// test column type
	if ( currentColumn != 0 && currentColumn -> getType() == Column::INTEGER ) {
			return currentColumn -> getData( line );
	}
	else {
		error( string("Table::getINTEGER(): wrong column type; cannot read data -- Object:") + getLabel() );
		return 0;
	}
}

double Table::getREAL( size_t col, size_t line ) {

	if( col >= columnVec.size() )
	{
		error( string("Table::getREAL(): column out of range -- Object:") + getLabel() );
		return 0.0;
	}

	if( line >= lineCount )
	{
		error( string("Table::getREAL(): line out of range -- Object:") + getLabel() );
		return 0.0;
	}

	// try to get the column col, check for the correct type, is 0 if wrong
	ColumnT<double>* currentColumn =
		dynamic_cast< ColumnT<double>* >( columnVec[ col ]  );

	// test column type
	if ( currentColumn != 0 && currentColumn -> getType() == Column::REAL ) {
		return currentColumn -> getData( line );
	}
	else {
		// cast from long type is also allowed
		ColumnT<long>* currentColumn =
			dynamic_cast< ColumnT<long>* >( columnVec[ col ] );

		if ( currentColumn != 0 && currentColumn -> getType() == Column::INTEGER ) {
			return (double) currentColumn -> getData( line );
		}
		else {
			error( string("Table::getREAL(): wrong column type; cannot read data -- Object:") + getLabel() );
			return 0.0;
		}
	}
}

string Table::getSTRING(size_t col, size_t line) {

	if( col >= columnVec.size() )
	{
		error( string("Table::getSTRING(): column out of range -- Object:") + getLabel() );
		return "ERROR";
	}

	if( line >= lineCount )
	{
		error( string("Table::getSTRING(): line out of range -- Object:") + getLabel() );
		return "ERROR";
	}

	// any column type accepted, conversion with ostringstream
	ostringstream convert;

	Column* currentColumn = columnVec[ col ];

	// checks for the correct type of column, returns 0 if wrong
	switch( currentColumn -> getType() ) {

	case Column::BAR:
	case Column::INTEGER:
		convert << dynamic_cast< ColumnT<long>* >( currentColumn ) -> getData( line );
		break;

	case Column::SIMTIME:
		convert << dynamic_cast< ColumnT<SimTime>* >( currentColumn ) -> getData( line );
		break;

	case Column::REAL:
		convert << dynamic_cast< ColumnT<double>* >( currentColumn ) -> getData( line );
		break;

	case Column::STRING:
		convert << dynamic_cast< ColumnT<string>* >( currentColumn ) -> getData( line );
		break;

	default:
		error( string("Table::getSTRING(): wrong column type; cannot read data -- Object:") + getLabel() );
		return "";
	}

	return convert.str();
}

// help procedures
void Table::advanceInputPosition() {

	// advance the inputPosition by one
	++inputPosition.col;

	// check for line end
	if( inputPosition.col >= getNumberOfColumns() ) {
		inputPosition.col = inputPosition.col % getNumberOfColumns();
		++inputPosition.line;
		++lineCount;
	}
}

void Table::advanceOutputPosition() {

	// advance the outputPosition by one
	++outputPosition.col;

	// check for line end
	if( outputPosition.col >= getNumberOfColumns() ) {
		outputPosition.col = outputPosition.col % getNumberOfColumns();
		++outputPosition.line;
	}

	// wrap output streaming at last line to start anew
	if( outputPosition.line >= getNumberOfLines() ) {
		outputPosition.line = 0;
	}
}

