//----------------------------------------------------------------------------
//	Copyright (C) 2002 - 2008 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Table.h

	\author Ronald Kluth

	\date created at 2008/03/10

	\brief Declaration of odemx::Table for the Report mechanism

	\sa Table.cpp

	\since 2.1
*/

#ifndef ODEMX_TABLE_INCLUDED
#define ODEMX_TABLE_INCLUDED

#include <vector>
#include <string>
#include <sstream>
#include <odemx/util/ErrorHandling.h>
#include <odemx/util/LabeledObject.h>
#include <odemx/util/SimTime.h>

namespace odemx {

	// forward declaration
	class Report;
	class Table;

	/** \class Column

		\ingroup util

		\author Ronald Kluth
		<!-- [\author <author>]* -->

		\brief provides column label and type describing column data

		<!-- [\note {notes}]* -->

		\sa Table, TableDefinition

		All columns in tables are associated to a data type like integer,
		real values or strings, and sometimes represent graphical bars.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.1
	*/
	class Column {
	public:

		/// Destruction
		virtual ~Column() {}

		/**
			\ingroup util

			\brief possible table column types

			Enumeration of the allowed column types in tables
		*/
		enum Type {
			INTEGER, ///< integer numbers
			REAL, ///< real numbers
			STRING, ///< text
			BAR, ///< graphical representation
			SIMTIME, ///< used for SimTime
			INVALID ///< used for undefined columns
		};

		/// get a string representation of the given column type \p t
		static const std::string typeToString( Type t );
		/// each column must know its caption label
		const Label& getLabel() { return label; }
		/// each column must know its type
		Column::Type getType() { return type; }

	protected:

		/// Constructor
		Column( const Label& l, Column::Type t ):
			label( l ),
			type( t )
			{}

	private:
		/// each column has a label for the table captions
		Label label;
		/// each column knows its type
		Column::Type type;
	};


	/** \class ColumnT

		\ingroup util

		\author Ronald Kluth
		<!-- [\author <author>]* -->

		\brief represents a table column and stores its data in a vector

		<!-- [\note {notes}]* -->

		\sa Table, TableDefinition

		ColumnT is the class that stores all table data. It requires a
		label and a column type, which is given during construction.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 2.1
	*/
	template < typename DataType >
	class ColumnT: public Column {
	public:

		/// Destructor
		virtual ~ColumnT() {}

	protected:
		friend class Table;

		friend Table& operator <<( Table& t, long d );
		friend Table& operator <<( Table& t, long long d );
		friend Table& operator <<( Table& t, double d );
		friend Table& operator <<( Table& t, std::string s );

		friend Table& operator >>(Table& t, long& to);
		friend Table& operator >>(Table& t, long long& to);
		friend Table& operator >>(Table& t, double& to);
		friend Table& operator >>(Table& t, std::string& to);

		/// Constructor, to be called by Table class only
		ColumnT( const Label& l, Column::Type t ):
			Column( l, t )
			{}

		/// append data to column, i.e. add content of next line in table
		void addData( DataType d ) { data.push_back( d ); }

		/// read data from column, the given line is range-checked
		DataType getData( size_t line )
		{
			if( line >= data.size() )
			{
				error( std::string("ColumnT::getData(): given line is out of range in Column ")
					   + getLabel() + " of type " + typeToString( getType() ) );
			}
			return data.at( line );
		}
	private:
		/// stores the column data, index represents lines in the Table
		std::vector< DataType > data;
	};


	/** \interface TableDefinition

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief The table structure is defined by TableDefinition.

		<!-- [\note {notes}]* -->

		\sa DynamicTableDefinition utTableDef

		The structure of a table consists of columns. Each column has
		a Label and a Column::Type. TableDefinition is an interface that
		is used to provide this information.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class TableDefinition {
	public:

		/**
			\ingroup util

			\brief control codes

			These control codes are used for operator << input
			into tables.
		*/
		enum ControlCode {
			ENDL, ///< end of table line
			TAB ///< skip column
		};

		/// Destructor
		virtual ~TableDefinition() {}
		/**
			\brief number of columns

			\return number of columns in table
		*/
		virtual size_t getNumberOfColumns() const = 0;

		/**
			\brief get label of column

			\return label or 0 if there is no such column
			\param i
				index of column (0..)
		*/
		virtual const std::string getLabelOfColumn(size_t i) const = 0;

		/**
			\brief get type of column

			\return Column::Type, Column::Type::INVALID if there is no
			such column

			\param i
				index of column (0..)
		*/
		virtual Column::Type getTypeOfColumn(size_t i) const = 0;

		friend bool operator ==(const TableDefinition& f, const TableDefinition& s);
	};

	/**
		\ingroup util

		\brief compare two TableDefinitions, used for grouping ReportProducers in the same table

		 \return true if number of columns, all labels and all types
		 are equal
	*/
	bool operator ==(const TableDefinition& f, const TableDefinition& s);

	/** \class DynamicTableDefinition

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief DynamicTableDefinition is a TableDefinition implementation for
		dynamic definitions.

		<!-- [\note {notes}]* -->

		<!-- [\sa {references to other classes}]* -->

		DynamicTableDefinition can be used as a TableDefinition in
		Report::createTable or Report::findTable. It can be used to build
		a TableDefinition at runtime.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class DynamicTableDefinition : public TableDefinition {
		// Interface
	public:
		virtual ~DynamicTableDefinition() {} ///< Destruction

		/**
			\name Implementation of TableDefinition interface
			\sa TableDefinition
			@{
		*/
		virtual size_t getNumberOfColumns() const;
		virtual const std::string getLabelOfColumn( size_t i ) const;
		virtual Column::Type getTypeOfColumn( size_t i ) const;
		//@}

		/**
			\brief add columns to the TableDefinition
			\param label
				column label
			\param type
				column type

			Use addColumn() to build a TableDefinition.
		*/
		void addColumn( const std::string& label, Column::Type type );

		// Implementation
	private:
		std::vector< std::string > columnLabels; ///< column labels
		std::vector< Column::Type > columnTypes; ///< column types
	};

	/** \class Table

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief Container for reported data.

		<!-- [\note {notes}]* -->

		\sa TableDefinition Report

		All data are reported in tables. The structure of a Table is defined
		by a TableDefinition. A Table cannot be created manually. To get a Table
		the user has to use a Report object.

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class Table {
	public:

		/// stores a Table input or output position when streaming data
		struct Position {
			Position(): col( 0 ), line( 0 ) {}

			size_t col;
			size_t line;
		};

		/// Construction is managed by Report
		Table( Label l, TableDefinition* d );

		/// Destruction is managed by Report
		~Table();

		/// \name Table attributes
		//@{

		/// get Label of table
		const Label getLabel() const { return label; }

		/**
		 * @brief get number of lines in table
		 *
		 * @note This function returns the number of completely filled lines.
		 * To finish a line, use TableDefinition::ControlCode.
		 */
		size_t getNumberOfLines() const { return lineCount; }

		/// get number of columns in table
		size_t getNumberOfColumns() const { return columnVec.size(); }

		/// get column labels
		const std::string getLabelOfColumn(size_t i) const { return columnVec[i] -> getLabel(); }

		/// get column types
		Column::Type getTypeOfColumn(size_t i) const { return columnVec[i] -> getType(); }
		//@}

		/**
			\name Input streaming

			The data can be entered into a table in a streaming like fashion.
			The data type has to match the column type. A ControlCode can be
			used to skip columns or the rest of a line. Skipped columns are
			filled with default	data <tt>(0, 0.0, "")</tt>.

			@{
		*/
		friend Table& operator <<(Table& t, int d);
		friend Table& operator <<(Table& t, unsigned int d);
		friend Table& operator <<(Table& t, long d);
		friend Table& operator <<(Table& t, long long d);
		friend Table& operator <<(Table& t, unsigned long d);
		friend Table& operator <<(Table& t, float d);
		friend Table& operator <<(Table& t, double d);
		friend Table& operator <<(Table& t, std::string d);
		friend Table& operator <<(Table& t, const char* d);
		friend Table& operator <<(Table& t, TableDefinition::ControlCode d);
		//@}

		/**
			\name Output streaming

			The data of a table can be read in a streaming like fashion.
			The data type has to match the column type.

			@{
		*/
		friend Table& operator >>(Table& t, long& to);
		friend Table& operator >>(Table& t, long long& to);
		friend Table& operator >>(Table& t, double& to);
		friend Table& operator >>(Table& t, std::string& to);
		//@}

		/**
			\name Output random access

			Get the data of a specific field in table.

			@{
		*/
		/**
			\brief Get SIMTIME value from field (\p col, \p line).
			\note Column col must be a SIMTIME column.
		*/
		SimTime getSIMTIME(size_t col, size_t line);

		/**
			\brief Get graphical BAR value from field (\p col, \p line).
			\note Column col must be a BAR column.
		*/
		long getBAR(size_t col, size_t line);

		/**
			\brief Get INTEGER value from field (\p col, \p line).
			\note Column col must be an INTEGER column.
		*/
		long getINTEGER(size_t col, size_t line);

		/**
			\brief Get REAL value from field (\p col, \p line).
			\note Column col must be an INTEGER or a REAL column.
		*/
		double getREAL(size_t col, size_t line);

		/**
			\brief Get a STRING value from field (\p col, \p line).
			\note Type conversion from BAR, INTEGER and REAL to STRING
			is done if required.
		*/
		std::string getSTRING(size_t col, size_t line);
		//@}

		/// return the current input column to add data
		Column* getCurrentInputColumn() {
			return columnVec[ inputPosition.col ];
		}
		/// return the current output column to read data from
		Column* getCurrentOutputColumn() {
			return columnVec[ outputPosition.col ];
		}
		/// return the current output streaming line
		size_t getOutputLine() {
			return outputPosition.line;
		}

		/// add default value (0, 0.0, "", 0) in column
		void addDefaultValue();
		/// increment input pointer for streaming, change line if necessary

	private:

		void advanceInputPosition();
		/// increment input pointer for streaming, change line if necessary
		void advanceOutputPosition();

		friend class Report;

		Label label; ///< table label
		TableDefinition* def; ///< table definition
		size_t lineCount; /// size of table
		std::vector< Column* > columnVec; ///< vector containing pointers to columns
		Position inputPosition; ///< input streaming position pointer
		Position outputPosition; ///< output streaming position pointer
	};

	Table& operator <<(Table& t, int d); ///< enter int value
	Table& operator <<(Table& t, unsigned int d); ///< enter unsigned int value
	Table& operator <<(Table& t, long d); ///< enter long value
	Table& operator <<(Table& t, long long d); ///< enter SimTime value
	Table& operator <<(Table& t, unsigned long d); ///< enter unsigned long value
	Table& operator <<(Table& t, float d); ///< enter float value
	Table& operator <<(Table& t, double d); ///< enter double value
	Table& operator <<(Table& t, std::string d); ///< enter std::string value
	Table& operator <<(Table& t, const char* d); ///< enter const char* value
	Table& operator <<(Table& t, TableDefinition::ControlCode d); ///< enter control codes

	Table& operator >>(Table& t, long& to); ///< read long value
	Table& operator >>(Table& t, long long& to); ///< read long long value
	Table& operator >>(Table& t, double& to); ///< read double value
	Table& operator >>(Table& t, std::string& to); ///< read std::string value

} // namespace odemx



#endif /* ODEMX_TABLE_INCLUDED */
