//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Trace.cpp

	\author Ralf Gerstenberger

	\date created at 2002/03/11

	\brief Trace mechanism implementation

	\sa Trace.h

	<!-- [detailed description] -->

	\since 1.0
*/

#ifdef _MSC_VER
#pragma warning(disable : 4530)
#endif

#include <odemx/util/Trace.h>

#include <algorithm>
#include <cstring>

using namespace std;
using namespace odemx;

Trace::Trace()
:	state(CREATED),
	openMark(false)
	{}

Trace::~Trace()
{
	if( state == RUNNING || state == PAUSED )
		warning( "~Trace(): trace was not stopped properly - call stopTrace()" );
}

//
// TraceConsumer management
//
void Trace::addConsumer(TraceConsumer* c)
{
	if( ! dynamic_cast< TraceConsumer* >( c ) )
	{
		error( "Trace::addConsumer(): failed to add TraceConsumer: wrong type" );
		return;
	}

	cons.push_back(c);

	// tell the new consumer the current trace state
	if (state==RUNNING || state==PAUSED)
		c->startTrace();

	if (state==PAUSED)
		c->pauseTrace();
}

void Trace::removeConsumer(TraceConsumer* c)
{
	if( ! dynamic_cast< TraceConsumer* >( c ) )
	{
		error( "Trace::removeConsumer(): failed to remove TraceConsumer: wrong type" );
		return;
	}

	// end the consumer's actions properly
	if (state==RUNNING || state==PAUSED)
		c->stopTrace();

	cons.remove(c);
}

const std::list<TraceConsumer*>& Trace::getConsumers() const
{
	return cons;
}

//
// Trace control
//
void Trace::startTrace() {

	if (state!=CREATED)
		return;

	state = RUNNING;

	list<TraceConsumer*>::iterator i;
	for (i=cons.begin(); i!=cons.end(); ++i)
		(*i)->startTrace();
}

void Trace::stopTrace() {

	if (state==PAUSED)
		continueTrace();

	if (state!=RUNNING)
		return;

	if (openMark)
		endMark();

	state = FINISHED;

	list<TraceConsumer*>::iterator i;
	for (i=cons.begin(); i!=cons.end(); ++i)
		(*i)->stopTrace();
}

void Trace::pauseTrace() {

	if (state!=RUNNING)
		return;

	if (openMark)
		endMark();

	state = PAUSED;

	list<TraceConsumer*>::iterator i;
	for (i=cons.begin(); i!=cons.end(); ++i)
		(*i)->pauseTrace();
}

void Trace::continueTrace() {

	if (state!=PAUSED)
		return;

	state = RUNNING;

	list<TraceConsumer*>::iterator i;
	for (i=cons.begin(); i!=cons.end(); ++i)
		(*i)->continueTrace();
}

//
// Simple trace marks are reported by the following methods.
//
void Trace::mark(const TraceProducer* sender, MarkType m, const std::string& comment /*=""*/) {
	if (!sendMarks())
		return;

	list<TraceConsumer*>::iterator i;

	for (i=cons.begin(); i!=cons.end(); ++i)
		(*i)->mark(sender, &m, comment);
}

void Trace::mark(const TraceProducer* sender, MarkType m, const TraceProducer* partner, const std::string& comment /*=""*/) {
	if (!sendMarks())
		return;

	list<TraceConsumer*>::iterator i;

	for (i=cons.begin(); i!=cons.end(); ++i)
		(*i)->mark(sender, &m, partner, comment);
}

void Trace::mark(const TraceProducer* sender, MarkType m, const TraceProducer* firstPartner, const TraceProducer* secondPartner, const std::string& comment /*=""*/) {
	if (!sendMarks())
		return;

	list<TraceConsumer*>::iterator i;

	for (i=cons.begin(); i!=cons.end(); ++i)
		(*i)->mark(sender, &m, firstPartner, secondPartner, comment);
}


//
// Composed trace marks are built with the following methods.
//
void Trace::beginMark(const TraceProducer* sender, MarkType m) {
	if (!sendMarks())
		return;

	if (openMark)
		endMark();

	list<TraceConsumer*>::iterator i;
	for (i=cons.begin(); i!=cons.end(); ++i) \
		(*i)->beginMark(sender, &m);

	openMark=true;
}

void Trace::endMark(const std::string& comment /*=""*/ ) {
	if (!sendMarks())
		return;

	if (!openMark) {
		error("Trace::endMark(): called without a corresponding beginMark()");
		return;
	}

	list<TraceConsumer*>::iterator i;
	for (i=cons.begin(); i!=cons.end(); ++i) \
		(*i)->endMark(comment);

	openMark=false;
}

void Trace::beginTag(Tag t) {
	if (!sendMarks())
		return;

	if (!openMark) {
		error("Trace::beginTag(): called without an open composed mark");
		return;
	}

	list<TraceConsumer*>::iterator i;
	for (i=cons.begin(); i!=cons.end(); ++i)
		(*i)->beginTag(t);
}

void Trace::endTag(Tag t) {
	if (!sendMarks())
		return;

	if (!openMark) {
		error("Trace::endTag(): called without an open composed mark");
		return;
	}

	list<TraceConsumer*>::iterator i;
	for (i=cons.begin(); i!=cons.end(); ++i)
		(*i)->endTag(t);
}

