//----------------------------------------------------------------------------
//	Copyright (C) 2002, 2004 Humboldt-Universitaet zu Berlin
//
//	This library is free software; you can redistribute it and/or
//	modify it under the terms of the GNU Lesser General Public
//	License as published by the Free Software Foundation; either
//	version 2.1 of the License, or (at your option) any later version.
//
//	This library is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//	Lesser General Public License for more details.
//
//	You should have received a copy of the GNU Lesser General Public
//	License along with this library; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
/**	\file Trace.h

	\author Ralf Gerstenberger
	<!-- [\author <author>]* -->

	\date created at 2002/03/10

	\brief Declaration of trace mark types

	\sa MarkType.h MarkType.cpp Trace.h Trace.cpp

	A Trace mechanism is used to report events and state changes of objects
	in a generalized way. Objects (TraceProducer) generate a sequence of marks,
	which are sent to a	trace manager (Trace). The trace manager broadcasts
	the marks to all registered trace handlers (TraceConsumer).

	Such a Mark is defined by its type (MarkType). A sender can use
	one of the predefined mark-functions to send a mark or compose a mark
	with the addition of tags.

	\since 1.0
*/

#ifndef ODEMX_TRACE_INCLUDED
#define ODEMX_TRACE_INCLUDED

#include <odemx/util/LabeledObject.h>
#include <odemx/util/TypedObject.h>
#include <odemx/util/MarkType.h>
#include <odemx/util/SimTime.h>
#include <odemx/util/ErrorHandling.h>

#include <string>
#include <list>
#include <cassert>

namespace odemx {

	// forward declaration
	class Trace;

	/** \interface TraceProducer

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief A TraceProducer generates marks that form the trace.

		<!-- [\note {notes}]* -->

		\sa Trace TraceConsumer MarkType Tag

		\copydoc Trace.h

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class TraceProducer: public virtual LabeledObject,
						 public virtual TypedObject {
	public:
		/**
			\brief A TraceProducer must have access to a Trace object.

			\return pointer to a Trace object
		*/
		virtual Trace* getTrace() const = 0;
	};

	/** \interface TraceConsumer

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief A TraceConsumer receives marks from the Trace it is added to.

		<!-- [\note {notes}]* -->

		\sa Trace TraceProducer MarkType Tag

  		\copydoc Trace.h

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class TraceConsumer {
	public:
		virtual ~TraceConsumer() { }
	private:
		// Only Trace may call any of these methods.
		friend class Trace;

		/**
			\name Simple trace marks.

			Simple trace marks should be handled by the following methods.

			@{
		*/
		/**
			\brief handle most simple trace mark

			\param sender
				pointer to sender of this mark
			\param m
				type of this mark
			\param comment
				comments
		*/
		virtual void mark(const TraceProducer* sender, const MarkType* m, const std::string& comment = "") = 0;

		// Marks that include information about associations of the producer:
		/// handle marks that includes information about associations of the producer
		virtual void mark(const TraceProducer* sender, const MarkType* m, const TraceProducer* partner, const std::string& comment = "") = 0;
		/// handle mark that includes information about associations of the producer
		virtual void mark(const TraceProducer* sender, const MarkType* m, const TraceProducer* firstPartner, const TraceProducer* secondPartner, const std::string& comment = "") = 0;

		// Marks that include information about attributes of the producer:
		/// handle bool-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, bool newValue, bool oldValue, const std::string& comment = "") = 0;
		/// handle char-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, char newValue, char oldValue, const std::string& comment = "") = 0;
		/// handle short-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, short newValue, short oldValue, const std::string& comment = "") = 0;
		/// handle int-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, int newValue, int oldValue, const std::string& comment = "") = 0;
		/// handle long-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, long newValue, long oldValue, const std::string& comment = "") = 0;
		/// handle SimTime-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, long long newValue, long long oldValue, const std::string& comment = "") = 0;
		/// handle float-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, float newValue, float oldValue, const std::string& comment = "") = 0;
		/// handle double-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, double newValue, double oldValue, const std::string& comment = "") = 0;
		/// handle const char*-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, const char* newValue, const char* oldValue, const std::string& comment = "") = 0;

		/// handle unsigned char-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned char newValue, unsigned char oldValue, const std::string& comment = "") = 0;
		/// handle unsigned short-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned short newValue, unsigned short oldValue, const std::string& comment = "") = 0;
		/// handle unsigned int-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned int newValue, unsigned int oldValue, const std::string& comment = "") = 0;
		/// handle unsigned long-attribute changes
		virtual void mark(const TraceProducer* sender, const MarkType* m, unsigned long newValue, unsigned long oldValue, const std::string& comment = "") = 0;
		//@}

		/**
			\name Composed trace marks.

			Composed trace marks ar transmitted with the following methods.
			A composed trace mark can be interpreted as a tree of tag-value pairs.

			@{
		*/
		/// indicates the construction of a new mark
		virtual void beginMark(const TraceProducer* sender, const MarkType* m) = 0;
		/// signals the end of construction
		virtual void endMark(const std::string& comment = "") = 0;

		/**
			\brief called to begin of the composed Tag \p t.

			\note A TraceConsumer may rely on correct order of beginTag()/endTag() calls (XML).
		*/
		virtual void beginTag(Tag t) = 0;
		/// end of Tag \p t
		virtual void endTag(Tag t) = 0;

		virtual void addTag(Tag t, bool value) = 0; ///< handle a Tag-bool value pair
		virtual void addTag(Tag t, char value) = 0; ///< handle a Tag-char value pair
		virtual void addTag(Tag t, short value) = 0; ///< handle a Tag-short value pair
		virtual void addTag(Tag t, int value) = 0; ///< handle a Tag-int value pair
		virtual void addTag(Tag t, long value) = 0; ///< handle a Tag-long value pair
		virtual void addTag(Tag t, long long value) = 0; ///< handle a Tag-long value pair
		virtual void addTag(Tag t, float value) = 0; ///< handle a Tag-float value pair
		virtual void addTag(Tag t, double value) = 0; ///< handle a Tag-double value pair
		virtual void addTag(Tag t, const char* value) = 0; ///< handle a Tag-const char* value pair
		virtual void addTag(Tag t, const TraceProducer* value) = 0; ///< handle a Tag-TraceProducer* value pair

		virtual void addTag(Tag t, unsigned char value) = 0; ///< handle a Tag-unsigned char value pair
		virtual void addTag(Tag t, unsigned short value) = 0; ///< handle a Tag-unsigned short value pair
		virtual void addTag(Tag t, unsigned int value) = 0; ///< handle a Tag-unsigned int value pair
		virtual void addTag(Tag t, unsigned long value) = 0; ///< handle a Tag-unsigned long value pair
		//@}

		/// \name Trace control
		//@{
		/**
			\brief This call marks the beginning of a trace

			A TraceConsumer should initialise for a new trace.
		*/
		virtual void startTrace() = 0;

		/**
			\brief This call is send after the traced finished

			A TraceConsumer may clean up.
		*/
		virtual void stopTrace() = 0;

		/**
			\brief This call signals a break in trace

			A TraceConsumer should not produce output after this call is send,
			but await further marks for updating internal data if required.
			(An open composed mark will be closed before this call.)
		*/
		virtual void pauseTrace() = 0;

		/**
			\brief This call signals the end of a break in trace

			A TraceConsumer should update its output if required,
			and report following marks.
		*/
		virtual void continueTrace() = 0;
		//@}
	};

	/** \class Trace

		\ingroup util

		\author Ralf Gerstenberger
		<!-- [\author <author>]* -->

		\brief The Trace receives marks from TraceProducer objects and informs
		registered TraceConsumer about these marks.

		<!-- [\note {notes}]* -->

		\sa TraceConsumer TraceProducer MarkType Tag

		\copydetails Trace.h

		<!-- [\warning {warnings}]* -->

		<!-- [\bug {bug description}]* -->

		<!-- [\test {testcase description}]* -->

		\since 1.0
	*/
	class Trace {
	public:
		Trace(); ///< Construction
		~Trace(); ///< Destruction

		/**
			\name Trace control

			The trace is started with startTrace() and finished with
			stopTrace(). pauseTrace() and continueTrace() should be
			used to	signal temporary deactivation of trace generation.

			\note In order to properly finish the actions of all trace
			consumers, stopTrace() must be called. This should not be done
			automatically in the destructor because this might lead to
			the calling of pure virtual functions on previously deleted
			consumers.

			\note An open composed mark will be closed by pauseTrace() and
			stopTrace().
			@{
		*/
		void startTrace(); ///< start trace
		void stopTrace(); ///< stop trace
		void pauseTrace(); ///< break trace
		void continueTrace(); ///< continue trace
		//@}

		/// \name TraceConsumer management
		//@{
		void addConsumer(TraceConsumer* c); ///< add Consumer
		void removeConsumer(TraceConsumer* c); ///< remove Consumer
		const std::list<TraceConsumer*>& getConsumers() const; ///< get a list of associated trace consumers
		//@}

		/**
			\name Simple trace marks

			Simple trace marks are reported by the following methods.
			@{
		*/
		/**
			\brief send most simple trace mark

			\param sender
				pointer to sender of this mark
			\param m
				type of this mark
			\param comment
				comments
		*/
		void mark(const TraceProducer* sender, MarkType m, const std::string& comment = "");

		// Marks that include information about associations of the producer:
		/// send marks that includes information about associations of the producer
		void mark(const TraceProducer* sender, MarkType m, const TraceProducer* partner, const std::string& comment = "");
		/// send marks that includes information about associations of the producer
		void mark(const TraceProducer* sender, MarkType m, const TraceProducer* firstPartner, const TraceProducer* secondPartner, const std::string& comment = "");

		/**
		 * @brief template method to mark attribute changes of arbitrary type
		 * @tparam AttributeType the value type of the attribute that changed
		 * @param sender pointer to the trace producer sending this mark
		 * @param m the type of the mark sent, i.e. which attribute changed
		 * @param newValue the new attribute value
		 * @param oldValue the previous value of the attribute
		 * @param comment a user-defined message to display for this change
		 */
		template < typename AttributeType >
		void mark( const TraceProducer* sender, MarkType m, AttributeType newValue, AttributeType oldValue, const std::string& comment = "" )
		{
			if (!sendMarks())
				return;

			std::list<TraceConsumer*>::iterator i;
			for (i=cons.begin(); i!=cons.end(); ++i)
				(*i)->mark(sender, &m, newValue, oldValue, comment);
		}
		//@}

		/**	\name Composed trace marks
			Composed trace marks are built with the following methods.
			A composed trace mark can be interpreted as a sequence of
			tag-value pairs.

			A structure can be built with beginTag() and endTag().
			It is recommended but not enforced that for every beginTag(X)
			also endTag(X) is called in	the appropriate order (XML).
			However, a TraceConsumer may rely on the correct order of beginTag()
			and endTag() calls.
			@{
		*/
		/**
			\brief begin a new composed mark

			The construction of a new composed mark starts with beginMark(...).
			If an older mark is still in construction (i.e. endMark(...) was
			not called)	it is closed.
		*/
		void beginMark(const TraceProducer* sender, MarkType m);

		/**
			\brief close a composed mark

			The construction of a composed mark is finished by endMark().
		*/
		void endMark(const std::string& comment = "");

		// With the following functions the composed mark is constructed.
		/**
			\brief called to begin of the composed Tag \p t.

			\note A TraceConsumer may rely on correct order of
			beginTag()/endTag() calls - consider XML, for example.
		*/
		void beginTag(Tag t); ///< begin a tag
		/// end of Tag \p t
		void endTag(Tag t); ///< end tag

		/**
		 * @brief add tag with value
		 * @tparam TagValueType the value type of the given tag
		 * @param t the tag type
		 * @param value the tag value
		 */
		template < typename TagValueType >
		void addTag(Tag t, TagValueType value) {
			if (!sendMarks())
				return;

			if (!openMark) {
				error( "Trace::addTag<>(): called without an open composed mark" );
				return;
			}

			std::list<TraceConsumer*>::iterator i;
			for (i=cons.begin(); i!=cons.end(); ++i)
				(*i)->addTag(t, value);
		}
		//@}

	private:
		std::list<TraceConsumer*> cons; ///< TraceConsumers
		enum STATE{CREATED, RUNNING, PAUSED, FINISHED} state; ///< Trace state
		bool openMark; ///< currently open composed mark

		/// helper method to handle trace pauses
		//bool sendMarks() {return (state==RUNNING || state==PAUSED);}
		bool sendMarks() { return state==RUNNING; }
	};
}

#endif

